(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var global = Package.meteor.global;
var meteorEnv = Package.meteor.meteorEnv;
var ECMAScript = Package.ecmascript.ECMAScript;
var check = Package.check.check;
var Match = Package.check.Match;
var Random = Package.random.Random;
var _ = Package.underscore._;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var meteorInstall = Package.modules.meteorInstall;
var meteorBabelHelpers = Package['babel-runtime'].meteorBabelHelpers;
var Promise = Package.promise.Promise;
var Symbol = Package['ecmascript-runtime-server'].Symbol;
var Map = Package['ecmascript-runtime-server'].Map;
var Set = Package['ecmascript-runtime-server'].Set;

/* Package-scope variables */
var SandstormPermissions, path;

var require = meteorInstall({"node_modules":{"meteor":{"sandstorm-permissions":{"permissions.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-permissions/permissions.js                                                                       //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
SandstormPermissions = {};                                                                                             // 19
                                                                                                                       //
var PermissionSet = function () {                                                                                      //
  // A wrapper around an array of booleans representing a set of permissions like "read" or                            // 22
  // "write". This might represent the permissions held by some user on some grain, or it might                        // 23
  // represent the permissions that one user has chosen to share with another user.                                    // 24
  //                                                                                                                   // 25
  // In our model, permissions are independent. You can have "read" without "write" and you can                        // 26
  // have "write" without "read". Many apps don't actually allow arbitrary permutations, and                           // 27
  // instead define "roles" like "editor" or "viewer", where "editor" implies both read and write                      // 28
  // permission while "viewer" implies only read. Roles, however, are just aliases for sets of                         // 29
  // permissions; all actual computation is done on permissions.                                                       // 30
  function PermissionSet(array) {                                                                                      // 32
    (0, _classCallCheck3.default)(this, PermissionSet);                                                                // 32
                                                                                                                       //
    if (!array) {                                                                                                      // 33
      this.array = [];                                                                                                 // 34
    } else if (array instanceof Array) {                                                                               // 35
      this.array = array.slice(0);                                                                                     // 36
    } else {                                                                                                           // 37
      throw new Error("don't know how to interpret as PermissionSet: " + array);                                       // 38
    }                                                                                                                  // 39
  }                                                                                                                    // 40
                                                                                                                       //
  PermissionSet.fromRoleAssignment = function () {                                                                     //
    function fromRoleAssignment(roleAssignment, viewInfo) {                                                            //
      // Create a PermissionSet based on a ViewSharingLink.RoleAssignment and a UiView.ViewInfo (as                    // 43
      // defined in grain.capnp). ViewInfo defines a mapping from roles to permission sets for a                       // 44
      // particular grain type. RoleAssignment represents the permissions passed from one user to                      // 45
      // another -- usually it specifies a single role, but sometimes also specifies permissions to                    // 46
      // add or remove as well.                                                                                        // 47
      //                                                                                                               // 48
      // A falsy value for `roleAssignment` is considered equivalent to a "none" value, which means                    // 49
      // that no role was explicitly chosen, so the default role should be assigned.                                   // 50
      var result = new PermissionSet([]);                                                                              // 52
                                                                                                                       //
      if (!roleAssignment || "none" in roleAssignment) {                                                               // 54
        // No role explicitly chosen, e.g. because the app did not define any roles at the time                        // 55
        // the sharing took place. Assign the default role, if there is one.                                           // 56
        if (viewInfo.roles) {                                                                                          // 58
          for (var ii = 0; ii < viewInfo.roles.length; ++ii) {                                                         // 59
            var roleDef = viewInfo.roles[ii];                                                                          // 60
                                                                                                                       //
            if (roleDef.default) {                                                                                     // 61
              result = new PermissionSet(roleDef.permissions);                                                         // 62
              break;                                                                                                   // 63
            }                                                                                                          // 64
          }                                                                                                            // 65
        }                                                                                                              // 66
      } else if ("allAccess" in roleAssignment) {                                                                      // 67
        // All permissions are shared, even if there is no explicitly-defined role for this.                           // 68
        var length = 0;                                                                                                // 70
                                                                                                                       //
        if (viewInfo.permissions) {                                                                                    // 71
          length = viewInfo.permissions.length;                                                                        // 72
        }                                                                                                              // 73
                                                                                                                       //
        var array = new Array(length);                                                                                 // 75
                                                                                                                       //
        for (var _ii = 0; _ii < array.length; ++_ii) {                                                                 // 76
          array[_ii] = true;                                                                                           // 77
        }                                                                                                              // 78
                                                                                                                       //
        result = new PermissionSet(array);                                                                             // 80
      } else if ("roleId" in roleAssignment && viewInfo.roles && viewInfo.roles.length > 0) {                          // 81
        // A specific role was chosen.                                                                                 // 82
        var _roleDef = viewInfo.roles[roleAssignment.roleId];                                                          // 84
                                                                                                                       //
        if (_roleDef) {                                                                                                // 85
          result = new PermissionSet(_roleDef.permissions);                                                            // 86
        }                                                                                                              // 87
      }                                                                                                                // 88
                                                                                                                       //
      if (roleAssignment) {                                                                                            // 90
        // Add or remove specific permissions. This is uncommon.                                                       // 91
        result.add(new PermissionSet(roleAssignment.addPermissionSet));                                                // 92
        result.remove(new PermissionSet(roleAssignment.removePermissionSet));                                          // 93
      }                                                                                                                // 94
                                                                                                                       //
      return result;                                                                                                   // 96
    }                                                                                                                  // 97
                                                                                                                       //
    return fromRoleAssignment;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  PermissionSet.prototype.isEmpty = function () {                                                                      //
    function isEmpty() {                                                                                               //
      var result = true;                                                                                               // 100
      this.array.forEach(function (p) {                                                                                // 101
        if (p) {                                                                                                       // 102
          result = false;                                                                                              // 103
        }                                                                                                              // 104
      });                                                                                                              // 105
      return result;                                                                                                   // 107
    }                                                                                                                  // 108
                                                                                                                       //
    return isEmpty;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  PermissionSet.prototype.isSubsetOf = function () {                                                                   //
    function isSubsetOf(other) {                                                                                       //
      check(other, PermissionSet);                                                                                     // 111
                                                                                                                       //
      for (var ii = 0; ii < this.array.length; ++ii) {                                                                 // 112
        var mine = this.array[ii];                                                                                     // 113
        var yours = other.array[ii] || false;                                                                          // 114
                                                                                                                       //
        if (mine && !yours) {                                                                                          // 115
          return false;                                                                                                // 116
        }                                                                                                              // 117
      }                                                                                                                // 118
                                                                                                                       //
      return true;                                                                                                     // 120
    }                                                                                                                  // 121
                                                                                                                       //
    return isSubsetOf;                                                                                                 //
  }(); // Methods for mutating a PermissionSet by combining it with another PermissionSet.                             //
  // These return a boolean indicating whether the operation had any effect.                                           // 124
                                                                                                                       //
                                                                                                                       //
  PermissionSet.prototype.add = function () {                                                                          //
    function add(other) {                                                                                              //
      check(other, PermissionSet);                                                                                     // 127
      var changed = false;                                                                                             // 128
                                                                                                                       //
      for (var ii = 0; ii < other.array.length; ++ii) {                                                                // 129
        var old = !!this.array[ii];                                                                                    // 130
        this.array[ii] = !!this.array[ii] || other.array[ii];                                                          // 131
                                                                                                                       //
        if (old !== this.array[ii]) {                                                                                  // 132
          changed = true;                                                                                              // 133
        }                                                                                                              // 134
      }                                                                                                                // 135
                                                                                                                       //
      return changed;                                                                                                  // 137
    }                                                                                                                  // 138
                                                                                                                       //
    return add;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  PermissionSet.prototype.remove = function () {                                                                       //
    function remove(other) {                                                                                           //
      check(other, PermissionSet);                                                                                     // 141
      var changed = false;                                                                                             // 142
                                                                                                                       //
      for (var ii = 0; ii < other.array.length && ii < this.array.length; ++ii) {                                      // 143
        var old = !!this.array[ii];                                                                                    // 144
        this.array[ii] = !!this.array[ii] && !other.array[ii];                                                         // 145
                                                                                                                       //
        if (old !== this.array[ii]) {                                                                                  // 146
          changed = true;                                                                                              // 147
        }                                                                                                              // 148
      }                                                                                                                // 149
                                                                                                                       //
      return changed;                                                                                                  // 151
    }                                                                                                                  // 152
                                                                                                                       //
    return remove;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  PermissionSet.prototype.intersect = function () {                                                                    //
    function intersect(other) {                                                                                        //
      check(other, PermissionSet);                                                                                     // 155
      var changed = false;                                                                                             // 156
                                                                                                                       //
      for (var ii = 0; ii < this.array.length; ++ii) {                                                                 // 157
        var old = !!this.array[ii];                                                                                    // 158
        this.array[ii] = !!this.array[ii] && other.array[ii];                                                          // 159
                                                                                                                       //
        if (old !== this.array[ii]) {                                                                                  // 160
          changed = true;                                                                                              // 161
        }                                                                                                              // 162
      }                                                                                                                // 163
                                                                                                                       //
      return changed;                                                                                                  // 165
    }                                                                                                                  // 166
                                                                                                                       //
    return intersect;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  return PermissionSet;                                                                                                //
}();                                                                                                                   //
                                                                                                                       //
var RequirementSet = function () {                                                                                     //
  // A conjunction of permissions for users on grains.                                                                 // 170
  //                                                                                                                   // 171
  // This typically represents a set of `MembraneRequirement`s, as defined in `supervisor.capnp`.                      // 172
  // These represent conditions under which some connection formed between grains remains valid.                       // 173
  // When a capability travels from grain to grain, it passes across these connections -- if any                       // 174
  // of the connections becomes invalid (is revoked), then the capability must be revoked as well.                     // 175
  // The word "membrane" comes from the concept of revokable membranes; the capability is passing                      // 176
  // across such membranes as it travels.                                                                              // 177
  //                                                                                                                   // 178
  // For example, a RequirementSet might represent the statement "Alice has read access to Foo, Bob                    // 179
  // has write access to Foo, and Bob has read access to Bar". Specifically, this example situation                    // 180
  // would come about if:                                                                                              // 181
  // - Bob used his read access to Bar to extract a capability from it.                                                // 182
  // - Bob embedded that capability into Foo, using his write access.                                                  // 183
  // - Alice extracted the capability from Foo, using her read access.                                                 // 184
  // If any of these permissions are revoked, then the capability needs to be revoked as well.                         // 185
  function RequirementSet() {                                                                                          // 187
    (0, _classCallCheck3.default)(this, RequirementSet);                                                               // 187
    this.permissionsHeldRequirements = {}; // Map from grain ID to objects of the form { users, tokens }, where `users` is a
    // map from account ID to PermissionSet and `tokens` is a map from token ID to PermissionSet.                      // 190
    // Represents a set of `permissionsHeld` requirements.                                                             // 191
                                                                                                                       //
    this.userIsAdminRequirements = {}; // Map from account ID to boolean. Represents a set of `userIsAdmin` requirements.
                                                                                                                       //
    this.tokenValidRequirements = {}; // Map from token ID to boolean. Represents a set of `tokenValid` requirements.  // 196
  }                                                                                                                    // 198
                                                                                                                       //
  RequirementSet.prototype.isEmpty = function () {                                                                     //
    function isEmpty() {                                                                                               //
      for (var tokenId in meteorBabelHelpers.sanitizeForInObject(this.tokenValidRequirements)) {                       // 201
        return false;                                                                                                  // 202
      }                                                                                                                // 203
                                                                                                                       //
      for (var grainId in meteorBabelHelpers.sanitizeForInObject(this.permissionsHeldRequirements)) {                  // 205
        if (Object.keys(this.permissionsHeldRequirements[grainId]).length > 0) {                                       // 206
          return false;                                                                                                // 207
        }                                                                                                              // 208
      }                                                                                                                // 209
                                                                                                                       //
      return true;                                                                                                     // 211
    }                                                                                                                  // 212
                                                                                                                       //
    return isEmpty;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  RequirementSet.prototype.addRequirements = function () {                                                             //
    function addRequirements(requirements) {                                                                           //
      var _this = this;                                                                                                // 214
                                                                                                                       //
      // Updates this RequirementSet to include the permissions required by `requirements`, which                      // 215
      // is a decoded Cap'n Proto List(MembraneRequirement).                                                           // 216
      if (!requirements) return;                                                                                       // 218
      requirements.forEach(function (requirement) {                                                                    // 219
        if (requirement.permissionsHeld) {                                                                             // 220
          var grainId = requirement.permissionsHeld.grainId;                                                           // 221
          var permissions = new PermissionSet(requirement.permissionsHeld.permissions);                                // 222
                                                                                                                       //
          if (!_this.permissionsHeldRequirements[grainId]) {                                                           // 224
            _this.permissionsHeldRequirements[grainId] = {                                                             // 225
              users: {},                                                                                               // 225
              tokens: {}                                                                                               // 225
            };                                                                                                         // 225
          }                                                                                                            // 226
                                                                                                                       //
          var grainReqs = _this.permissionsHeldRequirements[grainId];                                                  // 228
                                                                                                                       //
          if (requirement.permissionsHeld.accountId) {                                                                 // 230
            var accountId = requirement.permissionsHeld.accountId;                                                     // 231
                                                                                                                       //
            if (!grainReqs.users[accountId]) {                                                                         // 232
              grainReqs.users[accountId] = new PermissionSet([]);                                                      // 233
            }                                                                                                          // 234
                                                                                                                       //
            grainReqs.users[accountId].add(permissions);                                                               // 236
          } else if (requirement.permissionsHeld.tokenId) {                                                            // 237
            var tokenId = requirement.permissionsHeld.tokenId;                                                         // 238
                                                                                                                       //
            if (!grainReqs.tokens[tokenId]) {                                                                          // 239
              grainReqs.tokens[tokenId] = new PermissionSet([]);                                                       // 240
            }                                                                                                          // 241
                                                                                                                       //
            grainReqs.tokens[tokenId].add(permissions);                                                                // 243
          } else {                                                                                                     // 245
            throw new Error("unrecognized permissionsHeld requirement", JSON.stringify(requirement));                  // 246
          }                                                                                                            // 247
        } else if (requirement.userIsAdmin) {                                                                          // 248
          _this.userIsAdminRequirements[requirement.userIsAdmin] = true;                                               // 249
        } else if (requirement.tokenValid) {                                                                           // 250
          _this.tokenValidRequirements[requirement.tokenValid] = true;                                                 // 251
        } else {                                                                                                       // 252
          throw new Error("unsupported requirement: " + JSON.toString(requirement));                                   // 253
        }                                                                                                              // 254
      });                                                                                                              // 255
    }                                                                                                                  // 256
                                                                                                                       //
    return addRequirements;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  RequirementSet.prototype.getGrainIds = function () {                                                                 //
    function getGrainIds() {                                                                                           //
      return Object.keys(this.permissionsHeldRequirements);                                                            // 259
    }                                                                                                                  // 260
                                                                                                                       //
    return getGrainIds;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  RequirementSet.prototype.getTokenIds = function () {                                                                 //
    function getTokenIds() {                                                                                           //
      return Object.keys(this.tokenValidRequirements);                                                                 // 263
    }                                                                                                                  // 264
                                                                                                                       //
    return getTokenIds;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  RequirementSet.prototype.forEach = function () {                                                                     //
    function forEach(func) {                                                                                           //
      for (var tokenId in meteorBabelHelpers.sanitizeForInObject(this.tokenValidRequirements)) {                       // 267
        func({                                                                                                         // 268
          tokenValid: tokenId                                                                                          // 268
        });                                                                                                            // 268
      }                                                                                                                // 269
                                                                                                                       //
      for (var accountId in meteorBabelHelpers.sanitizeForInObject(this.userIsAdminRequirements)) {                    // 271
        func({                                                                                                         // 272
          userIsAdmin: accountId                                                                                       // 272
        });                                                                                                            // 272
      }                                                                                                                // 273
                                                                                                                       //
      for (var grainId in meteorBabelHelpers.sanitizeForInObject(this.permissionsHeldRequirements)) {                  // 275
        for (var _accountId in meteorBabelHelpers.sanitizeForInObject(this.permissionsHeldRequirements[grainId].users)) {
          var permissionSet = this.permissionsHeldRequirements[grainId].users[_accountId];                             // 277
          func({                                                                                                       // 278
            permissionsHeld: {                                                                                         // 278
              grainId: grainId,                                                                                        // 278
              accountId: _accountId,                                                                                   // 278
              permissionSet: permissionSet                                                                             // 278
            }                                                                                                          // 278
          });                                                                                                          // 278
        }                                                                                                              // 279
                                                                                                                       //
        for (var _tokenId in meteorBabelHelpers.sanitizeForInObject(this.permissionsHeldRequirements[grainId].tokens)) {
          var _permissionSet = this.permissionsHeldRequirements[grainId].tokens[_tokenId];                             // 282
          func({                                                                                                       // 283
            permissionsHeld: {                                                                                         // 283
              grainId: grainId,                                                                                        // 283
              tokenId: _tokenId,                                                                                       // 283
              permissionSet: _permissionSet                                                                            // 283
            }                                                                                                          // 283
          });                                                                                                          // 283
        }                                                                                                              // 284
      }                                                                                                                // 285
    }                                                                                                                  // 286
                                                                                                                       //
    return forEach;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  return RequirementSet;                                                                                               //
}(); // A permission ID is either the string "canAccess", corresponding to the implicit "can access                    //
// the grain at all" permission, or a non-negative integer, corresponding to a permission                              // 290
// defined by the app in its manifest.                                                                                 // 291
                                                                                                                       //
                                                                                                                       //
var PermissionId = Match.OneOf("canAccess", Match.Integer);                                                            // 292
                                                                                                                       //
function forEachPermission(permissions, func) {                                                                        // 294
  // Calls `func` on "canAccess" and each permission ID that corresponds to a `true`                                   // 295
  // value in `permissions`.                                                                                           // 296
  check(permissions, [Boolean]);                                                                                       // 297
  func("canAccess");                                                                                                   // 298
                                                                                                                       //
  for (var ii = 0; ii < permissions.length; ++ii) {                                                                    // 299
    if (permissions[ii]) {                                                                                             // 300
      func(ii);                                                                                                        // 301
    }                                                                                                                  // 302
  }                                                                                                                    // 303
} // A vertex is a principal in the sharing graph. A "vertex ID" is either "i:" + an account ID,                       // 304
// or "t:" + a token ID. In some limited contexts, "o:Owner" is also allowed, signifying the                           // 307
// *account* of the grain owner, from which all permissions flow.                                                      // 308
                                                                                                                       //
                                                                                                                       //
var vertexIdOfTokenOwner = function (token) {                                                                          // 310
  // Returns which vertex recieves permissions from this token.                                                        // 311
  var result = "t:" + token._id; // the bearer of the token                                                            // 313
                                                                                                                       //
  if (token.owner && token.owner.user) {                                                                               // 314
    result = "i:" + token.owner.user.accountId; // the user that owns the token                                        // 315
  }                                                                                                                    // 316
                                                                                                                       //
  return result;                                                                                                       // 318
};                                                                                                                     // 319
                                                                                                                       //
var vertexIdOfPermissionsHeld = function (held) {                                                                      // 321
  if (held.accountId) {                                                                                                // 322
    return "i:" + held.accountId;                                                                                      // 323
  } else if (held.tokenId) {                                                                                           // 324
    return "t:" + held.tokenId;                                                                                        // 325
  } else {                                                                                                             // 326
    throw new Error("Unrecognized permissionsHeld: " + JSON.stringify(held));                                          // 327
  }                                                                                                                    // 328
};                                                                                                                     // 329
                                                                                                                       //
var Variable = // Our permissions computation can be framed as a propositional HORNSAT problem; this `Variable`        //
// class represents a variable in that sense. There is a variable for every (grain ID, vertex ID,                      // 333
// permission ID) triple. There is also a variable for every non-UiView ApiToken, to account for                       // 334
// `tokenValid` requirements. In any given computation, we only explicitly construct those                             // 335
// variables that we know might actually be relevant to the result.                                                    // 336
//                                                                                                                     // 337
// The value of a variable represents an answer to the question "does this vertex in the sharing                       // 338
// graph receive this permission at this grain?" We start out by setting all variables to `false`,                     // 339
// and we only set a variable to `true` when an edge in the sharing graph forces us to. If                             // 340
// this forward-chaining eventually forces us to set our end goal variables to `true`, then the                        // 341
// HORNSAT problem is unsatisfiable and we have proved what we wanted. Otherwise, the HORNSAT                          // 342
// problem is satisfiable, i.e. there is a consistent way to set values to variables in which                          // 343
// our goal nodes do *not* receive the permissions we wanted them to receive.                                          // 344
function () {                                                                                                          //
  function Variable() {                                                                                                // 346
    (0, _classCallCheck3.default)(this, Variable);                                                                     // 346
    this.value = false;                                                                                                // 347
    this.directDependents = []; // List of token IDs for outgoing edges that need to be looked at once this variable gets set
    // to `true`.                                                                                                      // 351
                                                                                                                       //
    this.requirementDependents = []; // List of token IDs for tokens that have requirements that get fulfilled once this variable
    // gets set to `true`.                                                                                             // 355
  }                                                                                                                    // 356
                                                                                                                       //
  return Variable;                                                                                                     //
}();                                                                                                                   //
                                                                                                                       //
var ActiveToken = function () {                                                                                        //
  // An "active token" is one that we allow to propagate permissions because we've decided                             // 360
  // that it might be relevant to our current computation. This class tracks which permissions                         // 361
  // the token can carry, which of those permissions we've actually proved to arrive at the source                     // 362
  // end of the token, and how many of the token's requirements are still unmet.                                       // 363
  function ActiveToken(tokenId, requirements, numUnmetRequirements, permissions, grainId, recipientId) {               // 365
    (0, _classCallCheck3.default)(this, ActiveToken);                                                                  // 365
    check(tokenId, String);                                                                                            // 366
    check(numUnmetRequirements, Match.Integer);                                                                        // 367
    check(grainId, String);                                                                                            // 368
    check(recipientId, String);                                                                                        // 369
    check(permissions, [Boolean]); // The permissions that this token is capable of carrying, in addition to the implicit
    // "canAccess" permission.                                                                                         // 373
                                                                                                                       //
    this.tokenId = tokenId;                                                                                            // 375
    this.requirements = requirements;                                                                                  // 376
    this.grainId = grainId;                                                                                            // 377
    this.recipientId = recipientId;                                                                                    // 378
    this.numUnmetRequirements = numUnmetRequirements; // How many of this token's requirements we have not yet proven to be met.
                                                                                                                       //
    this.receivedPermissions = new Array(); // A map from permission ID to boolean. If a permission ID is present in this map,
    // then this token's roleAssignment includes that permission. If the corresponding                                 // 385
    // value is `true`, then we've proven that this token receives this permission and                                 // 386
    // therefore propagates it if the requirements are met.                                                            // 387
                                                                                                                       //
    this.receivedPermissions.canAccess = false;                                                                        // 389
                                                                                                                       //
    for (var ii = 0; ii < permissions.length; ++ii) {                                                                  // 390
      if (permissions[ii]) {                                                                                           // 391
        this.receivedPermissions[ii] = false;                                                                          // 392
      }                                                                                                                // 393
    }                                                                                                                  // 394
  }                                                                                                                    // 395
                                                                                                                       //
  ActiveToken.prototype.requirementsAreMet = function () {                                                             //
    function requirementsAreMet() {                                                                                    //
      // Have we yet proved that all of this token's requirements are met?                                             // 398
      return this.numUnmetRequirements == 0;                                                                           // 399
    }                                                                                                                  // 400
                                                                                                                       //
    return requirementsAreMet;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  ActiveToken.prototype.decrementRequirements = function () {                                                          //
    function decrementRequirements(context) {                                                                          //
      var _this2 = this;                                                                                               // 402
                                                                                                                       //
      // Decrements the number of unmet requirements of this token.                                                    // 403
      check(context, Context);                                                                                         // 404
      this.numUnmetRequirements -= 1;                                                                                  // 405
                                                                                                                       //
      if (this.numUnmetRequirements < 0) {                                                                             // 406
        throw new Meteor.Error(500, "numUnmetRequirements is negative");                                               // 407
      }                                                                                                                // 408
                                                                                                                       //
      if (this.requirementsAreMet()) {                                                                                 // 410
        // This was the last missing requirement for this token, therefore we've triggered the                         // 411
        // outgoing edges from the token corresponding to each known incoming permission.                              // 412
        this.forEachReceivedPermission(function (permissionId) {                                                       // 413
          // We've triggered a new edge! Push it onto the queue.                                                       // 414
          context.setToTrueStack.push({                                                                                // 415
            grainId: _this2.grainId,                                                                                   // 415
            vertexId: _this2.recipientId,                                                                              // 416
            permissionId: permissionId,                                                                                // 417
            responsibleTokenId: _this2.tokenId                                                                         // 418
          });                                                                                                          // 415
        });                                                                                                            // 419
      }                                                                                                                // 420
    }                                                                                                                  // 421
                                                                                                                       //
    return decrementRequirements;                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  ActiveToken.prototype.setReceivesPermission = function () {                                                          //
    function setReceivesPermission(permissionId, context) {                                                            //
      // Records that this token receives a permission.                                                                // 424
      check(permissionId, PermissionId);                                                                               // 425
      check(context, Context);                                                                                         // 426
                                                                                                                       //
      if (permissionId in this.receivedPermissions) {                                                                  // 427
        this.receivedPermissions[permissionId] = true;                                                                 // 428
      }                                                                                                                // 429
                                                                                                                       //
      if (this.numUnmetRequirements == 0) {                                                                            // 431
        context.setToTrueStack.push({                                                                                  // 432
          grainId: this.grainId,                                                                                       // 432
          vertexId: this.recipientId,                                                                                  // 433
          permissionId: permissionId,                                                                                  // 434
          responsibleTokenId: this.tokenId                                                                             // 435
        });                                                                                                            // 432
      } else {                                                                                                         // 436
        // Since we know permissions flow to this token, and the token is active (meaning it                           // 437
        // may be relevant to the overall computation), we now know that it's worth checking                           // 438
        // the token's requirements. Add them to `unmetRequirements`. Note that these requirement                      // 439
        // could actually be ones that we've checked already, but we'll figure that out once                           // 440
        // processUnmetRequirements() runs and discovers it isn't activating any new tokens.                           // 441
        //                                                                                                             // 442
        // TODO(perf): This is an overestimate. It might be worthwhile to keep track of which                          // 443
        //   requirements are already met, so that we don't need to add them here.                                     // 444
        context.unmetRequirements.addRequirements(this.requirements);                                                  // 446
      }                                                                                                                // 447
    }                                                                                                                  // 448
                                                                                                                       //
    return setReceivesPermission;                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  ActiveToken.prototype.forEachReceivedPermission = function () {                                                      //
    function forEachReceivedPermission(func) {                                                                         //
      // Calls `func` on each permission ID that we've proven this token receives.                                     // 451
      if (this.receivedPermissions.canAccess) {                                                                        // 453
        func("canAccess");                                                                                             // 454
      } // Note: Array.forEach() does hit the "canAccess" permission ID.                                               // 455
                                                                                                                       //
                                                                                                                       //
      this.receivedPermissions.forEach(function (received, permissionId) {                                             // 458
        if (received) {                                                                                                // 459
          func(permissionId);                                                                                          // 460
        }                                                                                                              // 461
      });                                                                                                              // 462
    }                                                                                                                  // 463
                                                                                                                       //
    return forEachReceivedPermission;                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  return ActiveToken;                                                                                                  //
}();                                                                                                                   //
                                                                                                                       //
var Context = function () {                                                                                            //
  // An ongoing permissions computation, including cached database state.                                              // 467
  function Context() {                                                                                                 // 469
    (0, _classCallCheck3.default)(this, Context);                                                                      // 469
    this.grains = {}; // Map from grain ID to entry in Grains table.                                                   // 470
                                                                                                                       //
    this.adminUsers = {}; // Set of account IDs for admins.                                                            // 471
                                                                                                                       //
    this.tokensById = {}; // Map from token ID to token.                                                               // 472
                                                                                                                       //
    this.tokensByRecipient = {}; // Map from grain ID and account ID to token array.                                   // 473
                                                                                                                       //
    this.variables = {}; // GrainId -> VertexId -> PermissionId -> Variable                                            // 475
    //                                                                                                                 // 477
    // We have a special fake grain ID "tokenValid", where we store the variables that track                           // 478
    // `tokenValid` requirements. Only the "canAccess" permission makes sense for that fake grain.                     // 479
                                                                                                                       //
    this.userIsAdminVariables = {}; // Account ID -> Variable                                                          // 481
                                                                                                                       //
    this.activeTokens = {}; // TokenId -> ActiveToken                                                                  // 484
                                                                                                                       //
    this.setToTrueStack = []; // Variables enqueued to be set to true.                                                 // 486
    // Array of { grainId: String, vertexId: String, permissionId: PermissionId,                                       // 488
    //            responsibleTokenId: Optional(String) }                                                               // 489
                                                                                                                       //
    this.unmetRequirements = new RequirementSet(); // As we run our forward-chaining algorithm, when we encounter a token with unmet requirements
    // we add those requirements to this set. Then, if we find that our current knowledge base                         // 493
    // is not large enough to prove our goal, we can expand our search by following these                              // 494
    // requirements backwards and activating tokens that might help prove that they are met.                           // 495
    // `unmetRequirements` is allowed to be an overestimate -- in fact, when adding requirements                       // 496
    // to `unmetRequirements` we generally do not check if they had already been proven. The next                      // 497
    // run of processUnmetRequirements(), however, will find that all of the tokens activated by                       // 498
    // the requirements were already active.                                                                           // 499
  }                                                                                                                    // 500
                                                                                                                       //
  Context.prototype.reset = function () {                                                                              //
    function reset() {                                                                                                 //
      // Resets all state except this.grains and this.adminUsers.                                                      // 503
      this.tokensById = {};                                                                                            // 504
      this.tokensByRecipient = {};                                                                                     // 505
      this.unmetRequirements = new RequirementSet();                                                                   // 506
      this.setToTrueStack = [];                                                                                        // 507
      this.variables = {};                                                                                             // 508
      this.activeTokens = {};                                                                                          // 509
    }                                                                                                                  // 510
                                                                                                                       //
    return reset;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.addToken = function () {                                                                           //
    function addToken(token) {                                                                                         //
      // Adds a token to `this.tokensById`. If the token is a UiView, also adds it to                                  // 513
      // `this.tokensByRecipient`. Does not activate the token.                                                        // 514
      var isUiView = token.grainId && !token.objectId;                                                                 // 516
      if (this.tokensById[token._id]) return;                                                                          // 518
      this.tokensById[token._id] = token;                                                                              // 520
                                                                                                                       //
      if (isUiView && token.owner && token.owner.user) {                                                               // 521
        if (!this.tokensByRecipient[token.grainId]) {                                                                  // 522
          this.tokensByRecipient[token.grainId] = {};                                                                  // 523
        }                                                                                                              // 524
                                                                                                                       //
        if (!this.tokensByRecipient[token.grainId][token.owner.user.accountId]) {                                      // 526
          this.tokensByRecipient[token.grainId][token.owner.user.accountId] = [];                                      // 527
        }                                                                                                              // 528
                                                                                                                       //
        this.tokensByRecipient[token.grainId][token.owner.user.accountId].push(token);                                 // 530
      }                                                                                                                // 531
    }                                                                                                                  // 532
                                                                                                                       //
    return addToken;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.addGrains = function () {                                                                          //
    function addGrains(db, grainIds) {                                                                                 //
      var _this3 = this;                                                                                               // 534
                                                                                                                       //
      // Retrieves grains from the database.                                                                           // 535
      check(db, SandstormDb);                                                                                          // 537
      check(grainIds, [String]);                                                                                       // 538
      if (grainIds.length == 0) return; // Nothing to do.                                                              // 539
                                                                                                                       //
      db.collections.grains.find({                                                                                     // 541
        _id: {                                                                                                         // 542
          $in: grainIds                                                                                                // 542
        },                                                                                                             // 542
        trashed: {                                                                                                     // 543
          $exists: false                                                                                               // 543
        },                                                                                                             // 543
        suspended: {                                                                                                   // 544
          $ne: true                                                                                                    // 544
        }                                                                                                              // 544
      }).forEach(function (grain) {                                                                                    // 541
        _this3.grains[grain._id] = grain;                                                                              // 546
      });                                                                                                              // 547
      var query = {                                                                                                    // 549
        grainId: {                                                                                                     // 549
          $in: grainIds                                                                                                // 549
        },                                                                                                             // 549
        revoked: {                                                                                                     // 550
          $ne: true                                                                                                    // 550
        },                                                                                                             // 550
        trashed: {                                                                                                     // 551
          $exists: false                                                                                               // 551
        },                                                                                                             // 551
        suspended: {                                                                                                   // 552
          $ne: true                                                                                                    // 552
        },                                                                                                             // 552
        objectId: {                                                                                                    // 553
          $exists: false                                                                                               // 553
        }                                                                                                              // 553
      };                                                                                                               // 549
      db.collections.apiTokens.find(query).forEach(function (token) {                                                  // 554
        return _this3.addToken(token);                                                                                 // 554
      });                                                                                                              // 554
    }                                                                                                                  // 555
                                                                                                                       //
    return addGrains;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.addTokensFromCursor = function () {                                                                //
    function addTokensFromCursor(cursor) {                                                                             //
      var _this4 = this;                                                                                               // 557
                                                                                                                       //
      cursor.forEach(function (token) {                                                                                // 558
        return _this4.addToken(token);                                                                                 // 558
      });                                                                                                              // 558
    }                                                                                                                  // 559
                                                                                                                       //
    return addTokensFromCursor;                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.activateOwnerEdges = function () {                                                                 //
    function activateOwnerEdges(grainId, edges) {                                                                      //
      var _this5 = this;                                                                                               // 561
                                                                                                                       //
      check(edges, [{                                                                                                  // 562
        accountId: String,                                                                                             // 562
        role: SandstormDb.prototype.roleAssignmentPattern                                                              // 562
      }]);                                                                                                             // 562
      edges.forEach(function (edge) {                                                                                  // 564
        var viewInfo = _this5.grains[grainId].cachedViewInfo || {};                                                    // 565
        var permissions = PermissionSet.fromRoleAssignment(edge.role, viewInfo);                                       // 566
        var vertexId = "i:" + edge.accountId;                                                                          // 567
        forEachPermission(permissions.array, function (permissionId) {                                                 // 568
          _this5.setToTrueStack.push({                                                                                 // 569
            grainId: grainId,                                                                                          // 570
            vertexId: vertexId,                                                                                        // 571
            permissionId: permissionId                                                                                 // 572
          });                                                                                                          // 569
        });                                                                                                            // 574
      });                                                                                                              // 575
      return edges.length > 0;                                                                                         // 577
    }                                                                                                                  // 578
                                                                                                                       //
    return activateOwnerEdges;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.registerInterestInRequirements = function () {                                                     //
    function registerInterestInRequirements(tokenId, requirements) {                                                   //
      var _this6 = this;                                                                                               // 580
                                                                                                                       //
      // Add all requirements.                                                                                         // 581
      var numUnmetRequirements = 0;                                                                                    // 582
                                                                                                                       //
      if (requirements) {                                                                                              // 583
        requirements.forEach(function (requirement) {                                                                  // 584
          if (requirement.permissionsHeld) {                                                                           // 585
            var reqGrainId = requirement.permissionsHeld.grainId;                                                      // 586
            var reqVertexId = vertexIdOfPermissionsHeld(requirement.permissionsHeld);                                  // 587
            var reqPermissions = requirement.permissionsHeld.permissions || [];                                        // 588
            forEachPermission(reqPermissions, function (permissionId) {                                                // 590
              var variable = _this6.getVariable(reqGrainId, reqVertexId, permissionId);                                // 591
                                                                                                                       //
              if (!variable.value) {                                                                                   // 592
                // This requirement hasn't been proven yet. We add it to the variable's list of                        // 593
                // dependents so that we get notified when that variable is proven, and we increment                   // 594
                // numUnmetRequirements to indicate how many notifications we need to get before we                    // 595
                // know all requirements are satisfied.                                                                // 596
                numUnmetRequirements += 1;                                                                             // 597
                variable.requirementDependents.push(tokenId);                                                          // 598
              }                                                                                                        // 599
            });                                                                                                        // 600
          } else if (requirement.userIsAdmin) {                                                                        // 601
            var accountId = requirement.userIsAdmin;                                                                   // 602
                                                                                                                       //
            var variable = _this6.getUserIsAdminVariable(accountId);                                                   // 603
                                                                                                                       //
            if (!variable.value) {                                                                                     // 604
              numUnmetRequirements += 1;                                                                               // 605
              variable.requirementDependents.push(tokenId);                                                            // 606
            }                                                                                                          // 607
          } else if (requirement.tokenValid) {                                                                         // 608
            var reqTokenId = requirement.tokenValid;                                                                   // 609
                                                                                                                       //
            var _variable = _this6.getVariable("tokenValid", "t:" + reqTokenId, "canAccess");                          // 610
                                                                                                                       //
            if (!_variable.value) {                                                                                    // 611
              numUnmetRequirements += 1;                                                                               // 612
                                                                                                                       //
              _variable.requirementDependents.push(tokenId);                                                           // 613
            }                                                                                                          // 614
          } else {                                                                                                     // 615
            throw new Error("unknown kind of requirement: " + JSON.stringify(requirement));                            // 616
          }                                                                                                            // 617
        });                                                                                                            // 618
      }                                                                                                                // 619
                                                                                                                       //
      return numUnmetRequirements;                                                                                     // 621
    }                                                                                                                  // 622
                                                                                                                       //
    return registerInterestInRequirements;                                                                             //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.activateToken = function () {                                                                      //
    function activateToken(tokenId) {                                                                                  //
      var _this7 = this;                                                                                               // 624
                                                                                                                       //
      // Includes a new token (which must already be in `this.tokensById`) in our computation. The                     // 625
      // tricky part here is dealing with our already-accumulated knowledge; we need to compute how                    // 626
      // many of the token's requirements are currently unmet and whether we need to push anything new                 // 627
      // onto `setToTrueStack`.                                                                                        // 628
      check(tokenId, String);                                                                                          // 630
                                                                                                                       //
      if (tokenId in this.activeTokens) {                                                                              // 631
        return false;                                                                                                  // 632
      }                                                                                                                // 633
                                                                                                                       //
      var token = this.tokensById[tokenId];                                                                            // 635
      var grainId = token.grainId;                                                                                     // 636
      var viewInfo = this.grains[grainId].cachedViewInfo || {};                                                        // 637
      var tokenPermissions = PermissionSet.fromRoleAssignment(token.roleAssignment, viewInfo);                         // 638
      var sharerId = token.parentToken ? "t:" + token.parentToken : "i:" + token.accountId;                            // 640
      var recipientId = vertexIdOfTokenOwner(token);                                                                   // 641
      var numUnmetRequirements = this.registerInterestInRequirements(tokenId, token.requirements);                     // 643
      var activeToken = new ActiveToken(tokenId, token.requirements, numUnmetRequirements, tokenPermissions.array, grainId, recipientId); // Add all edges represented by this token (one for each permission).
                                                                                                                       //
      forEachPermission(tokenPermissions.array, function (permissionId) {                                              // 648
        var recipientVariable = _this7.getVariable(grainId, recipientId, permissionId);                                // 649
                                                                                                                       //
        if (!recipientVariable.value) {                                                                                // 650
          var sharerVariable = _this7.getVariable(grainId, sharerId, permissionId);                                    // 651
                                                                                                                       //
          if (!sharerVariable.value) {                                                                                 // 652
            // Not proven yet that the source has this permission, so add to its direct dependents so                  // 653
            // that we get notified if that changes.                                                                   // 654
            sharerVariable.directDependents.push(tokenId);                                                             // 655
          } else {                                                                                                     // 656
            // The source has already been proven.                                                                     // 657
            activeToken.setReceivesPermission(permissionId, _this7);                                                   // 658
          }                                                                                                            // 659
        }                                                                                                              // 660
      });                                                                                                              // 661
      this.activeTokens[tokenId] = activeToken;                                                                        // 663
      return true;                                                                                                     // 665
    }                                                                                                                  // 666
                                                                                                                       //
    return activateToken;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.activateTokenValidToken = function () {                                                            //
    function activateTokenValidToken(tokenId) {                                                                        //
      // Like `activateToken()`, but for a token that appears in a `tokenValid` requirement.                           // 669
      // Such tokens require some special handling to fit into our computation.                                        // 670
      check(tokenId, String);                                                                                          // 672
                                                                                                                       //
      if (tokenId in this.activeTokens) {                                                                              // 673
        return false;                                                                                                  // 674
      }                                                                                                                // 675
                                                                                                                       //
      var token = this.tokensById[tokenId];                                                                            // 677
                                                                                                                       //
      if (!token) {                                                                                                    // 678
        return false;                                                                                                  // 679
      }                                                                                                                // 680
                                                                                                                       //
      var numUnmetRequirements = this.registerInterestInRequirements(tokenId, token.requirements);                     // 682
      var activeToken = new ActiveToken(tokenId, token.requirements, numUnmetRequirements, [], "tokenValid", "t:" + tokenId);
      var recipientId = "t:" + tokenId;                                                                                // 686
      var sharerVariable = token.parentToken && this.getVariable("tokenValid", "t:" + token.parentToken, "canAccess");
      var recipientVariable = this.getVariable("tokenValid", recipientId, "canAccess");                                // 690
                                                                                                                       //
      if (!recipientVariable.value) {                                                                                  // 691
        if (!token.parentToken || sharerVariable && sharerVariable.value) {                                            // 692
          // The source has already been proven.                                                                       // 693
          activeToken.setReceivesPermission("canAccess", this);                                                        // 694
        } else {                                                                                                       // 695
          // Not proven yet that the source has this permission, so add to its direct dependents so                    // 696
          // that we get notified if that changes.                                                                     // 697
          sharerVariable.directDependents.push(tokenId);                                                               // 698
        }                                                                                                              // 699
      }                                                                                                                // 700
                                                                                                                       //
      this.activeTokens[tokenId] = activeToken;                                                                        // 702
      return true;                                                                                                     // 704
    }                                                                                                                  // 705
                                                                                                                       //
    return activateTokenValidToken;                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.getVariable = function () {                                                                        //
    function getVariable(grainId, vertexId, permissionId) {                                                            //
      check(grainId, String);                                                                                          // 708
      check(vertexId, String);                                                                                         // 709
      check(permissionId, PermissionId);                                                                               // 710
                                                                                                                       //
      if (!this.variables[grainId]) {                                                                                  // 712
        this.variables[grainId] = {};                                                                                  // 713
      }                                                                                                                // 714
                                                                                                                       //
      if (!this.variables[grainId][vertexId]) {                                                                        // 716
        this.variables[grainId][vertexId] = new Array();                                                               // 717
      }                                                                                                                // 718
                                                                                                                       //
      if (!this.variables[grainId][vertexId][permissionId]) {                                                          // 720
        this.variables[grainId][vertexId][permissionId] = new Variable();                                              // 721
      }                                                                                                                // 722
                                                                                                                       //
      return this.variables[grainId][vertexId][permissionId];                                                          // 724
    }                                                                                                                  // 725
                                                                                                                       //
    return getVariable;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.getUserIsAdminVariable = function () {                                                             //
    function getUserIsAdminVariable(accountId) {                                                                       //
      check(accountId, String);                                                                                        // 728
                                                                                                                       //
      if (!this.userIsAdminVariables[accountId]) {                                                                     // 730
        this.userIsAdminVariables[accountId] = new Variable();                                                         // 731
      }                                                                                                                // 732
                                                                                                                       //
      return this.userIsAdminVariables[accountId];                                                                     // 734
    }                                                                                                                  // 735
                                                                                                                       //
    return getUserIsAdminVariable;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.getPermissions = function () {                                                                     //
    function getPermissions(grainId, vertexId) {                                                                       //
      // Looks up the permissions that have already been proven for the `vertexId` on `grainId`.                       // 738
      check(grainId, String);                                                                                          // 740
      check(vertexId, String);                                                                                         // 741
                                                                                                                       //
      if (!this.getVariable(grainId, vertexId, "canAccess").value) {                                                   // 743
        return null;                                                                                                   // 744
      } else {                                                                                                         // 745
        var length = this.variables[grainId][vertexId].length;                                                         // 746
        var permissions = new Array(length);                                                                           // 747
                                                                                                                       //
        for (var idx = 0; idx < length; ++idx) {                                                                       // 748
          if (this.variables[grainId][vertexId][idx]) {                                                                // 749
            permissions[idx] = !!this.variables[grainId][vertexId][idx].value;                                         // 750
          } else {                                                                                                     // 751
            permissions[idx] = false;                                                                                  // 752
          }                                                                                                            // 753
        }                                                                                                              // 754
                                                                                                                       //
        return new PermissionSet(permissions);                                                                         // 756
      }                                                                                                                // 757
    }                                                                                                                  // 758
                                                                                                                       //
    return getPermissions;                                                                                             //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.runForwardChaining = function () {                                                                 //
    function runForwardChaining(grainId, vertexId, permissionSet) {                                                    //
      var _this8 = this;                                                                                               // 760
                                                                                                                       //
      // Runs forward-chaining, consuming elements from `this.setToTrueStack` and propagating                          // 761
      // their permissions, until we've proven that `(grainId, vertexId)` has the permissions in                       // 762
      // `permissionSet`, or until we've exhausted `this.setToTrueStack`.                                              // 763
      //                                                                                                               // 764
      // TODO(perf): Exit early if we've already proven that permissionSet is fulfilled.                               // 765
      check(grainId, String);                                                                                          // 767
      check(vertexId, String);                                                                                         // 768
      check(permissionSet, PermissionSet);                                                                             // 769
                                                                                                                       //
      if (permissionSet.array.length > 0) {                                                                            // 771
        // Make sure that the result of this call, retrieved through `this.getPermissions()`,                          // 772
        // will have a full array of permissions, even if they won't all be set to `true`.                             // 773
        this.getVariable(grainId, vertexId, permissionSet.array.length - 1);                                           // 774
      }                                                                                                                // 775
                                                                                                                       //
      var _loop = function () {                                                                                        // 760
        var current = _this8.setToTrueStack.pop();                                                                     // 778
                                                                                                                       //
        var variable = _this8.getVariable(current.grainId, current.vertexId, current.permissionId);                    // 779
                                                                                                                       //
        if (variable.value) {                                                                                          // 780
          return "continue";                                                                                           // 781
        } // For the first time, this vertex is now known to be satisfied.                                             // 782
                                                                                                                       //
                                                                                                                       //
        variable.value = true; // For each edge (token) whose source is this vertex, mark that the input permission is
        // fulfilled.                                                                                                  // 788
                                                                                                                       //
        variable.responsibleTokenId = current.responsibleTokenId;                                                      // 789
        variable.directDependents.forEach(function (tokenId) {                                                         // 790
          // We know this permission must be met now. We also know that the permission was not met                     // 791
          // for this token previously becaues a token has exactly one source vertex for each                          // 792
          // permission, and we're processing that vertex now.                                                         // 793
          _this8.activeTokens[tokenId].setReceivesPermission(current.permissionId, _this8);                            // 795
        }); // For each token that has a requirement on this vertex, mark that one of its requirements                 // 796
        // has ben fulfilled.                                                                                          // 799
                                                                                                                       //
        variable.requirementDependents.forEach(function (tokenId) {                                                    // 800
          _this8.activeTokens[tokenId].decrementRequirements(_this8);                                                  // 801
        });                                                                                                            // 802
      };                                                                                                               // 760
                                                                                                                       //
      while (this.setToTrueStack.length > 0) {                                                                         // 777
        var _ret = _loop();                                                                                            // 777
                                                                                                                       //
        if (_ret === "continue") continue;                                                                             // 777
      }                                                                                                                // 803
                                                                                                                       //
      return this.getPermissions(grainId, vertexId);                                                                   // 805
    }                                                                                                                  // 806
                                                                                                                       //
    return runForwardChaining;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.activateRelevantTokens = function () {                                                             //
    function activateRelevantTokens(grainId, vertexId) {                                                               //
      var _this9 = this;                                                                                               // 808
                                                                                                                       //
      // Returns true if more computation might yield more progress.                                                   // 809
      check(grainId, String);                                                                                          // 811
      check(vertexId, String);                                                                                         // 812
      var result = false;                                                                                              // 814
      var relevant = computeRelevantTokens(this, grainId, vertexId);                                                   // 815
                                                                                                                       //
      if (this.activateOwnerEdges(grainId, relevant.ownerEdges)) {                                                     // 816
        result = true;                                                                                                 // 817
      }                                                                                                                // 818
                                                                                                                       //
      relevant.tokenIds.forEach(function (tokenId) {                                                                   // 820
        if (_this9.activateToken(tokenId)) {                                                                           // 821
          result = true;                                                                                               // 822
        }                                                                                                              // 823
      });                                                                                                              // 824
      return result;                                                                                                   // 826
    }                                                                                                                  // 827
                                                                                                                       //
    return activateRelevantTokens;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.processUnmetRequirements = function () {                                                           //
    function processUnmetRequirements(db) {                                                                            //
      var _this10 = this;                                                                                              // 829
                                                                                                                       //
      // Activate all tokens relevant to the requirements in `unmetRequirements`. May do a database                    // 830
      // query to look up new tokens.                                                                                  // 831
      //                                                                                                               // 832
      // Returns true if more computation might yield more progress, i.e., at least one token was                      // 833
      // newly-activated.                                                                                              // 834
      var grainIds = this.unmetRequirements.getGrainIds().filter(function (grainId) {                                  // 836
        return !(grainId in _this10.grains);                                                                           // 837
      });                                                                                                              // 837
                                                                                                                       //
      if (db) {                                                                                                        // 839
        this.addGrains(db, grainIds);                                                                                  // 840
      }                                                                                                                // 841
                                                                                                                       //
      var result = false; // As we process our requirements, we might find new unmet requirments.                      // 843
      // We defer processing those until the next time around.                                                         // 846
                                                                                                                       //
      var oldUnmetRequirements = this.unmetRequirements;                                                               // 847
      this.unmetRequirements = new RequirementSet();                                                                   // 848
      oldUnmetRequirements.forEach(function (req) {                                                                    // 850
        if (req.permissionsHeld) {                                                                                     // 851
          var next = req.permissionsHeld;                                                                              // 852
          var nextVertexId = vertexIdOfPermissionsHeld(next); // Activate all tokens relevant to this requirement.     // 853
          // TODO(perf): We're actually overestimating, because we're activating tokens relevant to                    // 855
          //   *all* permissions rather than just the permissions specified by the requirement. E.g.                   // 856
          //   if the requirement is that Bob has write access, then tokens which only transmit read                   // 857
          //   access are not relevant, but we'll end up activating them anyway.                                       // 858
                                                                                                                       //
          if (_this10.activateRelevantTokens(next.grainId, nextVertexId)) {                                            // 859
            result = true;                                                                                             // 860
          }                                                                                                            // 861
        } else if (req.tokenValid) {                                                                                   // 862
          // Active the token and all of its transitive parents.                                                       // 863
          var currentTokenId = req.tokenValid;                                                                         // 864
                                                                                                                       //
          while (true) {                                                                                               // 865
            var currentToken = _this10.tokensById[currentTokenId];                                                     // 866
                                                                                                                       //
            if (!currentToken && db) {                                                                                 // 867
              currentToken = db.collections.apiTokens.findOne({                                                        // 868
                _id: currentTokenId,                                                                                   // 869
                revoked: {                                                                                             // 870
                  $ne: true                                                                                            // 870
                },                                                                                                     // 870
                suspended: {                                                                                           // 871
                  $ne: true                                                                                            // 871
                }                                                                                                      // 871
              });                                                                                                      // 868
              _this10.tokensById[currentTokenId] = currentToken;                                                       // 873
            }                                                                                                          // 874
                                                                                                                       //
            if (!currentToken) {                                                                                       // 876
              break;                                                                                                   // 877
            }                                                                                                          // 878
                                                                                                                       //
            if (_this10.activateTokenValidToken(currentTokenId)) {                                                     // 880
              result = true;                                                                                           // 881
            }                                                                                                          // 882
                                                                                                                       //
            if (currentToken.parentToken) {                                                                            // 884
              currentTokenId = currentToken.parentToken;                                                               // 885
            } else {                                                                                                   // 886
              break;                                                                                                   // 887
            }                                                                                                          // 888
          }                                                                                                            // 889
        } else if (req.userIsAdmin) {                                                                                  // 890
          var accountId = req.userIsAdmin;                                                                             // 891
                                                                                                                       //
          if (!(accountId in _this10.adminUsers) && db) {                                                              // 892
            Meteor.users.find({                                                                                        // 893
              isAdmin: true                                                                                            // 893
            }).forEach(function (user) {                                                                               // 893
              _this10.adminUsers[user._id] = true;                                                                     // 894
            });                                                                                                        // 895
          }                                                                                                            // 896
                                                                                                                       //
          if (_this10.adminUsers[accountId]) {                                                                         // 898
            var variable = _this10.getUserIsAdminVariable(accountId);                                                  // 899
                                                                                                                       //
            variable.value = true; // Might have some requirement dependents. Cannot have any direct depependents.     // 900
                                                                                                                       //
            variable.requirementDependents.forEach(function (tokenId) {                                                // 903
              result = true;                                                                                           // 904
                                                                                                                       //
              _this10.activeTokens[tokenId].decrementRequirements(_this10);                                            // 905
            });                                                                                                        // 906
          }                                                                                                            // 907
        } else {                                                                                                       // 908
          throw new Error("unknown kind of requirement: " + JSON.stringify(req));                                      // 909
        }                                                                                                              // 910
      });                                                                                                              // 911
      return result;                                                                                                   // 913
    }                                                                                                                  // 914
                                                                                                                       //
    return processUnmetRequirements;                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.tryToProve = function () {                                                                         //
    function tryToProve(grainId, vertexId, permissionSet, db) {                                                        //
      // Tries to prove that `vertexId` has the given permissions on the given grain. Returns a                        // 917
      // `PermissionSet` representing the permissions proven, or null if it has not been proved                        // 918
      // yet that the vertex even has access to the grain.                                                             // 919
      check(grainId, String);                                                                                          // 921
      check(vertexId, String);                                                                                         // 922
      check(permissionSet, PermissionSet);                                                                             // 923
      check(db, Match.OneOf(undefined, SandstormDb)); // If `db` is not provided, then this function will make no database queries.
                                                                                                                       //
      if (db) {                                                                                                        // 927
        this.addGrains(db, [grainId]);                                                                                 // 928
      }                                                                                                                // 929
                                                                                                                       //
      this.activateRelevantTokens(grainId, vertexId);                                                                  // 931
                                                                                                                       //
      while (true) {                                                                                                   // 932
        var result = this.runForwardChaining(grainId, vertexId, permissionSet);                                        // 933
                                                                                                                       //
        if (result && permissionSet.isSubsetOf(result)) {                                                              // 934
          return result;                                                                                               // 935
        }                                                                                                              // 936
                                                                                                                       //
        if (!this.processUnmetRequirements(db)) {                                                                      // 938
          return result;                                                                                               // 939
        }                                                                                                              // 940
      }                                                                                                                // 941
    }                                                                                                                  // 942
                                                                                                                       //
    return tryToProve;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  Context.prototype.getResponsibleTokens = function () {                                                               //
    function getResponsibleTokens(grainId, vertexId) {                                                                 //
      // For the permissions that we've already proven must be held by `vertexId`, transitively finds                  // 945
      // the tokens that we have used in that proof, including tokens responsible for fulfilling                       // 946
      // membrane requirements.                                                                                        // 947
      //                                                                                                               // 948
      // Whenever we prove a fact, we keep track of the immediately responsible token for that fact,                   // 949
      // This function works by walking backwards in the sharing graph, following this trail of                        // 950
      // "responsible tokens".                                                                                         // 951
      //                                                                                                               // 952
      // Returns an object with two fields:                                                                            // 953
      //    tokenIds: list of IDs of the responsible tokens.                                                           // 954
      //    grainIds: list of IDs of all relevant grains.                                                              // 955
      check(grainId, String);                                                                                          // 957
      check(vertexId, String);                                                                                         // 958
      var stack = []; // [{ grainId: String, vertexId: String, permissionId: PermissionId }]                           // 960
                                                                                                                       //
      var visited = new Array(); // grainId -> vertexId -> permissionId -> bool;                                       // 961
                                                                                                                       //
      function pushVertex(grainId, vertexId, permissionId) {                                                           // 963
        if (!visited[grainId]) {                                                                                       // 964
          visited[grainId] = {};                                                                                       // 965
        }                                                                                                              // 966
                                                                                                                       //
        if (!visited[grainId][vertexId]) {                                                                             // 968
          visited[grainId][vertexId] = {};                                                                             // 969
        }                                                                                                              // 970
                                                                                                                       //
        var vertex = visited[grainId][vertexId];                                                                       // 972
                                                                                                                       //
        if (!vertex[permissionId]) {                                                                                   // 973
          vertex[permissionId] = true;                                                                                 // 974
          stack.push({                                                                                                 // 975
            grainId: grainId,                                                                                          // 975
            vertexId: vertexId,                                                                                        // 975
            permissionId: permissionId                                                                                 // 975
          });                                                                                                          // 975
        }                                                                                                              // 976
      }                                                                                                                // 977
                                                                                                                       //
      var neededTokens = {}; // TokenId -> bool                                                                        // 979
                                                                                                                       //
      forEachPermission(this.getPermissions(grainId, vertexId).array, function (permissionId) {                        // 981
        pushVertex(grainId, vertexId, permissionId);                                                                   // 982
      });                                                                                                              // 983
                                                                                                                       //
      while (stack.length > 0) {                                                                                       // 985
        var _current = stack.pop();                                                                                    // 986
                                                                                                                       //
        var variable = this.getVariable(_current.grainId, _current.vertexId, _current.permissionId);                   // 987
        var tokenId = variable.responsibleTokenId;                                                                     // 988
                                                                                                                       //
        if (tokenId) {                                                                                                 // 989
          var token = this.tokensById[tokenId];                                                                        // 990
                                                                                                                       //
          if (token.grainId && !token.objectId) {                                                                      // 992
            var sharerId = token.parentToken ? "t:" + token.parentToken : "i:" + token.accountId;                      // 993
            pushVertex(token.grainId, sharerId, _current.permissionId);                                                // 994
          } else if (token.parentToken) {                                                                              // 995
            pushVertex("tokenValid", "t:" + token.parentToken, "canAccess");                                           // 996
          }                                                                                                            // 997
                                                                                                                       //
          if (!neededTokens[tokenId]) {                                                                                // 999
            neededTokens[tokenId] = true;                                                                              // 1000
                                                                                                                       //
            if (token.requirements) {                                                                                  // 1001
              token.requirements.forEach(function (requirement) {                                                      // 1002
                if (requirement.permissionsHeld) {                                                                     // 1003
                  var held = requirement.permissionsHeld;                                                              // 1004
                  var reqVertexId = held.accountId ? "i:" + held.accountId : "t:" + held.tokenId;                      // 1005
                  forEachPermission(requirement.permissionsHeld.permissions, function (permissionId) {                 // 1006
                    pushVertex(requirement.permissionsHeld.grainId, reqVertexId, permissionId);                        // 1007
                  });                                                                                                  // 1008
                } else if (requirement.tokenValid) {                                                                   // 1009
                  pushVertex("tokenValid", "t:" + requirement.tokenValid, "canAccess");                                // 1010
                }                                                                                                      // 1011
              });                                                                                                      // 1012
            }                                                                                                          // 1013
          }                                                                                                            // 1014
        }                                                                                                              // 1015
      }                                                                                                                // 1016
                                                                                                                       //
      return {                                                                                                         // 1018
        tokenIds: Object.keys(neededTokens),                                                                           // 1018
        grainIds: Object.keys(visited)                                                                                 // 1018
      };                                                                                                               // 1018
    }                                                                                                                  // 1019
                                                                                                                       //
    return getResponsibleTokens;                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  return Context;                                                                                                      //
}();                                                                                                                   //
                                                                                                                       //
function computeRelevantTokens(context, grainId, vertexId) {                                                           // 1022
  // Finds all tokens in `context` that could possibly carry permissions of the grain `grainId` to                     // 1023
  // the vertex `vertexId` -- that is, all tokens that are contained in a path starting at the                         // 1024
  // grain owner and ending at `vertexId`. Ignores any requirements that those tokens might be                         // 1025
  // conditional upon.                                                                                                 // 1026
  //                                                                                                                   // 1027
  // Returns an object with two fields:                                                                                // 1028
  //    tokenIds: list of relevant token IDs.                                                                          // 1029
  //    ownerEdges: objects of the form { accountId: String, role: RoleAssignment }, representing                      // 1030
  //                initial pseudo-edges in the graph. `accountId` is typically                                        // 1031
  //                the grain's owning user, but for the case of a legacy public grain it could                        // 1032
  //                be any user.                                                                                       // 1033
  //                                                                                                                   // 1034
  // Works by traversing the sharing graph twice: first backwards starting from `vertexId`, then                       // 1035
  // forwards starting from the grain owner using only those tokens touched in the first step.                         // 1036
  //                                                                                                                   // 1037
  // `context` contains all the information from the database which is available for now. This call                    // 1038
  // will not make any new database lookups; edges not listed in `context` will not be considered                      // 1039
  // (as if they'd been revoked).                                                                                      // 1040
  check(context, Context);                                                                                             // 1042
  check(grainId, String);                                                                                              // 1043
  check(vertexId, String);                                                                                             // 1044
  var grain = context.grains[grainId];                                                                                 // 1046
  if (!grain) return {                                                                                                 // 1047
    tokenIds: [],                                                                                                      // 1047
    ownerEdges: []                                                                                                     // 1047
  };                                                                                                                   // 1047
  var viewInfo = grain.cachedViewInfo || {};                                                                           // 1048
  var vertexStack = []; // Vertex IDs that we need to explore.                                                         // 1050
                                                                                                                       //
  var visitedVertexIds = {}; // Set of vertex IDs that we have already enqueued to get explored.                       // 1051
                                                                                                                       //
  visitedVertexIds[vertexId] = true;                                                                                   // 1053
  vertexStack.push(vertexId);                                                                                          // 1054
  var visitedTokensBySharerId = {};                                                                                    // 1056
  var ownerEdges = []; // Repeatedly pop a vertex from the stack, find all its incoming edges (i.e. all other vertexes
  // that share permissions to this vertex), and push those vertexes onto the stack.                                   // 1060
                                                                                                                       //
  while (vertexStack.length > 0) {                                                                                     // 1061
    // List of edges in the sharing graph ending at this vertex. Each is an object with the fields:                    // 1065
    //     sharerId: The vertex ID of the edge's source.                                                               // 1066
    //     token: the token object backing this edge, if there is one.                                                 // 1067
    var tokenToEdge = function (token) {                                                                               // 1061
      // Convert an ApiToken into an edge.                                                                             // 1070
      return {                                                                                                         // 1071
        token: token,                                                                                                  // 1072
        sharerId: token.parentToken ? "t:" + token.parentToken : "i:" + token.accountId                                // 1073
      };                                                                                                               // 1071
    };                                                                                                                 // 1075
                                                                                                                       //
    var _vertexId = vertexStack.pop();                                                                                 // 1062
                                                                                                                       //
    var incomingEdges = [];                                                                                            // 1064
                                                                                                                       //
    if (_vertexId.slice(0, 2) === "o:") {                                                                              // 1077
      // Owner. We don't need to do anything.                                                                          // 1078
      incomingEdges = [];                                                                                              // 1079
    } else if (_vertexId.slice(0, 2) === "t:") {                                                                       // 1080
      // A webkey token. Extract it from the context (or ignore if it isn't present).                                  // 1081
      var token = context.tokensById[_vertexId.slice(2)];                                                              // 1082
                                                                                                                       //
      if (token) {                                                                                                     // 1083
        incomingEdges = [tokenToEdge(token)];                                                                          // 1084
      }                                                                                                                // 1085
    } else if (_vertexId.slice(0, 2) === "i:") {                                                                       // 1086
      // A user.                                                                                                       // 1087
      var accountId = _vertexId.slice(2);                                                                              // 1088
                                                                                                                       //
      if (accountId === grain.userId) {                                                                                // 1089
        // This is the owner.                                                                                          // 1090
        incomingEdges = [{                                                                                             // 1091
          sharerId: "o:Owner"                                                                                          // 1091
        }];                                                                                                            // 1091
        ownerEdges.push({                                                                                              // 1092
          accountId: accountId,                                                                                        // 1092
          role: {                                                                                                      // 1092
            allAccess: null                                                                                            // 1092
          }                                                                                                            // 1092
        });                                                                                                            // 1092
      } else if (!grain.private) {                                                                                     // 1093
        // This is a legacy "public" grain, meaning that any user who knows the grain ID receives                      // 1094
        // the grain's default role. If the user doesn't know the grain ID then they are unable                        // 1095
        // to express a request to open the grain in the first place and we'll never get to the                        // 1096
        // point of this permissions computation, so for this purpose we can assume all users                          // 1097
        // have the default role. (Similarly, a user who doesn't know the grain ID couldn't                            // 1098
        // possibly be the subject of any MembraneRequirements against the grain because they                          // 1099
        // have never interacted with the grain and so couldn't have caused such                                       // 1100
        // MembraneRequiments to come about. Note that this is kind of shaky non-local reasoning,                      // 1101
        // but literally no such legacy grain has been created since early 2015 and none will ever                     // 1102
        // be created again, so it's not a huge deal.)                                                                 // 1103
        incomingEdges = [{                                                                                             // 1104
          sharerId: "o:Owner"                                                                                          // 1104
        }];                                                                                                            // 1104
        ownerEdges.push({                                                                                              // 1105
          accountId: accountId,                                                                                        // 1105
          role: {                                                                                                      // 1105
            none: null                                                                                                 // 1105
          }                                                                                                            // 1105
        });                                                                                                            // 1105
      } else {                                                                                                         // 1106
        // Not a special case. Gather all tokens where this user is the recipient.                                     // 1107
        incomingEdges = ((context.tokensByRecipient[grainId] || {})[_vertexId.slice(2)] || []).map(tokenToEdge);       // 1108
      }                                                                                                                // 1110
    } else {                                                                                                           // 1111
      throw new Meteor.Error(500, "Unrecognized vertex ID: " + _vertexId);                                             // 1112
    } // For each edge incoming to this vertex, backpropagate this vertex's PermissionFlow to the                      // 1113
    // source vertex, joining it with the edge's constraints.                                                          // 1116
                                                                                                                       //
                                                                                                                       //
    incomingEdges.forEach(function (edge) {                                                                            // 1117
      var sharerId = edge.sharerId;                                                                                    // 1118
                                                                                                                       //
      if (edge.token) {                                                                                                // 1119
        if (!visitedTokensBySharerId[sharerId]) {                                                                      // 1120
          visitedTokensBySharerId[sharerId] = {};                                                                      // 1121
        }                                                                                                              // 1122
                                                                                                                       //
        visitedTokensBySharerId[sharerId][edge.token._id] = edge.token;                                                // 1124
      }                                                                                                                // 1125
                                                                                                                       //
      if (!visitedVertexIds[sharerId]) {                                                                               // 1127
        // Never saw this vertex before.                                                                               // 1128
        visitedVertexIds[sharerId] = true;                                                                             // 1129
        vertexStack.push(sharerId);                                                                                    // 1130
      }                                                                                                                // 1131
    });                                                                                                                // 1133
  } // Now walk forward from the owner.                                                                                // 1134
                                                                                                                       //
                                                                                                                       //
  var relevantTokens = {};                                                                                             // 1137
  var visitedSharers = {};                                                                                             // 1138
  var sharerStack = [];                                                                                                // 1140
  sharerStack.push("i:" + grain.userId);                                                                               // 1141
                                                                                                                       //
  while (sharerStack.length > 0) {                                                                                     // 1142
    var sharerId = sharerStack.pop();                                                                                  // 1143
                                                                                                                       //
    for (var tokenId in meteorBabelHelpers.sanitizeForInObject(visitedTokensBySharerId[sharerId])) {                   // 1144
      relevantTokens[tokenId] = true;                                                                                  // 1145
      var _token = visitedTokensBySharerId[sharerId][tokenId];                                                         // 1146
      var recipientId = vertexIdOfTokenOwner(_token);                                                                  // 1147
                                                                                                                       //
      if (!visitedSharers[recipientId]) {                                                                              // 1149
        visitedSharers[recipientId] = true;                                                                            // 1150
        sharerStack.push(recipientId);                                                                                 // 1151
      }                                                                                                                // 1152
    }                                                                                                                  // 1153
  }                                                                                                                    // 1154
                                                                                                                       //
  return {                                                                                                             // 1156
    tokenIds: Object.keys(relevantTokens),                                                                             // 1157
    ownerEdges: ownerEdges                                                                                             // 1158
  };                                                                                                                   // 1156
}                                                                                                                      // 1160
                                                                                                                       //
var vertexPattern = Match.OneOf({                                                                                      // 1162
  token: Match.ObjectIncluding({                                                                                       // 1163
    _id: String,                                                                                                       // 1163
    grainId: String                                                                                                    // 1163
  })                                                                                                                   // 1163
}, {                                                                                                                   // 1163
  grain: Match.ObjectIncluding({                                                                                       // 1165
    _id: String,                                                                                                       // 1166
    accountId: Match.OneOf(String, null, undefined)                                                                    // 1167
  })                                                                                                                   // 1165
}); // A vertex in the sharing graph is a principal, e.g. a user or a token. Complicating                              // 1164
// matters, we may have to traverse sharing graphs for multiple grains in the same computation. A                      // 1172
// token is specific to one grain, but a user of course can have access to multiple grains, so in                      // 1173
// the case of a user we represent the vertex as a (user, grain) pair.                                                 // 1174
//                                                                                                                     // 1175
// TODO(cleanup): Perhaps `grain` should be renamed to `user`? In the common case                                      // 1176
//   where only a single grain's shares need to be considered, it feels weird to think of the                          // 1177
//   grain ID as being the primary distinguishing feature of the vertex.                                               // 1178
                                                                                                                       //
SandstormPermissions.mayOpenGrain = function (db, vertex) {                                                            // 1180
  // Determines whether the vertex is allowed to open the grain. May make multiple database                            // 1181
  // queries.                                                                                                          // 1182
  check(vertex, vertexPattern);                                                                                        // 1184
  var grainId = vertex.token ? vertex.token.grainId : vertex.grain._id;                                                // 1185
  var vertexId = vertex.token ? "t:" + vertex.token._id : "i:" + vertex.grain.accountId;                               // 1186
  var context = new Context();                                                                                         // 1187
  var emptyPermissions = new PermissionSet([]);                                                                        // 1188
  return !!context.tryToProve(grainId, vertexId, emptyPermissions, db);                                                // 1189
};                                                                                                                     // 1190
                                                                                                                       //
var CompoundObserveHandle = function () {                                                                              //
  function CompoundObserveHandle() {                                                                                   // 1193
    (0, _classCallCheck3.default)(this, CompoundObserveHandle);                                                        // 1193
    this._handles = [];                                                                                                // 1194
  }                                                                                                                    // 1195
                                                                                                                       //
  CompoundObserveHandle.prototype.push = function () {                                                                 //
    function push(handle) {                                                                                            //
      this._handles.push(handle);                                                                                      // 1198
    }                                                                                                                  // 1199
                                                                                                                       //
    return push;                                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  CompoundObserveHandle.prototype.stop = function () {                                                                 //
    function stop() {                                                                                                  //
      this._handles.forEach(function (handle) {                                                                        // 1202
        return handle.stop();                                                                                          // 1202
      });                                                                                                              // 1202
    }                                                                                                                  // 1203
                                                                                                                       //
    return stop;                                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  return CompoundObserveHandle;                                                                                        //
}();                                                                                                                   //
                                                                                                                       //
SandstormPermissions.grainPermissions = function (db, vertex, viewInfo, onInvalidated) {                               // 1206
  // Computes the set of permissions received by `vertex`. Returns an object with a                                    // 1207
  // `permissions` field containing the computed permissions. If the field is null then                                // 1208
  // the `vertex` does not even have the base "allowed to access at all" permission.                                   // 1209
  //                                                                                                                   // 1210
  // `onInvalidated` is an optional callback. If provided, it will be called when the result                           // 1211
  // has been invalidated. If `onValidated` is provided, the result of `grainPermissions` will                         // 1212
  // have a `observeHandle` field, containing an object with a `stop()` method that must be                            // 1213
  // called once the computation becomes so longer relevant.                                                           // 1214
  check(db, SandstormDb);                                                                                              // 1216
  check(vertex, vertexPattern);                                                                                        // 1217
  var grainId = vertex.token ? vertex.token.grainId : vertex.grain._id;                                                // 1218
  var vertexId = vertex.token ? "t:" + vertex.token._id : "i:" + vertex.grain.accountId;                               // 1219
  var resultPermissions = void 0;                                                                                      // 1221
  var observeHandle = void 0;                                                                                          // 1222
  var onInvalidatedActive = false;                                                                                     // 1223
  var startTime = new Date();                                                                                          // 1225
                                                                                                                       //
  function measureElapsedTime(result) {                                                                                // 1226
    var elapsedMilliseconds = new Date() - startTime;                                                                  // 1227
                                                                                                                       //
    if (elapsedMilliseconds > 200) {                                                                                   // 1228
      console.log("Warning: SandstormPermissions.grainPermissions() took " + elapsedMilliseconds + " milliseconds to complete for the vertex " + JSON.stringify(vertex));
    }                                                                                                                  // 1231
  } // Our computation proceeds in two phases. In the first, we determine which permissions the vertex                 // 1232
  // appears to have and we compute set of tokens which appears to be sufficent to prove those                         // 1235
  // permissions. However, concurrent database modifications may render the computation invalid, so                    // 1236
  // in the second phases we verify that our proof is still valid and arrange for onInvalidated                        // 1237
  // to be called when necessary. This is an optimisistic approach to concurrency. If the                              // 1238
  // verification phase fails, we try again before giving up entirely.                                                 // 1239
                                                                                                                       //
                                                                                                                       //
  var _loop2 = function (attemptCount) {                                                                               // 1206
    if (observeHandle) {                                                                                               // 1241
      observeHandle.stop();                                                                                            // 1242
      observeHandle = null;                                                                                            // 1243
    }                                                                                                                  // 1244
                                                                                                                       //
    var context = new Context();                                                                                       // 1246
    var allPermissions = PermissionSet.fromRoleAssignment({                                                            // 1247
      allAccess: null                                                                                                  // 1247
    }, viewInfo);                                                                                                      // 1247
    var firstPhasePermissions = context.tryToProve(grainId, vertexId, allPermissions, db);                             // 1248
    if (!firstPhasePermissions) return "break"; // No permissions; give up now.                                        // 1250
                                                                                                                       //
    var needed = context.getResponsibleTokens(grainId, vertexId);                                                      // 1252
    var neededTokens = needed.tokenIds;                                                                                // 1253
    var neededGrains = needed.grainIds; // Phase 2: Now let's verify those permissions.                                // 1254
                                                                                                                       //
    context.reset();                                                                                                   // 1258
    var invalidated = false;                                                                                           // 1260
                                                                                                                       //
    function guardedOnInvalidated() {                                                                                  // 1261
      var shouldCall = onInvalidatedActive && !invalidated;                                                            // 1262
      invalidated = true;                                                                                              // 1263
                                                                                                                       //
      if (shouldCall) {                                                                                                // 1264
        onInvalidated();                                                                                               // 1265
      }                                                                                                                // 1266
    }                                                                                                                  // 1267
                                                                                                                       //
    var tokenCursor = db.collections.apiTokens.find({                                                                  // 1269
      _id: {                                                                                                           // 1270
        $in: neededTokens                                                                                              // 1270
      },                                                                                                               // 1270
      revoked: {                                                                                                       // 1271
        $ne: true                                                                                                      // 1271
      },                                                                                                               // 1271
      suspended: {                                                                                                     // 1272
        $ne: true                                                                                                      // 1272
      },                                                                                                               // 1272
      objectId: {                                                                                                      // 1273
        $exists: false                                                                                                 // 1273
      }                                                                                                                // 1273
    });                                                                                                                // 1269
                                                                                                                       //
    if (onInvalidated) {                                                                                               // 1276
      observeHandle = new CompoundObserveHandle();                                                                     // 1277
      observeHandle.push(tokenCursor.observe({                                                                         // 1278
        changed: function (newApiToken, oldApiToken) {                                                                 // 1279
          if (newApiToken.trashed || !_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.suspended, oldApiToken.suspended) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
            observeHandle.stop();                                                                                      // 1284
            guardedOnInvalidated();                                                                                    // 1285
          }                                                                                                            // 1286
        },                                                                                                             // 1287
        removed: function (oldApiToken) {                                                                              // 1289
          observeHandle.stop();                                                                                        // 1290
          guardedOnInvalidated();                                                                                      // 1291
        }                                                                                                              // 1292
      }));                                                                                                             // 1278
      var grainCursor = db.collections.grains.find({                                                                   // 1295
        _id: {                                                                                                         // 1295
          $in: neededGrains                                                                                            // 1295
        }                                                                                                              // 1295
      });                                                                                                              // 1295
      observeHandle.push(grainCursor.observe({                                                                         // 1296
        changed: function (newGrain, oldGrain) {                                                                       // 1297
          if (newGrain.trashed || newGrain.suspended || !oldGrain.private && newGrain.private) {                       // 1298
            observeHandle.stop();                                                                                      // 1301
            guardedOnInvalidated();                                                                                    // 1302
          }                                                                                                            // 1303
        },                                                                                                             // 1304
        removed: function (oldGrain) {                                                                                 // 1306
          observeHandle.stop();                                                                                        // 1307
          guardedOnInvalidated();                                                                                      // 1308
        }                                                                                                              // 1309
      }));                                                                                                             // 1296
    }                                                                                                                  // 1311
                                                                                                                       //
    context.addTokensFromCursor(tokenCursor); // TODO(someday): Also account for accounts losing admin privileges. Currently we do not call
    //   `onInvalided()` on such events. We would need to set up more cursor observers.                                // 1316
                                                                                                                       //
    resultPermissions = context.tryToProve(grainId, vertexId, firstPhasePermissions);                                  // 1318
                                                                                                                       //
    if (resultPermissions && firstPhasePermissions.isSubsetOf(resultPermissions)) {                                    // 1320
      // We've confirmed the permissions that we found the in the first phase. Done!                                   // 1321
      return "break";                                                                                                  // 1322
    }                                                                                                                  // 1323
  };                                                                                                                   // 1206
                                                                                                                       //
  for (var attemptCount = 0; attemptCount < 3; ++attemptCount) {                                                       // 1240
    var _ret2 = _loop2(attemptCount);                                                                                  // 1240
                                                                                                                       //
    if (_ret2 === "break") break;                                                                                      // 1240
  } // for (let attemptCount ...) {                                                                                    // 1324
                                                                                                                       //
                                                                                                                       //
  onInvalidatedActive = true;                                                                                          // 1326
  var result = {};                                                                                                     // 1327
  result.permissions = resultPermissions && resultPermissions.array || null;                                           // 1328
  result.observeHandle = observeHandle || new CompoundObserveHandle();                                                 // 1329
  measureElapsedTime();                                                                                                // 1330
  return result;                                                                                                       // 1331
};                                                                                                                     // 1332
                                                                                                                       //
SandstormPermissions.downstreamTokens = function (db, root) {                                                          // 1334
  // Computes a list of the UiView tokens that are downstream in the sharing graph from a given                        // 1335
  // source. The source, `root`, can either be a token or a (grain, user) pair. The exact format                       // 1336
  // of `root` is specified in the `check()` invocation below.                                                         // 1337
  //                                                                                                                   // 1338
  // TODO(someday): Account for membrane requirements in this computation.                                             // 1339
  check(root, Match.OneOf({                                                                                            // 1341
    token: Match.ObjectIncluding({                                                                                     // 1341
      _id: String,                                                                                                     // 1341
      grainId: String                                                                                                  // 1341
    })                                                                                                                 // 1341
  }, {                                                                                                                 // 1341
    grain: Match.ObjectIncluding({                                                                                     // 1342
      _id: String,                                                                                                     // 1342
      accountId: String                                                                                                // 1342
    })                                                                                                                 // 1342
  }));                                                                                                                 // 1342
  var result = [];                                                                                                     // 1344
  var tokenStack = [];                                                                                                 // 1345
  var stackedTokens = {};                                                                                              // 1346
  var tokensBySharer = {};                                                                                             // 1347
  var tokensByParent = {};                                                                                             // 1348
  var tokensById = {};                                                                                                 // 1349
                                                                                                                       //
  function addChildren(tokenId) {                                                                                      // 1351
    var children = tokensByParent[tokenId];                                                                            // 1352
                                                                                                                       //
    if (children) {                                                                                                    // 1353
      children.forEach(function (child) {                                                                              // 1354
        if (!stackedTokens[child._id]) {                                                                               // 1355
          tokenStack.push(child);                                                                                      // 1356
          stackedTokens[child._id] = true;                                                                             // 1357
        }                                                                                                              // 1358
      });                                                                                                              // 1359
    }                                                                                                                  // 1360
  }                                                                                                                    // 1361
                                                                                                                       //
  function addSharedTokens(sharer) {                                                                                   // 1363
    var sharedTokens = tokensBySharer[sharer];                                                                         // 1364
                                                                                                                       //
    if (sharedTokens) {                                                                                                // 1365
      sharedTokens.forEach(function (sharedToken) {                                                                    // 1366
        if (!stackedTokens[sharedToken._id]) {                                                                         // 1367
          tokenStack.push(sharedToken);                                                                                // 1368
          stackedTokens[sharedToken._id] = true;                                                                       // 1369
        }                                                                                                              // 1370
      });                                                                                                              // 1371
    }                                                                                                                  // 1372
  }                                                                                                                    // 1373
                                                                                                                       //
  var grainId = root.token ? root.token.grainId : root.grain._id;                                                      // 1375
  var grain = db.getGrain(grainId);                                                                                    // 1376
                                                                                                                       //
  if (!grain || !grain.private) {                                                                                      // 1377
    return result;                                                                                                     // 1377
  }                                                                                                                    // 1377
                                                                                                                       //
  db.collections.apiTokens.find({                                                                                      // 1379
    grainId: grainId,                                                                                                  // 1379
    revoked: {                                                                                                         // 1380
      $ne: true                                                                                                        // 1380
    },                                                                                                                 // 1380
    suspended: {                                                                                                       // 1381
      $ne: true                                                                                                        // 1381
    }                                                                                                                  // 1381
  }).forEach(function (token) {                                                                                        // 1379
    tokensById[token._id] = token;                                                                                     // 1383
                                                                                                                       //
    if (token.parentToken) {                                                                                           // 1384
      if (!tokensByParent[token.parentToken]) {                                                                        // 1385
        tokensByParent[token.parentToken] = [];                                                                        // 1386
      }                                                                                                                // 1387
                                                                                                                       //
      tokensByParent[token.parentToken].push(token);                                                                   // 1389
    } else if (token.accountId) {                                                                                      // 1390
      if (!tokensBySharer[token.accountId]) {                                                                          // 1391
        tokensBySharer[token.accountId] = [];                                                                          // 1392
      }                                                                                                                // 1393
                                                                                                                       //
      tokensBySharer[token.accountId].push(token);                                                                     // 1395
    }                                                                                                                  // 1396
  });                                                                                                                  // 1397
                                                                                                                       //
  if (root.token) {                                                                                                    // 1399
    addChildren(root.token._id);                                                                                       // 1400
  } else if (root.grain) {                                                                                             // 1401
    addSharedTokens(root.grain.accountId);                                                                             // 1402
  }                                                                                                                    // 1403
                                                                                                                       //
  while (tokenStack.length > 0) {                                                                                      // 1405
    var token = tokenStack.pop();                                                                                      // 1406
    result.push(token);                                                                                                // 1407
    addChildren(token._id);                                                                                            // 1408
                                                                                                                       //
    if (token.owner && token.owner.user) {                                                                             // 1409
      addSharedTokens(token.owner.user.accountId);                                                                     // 1410
    }                                                                                                                  // 1411
  }                                                                                                                    // 1412
                                                                                                                       //
  return result;                                                                                                       // 1414
};                                                                                                                     // 1415
                                                                                                                       //
var HeaderSafeString = Match.Where(function (str) {                                                                    // 1417
  check(str, String);                                                                                                  // 1418
  return str.match(/^[\x20-\x7E]*$/);                                                                                  // 1419
});                                                                                                                    // 1420
var DavClass = Match.Where(function (str) {                                                                            // 1422
  check(str, String);                                                                                                  // 1423
  return str.match(/^[a-zA-Z0-9!#$%&'*+.^_`|~-]+$/) || str.match(/^<[\x21-\x7E]*>$/); // supposed to be a URL          // 1424
});                                                                                                                    // 1426
var ResourceMap = Match.Where(function (map) {                                                                         // 1428
  for (path in meteorBabelHelpers.sanitizeForInObject(map)) {                                                          // 1429
    if (!path.match(/^\/[\x21-\x7E]*$/)) {                                                                             // 1430
      return false;                                                                                                    // 1431
    }                                                                                                                  // 1432
                                                                                                                       //
    check(map[path], {                                                                                                 // 1434
      type: HeaderSafeString,                                                                                          // 1435
      language: Match.Optional(HeaderSafeString),                                                                      // 1436
      encoding: Match.Optional(HeaderSafeString),                                                                      // 1437
      body: String                                                                                                     // 1438
    });                                                                                                                // 1434
  }                                                                                                                    // 1440
                                                                                                                       //
  return true;                                                                                                         // 1442
});                                                                                                                    // 1443
var LocalizedString = {                                                                                                // 1445
  defaultText: String,                                                                                                 // 1446
  localizations: Match.Optional([{                                                                                     // 1447
    locale: String,                                                                                                    // 1448
    text: String                                                                                                       // 1448
  }])                                                                                                                  // 1448
};                                                                                                                     // 1445
                                                                                                                       //
SandstormPermissions.createNewApiToken = function (db, provider, grainId, petname, roleAssignment, owner, unauthenticated) {
  // Creates a new UiView API token. If `rawParentToken` is set, creates a child token.                                // 1454
  check(grainId, String);                                                                                              // 1455
  check(petname, String);                                                                                              // 1456
  check(roleAssignment, db.roleAssignmentPattern); // Meteor bug #3877: we get null here instead of undefined when we  // 1457
  // explicitly pass in undefined.                                                                                     // 1459
                                                                                                                       //
  check(provider, Match.OneOf({                                                                                        // 1460
    accountId: String,                                                                                                 // 1460
    identityId: Match.Optional(String) /*obsolete*/                                                                    // 1460
  }, {                                                                                                                 // 1460
    rawParentToken: Match.OneOf(String, Buffer)                                                                        // 1461
  }));                                                                                                                 // 1461
  check(owner, Match.OneOf({                                                                                           // 1462
    webkey: Match.OneOf(null, {                                                                                        // 1463
      forSharing: Boolean,                                                                                             // 1464
      expiresIfUnusedDuration: Match.Optional(Number)                                                                  // 1465
    })                                                                                                                 // 1463
  }, {                                                                                                                 // 1462
    user: {                                                                                                            // 1468
      accountId: String,                                                                                               // 1469
      identityId: Match.Optional(String),                                                                              // 1470
      // obsolete                                                                                                      // 1470
      title: String,                                                                                                   // 1471
      renamed: Match.Optional(Boolean),                                                                                // 1472
      upstreamTitle: Match.Optional(String),                                                                           // 1473
      seenAllActivity: Match.Optional(Boolean)                                                                         // 1474
    }                                                                                                                  // 1468
  }, {                                                                                                                 // 1467
    grain: {                                                                                                           // 1477
      grainId: String,                                                                                                 // 1478
      saveLabel: LocalizedString,                                                                                      // 1479
      introducerIdentity: Match.Optional(String),                                                                      // 1480
      // obsolete                                                                                                      // 1480
      introducerUser: Match.Optional(String) // obsolete                                                               // 1481
                                                                                                                       //
    }                                                                                                                  // 1477
  }, {                                                                                                                 // 1476
    clientPowerboxRequest: {                                                                                           // 1484
      grainId: String,                                                                                                 // 1485
      sessionId: String,                                                                                               // 1486
      introducerIdentity: Match.Optional(String) // obsolete                                                           // 1487
                                                                                                                       //
    }                                                                                                                  // 1484
  }, {                                                                                                                 // 1483
    frontend: null                                                                                                     // 1490
  }));                                                                                                                 // 1489
  check(unauthenticated, Match.OneOf(undefined, null, {                                                                // 1493
    options: Match.Optional({                                                                                          // 1494
      dav: [Match.Optional(DavClass)]                                                                                  // 1494
    }),                                                                                                                // 1494
    resources: Match.Optional(ResourceMap)                                                                             // 1495
  }));                                                                                                                 // 1493
                                                                                                                       //
  if (unauthenticated && JSON.stringify(unauthenticated).length > 4096) {                                              // 1498
    throw new Meteor.Error(400, "Unauthenticated params too large; limit 4kb.");                                       // 1499
  }                                                                                                                    // 1500
                                                                                                                       //
  var grain = db.getGrain(grainId);                                                                                    // 1502
                                                                                                                       //
  if (!grain) {                                                                                                        // 1503
    throw new Meteor.Error(403, "Unauthorized", "No grain found.");                                                    // 1504
  }                                                                                                                    // 1505
                                                                                                                       //
  var token = Random.secret();                                                                                         // 1507
                                                                                                                       //
  if (encodeURIComponent(token) !== token) {                                                                           // 1508
    // Sandstorm guarantees that tokens with a `clientPowerboxRequest` owner are URL-safe.                             // 1509
    // `Random.secret()` only uses base64url characters, so we should never get here.                                  // 1510
    throw new Meteor.Error(500, "Random.secret() returned a non-URL safe token: " + token);                            // 1511
  }                                                                                                                    // 1512
                                                                                                                       //
  var apiToken = {                                                                                                     // 1514
    _id: Crypto.createHash("sha256").update(token).digest("base64"),                                                   // 1515
    grainId: grainId,                                                                                                  // 1516
    roleAssignment: roleAssignment,                                                                                    // 1517
    petname: petname,                                                                                                  // 1518
    created: new Date(),                                                                                               // 1519
    expires: null                                                                                                      // 1520
  };                                                                                                                   // 1514
  var result = {};                                                                                                     // 1523
  var parentForSharing = false;                                                                                        // 1524
                                                                                                                       //
  if (provider.rawParentToken) {                                                                                       // 1525
    var parentToken = Crypto.createHash("sha256").update(provider.rawParentToken).digest("base64");                    // 1526
    var parentApiToken = db.collections.apiTokens.findOne({                                                            // 1527
      _id: parentToken,                                                                                                // 1528
      grainId: grainId,                                                                                                // 1528
      objectId: {                                                                                                      // 1528
        $exists: false                                                                                                 // 1528
      }                                                                                                                // 1528
    });                                                                                                                // 1528
                                                                                                                       //
    if (!parentApiToken) {                                                                                             // 1529
      throw new Meteor.Error(403, "No such parent token found.");                                                      // 1530
    }                                                                                                                  // 1531
                                                                                                                       //
    if (parentApiToken.forSharing) {                                                                                   // 1533
      parentForSharing = true;                                                                                         // 1534
    }                                                                                                                  // 1535
                                                                                                                       //
    apiToken.accountId = parentApiToken.accountId;                                                                     // 1537
    apiToken.parentToken = parentToken;                                                                                // 1539
    result.parentApiToken = parentApiToken;                                                                            // 1540
  } else if (provider.accountId) {                                                                                     // 1541
    apiToken.accountId = provider.accountId;                                                                           // 1542
  }                                                                                                                    // 1543
                                                                                                                       //
  if (owner.webkey) {                                                                                                  // 1545
    // Non-null webkey is a special case not covered in ApiTokenOwner.                                                 // 1546
    // TODO(cleanup): Maybe ApiTokenOwner.webkey should be extended with these fields?                                 // 1547
    apiToken.owner = {                                                                                                 // 1548
      webkey: null                                                                                                     // 1548
    };                                                                                                                 // 1548
    apiToken.forSharing = parentForSharing || owner.webkey.forSharing;                                                 // 1549
                                                                                                                       //
    if (owner.webkey.expiresIfUnusedDuration) {                                                                        // 1550
      apiToken.expiresIfUnused = new Date(Date.now() + owner.webkey.expiresIfUnusedDuration);                          // 1551
    }                                                                                                                  // 1552
  } else if (owner.user) {                                                                                             // 1553
    // Determine the user's identity ID (their user ID as seen by the grain).                                          // 1554
    var identityId = db.getOrGenerateIdentityId(owner.user.accountId, grain);                                          // 1555
    var grainInfo = db.getDenormalizedGrainInfo(grainId);                                                              // 1557
    apiToken.owner = {                                                                                                 // 1558
      user: {                                                                                                          // 1559
        accountId: owner.user.accountId,                                                                               // 1560
        identityId: identityId,                                                                                        // 1561
        title: owner.user.title,                                                                                       // 1562
        denormalizedGrainMetadata: grainInfo                                                                           // 1563
      }                                                                                                                // 1559
    };                                                                                                                 // 1558
                                                                                                                       //
    if (grain.title !== owner.user.title) {                                                                            // 1567
      apiToken.owner.user.upstreamTitle = grain.title;                                                                 // 1568
    }                                                                                                                  // 1569
  } else {                                                                                                             // 1570
    // Note: Also covers the case of `webkey: null`.                                                                   // 1571
    apiToken.owner = owner;                                                                                            // 1572
  }                                                                                                                    // 1573
                                                                                                                       //
  if (unauthenticated) {                                                                                               // 1575
    var apiHost = {                                                                                                    // 1576
      _id: db.apiHostIdHashForToken(token),                                                                            // 1577
      hash2: Crypto.createHash("sha256").update(apiToken._id).digest("base64")                                         // 1578
    };                                                                                                                 // 1576
                                                                                                                       //
    if (unauthenticated.options) {                                                                                     // 1580
      apiHost.options = unauthenticated.options;                                                                       // 1581
    }                                                                                                                  // 1582
                                                                                                                       //
    if (unauthenticated.resources) {                                                                                   // 1584
      // Mongo requires keys in objects to be escaped. Ugh.                                                            // 1585
      apiHost.resources = {};                                                                                          // 1586
                                                                                                                       //
      for (var key in meteorBabelHelpers.sanitizeForInObject(unauthenticated.resources)) {                             // 1587
        apiHost.resources[SandstormDb.escapeMongoKey(key)] = unauthenticated.resources[key];                           // 1588
      }                                                                                                                // 1589
    }                                                                                                                  // 1590
                                                                                                                       //
    db.collections.apiHosts.insert(apiHost);                                                                           // 1592
    apiToken.hasApiHost = true;                                                                                        // 1593
  }                                                                                                                    // 1594
                                                                                                                       //
  db.collections.apiTokens.insert(apiToken);                                                                           // 1596
  result.id = apiToken._id;                                                                                            // 1598
  result.token = token;                                                                                                // 1599
  return result;                                                                                                       // 1600
}; // Make self-destructing tokens actually self-destruct, so they don't                                               // 1601
// clutter the token list view.                                                                                        // 1604
                                                                                                                       //
                                                                                                                       //
SandstormPermissions.cleanupSelfDestructing = function (db) {                                                          // 1605
  return function () {                                                                                                 // 1606
    var now = new Date();                                                                                              // 1607
    db.removeApiTokens({                                                                                               // 1608
      expiresIfUnused: {                                                                                               // 1608
        $lt: now                                                                                                       // 1608
      }                                                                                                                // 1608
    });                                                                                                                // 1608
  };                                                                                                                   // 1609
};                                                                                                                     // 1610
                                                                                                                       //
SandstormPermissions.cleanupClientPowerboxTokens = function (db) {                                                     // 1612
  return function () {                                                                                                 // 1613
    var tenMinutesAgo = new Date(Date.now() - 1000 * 60 * 10);                                                         // 1614
    db.removeApiTokens({                                                                                               // 1615
      $or: [{                                                                                                          // 1616
        "owner.clientPowerboxRequest": {                                                                               // 1617
          $exists: true                                                                                                // 1617
        }                                                                                                              // 1617
      }, {                                                                                                             // 1617
        "owner.clientPowerboxOffer": {                                                                                 // 1618
          $exists: true                                                                                                // 1618
        }                                                                                                              // 1618
      }],                                                                                                              // 1618
      created: {                                                                                                       // 1620
        $lt: tenMinutesAgo                                                                                             // 1620
      }                                                                                                                // 1620
    });                                                                                                                // 1615
  };                                                                                                                   // 1622
};                                                                                                                     // 1623
                                                                                                                       //
Meteor.methods({                                                                                                       // 1625
  transitiveShares: function (obsolete, grainId) {                                                                     // 1626
    check(grainId, String);                                                                                            // 1627
                                                                                                                       //
    if (this.userId) {                                                                                                 // 1628
      var db = this.connection.sandstormDb;                                                                            // 1629
      return SandstormPermissions.downstreamTokens(db, {                                                               // 1630
        grain: {                                                                                                       // 1631
          _id: grainId,                                                                                                // 1631
          accountId: this.userId                                                                                       // 1631
        }                                                                                                              // 1631
      });                                                                                                              // 1631
    }                                                                                                                  // 1632
  },                                                                                                                   // 1633
  newApiToken: function (provider, grainId, petname, roleAssignment, owner, unauthenticated) {                         // 1635
    check(provider, Match.OneOf({                                                                                      // 1636
      identityId: String                                                                                               // 1636
    }, // obsolete                                                                                                     // 1636
    {                                                                                                                  // 1637
      accountId: String                                                                                                // 1637
    }, {                                                                                                               // 1637
      rawParentToken: String                                                                                           // 1638
    }));                                                                                                               // 1638
                                                                                                                       //
    if (!owner.user && !owner.webkey) {                                                                                // 1639
      throw new Meteor.Error(403, "'webkey' and 'user' are the only allowed owners in newApiToken()");                 // 1640
    } // other check()s happen in SandstormPermissions.createNewApiToken().                                            // 1642
                                                                                                                       //
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 1645
                                                                                                                       //
    if (provider.identityId) {                                                                                         // 1646
      provider.accountId = this.userId;                                                                                // 1647
      delete provider.identityId;                                                                                      // 1648
    } else if (provider.accountId) {                                                                                   // 1649
      if (provider.accountId !== this.userId) {                                                                        // 1650
        throw new Meteor.Error(403, "Not the current user: " + provider.accountId);                                    // 1651
      }                                                                                                                // 1652
    }                                                                                                                  // 1653
                                                                                                                       //
    return SandstormPermissions.createNewApiToken(this.connection.sandstormDb, provider, grainId, petname, roleAssignment, owner, unauthenticated);
  },                                                                                                                   // 1658
  updateApiToken: function (token, newFields) {                                                                        // 1660
    var db = this.connection.sandstormDb;                                                                              // 1661
    check(token, String);                                                                                              // 1663
    check(newFields, {                                                                                                 // 1664
      petname: Match.Optional(String),                                                                                 // 1664
      roleAssignment: Match.Optional(db.roleAssignmentPattern),                                                        // 1665
      revoked: Match.Optional(Boolean),                                                                                // 1666
      suspended: Match.Optional(Boolean)                                                                               // 1667
    });                                                                                                                // 1664
                                                                                                                       //
    if (!this.userId) {                                                                                                // 1669
      throw new Meteor.Error(403, "Must be logged in to modify a token");                                              // 1670
    }                                                                                                                  // 1671
                                                                                                                       //
    var apiToken = db.collections.apiTokens.findOne(token);                                                            // 1673
                                                                                                                       //
    if (!apiToken) {                                                                                                   // 1674
      throw new Meteor.Error(404, "No such token found.");                                                             // 1675
    }                                                                                                                  // 1676
                                                                                                                       //
    if (apiToken.accountId === this.userId) {                                                                          // 1678
      var modifier = {                                                                                                 // 1679
        $set: newFields                                                                                                // 1679
      };                                                                                                               // 1679
      db.collections.apiTokens.update(token, modifier);                                                                // 1680
    } else {                                                                                                           // 1681
      throw new Meteor.Error(403, "User not authorized to modify this token.");                                        // 1682
    }                                                                                                                  // 1683
  }                                                                                                                    // 1684
});                                                                                                                    // 1625
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}}}},{
  "extensions": [
    ".js",
    ".json"
  ]
});
require("./node_modules/meteor/sandstorm-permissions/permissions.js");

/* Exports */
if (typeof Package === 'undefined') Package = {};
(function (pkg, symbols) {
  for (var s in symbols)
    (s in pkg) || (pkg[s] = symbols[s]);
})(Package['sandstorm-permissions'] = {}, {
  SandstormPermissions: SandstormPermissions
});

})();

//# sourceMappingURL=sandstorm-permissions.js.map
