var require = meteorInstall({"imports":{"server":{"accounts":{"ldap.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/ldap.js                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  LDAP: function () {                                                                                                  // 1
    return LDAP;                                                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var ldapjs = void 0;                                                                                                   // 1
module.watch(require("ldapjs"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    ldapjs = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 1
                                                                                                                       //
// At a minimum, set up LDAP_DEFAULTS.url and .dn according to                                                         // 4
// your needs. url should appear as 'ldap://your.url.here'                                                             // 5
// dn should appear in normal ldap format of comma separated attribute=value                                           // 6
// e.g. 'uid=someuser,cn=users,dc=somevalue'                                                                           // 7
LDAP_DEFAULTS = {                                                                                                      // 8
  url: false,                                                                                                          // 9
  port: "389",                                                                                                         // 10
  dn: false,                                                                                                           // 11
  searchDN: false,                                                                                                     // 12
  searchCredentials: false,                                                                                            // 13
  createNewUser: true,                                                                                                 // 14
  defaultDomain: false,                                                                                                // 15
  searchResultsProfileMap: false,                                                                                      // 16
  base: null,                                                                                                          // 17
  search: "(objectclass=*)",                                                                                           // 18
  ldapsCertificate: false                                                                                              // 19
}; /**                                                                                                                 // 8
    @class LDAP                                                                                                        //
    @constructor                                                                                                       //
    */                                                                                                                 //
                                                                                                                       //
function LDAP() {                                                                                                      // 26
  // Set options                                                                                                       // 27
  this.options = _.clone(LDAP_DEFAULTS);                                                                               // 28
}                                                                                                                      // 29
                                                                                                                       //
; /**                                                                                                                  // 29
   * Attempt to bind (authenticate) ldap                                                                               //
   * and perform a dn search if specified                                                                              //
   *                                                                                                                   //
   * @method ldapCheck                                                                                                 //
   *                                                                                                                   //
   * @param {Object} options  Object with username, ldapPass and overrides for LDAP_DEFAULTS object.                   //
   * Additionally the searchBeforeBind parameter can be specified, which is used to search for the DN                  //
   * if not provided.                                                                                                  //
   */                                                                                                                  //
                                                                                                                       //
LDAP.prototype.ldapCheck = function (db, options) {                                                                    // 41
  var _this = this;                                                                                                    // 43
                                                                                                                       //
  options = options || {};                                                                                             // 45
                                                                                                                       //
  if (options.hasOwnProperty("username") && options.hasOwnProperty("ldapPass") || options.hasOwnProperty("searchUsername")) {
    _this.options.base = db.getLdapBase();                                                                             // 49
    _this.options.url = db.getLdapUrl();                                                                               // 50
    _this.options.searchBeforeBind = {};                                                                               // 51
    _this.options.searchBeforeBind[options.searchUsernameField || db.getLdapSearchUsername()] = options.searchUsername || options.username;
    _this.options.filter = db.getLdapFilter() || "(objectclass=*)";                                                    // 54
    _this.options.searchBindDn = db.getLdapSearchBindDn();                                                             // 55
    _this.options.searchBindPassword = db.getLdapSearchBindPassword();                                                 // 56
    var resolved = false;                                                                                              // 58
    var ldapAsyncFut = new Future(); // Create ldap client                                                             // 59
                                                                                                                       //
    var fullUrl = db.getLdapUrl();                                                                                     // 62
    var client = null;                                                                                                 // 63
                                                                                                                       //
    var errorFunc = function (err) {                                                                                   // 65
      if (err) {                                                                                                       // 66
        if (resolved) return;                                                                                          // 67
        resolved = true;                                                                                               // 68
        ldapAsyncFut.return({                                                                                          // 69
          error: err                                                                                                   // 70
        });                                                                                                            // 69
      }                                                                                                                // 72
    };                                                                                                                 // 73
                                                                                                                       //
    if (fullUrl.indexOf("ldaps://") === 0) {                                                                           // 75
      var tlsOptions = {};                                                                                             // 76
      var cert = db.getLdapCaCert();                                                                                   // 77
                                                                                                                       //
      if (cert) {                                                                                                      // 78
        tlsOptions.ca = cert;                                                                                          // 79
      }                                                                                                                // 80
                                                                                                                       //
      client = ldapjs.createClient({                                                                                   // 82
        url: fullUrl,                                                                                                  // 83
        tlsOptions: tlsOptions                                                                                         // 84
      }, errorFunc);                                                                                                   // 82
    } else {                                                                                                           // 86
      client = ldapjs.createClient({                                                                                   // 87
        url: fullUrl                                                                                                   // 88
      }, errorFunc);                                                                                                   // 87
    }                                                                                                                  // 90
                                                                                                                       //
    client.on("error", errorFunc);                                                                                     // 92
    var username = options.username;                                                                                   // 94
    var domain = _this.options.defaultDomain;                                                                          // 95
                                                                                                                       //
    if (!options.hasOwnProperty("searchUsername")) {                                                                   // 97
      // Slide @xyz.whatever from username if it was passed in                                                         // 98
      // and replace it with the domain specified in defaults                                                          // 99
      var emailSliceIndex = options.username.indexOf("@"); // If user appended email domain, strip it out              // 100
      // And use the defaults.defaultDomain if set                                                                     // 103
                                                                                                                       //
      if (emailSliceIndex !== -1) {                                                                                    // 104
        username = options.username.substring(0, emailSliceIndex);                                                     // 105
        domain = domain || options.username.substring(emailSliceIndex + 1, options.username.length);                   // 106
      } else {                                                                                                         // 107
        username = options.username;                                                                                   // 108
      }                                                                                                                // 109
    }                                                                                                                  // 110
                                                                                                                       //
    if (_this.options.searchBindDn) {                                                                                  // 112
      var ldapBindFut = new Future();                                                                                  // 113
      client.bind(_this.options.searchBindDn, _this.options.searchBindPassword, function (err) {                       // 114
        if (err) {                                                                                                     // 116
          ldapBindFut.throw(err);                                                                                      // 117
        } else {                                                                                                       // 118
          ldapBindFut.return();                                                                                        // 119
        }                                                                                                              // 120
      });                                                                                                              // 121
                                                                                                                       //
      try {                                                                                                            // 124
        ldapBindFut.wait();                                                                                            // 125
      } catch (err) {                                                                                                  // 126
        return {                                                                                                       // 127
          error: err                                                                                                   // 127
        };                                                                                                             // 127
      }                                                                                                                // 128
    } // initialize result                                                                                             // 129
                                                                                                                       //
                                                                                                                       //
    var retObject = {                                                                                                  // 131
      username: username,                                                                                              // 132
      email: domain ? username + "@" + domain : false,                                                                 // 133
      emptySearch: true,                                                                                               // 134
      searchResults: {}                                                                                                // 135
    };                                                                                                                 // 131
    var filter = _this.options.filter;                                                                                 // 138
    Object.keys(_this.options.searchBeforeBind).forEach(function (searchKey) {                                         // 139
      filter = "&" + filter + "(" + searchKey + "=" + _this.options.searchBeforeBind[searchKey] + ")";                 // 140
    });                                                                                                                // 141
    var searchOptions = {                                                                                              // 143
      scope: "sub",                                                                                                    // 144
      sizeLimit: 1,                                                                                                    // 145
      filter: filter                                                                                                   // 146
    }; // perform LDAP search to determine DN                                                                          // 143
                                                                                                                       //
    client.search(_this.options.base, searchOptions, function (err, res) {                                             // 150
      if (err) {                                                                                                       // 151
        if (resolved) return;                                                                                          // 152
        resolved = true;                                                                                               // 153
        ldapAsyncFut.return({                                                                                          // 154
          error: err                                                                                                   // 155
        });                                                                                                            // 154
        return;                                                                                                        // 157
      }                                                                                                                // 158
                                                                                                                       //
      retObject.emptySearch = true;                                                                                    // 160
      res.on("searchEntry", function (entry) {                                                                         // 161
        retObject.dn = entry.objectName;                                                                               // 162
        retObject.username = retObject.dn;                                                                             // 163
        retObject.emptySearch = false;                                                                                 // 164
        retObject.searchResults = _.omit(entry.object, "userPassword");                                                // 166
                                                                                                                       //
        if (options.hasOwnProperty("searchUsername")) {                                                                // 168
          // This was only a search, return immediately                                                                // 169
          resolved = true;                                                                                             // 170
          ldapAsyncFut.return(retObject);                                                                              // 171
          return;                                                                                                      // 172
        } // use the determined DN to bind                                                                             // 173
                                                                                                                       //
                                                                                                                       //
        client.bind(entry.objectName, options.ldapPass, function (err) {                                               // 176
          try {                                                                                                        // 177
            if (err) {                                                                                                 // 178
              throw new Meteor.Error(err.code, err.message);                                                           // 179
            } else {                                                                                                   // 180
              resolved = true;                                                                                         // 181
              ldapAsyncFut.return(retObject);                                                                          // 182
            }                                                                                                          // 183
          } catch (e) {                                                                                                // 184
            if (resolved) return;                                                                                      // 186
            resolved = true;                                                                                           // 187
            ldapAsyncFut.return({                                                                                      // 188
              error: e                                                                                                 // 189
            });                                                                                                        // 188
          }                                                                                                            // 191
        });                                                                                                            // 192
      });                                                                                                              // 193
      res.on("error", errorFunc); // If no dn is found, return as is.                                                  // 195
                                                                                                                       //
      res.on("end", function (result) {                                                                                // 198
        if (retObject.dn === undefined) {                                                                              // 199
          resolved = true;                                                                                             // 200
          ldapAsyncFut.return(retObject);                                                                              // 201
        }                                                                                                              // 202
      });                                                                                                              // 203
    });                                                                                                                // 204
    return ldapAsyncFut.wait();                                                                                        // 206
  } else {                                                                                                             // 208
    throw new Meteor.Error(403, "Missing LDAP Auth Parameter");                                                        // 209
  }                                                                                                                    // 210
};                                                                                                                     // 212
                                                                                                                       //
LDAP.prototype.updateUserQuota = function (db, user) {                                                                 // 214
  var fallback = {                                                                                                     // 215
    storage: user.cachedStorageQuota || 0,                                                                             // 216
    grains: Infinity,                                                                                                  // 217
    compute: Infinity                                                                                                  // 218
  };                                                                                                                   // 215
  var setting = db.collections.settings.findOne({                                                                      // 221
    _id: "quotaLdapAttribute"                                                                                          // 221
  });                                                                                                                  // 221
  if (!setting || !setting.value) return fallback; // TODO(someday): don't just assume the first login identity is the primary identity?
                                                                                                                       //
  var email = db.getPrimaryEmail(user._id, user.loginCredentials[0].id);                                               // 225
  if (!email) return fallback;                                                                                         // 226
  var ldapUser = void 0;                                                                                               // 228
                                                                                                                       //
  try {                                                                                                                // 229
    ldapUser = this.ldapCheck(db, {                                                                                    // 230
      searchUsername: email,                                                                                           // 230
      searchUsernameField: "mail"                                                                                      // 230
    });                                                                                                                // 230
  } catch (err) {                                                                                                      // 231
    console.error("Error looking up quota from LDAP");                                                                 // 232
    return fallback;                                                                                                   // 233
  }                                                                                                                    // 234
                                                                                                                       //
  if (!ldapUser || !ldapUser.searchResults) return fallback;                                                           // 236
  var newStorageQuota = +ldapUser.searchResults[setting.value];                                                        // 238
                                                                                                                       //
  if (newStorageQuota !== user.cachedStorageQuota) {                                                                   // 239
    Meteor.users.update({                                                                                              // 240
      _id: user._id                                                                                                    // 240
    }, {                                                                                                               // 240
      $set: {                                                                                                          // 240
        cachedStorageQuota: newStorageQuota                                                                            // 240
      }                                                                                                                // 240
    });                                                                                                                // 240
  } // TODO(someday): cache timestamp as well and only check/update if greater than 60s ago                            // 241
                                                                                                                       //
                                                                                                                       //
  return {                                                                                                             // 244
    storage: newStorageQuota,                                                                                          // 245
    grains: Infinity,                                                                                                  // 246
    compute: Infinity                                                                                                  // 247
  };                                                                                                                   // 244
};                                                                                                                     // 249
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"picture.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/picture.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  userPictureUrl: function () {                                                                                        // 1
    return userPictureUrl;                                                                                             // 1
  },                                                                                                                   // 1
  fetchPicture: function () {                                                                                          // 1
    return fetchPicture;                                                                                               // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var HTTP = void 0;                                                                                                     // 1
module.watch(require("meteor/http"), {                                                                                 // 1
  HTTP: function (v) {                                                                                                 // 1
    HTTP = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var userPictureUrl = function (user) {                                                                                 // 3
  if (user.services && !(user.profile && user.profile.picture)) {                                                      // 4
    // Try to determine user's avatar URL from login service.                                                          // 5
    var google = user.services.google;                                                                                 // 7
                                                                                                                       //
    if (google && google.picture) {                                                                                    // 8
      return google.picture;                                                                                           // 9
    }                                                                                                                  // 10
                                                                                                                       //
    var github = user.services.github;                                                                                 // 12
                                                                                                                       //
    if (github && github.id) {                                                                                         // 13
      return "https://avatars.githubusercontent.com/u/" + github.id;                                                   // 14
    } // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be                    // 15
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                          // 18
    // Github are different because they are actually the identity providers, so they already know                     // 19
    // the user logged in.                                                                                             // 20
                                                                                                                       //
  }                                                                                                                    // 21
};                                                                                                                     // 22
                                                                                                                       //
var fetchPicture = function (db, url) {                                                                                // 24
  try {                                                                                                                // 25
    var result = HTTP.get(url, {                                                                                       // 26
      npmRequestOptions: {                                                                                             // 27
        encoding: null                                                                                                 // 27
      },                                                                                                               // 27
      timeout: 5000                                                                                                    // 28
    });                                                                                                                // 26
    var metadata = {};                                                                                                 // 31
    metadata.mimeType = result.headers["content-type"];                                                                // 33
                                                                                                                       //
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 && metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                                  // 36
    }                                                                                                                  // 37
                                                                                                                       //
    var enc = result.headers["content-encoding"];                                                                      // 39
                                                                                                                       //
    if (enc && enc !== "identity") {                                                                                   // 40
      metadata.encoding = enc;                                                                                         // 41
    }                                                                                                                  // 42
                                                                                                                       //
    return db.addStaticAsset(metadata, result.content);                                                                // 44
  } catch (err) {                                                                                                      // 45
    console.error("failed to fetch user profile picture:", url, err.stack);                                            // 46
  }                                                                                                                    // 47
};                                                                                                                     // 48
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"saml-utils.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/saml-utils.js                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  SAML: function () {                                                                                                  // 1
    return SAML;                                                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var crypto = void 0;                                                                                                   // 1
module.watch(require("crypto"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var querystring = void 0;                                                                                              // 1
module.watch(require("querystring"), {                                                                                 // 1
  "default": function (v) {                                                                                            // 1
    querystring = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module.watch(require("url"), {                                                                                         // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var xml2js = void 0;                                                                                                   // 1
module.watch(require("xml2js"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    xml2js = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var xmlbuilder = void 0;                                                                                               // 1
module.watch(require("xmlbuilder"), {                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    xmlbuilder = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var xmlCrypto = void 0;                                                                                                // 1
module.watch(require("xml-crypto"), {                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    xmlCrypto = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var xmldom = void 0;                                                                                                   // 1
module.watch(require("xmldom"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    xmldom = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var zlib = void 0;                                                                                                     // 1
module.watch(require("zlib"), {                                                                                        // 1
  "default": function (v) {                                                                                            // 1
    zlib = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 7);                                                                                                                 // 1
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;                                                               // 10
                                                                                                                       //
var SAML = function (options) {                                                                                        // 12
  this.options = this.initialize(options);                                                                             // 13
};                                                                                                                     // 14
                                                                                                                       //
SAML.prototype.initialize = function (options) {                                                                       // 16
  if (!options) {                                                                                                      // 17
    options = {};                                                                                                      // 18
  }                                                                                                                    // 19
                                                                                                                       //
  if (!options.protocol) {                                                                                             // 21
    options.protocol = "https://";                                                                                     // 22
  }                                                                                                                    // 23
                                                                                                                       //
  if (!options.path) {                                                                                                 // 25
    options.path = "/saml/consume";                                                                                    // 26
  }                                                                                                                    // 27
                                                                                                                       //
  if (!options.issuer) {                                                                                               // 29
    options.issuer = "onelogin_saml";                                                                                  // 30
  }                                                                                                                    // 31
                                                                                                                       //
  if (options.identifierFormat === undefined) {                                                                        // 33
    options.identifierFormat = "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent";                                 // 34
  }                                                                                                                    // 35
                                                                                                                       //
  return options;                                                                                                      // 37
};                                                                                                                     // 38
                                                                                                                       //
SAML.prototype.generateUniqueID = function () {                                                                        // 40
  var chars = "abcdef0123456789";                                                                                      // 41
  var uniqueID = "";                                                                                                   // 42
                                                                                                                       //
  for (var i = 0; i < 20; i++) {                                                                                       // 43
    uniqueID += chars.substr(Math.floor(Math.random() * 15), 1);                                                       // 44
  }                                                                                                                    // 45
                                                                                                                       //
  return uniqueID;                                                                                                     // 47
};                                                                                                                     // 48
                                                                                                                       //
SAML.prototype.generateInstant = function () {                                                                         // 50
  var date = new Date();                                                                                               // 51
  return date.getUTCFullYear() + "-" + ("0" + (date.getUTCMonth() + 1)).slice(-2) + "-" + ("0" + date.getUTCDate()).slice(-2) + "T" + ("0" + date.getUTCHours()).slice(-2) + ":" + ("0" + date.getUTCMinutes()).slice(-2) + ":" + ("0" + date.getUTCSeconds()).slice(-2) + "Z";
};                                                                                                                     // 53
                                                                                                                       //
SAML.prototype.signRequest = function (xml) {                                                                          // 55
  var signer = crypto.createSign("RSA-SHA1");                                                                          // 56
  signer.update(xml);                                                                                                  // 57
  return signer.sign(this.options.privateCert, "base64");                                                              // 58
};                                                                                                                     // 59
                                                                                                                       //
SAML.prototype.generateAuthorizeRequest = function (req) {                                                             // 61
  var id = "_" + this.generateUniqueID();                                                                              // 62
  var instant = this.generateInstant();                                                                                // 63
  var callbackUrl = void 0; // Post-auth destination                                                                   // 65
                                                                                                                       //
  if (this.options.callbackUrl) {                                                                                      // 67
    callbackUrl = this.options.callbackUrl;                                                                            // 68
  } else {                                                                                                             // 69
    callbackUrl = this.options.protocol + req.headers.host + this.options.path;                                        // 70
  }                                                                                                                    // 71
                                                                                                                       //
  if (this.options.id) id = this.options.id;                                                                           // 73
  var request = "<samlp:AuthnRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" ProtocolBinding=\"urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST\" AssertionConsumerServiceURL=\"" + callbackUrl + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>\n";
                                                                                                                       //
  if (this.options.identifierFormat) {                                                                                 // 82
    request += "<samlp:NameIDPolicy xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" Format=\"" + this.options.identifierFormat + "\" AllowCreate=\"true\"></samlp:NameIDPolicy>\n";
  }                                                                                                                    // 85
                                                                                                                       //
  request += "</samlp:AuthnRequest>";                                                                                  // 87
  return request;                                                                                                      // 89
};                                                                                                                     // 90
                                                                                                                       //
SAML.prototype.generateLogoutRequest = function (req) {                                                                // 92
  var id = "_" + this.generateUniqueID();                                                                              // 93
  var instant = this.generateInstant(); //samlp:LogoutRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
  // ID="_135ad2fd-b275-4428-b5d6-3ac3361c3a7f" Version="2.0" Destination="https://idphost/adfs/ls/"                   // 97
  //IssueInstant="2008-06-03T12:59:57Z"><saml:Issuer>myhost</saml:Issuer><NameID Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
  //NameQualifier="https://idphost/adfs/ls/">myemail@mydomain.com</NameID<samlp:SessionIndex>_0628125f-7f95-42cc-ad8e-fde86ae90bbe
  //</samlp:SessionIndex></samlp:LogoutRequest>                                                                        // 100
                                                                                                                       //
  var request = "<samlp:LogoutRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" " + "xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>" + "<saml:NameID Format=\"" + req.user.nameIDFormat + "\">" + req.user.nameID + "</saml:NameID>" + "</samlp:LogoutRequest>";
  return request;                                                                                                      // 108
};                                                                                                                     // 109
                                                                                                                       //
SAML.prototype.requestToUrl = function (request, operation, callback) {                                                // 111
  var _this = this;                                                                                                    // 112
                                                                                                                       //
  zlib.deflateRaw(request, function (err, buffer) {                                                                    // 113
    if (err) {                                                                                                         // 114
      return callback(err);                                                                                            // 115
    }                                                                                                                  // 116
                                                                                                                       //
    var base64 = buffer.toString("base64");                                                                            // 118
    var target = _this.options.entryPoint;                                                                             // 119
                                                                                                                       //
    if (operation === "logout") {                                                                                      // 121
      if (_this.options.logoutUrl) {                                                                                   // 122
        target = _this.options.logoutUrl;                                                                              // 123
      }                                                                                                                // 124
    }                                                                                                                  // 125
                                                                                                                       //
    if (target.indexOf("?") > 0) target += "&";else target += "?";                                                     // 127
    var samlRequest = {                                                                                                // 132
      SAMLRequest: base64                                                                                              // 133
    };                                                                                                                 // 132
                                                                                                                       //
    if (_this.options.privateCert) {                                                                                   // 136
      samlRequest.SigAlg = "http://www.w3.org/2000/09/xmldsig#rsa-sha1";                                               // 137
      samlRequest.Signature = _this.signRequest(querystring.stringify(samlRequest));                                   // 138
    }                                                                                                                  // 139
                                                                                                                       //
    target += querystring.stringify(samlRequest);                                                                      // 141
    callback(null, target);                                                                                            // 143
  });                                                                                                                  // 144
};                                                                                                                     // 145
                                                                                                                       //
SAML.prototype.getAuthorizeUrl = function (req, callback) {                                                            // 147
  var request = this.generateAuthorizeRequest(req);                                                                    // 148
  this.requestToUrl(request, "authorize", callback);                                                                   // 150
};                                                                                                                     // 151
                                                                                                                       //
SAML.prototype.getLogoutUrl = function (req, callback) {                                                               // 153
  var request = this.generateLogoutRequest(req);                                                                       // 154
  this.requestToUrl(request, "logout", callback);                                                                      // 156
};                                                                                                                     // 157
                                                                                                                       //
SAML.prototype.certToPEM = function (cert) {                                                                           // 159
  cert = cert.match(/.{1,64}/g).join("\n");                                                                            // 160
  cert = "-----BEGIN CERTIFICATE-----\n" + cert;                                                                       // 161
  cert = cert + "\n-----END CERTIFICATE-----\n";                                                                       // 162
  return cert;                                                                                                         // 163
};                                                                                                                     // 164
                                                                                                                       //
SAML.prototype.validateSignature = function (xml, cert) {                                                              // 166
  var _this = this;                                                                                                    // 167
                                                                                                                       //
  var doc = new xmldom.DOMParser().parseFromString(xml);                                                               // 168
  var signature = xmlCrypto.xpath(doc, "//*[local-name(.)='Signature' and namespace-uri(.)='http://www.w3.org/2000/09/xmldsig#']")[0];
  var sig = new xmlCrypto.SignedXml();                                                                                 // 170
  sig.keyInfoProvider = {                                                                                              // 171
    getKeyInfo: function (key) {                                                                                       // 172
      return "<X509Data></X509Data>";                                                                                  // 173
    },                                                                                                                 // 174
    getKey: function (keyInfo) {                                                                                       // 176
      return _this.certToPEM(cert);                                                                                    // 177
    }                                                                                                                  // 178
  };                                                                                                                   // 171
  sig.loadSignature(signature.toString());                                                                             // 180
  return sig.checkSignature(xml);                                                                                      // 181
};                                                                                                                     // 182
                                                                                                                       //
SAML.prototype.getElement = function (parentElement, elementName) {                                                    // 184
  if (parentElement["saml:" + elementName]) {                                                                          // 185
    return parentElement["saml:" + elementName];                                                                       // 186
  } else if (parentElement["samlp:" + elementName]) {                                                                  // 187
    return parentElement["samlp:" + elementName];                                                                      // 188
  } else if (parentElement["saml2p:" + elementName]) {                                                                 // 189
    return parentElement["saml2p:" + elementName];                                                                     // 190
  } else if (parentElement["saml2:" + elementName]) {                                                                  // 191
    return parentElement["saml2:" + elementName];                                                                      // 192
  }                                                                                                                    // 193
                                                                                                                       //
  return parentElement[elementName];                                                                                   // 195
};                                                                                                                     // 196
                                                                                                                       //
SAML.prototype.parseLogoutRequest = function (xml, callback) {                                                         // 198
  var _this = this;                                                                                                    // 199
                                                                                                                       //
  var parser = new xml2js.Parser({                                                                                     // 200
    explicitRoot: true                                                                                                 // 200
  });                                                                                                                  // 200
  parser.parseString(xml, function (err, doc) {                                                                        // 201
    if (err) {                                                                                                         // 202
      callback(err);                                                                                                   // 203
      return;                                                                                                          // 204
    }                                                                                                                  // 205
                                                                                                                       //
    var request = _this.getElement(doc, "LogoutRequest");                                                              // 207
                                                                                                                       //
    var nameId = _this.getElement(request, "NameID")[0];                                                               // 208
                                                                                                                       //
    callback(null, nameId && nameId._);                                                                                // 209
  });                                                                                                                  // 210
};                                                                                                                     // 211
                                                                                                                       //
SAML.prototype.validateResponse = function (samlResponse, callback) {                                                  // 213
  var _this = this;                                                                                                    // 214
                                                                                                                       //
  var xml = new Buffer(samlResponse, "base64").toString("utf8");                                                       // 215
  var parser = new xml2js.Parser({                                                                                     // 216
    explicitRoot: true                                                                                                 // 216
  });                                                                                                                  // 216
  parser.parseString(xml, function (err, doc) {                                                                        // 217
    // Verify signature                                                                                                // 218
    if (_this.options.cert && !_this.validateSignature(xml, _this.options.cert)) {                                     // 219
      return callback(new Error("Invalid signature"), null, false, xml);                                               // 220
    }                                                                                                                  // 221
                                                                                                                       //
    var response = _this.getElement(doc, "Response");                                                                  // 223
                                                                                                                       //
    if (response) {                                                                                                    // 224
      var assertion = _this.getElement(response, "Assertion");                                                         // 225
                                                                                                                       //
      if (!assertion) {                                                                                                // 226
        return callback(new Error("Missing SAML assertion"), null, false, xml);                                        // 227
      }                                                                                                                // 228
                                                                                                                       //
      profile = {};                                                                                                    // 230
                                                                                                                       //
      if (response.$ && response.$.InResponseTo) {                                                                     // 232
        profile.inResponseToId = response.$.InResponseTo;                                                              // 233
      }                                                                                                                // 234
                                                                                                                       //
      if (response.$ && response.$.Destination) {                                                                      // 236
        if (!response.$.Destination.startsWith(process.env.ROOT_URL)) {                                                // 237
          return callback(new Error("SAML Response received with invalid Destination: " + response.$.Destination), null, false, xml);
        }                                                                                                              // 240
      }                                                                                                                // 241
                                                                                                                       //
      var issuer = _this.getElement(assertion[0], "Issuer");                                                           // 243
                                                                                                                       //
      if (issuer) {                                                                                                    // 244
        profile.issuer = issuer[0];                                                                                    // 245
      }                                                                                                                // 246
                                                                                                                       //
      var subject = _this.getElement(assertion[0], "Subject");                                                         // 248
                                                                                                                       //
      if (subject) {                                                                                                   // 249
        var nameID = _this.getElement(subject[0], "NameID");                                                           // 250
                                                                                                                       //
        if (nameID) {                                                                                                  // 251
          profile.nameID = nameID[0]._;                                                                                // 252
                                                                                                                       //
          if (nameID[0].$.Format) {                                                                                    // 254
            profile.nameIDFormat = nameID[0].$.Format;                                                                 // 255
                                                                                                                       //
            if (profile.nameIDFormat.toLowerCase().indexOf("transient") !== -1) {                                      // 256
              return callback(new Error("SAML returned a transient NameID. Sandstorm requires a persistent NameID. " + "Please check your IdP config."), null, false, xml);
            }                                                                                                          // 260
          }                                                                                                            // 261
        }                                                                                                              // 262
                                                                                                                       //
        var subjectConfirmation = _this.getElement(subject[0], "SubjectConfirmation");                                 // 264
                                                                                                                       //
        if (subjectConfirmation) {                                                                                     // 265
          var subjectConfirmationData = _this.getElement(subjectConfirmation[0], "SubjectConfirmationData")[0];        // 266
                                                                                                                       //
          if (subjectConfirmationData) {                                                                               // 268
            var recipient = subjectConfirmationData.$.Recipient;                                                       // 269
                                                                                                                       //
            if (recipient && !recipient.startsWith(process.env.ROOT_URL)) {                                            // 270
              return callback(new Error("SAML sent to wrong recipient"), null, false, xml);                            // 271
            }                                                                                                          // 272
                                                                                                                       //
            var nowMs = Date.now();                                                                                    // 274
            var notBefore = subjectConfirmationData.$.NotBefore;                                                       // 275
                                                                                                                       //
            if (notBefore && nowMs < Date.parse(notBefore)) {                                                          // 276
              return callback(new Error("SAML assertion was signed for the future."), null, false, xml);               // 277
            }                                                                                                          // 278
                                                                                                                       //
            var notOnOrAfter = subjectConfirmationData.$.NotOnOrAfter;                                                 // 280
                                                                                                                       //
            if (notOnOrAfter && nowMs >= Date.parse(notOnOrAfter)) {                                                   // 281
              return callback(new Error("SAML assertion was signed for the past."), null, false, xml);                 // 282
            }                                                                                                          // 283
          }                                                                                                            // 284
        }                                                                                                              // 285
      }                                                                                                                // 286
                                                                                                                       //
      var conditions = _this.getElement(assertion[0], "Conditions")[0];                                                // 288
                                                                                                                       //
      if (conditions) {                                                                                                // 289
        for (var key in meteorBabelHelpers.sanitizeForInObject(conditions)) {                                          // 290
          if (conditions.hasOwnProperty(key)) {                                                                        // 291
            var value = conditions[key];                                                                               // 292
                                                                                                                       //
            if (key === "$") {                                                                                         // 293
              var _nowMs = Date.now();                                                                                 // 294
                                                                                                                       //
              var _notBefore = value.NotBefore;                                                                        // 295
                                                                                                                       //
              if (_notBefore && _nowMs < Date.parse(_notBefore)) {                                                     // 296
                return callback(new Error("SAML condition NotBefore is in the future."), null, false, xml);            // 297
              }                                                                                                        // 299
                                                                                                                       //
              var _notOnOrAfter = value.NotOnOrAfter;                                                                  // 301
                                                                                                                       //
              if (_notOnOrAfter && _nowMs >= Date.parse(_notOnOrAfter)) {                                              // 302
                return callback(new Error("SAML condition notOnOrAfter is in the past."), null, false, xml);           // 303
              }                                                                                                        // 305
            } else if (key.endsWith("AudienceRestriction") || key.endsWith("OneTimeUse") || key.endsWith("ProxyRestriction")) {
              continue; // Do nothing.                                                                                 // 309
              // We already check both Destination and SubjectConfirmation.Recipient, so                               // 311
              // the Audience constraint isn't as important, and is tricky to handle                                   // 312
              // correctly.                                                                                            // 313
              // OneTimeUse is actually very tricky to get right, and requires a cache of                              // 314
              // previously seen assertions (or at least their hashes). In Sandstorm's case,                           // 315
              // we would probably need to stuff it in Mongo to make it robust across multiple                         // 316
              // front-ends. It is a very rarely used constraint anyways.                                              // 317
              // ProxyRestriction constraints only govern the future use of the assertion by the                       // 318
              // SP. We don't fall under this constraint so it's meaningless to us. As per the                         // 319
              // spec, it is always considered valid.                                                                  // 320
            } else {                                                                                                   // 321
              return callback(new Error("Unrecognized SAML constraint: " + key), null, false, xml);                    // 322
            }                                                                                                          // 323
          }                                                                                                            // 324
        }                                                                                                              // 325
      }                                                                                                                // 326
                                                                                                                       //
      var attributeStatement = _this.getElement(assertion[0], "AttributeStatement");                                   // 328
                                                                                                                       //
      if (attributeStatement) {                                                                                        // 329
        var attributes = _this.getElement(attributeStatement[0], "Attribute");                                         // 330
                                                                                                                       //
        if (attributes) {                                                                                              // 332
          attributes.forEach(function (attribute) {                                                                    // 333
            var value = _this.getElement(attribute, "AttributeValue");                                                 // 334
                                                                                                                       //
            if (typeof value[0] === "string") {                                                                        // 335
              profile[attribute.$.Name] = value[0];                                                                    // 336
            } else {                                                                                                   // 337
              profile[attribute.$.Name] = value[0]._;                                                                  // 338
            }                                                                                                          // 339
          });                                                                                                          // 340
        }                                                                                                              // 341
                                                                                                                       //
        if (!profile.mail && profile["urn:oid:0.9.2342.19200300.100.1.3"]) {                                           // 343
          // See http://www.incommonfederation.org/attributesummary.html for definition of attribute OIDs              // 344
          profile.mail = profile["urn:oid:0.9.2342.19200300.100.1.3"];                                                 // 345
        }                                                                                                              // 346
                                                                                                                       //
        if (!profile.email && profile.mail) {                                                                          // 348
          profile.email = profile.mail;                                                                                // 349
        }                                                                                                              // 350
                                                                                                                       //
        var microsoftEmail = profile["http://schemas.xmlsoap.org/ws/2005/05/identity/claims/emailaddress"];            // 352
                                                                                                                       //
        if (!profile.email && microsoftEmail) {                                                                        // 353
          profile.email = microsoftEmail;                                                                              // 354
        }                                                                                                              // 355
                                                                                                                       //
        var azureAdEmail = profile["http://schemas.xmlsoap.org/ws/2005/05/identity/claims/name"];                      // 357
                                                                                                                       //
        if (!profile.email && azureAdEmail) {                                                                          // 358
          profile.email = azureAdEmail;                                                                                // 359
        }                                                                                                              // 360
                                                                                                                       //
        var microsoftDisplayName = profile["http://schemas.microsoft.com/identity/claims/displayname"];                // 362
                                                                                                                       //
        if (!profile.displayName && microsoftDisplayName) {                                                            // 363
          profile.displayName = microsoftDisplayName;                                                                  // 364
        }                                                                                                              // 365
      }                                                                                                                // 366
                                                                                                                       //
      if (!profile.email && profile.nameID && profile.nameIDFormat && profile.nameIDFormat.indexOf("emailAddress") >= 0) {
        profile.email = profile.nameID;                                                                                // 369
      }                                                                                                                // 370
                                                                                                                       //
      callback(null, profile, false, xml);                                                                             // 372
    } else {                                                                                                           // 373
      var logoutResponse = _this.getElement(doc, "LogoutResponse");                                                    // 374
                                                                                                                       //
      if (logoutResponse) {                                                                                            // 376
        callback(null, null, true, xml);                                                                               // 377
      } else {                                                                                                         // 378
        return callback(new Error("Unknown SAML response message"), null, false, xml);                                 // 379
      }                                                                                                                // 380
    }                                                                                                                  // 382
  });                                                                                                                  // 384
};                                                                                                                     // 385
                                                                                                                       //
SAML.prototype.generateServiceProviderMetadata = function () {                                                         // 387
  var entityId = this.options.issuer;                                                                                  // 388
  var metadata = {                                                                                                     // 389
    "EntityDescriptor": {                                                                                              // 390
      "@xmlns": "urn:oasis:names:tc:SAML:2.0:metadata",                                                                // 391
      "@xmlns:ds": "http://www.w3.org/2000/09/xmldsig#",                                                               // 392
      "@entityID": entityId,                                                                                           // 393
      "@ID": entityId.replace(/\W/g, "_"),                                                                             // 394
      "SPSSODescriptor": {                                                                                             // 395
        "@protocolSupportEnumeration": "urn:oasis:names:tc:SAML:2.0:protocol"                                          // 396
      }                                                                                                                // 395
    }                                                                                                                  // 390
  };                                                                                                                   // 389
  var decryptionCert = this.options.localCertPublic; // TODO(someday); support local certs                             // 401
                                                                                                                       //
  if (decryptionCert) {                                                                                                // 402
    decryptionCert = decryptionCert.replace(/-+BEGIN CERTIFICATE-+\r?\n?/, "");                                        // 403
    decryptionCert = decryptionCert.replace(/-+END CERTIFICATE-+\r?\n?/, "");                                          // 404
    decryptionCert = decryptionCert.replace(/\r\n/g, "\n");                                                            // 405
    metadata.EntityDescriptor.SPSSODescriptor.KeyDescriptor = {                                                        // 407
      "ds:KeyInfo": {                                                                                                  // 408
        "ds:X509Data": {                                                                                               // 409
          "ds:X509Certificate": {                                                                                      // 410
            "#text": decryptionCert                                                                                    // 411
          }                                                                                                            // 410
        }                                                                                                              // 409
      },                                                                                                               // 408
      "EncryptionMethod": [// this should be the set that the xmlenc library supports                                  // 415
      {                                                                                                                // 417
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#aes256-cbc"                                                    // 417
      }, {                                                                                                             // 417
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#aes128-cbc"                                                    // 418
      }, {                                                                                                             // 418
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#tripledes-cbc"                                                 // 419
      }]                                                                                                               // 419
    };                                                                                                                 // 407
  }                                                                                                                    // 422
                                                                                                                       //
  metadata.EntityDescriptor.SPSSODescriptor.SingleLogoutService = {                                                    // 424
    "@Binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",                                                  // 425
    "@Location": Meteor.absoluteUrl("saml/logout/default")                                                             // 426
  };                                                                                                                   // 424
  metadata.EntityDescriptor.SPSSODescriptor.NameIDFormat = this.options.identifierFormat;                              // 429
  metadata.EntityDescriptor.SPSSODescriptor.AssertionConsumerService = {                                               // 430
    "@index": "1",                                                                                                     // 431
    "@isDefault": "true",                                                                                              // 432
    "@Binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",                                                      // 433
    "@Location": Meteor.absoluteUrl("_saml/validate/default")                                                          // 434
  };                                                                                                                   // 430
  return xmlbuilder.create(metadata).end({                                                                             // 437
    pretty: true,                                                                                                      // 437
    indent: "  ",                                                                                                      // 437
    newline: "\n"                                                                                                      // 437
  });                                                                                                                  // 437
};                                                                                                                     // 438
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"async-helpers.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/async-helpers.js                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  inMeteor: function () {                                                                                              // 1
    return inMeteor;                                                                                                   // 1
  },                                                                                                                   // 1
  promiseToFuture: function () {                                                                                       // 1
    return promiseToFuture;                                                                                            // 1
  },                                                                                                                   // 1
  waitPromise: function () {                                                                                           // 1
    return waitPromise;                                                                                                // 1
  },                                                                                                                   // 1
  onInMeteor: function () {                                                                                            // 1
    return onInMeteor;                                                                                                 // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 3
                                                                                                                       //
var inMeteorListener = undefined;                                                                                      // 5
                                                                                                                       //
var onInMeteor = function (callback) {                                                                                 // 6
  inMeteorListener = callback;                                                                                         // 7
}; // Meteor context <-> Async Node.js context adapters                                                                // 8
                                                                                                                       //
                                                                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 11
  callback();                                                                                                          // 12
});                                                                                                                    // 13
                                                                                                                       //
var inMeteor = function (callback) {                                                                                   // 15
  if (inMeteorListener) {                                                                                              // 16
    inMeteorListener();                                                                                                // 17
  } // Calls the callback in a Meteor context.  Returns a Promise for its result.                                      // 18
                                                                                                                       //
                                                                                                                       //
  return new Promise(function (resolve, reject) {                                                                      // 21
    inMeteorInternal(function () {                                                                                     // 22
      try {                                                                                                            // 23
        resolve(callback());                                                                                           // 24
      } catch (err) {                                                                                                  // 25
        reject(err);                                                                                                   // 26
      }                                                                                                                // 27
    });                                                                                                                // 28
  });                                                                                                                  // 29
};                                                                                                                     // 30
                                                                                                                       //
var promiseToFuture = function (promise) {                                                                             // 32
  var result = new Future();                                                                                           // 33
  promise.then(result.return.bind(result), result.throw.bind(result));                                                 // 34
  return result;                                                                                                       // 35
};                                                                                                                     // 36
                                                                                                                       //
var waitPromise = function (promise) {                                                                                 // 38
  return promiseToFuture(promise).wait();                                                                              // 39
};                                                                                                                     // 40
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"auth.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/auth.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  checkAuth: function () {                                                                                             // 1
    return checkAuth;                                                                                                  // 1
  },                                                                                                                   // 1
  clearAdminToken: function () {                                                                                       // 1
    return clearAdminToken;                                                                                            // 1
  },                                                                                                                   // 1
  tokenIsValid: function () {                                                                                          // 1
    return tokenIsValid;                                                                                               // 1
  },                                                                                                                   // 1
  tokenIsSetupSession: function () {                                                                                   // 1
    return tokenIsSetupSession;                                                                                        // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  Match: function (v) {                                                                                                // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  check: function (v) {                                                                                                // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.watch(require("crypto"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.watch(require("fs"), {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var SANDSTORM_VARDIR = void 0;                                                                                         // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_VARDIR: function (v) {                                                                                     // 1
    SANDSTORM_VARDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
// TODO(cleanup): globalDb is still an unbound global, but extracting it is Hard.                                      // 6
var ADMIN_TOKEN_EXPIRATION_TIME = 60 * 60 * 1000;                                                                      // 8
var SANDSTORM_ADMIN_TOKEN = SANDSTORM_VARDIR + "/adminToken";                                                          // 9
                                                                                                                       //
function clearAdminToken(token) {                                                                                      // 11
  if (tokenIsSetupSession(token)) {                                                                                    // 12
    var hash = Crypto.createHash("sha256").update(token).digest("base64");                                             // 13
    globalDb.collections.setupSession.remove({                                                                         // 14
      _id: "current-session",                                                                                          // 15
      hashedSessionId: hash                                                                                            // 16
    });                                                                                                                // 14
  }                                                                                                                    // 18
                                                                                                                       //
  if (tokenIsValid(token)) {                                                                                           // 20
    Fs.unlinkSync(SANDSTORM_ADMIN_TOKEN);                                                                              // 21
    console.log("Admin token deleted.");                                                                               // 22
  }                                                                                                                    // 23
}                                                                                                                      // 24
                                                                                                                       //
function tokenIsValid(token) {                                                                                         // 26
  if (token && Fs.existsSync(SANDSTORM_ADMIN_TOKEN)) {                                                                 // 27
    var stats = Fs.statSync(SANDSTORM_ADMIN_TOKEN);                                                                    // 28
    var expireTime = new Date(Date.now() - ADMIN_TOKEN_EXPIRATION_TIME);                                               // 29
                                                                                                                       //
    if (stats.mtime < expireTime) {                                                                                    // 30
      return false;                                                                                                    // 31
    } else {                                                                                                           // 32
      return Fs.readFileSync(SANDSTORM_ADMIN_TOKEN, {                                                                  // 33
        encoding: "utf8"                                                                                               // 33
      }) === token;                                                                                                    // 33
    }                                                                                                                  // 34
  } else {                                                                                                             // 35
    return false;                                                                                                      // 36
  }                                                                                                                    // 37
}                                                                                                                      // 38
                                                                                                                       //
function tokenIsSetupSession(token) {                                                                                  // 40
  if (token) {                                                                                                         // 41
    var setupSession = globalDb.collections.setupSession.findOne({                                                     // 42
      _id: "current-session"                                                                                           // 42
    });                                                                                                                // 42
                                                                                                                       //
    if (setupSession) {                                                                                                // 43
      var hash = Crypto.createHash("sha256").update(token).digest("base64");                                           // 44
      var now = new Date();                                                                                            // 45
      var sessionLifetime = 24 * 60 * 60 * 1000; // length of setup session validity, in milliseconds: 1 day           // 46
                                                                                                                       //
      if (setupSession.hashedSessionId === hash && now - setupSession.creationDate < sessionLifetime) {                // 47
        return true;                                                                                                   // 48
      }                                                                                                                // 49
    }                                                                                                                  // 50
  }                                                                                                                    // 51
                                                                                                                       //
  return false;                                                                                                        // 53
}                                                                                                                      // 54
                                                                                                                       //
function checkAuth(token) {                                                                                            // 56
  check(token, Match.OneOf(undefined, null, String));                                                                  // 57
                                                                                                                       //
  if (!isAdmin() && !tokenIsValid(token) && !tokenIsSetupSession(token)) {                                             // 58
    throw new Meteor.Error(403, "User must be admin or provide a valid token");                                        // 59
  }                                                                                                                    // 60
}                                                                                                                      // 61
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"backend.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/backend.js                                                                                           //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  SandstormBackend: function () {                                                                                      // 1
    return SandstormBackend;                                                                                           // 1
  },                                                                                                                   // 1
  shouldRestartGrain: function () {                                                                                    // 1
    return shouldRestartGrain;                                                                                         // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_ALTHOME: function (v) {                                                                                    // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    promiseToFuture = void 0,                                                                                          // 1
    waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  promiseToFuture: function (v) {                                                                                      // 1
    promiseToFuture = v;                                                                                               // 1
  },                                                                                                                   // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 21
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 22
                                                                                                                       //
var storageUsageUnimplemented = false;                                                                                 // 24
                                                                                                                       //
function generateSessionId(grainId, userId, packageSalt, clientSalt) {                                                 // 26
  var sessionParts = [grainId, clientSalt];                                                                            // 27
                                                                                                                       //
  if (userId) {                                                                                                        // 28
    sessionParts.push(userId);                                                                                         // 29
  }                                                                                                                    // 30
                                                                                                                       //
  if (packageSalt) {                                                                                                   // 32
    sessionParts.push(packageSalt);                                                                                    // 33
  }                                                                                                                    // 34
                                                                                                                       //
  var sessionInput = sessionParts.join(":");                                                                           // 36
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 37
}                                                                                                                      // 38
                                                                                                                       //
var shouldRestartGrain = function (error, retryCount) {                                                                // 40
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      // 41
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          // 42
  // already gone through this cycle (should be zero for the first call).                                              // 43
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 44
};                                                                                                                     // 45
                                                                                                                       //
var SandstormBackend = function () {                                                                                   //
  function SandstormBackend(db, backendCap) {                                                                          // 48
    (0, _classCallCheck3.default)(this, SandstormBackend);                                                             // 48
    this._db = db;                                                                                                     // 49
    this._backendCap = backendCap;                                                                                     // 50
  }                                                                                                                    // 51
                                                                                                                       //
  SandstormBackend.prototype.cap = function () {                                                                       //
    function cap() {                                                                                                   //
      return this._backendCap;                                                                                         // 54
    }                                                                                                                  // 55
                                                                                                                       //
    return cap;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.deleteUser = function () {                                                                //
    function deleteUser(userId) {                                                                                      //
      return waitPromise(this._backendCap.deleteUser(userId));                                                         // 58
    }                                                                                                                  // 59
                                                                                                                       //
    return deleteUser;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.shutdownGrain = function () {                                                             //
    function shutdownGrain(grainId, ownerId, keepSessions) {                                                           //
      if (!keepSessions) {                                                                                             // 62
        Sessions.remove({                                                                                              // 63
          grainId: grainId                                                                                             // 63
        });                                                                                                            // 63
      }                                                                                                                // 64
                                                                                                                       //
      var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                              // 66
                                                                                                                       //
      return grain.shutdown().then(function () {                                                                       // 67
        grain.close();                                                                                                 // 68
        throw new Error("expected shutdown() to throw disconnected");                                                  // 69
      }, function (err) {                                                                                              // 70
        grain.close();                                                                                                 // 72
                                                                                                                       //
        if (err.kjType !== "disconnected") {                                                                           // 73
          throw err;                                                                                                   // 74
        }                                                                                                              // 75
      });                                                                                                              // 76
    }                                                                                                                  // 77
                                                                                                                       //
    return shutdownGrain;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.deleteGrain = function () {                                                               //
    function deleteGrain(grainId, ownerId) {                                                                           //
      // We leave it up to the caller if they want to actually wait, but some don't so we report                       // 80
      // exceptions.                                                                                                   // 81
      return this._backendCap.deleteGrain(ownerId, grainId).catch(function (err) {                                     // 82
        console.error("problem deleting grain " + grainId + ":", err.message);                                         // 83
        throw err;                                                                                                     // 84
      });                                                                                                              // 85
    }                                                                                                                  // 86
                                                                                                                       //
    return deleteGrain;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.maybeRetryUseGrain = function () {                                                        //
    function maybeRetryUseGrain(grainId, cb, retryCount, err) {                                                        //
      var _this = this;                                                                                                // 88
                                                                                                                       //
      if (shouldRestartGrain(err, retryCount)) {                                                                       // 89
        return inMeteor(function () {                                                                                  // 90
          return cb(_this.continueGrain(grainId).supervisor).catch(_this.maybeRetryUseGrain.bind(_this, grainId, cb, retryCount + 1));
        });                                                                                                            // 93
      } else {                                                                                                         // 94
        throw err;                                                                                                     // 95
      }                                                                                                                // 96
    }                                                                                                                  // 97
                                                                                                                       //
    return maybeRetryUseGrain;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.useGrain = function () {                                                                  //
    function useGrain(grainId, cb) {                                                                                   //
      var _this2 = this;                                                                                               // 99
                                                                                                                       //
      // This will open a grain for you, handling restarts if needed, and call the passed function with                // 100
      // the supervisor capability as the only parameter. The callback must return a promise that used                 // 101
      // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                   // 102
      // This function returns the same promise that your callback returns.                                            // 103
      //                                                                                                               // 104
      // This function is NOT expected to be run in a meteor context.                                                  // 105
      return inMeteor(function () {                                                                                    // 106
        return cb(_this2.continueGrain(grainId).supervisor).catch(_this2.maybeRetryUseGrain.bind(_this2, grainId, cb, 0));
      });                                                                                                              // 109
    }                                                                                                                  // 110
                                                                                                                       //
    return useGrain;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.continueGrain = function () {                                                             //
    function continueGrain(grainId) {                                                                                  //
      var grain = Grains.findOne(grainId);                                                                             // 113
                                                                                                                       //
      if (!grain) {                                                                                                    // 114
        throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                        // 115
      }                                                                                                                // 116
                                                                                                                       //
      if (grain.trashed) {                                                                                             // 118
        throw new Meteor.Error(403, "Grain is in the trash bin", "Grain ID: " + grainId);                              // 119
      } // If a DevPackage with the same app ID is currently active, we let it override the installed                  // 120
      // package, so that the grain runs using the dev app.                                                            // 123
                                                                                                                       //
                                                                                                                       //
      var devPackage = DevPackages.findOne({                                                                           // 124
        appId: grain.appId                                                                                             // 124
      });                                                                                                              // 124
      var isDev = void 0;                                                                                              // 125
      var mountProc = void 0;                                                                                          // 126
      var pkg = void 0;                                                                                                // 127
                                                                                                                       //
      if (devPackage) {                                                                                                // 128
        isDev = true;                                                                                                  // 129
        pkg = devPackage;                                                                                              // 130
        mountProc = pkg.mountProc;                                                                                     // 131
      } else {                                                                                                         // 132
        pkg = Packages.findOne(grain.packageId);                                                                       // 133
                                                                                                                       //
        if (!pkg) {                                                                                                    // 134
          throw new Meteor.Error(500, "Grain's package not installed", "Package ID: " + grain.packageId);              // 135
        }                                                                                                              // 137
      }                                                                                                                // 138
                                                                                                                       //
      var manifest = pkg.manifest;                                                                                     // 140
      var packageId = pkg._id;                                                                                         // 141
                                                                                                                       //
      if (!("continueCommand" in manifest)) {                                                                          // 143
        throw new Meteor.Error(500, "Package manifest defines no continueCommand.", "Package ID: " + packageId);       // 144
      }                                                                                                                // 146
                                                                                                                       //
      var result = this.startGrainInternal(packageId, grainId, grain.userId, manifest.continueCommand, false, isDev, mountProc);
      result.packageSalt = isDev ? pkg._id : grain.packageSalt;                                                        // 150
      return result;                                                                                                   // 151
    }                                                                                                                  // 152
                                                                                                                       //
    return continueGrain;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.startGrainInternal = function () {                                                        //
    function startGrainInternal(packageId, grainId, ownerId, command, isNew, isDev, mountProc) {                       //
      // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                    // 155
      // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                 // 156
      // user) and `supervisor` (the supervisor capability).                                                           // 157
      if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                 // 159
        throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.\n" + "Please ask them to upgrade.");
      } // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather              // 163
      //   than just "argv".                                                                                           // 166
                                                                                                                       //
                                                                                                                       //
      if ("args" in command) {                                                                                         // 167
        if (!("argv" in command)) {                                                                                    // 168
          command.argv = command.args;                                                                                 // 169
        }                                                                                                              // 170
                                                                                                                       //
        delete command.args;                                                                                           // 172
      }                                                                                                                // 173
                                                                                                                       //
      if ("executablePath" in command) {                                                                               // 175
        if (!("deprecatedExecutablePath" in command)) {                                                                // 176
          command.deprecatedExecutablePath = command.executablePath;                                                   // 177
        }                                                                                                              // 178
                                                                                                                       //
        delete command.executablePath;                                                                                 // 180
      }                                                                                                                // 181
                                                                                                                       //
      return this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev, mountProc);               // 183
    }                                                                                                                  // 184
                                                                                                                       //
    return startGrainInternal;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.updateLastActive = function () {                                                          //
    function updateLastActive(grainId, userId, obsolete) {                                                             //
      // Update the lastActive date on the grain, any relevant API tokens, and the user,                               // 187
      // and also update the user's storage usage.                                                                     // 188
      var storagePromise = undefined;                                                                                  // 190
      var ownerId = undefined;                                                                                         // 191
                                                                                                                       //
      if (this._db.isQuotaEnabled() && !storageUsageUnimplemented) {                                                   // 192
        ownerId = Grains.findOne(grainId).userId;                                                                      // 193
        storagePromise = this._backendCap.getUserStorageUsage(ownerId);                                                // 194
      }                                                                                                                // 195
                                                                                                                       //
      var now = new Date();                                                                                            // 197
                                                                                                                       //
      if (Grains.update(grainId, {                                                                                     // 198
        $set: {                                                                                                        // 198
          lastUsed: now                                                                                                // 198
        }                                                                                                              // 198
      }) === 0) {                                                                                                      // 198
        // Grain must have been deleted. Ignore.                                                                       // 199
        return;                                                                                                        // 200
      }                                                                                                                // 201
                                                                                                                       //
      if (userId) {                                                                                                    // 203
        Meteor.users.update(userId, {                                                                                  // 204
          $set: {                                                                                                      // 204
            lastActive: now                                                                                            // 204
          }                                                                                                            // 204
        }); // Update any API tokens that match this user/grain pairing as well                                        // 204
                                                                                                                       //
        ApiTokens.update({                                                                                             // 207
          grainId: grainId,                                                                                            // 207
          "owner.user.accountId": userId                                                                               // 207
        }, {                                                                                                           // 207
          $set: {                                                                                                      // 208
            lastUsed: now                                                                                              // 208
          }                                                                                                            // 208
        }, {                                                                                                           // 208
          multi: true                                                                                                  // 209
        });                                                                                                            // 209
      }                                                                                                                // 210
                                                                                                                       //
      if (storagePromise) {                                                                                            // 212
        try {                                                                                                          // 213
          var size = parseInt(waitPromise(storagePromise).size);                                                       // 214
          Meteor.users.update(ownerId, {                                                                               // 215
            $set: {                                                                                                    // 215
              storageUsage: size                                                                                       // 215
            }                                                                                                          // 215
          }); // TODO(security): Consider actively killing grains if the user is excessively over quota?               // 215
          //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                // 217
        } catch (err) {                                                                                                // 218
          if (err.kjType === "unimplemented") {                                                                        // 219
            storageUsageUnimplemented = true;                                                                          // 220
          } else {                                                                                                     // 221
            console.error("error getting user storage usage:", err.stack);                                             // 222
          }                                                                                                            // 223
        }                                                                                                              // 224
      }                                                                                                                // 225
    }                                                                                                                  // 226
                                                                                                                       //
    return updateLastActive;                                                                                           //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.openSessionInternal = function () {                                                       //
    function openSessionInternal(grain, userId, revealIdentity, title, apiToken, cachedSalt, sessionFields) {          //
      var grainId = grain._id; // Start the grain if it is not running. This is an optimization: if we didn't start it here,
      // it would start on the first request to the session host, but we'd like to get started before                  // 232
      // the round trip.                                                                                               // 233
                                                                                                                       //
      var _continueGrain = this.continueGrain(grainId),                                                                // 228
          supervisor = _continueGrain.supervisor,                                                                      // 228
          packageSalt = _continueGrain.packageSalt;                                                                    // 228
                                                                                                                       //
      this.updateLastActive(grainId, userId);                                                                          // 236
                                                                                                                       //
      var identityId = userId && revealIdentity && this._db.getOrGenerateIdentityId(userId, grain);                    // 238
                                                                                                                       //
      cachedSalt = cachedSalt || Random.id(22);                                                                        // 240
      var sessionId = generateSessionId(grainId, userId, packageSalt, cachedSalt);                                     // 241
      var session = Sessions.findOne({                                                                                 // 242
        _id: sessionId                                                                                                 // 242
      });                                                                                                              // 242
                                                                                                                       //
      if (session) {                                                                                                   // 243
        // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                 // 244
        if (session.userId && session.userId !== userId || session.identityId && session.identityId !== identityId || session.grainId !== grainId) {
          var e = new Meteor.Error(500, "Duplicate SessionId");                                                        // 248
          console.error(e);                                                                                            // 249
          throw e;                                                                                                     // 250
        } else {                                                                                                       // 251
          return {                                                                                                     // 252
            supervisor: supervisor,                                                                                    // 253
            methodResult: {                                                                                            // 254
              sessionId: session._id,                                                                                  // 255
              title: title,                                                                                            // 256
              grainId: grainId,                                                                                        // 257
              hostId: session.hostId,                                                                                  // 258
              tabId: session.tabId,                                                                                    // 259
              salt: cachedSalt                                                                                         // 260
            }                                                                                                          // 254
          };                                                                                                           // 252
        }                                                                                                              // 263
      } // TODO(someday): Allow caller to specify the parent session from which to inherit the tab ID, or              // 264
      //   something.                                                                                                  // 267
                                                                                                                       //
                                                                                                                       //
      session = {                                                                                                      // 269
        _id: sessionId,                                                                                                // 270
        grainId: grainId,                                                                                              // 271
        hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                              // 272
        tabId: Crypto.createHash("sha256").update("tab:").update(sessionId).digest("hex").slice(0, 32),                // 273
        timestamp: new Date().getTime(),                                                                               // 274
        hasLoaded: false                                                                                               // 275
      };                                                                                                               // 269
                                                                                                                       //
      if (userId) {                                                                                                    // 278
        session.userId = userId;                                                                                       // 279
      }                                                                                                                // 280
                                                                                                                       //
      if (identityId) {                                                                                                // 282
        session.identityId = identityId;                                                                               // 283
      }                                                                                                                // 284
                                                                                                                       //
      if (apiToken) {                                                                                                  // 286
        session.hashedToken = apiToken._id;                                                                            // 287
      }                                                                                                                // 288
                                                                                                                       //
      if (sessionFields) {                                                                                             // 290
        _.extend(session, sessionFields);                                                                              // 291
      }                                                                                                                // 292
                                                                                                                       //
      Sessions.insert(session);                                                                                        // 294
      return {                                                                                                         // 296
        supervisor: supervisor,                                                                                        // 297
        methodResult: {                                                                                                // 298
          sessionId: session._id,                                                                                      // 299
          title: title,                                                                                                // 300
          grainId: grainId,                                                                                            // 301
          hostId: session.hostId,                                                                                      // 302
          tabId: session.tabId,                                                                                        // 303
          salt: cachedSalt,                                                                                            // 304
          identityId: identityId                                                                                       // 305
        }                                                                                                              // 298
      };                                                                                                               // 296
    }                                                                                                                  // 308
                                                                                                                       //
    return openSessionInternal;                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormBackend;                                                                                             //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 309
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"constants.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/constants.js                                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  SANDSTORM_ALTHOME: function () {                                                                                     // 1
    return SANDSTORM_ALTHOME;                                                                                          // 1
  },                                                                                                                   // 1
  SANDSTORM_LOGDIR: function () {                                                                                      // 1
    return SANDSTORM_LOGDIR;                                                                                           // 1
  },                                                                                                                   // 1
  SANDSTORM_VARDIR: function () {                                                                                      // 1
    return SANDSTORM_VARDIR;                                                                                           // 1
  },                                                                                                                   // 1
  staticAssetHost: function () {                                                                                       // 1
    return staticAssetHost;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module.watch(require("url"), {                                                                                         // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                       // 22
var SANDSTORM_LOGDIR = (SANDSTORM_ALTHOME || "") + "/var/log";                                                         // 23
var SANDSTORM_VARDIR = (SANDSTORM_ALTHOME || "") + "/var/sandstorm"; // TODO: dedupe this and the equivalent copy in sandstorm-db
                                                                                                                       //
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                                       // 27
var staticAssetHost = Url.parse(process.env.ROOT_URL).protocol + "//" + wildcardHost[0] + "static" + wildcardHost[1];  // 29
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"desktop-notifications.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/desktop-notifications.js                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  createAppActivityDesktopNotification: function () {                                                                  // 1
    return createAppActivityDesktopNotification;                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  Match: function (v) {                                                                                                // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  check: function (v) {                                                                                                // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var createAppActivityDesktopNotification = function (options) {                                                        // 19
  check(options, {                                                                                                     // 20
    userId: String,                                                                                                    // 21
    notificationId: String,                                                                                            // 22
    appActivity: {                                                                                                     // 23
      user: Match.OneOf(undefined, {                                                                                   // 24
        accountId: String,                                                                                             // 27
        name: String,                                                                                                  // 28
        avatarUrl: String                                                                                              // 29
      }, {                                                                                                             // 26
        anonymous: true                                                                                                // 31
      }),                                                                                                              // 31
      grainId: String,                                                                                                 // 33
      path: String,                                                                                                    // 34
      body: Match.ObjectIncluding({                                                                                    // 35
        defaultText: String                                                                                            // 36
      }),                                                                                                              // 35
      actionText: Match.ObjectIncluding({                                                                              // 38
        defaultText: String                                                                                            // 39
      })                                                                                                               // 38
    }                                                                                                                  // 23
  });                                                                                                                  // 20
  globalDb.collections.desktopNotifications.insert({                                                                   // 44
    userId: options.userId,                                                                                            // 45
    notificationId: options.notificationId,                                                                            // 46
    creationDate: new Date(),                                                                                          // 47
    appActivity: options.appActivity                                                                                   // 48
  });                                                                                                                  // 44
};                                                                                                                     // 50
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"email.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/email.js                                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  send: function () {                                                                                                  // 1
    return send;                                                                                                       // 1
  },                                                                                                                   // 1
  rawSend: function () {                                                                                               // 1
    return rawSend;                                                                                                    // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var nodemailer = void 0;                                                                                               // 1
module.watch(require("nodemailer"), {                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    nodemailer = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var smtpPool = void 0;                                                                                                 // 1
module.watch(require("nodemailer-smtp-pool"), {                                                                        // 1
  "default": function (v) {                                                                                            // 1
    smtpPool = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 5
                                                                                                                       //
var getSmtpConfig = function () {                                                                                      // 7
  var config = Settings.findOne({                                                                                      // 8
    _id: "smtpConfig"                                                                                                  // 8
  });                                                                                                                  // 8
  return config && config.value;                                                                                       // 9
};                                                                                                                     // 10
                                                                                                                       //
var makePool = function (mailConfig) {                                                                                 // 12
  if (!mailConfig.hostname) {                                                                                          // 13
    throw new Error("This Sandstorm server has not been configured to send email.");                                   // 14
  }                                                                                                                    // 15
                                                                                                                       //
  var auth = false;                                                                                                    // 17
                                                                                                                       //
  if (mailConfig.auth && (mailConfig.auth.user || mailConfig.auth.pass)) {                                             // 18
    auth = mailConfig.auth;                                                                                            // 19
  }                                                                                                                    // 20
                                                                                                                       //
  var secure = mailConfig.port === 465;                                                                                // 22
  var tlsOptions = {                                                                                                   // 23
    // Previously, node 0.10 did not attempt to validate certificates received when connecting                         // 24
    // with STARTTLS, so to avoid regressing we need to preserve that behavior here for now.                           // 25
    rejectUnauthorized: false                                                                                          // 26
  };                                                                                                                   // 23
  var pool = nodemailer.createTransport(smtpPool({                                                                     // 29
    host: mailConfig.hostname,                                                                                         // 30
    port: mailConfig.port,                                                                                             // 31
    secure: secure,                                                                                                    // 32
    tls: tlsOptions,                                                                                                   // 33
    auth: auth // TODO(someday): allow maxConnections to be configured?                                                // 34
                                                                                                                       //
  }));                                                                                                                 // 29
  pool._futureWrappedSendMail = _.bind(Future.wrap(pool.sendMail), pool);                                              // 38
  return pool;                                                                                                         // 39
}; // We construct the SMTP pool at the first call to Email.send, so that                                              // 40
// other code like migrations can modify the SMTP configuration.                                                       // 43
                                                                                                                       //
                                                                                                                       //
var pool = void 0;                                                                                                     // 44
var configured = false;                                                                                                // 45
Meteor.startup(function () {                                                                                           // 47
  Settings.find({                                                                                                      // 48
    _id: "smtpConfig"                                                                                                  // 48
  }).observeChanges({                                                                                                  // 48
    removed: function () {                                                                                             // 49
      configured = false;                                                                                              // 50
    },                                                                                                                 // 51
    changed: function () {                                                                                             // 53
      configured = false;                                                                                              // 54
    },                                                                                                                 // 55
    added: function () {                                                                                               // 57
      configured = false;                                                                                              // 58
    }                                                                                                                  // 59
  });                                                                                                                  // 48
});                                                                                                                    // 61
                                                                                                                       //
var getPool = function (smtpConfig) {                                                                                  // 63
  if (smtpConfig) {                                                                                                    // 64
    return makePool(smtpConfig);                                                                                       // 65
  } else if (!configured) {                                                                                            // 66
    configured = true;                                                                                                 // 67
    var config = getSmtpConfig();                                                                                      // 68
                                                                                                                       //
    if (config) {                                                                                                      // 69
      pool = makePool(config);                                                                                         // 70
    }                                                                                                                  // 71
  }                                                                                                                    // 72
                                                                                                                       //
  return pool;                                                                                                         // 74
};                                                                                                                     // 75
                                                                                                                       //
var smtpSend = function (pool, mailOptions) {                                                                          // 77
  pool._futureWrappedSendMail(mailOptions).wait();                                                                     // 78
}; // From http://emailregex.com/, which claims this is the W3C standard for the HTML input element,                   // 79
// although their link is broken and I can find no evidence that this is a standard. The page                          // 82
// lists several regexes, ostensibly in syntaxes intended for different programming languages,                         // 83
// but each regex is in fact substantially different for no apparent reason.                                           // 84
//                                                                                                                     // 85
// The most important thing here is that we disallow separators that might allow a user to confuse                     // 86
// nodemailer into thinking the address is a list. Unfortunately, nodemailer will happily separate                     // 87
// strings into lists splitting on all kinds of separator characters, such as commas, semicolons,                      // 88
// etc. This regex should accomplish that both by disallowing the separators, and by disallowing                       // 89
// multiple @ signs. The rest is for show.                                                                             // 90
                                                                                                                       //
                                                                                                                       //
var EMAIL_REGEX = /^[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$/;                              // 91
                                                                                                                       //
function validateEmail(email) {                                                                                        // 93
  if (email instanceof Array) {                                                                                        // 94
    email.forEach(validateEmail);                                                                                      // 95
  } else if ((typeof email === "undefined" ? "undefined" : (0, _typeof3.default)(email)) === "object" && "address" in email) {
    validateEmail(email.address);                                                                                      // 97
  } else if (email) {                                                                                                  // 98
    check(email, String);                                                                                              // 99
                                                                                                                       //
    if (!email.match(EMAIL_REGEX)) {                                                                                   // 101
      throw new Meteor.Error(400, "invalid e-mail address");                                                           // 102
    }                                                                                                                  // 103
  }                                                                                                                    // 104
}                                                                                                                      // 105
                                                                                                                       //
var rawSend = function (mailOptions, smtpConfig) {                                                                     // 107
  // Sends an email mailOptions object structured as described in                                                      // 108
  // https://github.com/nodemailer/mailcomposer#e-mail-message-fields                                                  // 109
  // across the transport described by smtpConfig.                                                                     // 110
  // For fields that are supposed to be lists of addresses, if only a single string is provided,                       // 112
  // wrap it in an array. This prevents nodemailer from interpreting the address as a                                  // 113
  // comma-separated list.                                                                                             // 114
  ["from", "to", "cc", "bcc", "replyTo"].forEach(function (field) {                                                    // 115
    validateEmail(mailOptions[field]);                                                                                 // 116
  });                                                                                                                  // 117
  var pool = getPool(smtpConfig);                                                                                      // 119
                                                                                                                       //
  if (pool) {                                                                                                          // 120
    smtpSend(pool, mailOptions);                                                                                       // 121
  } else {                                                                                                             // 122
    throw new Error("SMTP pool is misconfigured.");                                                                    // 123
  }                                                                                                                    // 124
}; // Old comment below                                                                                                // 125
/**                                                                                                                    // 128
 * Send an email.                                                                                                      //
 *                                                                                                                     //
 * Connects to the mail server configured via the MAIL_URL environment                                                 //
 * variable. If unset, prints formatted message to stdout. The "from" option                                           //
 * is required, and at least one of "to", "cc", and "bcc" must be provided;                                            //
 * all other options are optional.                                                                                     //
 *                                                                                                                     //
 * @param options                                                                                                      //
 * @param options.from {String} RFC5322 "From:" address                                                                //
 * @param options.to {String|String[]} RFC5322 "To:" address[es]                                                       //
 * @param options.cc {String|String[]} RFC5322 "Cc:" address[es]                                                       //
 * @param options.bcc {String|String[]} RFC5322 "Bcc:" address[es]                                                     //
 * @param options.replyTo {String|String[]} RFC5322 "Reply-To:" address[es]                                            //
 * @param options.subject {String} RFC5322 "Subject:" line                                                             //
 * @param options.text {String} RFC5322 mail body (plain text)                                                         //
 * @param options.html {String} RFC5322 mail body (HTML)                                                               //
 * @param options.headers {Object} custom RFC5322 headers (dictionary)                                                 //
 */ // New API doc comment below                                                                                       //
/**                                                                                                                    // 149
 * @summary Send an email. Throws an `Error` on failure to contact mail server                                         //
 * or if mail server returns an error. All fields should match                                                         //
 * [RFC5322](http://tools.ietf.org/html/rfc5322) specification.                                                        //
 * @locus Server                                                                                                       //
 * @param {Object} options                                                                                             //
 * @param {String} options.from "From:" address (required)                                                             //
 * @param {String|String[]} options.to,cc,bcc,replyTo                                                                  //
 *   "To:", "Cc:", "Bcc:", and "Reply-To:" addresses                                                                   //
 * @param {String} [options.subject]  "Subject:" line                                                                  //
 * @param {String} [options.text|html] Mail body (in plain text or HTML)                                               //
 * @param {Object} [options.headers] Dictionary of custom headers                                                      //
 * @param {Object} [options.smtpConfig] SMTP server to use. Otherwise defaults to configured one.                      //
 * @param {String} [options.smtpConfig.hostname] SMTP server hostname.                                                 //
 * @param {Number} [options.smtpConfig.port] SMTP server port.                                                         //
 * @param {Object} [options.smtpConfig.auth] SMTP server authentication tokens.  Optional.                             //
 * @param {String} [options.smtpConfig.auth.user] Username of user to log in to SMTP server as.  Optional.             //
 * @param {String} [options.smtpConfig.auth.pass] Password of user to log in to SMTP server as.  Optional.             //
 * @param {Object} [options.attachments] Attachments. See:                                                             //
 *   https://github.com/nodemailer/mailcomposer/tree/v0.1.15#add-attachments                                           //
 * @param {String} [options.envelopeFrom] Envelope sender.                                                             //
 */                                                                                                                    //
                                                                                                                       //
var send = function (options) {                                                                                        // 171
  // Unpack options                                                                                                    // 172
  var from = options.from,                                                                                             // 171
      to = options.to,                                                                                                 // 171
      cc = options.cc,                                                                                                 // 171
      bcc = options.bcc,                                                                                               // 171
      replyTo = options.replyTo,                                                                                       // 171
      subject = options.subject,                                                                                       // 171
      text = options.text,                                                                                             // 171
      html = options.html,                                                                                             // 171
      envelopeFrom = options.envelopeFrom,                                                                             // 171
      headers = options.headers,                                                                                       // 171
      attachments = options.attachments,                                                                               // 171
      smtpConfig = options.smtpConfig;                                                                                 // 171
  var opts = {                                                                                                         // 188
    from: from,                                                                                                        // 189
    to: to,                                                                                                            // 190
    cc: cc,                                                                                                            // 191
    bcc: bcc,                                                                                                          // 192
    replyTo: replyTo,                                                                                                  // 193
    subject: subject,                                                                                                  // 194
    text: text,                                                                                                        // 195
    html: html,                                                                                                        // 196
    headers: headers,                                                                                                  // 197
    attachments: attachments                                                                                           // 198
  };                                                                                                                   // 188
                                                                                                                       //
  if (envelopeFrom) {                                                                                                  // 201
    opts.envelope = {                                                                                                  // 202
      from: envelopeFrom,                                                                                              // 203
      to: to,                                                                                                          // 204
      cc: cc,                                                                                                          // 205
      bcc: bcc                                                                                                         // 206
    };                                                                                                                 // 202
  }                                                                                                                    // 208
                                                                                                                       //
  rawSend(opts, smtpConfig);                                                                                           // 210
};                                                                                                                     // 211
                                                                                                                       //
// TODO(cleanup): Remove this once BlackrockPayments code finds a better way to import it.                             // 215
global.SandstormEmail = {                                                                                              // 216
  send: send                                                                                                           // 216
};                                                                                                                     // 216
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"frontend-ref.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/frontend-ref.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  FrontendRefRegistry: function () {                                                                                   // 1
    return FrontendRefRegistry;                                                                                        // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var checkRequirements = void 0;                                                                                        // 1
module.watch(require("./persistent.js"), {                                                                             // 1
  checkRequirements: function (v) {                                                                                    // 1
    checkRequirements = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var FrontendRefRegistry = function () {                                                                                //
  function FrontendRefRegistry() {                                                                                     // 20
    (0, _classCallCheck3.default)(this, FrontendRefRegistry);                                                          // 20
    this._frontendRefHandlers = {};                                                                                    // 21
    this._typeIdHandlers = {};                                                                                         // 22
  }                                                                                                                    // 23
                                                                                                                       //
  FrontendRefRegistry.prototype.create = function () {                                                                 //
    function create(db, frontendRef, requirements) {                                                                   //
      // Construct a fresh capability based on a `frontendRef` value.                                                  // 26
      checkRequirements(requirements);                                                                                 // 28
      var saveTemplate = {                                                                                             // 30
        frontendRef: frontendRef                                                                                       // 30
      };                                                                                                               // 30
                                                                                                                       //
      if (requirements) {                                                                                              // 31
        saveTemplate.requirements = requirements;                                                                      // 32
      }                                                                                                                // 33
                                                                                                                       //
      return this.restore(db, saveTemplate, frontendRef);                                                              // 35
    }                                                                                                                  // 36
                                                                                                                       //
    return create;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.restore = function () {                                                                //
    function restore(db, saveTemplate, frontendRef) {                                                                  //
      // Restores a frontendRef capability using the appropriate registered handler.                                   // 39
      var keys = Object.keys(frontendRef);                                                                             // 41
                                                                                                                       //
      if (keys.length != 1) {                                                                                          // 42
        throw new Error("invalid frontendRef: " + JSON.stringify(frontendRef));                                        // 43
      }                                                                                                                // 44
                                                                                                                       //
      var key = keys[0];                                                                                               // 46
      var handler = this._frontendRefHandlers[key];                                                                    // 47
                                                                                                                       //
      if (!handler) {                                                                                                  // 48
        throw new Error("invalid frontendRef: " + JSON.stringify(frontendRef));                                        // 49
      }                                                                                                                // 50
                                                                                                                       //
      return handler.restore(db, saveTemplate, frontendRef[key]);                                                      // 52
    }                                                                                                                  // 53
                                                                                                                       //
    return restore;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.query = function () {                                                                  //
    function query(db, userAccountId, tag) {                                                                           //
      // Performs a powerbox query using the appropriate registered handler.                                           // 56
      var handler = this._typeIdHandlers[tag.id];                                                                      // 58
                                                                                                                       //
      if (handler) {                                                                                                   // 59
        return handler.query(db, userAccountId, tag.value);                                                            // 60
      } else {                                                                                                         // 61
        return []; // no matches                                                                                       // 62
      }                                                                                                                // 63
    }                                                                                                                  // 64
                                                                                                                       //
    return query;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.validate = function () {                                                               //
    function validate(db, session, frontendRefRequest) {                                                               //
      // Validates a powerbox request on the given `session` (a record from the Sessions table)                        // 67
      // requesting the creation of the given frontendRef type. `frontendRefRequest` is expected to                    // 68
      // come directly from the client; calling validate() verifies that it is well-formed and                         // 69
      // carries out any type-specific server-side work needed to construct the desired frontendRef.                   // 70
      // `validate()` also returns an array of MembraneRequirements that shall apply to the                            // 71
      // capability. These requirements will not yet have been checked, so the caller should check                     // 72
      // them before proceeding.                                                                                       // 73
      var keys = Object.keys(frontendRefRequest);                                                                      // 75
                                                                                                                       //
      if (keys.length != 1) {                                                                                          // 76
        throw new Error("invalid frontendRefRequest: " + JSON.stringify(frontendRefRequest));                          // 77
      }                                                                                                                // 78
                                                                                                                       //
      var key = keys[0];                                                                                               // 80
      var handler = this._frontendRefHandlers[key];                                                                    // 81
                                                                                                                       //
      if (!handler) {                                                                                                  // 82
        throw new Error("invalid frontendRefRequest: " + JSON.stringify(frontendRefRequest));                          // 83
      }                                                                                                                // 84
                                                                                                                       //
      if (!handler.validate) {                                                                                         // 86
        throw new Error("frontendRef type cannot be created via powerbox: " + JSON.stringify(frontendRefRequest));     // 87
      }                                                                                                                // 89
                                                                                                                       //
      var _handler$validate = handler.validate(db, session, frontendRefRequest[key]),                                  // 66
          descriptor = _handler$validate.descriptor,                                                                   // 66
          requirements = _handler$validate.requirements,                                                               // 66
          frontendRef = _handler$validate.frontendRef;                                                                 // 66
                                                                                                                       //
      var result = {                                                                                                   // 94
        descriptor: descriptor,                                                                                        // 94
        requirements: requirements,                                                                                    // 94
        frontendRef: {}                                                                                                // 94
      };                                                                                                               // 94
      result.frontendRef[key] = frontendRef;                                                                           // 95
      return result;                                                                                                   // 96
    }                                                                                                                  // 97
                                                                                                                       //
    return validate;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.register = function () {                                                               //
    function register(object) {                                                                                        //
      // Register callbacks related to a particular frontendRef type. The object has the fields:                       // 100
      //   `frontendRefField`: Name of the field of `ApiTokens.frontendRef` that is filled in for                      // 101
      //       this type. Only needed if `create` and/or `validate` handlers are defined.                              // 102
      //   `typeId`: Type ID of powerbox tags handled by the `query` callback. Stringified decimal                     // 103
      //       64-bit integer. Only needed if `query` is defined.                                                      // 104
      //   `restore`: Callback to construct a capability of this type when restoring a saved                           // 105
      //       capability. Has signature `(db, saveTemplate, value) -> capability`, where:                             // 106
      //     `value`: The value of the single field of `ApiTokens.frontendRef` for this capability.                    // 107
      //     `saveTemplate`: The token template to pass to the PersistentImpl constructor.                             // 108
      //     `capability` (returned): A Cap'n Proto capability implementing SystemPersistent along                     // 109
      //         with whatever other interfaces are appropriate for the ref type.                                      // 110
      //   `validate`: Callback to validate a powerbox request for a new capability of this type.                      // 111
      //       Has signature `(db, session, request) -> {descriptor, requirements, frontendRef}`,                      // 112
      //       where:                                                                                                  // 113
      //     `request` is type-specific information describing the requested capability. The                           // 114
      //         callback *must* type-check this value, and should throw an exception if it is                         // 115
      //         not valid.                                                                                            // 116
      //     `session` is the record from the Sessions table of the UI session where the powerbox                      // 117
      //         request occurred.                                                                                     // 118
      //     `descriptor` (returned) is the JSON-encoded PowerboxDescriptor for the capability. Note                   // 119
      //         that if the descriptor contains any `tag.value`s, they of course need to be                           // 120
      //         presented as capnp-encoded Buffers.                                                                   // 121
      //     `requirements` (returned) is an array of MembraneRequirements which should apply to the                   // 122
      //         new capability. Note that these requirements will be checked immediately and the                      // 123
      //         powerbox request will fail if they aren't met.                                                        // 124
      //     `frontendRef` (returned) is the value that will be written for the key specified                          // 125
      //         by `frontendRefField` in the single-key object `ApiTokens.frontendRef`.                               // 126
      //    `query`: Callback to populate options for a powerbox request for this type ID. Has                         // 127
      //        signature `(db, userAccountId, tagValue) -> options`, where:                                           // 128
      //      `tagValue`: A Buffer of the Cap'n-Proto-encoded `PowerboxDescriptor.Tag.value`.                          // 129
      //      `options` (returned): An array of objects representing the options that should be                        // 130
      //          offered to the user for this query. See the `powerboxOptions` Meteor publish in                      // 131
      //          powerbox-server.js for a full description of the fields of each option.                              // 132
      if (object.frontendRefField) {                                                                                   // 134
        if (object.frontendRefField in this._frontendRefHandlers) {                                                    // 135
          throw new Error("frontendRef handler already registered: " + object.frontendRefField);                       // 136
        }                                                                                                              // 137
                                                                                                                       //
        this._frontendRefHandlers[object.frontendRefField] = object;                                                   // 139
      }                                                                                                                // 140
                                                                                                                       //
      if (object.typeId) {                                                                                             // 142
        if (object.typeId in this._typeIdHandlers) {                                                                   // 143
          throw new Error("typeId handler already registered: " + object.typeId);                                      // 144
        }                                                                                                              // 145
                                                                                                                       //
        this._typeIdHandlers[object.typeId] = object;                                                                  // 147
      }                                                                                                                // 148
    }                                                                                                                  // 149
                                                                                                                       //
    return register;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  return FrontendRefRegistry;                                                                                          //
}();                                                                                                                   //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"header-whitelist.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/header-whitelist.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  HeaderWhitelist: function () {                                                                                       // 1
    return HeaderWhitelist;                                                                                            // 1
  },                                                                                                                   // 1
  REQUEST_HEADER_WHITELIST: function () {                                                                              // 1
    return REQUEST_HEADER_WHITELIST;                                                                                   // 1
  },                                                                                                                   // 1
  RESPONSE_HEADER_WHITELIST: function () {                                                                             // 1
    return RESPONSE_HEADER_WHITELIST;                                                                                  // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2017 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Capnp = Npm.require("capnp");                                                                                      // 17
                                                                                                                       //
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;                                         // 18
                                                                                                                       //
var HeaderWhitelist = function () {                                                                                    //
  function HeaderWhitelist(list) {                                                                                     // 21
    var _this = this;                                                                                                  // 21
                                                                                                                       //
    (0, _classCallCheck3.default)(this, HeaderWhitelist);                                                              // 21
    this._headers = {};                                                                                                // 22
    this._prefixes = [];                                                                                               // 23
    list.forEach(function (pattern) {                                                                                  // 25
      if (pattern.endsWith("*")) {                                                                                     // 26
        _this._prefixes.push(pattern.slice(0, -1));                                                                    // 27
      } else {                                                                                                         // 28
        _this._headers[pattern] = true;                                                                                // 29
      }                                                                                                                // 30
    });                                                                                                                // 31
  }                                                                                                                    // 32
                                                                                                                       //
  HeaderWhitelist.prototype.matches = function () {                                                                    //
    function matches(header) {                                                                                         //
      header = header.toLowerCase();                                                                                   // 35
      if (this._headers[header]) return true;                                                                          // 36
                                                                                                                       //
      for (var i in meteorBabelHelpers.sanitizeForInObject(this._prefixes)) {                                          // 38
        if (header.startsWith(this._prefixes[i])) {                                                                    // 39
          return true;                                                                                                 // 40
        }                                                                                                              // 41
      }                                                                                                                // 42
                                                                                                                       //
      ;                                                                                                                // 42
      return false;                                                                                                    // 44
    }                                                                                                                  // 45
                                                                                                                       //
    return matches;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  return HeaderWhitelist;                                                                                              //
}();                                                                                                                   //
                                                                                                                       //
var REQUEST_HEADER_WHITELIST = new HeaderWhitelist(WebSession.Context.headerWhitelist);                                // 48
var RESPONSE_HEADER_WHITELIST = new HeaderWhitelist(WebSession.Response.headerWhitelist);                              // 49
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"migrations.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/migrations.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  migrateToLatest: function () {                                                                                       // 1
    return migrateToLatest;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var _ = void 0;                                                                                                        // 1
                                                                                                                       //
module.watch(require("meteor/underscore"), {                                                                           // 1
  _: function (v) {                                                                                                    // 1
    _ = v;                                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Match = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  Match: function (v) {                                                                                                // 1
    Match = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var userPictureUrl = void 0,                                                                                           // 1
    fetchPicture = void 0;                                                                                             // 1
module.watch(require("/imports/server/accounts/picture.js"), {                                                         // 1
  userPictureUrl: function (v) {                                                                                       // 1
    userPictureUrl = v;                                                                                                // 1
  },                                                                                                                   // 1
  fetchPicture: function (v) {                                                                                         // 1
    fetchPicture = v;                                                                                                  // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var PRIVATE_IPV4_ADDRESSES = void 0,                                                                                   // 1
    PRIVATE_IPV6_ADDRESSES = void 0;                                                                                   // 1
module.watch(require("/imports/constants.js"), {                                                                       // 1
  PRIVATE_IPV4_ADDRESSES: function (v) {                                                                               // 1
    PRIVATE_IPV4_ADDRESSES = v;                                                                                        // 1
  },                                                                                                                   // 1
  PRIVATE_IPV6_ADDRESSES: function (v) {                                                                               // 1
    PRIVATE_IPV6_ADDRESSES = v;                                                                                        // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 14
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 15
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 16
                                                                                                                       //
var updateLoginStyleToRedirect = function (db, backend) {                                                              // 18
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                           // 19
  ["google", "github"].forEach(function (serviceName) {                                                                // 20
    var config = configurations.findOne({                                                                              // 21
      service: serviceName                                                                                             // 21
    });                                                                                                                // 21
                                                                                                                       //
    if (config && config.loginStyle !== "redirect") {                                                                  // 22
      configurations.update({                                                                                          // 23
        service: serviceName                                                                                           // 23
      }, {                                                                                                             // 23
        $set: {                                                                                                        // 23
          loginStyle: "redirect"                                                                                       // 23
        }                                                                                                              // 23
      });                                                                                                              // 23
    }                                                                                                                  // 24
  });                                                                                                                  // 25
};                                                                                                                     // 26
                                                                                                                       //
var enableLegacyOAuthProvidersIfNotInSettings = function (db, backend) {                                               // 28
  // In the before time, Google and Github login were enabled by default.                                              // 29
  //                                                                                                                   // 30
  // This actually didn't make much sense, required the first user to configure                                        // 31
  // OAuth, and had some trust-the-first-user properties that weren't totally                                          // 32
  // secure.                                                                                                           // 33
  //                                                                                                                   // 34
  // Now, we have admin-token, so we wish to disable all logins by default                                             // 35
  // (since they need to be configured anyway) but since users may have never                                          // 36
  // explicitly told Sandstorm that Google or Github login should be enabled,                                          // 37
  // we can't just use the value in the Settings collection, since it might                                            // 38
  // never have been set.                                                                                              // 39
  //                                                                                                                   // 40
  // Thus, if the service is configured but absent from Settings, we should                                            // 41
  // explicitly enable it in Settings, and then the rest of the logic can just                                         // 42
  // depend on what value is in Settings and default to false without breaking                                         // 43
  // user installations.                                                                                               // 44
  var configurations = ServiceConfiguration.configurations;                                                            // 45
  ["google", "github"].forEach(function (serviceName) {                                                                // 46
    var config = configurations.findOne({                                                                              // 47
      service: serviceName                                                                                             // 47
    });                                                                                                                // 47
    var serviceConfig = db.collections.settings.findOne({                                                              // 48
      _id: serviceName                                                                                                 // 48
    });                                                                                                                // 48
                                                                                                                       //
    if (config && !serviceConfig) {                                                                                    // 49
      // Only explicitly enable the login service if:                                                                  // 50
      // 1) the service is already configured                                                                          // 51
      // 2) there is no sandstorm configuration already present (the user was                                          // 52
      //    using the previous default behavior).                                                                      // 53
      db.collections.settings.insert({                                                                                 // 54
        _id: serviceName,                                                                                              // 54
        value: true                                                                                                    // 54
      });                                                                                                              // 54
    }                                                                                                                  // 55
  });                                                                                                                  // 56
};                                                                                                                     // 57
                                                                                                                       //
var denormalizeInviteInfo = function (db, backend) {                                                                   // 59
  // When a user is invited via a signup token, the `signupKey` field of their user table entry                        // 60
  // has always been populated to indicate the key they used. This points into the SignupKeys table                    // 61
  // which has more information about the key, namely a freeform note entered by the admin when                        // 62
  // they created the key. In the case that the email invite form was used, the note has the form                      // 63
  // "E-mail invite to <address>".                                                                                     // 64
  //                                                                                                                   // 65
  // Later, we decided it was useful to indicate in the users table visible to the admin                               // 66
  // information about the invite terms. Namely, for email invites we want to show the address                         // 67
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                          // 68
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing                        // 69
  // from the SignupKeys table.                                                                                        // 70
  db.collections.users.find().forEach(function (user) {                                                                // 72
    if (user.signupKey && typeof user.signupKey === "string" && user.signupKey !== "admin") {                          // 73
      var signupInfo = db.collections.signupKeys.findOne(user.signupKey);                                              // 74
                                                                                                                       //
      if (signupInfo && signupInfo.note) {                                                                             // 75
        var newFields = {                                                                                              // 76
          signupNote: signupInfo.note                                                                                  // 76
        };                                                                                                             // 76
        var prefix = "E-mail invite to ";                                                                              // 78
                                                                                                                       //
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                               // 79
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                                // 80
        }                                                                                                              // 81
                                                                                                                       //
        db.collections.users.update(user._id, {                                                                        // 83
          $set: newFields                                                                                              // 83
        });                                                                                                            // 83
      }                                                                                                                // 84
    }                                                                                                                  // 85
  });                                                                                                                  // 86
};                                                                                                                     // 87
                                                                                                                       //
var mergeRoleAssignmentsIntoApiTokens = function (db, backend) {                                                       // 89
  db.collections.roleAssignments.find().forEach(function (roleAssignment) {                                            // 90
    db.collections.apiTokens.insert({                                                                                  // 91
      grainId: roleAssignment.grainId,                                                                                 // 92
      userId: roleAssignment.sharer,                                                                                   // 93
      roleAssignment: roleAssignment.roleAssignment,                                                                   // 94
      petname: roleAssignment.petname,                                                                                 // 95
      created: roleAssignment.created,                                                                                 // 96
      owner: {                                                                                                         // 97
        user: {                                                                                                        // 98
          userId: roleAssignment.recipient,                                                                            // 99
          title: roleAssignment.title                                                                                  // 100
        }                                                                                                              // 98
      }                                                                                                                // 97
    });                                                                                                                // 91
  });                                                                                                                  // 104
};                                                                                                                     // 105
                                                                                                                       //
var fixOasisStorageUsageStats = function (db, backend) {}; // This migration only pertained to Oasis and it was successfully applied there. Since it referred
// to some global variables that we later wanted to remove and/or rename, we've since replaced it                      // 109
// with a no-op.                                                                                                       // 110
                                                                                                                       //
                                                                                                                       //
var fetchProfilePictures = function (db, backend) {                                                                    // 112
  db.collections.users.find({}).forEach(function (user) {                                                              // 113
    var url = userPictureUrl(user);                                                                                    // 114
                                                                                                                       //
    if (url) {                                                                                                         // 115
      console.log("Fetching user picture:", url);                                                                      // 116
      var assetId = fetchPicture(db, url);                                                                             // 117
                                                                                                                       //
      if (assetId) {                                                                                                   // 118
        db.collections.users.update(user._id, {                                                                        // 119
          $set: {                                                                                                      // 119
            "profile.picture": assetId                                                                                 // 119
          }                                                                                                            // 119
        });                                                                                                            // 119
      }                                                                                                                // 120
    }                                                                                                                  // 121
  });                                                                                                                  // 122
};                                                                                                                     // 123
                                                                                                                       //
var assignPlans = function (db, backend) {// This was a one-time migration intended to be applied on Oasis to existing users.
  // It has run, so we only need this stub function here.                                                              // 127
};                                                                                                                     // 128
                                                                                                                       //
var removeKeyrings = function (db, backend) {                                                                          // 130
  // These blobs full of public keys were not intended to find their way into mongo and while                          // 131
  // harmless they slow things down because they're huge. Remove them.                                                 // 132
  db.collections.packages.update({                                                                                     // 133
    "manifest.metadata.pgpKeyring": {                                                                                  // 133
      $exists: true                                                                                                    // 133
    }                                                                                                                  // 133
  }, {                                                                                                                 // 133
    $unset: {                                                                                                          // 134
      "manifest.metadata.pgpKeyring": ""                                                                               // 134
    }                                                                                                                  // 134
  }, {                                                                                                                 // 134
    multi: true                                                                                                        // 135
  });                                                                                                                  // 135
};                                                                                                                     // 136
                                                                                                                       //
var useLocalizedTextInUserActions = function (db, backend) {                                                           // 138
  function toLocalizedText(newObj, oldObj, field) {                                                                    // 139
    if (field in oldObj) {                                                                                             // 140
      if (typeof oldObj[field] === "string") {                                                                         // 141
        newObj[field] = {                                                                                              // 142
          defaultText: oldObj[field]                                                                                   // 142
        };                                                                                                             // 142
      } else {                                                                                                         // 143
        newObj[field] = oldObj[field];                                                                                 // 144
      }                                                                                                                // 145
    }                                                                                                                  // 146
  }                                                                                                                    // 147
                                                                                                                       //
  db.collections.userActions.find({}).forEach(function (userAction) {                                                  // 149
    var fields = {};                                                                                                   // 150
    toLocalizedText(fields, userAction, "appTitle");                                                                   // 151
    toLocalizedText(fields, userAction, "title");                                                                      // 152
    toLocalizedText(fields, userAction, "nounPhrase");                                                                 // 153
    db.collections.userActions.update(userAction._id, {                                                                // 154
      $set: fields                                                                                                     // 154
    });                                                                                                                // 154
  });                                                                                                                  // 155
};                                                                                                                     // 156
                                                                                                                       //
var verifyAllPgpSignatures = function (db, backend) {                                                                  // 158
  db.collections.packages.find({}).forEach(function (pkg) {                                                            // 159
    try {                                                                                                              // 160
      console.log("checking PGP signature for package:", pkg._id);                                                     // 161
      var info = waitPromise(backend.cap().tryGetPackage(pkg._id));                                                    // 162
                                                                                                                       //
      if (info.authorPgpKeyFingerprint) {                                                                              // 163
        console.log("  " + info.authorPgpKeyFingerprint);                                                              // 164
        db.collections.packages.update(pkg._id, {                                                                      // 165
          $set: {                                                                                                      // 166
            authorPgpKeyFingerprint: info.authorPgpKeyFingerprint                                                      // 166
          }                                                                                                            // 166
        });                                                                                                            // 166
      } else {                                                                                                         // 167
        console.log("  no signature");                                                                                 // 168
      }                                                                                                                // 169
    } catch (err) {                                                                                                    // 170
      console.error(err.stack);                                                                                        // 171
    }                                                                                                                  // 172
  });                                                                                                                  // 173
};                                                                                                                     // 174
                                                                                                                       //
var splitUserIdsIntoAccountIdsAndIdentityIds = function (db, backend) {                                                // 176
  db.collections.users.find().forEach(function (user) {                                                                // 177
    var identity = {};                                                                                                 // 178
    var serviceUserId = void 0;                                                                                        // 179
                                                                                                                       //
    if ("devName" in user) {                                                                                           // 180
      identity.service = "dev";                                                                                        // 181
      serviceUserId = user.devName;                                                                                    // 182
    } else if ("expires" in user) {                                                                                    // 183
      identity.service = "demo";                                                                                       // 184
      serviceUserId = user._id;                                                                                        // 185
    } else if (user.services && "google" in user.services) {                                                           // 186
      identity.service = "google";                                                                                     // 187
                                                                                                                       //
      if (user.services.google.email && user.services.google.verified_email) {                                         // 188
        // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                          // 188
        identity.verifiedEmail = user.services.google.email;                                                           // 189
      }                                                                                                                // 190
                                                                                                                       //
      serviceUserId = user.services.google.id;                                                                         // 192
    } else if (user.services && "github" in user.services) {                                                           // 193
      identity.service = "github";                                                                                     // 194
      identity.unverifiedEmail = user.services.github.email;                                                           // 195
      serviceUserId = user.services.github.id;                                                                         // 196
    } else if (user.services && "emailToken" in user.services) {                                                       // 197
      identity.service = "emailToken";                                                                                 // 198
      identity.verifiedEmail = user.services.emailToken.email;                                                         // 199
      serviceUserId = user.services.emailToken.email;                                                                  // 200
    }                                                                                                                  // 201
                                                                                                                       //
    identity.id = Crypto.createHash("sha256").update(identity.service + ":" + serviceUserId).digest("hex");            // 203
                                                                                                                       //
    if (user.profile) {                                                                                                // 206
      if (user.profile.name) {                                                                                         // 207
        identity.name = user.profile.name;                                                                             // 208
      }                                                                                                                // 209
                                                                                                                       //
      if (user.profile.handle) {                                                                                       // 211
        identity.handle = user.profile.handle;                                                                         // 212
      }                                                                                                                // 213
                                                                                                                       //
      if (user.profile.picture) {                                                                                      // 215
        identity.picture = user.profile.picture;                                                                       // 216
      }                                                                                                                // 217
                                                                                                                       //
      if (user.profile.pronoun) {                                                                                      // 219
        identity.pronoun = user.profile.pronoun;                                                                       // 220
      }                                                                                                                // 221
                                                                                                                       //
      if (user.profile.email) {                                                                                        // 223
        identity.unverifiedEmail = user.profile.email;                                                                 // 224
      }                                                                                                                // 225
    }                                                                                                                  // 226
                                                                                                                       //
    identity.main = true;                                                                                              // 228
    db.collections.users.update(user._id, {                                                                            // 230
      $set: {                                                                                                          // 230
        identities: [identity]                                                                                         // 230
      }                                                                                                                // 230
    });                                                                                                                // 230
    db.collections.grains.update({                                                                                     // 232
      userId: user._id                                                                                                 // 232
    }, {                                                                                                               // 232
      $set: {                                                                                                          // 232
        identityId: identity.id                                                                                        // 232
      }                                                                                                                // 232
    }, {                                                                                                               // 232
      multi: true                                                                                                      // 232
    });                                                                                                                // 232
    db.collections.sessions.update({                                                                                   // 233
      userId: user._id                                                                                                 // 233
    }, {                                                                                                               // 233
      $set: {                                                                                                          // 233
        identityId: identity.id                                                                                        // 233
      }                                                                                                                // 233
    }, {                                                                                                               // 233
      multi: true                                                                                                      // 233
    });                                                                                                                // 233
    db.collections.apiTokens.update({                                                                                  // 234
      userId: user._id                                                                                                 // 234
    }, {                                                                                                               // 234
      $set: {                                                                                                          // 235
        identityId: identity.id                                                                                        // 235
      }                                                                                                                // 235
    }, {                                                                                                               // 235
      multi: true                                                                                                      // 236
    });                                                                                                                // 236
    db.collections.apiTokens.update({                                                                                  // 237
      "owner.user.userId": user._id                                                                                    // 237
    }, {                                                                                                               // 237
      $set: {                                                                                                          // 238
        "owner.user.identityId": identity.id                                                                           // 238
      }                                                                                                                // 238
    }, {                                                                                                               // 238
      multi: true                                                                                                      // 239
    });                                                                                                                // 239
    db.collections.apiTokens.update({                                                                                  // 240
      "owner.grain.introducerUser": user._id                                                                           // 240
    }, {                                                                                                               // 240
      $set: {                                                                                                          // 241
        "owner.grain.introducerIdentity": identity.id                                                                  // 241
      }                                                                                                                // 241
    }, {                                                                                                               // 241
      multi: true                                                                                                      // 242
    });                                                                                                                // 242
                                                                                                                       //
    while (db.collections.apiTokens.update({                                                                           // 244
      "requirements.permissionsHeld.userId": user._id                                                                  // 245
    }, {                                                                                                               // 244
      $set: {                                                                                                          // 247
        "requirements.$.permissionsHeld.identityId": identity.id                                                       // 247
      },                                                                                                               // 247
      $unset: {                                                                                                        // 248
        "requirements.$.permissionsHeld.userId": 1                                                                     // 248
      }                                                                                                                // 248
    }, {                                                                                                               // 246
      multi: true                                                                                                      // 250
    }) > 0) {} // The `$` operator modifies the first element in the array that matches the query. Since               // 249
    // there may be many matches, we need to repeat until no documents are modified.                                   // 253
                                                                                                                       //
  });                                                                                                                  // 255
  db.collections.apiTokens.remove({                                                                                    // 257
    userInfo: {                                                                                                        // 257
      $exists: true                                                                                                    // 257
    }                                                                                                                  // 257
  }); // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                          // 257
  // that this field could show up in the database was in this deprecated, no-longer-functional                        // 259
  // form of API token.                                                                                                // 260
};                                                                                                                     // 261
                                                                                                                       //
var appUpdateSettings = function (db, backend) {                                                                       // 263
  db.collections.settings.insert({                                                                                     // 264
    _id: "appMarketUrl",                                                                                               // 264
    value: "https://apps.sandstorm.io"                                                                                 // 264
  });                                                                                                                  // 264
  db.collections.settings.insert({                                                                                     // 265
    _id: "appIndexUrl",                                                                                                // 265
    value: "https://app-index.sandstorm.io"                                                                            // 265
  });                                                                                                                  // 265
  db.collections.settings.insert({                                                                                     // 266
    _id: "appUpdatesEnabled",                                                                                          // 266
    value: true                                                                                                        // 266
  });                                                                                                                  // 266
};                                                                                                                     // 267
                                                                                                                       //
var moveDevAndEmailLoginDataIntoIdentities = function (db, backend) {                                                  // 269
  db.collections.users.find().forEach(function (user) {                                                                // 270
    if (user.identities.length !== 1) {                                                                                // 271
      throw new Error("User does not have exactly one identity: ", user);                                              // 272
    }                                                                                                                  // 273
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 275
                                                                                                                       //
    if (Match.test(identity.service, Object)) {                                                                        // 276
      return;                                                                                                          // 276
    } // Already migrated.                                                                                             // 276
                                                                                                                       //
                                                                                                                       //
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");                   // 278
                                                                                                                       //
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                                    // 279
    var serviceObject = {};                                                                                            // 281
    var fieldsToUnset = {};                                                                                            // 282
                                                                                                                       //
    if (identity.service === "dev") {                                                                                  // 284
      serviceObject.name = user.devName;                                                                               // 285
      fieldsToUnset.devName = 1;                                                                                       // 286
    } else if (identity.service === "emailToken") {                                                                    // 287
      serviceObject.tokens = user.services.emailToken.tokens;                                                          // 288
      serviceObject.email = user.services.emailToken.email;                                                            // 289
      fieldsToUnset["services.emailToken"] = 1;                                                                        // 290
    }                                                                                                                  // 291
                                                                                                                       //
    newIdentity.service = {};                                                                                          // 293
    newIdentity.service[identity.service] = serviceObject;                                                             // 294
    var modifier = {                                                                                                   // 296
      $set: {                                                                                                          // 296
        identities: [newIdentity]                                                                                      // 296
      }                                                                                                                // 296
    };                                                                                                                 // 296
                                                                                                                       //
    if (Object.keys(fieldsToUnset).length > 0) {                                                                       // 297
      modifier.$unset = fieldsToUnset;                                                                                 // 298
    }                                                                                                                  // 299
                                                                                                                       //
    db.collections.users.update({                                                                                      // 301
      _id: user._id                                                                                                    // 301
    }, modifier);                                                                                                      // 301
  });                                                                                                                  // 302
};                                                                                                                     // 303
                                                                                                                       //
var repairEmailIdentityIds = function (db, backend) {                                                                  // 305
  db.collections.users.find({                                                                                          // 306
    "identities.service.emailToken": {                                                                                 // 306
      $exists: 1                                                                                                       // 306
    }                                                                                                                  // 306
  }).forEach(function (user) {                                                                                         // 306
    if (user.identities.length !== 1) {                                                                                // 307
      throw new Error("User does not have exactly one identity: ", user);                                              // 308
    }                                                                                                                  // 309
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 311
                                                                                                                       //
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail", "profile");               // 312
                                                                                                                       //
    newIdentity.service = {                                                                                            // 314
      email: identity.service.emailToken                                                                               // 314
    };                                                                                                                 // 314
    newIdentity.id = Crypto.createHash("sha256").update("email:" + identity.service.emailToken.email).digest("hex");   // 315
    db.collections.grains.update({                                                                                     // 318
      identityId: identity.id                                                                                          // 318
    }, {                                                                                                               // 318
      $set: {                                                                                                          // 318
        identityId: newIdentity.id                                                                                     // 318
      }                                                                                                                // 318
    }, {                                                                                                               // 318
      multi: true                                                                                                      // 318
    });                                                                                                                // 318
    db.collections.sessions.update({                                                                                   // 319
      identityId: identity.id                                                                                          // 319
    }, {                                                                                                               // 319
      $set: {                                                                                                          // 319
        identityId: newIdentity.id                                                                                     // 319
      }                                                                                                                // 319
    }, {                                                                                                               // 319
      multi: true                                                                                                      // 319
    });                                                                                                                // 319
    db.collections.apiTokens.update({                                                                                  // 320
      identityId: identity.id                                                                                          // 320
    }, {                                                                                                               // 320
      $set: {                                                                                                          // 321
        identityId: newIdentity.id                                                                                     // 321
      }                                                                                                                // 321
    }, {                                                                                                               // 321
      multi: true                                                                                                      // 322
    });                                                                                                                // 322
    db.collections.apiTokens.update({                                                                                  // 323
      "owner.user.identityId": identity.id                                                                             // 323
    }, {                                                                                                               // 323
      $set: {                                                                                                          // 324
        "owner.user.identityId": newIdentity.id                                                                        // 324
      }                                                                                                                // 324
    }, {                                                                                                               // 324
      multi: true                                                                                                      // 325
    });                                                                                                                // 325
    db.collections.apiTokens.update({                                                                                  // 326
      "owner.grain.introducerIdentity": identity.id                                                                    // 326
    }, {                                                                                                               // 326
      $set: {                                                                                                          // 327
        "owner.grain.introducerIdentity": newIdentity.id                                                               // 327
      }                                                                                                                // 327
    }, {                                                                                                               // 327
      multi: true                                                                                                      // 328
    });                                                                                                                // 328
                                                                                                                       //
    while (db.collections.apiTokens.update({                                                                           // 330
      "requirements.permissionsHeld.identityId": identity.id                                                           // 330
    }, {                                                                                                               // 330
      $set: {                                                                                                          // 331
        "requirements.$.permissionsHeld.identityId": newIdentity.id                                                    // 331
      }                                                                                                                // 331
    }, {                                                                                                               // 331
      multi: true                                                                                                      // 332
    }) > 0) {}                                                                                                         // 332
                                                                                                                       //
    db.collections.users.update({                                                                                      // 334
      _id: user._id                                                                                                    // 334
    }, {                                                                                                               // 334
      $set: {                                                                                                          // 334
        identities: [newIdentity]                                                                                      // 334
      }                                                                                                                // 334
    });                                                                                                                // 334
  });                                                                                                                  // 335
};                                                                                                                     // 336
                                                                                                                       //
var splitAccountUsersAndIdentityUsers = function (db, backend) {                                                       // 338
  db.collections.users.find({                                                                                          // 339
    identities: {                                                                                                      // 339
      $exists: true                                                                                                    // 339
    }                                                                                                                  // 339
  }).forEach(function (user) {                                                                                         // 339
    if (user.identities.length !== 1) {                                                                                // 340
      throw new Error("User does not have exactly one identity: ", user);                                              // 341
    }                                                                                                                  // 342
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 344
                                                                                                                       //
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires");                                             // 345
                                                                                                                       //
    identityUser._id = identity.id;                                                                                    // 346
    identityUser.profile = identity.profile;                                                                           // 347
                                                                                                                       //
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                                       // 348
                                                                                                                       //
    identityUser.profile.service = Object.keys(identity.service)[0]; // Updating this user needs to be a two step process because the `services` field typically
    // contains subfields that are constrained to be unique by Mongo indices.                                          // 352
                                                                                                                       //
    identityUser.stagedServices = _.omit(user.services, "resume");                                                     // 353
                                                                                                                       //
    if (identity.service.dev) {                                                                                        // 354
      identityUser.stagedServices.dev = identity.service.dev;                                                          // 355
    } else if (identity.service.email) {                                                                               // 356
      identityUser.stagedServices.email = identity.service.email;                                                      // 357
    }                                                                                                                  // 358
                                                                                                                       //
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires", "isAdmin", "signupKey", "signupNote", "signupEmail", "plan", "storageUsage", "isAppDemoUser", "appDemoId", "payments", "dailySentMailCount", "hasCompletedSignup");
                                                                                                                       //
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                            // 364
    accountUser.nonloginIdentities = [];                                                                               // 365
                                                                                                                       //
    if (user.services && user.services.resume) {                                                                       // 366
      accountUser.services = {                                                                                         // 367
        resume: user.services.resume                                                                                   // 367
      };                                                                                                               // 367
    }                                                                                                                  // 368
                                                                                                                       //
    accountUser.stashedOldUser = user;                                                                                 // 370
    db.collections.apiTokens.update({                                                                                  // 372
      identityId: identityUser._id                                                                                     // 373
    }, {                                                                                                               // 373
      $set: {                                                                                                          // 374
        accountId: user._id                                                                                            // 374
      }                                                                                                                // 374
    }, {                                                                                                               // 374
      multi: true                                                                                                      // 375
    });                                                                                                                // 375
    db.collections.users.upsert({                                                                                      // 377
      _id: identityUser._id                                                                                            // 377
    }, identityUser);                                                                                                  // 377
    db.collections.users.update({                                                                                      // 378
      _id: user._id                                                                                                    // 378
    }, accountUser);                                                                                                   // 378
  });                                                                                                                  // 379
  db.collections.users.find({                                                                                          // 381
    stagedServices: {                                                                                                  // 381
      $exists: true                                                                                                    // 381
    }                                                                                                                  // 381
  }).forEach(function (identity) {                                                                                     // 381
    db.collections.users.update({                                                                                      // 382
      _id: identity._id                                                                                                // 382
    }, {                                                                                                               // 382
      $unset: {                                                                                                        // 383
        stagedServices: 1                                                                                              // 383
      },                                                                                                               // 383
      $set: {                                                                                                          // 384
        services: identity.stagedServices                                                                              // 384
      }                                                                                                                // 384
    });                                                                                                                // 382
  });                                                                                                                  // 386
};                                                                                                                     // 387
                                                                                                                       //
var populateContactsFromApiTokens = function (db, backend) {                                                           // 389
  db.collections.apiTokens.find({                                                                                      // 390
    "owner.user.identityId": {                                                                                         // 391
      $exists: 1                                                                                                       // 391
    },                                                                                                                 // 391
    accountId: {                                                                                                       // 392
      $exists: 1                                                                                                       // 392
    }                                                                                                                  // 392
  }).forEach(function (token) {                                                                                        // 390
    var identityId = token.owner.user.identityId;                                                                      // 394
    var identity = Meteor.users.findOne({                                                                              // 395
      _id: identityId                                                                                                  // 395
    });                                                                                                                // 395
                                                                                                                       //
    if (identity) {                                                                                                    // 396
      var profile = identity.profile;                                                                                  // 397
      SandstormDb.fillInProfileDefaults(identity, profile);                                                            // 398
      db.collections.contacts.upsert({                                                                                 // 399
        ownerId: token.accountId,                                                                                      // 399
        identityId: identityId                                                                                         // 399
      }, {                                                                                                             // 399
        ownerId: token.accountId,                                                                                      // 400
        petname: profile && profile.name,                                                                              // 401
        created: new Date(),                                                                                           // 402
        identityId: identityId,                                                                                        // 403
        profile: profile                                                                                               // 404
      });                                                                                                              // 399
    }                                                                                                                  // 406
  });                                                                                                                  // 407
};                                                                                                                     // 408
                                                                                                                       //
var cleanUpApiTokens = function (db, backend) {                                                                        // 410
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases                       // 411
  // where the user still existed in the database.                                                                     // 412
  db.collections.apiTokens.remove({                                                                                    // 413
    userId: {                                                                                                          // 414
      $exists: true                                                                                                    // 414
    },                                                                                                                 // 414
    identityId: {                                                                                                      // 415
      $exists: false                                                                                                   // 415
    }                                                                                                                  // 415
  });                                                                                                                  // 413
  db.collections.apiTokens.remove({                                                                                    // 417
    "owner.user.userId": {                                                                                             // 418
      $exists: true                                                                                                    // 418
    },                                                                                                                 // 418
    "owner.user.identityId": {                                                                                         // 419
      $exists: false                                                                                                   // 419
    }                                                                                                                  // 419
  }); // For a while we were accidentally setting `appIcon` instead of `icon`.                                         // 417
                                                                                                                       //
  db.collections.apiTokens.find({                                                                                      // 423
    "owner.user.denormalizedGrainMetadata.appIcon": {                                                                  // 424
      $exists: true                                                                                                    // 424
    }                                                                                                                  // 424
  }).forEach(function (apiToken) {                                                                                     // 423
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                                  // 426
    db.collections.apiTokens.update({                                                                                  // 427
      _id: apiToken._id                                                                                                // 427
    }, {                                                                                                               // 427
      $set: {                                                                                                          // 428
        "owner.user.denormalizedGrainMetadata.icon": icon                                                              // 428
      },                                                                                                               // 428
      $unset: {                                                                                                        // 429
        "owner.user.denormalizedGrainMetadata.appIcon": true                                                           // 429
      }                                                                                                                // 429
    });                                                                                                                // 427
  }); // For a while the `identityId` field of child UiView tokens was not getting set.                                // 431
                                                                                                                       //
  function repairChain(parentToken) {                                                                                  // 434
    db.collections.apiTokens.find({                                                                                    // 435
      parentToken: parentToken._id,                                                                                    // 436
      grainId: {                                                                                                       // 437
        $exists: true                                                                                                  // 437
      },                                                                                                               // 437
      identityId: {                                                                                                    // 438
        $exists: false                                                                                                 // 438
      }                                                                                                                // 438
    }).forEach(function (childToken) {                                                                                 // 435
      db.collections.apiTokens.update({                                                                                // 440
        _id: childToken._id                                                                                            // 440
      }, {                                                                                                             // 440
        $set: {                                                                                                        // 440
          identityId: parentToken.identityId                                                                           // 440
        }                                                                                                              // 440
      });                                                                                                              // 440
      repairChain(childToken);                                                                                         // 441
    });                                                                                                                // 442
  }                                                                                                                    // 443
                                                                                                                       //
  db.collections.apiTokens.find({                                                                                      // 445
    grainId: {                                                                                                         // 446
      $exists: true                                                                                                    // 446
    },                                                                                                                 // 446
    identityId: {                                                                                                      // 447
      $exists: true                                                                                                    // 447
    },                                                                                                                 // 447
    parentToken: {                                                                                                     // 448
      $exists: false                                                                                                   // 448
    }                                                                                                                  // 448
  }).forEach(repairChain);                                                                                             // 445
};                                                                                                                     // 450
                                                                                                                       //
var initServerTitleAndReturnAddress = function (db, backend) {                                                         // 452
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                             // 453
  db.collections.settings.insert({                                                                                     // 454
    _id: "serverTitle",                                                                                                // 454
    value: hostname                                                                                                    // 454
  });                                                                                                                  // 454
  db.collections.settings.insert({                                                                                     // 455
    _id: "returnAddress",                                                                                              // 455
    value: "no-reply@" + hostname                                                                                      // 455
  });                                                                                                                  // 455
};                                                                                                                     // 456
                                                                                                                       //
var sendReferralNotifications = function (db, backend) {                                                               // 458
  if (db.isReferralEnabled()) {                                                                                        // 459
    db.collections.users.find({                                                                                        // 460
      loginIdentities: {                                                                                               // 461
        $exists: true                                                                                                  // 461
      },                                                                                                               // 461
      expires: {                                                                                                       // 462
        $exists: false                                                                                                 // 462
      }                                                                                                                // 462
    }, {                                                                                                               // 460
      fields: {                                                                                                        // 463
        _id: 1                                                                                                         // 463
      }                                                                                                                // 463
    }).forEach(function (user) {                                                                                       // 463
      db.sendReferralProgramNotification(user._id);                                                                    // 464
    });                                                                                                                // 465
  }                                                                                                                    // 466
};                                                                                                                     // 467
                                                                                                                       //
var assignBonuses = function (db, backend) {// This was a one-time migration intended to be applied on Oasis to existing users.
  // It has run, so we only need this stub function here.                                                              // 471
};                                                                                                                     // 472
                                                                                                                       //
var splitSmtpUrl = function (db, backend) {                                                                            // 474
  var smtpUrlSetting = db.collections.settings.findOne({                                                               // 475
    _id: "smtpUrl"                                                                                                     // 475
  });                                                                                                                  // 475
  var smtpUrl = smtpUrlSetting ? smtpUrlSetting.value : process.env.MAIL_URL;                                          // 476
  var returnAddress = db.collections.settings.findOne({                                                                // 477
    _id: "returnAddress"                                                                                               // 477
  }); // Default values.                                                                                               // 477
                                                                                                                       //
  var smtpConfig = {                                                                                                   // 480
    hostname: "localhost",                                                                                             // 481
    port: "25",                                                                                                        // 482
    auth: undefined,                                                                                                   // 483
    returnAddress: returnAddress.value                                                                                 // 484
  };                                                                                                                   // 480
  var parsed = void 0;                                                                                                 // 487
                                                                                                                       //
  try {                                                                                                                // 488
    parsed = smtpUrl && Url.parse(smtpUrl);                                                                            // 489
  } catch (e) {}                                                                                                       // 490
                                                                                                                       //
  if (parsed) {                                                                                                        // 492
    // If there was a SMTP URL previously defined, import its data.                                                    // 493
    var auth = undefined;                                                                                              // 494
                                                                                                                       //
    if (parsed.auth) {                                                                                                 // 495
      var colonIndex = parsed.auth.indexOf(":");                                                                       // 496
      var user = undefined;                                                                                            // 497
      var pass = undefined;                                                                                            // 498
                                                                                                                       //
      if (colonIndex !== -1) {                                                                                         // 499
        user = parsed.auth.slice(0, colonIndex);                                                                       // 500
        pass = parsed.auth.slice(colonIndex + 1);                                                                      // 501
      }                                                                                                                // 502
                                                                                                                       //
      auth = {                                                                                                         // 504
        user: user,                                                                                                    // 505
        pass: pass                                                                                                     // 506
      };                                                                                                               // 504
    } // Override defaults with previous config's values.                                                              // 508
                                                                                                                       //
                                                                                                                       //
    smtpConfig.hostname = parsed.hostname || "localhost";                                                              // 511
    smtpConfig.port = parsed.port || "25";                                                                             // 512
    smtpConfig.auth = auth;                                                                                            // 513
  }                                                                                                                    // 514
                                                                                                                       //
  db.collections.settings.upsert({                                                                                     // 516
    _id: "smtpConfig"                                                                                                  // 516
  }, {                                                                                                                 // 516
    value: smtpConfig                                                                                                  // 516
  });                                                                                                                  // 516
  db.collections.settings.remove({                                                                                     // 517
    _id: "returnAddress"                                                                                               // 517
  });                                                                                                                  // 517
  db.collections.settings.remove({                                                                                     // 518
    _id: "smtpUrl"                                                                                                     // 518
  });                                                                                                                  // 518
};                                                                                                                     // 519
                                                                                                                       //
var smtpPortShouldBeNumber = function (db, backend) {                                                                  // 521
  var entry = db.collections.settings.findOne({                                                                        // 522
    _id: "smtpConfig"                                                                                                  // 522
  });                                                                                                                  // 522
                                                                                                                       //
  if (entry) {                                                                                                         // 523
    var setting = entry.value;                                                                                         // 524
                                                                                                                       //
    if (setting.port) {                                                                                                // 525
      setting.port = _.isNumber(setting.port) ? setting.port : parseInt(setting.port);                                 // 526
      db.collections.settings.upsert({                                                                                 // 527
        _id: "smtpConfig"                                                                                              // 527
      }, {                                                                                                             // 527
        value: setting                                                                                                 // 527
      });                                                                                                              // 527
    }                                                                                                                  // 528
  }                                                                                                                    // 529
};                                                                                                                     // 530
                                                                                                                       //
var consolidateOrgSettings = function (db, backend) {                                                                  // 532
  var settings = db.collections.settings;                                                                              // 533
  var orgGoogleDomain = settings.findOne({                                                                             // 534
    _id: "organizationGoogle"                                                                                          // 534
  });                                                                                                                  // 534
  var orgEmailDomain = settings.findOne({                                                                              // 535
    _id: "organizationEmail"                                                                                           // 535
  });                                                                                                                  // 535
  var orgLdap = settings.findOne({                                                                                     // 536
    _id: "organizationLdap"                                                                                            // 536
  });                                                                                                                  // 536
  var orgSaml = settings.findOne({                                                                                     // 537
    _id: "organizationSaml"                                                                                            // 537
  });                                                                                                                  // 537
  var orgMembership = {                                                                                                // 539
    google: {                                                                                                          // 540
      enabled: orgGoogleDomain ? !!orgGoogleDomain.value : false,                                                      // 541
      domain: orgGoogleDomain ? orgGoogleDomain.value : ""                                                             // 542
    },                                                                                                                 // 540
    email: {                                                                                                           // 544
      enabled: orgEmailDomain ? !!orgEmailDomain.value : false,                                                        // 545
      domain: orgEmailDomain ? orgEmailDomain.value : ""                                                               // 546
    },                                                                                                                 // 544
    ldap: {                                                                                                            // 548
      enabled: orgLdap ? orgLdap.value : false                                                                         // 549
    },                                                                                                                 // 548
    saml: {                                                                                                            // 551
      enabled: orgSaml ? orgSaml.value : false                                                                         // 552
    }                                                                                                                  // 551
  };                                                                                                                   // 539
  settings.upsert({                                                                                                    // 556
    _id: "organizationMembership"                                                                                      // 556
  }, {                                                                                                                 // 556
    value: orgMembership                                                                                               // 556
  });                                                                                                                  // 556
  settings.remove({                                                                                                    // 557
    _id: "organizationGoogle"                                                                                          // 557
  });                                                                                                                  // 557
  settings.remove({                                                                                                    // 558
    _id: "organizationEmail"                                                                                           // 558
  });                                                                                                                  // 558
  settings.remove({                                                                                                    // 559
    _id: "organizationLdap"                                                                                            // 559
  });                                                                                                                  // 559
  settings.remove({                                                                                                    // 560
    _id: "organizationSaml"                                                                                            // 560
  });                                                                                                                  // 560
};                                                                                                                     // 561
                                                                                                                       //
var unsetSmtpDefaultHostnameIfNoUsersExist = function (db, backend) {                                                  // 563
  // We don't actually want to have the default hostname "localhost" set.                                              // 564
  // If the user has already finished configuring their server, then this migration should do                          // 565
  // nothing (since we might break their deployment), but for new installs (which will have no users                   // 566
  // at the time this migration runs) we'll unset the hostname if it's still the previously-filled                     // 567
  // default value.                                                                                                    // 568
  var hasUsers = db.collections.users.findOne();                                                                       // 569
                                                                                                                       //
  if (!hasUsers) {                                                                                                     // 570
    var entry = db.collections.settings.findOne({                                                                      // 571
      _id: "smtpConfig"                                                                                                // 571
    });                                                                                                                // 571
    var smtpConfig = entry.value;                                                                                      // 572
                                                                                                                       //
    if (smtpConfig.hostname === "localhost") {                                                                         // 573
      smtpConfig.hostname = "";                                                                                        // 574
      db.collections.settings.upsert({                                                                                 // 575
        _id: "smtpConfig"                                                                                              // 575
      }, {                                                                                                             // 575
        value: smtpConfig                                                                                              // 575
      });                                                                                                              // 575
    }                                                                                                                  // 576
  }                                                                                                                    // 577
};                                                                                                                     // 578
                                                                                                                       //
var extractLastUsedFromApiTokenOwner = function (db, backend) {                                                        // 580
  // We used to store lastUsed as a field on owner.user.  It makes more sense to store lastUsed on                     // 581
  // the apiToken as a whole.  This migration hoists such values from owner.user onto the apiToken                     // 582
  // itself.                                                                                                           // 583
  db.collections.apiTokens.find({                                                                                      // 584
    "owner.user": {                                                                                                    // 584
      $exists: true                                                                                                    // 584
    }                                                                                                                  // 584
  }).forEach(function (token) {                                                                                        // 584
    var lastUsed = token.owner.user.lastUsed;                                                                          // 585
    db.collections.apiTokens.update({                                                                                  // 586
      _id: token._id                                                                                                   // 586
    }, {                                                                                                               // 586
      $set: {                                                                                                          // 587
        lastUsed: lastUsed                                                                                             // 587
      },                                                                                                               // 587
      $unset: {                                                                                                        // 588
        "owner.user.lastUsed": true                                                                                    // 588
      }                                                                                                                // 588
    });                                                                                                                // 586
  });                                                                                                                  // 590
};                                                                                                                     // 591
                                                                                                                       //
var setUpstreamTitles = function (db, backend) {                                                                       // 593
  // Initializes the `upstreamTitle` and `renamed` fields of `ApiToken.owner.user`.                                    // 594
  var apiTokensRaw = db.collections.apiTokens.rawCollection();                                                         // 596
  var aggregateApiTokens = Meteor.wrapAsync(apiTokensRaw.aggregate, apiTokensRaw); // First, construct a list of all *shared* grains. We will need to do a separate update()
  // for each of these, so we'd like to skip those which have nothing to update.                                       // 600
                                                                                                                       //
  var grainIds = aggregateApiTokens([{                                                                                 // 601
    $match: {                                                                                                          // 602
      "owner.user": {                                                                                                  // 602
        $exists: true                                                                                                  // 602
      },                                                                                                               // 602
      grainId: {                                                                                                       // 602
        $exists: true                                                                                                  // 602
      }                                                                                                                // 602
    }                                                                                                                  // 602
  }, {                                                                                                                 // 602
    $group: {                                                                                                          // 603
      _id: "$grainId"                                                                                                  // 603
    }                                                                                                                  // 603
  }]).map(function (grain) {                                                                                           // 603
    return grain._id;                                                                                                  // 604
  });                                                                                                                  // 604
  var count = 0;                                                                                                       // 606
  db.collections.grains.find({                                                                                         // 607
    _id: {                                                                                                             // 607
      $in: grainIds                                                                                                    // 607
    }                                                                                                                  // 607
  }, {                                                                                                                 // 607
    fields: {                                                                                                          // 607
      title: 1                                                                                                         // 607
    }                                                                                                                  // 607
  }).forEach(function (grain) {                                                                                        // 607
    if (count % 100 == 0) {                                                                                            // 608
      console.log(count + " / " + grainIds.length);                                                                    // 609
    }                                                                                                                  // 610
                                                                                                                       //
    ++count; // For ApiTokens whose petname titles do not match the upstream title, we need to set                     // 612
    // `upstreamTitle` and `renamed`. We have no way of knowing which of the two names was the                         // 615
    // original name, so we don't know whether it was the owner or the receiver who renamed their                      // 616
    // copy post-sharaing. We assume it was the owner, because:                                                        // 617
    // 1. It's probably unusual for people to try to rename a grain they don't own.                                    // 618
    // 2. This results in UI that is reasonably non-confusing even if we guessed wrong. Namely,                        // 619
    //    the user sees "Owner's title (was: User's title)", which is reasonably OK even if it                         // 620
    //    was the user who had renamed their copy. They can rename it again if they like. On the                       // 621
    //    other hand, if we guessed wrongly in the other direction, the user would see                                 // 622
    //    "User's title (renamed from: Owners title)", which would be wrong if it was in fact the                      // 623
    //    owner who renamed post-sharing.                                                                              // 624
                                                                                                                       //
    db.collections.apiTokens.update({                                                                                  // 625
      grainId: grain._id,                                                                                              // 626
      "owner.user.title": {                                                                                            // 627
        $exists: true,                                                                                                 // 627
        $ne: grain.title                                                                                               // 627
      }                                                                                                                // 627
    }, {                                                                                                               // 625
      $set: {                                                                                                          // 628
        "owner.user.upstreamTitle": grain.title                                                                        // 628
      }                                                                                                                // 628
    }, {                                                                                                               // 628
      multi: true                                                                                                      // 628
    });                                                                                                                // 628
  });                                                                                                                  // 629
};                                                                                                                     // 630
                                                                                                                       //
var markAllRead = function (db, backend) {                                                                             // 632
  // Mark as "read" all grains and tokens that predate the creation of read/unread status.                             // 633
  // Otherwise it's pretty annoying to see all your old grains look like they have activity.                           // 634
  db.collections.grains.update({}, {                                                                                   // 636
    $set: {                                                                                                            // 636
      ownerSeenAllActivity: true                                                                                       // 636
    }                                                                                                                  // 636
  }, {                                                                                                                 // 636
    multi: true                                                                                                        // 636
  });                                                                                                                  // 636
  db.collections.apiTokens.update({                                                                                    // 637
    "owner.user": {                                                                                                    // 637
      $exists: true                                                                                                    // 637
    }                                                                                                                  // 637
  }, {                                                                                                                 // 637
    $set: {                                                                                                            // 638
      "owner.user.seenAllActivity": true                                                                               // 638
    }                                                                                                                  // 638
  }, {                                                                                                                 // 638
    multi: true                                                                                                        // 639
  });                                                                                                                  // 639
};                                                                                                                     // 640
                                                                                                                       //
var clearAppIndex = function (db, backend) {                                                                           // 642
  // Due to a bug in the app update code, some app update notifications that the user accepted                         // 643
  // around July 9-16, 2016 may not have applied. We have no way of knowing exactly which updates                      // 644
  // the user accepted but didn't receive. Instead, to recover, we are clearing the local cache of                     // 645
  // the app index one time. This way, the next time the app index is re-fetched, the server will                      // 646
  // interpret all of the apps in the index as updated and will re-deliver notifications to                            // 647
  // everyone. This may mean users get notifications that they previously dismissed, but they can                      // 648
  // click "dismiss" again easily enough.                                                                              // 649
  db.collections.appIndex.remove({});                                                                                  // 651
};                                                                                                                     // 652
                                                                                                                       //
var assignEmailVerifierIds = function (db, backend) {                                                                  // 654
  // Originally, the ID of an EmailVerifier was actually the _id of the root token from which it                       // 655
  // was restored. This was broken, though: Conceptually, it meant that you couldn't have a working                    // 656
  // EmailVerifier that had not been restore()d from disk. In practice, that wasn't a problem due                      // 657
  // to the fact that powerbox selections always involve a restore(), but a different problem                          // 658
  // existed: the capability returned by claimRequest() would return one ID, but after save()ing                       // 659
  // and restore()ing that capability, the ID would be different, because the new capability would                     // 660
  // be a copy -- not a child -- of the original. Additionally, the whole thing made various code                      // 661
  // ugly because it was puncturing layers of abstraction. So, we switched to doing the right                          // 662
  // thing: assigning an ID to the EmailVerifier on first creation and storing it separately.                          // 663
  db.collections.apiTokens.find({                                                                                      // 665
    "frontendRef.emailVerifier": {                                                                                     // 665
      $exists: true                                                                                                    // 665
    }                                                                                                                  // 665
  }).forEach(function (token) {                                                                                        // 665
    db.collections.apiTokens.update(token._id, {                                                                       // 666
      $set: {                                                                                                          // 666
        "frontendRef.emailVerifier.id": token._id                                                                      // 666
      }                                                                                                                // 666
    });                                                                                                                // 666
  });                                                                                                                  // 667
};                                                                                                                     // 668
                                                                                                                       //
var startPreinstallingApps = function (db, backend) {                                                                  // 670
  // This isn't really a normal migration. It will run only on brand new servers, and it has to                        // 671
  // run after the `clearAppIndex` migration because it relies on populating AppIndex.                                 // 672
  var startPreinstallingAppsHelper = function () {                                                                     // 674
    db.updateAppIndex();                                                                                               // 675
    var preinstalledApps = db.collections.appIndex.find({                                                              // 677
      _id: {                                                                                                           // 677
        $in: db.getProductivitySuiteAppIds().concat(db.getSystemSuiteAppIds())                                         // 678
      }                                                                                                                // 677
    }).fetch();                                                                                                        // 677
                                                                                                                       //
    var appAndPackageIds = _.map(preinstalledApps, function (app) {                                                    // 681
      return {                                                                                                         // 682
        appId: app.appId,                                                                                              // 683
        packageId: app.packageId                                                                                       // 684
      };                                                                                                               // 682
    });                                                                                                                // 686
                                                                                                                       //
    db.setPreinstalledApps(appAndPackageIds);                                                                          // 688
  };                                                                                                                   // 689
                                                                                                                       //
  if (!Meteor.settings.public.isTesting && !db.allowDevAccounts()) {                                                   // 691
    // We want preinstalling apps to run async and not block startup.                                                  // 692
    Meteor.setTimeout(startPreinstallingAppsHelper, 0);                                                                // 693
  }                                                                                                                    // 694
};                                                                                                                     // 695
                                                                                                                       //
var setNewServer = function (db, backend) {                                                                            // 697
  // This migration only applies to "old" servers. New servers will set                                                // 698
  // new_server_migrations_applied to false before any migrations run.                                                 // 699
  if (!db.collections.migrations.findOne({                                                                             // 700
    _id: "new_server_migrations_applied"                                                                               // 700
  })) {                                                                                                                // 700
    db.collections.migrations.insert({                                                                                 // 701
      _id: "new_server_migrations_applied",                                                                            // 701
      value: true                                                                                                      // 701
    });                                                                                                                // 701
  }                                                                                                                    // 702
};                                                                                                                     // 703
                                                                                                                       //
var addMembraneRequirementsToIdentities = function (db, backend) {                                                     // 705
  var query = {                                                                                                        // 706
    "frontendRef.identity": {                                                                                          // 707
      $exists: true                                                                                                    // 707
    },                                                                                                                 // 707
    "owner.grain.grainId": {                                                                                           // 708
      $exists: true                                                                                                    // 708
    },                                                                                                                 // 708
    "requirements.0": {                                                                                                // 709
      $exists: false                                                                                                   // 709
    }                                                                                                                  // 709
  };                                                                                                                   // 706
  db.collections.apiTokens.find(query).map(function (apiToken) {                                                       // 712
    db.collections.apiTokens.update({                                                                                  // 713
      _id: apiToken._id                                                                                                // 714
    }, {                                                                                                               // 714
      $push: {                                                                                                         // 716
        requirements: {                                                                                                // 717
          permissionsHeld: {                                                                                           // 718
            grainId: apiToken.owner.grain.grainId,                                                                     // 719
            identityId: apiToken.frontendRef.identity,                                                                 // 720
            permissions: []                                                                                            // 721
          }                                                                                                            // 718
        }                                                                                                              // 717
      }                                                                                                                // 716
    });                                                                                                                // 715
  });                                                                                                                  // 727
};                                                                                                                     // 728
                                                                                                                       //
var addEncryptionToFrontendRefIpNetwork = function (db, backend) {                                                     // 730
  db.collections.apiTokens.find({                                                                                      // 731
    "frontendRef.ipNetwork": true                                                                                      // 731
  }).map(function (apiToken) {                                                                                         // 731
    db.collections.apiTokens.update({                                                                                  // 732
      _id: apiToken._id                                                                                                // 733
    }, {                                                                                                               // 733
      $set: {                                                                                                          // 734
        "frontendRef.ipNetwork": {                                                                                     // 734
          encryption: {                                                                                                // 734
            none: null                                                                                                 // 734
          }                                                                                                            // 734
        }                                                                                                              // 734
      }                                                                                                                // 734
    });                                                                                                                // 734
  });                                                                                                                  // 735
};                                                                                                                     // 736
                                                                                                                       //
function backgroundFillInGrainSizes(db, backend) {                                                                     // 738
  // Fill in sizes for all grains that don't have them. Since computing a grain size requires a                        // 739
  // directory walk, we don't want to do them all at once. Instead, we compute one a second until                      // 740
  // all grains have sizes.                                                                                            // 741
  try {                                                                                                                // 743
    var grain = db.collections.grains.findOne({                                                                        // 744
      size: {                                                                                                          // 744
        $exists: false                                                                                                 // 744
      }                                                                                                                // 744
    }, {                                                                                                               // 744
      fields: {                                                                                                        // 744
        _id: 1,                                                                                                        // 744
        userId: 1                                                                                                      // 744
      }                                                                                                                // 744
    });                                                                                                                // 744
                                                                                                                       //
    if (grain) {                                                                                                       // 746
      // Compute size!                                                                                                 // 747
      try {                                                                                                            // 748
        var result = waitPromise(backend.cap().getGrainStorageUsage(grain.userId, grain._id));                         // 749
        db.collections.grains.update({                                                                                 // 751
          _id: grain._id,                                                                                              // 751
          size: {                                                                                                      // 751
            $exists: false                                                                                             // 751
          }                                                                                                            // 751
        }, {                                                                                                           // 751
          $set: {                                                                                                      // 752
            size: parseInt(result.size)                                                                                // 752
          }                                                                                                            // 752
        });                                                                                                            // 752
      } catch (err) {                                                                                                  // 753
        if (err.kjType === "failed") {                                                                                 // 754
          // Backend had a problem. Maybe the grain doesn't actually exist on disk and the database                    // 755
          // is messed up. We'll set the size to zero and move on.                                                     // 756
          console.error("Error while backfilling grain size for", grain._id, ":", err.stack);                          // 757
          db.collections.grains.update({                                                                               // 758
            _id: grain._id,                                                                                            // 758
            size: {                                                                                                    // 758
              $exists: false                                                                                           // 758
            }                                                                                                          // 758
          }, {                                                                                                         // 758
            $set: {                                                                                                    // 758
              size: 0                                                                                                  // 758
            }                                                                                                          // 758
          });                                                                                                          // 758
        } else {                                                                                                       // 759
          // Rethrow on disconnected / overloaded / unimplemented.                                                     // 760
          throw err;                                                                                                   // 761
        }                                                                                                              // 762
      } // Do another one in a second.                                                                                 // 763
                                                                                                                       //
                                                                                                                       //
      Meteor.setTimeout(backgroundFillInGrainSizes.bind(this, db, backend), 1000);                                     // 766
    }                                                                                                                  // 767
  } catch (err) {                                                                                                      // 768
    // We'll just stop for now, to avoid spamming logs if this error persists. Next time the server                    // 769
    // restarts, the migration will continue.                                                                          // 770
    console.error("Error while backfilling grain sizes:", err.stack);                                                  // 771
  }                                                                                                                    // 772
}                                                                                                                      // 773
                                                                                                                       //
function removeFeatureKeys(db, backend) {                                                                              // 775
  // Remove obsolete data related to the Sandstorm for Work paywall, which was eliminated.                             // 776
  db.notifications.remove({                                                                                            // 778
    "admin.type": "cantRenewFeatureKey"                                                                                // 778
  });                                                                                                                  // 778
  db.notifications.remove({                                                                                            // 779
    "admin.type": "trialFeatureKeyExpired"                                                                             // 779
  });                                                                                                                  // 779
}                                                                                                                      // 780
                                                                                                                       //
function setIpBlacklist(db, backend) {                                                                                 // 782
  if (Meteor.settings.public.isTesting) {                                                                              // 783
    db.collections.settings.insert({                                                                                   // 784
      _id: "ipBlacklist",                                                                                              // 784
      value: "192.168.0.0/16"                                                                                          // 784
    });                                                                                                                // 784
  } else {                                                                                                             // 785
    var defaultIpBlacklist = PRIVATE_IPV4_ADDRESSES.concat(PRIVATE_IPV6_ADDRESSES).join("\n");                         // 786
    db.collections.settings.insert({                                                                                   // 787
      _id: "ipBlacklist",                                                                                              // 787
      value: defaultIpBlacklist                                                                                        // 787
    });                                                                                                                // 787
  }                                                                                                                    // 788
}                                                                                                                      // 789
                                                                                                                       //
function getUserIdentityIds(user) {                                                                                    // 791
  // Formerly SandstormDb.getUserIdentityIds(), from before the identity refactor.                                     // 792
  if (user && user.loginIdentities) {                                                                                  // 794
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                              // 795
  } else {                                                                                                             // 796
    return [];                                                                                                         // 797
  }                                                                                                                    // 798
}                                                                                                                      // 799
                                                                                                                       //
function notifyIdentityChanges(db, backend) {                                                                          // 801
  // Notify users who might be affected by the identity model changes.                                                 // 802
  //                                                                                                                   // 803
  // Two types of users are affected:                                                                                  // 804
  // - Users who have multiple identities with differing names.                                                        // 805
  // - Users who have identities that are shared with other users.                                                     // 806
  //                                                                                                                   // 807
  // However, the second group seems like it must be a subset of the first group. So we only check                     // 808
  // for the first.                                                                                                    // 809
  var names = {};                                                                                                      // 811
  Meteor.users.find({                                                                                                  // 812
    "profile.name": {                                                                                                  // 812
      $exists: true                                                                                                    // 812
    }                                                                                                                  // 812
  }, {                                                                                                                 // 812
    fields: {                                                                                                          // 812
      "profile.name": 1                                                                                                // 812
    }                                                                                                                  // 812
  }).forEach(function (user) {                                                                                         // 812
    names[user._id] = user.profile.name;                                                                               // 814
  });                                                                                                                  // 815
  Meteor.users.find({                                                                                                  // 817
    loginIdentities: {                                                                                                 // 817
      $exists: true                                                                                                    // 817
    }                                                                                                                  // 817
  }, {                                                                                                                 // 817
    fields: {                                                                                                          // 818
      loginIdentities: 1,                                                                                              // 818
      nonloginIdentities: 1                                                                                            // 818
    }                                                                                                                  // 818
  }).forEach(function (user) {                                                                                         // 818
    var previousName = null;                                                                                           // 819
    var needsNotification = false;                                                                                     // 820
    getUserIdentityIds(user).forEach(function (identityId) {                                                           // 821
      var name = names[identityId];                                                                                    // 822
                                                                                                                       //
      if (!name || previousName && previousName !== name) {                                                            // 823
        needsNotification = true;                                                                                      // 824
      }                                                                                                                // 825
                                                                                                                       //
      previousName = name;                                                                                             // 827
    });                                                                                                                // 828
                                                                                                                       //
    if (needsNotification) {                                                                                           // 830
      db.collections.notifications.upsert({                                                                            // 831
        userId: user._id,                                                                                              // 832
        identityChanges: true                                                                                          // 833
      }, {                                                                                                             // 831
        userId: user._id,                                                                                              // 835
        identityChanges: true,                                                                                         // 836
        timestamp: new Date(),                                                                                         // 837
        isUnread: true                                                                                                 // 838
      });                                                                                                              // 834
    }                                                                                                                  // 840
  });                                                                                                                  // 841
}                                                                                                                      // 842
                                                                                                                       //
Mongo.Collection.prototype.ensureDroppedIndex = function () {                                                          // 844
  try {                                                                                                                // 845
    this._dropIndex.apply(this, arguments);                                                                            // 846
  } catch (err) {// ignore (probably, index didn't exist)                                                              // 847
  }                                                                                                                    // 849
};                                                                                                                     // 850
                                                                                                                       //
function onePersonaPerAccountPreCleanup(db, backend) {                                                                 // 852
  // Removes some already-obsolete data from the database before attempting the                                        // 853
  // one-persona-per-account migration.                                                                                // 854
  // Remove long-obsolete index.                                                                                       // 856
  Meteor.users.ensureDroppedIndex({                                                                                    // 857
    "identities.id": 1                                                                                                 // 857
  }); // Remove `stashedOldUser`, which is long-obsolete.                                                              // 857
                                                                                                                       //
  Meteor.users.update({                                                                                                // 860
    stashedOldUser: {                                                                                                  // 860
      $exists: true                                                                                                    // 860
    }                                                                                                                  // 860
  }, {                                                                                                                 // 860
    $unset: {                                                                                                          // 861
      stashedOldUser: 1                                                                                                // 861
    }                                                                                                                  // 861
  }, {                                                                                                                 // 861
    multi: true                                                                                                        // 862
  }); // Remove ApiTokens which have an identityId but not an accountId. These tokens could only exist                 // 862
  // if they were created in between `splitUserIdsIntoAccountIdsAndIdentityIds` and                                    // 865
  // `splitAccountUsersAndIdentityUsers` (late 2015), and if, during that time, the user who                           // 866
  // created the ApiToken was deleted (which, at the time, was only possible for demo users, or                        // 867
  // through direct database manipulation). These tokens are all invalid and couldn't possibly have                    // 868
  // been used since the user was deleted.                                                                             // 869
                                                                                                                       //
  db.collections.apiTokens.remove({                                                                                    // 870
    identityId: {                                                                                                      // 870
      $exists: true                                                                                                    // 870
    },                                                                                                                 // 870
    accountId: {                                                                                                       // 870
      $exists: false                                                                                                   // 870
    }                                                                                                                  // 870
  }); // Remove ApiTokens that have the obsolete owner.grain.introducerIdentity field as these tokens                  // 870
  // are not allowed to be restored anyway.                                                                            // 873
                                                                                                                       //
  db.collections.apiTokens.remove({                                                                                    // 874
    "owner.grain.introducerIdentity": {                                                                                // 874
      $exists: true                                                                                                    // 874
    }                                                                                                                  // 874
  }); // Make sure all demo credentials have a "services.demo" entry, to be consistent with all other                  // 874
  // service types.                                                                                                    // 877
                                                                                                                       //
  Meteor.users.update({                                                                                                // 878
    "profile.service": "demo"                                                                                          // 878
  }, {                                                                                                                 // 878
    $set: {                                                                                                            // 879
      "services.demo": {}                                                                                              // 879
    }                                                                                                                  // 879
  }, {                                                                                                                 // 879
    multi: true                                                                                                        // 880
  });                                                                                                                  // 880
}                                                                                                                      // 882
                                                                                                                       //
function forEachProgress(title, cursor, func) {                                                                        // 884
  console.log(title);                                                                                                  // 885
  var total = cursor.count();                                                                                          // 887
  var count = 0;                                                                                                       // 888
  cursor.forEach(function (doc) {                                                                                      // 890
    func(doc);                                                                                                         // 891
    if (++count % 100 == 0) console.log("   ", count, "/", total);                                                     // 892
  });                                                                                                                  // 893
}                                                                                                                      // 894
                                                                                                                       //
function onePersonaPerAccount(db, backend) {                                                                           // 896
  // THIS IS A MAJOR CHANGE: https://sandstorm.io/news/2017-05-08-refactoring-identities                               // 897
  console.log("** Migrating to new identity model! **");                                                               // 899
  console.log("see: https://sandstorm.io/news/2017-05-08-refactoring-identities");                                     // 900
  console.log("tagging accounts...");                                                                                  // 902
  Meteor.users.update({                                                                                                // 903
    type: {                                                                                                            // 903
      $exists: false                                                                                                   // 903
    },                                                                                                                 // 903
    loginIdentities: {                                                                                                 // 903
      $exists: true                                                                                                    // 903
    }                                                                                                                  // 903
  }, {                                                                                                                 // 903
    $set: {                                                                                                            // 904
      type: "account"                                                                                                  // 904
    }                                                                                                                  // 904
  }, {                                                                                                                 // 904
    multi: true                                                                                                        // 905
  });                                                                                                                  // 905
  console.log("tagging credentials...");                                                                               // 907
  Meteor.users.update({                                                                                                // 908
    type: {                                                                                                            // 908
      $exists: false                                                                                                   // 908
    },                                                                                                                 // 908
    profile: {                                                                                                         // 908
      $exists: true                                                                                                    // 908
    }                                                                                                                  // 908
  }, {                                                                                                                 // 908
    $set: {                                                                                                            // 909
      type: "credential"                                                                                               // 909
    }                                                                                                                  // 909
  }, {                                                                                                                 // 909
    multi: true                                                                                                        // 910
  }); // Map each identity ID to the list of connected accounts.                                                       // 910
                                                                                                                       //
  console.log("building identity map...");                                                                             // 913
  var identityToAccount = {};                                                                                          // 914
  var needSort = [];                                                                                                   // 915
  Meteor.users.find({                                                                                                  // 916
    type: "account"                                                                                                    // 916
  }).forEach(function (user) {                                                                                         // 916
    var userInfo = {                                                                                                   // 917
      id: user._id,                                                                                                    // 917
      lastActive: user.lastActive || user.createdAt                                                                    // 917
    };                                                                                                                 // 917
                                                                                                                       //
    function handleIdentity(identity) {                                                                                // 919
      if (!identityToAccount[identity.id]) {                                                                           // 920
        identityToAccount[identity.id] = [userInfo];                                                                   // 921
      } else {                                                                                                         // 922
        var list = identityToAccount[identity.id];                                                                     // 923
        if (list.length == 1) needSort.push(list);                                                                     // 924
        list.push(userInfo);                                                                                           // 925
      }                                                                                                                // 926
    }                                                                                                                  // 927
                                                                                                                       //
    user.loginIdentities.forEach(handleIdentity);                                                                      // 928
    user.nonloginIdentities.forEach(handleIdentity);                                                                   // 929
  }); // For identities attached to multiple accounts, we want the most-recently-active account to sort                // 930
  // first. In cases where we need to replace an identity with exactly one account, we'll use that                     // 933
  // first option.                                                                                                     // 934
                                                                                                                       //
  needSort.forEach(function (item) {                                                                                   // 935
    item.sort(function (a, b) {                                                                                        // 936
      if (a.lastActive > b.lastActive) {                                                                               // 937
        return -1;                                                                                                     // 938
      } else if (a.lastActive < b.lastActive) {                                                                        // 939
        return 1;                                                                                                      // 940
      } else {                                                                                                         // 941
        return 0;                                                                                                      // 942
      }                                                                                                                // 943
    });                                                                                                                // 944
  });                                                                                                                  // 945
                                                                                                                       //
  function accountForIdentity(identityId) {                                                                            // 947
    var ids = identityToAccount[identityId];                                                                           // 948
                                                                                                                       //
    if (ids) {                                                                                                         // 949
      return ids[0].id;                                                                                                // 950
    } else {                                                                                                           // 951
      console.error("no such identity:", identityId);                                                                  // 952
      return "invalid-" + identityId;                                                                                  // 953
    }                                                                                                                  // 954
  }                                                                                                                    // 955
                                                                                                                       //
  function accountListForIdentity(identityId) {                                                                        // 957
    var ids = identityToAccount[identityId];                                                                           // 958
                                                                                                                       //
    if (ids) {                                                                                                         // 959
      return ids.map(function (i) {                                                                                    // 960
        return i.id;                                                                                                   // 960
      });                                                                                                              // 960
    } else {                                                                                                           // 961
      console.error("no such identity:", identityId);                                                                  // 962
      return ["invalid-" + identityId];                                                                                // 963
    }                                                                                                                  // 964
  }                                                                                                                    // 965
                                                                                                                       //
  forEachProgress("migrating users...", Meteor.users.find({                                                            // 967
    type: "account"                                                                                                    // 968
  }), function (user) {                                                                                                // 968
    // Fetch all the user's login identities.                                                                          // 970
    var identities = Meteor.users.find({                                                                               // 971
      _id: {                                                                                                           // 972
        $in: user.loginIdentities.map(function (identity) {                                                            // 972
          return identity.id;                                                                                          // 972
        })                                                                                                             // 972
      },                                                                                                               // 972
      profile: {                                                                                                       // 973
        $exists: true                                                                                                  // 973
      }                                                                                                                // 973
    }).fetch(); // Fill out the profiles for each identity.                                                            // 972
                                                                                                                       //
    identities.forEach(function (identity) {                                                                           // 976
      SandstormDb.fillInProfileDefaults(identity, identity.profile);                                                   // 977
    }); // Find the best profile among them.                                                                           // 978
                                                                                                                       //
    var profile = void 0;                                                                                              // 981
                                                                                                                       //
    if (identities.length == 0) {                                                                                      // 982
      // no profiles???                                                                                                // 983
      profile = null;                                                                                                  // 984
    } else if (identities.length == 1) {                                                                               // 985
      profile = identities[0].profile;                                                                                 // 986
    } else {                                                                                                           // 987
      // Multi-identity user. Try to find the "best" identity.                                                         // 988
      var maxScore = -1;                                                                                               // 990
      identities.forEach(function (identity) {                                                                         // 992
        // Count total grains using this identity.                                                                     // 993
        var score = db.collections.grains.find({                                                                       // 994
          userId: user._id,                                                                                            // 995
          identityId: identity._id                                                                                     // 995
        }).count() + db.collections.apiTokens.find({                                                                   // 995
          "owner.user.identityId": identity._id                                                                        // 996
        }).count(); // Avoid choosing demo user, unless they've really used it a lot.                                  // 996
                                                                                                                       //
        if (identity.profile.name !== "Demo User") {                                                                   // 999
          score += 10;                                                                                                 // 1000
        }                                                                                                              // 1001
                                                                                                                       //
        if (score > maxScore) {                                                                                        // 1003
          profile = identity.profile;                                                                                  // 1004
          maxScore = score;                                                                                            // 1005
        }                                                                                                              // 1006
      });                                                                                                              // 1007
    }                                                                                                                  // 1008
                                                                                                                       //
    if (profile) {                                                                                                     // 1010
      delete profile.service;                                                                                          // 1011
    } else {                                                                                                           // 1012
      console.warn("no suitable profile found for user account:", user._id);                                           // 1013
      profile = {                                                                                                      // 1014
        name: "Unknown",                                                                                               // 1015
        handle: "unknown",                                                                                             // 1016
        pronoun: "neutral",                                                                                            // 1017
        identicon: Crypto.randomBytes(32).toString("hex")                                                              // 1018
      };                                                                                                               // 1014
    }                                                                                                                  // 1020
                                                                                                                       //
    var mod = {                                                                                                        // 1022
      profile: profile                                                                                                 // 1022
    }; // Also figure out referrals.                                                                                   // 1022
                                                                                                                       //
    var referrers = Meteor.users.find({                                                                                // 1025
      _id: {                                                                                                           // 1026
        $in: getUserIdentityIds(user)                                                                                  // 1026
      },                                                                                                               // 1026
      referredBy: {                                                                                                    // 1026
        $exists: true                                                                                                  // 1026
      }                                                                                                                // 1026
    }, {                                                                                                               // 1026
      fields: {                                                                                                        // 1027
        referredBy: 1                                                                                                  // 1027
      }                                                                                                                // 1027
    }).map(function (id) {                                                                                             // 1027
      return id.referredBy;                                                                                            // 1028
    });                                                                                                                // 1028
                                                                                                                       //
    if (referrers.length > 0) {                                                                                        // 1029
      mod.referredBy = referrers[0];                                                                                   // 1030
    }                                                                                                                  // 1031
                                                                                                                       //
    if (user.referredIdentityIds) {                                                                                    // 1033
      mod.referredAccountIds = user.referredIdentityIds.map(accountForIdentity);                                       // 1034
    }                                                                                                                  // 1035
                                                                                                                       //
    Meteor.users.update({                                                                                              // 1037
      _id: user._id                                                                                                    // 1037
    }, {                                                                                                               // 1037
      $set: mod                                                                                                        // 1037
    });                                                                                                                // 1037
  });                                                                                                                  // 1038
  forEachProgress("migrating ApiTokens...", db.collections.apiTokens.find({                                            // 1040
    "owner.user.identityId": {                                                                                         // 1041
      $exists: true                                                                                                    // 1041
    }                                                                                                                  // 1041
  }), function (token) {                                                                                               // 1041
    var accounts = accountListForIdentity(token.owner.user.identityId);                                                // 1043
    db.collections.apiTokens.update({                                                                                  // 1045
      _id: token._id                                                                                                   // 1045
    }, {                                                                                                               // 1045
      $set: {                                                                                                          // 1046
        "owner.user.accountId": accounts[0]                                                                            // 1046
      }                                                                                                                // 1046
    });                                                                                                                // 1046
                                                                                                                       //
    if (accounts.length > 1 && token.grainId && token.accountId) {                                                     // 1048
      // Shared to an identity that has multiple accounts. We need to denormalize the share to                         // 1049
      // target each account individually.                                                                             // 1050
      //                                                                                                               // 1051
      // Note: A token with owner.user.identityId is always a UiView share token and so always has                     // 1052
      //   accountId and grainId. There are zero counterexamples in Oasis. But if a counterexample                     // 1053
      //   existed, the code below might compound the confusion, so we skip it.                                        // 1054
      delete token._id;                                                                                                // 1056
      accounts.slice(1).forEach(function (account) {                                                                   // 1057
        // For idempotency purposes, don't insert if a similar token already exists.                                   // 1058
        if (!db.collections.apiTokens.findOne({                                                                        // 1059
          grainId: token.grainId,                                                                                      // 1060
          accountId: token.accountId,                                                                                  // 1061
          "owner.user.accountId": account                                                                              // 1062
        })) {                                                                                                          // 1059
          token.owner.user.accountId = account;                                                                        // 1064
          db.collections.apiTokens.insert(token);                                                                      // 1065
        }                                                                                                              // 1066
      });                                                                                                              // 1067
    }                                                                                                                  // 1068
  });                                                                                                                  // 1069
  forEachProgress("migrating membrane requirements...", db.collections.apiTokens.find({                                // 1071
    "requirements.permissionsHeld.identityId": {                                                                       // 1072
      $exists: true                                                                                                    // 1072
    }                                                                                                                  // 1072
  }), function (token) {                                                                                               // 1072
    token.requirements.forEach(function (requirement) {                                                                // 1074
      if (requirement.permissionsHeld && requirement.permissionsHeld.identityId) {                                     // 1075
        requirement.permissionsHeld.accountId = accountForIdentity(requirement.permissionsHeld.identityId);            // 1076
      }                                                                                                                // 1078
    });                                                                                                                // 1079
    db.collections.apiTokens.update({                                                                                  // 1081
      _id: token._id                                                                                                   // 1081
    }, {                                                                                                               // 1081
      $set: {                                                                                                          // 1082
        requirements: token.requirements                                                                               // 1082
      }                                                                                                                // 1082
    });                                                                                                                // 1082
  });                                                                                                                  // 1083
  forEachProgress("migrating identity capabilities...", db.collections.apiTokens.find({                                // 1085
    "frontendRef.identity": {                                                                                          // 1086
      $exists: true                                                                                                    // 1086
    }                                                                                                                  // 1086
  }), function (token) {                                                                                               // 1086
    db.collections.apiTokens.update({                                                                                  // 1088
      _id: token._id                                                                                                   // 1088
    }, {                                                                                                               // 1088
      $set: {                                                                                                          // 1089
        "frontendRef.identity": accountForIdentity(token.frontendRef.identity)                                         // 1089
      }                                                                                                                // 1089
    });                                                                                                                // 1089
  });                                                                                                                  // 1090
  forEachProgress("migrating contacts...", db.collections.contacts.find(), function (contact) {                        // 1092
    db.collections.contacts.update({                                                                                   // 1095
      _id: contact._id                                                                                                 // 1095
    }, {                                                                                                               // 1095
      $set: {                                                                                                          // 1096
        accountId: accountForIdentity(contact.identityId)                                                              // 1096
      }                                                                                                                // 1096
    });                                                                                                                // 1096
  });                                                                                                                  // 1097
  forEachProgress("migrating notifications...", db.collections.notifications.find({                                    // 1099
    initiatingIdentity: {                                                                                              // 1100
      $exists: true                                                                                                    // 1100
    }                                                                                                                  // 1100
  }), function (notification) {                                                                                        // 1100
    db.collections.notifications.update({                                                                              // 1102
      _id: notification._id                                                                                            // 1102
    }, {                                                                                                               // 1102
      $set: {                                                                                                          // 1103
        initiatingAccount: accountForIdentity(notification.initiatingIdentity)                                         // 1103
      }                                                                                                                // 1103
    });                                                                                                                // 1103
  });                                                                                                                  // 1104
  forEachProgress("migrating desktop notifications...", db.collections.desktopNotifications.find({                     // 1106
    "appActivity.user.identityId": {                                                                                   // 1108
      $exists: true                                                                                                    // 1108
    }                                                                                                                  // 1108
  }), function (notification) {                                                                                        // 1108
    db.collections.desktopNotifications.update({                                                                       // 1110
      _id: notification._id                                                                                            // 1110
    }, {                                                                                                               // 1110
      $set: {                                                                                                          // 1111
        "appActivity.user.accountId": accountForIdentity(notification.appActivity.user.identityId)                     // 1111
      }                                                                                                                // 1111
    });                                                                                                                // 1111
  });                                                                                                                  // 1113
  forEachProgress("migrating subscriptions...", db.collections.activitySubscriptions.find({                            // 1115
    identityId: {                                                                                                      // 1116
      $exists: true                                                                                                    // 1116
    }                                                                                                                  // 1116
  }), function (subscription) {                                                                                        // 1116
    db.collections.activitySubscriptions.update({                                                                      // 1118
      _id: subscription._id                                                                                            // 1118
    }, {                                                                                                               // 1118
      $set: {                                                                                                          // 1119
        accountId: accountForIdentity(subscription.identityId)                                                         // 1119
      }                                                                                                                // 1119
    });                                                                                                                // 1119
  });                                                                                                                  // 1120
}                                                                                                                      // 1121
                                                                                                                       //
function onePersonaPerAccountPostCleanup(db, backend) {                                                                // 1123
  // Drop obsolete indices.                                                                                            // 1124
  db.collections.apiTokens.ensureDroppedIndex({                                                                        // 1125
    "owner.user.identityId": 1                                                                                         // 1125
  });                                                                                                                  // 1125
  db.collections.activitySubscriptions.ensureDroppedIndex({                                                            // 1126
    "identityId": 1                                                                                                    // 1126
  });                                                                                                                  // 1126
  Meteor.users.ensureDroppedIndex({                                                                                    // 1127
    "loginIdentities.id": 1                                                                                            // 1127
  });                                                                                                                  // 1127
  Meteor.users.ensureDroppedIndex({                                                                                    // 1128
    "nonloginIdentities.id": 1                                                                                         // 1128
  });                                                                                                                  // 1128
  Meteor.users.update({                                                                                                // 1130
    type: "account"                                                                                                    // 1130
  }, {                                                                                                                 // 1130
    $rename: {                                                                                                         // 1131
      loginIdentities: "loginCredentials",                                                                             // 1131
      nonloginIdentities: "nonloginCredentials"                                                                        // 1132
    },                                                                                                                 // 1131
    $unset: {                                                                                                          // 1133
      referredIdentityIds: 1                                                                                           // 1133
    }                                                                                                                  // 1133
  }, {                                                                                                                 // 1131
    multi: true                                                                                                        // 1134
  }); // Note that we intentionally don't unset profiles from credentials for now, as this is a place                  // 1134
  // where we could legitimately be losing data since only of the user's identity profiles gets                        // 1137
  // promoted to their account. We could remove them in a later pass, once we're sure we no longer                     // 1138
  // need that data. (Also unclear if Meteor's accounts system will get confused if profiles                           // 1139
  // disappear...) (Also SansdtormDb.fillInIntrinsicName() is specific to credentials and still                        // 1140
  // reads the profile a bit...)                                                                                       // 1141
                                                                                                                       //
  db.collections.notifications.update({}, {                                                                            // 1143
    $unset: {                                                                                                          // 1144
      initiatingIdentity: 1                                                                                            // 1144
    }                                                                                                                  // 1144
  }, {                                                                                                                 // 1144
    multi: true                                                                                                        // 1145
  });                                                                                                                  // 1145
  db.collections.apiTokens.update({                                                                                    // 1147
    identityId: {                                                                                                      // 1147
      $exists: true                                                                                                    // 1147
    }                                                                                                                  // 1147
  }, {                                                                                                                 // 1147
    $unset: {                                                                                                          // 1148
      identityId: 1                                                                                                    // 1148
    }                                                                                                                  // 1148
  }, {                                                                                                                 // 1148
    multi: true                                                                                                        // 1149
  });                                                                                                                  // 1149
  db.collections.apiTokens.update({                                                                                    // 1151
    "requirements.permissionsHeld.identityId": {                                                                       // 1151
      $exists: true                                                                                                    // 1151
    }                                                                                                                  // 1151
  }, {                                                                                                                 // 1151
    $unset: {                                                                                                          // 1152
      "requirements.permissionsHeld.identityId": 1                                                                     // 1152
    }                                                                                                                  // 1152
  }, {                                                                                                                 // 1152
    multi: true                                                                                                        // 1153
  });                                                                                                                  // 1153
} // TODO(cleanup): Delete profiles from credentials. (Make sure nothing depends on them.)                             // 1154
// TODO(cleanup): Delete all demo credentials since they aren't really needed anymore. Remove them                     // 1157
//   from associated account nonloginCredentials.                                                                      // 1158
// This must come after all the functions named within are defined.                                                    // 1160
// Only append to this list!  Do not modify or remove list entries;                                                    // 1161
// doing so is likely change the meaning and semantics of user databases.                                              // 1162
                                                                                                                       //
                                                                                                                       //
var MIGRATIONS = [updateLoginStyleToRedirect, enableLegacyOAuthProvidersIfNotInSettings, denormalizeInviteInfo, mergeRoleAssignmentsIntoApiTokens, fixOasisStorageUsageStats, fetchProfilePictures, assignPlans, removeKeyrings, useLocalizedTextInUserActions, verifyAllPgpSignatures, splitUserIdsIntoAccountIdsAndIdentityIds, appUpdateSettings, moveDevAndEmailLoginDataIntoIdentities, repairEmailIdentityIds, splitAccountUsersAndIdentityUsers, populateContactsFromApiTokens, cleanUpApiTokens, initServerTitleAndReturnAddress, sendReferralNotifications, assignBonuses, splitSmtpUrl, smtpPortShouldBeNumber, consolidateOrgSettings, unsetSmtpDefaultHostnameIfNoUsersExist, extractLastUsedFromApiTokenOwner, setUpstreamTitles, markAllRead, clearAppIndex, assignEmailVerifierIds, setNewServer, addMembraneRequirementsToIdentities, addEncryptionToFrontendRefIpNetwork, setIpBlacklist, notifyIdentityChanges, onePersonaPerAccountPreCleanup, onePersonaPerAccount, onePersonaPerAccountPostCleanup];
var NEW_SERVER_STARTUP = [startPreinstallingApps];                                                                     // 1203
                                                                                                                       //
var migrateToLatest = function (db, backend) {                                                                         // 1207
  if (Meteor.settings.replicaNumber) {                                                                                 // 1208
    // This is a replica. Wait for the first replica to perform migrations.                                            // 1209
    console.log("Waiting for migrations on replica zero...");                                                          // 1211
    var done = new Future();                                                                                           // 1213
                                                                                                                       //
    var change = function (doc) {                                                                                      // 1214
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                             // 1215
      if (doc.value >= MIGRATIONS.length) done.return();                                                               // 1216
    };                                                                                                                 // 1217
                                                                                                                       //
    var observer = db.collections.migrations.find({                                                                    // 1219
      _id: "migrations_applied"                                                                                        // 1219
    }).observe({                                                                                                       // 1219
      added: change,                                                                                                   // 1220
      changed: change                                                                                                  // 1221
    });                                                                                                                // 1219
    var newServerDone = new Future();                                                                                  // 1224
                                                                                                                       //
    var newServerChange = function (doc) {                                                                             // 1225
      if (doc.value) {                                                                                                 // 1226
        console.log("New server migrations applied elsewhere");                                                        // 1227
        newServerDone.return();                                                                                        // 1228
      }                                                                                                                // 1229
    };                                                                                                                 // 1230
                                                                                                                       //
    var newServerObserver = db.collections.migrations.find({                                                           // 1232
      _id: "new_server_migrations_applied"                                                                             // 1232
    }).observe({                                                                                                       // 1232
      added: newServerChange,                                                                                          // 1233
      changed: newServerChange                                                                                         // 1234
    });                                                                                                                // 1232
    done.wait();                                                                                                       // 1237
    observer.stop();                                                                                                   // 1238
    newServerDone.wait();                                                                                              // 1239
    newServerObserver.stop();                                                                                          // 1240
    console.log("Migrations have completed on replica zero.");                                                         // 1241
  } else {                                                                                                             // 1242
    var applied = db.collections.migrations.findOne({                                                                  // 1243
      _id: "migrations_applied"                                                                                        // 1243
    });                                                                                                                // 1243
    var start = void 0;                                                                                                // 1244
                                                                                                                       //
    if (!applied) {                                                                                                    // 1245
      // Migrations table is not yet seeded with a value.  This means it has                                           // 1246
      // applied 0 migrations.  Persist this.                                                                          // 1247
      db.collections.migrations.insert({                                                                               // 1248
        _id: "migrations_applied",                                                                                     // 1248
        value: 0                                                                                                       // 1248
      });                                                                                                              // 1248
      start = 0; // This also means this is a brand new server                                                         // 1249
                                                                                                                       //
      db.collections.migrations.insert({                                                                               // 1252
        _id: "new_server_migrations_applied",                                                                          // 1252
        value: false                                                                                                   // 1252
      });                                                                                                              // 1252
    } else {                                                                                                           // 1253
      start = applied.value;                                                                                           // 1254
    }                                                                                                                  // 1255
                                                                                                                       //
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                                     // 1257
                                                                                                                       //
    for (var i = start; i < MIGRATIONS.length; i++) {                                                                  // 1259
      // Apply migration i, then record that migration i was successfully run.                                         // 1260
      console.log("Applying migration " + (i + 1));                                                                    // 1261
      MIGRATIONS[i](db, backend);                                                                                      // 1262
      db.collections.migrations.update({                                                                               // 1263
        _id: "migrations_applied"                                                                                      // 1263
      }, {                                                                                                             // 1263
        $set: {                                                                                                        // 1263
          value: i + 1                                                                                                 // 1263
        }                                                                                                              // 1263
      });                                                                                                              // 1263
      console.log("Applied migration " + (i + 1));                                                                     // 1264
    }                                                                                                                  // 1265
                                                                                                                       //
    if (!db.collections.migrations.findOne({                                                                           // 1267
      _id: "new_server_migrations_applied"                                                                             // 1267
    }).value) {                                                                                                        // 1267
      // new_server_migrations_applied is guaranteed to exist since we have a migration that                           // 1268
      // ensures it.                                                                                                   // 1269
      for (var _i = 0; _i < NEW_SERVER_STARTUP.length; _i++) {                                                         // 1270
        console.log("Running new server startup function " + (_i + 1));                                                // 1271
                                                                                                                       //
        NEW_SERVER_STARTUP[_i](db, backend);                                                                           // 1272
                                                                                                                       //
        console.log("Running new server startup function " + (_i + 1));                                                // 1273
      }                                                                                                                // 1274
                                                                                                                       //
      db.collections.migrations.update({                                                                               // 1276
        _id: "new_server_migrations_applied"                                                                           // 1276
      }, {                                                                                                             // 1276
        $set: {                                                                                                        // 1276
          value: true                                                                                                  // 1276
        }                                                                                                              // 1276
      });                                                                                                              // 1276
    } // Start background migrations.                                                                                  // 1277
                                                                                                                       //
                                                                                                                       //
    backgroundFillInGrainSizes(db, backend);                                                                           // 1280
  }                                                                                                                    // 1281
};                                                                                                                     // 1282
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"networking.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/networking.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var module1 = module;                                                                                                  // 1
module1.export({                                                                                                       // 1
  ssrfSafeLookup: function () {                                                                                        // 1
    return ssrfSafeLookup;                                                                                             // 1
  },                                                                                                                   // 1
  ssrfSafeLookupOrProxy: function () {                                                                                 // 1
    return ssrfSafeLookupOrProxy;                                                                                      // 1
  },                                                                                                                   // 1
  monkeyPatchHttp: function () {                                                                                       // 1
    return monkeyPatchHttp;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var SPECIAL_IPV4_ADDRESSES = void 0,                                                                                   // 1
    SPECIAL_IPV6_ADDRESSES = void 0;                                                                                   // 1
module1.watch(require("/imports/constants.js"), {                                                                      // 1
  SPECIAL_IPV4_ADDRESSES: function (v) {                                                                               // 1
    SPECIAL_IPV4_ADDRESSES = v;                                                                                        // 1
  },                                                                                                                   // 1
  SPECIAL_IPV6_ADDRESSES: function (v) {                                                                               // 1
    SPECIAL_IPV6_ADDRESSES = v;                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Dns = Npm.require("dns");                                                                                          // 19
                                                                                                                       //
var Ip = Npm.require("ip");                                                                                            // 20
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 21
                                                                                                                       //
var lookupInFiber = Meteor.wrapAsync(Dns.lookup, Dns);                                                                 // 23
                                                                                                                       //
function getNetwork(addr, bits) {                                                                                      // 25
  // npm ip's "mask" and "cidr" functions are broken for ipv6. :(                                                      // 26
  var parsed = Ip.toBuffer(addr);                                                                                      // 28
                                                                                                                       //
  for (var i = Math.ceil(bits / 8); i < parsed.length; i++) {                                                          // 30
    parsed[i] = 0;                                                                                                     // 31
  }                                                                                                                    // 32
                                                                                                                       //
  var n = Math.floor(bits / 8);                                                                                        // 34
                                                                                                                       //
  if (n < parsed.length) {                                                                                             // 35
    parsed[n] = parsed[n] & 0xff << 8 - bits % 8;                                                                      // 36
  }                                                                                                                    // 37
                                                                                                                       //
  return parsed;                                                                                                       // 39
}                                                                                                                      // 40
                                                                                                                       //
function parseCidr(cidr) {                                                                                             // 42
  // Given a CIDR-format network specification, return a function which, given an address string,                      // 43
  // returns true if the address is in this network.                                                                   // 44
  // The "ip" NPM module's CIDR handling unfortunately is very broken for IPv6. Many bugs have been                    // 46
  // filed but it remains broken.                                                                                      // 47
  try {                                                                                                                // 49
    cidr = cidr.trim();                                                                                                // 50
    if (cidr === "") return null;                                                                                      // 51
    var parts = cidr.split("/");                                                                                       // 53
                                                                                                                       //
    if (parts.length === 1) {                                                                                          // 54
      // Bare address.                                                                                                 // 55
      return function (addr) {                                                                                         // 56
        return Ip.isEqual(cidr, addr);                                                                                 // 56
      };                                                                                                               // 56
    } else if (parts.length === 2) {                                                                                   // 57
      var bits = parseInt(parts[1], 10);                                                                               // 58
      if (bits !== bits) throw new Error("value after slash must be an integer");                                      // 59
      var network = getNetwork(parts[0], bits);                                                                        // 60
      return function (addr) {                                                                                         // 61
        return network.equals(getNetwork(addr, bits));                                                                 // 62
      };                                                                                                               // 63
    } else {                                                                                                           // 64
      throw new Error("too many slashes");                                                                             // 65
    }                                                                                                                  // 66
  } catch (err) {                                                                                                      // 67
    console.error("invalid network specification in IP blacklist:", cidr, err);                                        // 68
    return null;                                                                                                       // 69
  }                                                                                                                    // 70
}                                                                                                                      // 71
                                                                                                                       //
SPECIAL_FILTERS = SPECIAL_IPV4_ADDRESSES.concat(SPECIAL_IPV6_ADDRESSES).map(parseCidr);                                // 73
                                                                                                                       //
function ssrfSafeLookup(db, url) {                                                                                     // 75
  // Given an HTTP/HTTPS URL, look up the hostname, verify it doesn't point to a blacklisted IP,                       // 76
  // then return an object of {url, host}, where `url` has the original hostname substituted with                      // 77
  // an IP address, and `host` is the original hostname suitable for sending in the `Host` header.                     // 78
  var parsedUrl = Url.parse(url);                                                                                      // 80
                                                                                                                       //
  if (parsedUrl.protocol !== "http:" && parsedUrl.protocol !== "https:") {                                             // 82
    throw new Error("not an HTTP nor HTTPS URL: " + url);                                                              // 83
  }                                                                                                                    // 84
                                                                                                                       //
  var addresses = lookupInFiber(parsedUrl.hostname, {                                                                  // 86
    all: true,                                                                                                         // 86
    hints: Dns.ADDRCONFIG                                                                                              // 86
  }); // TODO(perf): Subscribe to blacklist changes so that we don't have to do a new lookup and                       // 86
  //   parse each time.                                                                                                // 89
                                                                                                                       //
  var blacklist = db.getSettingWithFallback("ipBlacklist", "").split("\n").map(parseCidr).filter(function (x) {        // 90
    return x;                                                                                                          // 91
  });                                                                                                                  // 91
                                                                                                                       //
  var _loop = function (i) {                                                                                           // 75
    var address = addresses[i];                                                                                        // 94
    if (address.family !== 4 && address.family !== 6) return "continue";                                               // 95
    var ok = true;                                                                                                     // 97
    blacklist.forEach(function (test) {                                                                                // 98
      if (test(address.address)) {                                                                                     // 98
        ok = false;                                                                                                    // 98
      }                                                                                                                // 98
    });                                                                                                                // 98
    SPECIAL_FILTERS.forEach(function (test) {                                                                          // 100
      if (test(address.address)) {                                                                                     // 100
        ok = false;                                                                                                    // 100
      }                                                                                                                // 100
    });                                                                                                                // 100
                                                                                                                       //
    if (ok) {                                                                                                          // 102
      var host = parsedUrl.host;                                                                                       // 103
      delete parsedUrl.host;                                                                                           // 104
      parsedUrl.hostname = address.address;                                                                            // 105
      return {                                                                                                         // 106
        v: {                                                                                                           // 106
          url: Url.format(parsedUrl),                                                                                  // 106
          host: host                                                                                                   // 106
        }                                                                                                              // 106
      };                                                                                                               // 106
    }                                                                                                                  // 107
  };                                                                                                                   // 75
                                                                                                                       //
  for (var i in meteorBabelHelpers.sanitizeForInObject(addresses)) {                                                   // 93
    var _ret = _loop(i);                                                                                               // 93
                                                                                                                       //
    switch (_ret) {                                                                                                    // 93
      case "continue":                                                                                                 // 93
        continue;                                                                                                      // 95
                                                                                                                       //
      default:                                                                                                         // 93
        if ((typeof _ret === "undefined" ? "undefined" : (0, _typeof3.default)(_ret)) === "object") return _ret.v;     // 93
    }                                                                                                                  // 93
  }                                                                                                                    // 108
                                                                                                                       //
  if (addresses.length > 0) {                                                                                          // 110
    throw new Meteor.Error(403, "can't connect to blacklisted private network address: " + parsedUrl.hostname + "; the Sandstorm server admin can change the blacklist in the admin settings");
  } else {                                                                                                             // 114
    throw new Meteor.Error(404, "host not found: " + parsedUrl.hostname);                                              // 115
  }                                                                                                                    // 116
}                                                                                                                      // 117
                                                                                                                       //
function ssrfSafeLookupOrProxy(db, url) {                                                                              // 119
  // If there is an HTTP proxy, then it will have to do the work of blacklisting IPs, because it's                     // 120
  // the proxy that does the DNS lookup.                                                                               // 121
  var httpProxy = process.env.HTTP_PROXY || process.env.http_proxy;                                                    // 122
  var httpsProxy = process.env.HTTPS_PROXY || process.env.https_proxy;                                                 // 123
                                                                                                                       //
  if (httpProxy && url.startsWith("http:")) {                                                                          // 125
    return {                                                                                                           // 126
      proxy: httpProxy                                                                                                 // 126
    };                                                                                                                 // 126
  } else if (httpsProxy && url.startsWith("https:")) {                                                                 // 127
    return {                                                                                                           // 128
      proxy: httpsProxy                                                                                                // 128
    };                                                                                                                 // 128
  } else {                                                                                                             // 129
    return ssrfSafeLookup(db, url);                                                                                    // 130
  }                                                                                                                    // 131
}                                                                                                                      // 132
                                                                                                                       //
function ssrfSafeHttp(originalHttpCall, db, method, url, options, callback) {                                          // 134
  if (typeof options === "function") {                                                                                 // 135
    callback = options;                                                                                                // 136
    options = undefined;                                                                                               // 137
  }                                                                                                                    // 138
                                                                                                                       //
  if (!options) options = {};                                                                                          // 140
                                                                                                                       //
  if (options.npmRequestOptions && options.npmRequestOptions.proxy) {                                                  // 142
    // Request already specifies a different proxy.                                                                    // 143
    return originalHttpCall(method, url, options, callback);                                                           // 144
  }                                                                                                                    // 145
                                                                                                                       //
  var safe = ssrfSafeLookupOrProxy(db, url);                                                                           // 147
                                                                                                                       //
  if (safe.proxy) {                                                                                                    // 149
    if (!options.npmRequestOptions) options.npmRequestOptions = {};                                                    // 150
    options.npmRequestOptions.proxy = safe.proxy;                                                                      // 151
    return originalHttpCall(method, url, options, callback);                                                           // 152
  } else {                                                                                                             // 153
    var _safe = ssrfSafeLookup(db, url);                                                                               // 154
                                                                                                                       //
    if (!options.headers) options.headers = {};                                                                        // 155
    options.headers.host = _safe.host;                                                                                 // 156
    options.servername = _safe.host.split(":")[0];                                                                     // 157
    return originalHttpCall(method, _safe.url, options, callback);                                                     // 158
  }                                                                                                                    // 159
}                                                                                                                      // 160
                                                                                                                       //
function monkeyPatchHttp(db, HTTP) {                                                                                   // 162
  var original = HTTP.call.bind(HTTP);                                                                                 // 163
  HTTP.call = ssrfSafeHttp.bind(this, original, db);                                                                   // 164
}                                                                                                                      // 165
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"observe-helpers.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/observe-helpers.js                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  fillUndefinedForChangedDoc: function () {                                                                            // 1
    return fillUndefinedForChangedDoc;                                                                                 // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
function fillUndefinedForChangedDoc(newDoc, oldDoc) {                                                                  // 17
  // In publishes, setting a field to `undefined` specifies that it is no longer present.                              // 18
  // Annoyingly, in the changed callback of observing a collection, it does not do this for you.                       // 19
  Object.keys(oldDoc).forEach(function (key) {                                                                         // 20
    if (!(key in newDoc)) {                                                                                            // 21
      newDoc[key] = undefined;                                                                                         // 22
    }                                                                                                                  // 23
  });                                                                                                                  // 24
}                                                                                                                      // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"persistent.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/persistent.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  PersistentImpl: function () {                                                                                        // 1
    return PersistentImpl;                                                                                             // 1
  },                                                                                                                   // 1
  hashSturdyRef: function () {                                                                                         // 1
    return hashSturdyRef;                                                                                              // 1
  },                                                                                                                   // 1
  generateSturdyRef: function () {                                                                                     // 1
    return generateSturdyRef;                                                                                          // 1
  },                                                                                                                   // 1
  checkRequirements: function () {                                                                                     // 1
    return checkRequirements;                                                                                          // 1
  },                                                                                                                   // 1
  fetchApiToken: function () {                                                                                         // 1
    return fetchApiToken;                                                                                              // 1
  },                                                                                                                   // 1
  insertApiToken: function () {                                                                                        // 1
    return insertApiToken;                                                                                             // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var inMeteor = void 0;                                                                                                 // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 19
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var privateDb = Symbol("PersistentImpl.db");                                                                           // 22
var privateTemplate = Symbol("PersistentImpl.template");                                                               // 23
var privateIsSaved = Symbol("PersistentImpl.isSaved");                                                                 // 24
                                                                                                                       //
var PersistentImpl = function () {                                                                                     //
  function PersistentImpl(db, saveTemplate) {                                                                          // 27
    (0, _classCallCheck3.default)(this, PersistentImpl);                                                               // 27
    this[privateDb] = db;                                                                                              // 28
    this[privateTemplate] = saveTemplate; // TODO(someday): This may or may not be quite right in cases where we're saving a root token
    //   by copy. Is the copy the "same" capability or not? Today, isSaved() is only used by                           // 32
    //   wakelock notification handles which are not issued through the powerbox, therefore do not                     // 33
    //   go through claimRequest(), therefore are not saved by copy, so the point is moot.                             // 34
                                                                                                                       //
    this[privateIsSaved] = !!saveTemplate.parentToken;                                                                 // 35
  }                                                                                                                    // 36
                                                                                                                       //
  PersistentImpl.prototype.isSaved = function () {                                                                     //
    function isSaved() {                                                                                               //
      // Can be called by subclasses to ask whether or not a record currently exists in the ApiTokens                  // 39
      // table representing this capability.                                                                           // 40
      return this[privateIsSaved];                                                                                     // 42
    }                                                                                                                  // 43
                                                                                                                       //
    return isSaved;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  PersistentImpl.prototype.save = function () {                                                                        //
    function save(params) {                                                                                            //
      var _this = this;                                                                                                // 45
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 46
        if (!params.sealFor) {                                                                                         // 47
          throw new Error("must specify 'sealFor'");                                                                   // 48
        }                                                                                                              // 49
                                                                                                                       //
        var db = _this[privateDb];                                                                                     // 51
                                                                                                                       //
        var newToken = _.clone(_this[privateTemplate]);                                                                // 53
                                                                                                                       //
        newToken.owner = params.sealFor;                                                                               // 54
                                                                                                                       //
        if (newToken.owner.user) {                                                                                     // 55
          if (!newToken.accountId || !newToken.grainId) {                                                              // 56
            throw new Error("can't save non-UiView with user as owner");                                               // 57
          } // Only "accountId" and "title" are allowed to be passed to save().                                        // 58
                                                                                                                       //
                                                                                                                       //
          var userOwner = _.pick(newToken.owner.user, "accountId", "title");                                           // 61
                                                                                                                       //
          var grain = db.getGrain(newToken.grainId);                                                                   // 63
                                                                                                                       //
          if (!grain) {                                                                                                // 64
            throw new Error("unknown grain ID");                                                                       // 65
          }                                                                                                            // 66
                                                                                                                       //
          userOwner.identityId = db.getOrGenerateIdentityId(userOwner.accountId, grain); // Fill in denormalizedGrainMetadata and upstreamTitle ourselves.
                                                                                                                       //
          userOwner.denormalizedGrainMetadata = db.getDenormalizedGrainInfo(newToken.grainId);                         // 71
                                                                                                                       //
          if (grain.title !== userOwner.title) {                                                                       // 73
            userOwner.upstreamTitle = grain.title;                                                                     // 74
          }                                                                                                            // 75
                                                                                                                       //
          newToken.owner.user = userOwner;                                                                             // 77
        }                                                                                                              // 78
                                                                                                                       //
        newToken.created = new Date();                                                                                 // 80
        var sturdyRef = insertApiToken(db, newToken);                                                                  // 81
        _this[privateIsSaved] = true;                                                                                  // 83
        return {                                                                                                       // 84
          sturdyRef: new Buffer(sturdyRef)                                                                             // 84
        };                                                                                                             // 84
      });                                                                                                              // 85
    }                                                                                                                  // 86
                                                                                                                       //
    return save;                                                                                                       //
  }(); // TODO(someday): Implement SystemPersistent.addRequirements().                                                 //
                                                                                                                       //
                                                                                                                       //
  return PersistentImpl;                                                                                               //
}();                                                                                                                   //
                                                                                                                       //
function hashSturdyRef(sturdyRef) {                                                                                    // 91
  return Crypto.createHash("sha256").update(sturdyRef).digest("base64");                                               // 92
}                                                                                                                      // 93
                                                                                                                       //
;                                                                                                                      // 93
                                                                                                                       //
function generateSturdyRef() {                                                                                         // 95
  return Random.secret();                                                                                              // 96
}                                                                                                                      // 97
                                                                                                                       //
function cryptApiToken(key, entry, cryptIn, cryptOut) {                                                                // 99
  // Encrypts or decrypts all fields of an ApiToken.                                                                   // 100
  // `cryptIn` translates a token in the input to a buffer.                                                            // 101
  // `cryptOut` translates a buffer to a token in the output.                                                          // 102
  check(key, String);                                                                                                  // 104
  check(entry, Object);                                                                                                // 105
  var nonce0 = new Buffer(8);                                                                                          // 107
  nonce0.fill(0);                                                                                                      // 108
  var keyBuf = new Buffer(32);                                                                                         // 110
  keyBuf.fill(0);                                                                                                      // 111
  keyBuf.write(key, 0, 32, "base64");                                                                                  // 112
                                                                                                                       //
  function encrypt0(token) {                                                                                           // 114
    return cryptOut(Capnp.chacha20(cryptIn(token), nonce0, keyBuf));                                                   // 115
  }                                                                                                                    // 116
                                                                                                                       //
  if (entry.parentTokenKey) {                                                                                          // 118
    entry.parentTokenKey = encrypt0(entry.parentTokenKey);                                                             // 119
  } else if (entry.frontendRef && entry.frontendRef.http) {                                                            // 120
    var http = entry.frontendRef.http;                                                                                 // 121
                                                                                                                       //
    if (http.auth) {                                                                                                   // 122
      var auth = http.auth;                                                                                            // 123
                                                                                                                       //
      if (auth.bearer) {                                                                                               // 124
        auth.bearer = encrypt0(auth.bearer);                                                                           // 125
      } else if (auth.basic) {                                                                                         // 126
        auth.basic.password = encrypt0(auth.basic.password);                                                           // 127
      } else if (auth.refresh) {                                                                                       // 128
        auth.refresh = encrypt0(auth.refresh);                                                                         // 129
      }                                                                                                                // 130
    }                                                                                                                  // 131
  }                                                                                                                    // 132
}                                                                                                                      // 133
                                                                                                                       //
function fetchApiToken(db, key, moreQuery) {                                                                           // 135
  // Reads an ApiToken from the database and decrypts its encrypted fields.                                            // 136
  function bufferToString(buf) {                                                                                       // 138
    // un-pad short secrets                                                                                            // 139
    var size = buf.length;                                                                                             // 140
                                                                                                                       //
    while (size > 0 && buf[size - 1] == 0) {                                                                           // 141
      --size;                                                                                                          // 142
    }                                                                                                                  // 143
                                                                                                                       //
    return buf.slice(0, size).toString("utf8");                                                                        // 145
  }                                                                                                                    // 146
                                                                                                                       //
  var query = {                                                                                                        // 148
    _id: hashSturdyRef(key)                                                                                            // 148
  };                                                                                                                   // 148
  Object.assign(query, moreQuery || {});                                                                               // 149
  var entry = db.collections.apiTokens.findOne(query);                                                                 // 150
                                                                                                                       //
  if (entry) {                                                                                                         // 151
    cryptApiToken(key, entry, function (x) {                                                                           // 152
      return x;                                                                                                        // 152
    }, bufferToString);                                                                                                // 152
  }                                                                                                                    // 153
                                                                                                                       //
  return entry;                                                                                                        // 155
}                                                                                                                      // 156
                                                                                                                       //
function insertApiToken(db, entry, key) {                                                                              // 158
  // Adds a new ApiToken to the database. `key`, if specified, *must* be a base64-encoded 256-bit                      // 159
  // value. If omitted, a key will be generated. Either way, the key is returned, and entry._id                        // 160
  // is filled in. Also, as a side effect, some fields of `entry` will become encrypted, but                           // 161
  // ideally callers should not depend on this behavior.                                                               // 162
  function stringToBuffer(str) {                                                                                       // 164
    var buf = new Buffer(str, "utf8");                                                                                 // 165
    if (buf.length >= 32) return buf;                                                                                  // 166
    var padded = new Buffer(32);                                                                                       // 168
    padded.fill(0);                                                                                                    // 169
    buf.copy(padded);                                                                                                  // 170
    return padded;                                                                                                     // 171
  }                                                                                                                    // 172
                                                                                                                       //
  if (!key) key = generateSturdyRef();                                                                                 // 174
  entry._id = hashSturdyRef(key);                                                                                      // 175
  cryptApiToken(key, entry, stringToBuffer, function (x) {                                                             // 176
    return x;                                                                                                          // 176
  });                                                                                                                  // 176
  db.collections.apiTokens.insert(entry);                                                                              // 177
  return key;                                                                                                          // 178
}                                                                                                                      // 179
                                                                                                                       //
function checkRequirements(db, requirements) {                                                                         // 181
  // Checks if the given list of MembraneRequirements are all satisfied, returning true if so and                      // 182
  // false otherwise.                                                                                                  // 183
  // TODO(cleanup): SandstormPermissions has a different way of checking the same requirements.                        // 185
  //   Reuse?                                                                                                          // 186
  // TODO(security): Eventually we want checkRequirements(), when it passes, to produce an observer                    // 188
  //   that can be used to receive a notification when the requirements may no longer be satisfied,                    // 189
  //   in order to revoke the live object. See RequirementObserver in supervisor.capnp.                                // 190
  if (!requirements) {                                                                                                 // 192
    return true;                                                                                                       // 193
  }                                                                                                                    // 194
                                                                                                                       //
  requirements.forEach(function (requirement) {                                                                        // 196
    if (requirement.tokenValid) {                                                                                      // 197
      var token = db.collections.apiTokens.findOne({                                                                   // 198
        _id: requirement.tokenValid,                                                                                   // 199
        revoked: {                                                                                                     // 199
          $ne: true                                                                                                    // 199
        }                                                                                                              // 199
      }, {                                                                                                             // 199
        fields: {                                                                                                      // 200
          requirements: 1                                                                                              // 200
        }                                                                                                              // 200
      });                                                                                                              // 200
                                                                                                                       //
      if (!token) {                                                                                                    // 201
        throw new Meteor.Error(403, "Capability revoked because the link through which it was introduced has been " + "revoked or deleted.");
      }                                                                                                                // 205
                                                                                                                       //
      checkRequirements(db, token.requirements);                                                                       // 207
                                                                                                                       //
      if (token.parentToken) {                                                                                         // 209
        checkRequirements(db, [{                                                                                       // 210
          tokenValid: token.parentToken                                                                                // 210
        }]);                                                                                                           // 210
      }                                                                                                                // 211
    } else if (requirement.permissionsHeld) {                                                                          // 212
      var p = requirement.permissionsHeld;                                                                             // 213
      var viewInfo = db.collections.grains.findOne(p.grainId, {                                                        // 214
        fields: {                                                                                                      // 215
          cachedViewInfo: 1                                                                                            // 215
        }                                                                                                              // 215
      }).cachedViewInfo;                                                                                               // 215
      var vertex = void 0;                                                                                             // 216
                                                                                                                       //
      if (p.accountId) {                                                                                               // 217
        vertex = {                                                                                                     // 218
          grain: {                                                                                                     // 218
            _id: p.grainId,                                                                                            // 218
            accountId: p.accountId                                                                                     // 218
          }                                                                                                            // 218
        };                                                                                                             // 218
      } else {                                                                                                         // 219
        vertex = {                                                                                                     // 220
          token: {                                                                                                     // 220
            _id: p.tokenId,                                                                                            // 220
            grainId: p.grainId                                                                                         // 220
          }                                                                                                            // 220
        };                                                                                                             // 220
      }                                                                                                                // 221
                                                                                                                       //
      var currentPermissions = SandstormPermissions.grainPermissions(db, vertex, viewInfo || {}).permissions;          // 223
                                                                                                                       //
      if (!currentPermissions) {                                                                                       // 225
        throw new Meteor.Error(403, "Capability revoked because a user involved in introducing it no longer has " + "the necessary permissions.");
      }                                                                                                                // 229
                                                                                                                       //
      var requiredPermissions = p.permissions || [];                                                                   // 231
                                                                                                                       //
      for (var ii = 0; ii < requiredPermissions.length; ++ii) {                                                        // 232
        if (requiredPermissions[ii] && !currentPermissions[ii]) {                                                      // 233
          throw new Meteor.Error(403, "Capability revoked because a user involved in introducing it no longer has " + "the necessary permissions.");
        }                                                                                                              // 237
      }                                                                                                                // 238
    } else if (requirement.userIsAdmin) {                                                                              // 239
      if (!db.isAdminById(requirement.userIsAdmin)) {                                                                  // 240
        throw new Meteor.Error(403, "Capability revoked because the user who created it has lost their admin " + "rights.");
      }                                                                                                                // 244
    } else {                                                                                                           // 245
      throw new Meteor.Error(403, "Unknown requirement type.");                                                        // 246
    }                                                                                                                  // 247
  });                                                                                                                  // 248
}                                                                                                                      // 249
                                                                                                                       //
;                                                                                                                      // 249
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"static-asset.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/static-asset.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  StaticAssetImpl: function () {                                                                                       // 1
    return StaticAssetImpl;                                                                                            // 1
  },                                                                                                                   // 1
  IdenticonStaticAssetImpl: function () {                                                                              // 1
    return IdenticonStaticAssetImpl;                                                                                   // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014-2016 Sandstorm Development Group, Inc. and contributors                                          // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Capnp = Npm.require("capnp");                                                                                      // 17
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 18
                                                                                                                       //
var StaticAsset = Capnp.importSystem("sandstorm/grain.capnp").StaticAsset;                                             // 19
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;                                                               // 21
                                                                                                                       //
var StaticAssetImpl = function () {                                                                                    //
  function StaticAssetImpl(assetId) {                                                                                  // 24
    (0, _classCallCheck3.default)(this, StaticAssetImpl);                                                              // 24
    check(assetId, String);                                                                                            // 25
    this._protocol = PROTOCOL.slice(0, -1);                                                                            // 26
    this._hostPath = makeWildcardHost("static") + "/" + assetId;                                                       // 27
  }                                                                                                                    // 28
                                                                                                                       //
  StaticAssetImpl.prototype.getUrl = function () {                                                                     //
    function getUrl() {                                                                                                //
      return {                                                                                                         // 31
        protocol: this._protocol,                                                                                      // 31
        hostPath: this._hostPath                                                                                       // 31
      };                                                                                                               // 31
    }                                                                                                                  // 32
                                                                                                                       //
    return getUrl;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  return StaticAssetImpl;                                                                                              //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 33
                                                                                                                       //
var IdenticonStaticAssetImpl = function () {                                                                           //
  function IdenticonStaticAssetImpl(hash, size) {                                                                      // 36
    (0, _classCallCheck3.default)(this, IdenticonStaticAssetImpl);                                                     // 36
    check(hash, String);                                                                                               // 37
    check(size, Match.Integer);                                                                                        // 38
    this._protocol = PROTOCOL.slice(0, -1);                                                                            // 39
    this._hostPath = makeWildcardHost("static") + "/identicon/" + hash + "?s=" + size;                                 // 40
  }                                                                                                                    // 41
                                                                                                                       //
  IdenticonStaticAssetImpl.prototype.getUrl = function () {                                                            //
    function getUrl() {                                                                                                //
      return {                                                                                                         // 44
        protocol: this._protocol,                                                                                      // 44
        hostPath: this._hostPath                                                                                       // 44
      };                                                                                                               // 44
    }                                                                                                                  // 45
                                                                                                                       //
    return getUrl;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  return IdenticonStaticAssetImpl;                                                                                     //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 46
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"constants.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/constants.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  ACCOUNT_DELETION_SUSPENSION_TIME: function () {                                                                      // 1
    return ACCOUNT_DELETION_SUSPENSION_TIME;                                                                           // 1
  },                                                                                                                   // 1
  PRIVATE_IPV4_ADDRESSES: function () {                                                                                // 1
    return PRIVATE_IPV4_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  PRIVATE_IPV6_ADDRESSES: function () {                                                                                // 1
    return PRIVATE_IPV6_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  SPECIAL_IPV4_ADDRESSES: function () {                                                                                // 1
    return SPECIAL_IPV4_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  SPECIAL_IPV6_ADDRESSES: function () {                                                                                // 1
    return SPECIAL_IPV6_ADDRESSES;                                                                                     // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var ACCOUNT_DELETION_SUSPENSION_TIME = 7 * 60 * 60 * 24 * 1000; // 7 days in ms                                        // 17
// Lists below developed from RFC6890, which is an overview of all special addresses.                                  // 19
                                                                                                                       //
var PRIVATE_IPV4_ADDRESSES = ["10.0.0.0/8", // RFC1918 reserved for internal network                                   // 20
"127.0.0.0/8", // RFC1122 loopback / localhost                                                                         // 22
"169.254.0.0/16", // RFC3927 "link local" (auto-configured LAN in absence of DHCP)                                     // 23
"172.16.0.0/12", // RFC1918 reserved for internal network                                                              // 24
"192.168.0.0/16"];                                                                                                     // 25
var PRIVATE_IPV6_ADDRESSES = ["::1/128", // RFC4291 loopback / localhost                                               // 28
"fc00::/7", // RFC4193 unique private network                                                                          // 30
"fe80::/10"];                                                                                                          // 31
var SPECIAL_IPV4_ADDRESSES = ["0.0.0.0/8", // RFC1122 "this host" / wildcard                                           // 34
"100.64.0.0/10", // RFC6598 "shared address space" for carrier-grade NAT                                               // 36
"192.0.0.0/24", // RFC6890 reserved for special protocols                                                              // 37
"192.0.2.0/24", // RFC5737 "example address" block 1 -- like example.com for IPs                                       // 38
"192.88.99.0/24", // RFC3068 6to4 relay                                                                                // 39
"198.18.0.0/15", // RFC2544 standard benchmarks                                                                        // 40
"198.51.100.0/24", // RFC5737 "example address" block 2 -- like example.com for IPs                                    // 41
"203.0.113.0/24", // RFC5737 "example address" block 3 -- like example.com for IPs                                     // 42
"224.0.0.0/4", // RFC1112 multicast                                                                                    // 43
"240.0.0.0/4", // RFC1112 multicast / reserved for future use                                                          // 44
"255.255.255.255/32"];                                                                                                 // 45
var SPECIAL_IPV6_ADDRESSES = ["::/128", // RFC4291 unspecified address / wildcard                                      // 48
"64:ff9b::/96", // RFC6052 IPv4-IPv6 translation                                                                       // 50
"::ffff:0:0/96", // RFC4291 IPv4-mapped address                                                                        // 51
// TODO(someday): I don't understand the difference between the above                                                  // 52
//     two. Both are described as mapping ip4 addresses into the ip6                                                   // 53
//     space. Perhaps this should be allowed, however, we'd need to                                                    // 54
//     filter the ip4 address against the ip4 blacklist, so special                                                    // 55
//     handling would be needed.                                                                                       // 56
"100::/64", // RFC6666 discard-only address block                                                                      // 57
"2001::/23", // RFC2928 reserved for special protocols                                                                 // 58
"2001:2::/48", // RFC5180 standard benchmarks                                                                          // 59
"2001:db8::/32", // RFC3849 "example address" block -- like example.com for IPs                                        // 60
"2001:10::/28", // RFC4843 ORCHID                                                                                      // 61
"2002::/16", // RFC3056 6to4 relay                                                                                     // 62
"ff00::/8"];                                                                                                           // 63
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"db-deprecated.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/db-deprecated.js                                                                                            //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file imports symbols from the sandstorm-db package. New code should use SandstormDb                            // 17
// directly.                                                                                                           // 18
// TODO(cleanup): Over time, eliminate the use of each of these assignments by using `globalDb`                        // 20
//   directly. For collections, prefer to update SandstormDb to provide methods for querying the                       // 21
//   collection rather than use `globalDb.collections` directly. When code is moved into packages,                     // 22
//   the `globalDb` global should NOT go with it; the package should expect a SandstormDb to be                        // 23
//   passed in, thus allowing mocking the database for unit tests.                                                     // 24
var quotaManager = void 0;                                                                                             // 26
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 27
  var LDAP = void 0;                                                                                                   // 1
  module.watch(require("/imports/server/accounts/ldap.js"), {                                                          // 1
    LDAP: function (v) {                                                                                               // 1
      LDAP = v;                                                                                                        // 1
    }                                                                                                                  // 1
  }, 0);                                                                                                               // 1
  // Imports are usually not allowed to occur in a block. However, it is the only way to do                            // 29
  // this under Meteor. Using // jscs:disable doesn't work for what it considers syntax violations,                    // 30
  // and so we've added this file to .jscsrc's excludedFiles explicitly.                                               // 31
  quotaManager = new LDAP();                                                                                           // 33
} else {                                                                                                               // 34
  quotaManager = {                                                                                                     // 35
    updateUserQuota: function (db, user) {                                                                             // 36
      return {                                                                                                         // 37
        storage: user.cachedStorageQuota || 0,                                                                         // 38
        grains: Infinity,                                                                                              // 39
        compute: Infinity                                                                                              // 40
      };                                                                                                               // 37
    }                                                                                                                  // 42
  };                                                                                                                   // 35
}                                                                                                                      // 44
                                                                                                                       //
globalDb = new SandstormDb(quotaManager);                                                                              // 46
Packages = globalDb.collections.packages;                                                                              // 48
DevPackages = globalDb.collections.devPackages;                                                                        // 49
UserActions = globalDb.collections.userActions;                                                                        // 50
Grains = globalDb.collections.grains;                                                                                  // 51
Contacts = globalDb.collections.contacts;                                                                              // 52
Sessions = globalDb.collections.sessions;                                                                              // 53
SignupKeys = globalDb.collections.signupKeys;                                                                          // 54
ActivityStats = globalDb.collections.activityStats;                                                                    // 55
DeleteStats = globalDb.collections.deleteStats;                                                                        // 56
FileTokens = globalDb.collections.fileTokens;                                                                          // 57
ApiTokens = globalDb.collections.apiTokens;                                                                            // 58
Notifications = globalDb.collections.notifications;                                                                    // 59
StatsTokens = globalDb.collections.statsTokens;                                                                        // 60
Misc = globalDb.collections.misc;                                                                                      // 61
Settings = globalDb.collections.settings;                                                                              // 62
currentUserGrains = globalDb.currentUserGrains.bind(globalDb);                                                         // 64
isDemoUser = globalDb.isDemoUser.bind(globalDb);                                                                       // 65
isSignedUp = globalDb.isSignedUp.bind(globalDb);                                                                       // 66
isSignedUpOrDemo = globalDb.isSignedUpOrDemo.bind(globalDb);                                                           // 67
isUserOverQuota = globalDb.isUserOverQuota.bind(globalDb);                                                             // 68
isUserExcessivelyOverQuota = globalDb.isUserExcessivelyOverQuota.bind(globalDb);                                       // 69
isAdmin = globalDb.isAdmin.bind(globalDb);                                                                             // 70
isAdminById = globalDb.isAdminById.bind(globalDb);                                                                     // 71
findAdminUserForToken = globalDb.findAdminUserForToken.bind(globalDb);                                                 // 72
matchWildcardHost = globalDb.matchWildcardHost.bind(globalDb);                                                         // 73
makeWildcardHost = globalDb.makeWildcardHost.bind(globalDb);                                                           // 74
allowDevAccounts = globalDb.allowDevAccounts.bind(globalDb);                                                           // 75
roleAssignmentPattern = globalDb.roleAssignmentPattern;                                                                // 76
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"demo.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/demo.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
                  allowDemo: function () {                                                                             // 1
                                    return allowDemo;                                                                  // 1
                  }                                                                                                    // 1
});                                                                                                                    // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var allowDemo = Meteor.settings && Meteor.settings.public && Meteor.settings.public.allowDemoAccounts;                 // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"server":{"accounts":{"credentials":{"credentials-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/credentials/credentials-server.js                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var SandstormBackend = void 0;                                                                                         // 1
module.watch(require("/imports/server/backend.js"), {                                                                  // 1
  SandstormBackend: function (v) {                                                                                     // 1
    SandstormBackend = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var linkCredentialToAccountInternal = function (db, backend, credentialId, accountId, allowLogin) {                    // 19
  // Links the credential to the account. If `allowLogin` is true, grants the credential login access                  // 20
  // if possible. Makes the account durable if it is a demo account.                                                   // 21
  check(db, SandstormDb);                                                                                              // 23
  check(backend, SandstormBackend);                                                                                    // 24
  check(credentialId, String);                                                                                         // 25
  check(accountId, String);                                                                                            // 26
  var accountUser = Meteor.users.findOne({                                                                             // 28
    _id: accountId                                                                                                     // 28
  });                                                                                                                  // 28
                                                                                                                       //
  if (!accountUser) {                                                                                                  // 29
    throw new Meteor.Error(404, "No account found with ID " + accountId);                                              // 30
  }                                                                                                                    // 31
                                                                                                                       //
  if (accountUser.type !== "account") {                                                                                // 33
    throw new Meteor.Error(400, "Cannot link a credential to another credential.");                                    // 34
  }                                                                                                                    // 35
                                                                                                                       //
  if (!!_.findWhere(accountUser.loginCredentials, {                                                                    // 37
    id: credentialId                                                                                                   // 37
  }) || !!_.findWhere(accountUser.nonloginCredentials, {                                                               // 37
    id: credentialId                                                                                                   // 38
  })) {                                                                                                                // 38
    throw new Meteor.Error("alreadyLinked", "Cannot link a credential that's alread linked to this account.");         // 39
  }                                                                                                                    // 41
                                                                                                                       //
  var credentialUser = Meteor.users.findOne({                                                                          // 43
    _id: credentialId                                                                                                  // 43
  });                                                                                                                  // 43
                                                                                                                       //
  if (!credentialUser) {                                                                                               // 45
    throw new Meteor.Error(404, "No credential found with ID " + credentialId);                                        // 46
  }                                                                                                                    // 47
                                                                                                                       //
  if (!credentialUser.profile) {                                                                                       // 49
    throw new Meteor.Error(400, "Cannot link an account to another account");                                          // 50
  }                                                                                                                    // 51
                                                                                                                       //
  db.deleteUnusedAccount(backend, credentialUser._id);                                                                 // 53
                                                                                                                       //
  if (Meteor.users.findOne({                                                                                           // 54
    "loginCredentials.id": credentialUser._id                                                                          // 54
  })) {                                                                                                                // 54
    throw new Meteor.Error(403, "Cannot link a credential that can already log into another account");                 // 55
  }                                                                                                                    // 57
                                                                                                                       //
  var alreadyLinked = !!Meteor.users.findOne({                                                                         // 59
    "nonloginCredentials.id": credentialUser._id                                                                       // 59
  });                                                                                                                  // 59
  var pushModifier = alreadyLinked || !allowLogin ? {                                                                  // 61
    nonloginCredentials: {                                                                                             // 62
      id: credentialUser._id                                                                                           // 62
    }                                                                                                                  // 62
  } : {                                                                                                                // 62
    loginCredentials: {                                                                                                // 63
      id: credentialUser._id                                                                                           // 63
    }                                                                                                                  // 63
  };                                                                                                                   // 63
  var modifier = void 0;                                                                                               // 65
                                                                                                                       //
  if (accountUser.expires) {                                                                                           // 66
    if (alreadyLinked) {                                                                                               // 67
      throw new Meteor.Error(403, "Cannot create an account for a credential that's " + "already linked to another account.");
    }                                                                                                                  // 70
                                                                                                                       //
    modifier = {                                                                                                       // 72
      $push: pushModifier,                                                                                             // 73
      $unset: {                                                                                                        // 74
        expires: 1                                                                                                     // 74
      },                                                                                                               // 74
      $set: {                                                                                                          // 75
        upgradedFromDemo: Date.now()                                                                                   // 75
      }                                                                                                                // 75
    };                                                                                                                 // 72
                                                                                                                       //
    if (db.isReferralEnabled()) {                                                                                      // 77
      // Demo users never got the referral notification. Send it now:                                                  // 78
      db.sendReferralProgramNotification(accountUser._id);                                                             // 79
    }                                                                                                                  // 80
  } else {                                                                                                             // 82
    modifier = {                                                                                                       // 83
      $push: pushModifier                                                                                              // 83
    };                                                                                                                 // 83
  } // Make sure not to add the same credential twice.                                                                 // 84
                                                                                                                       //
                                                                                                                       //
  Meteor.users.update({                                                                                                // 87
    _id: accountUser._id,                                                                                              // 87
    "nonloginCredentials.id": {                                                                                        // 88
      $ne: credentialUser._id                                                                                          // 88
    },                                                                                                                 // 88
    "loginCredentials.id": {                                                                                           // 89
      $ne: credentialUser._id                                                                                          // 89
    }                                                                                                                  // 89
  }, modifier);                                                                                                        // 87
                                                                                                                       //
  if (accountUser.expires) {                                                                                           // 92
    var demoCredentialId = SandstormDb.getUserCredentialIds(accountUser)[0];                                           // 93
    Meteor.users.update({                                                                                              // 94
      _id: demoCredentialId                                                                                            // 94
    }, {                                                                                                               // 94
      $unset: {                                                                                                        // 95
        expires: 1                                                                                                     // 95
      },                                                                                                               // 95
      $set: {                                                                                                          // 96
        upgradedFromDemo: Date.now()                                                                                   // 96
      }                                                                                                                // 96
    }); // The account's existing profile is just "Demo User". Import the new credential's profile.                    // 95
                                                                                                                       //
    SandstormDb.fillInProfileDefaults(credentialUser, credentialUser.profile); // Mark the demo credential as nonlogin. It'd be nicer if the credential started out as nonlogin,
    // but to get that to work we would need to adjust the account creation and first login logic.                     // 102
                                                                                                                       //
    Meteor.users.update({                                                                                              // 103
      _id: accountUser._id,                                                                                            // 103
      "loginCredentials.id": demoCredentialId,                                                                         // 104
      "nonloginCredentials.id": {                                                                                      // 105
        $not: {                                                                                                        // 105
          $eq: demoCredentialId                                                                                        // 105
        }                                                                                                              // 105
      }                                                                                                                // 105
    }, {                                                                                                               // 103
      $pull: {                                                                                                         // 106
        loginCredentials: {                                                                                            // 106
          id: demoCredentialId                                                                                         // 106
        }                                                                                                              // 106
      },                                                                                                               // 106
      $push: {                                                                                                         // 107
        nonloginCredentials: {                                                                                         // 107
          id: demoCredentialId                                                                                         // 107
        }                                                                                                              // 107
      },                                                                                                               // 107
      $set: {                                                                                                          // 108
        profile: credentialUser.profile                                                                                // 108
      }                                                                                                                // 108
    });                                                                                                                // 106
  }                                                                                                                    // 109
};                                                                                                                     // 110
                                                                                                                       //
Meteor.methods({                                                                                                       // 112
  loginWithCredential: function (accountUserId) {                                                                      // 113
    // Logs into the account with ID `accountUserId`. Throws an exception if the current user is                       // 114
    // not a credential user listed in the account's `loginCredentials` field. This method is not                      // 115
    // intended to be called directly; client-side code should only invoke it through                                  // 116
    // `Meteor.loginWithCredential()`, which additionally maintains the standard Meteor client-side                    // 117
    // login state.                                                                                                    // 118
    check(accountUserId, String);                                                                                      // 120
    var credentialUser = Meteor.user();                                                                                // 122
                                                                                                                       //
    if (!credentialUser || !credentialUser.profile) {                                                                  // 123
      throw new Meteor.Error(403, "Must be already logged in as an credential.");                                      // 124
    }                                                                                                                  // 125
                                                                                                                       //
    var accountUser = Meteor.users.findOne(accountUserId);                                                             // 127
                                                                                                                       //
    if (!accountUser) {                                                                                                // 128
      throw new Meteor.Error(404, "No such user found: " + accountUserId);                                             // 129
    }                                                                                                                  // 130
                                                                                                                       //
    var linkedCredential = _.findWhere(accountUser.loginCredentials, {                                                 // 132
      id: credentialUser._id                                                                                           // 132
    });                                                                                                                // 132
                                                                                                                       //
    if (!linkedCredential) {                                                                                           // 134
      throw new Meteor.Error(403, "Current credential is not a login credential for account " + accountUserId);        // 135
    }                                                                                                                  // 137
                                                                                                                       //
    return Accounts._loginMethod(this, "loginWithCredential", [accountUserId], "credential", function () {             // 139
      return {                                                                                                         // 140
        userId: accountUserId                                                                                          // 140
      };                                                                                                               // 140
    });                                                                                                                // 140
  },                                                                                                                   // 141
  createAccountForCredential: function () {                                                                            // 143
    // Creates a new account for the currently-logged-in credential.                                                   // 144
    var user = Meteor.user();                                                                                          // 146
                                                                                                                       //
    if (user.type !== "credential") {                                                                                  // 147
      throw new Meteor.Error(403, "Must be logged in as a credential in order to create an account.");                 // 148
    }                                                                                                                  // 149
                                                                                                                       //
    if (Meteor.users.findOne({                                                                                         // 151
      $or: [{                                                                                                          // 152
        "loginCredentials.id": user._id                                                                                // 153
      }, {                                                                                                             // 153
        "nonloginCredentials.id": user._id                                                                             // 154
      }]                                                                                                               // 154
    })) {                                                                                                              // 151
      throw new Meteor.Error(403, "Cannot create an account for a credential that's already " + "linked to another account.");
    }                                                                                                                  // 159
                                                                                                                       //
    var newUser = {                                                                                                    // 161
      type: "account",                                                                                                 // 162
      loginCredentials: [{                                                                                             // 163
        id: user._id                                                                                                   // 163
      }],                                                                                                              // 163
      nonloginCredentials: []                                                                                          // 164
    };                                                                                                                 // 161
                                                                                                                       //
    if (user.services.dev) {                                                                                           // 166
      newUser.signupKey = "devAccounts";                                                                               // 167
                                                                                                                       //
      if (user.services.dev.isAdmin) {                                                                                 // 168
        newUser.isAdmin = true;                                                                                        // 169
      }                                                                                                                // 170
                                                                                                                       //
      if (user.services.dev.hasCompletedSignup) {                                                                      // 172
        newUser.hasCompletedSignup = true;                                                                             // 173
      }                                                                                                                // 174
    } else if (user.expires) {                                                                                         // 175
      // Demo user.                                                                                                    // 176
      newUser.expires = user.expires;                                                                                  // 177
                                                                                                                       //
      if (!!user.appDemoId) {                                                                                          // 178
        newUser.appDemoId = user.appDemoId;                                                                            // 179
      }                                                                                                                // 180
    }                                                                                                                  // 181
                                                                                                                       //
    SandstormDb.fillInProfileDefaults(user, user.profile);                                                             // 183
    var options = {                                                                                                    // 184
      profile: user.profile                                                                                            // 184
    }; // This will throw an error if the credential has been added as a login credential to some                      // 184
    // other account while we were executing the body of this method.                                                  // 187
                                                                                                                       //
    return Accounts.insertUserDoc(options, newUser);                                                                   // 188
  },                                                                                                                   // 189
  linkCredentialToAccount: function (token) {                                                                          // 191
    // Links the credential of the current user to the account that has `token` as a resume token.                     // 192
    // If the account is a demo account, makes the account durable and gives the credential login                      // 193
    // access to it.                                                                                                   // 194
    check(token, String);                                                                                              // 196
                                                                                                                       //
    if (!this.userId) {                                                                                                // 198
      throw new Meteor.Error(403, "Cannot link to account if not logged in.");                                         // 199
    }                                                                                                                  // 200
                                                                                                                       //
    var hashed = Accounts._hashLoginToken(token);                                                                      // 202
                                                                                                                       //
    var accountUser = Meteor.users.findOne({                                                                           // 203
      "services.resume.loginTokens.hashedToken": hashed                                                                // 203
    });                                                                                                                // 203
    linkCredentialToAccountInternal(this.connection.sandstormDb, this.connection.sandstormBackend, this.userId, accountUser._id, true);
  },                                                                                                                   // 207
  unlinkCredential: function (accountUserId, credentialId) {                                                           // 209
    // Unlinks the credential with ID `credentialId` from the account with ID `accountUserId`.                         // 210
    check(credentialId, String);                                                                                       // 212
    check(accountUserId, String);                                                                                      // 213
                                                                                                                       //
    if (!this.userId) {                                                                                                // 215
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 216
    }                                                                                                                  // 217
                                                                                                                       //
    if (!this.connection.sandstormDb.userHasCredential(this.userId, credentialId)) {                                   // 219
      throw new Meteor.Error(403, "Current user does not own credential " + credentialId);                             // 220
    }                                                                                                                  // 221
                                                                                                                       //
    var credentialUser = Meteor.users.findOne({                                                                        // 223
      _id: credentialId                                                                                                // 223
    });                                                                                                                // 223
    Meteor.users.update({                                                                                              // 224
      _id: accountUserId                                                                                               // 225
    }, {                                                                                                               // 224
      $pull: {                                                                                                         // 227
        nonloginCredentials: {                                                                                         // 228
          id: credentialId                                                                                             // 228
        },                                                                                                             // 228
        loginCredentials: {                                                                                            // 229
          id: credentialId                                                                                             // 229
        }                                                                                                              // 229
      }                                                                                                                // 227
    });                                                                                                                // 226
  },                                                                                                                   // 232
  setCredentialAllowsLogin: function (credentialId, allowLogin) {                                                      // 234
    // Sets whether the current account allows the credential with ID `credentialId` to log in.                        // 235
    check(credentialId, String);                                                                                       // 237
    check(allowLogin, Boolean);                                                                                        // 238
                                                                                                                       //
    if (!this.userId) {                                                                                                // 239
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 240
    }                                                                                                                  // 241
                                                                                                                       //
    if (!this.connection.sandstormDb.userHasCredential(this.userId, credentialId)) {                                   // 243
      throw new Meteor.Error(403, "Current user does not own credential " + credentialId);                             // 244
    }                                                                                                                  // 245
                                                                                                                       //
    if (allowLogin) {                                                                                                  // 247
      Meteor.users.update({                                                                                            // 248
        _id: this.userId,                                                                                              // 248
        "nonloginCredentials.id": credentialId,                                                                        // 249
        "loginCredentials.id": {                                                                                       // 250
          $not: {                                                                                                      // 250
            $eq: credentialId                                                                                          // 250
          }                                                                                                            // 250
        }                                                                                                              // 250
      }, {                                                                                                             // 248
        $pull: {                                                                                                       // 251
          nonloginCredentials: {                                                                                       // 251
            id: credentialId                                                                                           // 251
          }                                                                                                            // 251
        },                                                                                                             // 251
        $push: {                                                                                                       // 252
          loginCredentials: {                                                                                          // 252
            id: credentialId                                                                                           // 252
          }                                                                                                            // 252
        }                                                                                                              // 252
      });                                                                                                              // 251
    } else {                                                                                                           // 253
      Meteor.users.update({                                                                                            // 254
        _id: this.userId,                                                                                              // 254
        "loginCredentials.id": credentialId,                                                                           // 255
        "nonloginCredentials.id": {                                                                                    // 256
          $not: {                                                                                                      // 256
            $eq: credentialId                                                                                          // 256
          }                                                                                                            // 256
        }                                                                                                              // 256
      }, {                                                                                                             // 254
        $pull: {                                                                                                       // 257
          loginCredentials: {                                                                                          // 257
            id: credentialId                                                                                           // 257
          }                                                                                                            // 257
        },                                                                                                             // 257
        $push: {                                                                                                       // 258
          nonloginCredentials: {                                                                                       // 258
            id: credentialId                                                                                           // 258
          }                                                                                                            // 258
        }                                                                                                              // 258
      });                                                                                                              // 257
    }                                                                                                                  // 259
  },                                                                                                                   // 260
  logoutCredentialsOfCurrentAccount: function () {                                                                     // 262
    // Logs out all credentials that are allowed to log in to the current account.                                     // 263
    var user = Meteor.user();                                                                                          // 264
                                                                                                                       //
    if (user && user.loginCredentials) {                                                                               // 265
      user.loginCredentials.forEach(function (credential) {                                                            // 266
        Meteor.users.update({                                                                                          // 267
          _id: credential.id                                                                                           // 267
        }, {                                                                                                           // 267
          $set: {                                                                                                      // 267
            "services.resume.loginTokens": []                                                                          // 267
          }                                                                                                            // 267
        });                                                                                                            // 267
      });                                                                                                              // 268
    }                                                                                                                  // 269
  }                                                                                                                    // 270
});                                                                                                                    // 112
                                                                                                                       //
Accounts.linkCredentialToAccount = function (db, backend, credentialId, accountId, allowLogin) {                       // 273
  // Links the credential to the account. If the account is a demo account, makes it durable.                          // 274
  // If `allowLogin` is true, attempts to give the credential login access.                                            // 275
  check(db, SandstormDb);                                                                                              // 276
  check(backend, SandstormBackend);                                                                                    // 277
  check(credentialId, String);                                                                                         // 278
  check(accountId, String);                                                                                            // 279
  check(allowLogin, Boolean);                                                                                          // 280
  linkCredentialToAccountInternal(db, backend, credentialId, accountId, allowLogin);                                   // 281
};                                                                                                                     // 282
                                                                                                                       //
Meteor.publish("accountsOfCredential", function (credentialId) {                                                       // 284
  check(credentialId, String);                                                                                         // 285
  if (!SandstormDb.ensureSubscriberHasCredential(this, credentialId)) return; // Map from credential ID to `true` for each credential we've published already.
                                                                                                                       //
  var loginCredentials = {};                                                                                           // 289
                                                                                                                       //
  var _this = this;                                                                                                    // 291
                                                                                                                       //
  function addCredentialsOfAccount(account) {                                                                          // 292
    account.loginCredentials.forEach(function (credential) {                                                           // 293
      if (!(credential.id in loginCredentials)) {                                                                      // 294
        var user = Meteor.users.findOne({                                                                              // 295
          _id: credential.id                                                                                           // 295
        });                                                                                                            // 295
                                                                                                                       //
        if (user) {                                                                                                    // 296
          user.intrinsicName = SandstormDb.getIntrinsicName(user);                                                     // 297
          user.loginId = SandstormDb.getLoginId(user);                                                                 // 298
          user.serviceName = SandstormDb.getServiceName(user);                                                         // 299
                                                                                                                       //
          var filteredUser = _.pick(user, "_id", "intrinsicName", "loginId", "serviceName");                           // 301
                                                                                                                       //
          filteredUser.loginAccountId = account._id;                                                                   // 302
          filteredUser.sourceCredentialId = credentialId;                                                              // 303
                                                                                                                       //
          _this.added("loginCredentialsOfLinkedAccounts", user._id, filteredUser);                                     // 304
        }                                                                                                              // 305
                                                                                                                       //
        loginCredentials[credential.id] = true;                                                                        // 307
      }                                                                                                                // 308
    });                                                                                                                // 309
  }                                                                                                                    // 310
                                                                                                                       //
  var cursor = Meteor.users.find({                                                                                     // 312
    $or: [{                                                                                                            // 313
      "loginCredentials.id": credentialId                                                                              // 314
    }, {                                                                                                               // 314
      "nonloginCredentials.id": credentialId                                                                           // 315
    }]                                                                                                                 // 315
  });                                                                                                                  // 312
  var handle = cursor.observe({                                                                                        // 319
    added: function (account) {                                                                                        // 320
      addCredentialsOfAccount(account);                                                                                // 321
    },                                                                                                                 // 322
    changed: function (newAccount, oldAccount) {                                                                       // 324
      addCredentialsOfAccount(newAccount);                                                                             // 325
    },                                                                                                                 // 326
    removed: function (account) {                                                                                      // 328
      account.loginCredentials.forEach(function (credential) {                                                         // 329
        if (credential.id in loginCredentials) {                                                                       // 330
          _this.removed("loginCredentialsOfLinkedAccounts", credential.id);                                            // 331
                                                                                                                       //
          delete loginCredentials[credential.id];                                                                      // 332
        }                                                                                                              // 333
      });                                                                                                              // 334
    }                                                                                                                  // 335
  });                                                                                                                  // 319
  this.ready();                                                                                                        // 337
  this.onStop(function () {                                                                                            // 339
    handle.stop();                                                                                                     // 340
    Object.keys(loginCredentials).forEach(function (credentialId) {                                                    // 341
      delete loginCredentials[credentialId];                                                                           // 342
    });                                                                                                                // 343
  });                                                                                                                  // 344
});                                                                                                                    // 345
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"email-token":{"token-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/email-token/token-server.js                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var crypto = void 0;                                                                                                   // 1
module.watch(require("crypto"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var sendEmail = void 0;                                                                                                // 1
module.watch(require("/imports/server/email.js"), {                                                                    // 1
  send: function (v) {                                                                                                 // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 5
                                                                                                                       //
var V1_ROUNDS = 4096; // Selected to take ~5msec at creation time (2016) on a developer's laptop.                      // 7
                                                                                                                       //
var V1_KEYSIZE = 32; // 256 bits / 8 bits/byte = 32 bytes                                                              // 8
                                                                                                                       //
var V1_HASHFUNC = "sha512"; // ^ hash function used with pbkdf2.  Chosen to be different from the function which maps the token
// to the value stored in the database.  Note that the first thing that pbkdf2 does is                                 // 11
// HMAC(HASHFUNC, key, salt), and the first thing that HMAC does is either pad or hash the key to                      // 12
// make it the appropriate width.  The result is that knowing sha256(key) and the salt is possibly                     // 13
// sufficient to reconstruct the output of pbkdf2().                                                                   // 14
                                                                                                                       //
var V1_CIPHER = "AES-256-CTR"; // cipher used                                                                          // 15
                                                                                                                       //
var TOKEN_EXPIRATION_MS = 60 * 60 * 1000;                                                                              // 17
                                                                                                                       //
var cleanupExpiredTokens = function () {                                                                               // 19
  Meteor.users.update({                                                                                                // 20
    "services.email.tokens.createdAt": {                                                                               // 21
      $lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)                                                                  // 22
    }                                                                                                                  // 21
  }, {                                                                                                                 // 20
    $pull: {                                                                                                           // 25
      "services.email.tokens": {                                                                                       // 26
        createdAt: {                                                                                                   // 27
          $lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)                                                              // 27
        }                                                                                                              // 27
      }                                                                                                                // 26
    }                                                                                                                  // 25
  }, {                                                                                                                 // 24
    multi: true                                                                                                        // 31
  });                                                                                                                  // 30
};                                                                                                                     // 33
                                                                                                                       //
Meteor.startup(cleanupExpiredTokens); // Tokens can actually last up to 2 * TOKEN_EXPIRATION_MS                        // 35
                                                                                                                       //
SandstormDb.periodicCleanup(TOKEN_EXPIRATION_MS, cleanupExpiredTokens);                                                // 37
                                                                                                                       //
var hashToken = function (token) {                                                                                     // 39
  return {                                                                                                             // 40
    digest: SHA256(token),                                                                                             // 41
    algorithm: "sha-256"                                                                                               // 42
  };                                                                                                                   // 40
};                                                                                                                     // 44
                                                                                                                       //
var checkToken = function (tokens, token) {                                                                            // 46
  // Looks for an object in `tokens` with `algorithm` and `digest` fields matching those in `token`.                   // 47
  // Returns the matching object, if one is found, or undefined if none match.                                         // 48
  var foundToken = undefined;                                                                                          // 49
  tokens.forEach(function (userToken) {                                                                                // 50
    if (userToken.algorithm === token.algorithm && userToken.digest === token.digest) {                                // 51
      foundToken = userToken;                                                                                          // 53
    }                                                                                                                  // 54
  });                                                                                                                  // 55
  return foundToken;                                                                                                   // 57
};                                                                                                                     // 58
                                                                                                                       //
var consumeToken = function (user, token) {                                                                            // 60
  var hashedToken = hashToken(token);                                                                                  // 61
  var foundToken = checkToken(user.services.email.tokens, hashedToken);                                                // 62
                                                                                                                       //
  if (foundToken !== undefined) {                                                                                      // 64
    Meteor.users.update({                                                                                              // 65
      _id: user._id                                                                                                    // 65
    }, {                                                                                                               // 65
      $pull: {                                                                                                         // 65
        "services.email.tokens": hashedToken                                                                           // 65
      }                                                                                                                // 65
    });                                                                                                                // 65
  }                                                                                                                    // 66
                                                                                                                       //
  return foundToken;                                                                                                   // 68
};                                                                                                                     // 69
                                                                                                                       //
var makeBox = function (token, plaintext) {                                                                            // 71
  // Encrypt plaintext symmetrically with a key derived from token.  Returns an object with                            // 72
  // ciphertext and associated data needed to decrypt later.                                                           // 73
  // Produce a symmetric key.  Note that the token itself does not have sufficient entropy to                          // 75
  // be used as a key directly, so we need to use a KDF with a strong random salt.                                     // 76
  // In the fullness of time, it might be nice to move away from using a KDF (which blocks the whole                   // 77
  // node process) in favor of the token itself having enough entropy to serve as the key itself.                      // 78
  // This would require lengthening the token, which would make the manual-code-entry workflow                         // 79
  // worse, so I'm punting on that for now.                                                                            // 80
  var salt = Random.secret(16);                                                                                        // 81
  var key = crypto.pbkdf2Sync(token, salt, V1_ROUNDS, V1_KEYSIZE, V1_HASHFUNC);                                        // 82
  var iv = crypto.randomBytes(16);                                                                                     // 83
  var cipher = crypto.createCipheriv(V1_CIPHER, key, iv);                                                              // 84
  var ciphertext = cipher.update(new Buffer(plaintext, "binary"));                                                     // 85
  return {                                                                                                             // 86
    version: 1,                                                                                                        // 87
    salt: salt,                                                                                                        // 88
    iv: iv.toString("base64"),                                                                                         // 89
    boxedValue: ciphertext.toString("base64")                                                                          // 90
  };                                                                                                                   // 86
};                                                                                                                     // 92
                                                                                                                       //
var tryUnbox = function (box, secret) {                                                                                // 94
  if (box) {                                                                                                           // 95
    if (box.version === 1) {                                                                                           // 96
      var key = crypto.pbkdf2Sync(secret, box.salt, V1_ROUNDS, V1_KEYSIZE, V1_HASHFUNC);                               // 97
      var iv = new Buffer(box.iv, "base64");                                                                           // 98
      var cipher = crypto.createDecipheriv(V1_CIPHER, key, iv);                                                        // 99
      var cipherText = new Buffer(box.boxedValue, "base64");                                                           // 100
      var plaintext = cipher.update(cipherText);                                                                       // 101
      return plaintext.toString("binary");                                                                             // 102
    }                                                                                                                  // 103
  } // If no box was provided, or it was of an unknown version, return no data.                                        // 104
                                                                                                                       //
                                                                                                                       //
  return;                                                                                                              // 107
}; // Handler to login with a token.                                                                                   // 108
                                                                                                                       //
                                                                                                                       //
Accounts.registerLoginHandler("email", function (options) {                                                            // 111
  if (!options.email) {                                                                                                // 112
    return undefined; // don't handle                                                                                  // 113
  }                                                                                                                    // 114
                                                                                                                       //
  if (!Accounts.loginServices.email.isEnabled()) {                                                                     // 116
    throw new Meteor.Error(403, "Email login service is disabled.");                                                   // 117
  }                                                                                                                    // 118
                                                                                                                       //
  options = options.email;                                                                                             // 120
  check(options, {                                                                                                     // 121
    email: String,                                                                                                     // 122
    token: String                                                                                                      // 123
  });                                                                                                                  // 121
  var user = Meteor.users.findOne({                                                                                    // 126
    "services.email.email": options.email                                                                              // 127
  }, {                                                                                                                 // 126
    fields: {                                                                                                          // 129
      "services.email": 1                                                                                              // 130
    }                                                                                                                  // 129
  });                                                                                                                  // 128
                                                                                                                       //
  if (!user) {                                                                                                         // 134
    console.error("User not found:", options.email);                                                                   // 135
    return {                                                                                                           // 136
      error: new Meteor.Error(403, "User not found")                                                                   // 137
    };                                                                                                                 // 136
  }                                                                                                                    // 139
                                                                                                                       //
  if (!user.services.email.tokens) {                                                                                   // 141
    console.error("User has no token set:", options.email);                                                            // 142
    return {                                                                                                           // 143
      error: new Meteor.Error(403, "User has no token set")                                                            // 144
    };                                                                                                                 // 143
  }                                                                                                                    // 146
                                                                                                                       //
  var tokenString = options.token.trim();                                                                              // 148
  var maybeToken = consumeToken(user, tokenString);                                                                    // 149
                                                                                                                       //
  if (!maybeToken) {                                                                                                   // 150
    console.error("Token not found:", options.email);                                                                  // 151
    return {                                                                                                           // 152
      error: new Meteor.Error(403, "Invalid authentication code")                                                      // 153
    };                                                                                                                 // 152
  } // Attempt to decrypt the resumePath, if provided.                                                                 // 155
                                                                                                                       //
                                                                                                                       //
  var resumePath = tryUnbox(maybeToken.secureBox, tokenString);                                                        // 158
  return {                                                                                                             // 160
    userId: user._id,                                                                                                  // 161
    options: {                                                                                                         // 162
      resumePath: resumePath                                                                                           // 163
    }                                                                                                                  // 162
  };                                                                                                                   // 160
});                                                                                                                    // 166
                                                                                                                       //
var makeTokenUrl = function (email, token, options) {                                                                  // 168
  if (options.linking) {                                                                                               // 169
    return options.rootUrl + "/_emailLinkCredential/" + encodeURIComponent(email) + "/" + encodeURIComponent(token) + "/" + Meteor.userId() + "?allowLogin=" + options.linking.allowLogin;
  } else {                                                                                                             // 173
    return options.rootUrl + "/_emailLogin/" + encodeURIComponent(email) + "/" + encodeURIComponent(token);            // 174
  }                                                                                                                    // 175
}; ///                                                                                                                 // 176
/// EMAIL VERIFICATION                                                                                                 // 179
///                                                                                                                    // 180
                                                                                                                       //
                                                                                                                       //
var sendTokenEmail = function (db, email, token, options) {                                                            // 181
  var subject = void 0;                                                                                                // 182
  var text = void 0;                                                                                                   // 183
  var rootHostname = Url.parse(options.rootUrl).hostname;                                                              // 185
                                                                                                                       //
  if (!options.linking) {                                                                                              // 187
    subject = "Log in to " + rootHostname;                                                                             // 188
    text = "To confirm this email address on ";                                                                        // 189
  } else {                                                                                                             // 190
    subject = "Confirm this email address on " + rootHostname;                                                         // 191
    text = "To confirm this email address on ";                                                                        // 192
  }                                                                                                                    // 193
                                                                                                                       //
  text = text + rootHostname + ", click on the following link:\n\n" + makeTokenUrl(email, token, options) + "\n\n" + "Alternatively, enter the following one-time authentication code into the log-in form:\n\n" + token;
  var sendOptions = {                                                                                                  // 200
    to: email,                                                                                                         // 201
    from: {                                                                                                            // 202
      name: globalDb.getServerTitle(),                                                                                 // 202
      address: db.getReturnAddress()                                                                                   // 202
    },                                                                                                                 // 202
    subject: subject,                                                                                                  // 203
    text: text                                                                                                         // 204
  };                                                                                                                   // 200
  sendEmail(sendOptions);                                                                                              // 207
};                                                                                                                     // 208
                                                                                                                       //
var parsedRootUrl = Url.parse(process.env.ROOT_URL); ///                                                               // 210
/// CREATING USERS                                                                                                     // 212
///                                                                                                                    // 213
// returns the user id                                                                                                 // 214
                                                                                                                       //
var createAndEmailTokenForUser = function (db, email, options) {                                                       // 215
  check(email, String);                                                                                                // 216
  check(options, {                                                                                                     // 217
    resumePath: String,                                                                                                // 218
    linking: Match.Optional({                                                                                          // 219
      allowLogin: Boolean                                                                                              // 219
    }),                                                                                                                // 219
    rootUrl: String                                                                                                    // 220
  });                                                                                                                  // 217
  var parsedUrl = Url.parse(options.rootUrl);                                                                          // 223
                                                                                                                       //
  if ((parsedUrl.hostname !== parsedRootUrl.hostname || parsedUrl.protocol !== parsedRootUrl.protocol) && !db.hostIsStandalone(parsedUrl.hostname)) {
    // Ignore port and only check hostname/protocol since IE will differ from other browsers and                       // 227
    // sometimes include port 80/443 and sometimes won't                                                               // 228
    throw new Meteor.Error(400, "rootUrl is not valid");                                                               // 229
  }                                                                                                                    // 230
                                                                                                                       //
  var atIndex = email.indexOf("@");                                                                                    // 232
                                                                                                                       //
  if (atIndex === -1) {                                                                                                // 233
    throw new Meteor.Error(400, "No @ symbol was found in your email");                                                // 234
  }                                                                                                                    // 235
                                                                                                                       //
  var user = Meteor.users.findOne({                                                                                    // 237
    "services.email.email": email                                                                                      // 237
  }, {                                                                                                                 // 237
    fields: {                                                                                                          // 238
      "services.email": 1                                                                                              // 238
    }                                                                                                                  // 238
  });                                                                                                                  // 238
  var userId = void 0; // TODO(someday): make this shorter, and handle requests that try to brute force it.            // 239
  // Alternately, require using the link over copy/pasting the code, and crank up the entropy.                         // 242
                                                                                                                       //
  var token = Random.id(12);                                                                                           // 243
  var tokenObj = hashToken(token);                                                                                     // 244
  tokenObj.createdAt = new Date();                                                                                     // 245
  tokenObj.secureBox = makeBox(token, options.resumePath);                                                             // 246
                                                                                                                       //
  if (user) {                                                                                                          // 248
    if (user.services.email.tokens && user.services.email.tokens.length > 2) {                                         // 249
      throw new Meteor.Error("alreadySentEmailToken", "It looks like we sent a log in email to this address not long " + "ago. Please use the one that was already sent (check your spam folder if you can't find " + "it), or wait a while and try again.");
    }                                                                                                                  // 255
                                                                                                                       //
    userId = user._id;                                                                                                 // 257
    Meteor.users.update({                                                                                              // 259
      _id: user._id                                                                                                    // 259
    }, {                                                                                                               // 259
      $push: {                                                                                                         // 259
        "services.email.tokens": tokenObj                                                                              // 259
      }                                                                                                                // 259
    });                                                                                                                // 259
  } else {                                                                                                             // 260
    var _options = {};                                                                                                 // 261
    user = {                                                                                                           // 262
      services: {                                                                                                      // 263
        email: {                                                                                                       // 264
          tokens: [tokenObj],                                                                                          // 265
          email: email                                                                                                 // 266
        }                                                                                                              // 264
      }                                                                                                                // 263
    };                                                                                                                 // 262
    userId = Accounts.insertUserDoc(_options, user);                                                                   // 271
  }                                                                                                                    // 272
                                                                                                                       //
  sendTokenEmail(db, email, token, options);                                                                           // 274
  return userId;                                                                                                       // 276
};                                                                                                                     // 277
                                                                                                                       //
Meteor.methods({                                                                                                       // 279
  createAndEmailTokenForUser: function (email, options) {                                                              // 280
    // method for create user. Requests come from the client.                                                          // 281
    // This method will create a user if it doesn't exist, otherwise it will generate a token.                         // 282
    // It will always send an email to the user                                                                        // 283
    check(email, String);                                                                                              // 285
    check(options, {                                                                                                   // 286
      resumePath: String,                                                                                              // 287
      linking: Match.Optional({                                                                                        // 288
        allowLogin: Boolean                                                                                            // 288
      }),                                                                                                              // 288
      rootUrl: String                                                                                                  // 289
    });                                                                                                                // 286
                                                                                                                       //
    if (!Accounts.loginServices.email.isEnabled()) {                                                                   // 292
      throw new Meteor.Error(403, "Email login service is disabled.");                                                 // 293
    } // Create user. result contains id and token.                                                                    // 294
                                                                                                                       //
                                                                                                                       //
    var user = createAndEmailTokenForUser(this.connection.sandstormDb, email, options);                                // 296
  },                                                                                                                   // 297
  linkEmailCredentialToAccount: function (email, token, allowLogin) {                                                  // 299
    // Links the email credential with address `email` and login token `token` to the current                          // 300
    // account.                                                                                                        // 301
    check(email, String);                                                                                              // 302
    check(token, String);                                                                                              // 303
    check(allowLogin, Boolean);                                                                                        // 304
    var account = Meteor.user();                                                                                       // 305
                                                                                                                       //
    if (!account || !account.loginCredentials) {                                                                       // 306
      throw new Meteor.Error(403, "Must be logged in to an account to link an email credential.");                     // 307
    }                                                                                                                  // 308
                                                                                                                       //
    var credential = Meteor.users.findOne({                                                                            // 310
      "services.email.email": email                                                                                    // 310
    }, {                                                                                                               // 310
      fields: {                                                                                                        // 311
        "services.email": 1                                                                                            // 311
      }                                                                                                                // 311
    });                                                                                                                // 311
                                                                                                                       //
    if (!credential) {                                                                                                 // 312
      throw new Meteor.Error(403, "Invalid authentication code.");                                                     // 313
    }                                                                                                                  // 314
                                                                                                                       //
    var maybeToken = consumeToken(credential, token.trim());                                                           // 316
                                                                                                                       //
    if (!maybeToken) {                                                                                                 // 317
      throw new Meteor.Error(403, "Invalid authentication code.");                                                     // 318
    }                                                                                                                  // 319
                                                                                                                       //
    Accounts.linkCredentialToAccount(this.connection.sandstormDb, this.connection.sandstormBackend, credential._id, account._id, allowLogin); // Return the resume path, if we have one.
                                                                                                                       //
    var resumePath = tryUnbox(maybeToken.secureBox, token);                                                            // 325
    return resumePath;                                                                                                 // 326
  }                                                                                                                    // 327
});                                                                                                                    // 279
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"ldap":{"ldap-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/ldap/ldap-server.js                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var LDAP = void 0;                                                                                                     // 1
module.watch(require("/imports/server/accounts/ldap.js"), {                                                            // 1
  LDAP: function (v) {                                                                                                 // 1
    LDAP = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
// Register login handler with Meteor                                                                                  // 3
Accounts.registerLoginHandler("ldap", function (loginRequest) {                                                        // 4
  // If 'ldap' isn't set in loginRequest object,                                                                       // 5
  // then this isn't the proper handler (return undefined)                                                             // 6
  if (!loginRequest.ldap) {                                                                                            // 7
    return undefined;                                                                                                  // 8
  }                                                                                                                    // 9
                                                                                                                       //
  if (!Accounts.loginServices.ldap.isEnabled()) {                                                                      // 11
    throw new Meteor.Error(403, "LDAP service is disabled.");                                                          // 12
  }                                                                                                                    // 13
                                                                                                                       //
  check(loginRequest, {                                                                                                // 15
    ldap: true,                                                                                                        // 16
    username: String,                                                                                                  // 17
    ldapPass: String                                                                                                   // 18
  }); // Instantiate LDAP                                                                                              // 15
                                                                                                                       //
  var ldapObj = new LDAP(); // Call ldapCheck and get response                                                         // 22
                                                                                                                       //
  var ldapResponse = ldapObj.ldapCheck(this.connection.sandstormDb, loginRequest);                                     // 25
                                                                                                                       //
  if (ldapResponse.error) {                                                                                            // 27
    return {                                                                                                           // 28
      userId: null,                                                                                                    // 29
      error: ldapResponse.error                                                                                        // 30
    };                                                                                                                 // 28
  } else if (ldapResponse.emptySearch) {                                                                               // 32
    return {                                                                                                           // 33
      userId: null,                                                                                                    // 34
      error: new Meteor.Error(403, "User not found in LDAP")                                                           // 35
    };                                                                                                                 // 33
  } else {                                                                                                             // 37
    // Set initial userId and token vals                                                                               // 38
    return Accounts.updateOrCreateUserFromExternalService("ldap", {                                                    // 39
      id: ldapResponse.dn,                                                                                             // 40
      username: loginRequest.username,                                                                                 // 40
      rawAttrs: ldapResponse.searchResults                                                                             // 41
    }, {});                                                                                                            // 40
  }                                                                                                                    // 42
});                                                                                                                    // 44
Meteor.methods({                                                                                                       // 46
  updateQuota: function () {                                                                                           // 47
    return this.connection.sandstormDb.updateUserQuota(Meteor.user()); // This is a no-op if settings aren't enabled   // 48
  }                                                                                                                    // 50
});                                                                                                                    // 46
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"saml":{"saml-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/saml/saml-server.js                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var module1 = module;                                                                                                  // 1
var Url = void 0;                                                                                                      // 1
module1.watch(require("url"), {                                                                                        // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var zlib = void 0;                                                                                                     // 1
module1.watch(require("zlib"), {                                                                                       // 1
  "default": function (v) {                                                                                            // 1
    zlib = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SAML = void 0;                                                                                                     // 1
module1.watch(require("/imports/server/accounts/saml-utils.js"), {                                                     // 1
  SAML: function (v) {                                                                                                 // 1
    SAML = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Fiber = Npm.require("fibers");                                                                                     // 5
                                                                                                                       //
if (!Accounts.saml) {                                                                                                  // 7
  Accounts.saml = {};                                                                                                  // 8
} // We need to use connect. Let's make sure we're using the same version as Meteor's WebApp module                    // 9
// uses. Fortunately, they let us extract it.                                                                          // 12
                                                                                                                       //
                                                                                                                       //
var connect = WebAppInternals.NpmModules.connect.module;                                                               // 13
RoutePolicy.declare("/_saml/", "network");                                                                             // 15
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname; // TODO(soon): This may need to be a Mongo collection in order to work when the frontend is
//   replicated (but currently SAML is not used on any Blackrock servers).                                             // 20
                                                                                                                       //
var _loginResultForCredentialToken = {};                                                                               // 21
                                                                                                                       //
var retrieveCredential = function (credentialToken) {                                                                  // 23
  var result = _loginResultForCredentialToken[credentialToken];                                                        // 24
  delete _loginResultForCredentialToken[credentialToken];                                                              // 25
  return result;                                                                                                       // 26
};                                                                                                                     // 27
                                                                                                                       //
Accounts.registerLoginHandler(function (loginRequest) {                                                                // 29
  if (!loginRequest.saml || !loginRequest.credentialToken) {                                                           // 30
    return undefined;                                                                                                  // 31
  }                                                                                                                    // 32
                                                                                                                       //
  if (!Accounts.loginServices.saml.isEnabled()) {                                                                      // 34
    throw new Meteor.Error(403, "SAML service is disabled.");                                                          // 35
  }                                                                                                                    // 36
                                                                                                                       //
  var loginResult = retrieveCredential(loginRequest.credentialToken);                                                  // 38
                                                                                                                       //
  if (!loginResult) {                                                                                                  // 39
    throw new Meteor.Error(500, "SAML login did not complete.");                                                       // 40
  } else if (loginResult.profile && loginResult.profile.email) {                                                       // 41
    var user = _.pick(loginResult.profile, "displayName", "email", "nameIDFormat");                                    // 42
                                                                                                                       //
    user.id = loginResult.profile.nameID;                                                                              // 43
    return Accounts.updateOrCreateUserFromExternalService("saml", user, {});                                           // 44
  } else {                                                                                                             // 45
    throw new Meteor.Error(500, "SAML profile did not contain an email address");                                      // 46
  }                                                                                                                    // 47
});                                                                                                                    // 48
                                                                                                                       //
var samlUrlToObject = function (url) {                                                                                 // 50
  // req.url will be "/_saml/<action>/<service name>/<credentialToken>"                                                // 51
  if (!url) return null;                                                                                               // 52
  var splitPath = url.split("/"); // Any non-saml request will continue down the default                               // 55
  // middlewares.                                                                                                      // 58
                                                                                                                       //
  if (splitPath[1] !== "_saml") return null;                                                                           // 59
  return {                                                                                                             // 62
    actionName: splitPath[2],                                                                                          // 63
    serviceName: splitPath[3],                                                                                         // 64
    credentialToken: splitPath[4]                                                                                      // 65
  };                                                                                                                   // 62
};                                                                                                                     // 67
                                                                                                                       //
var closePopup = function (res, err) {                                                                                 // 69
  res.writeHead(200, {                                                                                                 // 70
    "Content-Type": "text/html"                                                                                        // 70
  });                                                                                                                  // 70
  var content = "<html><head><script>window.close()</script></head></html>";                                           // 71
  if (err) content = "<html><body><h2>Sorry, an error occured</h2><div>" + err + '</div><a onclick="window.close();">Close Window</a> <a href="/admin/status" target="_blank">View system log (admin only)</a></body></html>';
  res.end(content, "utf-8");                                                                                           // 75
};                                                                                                                     // 76
                                                                                                                       //
var generateService = function () {                                                                                    // 78
  // TODO(cleanup): Inject the db.                                                                                     // 79
  var db = new SandstormDb();                                                                                          // 80
  var entityId = db.getSamlEntityId();                                                                                 // 82
  var service = {                                                                                                      // 83
    "provider": "default",                                                                                             // 84
    "entryPoint": db.getSamlEntryPoint(),                                                                              // 85
    "logoutUrl": db.getSamlLogout(),                                                                                   // 86
    // TODO(someday): find a better way to inject the DB                                                               // 87
    "issuer": entityId || HOSTNAME,                                                                                    // 88
    // If the certificate has "-----BEGIN CERTIFICATE-----" markers, automatically remove those.                       // 89
    "cert": db.getSamlPublicCert().replace(/-[^\n]*-/g, "").trim()                                                     // 90
  };                                                                                                                   // 83
  return service;                                                                                                      // 92
};                                                                                                                     // 93
                                                                                                                       //
var middleware = function (req, res, next) {                                                                           // 95
  // Make sure to catch any exceptions because otherwise we'd crash                                                    // 96
  // the runner                                                                                                        // 97
  try {                                                                                                                // 98
    var samlObject = samlUrlToObject(req.url);                                                                         // 99
                                                                                                                       //
    if (!samlObject || !samlObject.serviceName) {                                                                      // 100
      next();                                                                                                          // 101
      return;                                                                                                          // 102
    }                                                                                                                  // 103
                                                                                                                       //
    if (samlObject.actionName === "config") {                                                                          // 105
      var _saml = new SAML(generateService());                                                                         // 106
                                                                                                                       //
      res.writeHead(200, {                                                                                             // 107
        "Content-Type": "text/xml"                                                                                     // 107
      });                                                                                                              // 107
      res.end(_saml.generateServiceProviderMetadata());                                                                // 108
      return;                                                                                                          // 109
    }                                                                                                                  // 110
                                                                                                                       //
    if (!Accounts.loginServices.saml.isEnabled()) {                                                                    // 112
      next();                                                                                                          // 113
      return;                                                                                                          // 114
    }                                                                                                                  // 115
                                                                                                                       //
    if (!samlObject.actionName) throw new Error("Missing SAML action");                                                // 117
    var service = generateService(); // Skip everything if there's no service set by the saml middleware               // 120
                                                                                                                       //
    if (!service || samlObject.serviceName !== service.provider) throw new Error("Unexpected SAML service " + samlObject.serviceName);
                                                                                                                       //
    if (samlObject.actionName === "authorize") {                                                                       // 126
      service.callbackUrl = Meteor.absoluteUrl("_saml/validate/" + service.provider);                                  // 127
      service.id = samlObject.credentialToken;                                                                         // 128
                                                                                                                       //
      var _saml2 = new SAML(service);                                                                                  // 129
                                                                                                                       //
      _saml2.getAuthorizeUrl(req, function (err, url) {                                                                // 130
        if (err) throw new Error("Unable to generate authorize url");                                                  // 131
        res.writeHead(302, {                                                                                           // 133
          "Location": url                                                                                              // 133
        });                                                                                                            // 133
        res.end();                                                                                                     // 134
      });                                                                                                              // 135
    } else if (samlObject.actionName === "validate") {                                                                 // 136
      var _saml3 = new SAML(service);                                                                                  // 137
                                                                                                                       //
      _saml3.validateResponse(req.body.SAMLResponse, function (err, profile, loggedOut, responseText) {                // 138
        if (err) {                                                                                                     // 140
          console.error("Error validating SAML response:", err.toString(), "\nFull SAML response XML:\n", responseText);
          throw new Error("Unable to validate SAML response.");                                                        // 143
        } // Do NOT use samlObject.credentialToken; it isn't signed!                                                   // 144
                                                                                                                       //
                                                                                                                       //
        var credentialToken = profile.inResponseToId || profile.InResponseTo;                                          // 147
                                                                                                                       //
        if (!credentialToken) {                                                                                        // 148
          throw new Error("SAML response missing InResponseTo attribute. Sandstorm does not support " + "IdP-initiated authentication; authentication requests must start " + "from the user choosing SAML login in the Sandstorm UI.");
        }                                                                                                              // 153
                                                                                                                       //
        _loginResultForCredentialToken[credentialToken] = {                                                            // 155
          profile: profile                                                                                             // 156
        };                                                                                                             // 155
        closePopup(res);                                                                                               // 159
      });                                                                                                              // 160
    } else {                                                                                                           // 161
      throw new Error("Unexpected SAML action " + samlObject.actionName);                                              // 162
    }                                                                                                                  // 163
  } catch (err) {                                                                                                      // 164
    console.error(err.stack);                                                                                          // 165
    closePopup(res, err);                                                                                              // 166
  }                                                                                                                    // 167
}; // Listen to incoming OAuth http requests                                                                           // 168
                                                                                                                       //
                                                                                                                       //
WebApp.connectHandlers.use(connect.urlencoded()).use(function (req, res, next) {                                       // 171
  // Need to create a Fiber since we're using synchronous http calls and nothing                                       // 172
  // else is wrapping this in a fiber automatically                                                                    // 173
  Fiber(function () {                                                                                                  // 174
    middleware(req, res, next);                                                                                        // 175
  }).run();                                                                                                            // 176
});                                                                                                                    // 177
Meteor.methods({                                                                                                       // 179
  generateSamlLogout: function () {                                                                                    // 180
    var service = generateService();                                                                                   // 181
                                                                                                                       //
    if (!service.logoutUrl) {                                                                                          // 182
      throw new Meteor.Error(500, "No SAML logout url specified");                                                     // 183
    }                                                                                                                  // 184
                                                                                                                       //
    var _saml = new SAML(service);                                                                                     // 186
                                                                                                                       //
    var credential = void 0;                                                                                           // 187
    Meteor.user().loginCredentials.forEach(function (_credential) {                                                    // 188
      var currCredential = Meteor.users.findOne({                                                                      // 189
        _id: _credential.id                                                                                            // 189
      });                                                                                                              // 189
                                                                                                                       //
      if (currCredential.services.saml) {                                                                              // 190
        credential = currCredential;                                                                                   // 191
      }                                                                                                                // 192
    }); // TODO(someday): handle user having more than one SAML credential                                             // 193
                                                                                                                       //
    return Meteor.wrapAsync(_saml.getLogoutUrl.bind(_saml))({                                                          // 196
      user: {                                                                                                          // 197
        nameID: credential.services.saml.id,                                                                           // 198
        nameIDFormat: credential.services.saml.nameIDFormat || "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent"  // 199
      }                                                                                                                // 197
    });                                                                                                                // 196
  },                                                                                                                   // 203
  validateSamlLogout: function (samlRequest) {                                                                         // 205
    check(samlRequest, String);                                                                                        // 206
    var db = this.connection.sandstormDb;                                                                              // 208
    var userId = Meteor.userId();                                                                                      // 209
                                                                                                                       //
    if (!userId) {                                                                                                     // 210
      return new Meteor.Error(403, "Non-logged in users can't logout.");                                               // 211
    }                                                                                                                  // 212
                                                                                                                       //
    var service = generateService();                                                                                   // 214
                                                                                                                       //
    var _saml = new SAML(service);                                                                                     // 215
                                                                                                                       //
    if (samlRequest) {                                                                                                 // 216
      var buf = new Buffer(samlRequest, "base64");                                                                     // 217
      var xml = zlib.inflateRawSync(buf).toString();                                                                   // 218
                                                                                                                       //
      _saml.parseLogoutRequest(xml, function (err, nameId) {                                                           // 219
        if (err) {                                                                                                     // 220
          console.error("Error validating SAML logout response:", err.toString(), "\nFull SAML response XML:\n", responseText);
          throw new Error("Unable to validate SAML logout response.");                                                 // 223
        }                                                                                                              // 224
                                                                                                                       //
        check(nameId, String);                                                                                         // 226
        var credential = db.collections.users.findOne({                                                                // 227
          "services.saml.id": nameId                                                                                   // 227
        }, {                                                                                                           // 227
          fields: {                                                                                                    // 228
            _id: 1                                                                                                     // 228
          }                                                                                                            // 228
        });                                                                                                            // 228
                                                                                                                       //
        if (!credential) {                                                                                             // 229
          return new Meteor.Error(400, "No credential found matching SAML nameID.");                                   // 230
        }                                                                                                              // 231
                                                                                                                       //
        var user = db.collections.users.findOne({                                                                      // 233
          "loginCredentials.id": credential._id                                                                        // 233
        }, {                                                                                                           // 233
          fields: {                                                                                                    // 234
            _id: 1                                                                                                     // 234
          }                                                                                                            // 234
        });                                                                                                            // 234
                                                                                                                       //
        if (!user) {                                                                                                   // 235
          return new Meteor.Error(403, "No user found for expected SAML credential.");                                 // 236
        }                                                                                                              // 237
                                                                                                                       //
        if (user._id !== userId) {                                                                                     // 239
          var txt = "SAML logout requested for wrong user: " + nameId + ", " + userId;                                 // 240
          console.error(txt);                                                                                          // 241
          throw new Error(txt);                                                                                        // 242
        }                                                                                                              // 243
      });                                                                                                              // 244
    }                                                                                                                  // 245
  }                                                                                                                    // 246
});                                                                                                                    // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"accounts-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-server.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Accounts = void 0;                                                                                                 // 1
module.watch(require("meteor/accounts-base"), {                                                                        // 1
  Accounts: function (v) {                                                                                             // 1
    Accounts = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var fetchPicture = void 0,                                                                                             // 1
    userPictureUrl = void 0;                                                                                           // 1
module.watch(require("/imports/server/accounts/picture.js"), {                                                         // 1
  fetchPicture: function (v) {                                                                                         // 1
    fetchPicture = v;                                                                                                  // 1
  },                                                                                                                   // 1
  userPictureUrl: function (v) {                                                                                       // 1
    userPictureUrl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 20
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 21
                                                                                                                       //
var ValidHandle = Match.Where(function (handle) {                                                                      // 23
  check(handle, String);                                                                                               // 24
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                         // 25
});                                                                                                                    // 26
Accounts.onCreateUser(function (options, user) {                                                                       // 28
  if (!user.type) {                                                                                                    // 29
    if (user.loginCredentials) throw new Error("please set user.type");                                                // 30
    user.type = "credential";                                                                                          // 32
  }                                                                                                                    // 33
                                                                                                                       //
  if (user.type === "account") {                                                                                       // 35
    // it's an account                                                                                                 // 36
    check(user, {                                                                                                      // 37
      _id: String,                                                                                                     // 37
      type: String,                                                                                                    // 38
      createdAt: Date,                                                                                                 // 39
      isAdmin: Match.Optional(Boolean),                                                                                // 40
      hasCompletedSignup: Match.Optional(Boolean),                                                                     // 41
      signupKey: Match.Optional(String),                                                                               // 42
      signupNote: Match.Optional(String),                                                                              // 43
      signupEmail: Match.Optional(String),                                                                             // 44
      expires: Match.Optional(Date),                                                                                   // 45
      appDemoId: Match.Optional(String),                                                                               // 46
      loginCredentials: [{                                                                                             // 47
        id: String                                                                                                     // 47
      }],                                                                                                              // 47
      nonloginCredentials: [{                                                                                          // 48
        id: String                                                                                                     // 48
      }]                                                                                                               // 48
    });                                                                                                                // 37
                                                                                                                       //
    if (globalDb.isReferralEnabled()) {                                                                                // 50
      user.experiments = user.experiments || {};                                                                       // 51
      user.experiments = {                                                                                             // 52
        freeGrainLimit: Math.random() < 0.1 ? 100 : "control"                                                          // 53
      };                                                                                                               // 52
                                                                                                                       //
      if (!("expires" in user)) {                                                                                      // 55
        globalDb.sendReferralProgramNotification(user._id);                                                            // 56
      }                                                                                                                // 57
    }                                                                                                                  // 58
                                                                                                                       //
    if (!options.profile) throw new Error("missing profile");                                                          // 60
    user.profile = options.profile;                                                                                    // 62
    globalDb.preinstallAppsForUser(user._id);                                                                          // 64
    return user;                                                                                                       // 66
  }                                                                                                                    // 67
                                                                                                                       //
  if (globalDb.getOrganizationDisallowGuests() && !globalDb.isCredentialInOrganization(user)) {                        // 69
    throw new Meteor.Error(400, "User not in organization.");                                                          // 71
  } // Check profile.                                                                                                  // 72
                                                                                                                       //
                                                                                                                       //
  if (options.profile) {                                                                                               // 75
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                               // 76
    check(options.profile, Match.ObjectIncluding({                                                                     // 77
      name: Match.OneOf(null, Match.Optional(String)),                                                                 // 78
      handle: Match.Optional(ValidHandle),                                                                             // 79
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot"))                                       // 80
    }));                                                                                                               // 77
  }                                                                                                                    // 82
                                                                                                                       //
  check(options.unverifiedEmail, Match.Optional(String));                                                              // 84
                                                                                                                       //
  if (options.unverifiedEmail) {                                                                                       // 86
    user.unverifiedEmail = options.unverifiedEmail;                                                                    // 87
  }                                                                                                                    // 88
                                                                                                                       //
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronoun"); // Try downloading avatar.                // 90
                                                                                                                       //
  var url = userPictureUrl(user);                                                                                      // 93
                                                                                                                       //
  if (url) {                                                                                                           // 94
    var assetId = fetchPicture(globalDb, url);                                                                         // 95
                                                                                                                       //
    if (assetId) {                                                                                                     // 96
      user.profile.picture = assetId;                                                                                  // 97
    }                                                                                                                  // 98
  }                                                                                                                    // 99
                                                                                                                       //
  var serviceUserId = void 0;                                                                                          // 101
                                                                                                                       //
  if (user.services && user.services.dev) {                                                                            // 102
    check(user.services.dev, {                                                                                         // 103
      name: String,                                                                                                    // 103
      isAdmin: Boolean,                                                                                                // 103
      hasCompletedSignup: Boolean                                                                                      // 103
    });                                                                                                                // 103
    serviceUserId = user.services.dev.name;                                                                            // 104
  } else if ("expires" in user) {                                                                                      // 105
    serviceUserId = user._id;                                                                                          // 106
    user.services = {                                                                                                  // 107
      demo: {}                                                                                                         // 107
    };                                                                                                                 // 107
  } else if (user.services && user.services.email) {                                                                   // 108
    check(user.services.email, {                                                                                       // 109
      email: String,                                                                                                   // 110
      tokens: [{                                                                                                       // 111
        digest: String,                                                                                                // 113
        algorithm: String,                                                                                             // 114
        createdAt: Date,                                                                                               // 115
        secureBox: Match.Optional({                                                                                    // 116
          version: Number,                                                                                             // 117
          salt: String,                                                                                                // 118
          iv: String,                                                                                                  // 119
          boxedValue: String                                                                                           // 120
        })                                                                                                             // 116
      }]                                                                                                               // 112
    });                                                                                                                // 109
    serviceUserId = user.services.email.email;                                                                         // 125
  } else if (user.services && "google" in user.services) {                                                             // 126
    serviceUserId = user.services.google.id;                                                                           // 127
  } else if (user.services && "github" in user.services) {                                                             // 128
    serviceUserId = user.services.github.id;                                                                           // 129
  } else if (user.services && "ldap" in user.services) {                                                               // 130
    serviceUserId = user.services.ldap.id;                                                                             // 131
  } else if (user.services && "saml" in user.services) {                                                               // 132
    serviceUserId = user.services.saml.id;                                                                             // 133
  } else {                                                                                                             // 134
    throw new Meteor.Error(400, "user does not have a recognized login provider: " + JSON.stringify(user));            // 135
  }                                                                                                                    // 137
                                                                                                                       //
  user._id = Crypto.createHash("sha256").update(SandstormDb.getServiceName(user) + ":" + serviceUserId).digest("hex");
  return user;                                                                                                         // 142
});                                                                                                                    // 143
Accounts.validateLoginAttempt(function (attempt) {                                                                     // 145
  if (!attempt.allowed) {                                                                                              // 146
    return false;                                                                                                      // 147
  }                                                                                                                    // 148
                                                                                                                       //
  var db = attempt.connection.sandstormDb;                                                                             // 150
  var user = attempt.user;                                                                                             // 151
                                                                                                                       //
  if (user.suspended && user.suspended.admin) {                                                                        // 153
    throw new Meteor.Error(403, "User has been suspended. Please contact the administrator " + "of this server if you believe this to be in error.");
  }                                                                                                                    // 156
                                                                                                                       //
  if (user.loginCredentials) {                                                                                         // 158
    // it's an account                                                                                                 // 159
    if (db.getOrganizationDisallowGuests() && !db.isUserInOrganization(user)) {                                        // 160
      throw new Meteor.Error(403, "User not in organization.");                                                        // 162
    }                                                                                                                  // 163
                                                                                                                       //
    db.updateUserQuota(user); // This is a no-op if settings aren't enabled                                            // 165
  } else {                                                                                                             // 166
    if (db.getOrganizationDisallowGuests() && !db.isCredentialInOrganization(user)) {                                  // 167
      throw new Meteor.Error(403, "User not in organization.");                                                        // 169
    }                                                                                                                  // 170
  }                                                                                                                    // 171
                                                                                                                       //
  return true;                                                                                                         // 173
});                                                                                                                    // 174
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"accounts-ui-methods.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-ui-methods.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file contains method definitions which are available on both client and server (on the                         // 17
// client for prediction purposes, on the server for actual execution).                                                // 18
var ValidHandle = Match.Where(function (handle) {                                                                      // 20
  check(handle, String);                                                                                               // 21
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                         // 22
});                                                                                                                    // 23
Meteor.methods({                                                                                                       // 25
  updateProfile: function (obsolete, profile) {                                                                        // 26
    check(profile, {                                                                                                   // 27
      name: String,                                                                                                    // 28
      handle: ValidHandle,                                                                                             // 29
      pronoun: Match.OneOf("male", "female", "neutral", "robot"),                                                      // 30
      unverifiedEmail: Match.Optional(String)                                                                          // 31
    });                                                                                                                // 27
                                                                                                                       //
    if (!this.userId) {                                                                                                // 34
      throw new Meteor.Error(403, "not logged in");                                                                    // 35
    }                                                                                                                  // 36
                                                                                                                       //
    var userToUpdate = Meteor.user();                                                                                  // 38
    var newValues = {                                                                                                  // 40
      "profile.name": profile.name,                                                                                    // 41
      "profile.handle": profile.handle,                                                                                // 42
      "profile.pronoun": profile.pronoun                                                                               // 43
    };                                                                                                                 // 40
    Meteor.users.update({                                                                                              // 46
      _id: userToUpdate._id                                                                                            // 46
    }, {                                                                                                               // 46
      $set: newValues                                                                                                  // 46
    });                                                                                                                // 46
                                                                                                                       //
    if (!Meteor.user().hasCompletedSignup) {                                                                           // 48
      Meteor.users.update({                                                                                            // 49
        _id: this.userId                                                                                               // 49
      }, {                                                                                                             // 49
        $set: {                                                                                                        // 49
          hasCompletedSignup: true                                                                                     // 49
        }                                                                                                              // 49
      });                                                                                                              // 49
    }                                                                                                                  // 50
  },                                                                                                                   // 51
  testFirstSignup: function (profile) {                                                                                // 53
    if (!this.userId) {                                                                                                // 54
      throw new Meteor.Error(403, "not logged in");                                                                    // 55
    }                                                                                                                  // 56
                                                                                                                       //
    Meteor.users.update(this.userId, {                                                                                 // 58
      $unset: {                                                                                                        // 58
        hasCompletedSignup: ""                                                                                         // 58
      }                                                                                                                // 58
    });                                                                                                                // 58
  },                                                                                                                   // 59
  uploadProfilePicture: function (obsolete) {                                                                          // 61
    if (!this.userId) {                                                                                                // 62
      throw new Meteor.Error(403, "not logged in");                                                                    // 63
    }                                                                                                                  // 64
                                                                                                                       //
    return this.connection.sandstormDb.newAssetUpload({                                                                // 66
      profilePicture: {                                                                                                // 67
        userId: this.userId                                                                                            // 67
      }                                                                                                                // 67
    });                                                                                                                // 66
  },                                                                                                                   // 69
  cancelUploadProfilePicture: function (id) {                                                                          // 71
    check(id, String);                                                                                                 // 72
    this.connection.sandstormDb.fulfillAssetUpload(id);                                                                // 73
  },                                                                                                                   // 74
  setPrimaryEmail: function (email) {                                                                                  // 76
    check(email, String);                                                                                              // 77
                                                                                                                       //
    if (!this.userId) {                                                                                                // 78
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 79
    }                                                                                                                  // 80
                                                                                                                       //
    var emails = SandstormDb.getUserEmails(Meteor.user());                                                             // 82
                                                                                                                       //
    if (!_.findWhere(emails, {                                                                                         // 83
      email: email,                                                                                                    // 83
      verified: true                                                                                                   // 83
    })) {                                                                                                              // 83
      throw new Meteor.Error(403, "Not a verified email of the current user: " + email);                               // 84
    }                                                                                                                  // 85
                                                                                                                       //
    Meteor.users.update({                                                                                              // 87
      _id: this.userId                                                                                                 // 87
    }, {                                                                                                               // 87
      $set: {                                                                                                          // 87
        primaryEmail: email                                                                                            // 87
      }                                                                                                                // 87
    });                                                                                                                // 87
  }                                                                                                                    // 88
});                                                                                                                    // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"accounts-ui-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-ui-server.js                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// Meteor permits users to modify their own profile by default, for some reason.                                       // 17
Meteor.users.deny({                                                                                                    // 18
  insert: function () {                                                                                                // 19
    return true;                                                                                                       // 19
  },                                                                                                                   // 19
  update: function () {                                                                                                // 21
    return true;                                                                                                       // 21
  },                                                                                                                   // 21
  remove: function () {                                                                                                // 23
    return true;                                                                                                       // 23
  },                                                                                                                   // 23
  fetch: []                                                                                                            // 25
});                                                                                                                    // 18
Meteor.publish("getMyUsage", function () {                                                                             // 28
  var db = this.connection.sandstormDb;                                                                                // 29
                                                                                                                       //
  if (this.userId) {                                                                                                   // 30
    // TODO(someday): Make this reactive.                                                                              // 31
    var user = Meteor.users.findOne(this.userId);                                                                      // 32
    var usage = db.getMyUsage(user);                                                                                   // 33
    this.added("users", this.userId, {                                                                                 // 34
      pseudoUsage: usage                                                                                               // 35
    });                                                                                                                // 34
  }                                                                                                                    // 37
                                                                                                                       //
  this.ready();                                                                                                        // 39
});                                                                                                                    // 40
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"admin":{"admin-alert-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/admin-alert-server.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  Match: function (v) {                                                                                                // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  check: function (v) {                                                                                                // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.watch(require("/imports/server/auth.js"), {                                                                     // 1
  checkAuth: function (v) {                                                                                            // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var maintenanceMessageShape = {                                                                                        // 5
  text: String,                                                                                                        // 6
  time: Match.OneOf(Date, undefined, null),                                                                            // 7
  url: Match.Optional(String)                                                                                          // 8
};                                                                                                                     // 5
Meteor.methods({                                                                                                       // 11
  setMaintenanceMessage: function (params) {                                                                           // 12
    checkAuth(undefined);                                                                                              // 13
    check(params, maintenanceMessageShape);                                                                            // 14
    var db = this.connection.sandstormDb; // TODO(soon): make this a single write to a single settings object          // 15
                                                                                                                       //
    db.collections.settings.upsert({                                                                                   // 17
      _id: "adminAlertTime"                                                                                            // 17
    }, {                                                                                                               // 17
      value: params.time                                                                                               // 17
    });                                                                                                                // 17
    db.collections.settings.upsert({                                                                                   // 18
      _id: "adminAlertUrl"                                                                                             // 18
    }, {                                                                                                               // 18
      value: params.url                                                                                                // 18
    });                                                                                                                // 18
    db.collections.settings.upsert({                                                                                   // 19
      _id: "adminAlert"                                                                                                // 19
    }, {                                                                                                               // 19
      value: params.text                                                                                               // 19
    });                                                                                                                // 19
  }                                                                                                                    // 20
});                                                                                                                    // 11
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"admin-user-invite.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/admin-user-invite.js                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
Meteor.publish("allInviteTokens", function () {                                                                        // 1
  var db = this.connection.sandstormDb;                                                                                // 2
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 3
    throw new Meteor.Error(403, "User must be admin to list invite tokens.");                                          // 4
  }                                                                                                                    // 5
                                                                                                                       //
  return db.collections.signupKeys.find({});                                                                           // 7
});                                                                                                                    // 8
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"network-capabilities-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/network-capabilities-server.js                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var check = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  check: function (v) {                                                                                                // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.watch(require("/imports/server/auth.js"), {                                                                     // 1
  checkAuth: function (v) {                                                                                            // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
Meteor.publish("adminGrains", function (grainIds) {                                                                    // 5
  // If the caller is an admin, publishes the Grains referred to by the provided list of grain IDs.                    // 6
  // Otherwise, does nothing.                                                                                          // 7
  if (!isAdminById(this.userId)) return [];                                                                            // 8
  check(grainIds, [String]);                                                                                           // 9
  var db = this.connection.sandstormDb;                                                                                // 11
  return db.collections.grains.find({                                                                                  // 12
    _id: {                                                                                                             // 13
      $in: grainIds                                                                                                    // 14
    }                                                                                                                  // 13
  }, {                                                                                                                 // 12
    fields: {                                                                                                          // 17
      title: 1,                                                                                                        // 18
      packageId: 1                                                                                                     // 19
    }                                                                                                                  // 17
  });                                                                                                                  // 16
});                                                                                                                    // 22
Meteor.publish("adminPackages", function (packageIds) {                                                                // 24
  // If the caller is an admin, publishes the Packages referred to by the provided list of package                     // 25
  // IDs.  Otherwise, does nothing.                                                                                    // 26
  if (!isAdminById(this.userId)) return [];                                                                            // 27
  check(packageIds, [String]);                                                                                         // 28
  var db = this.connection.sandstormDb;                                                                                // 30
  return db.collections.packages.find({                                                                                // 31
    _id: {                                                                                                             // 32
      $in: packageIds                                                                                                  // 33
    }                                                                                                                  // 32
  }, {                                                                                                                 // 31
    fields: {                                                                                                          // 36
      manifest: 1                                                                                                      // 37
    }                                                                                                                  // 36
  });                                                                                                                  // 35
});                                                                                                                    // 40
Meteor.publish("adminProfiles", function (userIds) {                                                                   // 42
  // If the caller is an admin, publishes the Users listed by ID in userIds.                                           // 43
  // Otherwise, does nothing.                                                                                          // 44
  if (!isAdminById(this.userId)) return [];                                                                            // 45
  check(userIds, [String]);                                                                                            // 46
  var db = this.connection.sandstormDb;                                                                                // 48
  return db.collections.users.find({                                                                                   // 49
    _id: {                                                                                                             // 50
      $in: userIds                                                                                                     // 51
    }                                                                                                                  // 50
  }, {                                                                                                                 // 49
    fields: {                                                                                                          // 53
      profile: 1                                                                                                       // 53
    }                                                                                                                  // 53
  });                                                                                                                  // 53
});                                                                                                                    // 54
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"personalization-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/personalization-server.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var check = void 0;                                                                                                    // 1
module.watch(require("meteor/check"), {                                                                                // 1
  check: function (v) {                                                                                                // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.watch(require("/imports/server/auth.js"), {                                                                     // 1
  checkAuth: function (v) {                                                                                            // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var personalizationMessageShape = {                                                                                    // 5
  serverTitle: String,                                                                                                 // 6
  splashUrl: String,                                                                                                   // 7
  signupDialog: String,                                                                                                // 8
  termsOfServiceUrl: String,                                                                                           // 9
  privacyPolicyUrl: String,                                                                                            // 10
  whitelabelCustomLoginProviderName: String,                                                                           // 12
  whitelabelHideSendFeedback: Boolean,                                                                                 // 13
  whitelabelHideTroubleshooting: Boolean,                                                                              // 14
  whiteLabelHideAbout: Boolean,                                                                                        // 15
  whitelabelUseServerTitleForHomeText: Boolean                                                                         // 16
};                                                                                                                     // 5
Meteor.methods({                                                                                                       // 19
  setPersonalizationSettings: function (params) {                                                                      // 20
    checkAuth(undefined);                                                                                              // 21
    check(params, personalizationMessageShape);                                                                        // 22
    var db = this.connection.sandstormDb; // TODO(soon): make this a single write to a single settings object          // 23
                                                                                                                       //
    db.collections.settings.upsert({                                                                                   // 25
      _id: "serverTitle"                                                                                               // 25
    }, {                                                                                                               // 25
      value: params.serverTitle                                                                                        // 25
    });                                                                                                                // 25
    db.collections.settings.upsert({                                                                                   // 26
      _id: "splashUrl"                                                                                                 // 26
    }, {                                                                                                               // 26
      value: params.splashUrl                                                                                          // 26
    });                                                                                                                // 26
    db.collections.settings.upsert({                                                                                   // 27
      _id: "signupDialog"                                                                                              // 27
    }, {                                                                                                               // 27
      value: params.signupDialog                                                                                       // 27
    });                                                                                                                // 27
    db.collections.settings.upsert({                                                                                   // 28
      _id: "termsUrl"                                                                                                  // 28
    }, {                                                                                                               // 28
      value: params.termsOfServiceUrl                                                                                  // 28
    });                                                                                                                // 28
    db.collections.settings.upsert({                                                                                   // 29
      _id: "privacyUrl"                                                                                                // 29
    }, {                                                                                                               // 29
      value: params.privacyPolicyUrl                                                                                   // 29
    });                                                                                                                // 29
    db.collections.settings.upsert({                                                                                   // 31
      _id: "whitelabelCustomLoginProviderName"                                                                         // 31
    }, {                                                                                                               // 31
      value: params.whitelabelCustomLoginProviderName                                                                  // 32
    });                                                                                                                // 32
    db.collections.settings.upsert({                                                                                   // 33
      _id: "whitelabelHideSendFeedback"                                                                                // 33
    }, {                                                                                                               // 33
      value: params.whitelabelHideSendFeedback                                                                         // 34
    });                                                                                                                // 34
    db.collections.settings.upsert({                                                                                   // 35
      _id: "whitelabelHideTroubleshooting"                                                                             // 35
    }, {                                                                                                               // 35
      value: params.whitelabelHideTroubleshooting                                                                      // 36
    });                                                                                                                // 36
    db.collections.settings.upsert({                                                                                   // 37
      _id: "whiteLabelHideAbout"                                                                                       // 37
    }, {                                                                                                               // 37
      value: params.whiteLabelHideAbout                                                                                // 38
    });                                                                                                                // 38
    db.collections.settings.upsert({                                                                                   // 39
      _id: "whitelabelUseServerTitleForHomeText"                                                                       // 39
    }, {                                                                                                               // 39
      value: params.whitelabelUseServerTitleForHomeText                                                                // 40
    });                                                                                                                // 40
  },                                                                                                                   // 41
  getWhitelabelLogoUploadToken: function () {                                                                          // 43
    checkAuth(undefined);                                                                                              // 44
    var db = this.connection.sandstormDb;                                                                              // 45
    return db.newAssetUpload({                                                                                         // 46
      loginLogo: {}                                                                                                    // 46
    });                                                                                                                // 46
  },                                                                                                                   // 47
  resetWhitelabelLogo: function () {                                                                                   // 49
    checkAuth(undefined);                                                                                              // 50
    var db = this.connection.sandstormDb;                                                                              // 51
    var result = globalDb.collections.settings.findAndModify({                                                         // 52
      query: {                                                                                                         // 53
        _id: "whitelabelCustomLogoAssetId"                                                                             // 53
      },                                                                                                               // 53
      remove: true,                                                                                                    // 54
      fields: {                                                                                                        // 55
        value: 1                                                                                                       // 55
      }                                                                                                                // 55
    });                                                                                                                // 52
                                                                                                                       //
    if (result.ok) {                                                                                                   // 58
      var old = result.value;                                                                                          // 59
                                                                                                                       //
      if (old) {                                                                                                       // 60
        db.unrefStaticAsset(old.value);                                                                                // 61
      }                                                                                                                // 62
    } else {                                                                                                           // 63
      throw new Meteor.Error(500, "Couldn't remove whitelabelCustomLogoAssetId");                                      // 64
    }                                                                                                                  // 65
  }                                                                                                                    // 66
});                                                                                                                    // 19
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"system-status-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/system-status-server.js                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Router = void 0;                                                                                                   // 1
module.watch(require("meteor/iron:router"), {                                                                          // 1
  Router: function (v) {                                                                                               // 1
    Router = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Random = void 0;                                                                                                   // 1
module.watch(require("meteor/random"), {                                                                               // 1
  Random: function (v) {                                                                                               // 1
    Random = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.watch(require("crypto"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.watch(require("fs"), {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.watch(require("/imports/server/auth.js"), {                                                                     // 1
  checkAuth: function (v) {                                                                                            // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var SANDSTORM_LOGDIR = void 0;                                                                                         // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_LOGDIR: function (v) {                                                                                     // 1
    SANDSTORM_LOGDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
                                                                                                                       //
var hashSessionId = function (sessionId) {                                                                             // 10
  return Crypto.createHash("sha256").update(sessionId).digest("base64");                                               // 11
}; // An in-memory map of random id -> timestamp, used to authorize requests to download the system                    // 12
// log.                                                                                                                // 15
                                                                                                                       //
                                                                                                                       //
var SYSTEM_LOG_DOWNLOAD_TOKENS = {};                                                                                   // 16
var SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION = 60000; // 60 seconds                                                 // 17
                                                                                                                       //
Meteor.methods({                                                                                                       // 19
  adminGetServerLogDownloadToken: function (setupToken) {                                                              // 20
    checkAuth(setupToken);                                                                                             // 21
    var token = Random.id(22);                                                                                         // 23
    SYSTEM_LOG_DOWNLOAD_TOKENS[token] = Date.now();                                                                    // 24
    setTimeout(function () {                                                                                           // 25
      if (SYSTEM_LOG_DOWNLOAD_TOKENS[token]) {                                                                         // 26
        delete SYSTEM_LOG_DOWNLOAD_TOKENS[token];                                                                      // 27
      }                                                                                                                // 28
    }, SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION * 2);                                                               // 29
    return token;                                                                                                      // 30
  }                                                                                                                    // 31
});                                                                                                                    // 19
Router.map(function () {                                                                                               // 34
  this.route("adminDownloadServerLog", {                                                                               // 35
    where: "server",                                                                                                   // 36
    path: "/admin/status/server-log/:tokenId",                                                                         // 37
    action: function () {                                                                                              // 38
      var token = this.params.tokenId;                                                                                 // 39
      var response = this.response;                                                                                    // 40
      var issueDate = SYSTEM_LOG_DOWNLOAD_TOKENS[token];                                                               // 41
      delete SYSTEM_LOG_DOWNLOAD_TOKENS[token]; // Clear the token from the in-memory map.                             // 42
                                                                                                                       //
      if (issueDate === undefined || issueDate + SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION < Date.now()) {           // 43
        // Require download to start within 60 seconds of the adminGetServerLogDownloadToken                           // 45
        // method being called.                                                                                        // 46
        response.writeHead(404, {                                                                                      // 47
          "Content-Type": "text/plain"                                                                                 // 47
        });                                                                                                            // 47
        return response.end("Invalid server log download token.");                                                     // 48
      }                                                                                                                // 49
                                                                                                                       //
      var logFilePath = SANDSTORM_LOGDIR + "/sandstorm.log";                                                           // 51
      var fd = Fs.openSync(logFilePath, "r");                                                                          // 52
      var stats = Fs.fstatSync(fd);                                                                                    // 53
      var initialSize = stats.size;                                                                                    // 54
      var readStream = Fs.createReadStream(undefined, {                                                                // 55
        fd: fd,                                                                                                        // 56
        start: 0,                                                                                                      // 57
        end: initialSize                                                                                               // 58
      });                                                                                                              // 55
      response.writeHead(200, {                                                                                        // 61
        "Content-Length": initialSize,                                                                                 // 62
        "Content-Type": "text/plain",                                                                                  // 63
        "content-Disposition": "attachment;filename=\"sandstorm.log\""                                                 // 64
      });                                                                                                              // 61
      readStream.pipe(response);                                                                                       // 66
    }                                                                                                                  // 67
  });                                                                                                                  // 35
});                                                                                                                    // 69
Meteor.publish("systemStatus", function () {                                                                           // 71
  // A pseudocollection containing the number of active grain sessions and unique                                      // 72
  // userIds associated with these sessions                                                                            // 73
  var db = this.connection.sandstormDb;                                                                                // 74
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 75
    throw new Meteor.Error(403, "User must be admin to view system status.");                                          // 76
  }                                                                                                                    // 77
                                                                                                                       //
  var _this = this;                                                                                                    // 79
                                                                                                                       //
  var userIdToSessionHashes = {}; // Maps userId => list of (sha256(sessionId))                                        // 80
                                                                                                                       //
  var grainIdToSessionHashes = {}; // Maps grainId => list of (sha256(sessionId))                                      // 81
                                                                                                                       //
  var userIdCount = 0;                                                                                                 // 83
  var grainIdCount = 0;                                                                                                // 84
  this.added("systemStatus", "globalStatus", {                                                                         // 86
    activeUsers: userIdCount,                                                                                          // 87
    activeGrains: grainIdCount                                                                                         // 88
  });                                                                                                                  // 86
  var query = db.collections.sessions.find();                                                                          // 91
  var handle = query.observe({                                                                                         // 92
    added: function (session) {                                                                                        // 93
      var hashedSessionId = hashSessionId(session._id);                                                                // 94
      var changed = {}; // Update userId cache.                                                                        // 95
                                                                                                                       //
      if (session.userId) {                                                                                            // 98
        if (userIdToSessionHashes[session.userId] === undefined) {                                                     // 99
          userIdToSessionHashes[session.userId] = [];                                                                  // 100
          userIdCount = userIdCount + 1;                                                                               // 101
          changed.activeUsers = userIdCount;                                                                           // 102
        }                                                                                                              // 103
                                                                                                                       //
        userIdToSessionHashes[session.userId] = _.union(userIdToSessionHashes[session.userId], [hashedSessionId]);     // 105
      } // Update grainId cache.                                                                                       // 107
                                                                                                                       //
                                                                                                                       //
      if (grainIdToSessionHashes[session.grainId] === undefined) {                                                     // 110
        grainIdToSessionHashes[session.grainId] = [];                                                                  // 111
        grainIdCount = grainIdCount + 1;                                                                               // 112
        changed.activeGrains = grainIdCount;                                                                           // 113
      }                                                                                                                // 114
                                                                                                                       //
      grainIdToSessionHashes[session.grainId] = _.union(grainIdToSessionHashes[session.grainId], [hashedSessionId]);   // 116
                                                                                                                       //
      if (!_.isEmpty(changed)) {                                                                                       // 119
        _this.changed("systemStatus", "globalStatus", changed);                                                        // 120
      }                                                                                                                // 121
    },                                                                                                                 // 122
    removed: function (session) {                                                                                      // 124
      var hashedSessionId = hashSessionId(session._id);                                                                // 125
      var changed = {};                                                                                                // 126
                                                                                                                       //
      if (session.userId) {                                                                                            // 128
        userIdToSessionHashes[session.userId] = _.without(userIdToSessionHashes[session.userId], hashedSessionId);     // 129
                                                                                                                       //
        if (userIdToSessionHashes[session.userId].length === 0) {                                                      // 132
          delete userIdToSessionHashes[session.userId];                                                                // 133
          userIdCount = userIdCount - 1;                                                                               // 134
          changed.activeUsers = userIdCount;                                                                           // 135
        }                                                                                                              // 136
      }                                                                                                                // 137
                                                                                                                       //
      grainIdToSessionHashes[session.grainId] = _.without(grainIdToSessionHashes[session.grainId], hashedSessionId);   // 139
                                                                                                                       //
      if (grainIdToSessionHashes[session.grainId].length === 0) {                                                      // 142
        delete grainIdToSessionHashes[session.grainId];                                                                // 143
        grainIdCount = grainIdCount - 1;                                                                               // 144
        changed.activeGrains = grainIdCount;                                                                           // 145
      }                                                                                                                // 146
                                                                                                                       //
      if (!_.isEmpty(changed)) {                                                                                       // 148
        _this.changed("systemStatus", "globalStatus", changed);                                                        // 149
      }                                                                                                                // 150
    }                                                                                                                  // 151
  });                                                                                                                  // 92
  this.onStop(function () {                                                                                            // 154
    handle.stop();                                                                                                     // 155
  });                                                                                                                  // 156
  this.ready();                                                                                                        // 157
});                                                                                                                    // 158
Meteor.publish("adminDemoUsers", function () {                                                                         // 160
  // Publishes expiry information about demo accounts to admins, so demo users can be counted                          // 161
  // clientside and reactively updated.                                                                                // 162
  var db = this.connection.sandstormDb;                                                                                // 163
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 164
    throw new Meteor.Error(403, "User must be admin to view system status.");                                          // 165
  }                                                                                                                    // 166
                                                                                                                       //
  return db.collections.users.find({                                                                                   // 168
    expires: {                                                                                                         // 169
      $gt: new Date()                                                                                                  // 170
    },                                                                                                                 // 169
    loginCredentials: {                                                                                                // 172
      $exists: true                                                                                                    // 173
    }                                                                                                                  // 172
  }, {                                                                                                                 // 168
    expires: 1,                                                                                                        // 176
    loginCredentials: 1                                                                                                // 177
  });                                                                                                                  // 175
});                                                                                                                    // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"drivers":{"external-ui-view.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/external-ui-view.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var PersistentImpl = void 0;                                                                                           // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var ssrfSafeLookup = void 0;                                                                                           // 1
module.watch(require("/imports/server/networking.js"), {                                                               // 1
  ssrfSafeLookup: function (v) {                                                                                       // 1
    ssrfSafeLookup = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var REQUEST_HEADER_WHITELIST = void 0,                                                                                 // 1
    RESPONSE_HEADER_WHITELIST = void 0;                                                                                // 1
module.watch(require("/imports/server/header-whitelist.js"), {                                                         // 1
  REQUEST_HEADER_WHITELIST: function (v) {                                                                             // 1
    REQUEST_HEADER_WHITELIST = v;                                                                                      // 1
  },                                                                                                                   // 1
  RESPONSE_HEADER_WHITELIST: function (v) {                                                                            // 1
    RESPONSE_HEADER_WHITELIST = v;                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 22
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 23
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 24
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 25
                                                                                                                       //
var Https = Npm.require("https");                                                                                      // 26
                                                                                                                       //
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;                                         // 27
var PersistentApiSession = Capnp.importSystem("sandstorm/api-session-impl.capnp").PersistentApiSession;                // 28
                                                                                                                       //
ExternalUiView = function () {                                                                                         // 31
  function ExternalUiView(url, token) {                                                                                // 32
    (0, _classCallCheck3.default)(this, ExternalUiView);                                                               // 32
    this.url = url;                                                                                                    // 33
    this.token = token;                                                                                                // 34
  }                                                                                                                    // 35
                                                                                                                       //
  ExternalUiView.prototype.newSession = function () {                                                                  // 31
    function newSession(userInfo, context, sessionType, sessionParams) {                                               // 31
      var _this2 = this;                                                                                               // 37
                                                                                                                       //
      if (sessionType !== ApiSession.typeId) {                                                                         // 38
        throw new Error("SessionType must be ApiSession.");                                                            // 39
      }                                                                                                                // 40
                                                                                                                       //
      var options = {};                                                                                                // 42
                                                                                                                       //
      if (this.token) {                                                                                                // 44
        options.headers = {                                                                                            // 45
          authorization: "Bearer " + this.token                                                                        // 46
        };                                                                                                             // 45
      }                                                                                                                // 48
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 50
        return {                                                                                                       // 51
          session: new Capnp.Capability(new ExternalWebSession(_this2.url, options, globalDb), ApiSession)             // 52
        };                                                                                                             // 51
      });                                                                                                              // 55
    }                                                                                                                  // 56
                                                                                                                       //
    return newSession;                                                                                                 // 31
  }();                                                                                                                 // 31
                                                                                                                       //
  return ExternalUiView;                                                                                               // 31
}();                                                                                                                   // 31
                                                                                                                       //
function getOAuthServiceInfo(url) {                                                                                    // 59
  // TODO(soon): Define a table somewhere (probably in a .capnp file) mapping API hosts to OAuth                       // 60
  //   metadata.                                                                                                       // 61
  if (url.startsWith("https://apidata.googleusercontent.com/") || url.startsWith("https://www.googleapis.com/")) {     // 62
    return {                                                                                                           // 64
      service: "google",                                                                                               // 65
      endpoint: "https://www.googleapis.com/oauth2/v4/token"                                                           // 66
    };                                                                                                                 // 64
  } else if (url.startsWith("https://api.github.com/users")) {                                                         // 68
    return {                                                                                                           // 69
      service: "github",                                                                                               // 70
      endpoint: "https://github.com/login/oauth/access_token"                                                          // 71
    };                                                                                                                 // 69
  } else {                                                                                                             // 73
    return null;                                                                                                       // 74
  }                                                                                                                    // 75
}                                                                                                                      // 76
                                                                                                                       //
function refreshOAuth(url, refreshToken) {                                                                             // 78
  // TODO(perf): Cache access tokens until they expire? Currently we re-do the refresh on every                        // 79
  //   restore. In particular, this means we always drop the first access token returned (which                        // 80
  //   is returned together with the refresh token) and then immediately request a new one.                            // 81
  var serviceInfo = getOAuthServiceInfo(url);                                                                          // 83
                                                                                                                       //
  if (!serviceInfo) {                                                                                                  // 84
    throw new Error("Don't know how to OAuth for: " + url);                                                            // 85
  }                                                                                                                    // 86
                                                                                                                       //
  var config = ServiceConfiguration.configurations.findOne({                                                           // 88
    service: serviceInfo.service                                                                                       // 88
  });                                                                                                                  // 88
                                                                                                                       //
  if (!config) {                                                                                                       // 89
    throw new Error("can't refresh OAuth token for service that isn't configured: " + serviceInfo.service);            // 90
  }                                                                                                                    // 92
                                                                                                                       //
  var response = HTTP.post(serviceInfo.endpoint, {                                                                     // 94
    headers: {                                                                                                         // 95
      "Content-Type": "application/x-www-form-urlencoded",                                                             // 96
      "Accept": "application/json"                                                                                     // 97
    },                                                                                                                 // 95
    content: "client_id=" + encodeURIComponent(config.clientId) + "&client_secret=" + encodeURIComponent(config.secret) + "&refresh_token=" + encodeURIComponent(refreshToken) + "&grant_type=refresh_token"
  });                                                                                                                  // 94
  return response.data;                                                                                                // 105
}                                                                                                                      // 106
                                                                                                                       //
function newExternalHttpSession(url, auth, db, saveTemplate) {                                                         // 108
  // `url` and `auth` are the corresponding members of `ApiToken.frontendRef.http`.                                    // 109
  var createCap = function (authorization) {                                                                           // 111
    return new Capnp.Capability(new ExternalWebSession(url, authorization ? {                                          // 112
      headers: {                                                                                                       // 113
        authorization: authorization                                                                                   // 113
      }                                                                                                                // 113
    } : {}, db, saveTemplate), PersistentApiSession);                                                                  // 113
  };                                                                                                                   // 115
                                                                                                                       //
  if (auth.refresh) {                                                                                                  // 117
    return createCap("Bearer " + refreshOAuth(url, auth.refresh).access_token);                                        // 118
  } else if (auth.bearer) {                                                                                            // 119
    return createCap("Bearer " + auth.bearer);                                                                         // 120
  } else if (auth.basic) {                                                                                             // 121
    var userpass = [auth.basic.username, auth.basic.password].join(":");                                               // 122
    return createCap("Basic " + new Buffer(userpass, "utf8").toString("base64"));                                      // 123
  } else {                                                                                                             // 124
    return createCap(null);                                                                                            // 125
  }                                                                                                                    // 126
}                                                                                                                      // 127
                                                                                                                       //
function registerHttpApiFrontendRef(registry) {                                                                        // 129
  registry.register({                                                                                                  // 130
    frontendRefField: "http",                                                                                          // 131
    typeId: ApiSession.typeId,                                                                                         // 132
    restore: function (db, saveTemplate, value) {                                                                      // 134
      return newExternalHttpSession(value.url, value.auth, db, saveTemplate);                                          // 135
    },                                                                                                                 // 136
    validate: function (db, session, request) {                                                                        // 138
      check(request, {                                                                                                 // 139
        url: String,                                                                                                   // 140
        auth: Match.OneOf({                                                                                            // 141
          none: null                                                                                                   // 142
        }, {                                                                                                           // 142
          bearer: String                                                                                               // 143
        }, {                                                                                                           // 143
          basic: {                                                                                                     // 144
            username: String,                                                                                          // 144
            password: String                                                                                           // 144
          }                                                                                                            // 144
        }, {                                                                                                           // 144
          oauth: {                                                                                                     // 145
            credentialToken: String,                                                                                   // 145
            credentialSecret: String                                                                                   // 145
          }                                                                                                            // 145
        })                                                                                                             // 145
      });                                                                                                              // 139
                                                                                                                       //
      if (!request.url.startsWith("https://") && !request.url.startsWith("http://")) {                                 // 148
        throw new Meteor.Error(400, "URL must be HTTP or HTTPS.");                                                     // 150
      } // Check for URL patterns.                                                                                     // 151
                                                                                                                       //
                                                                                                                       //
      var parsedUrl = Url.parse(request.url);                                                                          // 154
                                                                                                                       //
      if (parsedUrl.auth) {                                                                                            // 156
        var parts = parsedUrl.auth.split(":");                                                                         // 157
                                                                                                                       //
        if (parts.length === 2) {                                                                                      // 158
          if ("none" in request.auth) {                                                                                // 159
            request.auth = {                                                                                           // 160
              basic: {                                                                                                 // 160
                username: parts[0],                                                                                    // 160
                password: parts[1]                                                                                     // 160
              }                                                                                                        // 160
            };                                                                                                         // 160
          } else {                                                                                                     // 161
            throw new Meteor.Error(400, "Can't support multiple authentication mechanisms at once");                   // 162
          }                                                                                                            // 163
        }                                                                                                              // 164
                                                                                                                       //
        parsedUrl.auth = null;                                                                                         // 166
        request.url = Url.format(parsedUrl);                                                                           // 167
      }                                                                                                                // 168
                                                                                                                       //
      if (parsedUrl.hash) {                                                                                            // 170
        if ("none" in request.auth) {                                                                                  // 171
          request.auth = {                                                                                             // 172
            bearer: parsedUrl.hash.slice(1)                                                                            // 172
          };                                                                                                           // 172
        } else {                                                                                                       // 173
          throw new Meteor.Error(400, "Can't support multiple authentication mechanisms at once");                     // 174
        }                                                                                                              // 175
                                                                                                                       //
        parsedUrl.hash = null;                                                                                         // 177
        request.url = Url.format(parsedUrl);                                                                           // 178
      }                                                                                                                // 179
                                                                                                                       //
      if (request.auth.oauth) {                                                                                        // 181
        // We did an OAuth handshake client-side.                                                                      // 182
        var oauthResult = OAuth.retrieveCredential(request.auth.oauth.credentialToken, request.auth.oauth.credentialSecret);
                                                                                                                       //
        if (oauthResult instanceof Error) {                                                                            // 186
          throw oauthResult;                                                                                           // 187
        }                                                                                                              // 188
                                                                                                                       //
        var serviceData = oauthResult.serviceData;                                                                     // 190
                                                                                                                       //
        if (serviceData.refreshToken) {                                                                                // 191
          request.auth = {                                                                                             // 192
            refresh: serviceData.refreshToken                                                                          // 192
          };                                                                                                           // 192
        } else {                                                                                                       // 193
          request.auth = {                                                                                             // 194
            bearer: serviceData.accessToken                                                                            // 194
          };                                                                                                           // 194
        } // TODO(security): We could maybe add a MembraneRequirement that this user account                           // 195
        //   possesses credentials for this OAuth service. Conversely, perhaps if an authCode was                      // 198
        //   specified, we should automatically add the associated credential to the user's                            // 199
        //   account? (As a non-login credential.)                                                                     // 200
                                                                                                                       //
      }                                                                                                                // 201
                                                                                                                       //
      var descriptor = {                                                                                               // 203
        tags: [{                                                                                                       // 203
          id: ApiSession.typeId                                                                                        // 203
        }]                                                                                                             // 203
      };                                                                                                               // 203
      return {                                                                                                         // 204
        descriptor: descriptor,                                                                                        // 204
        requirements: [],                                                                                              // 204
        frontendRef: request                                                                                           // 204
      };                                                                                                               // 204
    },                                                                                                                 // 205
    query: function (db, userAccountId, tagValue) {                                                                    // 207
      var tag = tagValue ? Capnp.parse(ApiSession.PowerboxTag, tagValue) : {};                                         // 208
      var options = [];                                                                                                // 210
                                                                                                                       //
      if (tag.canonicalUrl && (tag.canonicalUrl.startsWith("https://") || tag.canonicalUrl.startsWith("http://"))) {   // 212
        var serviceInfo = getOAuthServiceInfo(tag.canonicalUrl);                                                       // 215
                                                                                                                       //
        if (serviceInfo && tag.oauthScopes) {                                                                          // 216
          // Note: We don't check if the service is configured, because it's useful to show the                        // 217
          //   user that the option exists but inform them that it will only work if the admin                         // 218
          //   configures this service.                                                                                // 219
          options.push({                                                                                               // 220
            _id: "http-oauth-" + tag.canonicalUrl,                                                                     // 221
            cardTemplate: "httpOAuthPowerboxCard",                                                                     // 222
            configureTemplate: "httpOAuthPowerboxConfiguration",                                                       // 223
            httpUrl: tag.canonicalUrl,                                                                                 // 224
            oauthServiceInfo: serviceInfo,                                                                             // 225
            oauthScopes: tag.oauthScopes                                                                               // 226
          });                                                                                                          // 220
        } else {                                                                                                       // 228
          // TODO(soon): Support tag.authentication.                                                                   // 229
          options.push({                                                                                               // 231
            _id: "http-url-" + tag.canonicalUrl,                                                                       // 232
            frontendRef: {                                                                                             // 233
              http: {                                                                                                  // 233
                url: tag.canonicalUrl,                                                                                 // 233
                auth: {                                                                                                // 233
                  none: null                                                                                           // 233
                }                                                                                                      // 233
              }                                                                                                        // 233
            },                                                                                                         // 233
            cardTemplate: "httpUrlPowerboxCard"                                                                        // 234
          });                                                                                                          // 231
        }                                                                                                              // 236
      } // Always offer the user the option to connect to an arbitrary URL of their choosing, even                     // 237
      // if canonicalUrl exists.                                                                                       // 240
                                                                                                                       //
                                                                                                                       //
      options.push({                                                                                                   // 241
        _id: "http-arbitrary",                                                                                         // 242
        cardTemplate: "httpArbitraryPowerboxCard",                                                                     // 243
        configureTemplate: "httpArbitraryPowerboxConfiguration"                                                        // 244
      });                                                                                                              // 241
      return options;                                                                                                  // 247
    }                                                                                                                  // 248
  });                                                                                                                  // 130
}                                                                                                                      // 250
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 252
  registerHttpApiFrontendRef(globalFrontendRefRegistry);                                                               // 252
}); // =======================================================================================                         // 252
                                                                                                                       //
var responseCodes = {                                                                                                  // 256
  200: {                                                                                                               // 257
    type: "content",                                                                                                   // 257
    code: "ok"                                                                                                         // 257
  },                                                                                                                   // 257
  201: {                                                                                                               // 258
    type: "content",                                                                                                   // 258
    code: "created"                                                                                                    // 258
  },                                                                                                                   // 258
  202: {                                                                                                               // 259
    type: "content",                                                                                                   // 259
    code: "accepted"                                                                                                   // 259
  },                                                                                                                   // 259
  204: {                                                                                                               // 260
    type: "noContent",                                                                                                 // 260
    shouldResetForm: false                                                                                             // 260
  },                                                                                                                   // 260
  205: {                                                                                                               // 261
    type: "noContent",                                                                                                 // 261
    shouldResetForm: true                                                                                              // 261
  },                                                                                                                   // 261
  // Unsupported until something demonstrates need.                                                                    // 263
  // 206: {type: 'noContent'},                                                                                         // 264
  // 300: {type: 'redirect'},                                                                                          // 265
  301: {                                                                                                               // 266
    type: "redirect",                                                                                                  // 266
    switchToGet: true,                                                                                                 // 266
    isPermanent: true                                                                                                  // 266
  },                                                                                                                   // 266
  302: {                                                                                                               // 267
    type: "redirect",                                                                                                  // 267
    switchToGet: true,                                                                                                 // 267
    isPermanent: false                                                                                                 // 267
  },                                                                                                                   // 267
  303: {                                                                                                               // 268
    type: "redirect",                                                                                                  // 268
    switchToGet: true,                                                                                                 // 268
    isPermanent: false                                                                                                 // 268
  },                                                                                                                   // 268
  304: {                                                                                                               // 270
    type: "preconditionFailed"                                                                                         // 270
  },                                                                                                                   // 270
  // Unsupported until something demonstrates need.                                                                    // 272
  // 305: {type: 'redirect'},                                                                                          // 273
  307: {                                                                                                               // 274
    type: "redirect",                                                                                                  // 274
    switchToGet: false,                                                                                                // 274
    isPermanent: false                                                                                                 // 274
  },                                                                                                                   // 274
  308: {                                                                                                               // 275
    type: "redirect",                                                                                                  // 275
    switchToGet: false,                                                                                                // 275
    isPermanent: true                                                                                                  // 275
  },                                                                                                                   // 275
  400: {                                                                                                               // 276
    type: "clientError",                                                                                               // 276
    clientErrorCode: "badRequest",                                                                                     // 276
    descriptionHtml: "Bad Request"                                                                                     // 276
  },                                                                                                                   // 276
  403: {                                                                                                               // 277
    type: "clientError",                                                                                               // 277
    clientErrorCode: "forbidden",                                                                                      // 277
    descriptionHtml: "Forbidden"                                                                                       // 277
  },                                                                                                                   // 277
  404: {                                                                                                               // 278
    type: "clientError",                                                                                               // 278
    clientErrorCode: "notFound",                                                                                       // 278
    descriptionHtml: "Not Found"                                                                                       // 278
  },                                                                                                                   // 278
  405: {                                                                                                               // 279
    type: "clientError",                                                                                               // 279
    clientErrorCode: "methodNotAllowed",                                                                               // 279
    descriptionHtml: "Method Not Allowed"                                                                              // 279
  },                                                                                                                   // 279
  406: {                                                                                                               // 280
    type: "clientError",                                                                                               // 280
    clientErrorCode: "notAcceptable",                                                                                  // 280
    descriptionHtml: "Not Acceptable"                                                                                  // 280
  },                                                                                                                   // 280
  409: {                                                                                                               // 281
    type: "clientError",                                                                                               // 281
    clientErrorCode: "conflict",                                                                                       // 281
    descriptionHtml: "Conflict"                                                                                        // 281
  },                                                                                                                   // 281
  410: {                                                                                                               // 282
    type: "clientError",                                                                                               // 282
    clientErrorCode: "gone",                                                                                           // 282
    descriptionHtml: "Gone"                                                                                            // 282
  },                                                                                                                   // 282
  412: {                                                                                                               // 283
    type: "preconditionFailed"                                                                                         // 283
  },                                                                                                                   // 283
  413: {                                                                                                               // 284
    type: "clientError",                                                                                               // 284
    clientErrorCode: "requestEntityTooLarge",                                                                          // 284
    descriptionHtml: "Request Entity Too Large"                                                                        // 284
  },                                                                                                                   // 284
  414: {                                                                                                               // 285
    type: "clientError",                                                                                               // 285
    clientErrorCode: "requestUriTooLong",                                                                              // 285
    descriptionHtml: "Request-URI Too Long"                                                                            // 285
  },                                                                                                                   // 285
  415: {                                                                                                               // 286
    type: "clientError",                                                                                               // 286
    clientErrorCode: "unsupportedMediaType",                                                                           // 286
    descriptionHtml: "Unsupported Media Type"                                                                          // 286
  },                                                                                                                   // 286
  418: {                                                                                                               // 287
    type: "clientError",                                                                                               // 287
    clientErrorCode: "imATeapot",                                                                                      // 287
    descriptionHtml: "I'm a teapot"                                                                                    // 287
  },                                                                                                                   // 287
  500: {                                                                                                               // 288
    type: "serverError"                                                                                                // 288
  },                                                                                                                   // 288
  501: {                                                                                                               // 289
    type: "serverError"                                                                                                // 289
  },                                                                                                                   // 289
  502: {                                                                                                               // 290
    type: "serverError"                                                                                                // 290
  },                                                                                                                   // 290
  503: {                                                                                                               // 291
    type: "serverError"                                                                                                // 291
  },                                                                                                                   // 291
  504: {                                                                                                               // 292
    type: "serverError"                                                                                                // 292
  },                                                                                                                   // 292
  505: {                                                                                                               // 293
    type: "serverError"                                                                                                // 293
  }                                                                                                                    // 293
};                                                                                                                     // 256
                                                                                                                       //
function composeETag(etag) {                                                                                           // 296
  if (etag.weak) {                                                                                                     // 297
    return "W/\"" + etag.value + "\"";                                                                                 // 298
  } else {                                                                                                             // 299
    return "\"" + etag.value + "\"";                                                                                   // 300
  }                                                                                                                    // 301
}                                                                                                                      // 302
                                                                                                                       //
function parseETag(input) {                                                                                            // 304
  var etag = {                                                                                                         // 305
    value: "",                                                                                                         // 305
    weak: false                                                                                                        // 305
  };                                                                                                                   // 305
  input = input.trim();                                                                                                // 307
                                                                                                                       //
  if (input.startsWith("W/")) {                                                                                        // 308
    input = input.slice(2);                                                                                            // 309
    etag.weak = true;                                                                                                  // 310
  }                                                                                                                    // 311
                                                                                                                       //
  if (!(input.startsWith("\"") && input.endsWith("\"") && input.length >= 2)) {                                        // 313
    // Invalid etag. Drop.                                                                                             // 314
    // (It would be nice to tell the developer about this but... how?)                                                 // 315
    return undefined;                                                                                                  // 316
  }                                                                                                                    // 317
                                                                                                                       //
  try {                                                                                                                // 319
    // Since the text starts with a quote, the only way it could parse as JSON is if it is a single                    // 320
    // string. This nicely handles escape sequences for us.                                                            // 321
    etag.value = JSON.parse(input);                                                                                    // 322
    return etag;                                                                                                       // 323
  } catch (err) {                                                                                                      // 324
    // Invalid etag.                                                                                                   // 325
    return undefined;                                                                                                  // 326
  }                                                                                                                    // 327
}                                                                                                                      // 328
                                                                                                                       //
ExternalWebSession = function (_PersistentImpl) {                                                                      // 330
  (0, _inherits3.default)(ExternalWebSession, _PersistentImpl);                                                        // 330
                                                                                                                       //
  function ExternalWebSession(url, options, db, saveTemplate) {                                                        // 331
    (0, _classCallCheck3.default)(this, ExternalWebSession);                                                           // 331
                                                                                                                       //
    // TODO(soon): Support HTTP proxy.                                                                                 // 334
    var _this3 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 331
                                                                                                                       //
    var safe = ssrfSafeLookup(db, url);                                                                                // 335
    if (!options) options = {};                                                                                        // 337
    if (!options.headers) options.headers = {};                                                                        // 338
    options.headers.host = safe.host;                                                                                  // 339
    options.servername = safe.host.split(":")[0];                                                                      // 340
                                                                                                                       //
    if (!saveTemplate) {                                                                                               // 342
      // enable backwards-compatibilty tweaks.                                                                         // 343
      _this3.fromHackSession = true;                                                                                   // 344
    }                                                                                                                  // 345
                                                                                                                       //
    var parsedUrl = Url.parse(safe.url);                                                                               // 347
    _this3.host = parsedUrl.hostname;                                                                                  // 348
                                                                                                                       //
    if (_this3.fromHackSession) {// HackSessionContext.getExternalUiView() apparently ignored any path on the URL. Whoops.
    } else {                                                                                                           // 351
      if (parsedUrl.path === "/") {                                                                                    // 352
        // The URL parser says path = "/" for both "http://foo" and "http://foo/". We want to be                       // 353
        // strict, though.                                                                                             // 354
        _this3.path = url.endsWith("/") ? "/" : "";                                                                    // 355
      } else {                                                                                                         // 356
        _this3.path = parsedUrl.path;                                                                                  // 357
      }                                                                                                                // 358
    }                                                                                                                  // 359
                                                                                                                       //
    _this3.port = parsedUrl.port;                                                                                      // 361
    _this3.protocol = parsedUrl.protocol;                                                                              // 362
    _this3.options = options;                                                                                          // 363
    return _this3;                                                                                                     // 331
  }                                                                                                                    // 364
                                                                                                                       //
  ExternalWebSession.prototype.get = function () {                                                                     // 330
    function get(path, context) {                                                                                      // 330
      return this._requestHelper("GET", path, context);                                                                // 367
    }                                                                                                                  // 368
                                                                                                                       //
    return get;                                                                                                        // 330
  }();                                                                                                                 // 330
                                                                                                                       //
  ExternalWebSession.prototype.post = function () {                                                                    // 330
    function post(path, content, context) {                                                                            // 330
      return this._requestHelper("POST", path, context, content.content, content.mimeType);                            // 371
    }                                                                                                                  // 372
                                                                                                                       //
    return post;                                                                                                       // 330
  }();                                                                                                                 // 330
                                                                                                                       //
  ExternalWebSession.prototype.put = function () {                                                                     // 330
    function put(path, content, context) {                                                                             // 330
      return this._requestHelper("PUT", path, context, content.content, content.mimeType);                             // 375
    }                                                                                                                  // 376
                                                                                                                       //
    return put;                                                                                                        // 330
  }();                                                                                                                 // 330
                                                                                                                       //
  ExternalWebSession.prototype.delete = function () {                                                                  // 330
    function _delete(path, context) {                                                                                  // 330
      return this._requestHelper("DELETE", path, context);                                                             // 379
    }                                                                                                                  // 380
                                                                                                                       //
    return _delete;                                                                                                    // 330
  }(); // TODO(someday): implement streaming and websockets for ExternalWebSession                                     // 330
  //postStreaming(path, mimeType, context) {                                                                           // 383
  //}                                                                                                                  // 384
  //putStreaming(path, mimeType, context) {                                                                            // 386
  //}                                                                                                                  // 387
  //openWebSocket(path, context, protocol, clientStream) {                                                             // 389
  //}                                                                                                                  // 390
                                                                                                                       //
                                                                                                                       //
  ExternalWebSession.prototype._requestHelper = function () {                                                          // 330
    function _requestHelper(method, path, context, content, contentType) {                                             // 330
      var _this4 = this;                                                                                               // 392
                                                                                                                       //
      var _this = this;                                                                                                // 393
                                                                                                                       //
      var session = _this;                                                                                             // 394
      return new Promise(function (resolve, reject) {                                                                  // 395
        var options = _.clone(session.options);                                                                        // 396
                                                                                                                       //
        options.headers = options.headers || {};                                                                       // 397
                                                                                                                       //
        if (!options.headers["user-agent"]) {                                                                          // 399
          options.headers["user-agent"] = "sandstorm app";                                                             // 400
        }                                                                                                              // 401
                                                                                                                       //
        if (_this4.fromHackSession) {                                                                                  // 403
          // According to the specification of `WebSession`, `path` should not contain a                               // 404
          // leading slash, and therefore we need to prepend "/". However, for a long time                             // 405
          // this implementation did not in fact prepend a "/". Since some apps might rely on                          // 406
          // that behavior, we only prepend "/" if the path does not start with "/".                                   // 407
          //                                                                                                           // 408
          // TODO(soon): Once apps have updated, prepend "/" unconditionally.                                          // 409
          options.path = path.startsWith("/") ? path : "/" + path;                                                     // 410
        } else {                                                                                                       // 411
          options.path = _this4.path + "/" + path;                                                                     // 412
        }                                                                                                              // 413
                                                                                                                       //
        options.method = method;                                                                                       // 415
                                                                                                                       //
        if (contentType) {                                                                                             // 416
          options.headers["content-type"] = contentType;                                                               // 417
        } // set accept header                                                                                         // 418
                                                                                                                       //
                                                                                                                       //
        if ("accept" in context) {                                                                                     // 421
          options.headers.accept = context.accept.map(function (acceptedType) {                                        // 422
            return acceptedType.mimeType + "; " + acceptedType.qValue;                                                 // 423
          }).join(", ");                                                                                               // 424
        } else if (!("accept" in options.headers)) {                                                                   // 425
          options.headers.accept = "*/*";                                                                              // 426
        } // set cookies                                                                                               // 427
                                                                                                                       //
                                                                                                                       //
        if (context.cookies && context.cookies.length > 0) {                                                           // 430
          options.headers.cookies = options.headers.cookies || "";                                                     // 431
          context.cookies.forEach(function (keyVal) {                                                                  // 432
            options.headers.cookies += keyVal.key + "=" + keyVal.val + ",";                                            // 433
          });                                                                                                          // 434
          options.headers.cookies = options.headers.cookies.slice(0, -1);                                              // 435
        } // set precondition                                                                                          // 436
                                                                                                                       //
                                                                                                                       //
        if (context.eTagPrecondition) {                                                                                // 439
          if (context.eTagPrecondition.none) {// nothing                                                               // 440
          } else if (context.eTagPrecondition.exists) {                                                                // 442
            options.headers["if-match"] = "*";                                                                         // 443
          } else if (context.eTagPrecondition.doesntExist) {                                                           // 444
            options.headers["if-none-match"] = "*";                                                                    // 445
          } else if (context.eTagPrecondition.matchesOneOf) {                                                          // 446
            options.headers["if-match"] = context.eTagPrecondition.matchesOneOf.map(composeETag).join(", ");           // 447
          } else if (context.eTagPrecondition.matchesNoneOf) {                                                         // 449
            options.headers["if-none-match"] = context.eTagPrecondition.matchesNoneOf.map(composeETag).join(", ");     // 450
          }                                                                                                            // 452
        } // set additional headers                                                                                    // 453
                                                                                                                       //
                                                                                                                       //
        (context.additionalHeaders || []).forEach(function (header) {                                                  // 456
          if (REQUEST_HEADER_WHITELIST.matches(header.name)) {                                                         // 457
            options.headers[header.name] = header.value;                                                               // 458
          }                                                                                                            // 459
        });                                                                                                            // 460
        options.host = session.host;                                                                                   // 462
        options.port = session.port;                                                                                   // 463
        var requestMethod = Http.request;                                                                              // 465
                                                                                                                       //
        if (session.protocol === "https:") {                                                                           // 466
          requestMethod = Https.request;                                                                               // 467
        }                                                                                                              // 468
                                                                                                                       //
        req = requestMethod(options, function (resp) {                                                                 // 470
          try {                                                                                                        // 471
            var buffers = [];                                                                                          // 472
            var statusInfo = responseCodes[resp.statusCode];                                                           // 473
            var rpcResponse = {};                                                                                      // 475
            rpcResponse.additionalHeaders = [];                                                                        // 477
                                                                                                                       //
            for (var headerName in meteorBabelHelpers.sanitizeForInObject(resp.headers)) {                             // 478
              if (RESPONSE_HEADER_WHITELIST.matches(headerName)) {                                                     // 479
                rpcResponse.additionalHeaders.push({                                                                   // 480
                  name: headerName,                                                                                    // 481
                  value: resp.headers[headerName]                                                                      // 482
                });                                                                                                    // 480
              }                                                                                                        // 484
            }                                                                                                          // 485
                                                                                                                       //
            resp.on("data", function (buf) {                                                                           // 487
              buffers.push(buf);                                                                                       // 488
            });                                                                                                        // 489
            resp.on("end", function () {                                                                               // 491
              try {                                                                                                    // 492
                var fillInErrorBody = function (error) {                                                               // 492
                  var contentType = resp.headers["content-type"];                                                      // 496
                                                                                                                       //
                  if (contentType && (contentType == "text/html" || contentType.startsWith("text/html;"))) {           // 497
                    // TODO(someday): Check for non-UTF-8 charset and translate?                                       // 499
                    error.descriptionHtml = data.toString("utf8");                                                     // 500
                  } else if (contentType || data.length > 0) {                                                         // 501
                    var _content = {                                                                                   // 502
                      data: data                                                                                       // 502
                    };                                                                                                 // 502
                    if (contentType) _content.mimeType = contentType;                                                  // 503
                    if ("content-encoding" in resp.headers) _content.encoding = resp.headers["content-encoding"];      // 504
                    if ("content-language" in resp.headers) _content.language = resp.headers["content-language"];      // 505
                    error.nonHtmlBody = _content;                                                                      // 506
                  }                                                                                                    // 507
                };                                                                                                     // 508
                                                                                                                       //
                var data = Buffer.concat(buffers);                                                                     // 493
                                                                                                                       //
                switch (statusInfo ? statusInfo.type : resp.statusCode) {                                              // 510
                  case "content":                                                                                      // 511
                    var _content2 = {};                                                                                // 512
                    rpcResponse.content = _content2;                                                                   // 513
                    _content2.statusCode = statusInfo.code;                                                            // 515
                    if ("content-encoding" in resp.headers) _content2.encoding = resp.headers["content-encoding"];     // 516
                    if ("content-language" in resp.headers) _content2.language = resp.headers["content-language"];     // 517
                    if ("content-type" in resp.headers) _content2.mimeType = resp.headers["content-type"];             // 518
                                                                                                                       //
                    if ("content-disposition" in resp.headers) {                                                       // 519
                      var disposition = resp.headers["content-disposition"];                                           // 520
                      var parts = disposition.split(";");                                                              // 521
                                                                                                                       //
                      if (parts[0].toLowerCase().trim() === "attachment") {                                            // 522
                        parts.forEach(function (part) {                                                                // 523
                          var splitPart = part.split("=");                                                             // 524
                                                                                                                       //
                          if (splitPart[0].toLowerCase().trim() === "filename") {                                      // 525
                            _content2.disposition = {                                                                  // 526
                              download: splitPart[1].trim()                                                            // 526
                            };                                                                                         // 526
                          }                                                                                            // 527
                        });                                                                                            // 528
                      }                                                                                                // 529
                    }                                                                                                  // 530
                                                                                                                       //
                    if (resp.headers.etag) {                                                                           // 532
                      _content2.eTag = parseETag(resp.headers.etag);                                                   // 533
                    }                                                                                                  // 534
                                                                                                                       //
                    _content2.body = {};                                                                               // 536
                    _content2.body.bytes = data;                                                                       // 537
                    resolve(rpcResponse);                                                                              // 539
                    break;                                                                                             // 540
                                                                                                                       //
                  case "noContent":                                                                                    // 541
                    var noContent = {};                                                                                // 542
                    rpcResponse.noContent = noContent;                                                                 // 543
                    noContent.setShouldResetForm = statusInfo.shouldResetForm;                                         // 544
                                                                                                                       //
                    if (resp.headers.etag) {                                                                           // 545
                      noContent.eTag = parseETag(resp.headers.etag);                                                   // 546
                    }                                                                                                  // 547
                                                                                                                       //
                    resolve(rpcResponse);                                                                              // 549
                    break;                                                                                             // 550
                                                                                                                       //
                  case "redirect":                                                                                     // 551
                    var redirect = {};                                                                                 // 552
                    rpcResponse.redirect = redirect;                                                                   // 553
                    redirect.isPermanent = statusInfo.isPermanent;                                                     // 554
                    redirect.switchToGet = statusInfo.switchToGet;                                                     // 555
                    if ("location" in resp.headers) redirect.location = resp.headers.location;                         // 556
                    resolve(rpcResponse);                                                                              // 557
                    break;                                                                                             // 558
                                                                                                                       //
                  case "clientError":                                                                                  // 559
                    var clientError = {};                                                                              // 560
                    rpcResponse.clientError = clientError;                                                             // 561
                    clientError.statusCode = statusInfo.clientErrorCode;                                               // 562
                    fillInErrorBody(clientError);                                                                      // 564
                    resolve(rpcResponse);                                                                              // 565
                    break;                                                                                             // 566
                                                                                                                       //
                  case "serverError":                                                                                  // 567
                    var serverError = {};                                                                              // 568
                    rpcResponse.serverError = serverError;                                                             // 569
                    fillInErrorBody(serverError);                                                                      // 570
                    resolve(rpcResponse);                                                                              // 571
                    break;                                                                                             // 572
                                                                                                                       //
                  case "preconditionFailed":                                                                           // 573
                    var preconditionFailed = {};                                                                       // 574
                    rpcResponse.preconditionFailed = preconditionFailed;                                               // 575
                                                                                                                       //
                    if (resp.headers.etag) {                                                                           // 576
                      preconditionFailed.matchingETag = parseETag(resp.headers.etag);                                  // 577
                    }                                                                                                  // 578
                                                                                                                       //
                    resolve(rpcResponse);                                                                              // 580
                    break;                                                                                             // 581
                  // TODO(soon): Handle token-expired errors by throwing DISCONNECTED -- this will                     // 583
                  //   force the client to reload the capability which will refresh the token.                         // 584
                                                                                                                       //
                  default:                                                                                             // 586
                    // ???                                                                                             // 586
                    var _err = new Error("Invalid status code " + resp.statusCode + " received in response.");         // 587
                                                                                                                       //
                    reject(_err);                                                                                      // 589
                    break;                                                                                             // 590
                }                                                                                                      // 510
              } catch (err) {                                                                                          // 592
                reject(err);                                                                                           // 593
              }                                                                                                        // 594
            });                                                                                                        // 595
          } catch (err) {                                                                                              // 596
            reject(err);                                                                                               // 597
          }                                                                                                            // 598
        });                                                                                                            // 599
        req.on("error", function (e) {                                                                                 // 601
          reject(e);                                                                                                   // 602
        });                                                                                                            // 603
        req.setTimeout(15000, function () {                                                                            // 605
          req.abort();                                                                                                 // 606
          err = new Error("Request timed out.");                                                                       // 607
          err.kjType = "overloaded";                                                                                   // 608
          reject(err);                                                                                                 // 609
        });                                                                                                            // 610
                                                                                                                       //
        if (content) {                                                                                                 // 612
          req.end(content);                                                                                            // 613
        } else {                                                                                                       // 614
          req.end();                                                                                                   // 615
        }                                                                                                              // 616
      });                                                                                                              // 617
    }                                                                                                                  // 618
                                                                                                                       //
    return _requestHelper;                                                                                             // 330
  }();                                                                                                                 // 330
                                                                                                                       //
  return ExternalWebSession;                                                                                           // 330
}(PersistentImpl);                                                                                                     // 330
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"ip.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/ip.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Bignum = void 0;                                                                                                   // 1
module.watch(require("bignum"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    Bignum = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 19
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 20
                                                                                                                       //
var Tls = Npm.require("tls");                                                                                          // 21
                                                                                                                       //
var Dgram = Npm.require("dgram");                                                                                      // 22
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 23
                                                                                                                       //
var IpRpc = Capnp.importSystem("sandstorm/ip.capnp");                                                                  // 25
                                                                                                                       //
ByteStreamConnection = function () {                                                                                   // 27
  function ByteStreamConnection(connection) {                                                                          // 28
    (0, _classCallCheck3.default)(this, ByteStreamConnection);                                                         // 28
    this.connection = connection;                                                                                      // 29
  }                                                                                                                    // 30
                                                                                                                       //
  ByteStreamConnection.prototype.done = function () {                                                                  // 27
    function done() {                                                                                                  // 27
      this.connection.end();                                                                                           // 33
    }                                                                                                                  // 34
                                                                                                                       //
    return done;                                                                                                       // 27
  }();                                                                                                                 // 27
                                                                                                                       //
  ByteStreamConnection.prototype.write = function () {                                                                 // 27
    function write(data) {                                                                                             // 27
      this.connection.write(data);                                                                                     // 37
    }                                                                                                                  // 38
                                                                                                                       //
    return write;                                                                                                      // 27
  }(); // expectSize not implemented                                                                                   // 27
  // expectSize(size) { }                                                                                              // 41
                                                                                                                       //
                                                                                                                       //
  return ByteStreamConnection;                                                                                         // 27
}();                                                                                                                   // 27
                                                                                                                       //
var IpInterfaceImpl = function (_PersistentImpl) {                                                                     //
  (0, _inherits3.default)(IpInterfaceImpl, _PersistentImpl);                                                           //
                                                                                                                       //
  function IpInterfaceImpl(db, saveTemplate) {                                                                         // 45
    (0, _classCallCheck3.default)(this, IpInterfaceImpl);                                                              // 45
    return (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));               // 45
  }                                                                                                                    // 47
                                                                                                                       //
  IpInterfaceImpl.prototype.listenTcp = function () {                                                                  //
    function listenTcp(portNum, port) {                                                                                //
      return new Promise(function (resolve, reject) {                                                                  // 50
        var resolved = false;                                                                                          // 51
        var server = Net.createServer(function (connection) {                                                          // 52
          var wrappedConnection = new ByteStreamConnection(connection);                                                // 53
          var upstream = port.connect(wrappedConnection).upstream;                                                     // 54
          connection.on("data", function (data) {                                                                      // 56
            upstream.write(data);                                                                                      // 57
          });                                                                                                          // 58
          connection.on("close", function (hadError) {                                                                 // 60
            upstream.done();                                                                                           // 61
          });                                                                                                          // 62
          connection.on("error", function (err) {                                                                      // 64
            if (resolved) {                                                                                            // 65
              connection.write = errorWrite;                                                                           // 66
            } else {                                                                                                   // 67
              // upstream hasn't been resolved yet, so it's safe to reject                                             // 68
              reject(err);                                                                                             // 69
            }                                                                                                          // 70
          });                                                                                                          // 71
        });                                                                                                            // 72
        server.listen(portNum, function () {                                                                           // 74
          resolved = true;                                                                                             // 75
          resolve({                                                                                                    // 76
            handle: server                                                                                             // 76
          }); // server has a close method which is all we want from a handle                                          // 76
        });                                                                                                            // 77
        server.on("error", function (err) {                                                                            // 79
          if (!resolved) {                                                                                             // 80
            reject(err);                                                                                               // 81
          }                                                                                                            // 82
        });                                                                                                            // 83
      });                                                                                                              // 84
    }                                                                                                                  // 85
                                                                                                                       //
    return listenTcp;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  IpInterfaceImpl.prototype.listenUdp = function () {                                                                  //
    function listenUdp(portNum, port) {                                                                                //
      return new Promise(function (resolve, reject) {                                                                  // 88
        var portMap = {};                                                                                              // 89
        var resolved = false;                                                                                          // 90
        var server = Dgram.createSocket("udp4"); // TODO(someday): handle ipv6 sockets too                             // 91
                                                                                                                       //
        server.bind(portNum);                                                                                          // 92
        server.on("listening", function () {                                                                           // 94
          // Although UDP is connectionless, we don't resolve until here so that we can handle bind                    // 95
          // errors such as invalid host                                                                               // 96
          resolved = true;                                                                                             // 97
          resolve({                                                                                                    // 98
            handle: server                                                                                             // 98
          }); // server has a close method which is all we want from a handle                                          // 98
        });                                                                                                            // 99
        server.on("error", function (err) {                                                                            // 101
          // TODO(someday): do something about errors after the promise is resolved                                    // 102
          if (!resolved) {                                                                                             // 103
            reject(err);                                                                                               // 104
          } else {                                                                                                     // 105
            console.error("error in listenUdp: " + err);                                                               // 106
          }                                                                                                            // 107
        });                                                                                                            // 108
        var returnMap = {};                                                                                            // 110
        server.on("message", function (msg, rinfo) {                                                                   // 111
          var address = rinfo.address + "]:" + rinfo.port;                                                             // 112
          var returnPort = returnMap[address];                                                                         // 113
                                                                                                                       //
          if (!returnPort) {                                                                                           // 115
            returnMap[address] = returnPort = new BoundUdpPortImpl(server, rinfo.address, rinfo.port);                 // 116
          }                                                                                                            // 117
                                                                                                                       //
          port.send(msg, returnPort);                                                                                  // 119
        });                                                                                                            // 120
      });                                                                                                              // 121
    }                                                                                                                  // 122
                                                                                                                       //
    return listenUdp;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  return IpInterfaceImpl;                                                                                              //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 123
//   (ugh).                                                                                                            // 126
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 127
  globalFrontendRefRegistry.register({                                                                                 // 128
    frontendRefField: "ipInterface",                                                                                   // 129
    typeId: IpRpc.IpInterface.typeId,                                                                                  // 130
    restore: function (db, saveTemplate) {                                                                             // 132
      return new Capnp.Capability(new IpInterfaceImpl(db, saveTemplate), IpRpc.PersistentIpInterface);                 // 133
    },                                                                                                                 // 135
    validate: function (db, session, value) {                                                                          // 137
      check(value, true);                                                                                              // 138
                                                                                                                       //
      if (!session.userId) {                                                                                           // 140
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 141
      }                                                                                                                // 142
                                                                                                                       //
      return {                                                                                                         // 144
        descriptor: {                                                                                                  // 145
          tags: [{                                                                                                     // 145
            id: IpRpc.IpInterface.typeId                                                                               // 145
          }]                                                                                                           // 145
        },                                                                                                             // 145
        requirements: [{                                                                                               // 146
          userIsAdmin: session.userId                                                                                  // 146
        }],                                                                                                            // 146
        frontendRef: value                                                                                             // 147
      };                                                                                                               // 144
    },                                                                                                                 // 149
    query: function (db, userId, value) {                                                                              // 151
      if (userId && Meteor.users.findOne(userId).isAdmin) {                                                            // 152
        return [{                                                                                                      // 153
          _id: "frontendref-ipinterface",                                                                              // 155
          frontendRef: {                                                                                               // 156
            ipInterface: true                                                                                          // 156
          },                                                                                                           // 156
          cardTemplate: "ipInterfacePowerboxCard"                                                                      // 157
        }];                                                                                                            // 154
      } else {                                                                                                         // 160
        return [];                                                                                                     // 161
      }                                                                                                                // 162
    }                                                                                                                  // 163
  });                                                                                                                  // 128
});                                                                                                                    // 165
                                                                                                                       //
BoundUdpPortImpl = function () {                                                                                       // 167
  function BoundUdpPortImpl(server, address, port) {                                                                   // 168
    (0, _classCallCheck3.default)(this, BoundUdpPortImpl);                                                             // 168
    this.server = server;                                                                                              // 169
    this.address = address;                                                                                            // 170
    this.port = port;                                                                                                  // 171
  }                                                                                                                    // 172
                                                                                                                       //
  BoundUdpPortImpl.prototype.send = function () {                                                                      // 167
    function send(message, returnPort) {                                                                               // 167
      // TODO(someday): this whole class is a hack to deal with the fact that we can't compare                         // 175
      // capabilities or build a map with them. What we should be doing is mapping all ports to                        // 176
      // their raw physical address/port, and using that here                                                          // 177
      this.server.send(message, 0, message.length, this.port, this.address);                                           // 178
    }                                                                                                                  // 179
                                                                                                                       //
    return send;                                                                                                       // 167
  }();                                                                                                                 // 167
                                                                                                                       //
  return BoundUdpPortImpl;                                                                                             // 167
}();                                                                                                                   // 167
                                                                                                                       //
var bits16 = Bignum(1).shiftLeft(16).sub(1);                                                                           // 182
var bits32 = Bignum(1).shiftLeft(32).sub(1);                                                                           // 183
                                                                                                                       //
var intToIpv4 = function (num) {                                                                                       // 185
  var part1 = num & 255;                                                                                               // 186
  var part2 = num >> 8 & 255;                                                                                          // 187
  var part3 = num >> 16 & 255;                                                                                         // 188
  var part4 = num >> 24 & 255;                                                                                         // 189
  return part4 + "." + part3 + "." + part2 + "." + part1;                                                              // 191
};                                                                                                                     // 192
                                                                                                                       //
var addressToString = function (address) {                                                                             // 194
  var ipv6num = Bignum(address.upper64).shiftLeft(64).add(Bignum(address.lower64));                                    // 195
                                                                                                                       //
  if (ipv6num.shiftRight(32).eq(bits16)) {                                                                             // 197
    // this is an ipv4 address, we should return it as such                                                            // 198
    var ipv4num = ipv6num.and(bits32).toNumber();                                                                      // 199
    return intToIpv4(ipv4num);                                                                                         // 200
  }                                                                                                                    // 201
                                                                                                                       //
  var hex = ipv6num.toString(16);                                                                                      // 203
  var numColons = 0;                                                                                                   // 204
  var out = "";                                                                                                        // 205
                                                                                                                       //
  for (var i = 0; i < hex.length; ++i) {                                                                               // 207
    // start with lower bits of address and build the output in reverse                                                // 208
    // this ensures that we can place a colon every 4 characters                                                       // 209
    out += hex[hex.length - 1 - i];                                                                                    // 210
                                                                                                                       //
    if ((i + 1) % 4 === 0) {                                                                                           // 211
      out += ":";                                                                                                      // 212
      ++numColons;                                                                                                     // 213
    }                                                                                                                  // 214
  } // Double colon represents all bits being 0                                                                        // 215
                                                                                                                       //
                                                                                                                       //
  if (numColons < 7) {                                                                                                 // 218
    out += "::";                                                                                                       // 219
  }                                                                                                                    // 220
                                                                                                                       //
  return out.split("").reverse().join("");                                                                             // 222
};                                                                                                                     // 223
                                                                                                                       //
var addressType = function (address) {                                                                                 // 225
  var type = "udp4"; // Check if it's an ipv6 address                                                                  // 226
  // TODO(someday): make this less hacky and change address to explicitly pass this information                        // 229
                                                                                                                       //
  if (address.indexOf(":") != -1) {                                                                                    // 230
    type = "udp6";                                                                                                     // 231
  }                                                                                                                    // 232
                                                                                                                       //
  return type;                                                                                                         // 234
};                                                                                                                     // 235
                                                                                                                       //
var IpNetworkImpl = function (_PersistentImpl2) {                                                                      //
  (0, _inherits3.default)(IpNetworkImpl, _PersistentImpl2);                                                            //
                                                                                                                       //
  function IpNetworkImpl(db, saveTemplate, tls) {                                                                      // 238
    (0, _classCallCheck3.default)(this, IpNetworkImpl);                                                                // 238
                                                                                                                       //
    var _this3 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));        // 238
                                                                                                                       //
    _this3.tls = tls;                                                                                                  // 240
    return _this3;                                                                                                     // 238
  }                                                                                                                    // 241
                                                                                                                       //
  IpNetworkImpl.prototype.getRemoteHost = function () {                                                                //
    function getRemoteHost(address) {                                                                                  //
      // Note that TLS typically authenticates hostnames, not raw IP addresses. So if `this.tls`                       // 244
      // is true, then the returned `IpRemoteHost` will likely refuse to make any connections.                         // 245
      // However, we keep that code path active because there is in principle no reason why it                         // 246
      // should always fail, and we wish to allow for a possible future in which we let users                          // 247
      // specify custom certificate authorities, in which case it might be more likely for TLS                         // 248
      // to be expected to authenticate raw IP addresses.                                                              // 249
      return {                                                                                                         // 251
        host: new IpRemoteHostImpl(addressToString(address), this.tls)                                                 // 251
      };                                                                                                               // 251
    }                                                                                                                  // 252
                                                                                                                       //
    return getRemoteHost;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  IpNetworkImpl.prototype.getRemoteHostByName = function () {                                                          //
    function getRemoteHostByName(address) {                                                                            //
      return {                                                                                                         // 255
        host: new IpRemoteHostImpl(address, this.tls)                                                                  // 255
      };                                                                                                               // 255
    }                                                                                                                  // 256
                                                                                                                       //
    return getRemoteHostByName;                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  return IpNetworkImpl;                                                                                                //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 257
//   (ugh).                                                                                                            // 260
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 261
  globalFrontendRefRegistry.register({                                                                                 // 262
    frontendRefField: "ipNetwork",                                                                                     // 263
    typeId: IpRpc.IpNetwork.typeId,                                                                                    // 264
    restore: function (db, saveTemplate, value) {                                                                      // 266
      return new Capnp.Capability(new IpNetworkImpl(db, saveTemplate, "tls" in value.encryption), IpRpc.PersistentIpNetwork);
    },                                                                                                                 // 270
    validate: function (db, session, value) {                                                                          // 272
      check(value, {                                                                                                   // 273
        encryption: Match.OneOf({                                                                                      // 273
          none: null                                                                                                   // 273
        }, {                                                                                                           // 273
          tls: null                                                                                                    // 273
        })                                                                                                             // 273
      });                                                                                                              // 273
                                                                                                                       //
      if (!session.userId) {                                                                                           // 275
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 276
      }                                                                                                                // 277
                                                                                                                       //
      return {                                                                                                         // 279
        descriptor: {                                                                                                  // 280
          tags: [{                                                                                                     // 281
            id: IpRpc.IpNetwork.typeId,                                                                                // 283
            value: Capnp.serialize(IpRpc.IpNetwork.PowerboxTag, value)                                                 // 284
          }]                                                                                                           // 282
        },                                                                                                             // 280
        requirements: [{                                                                                               // 290
          userIsAdmin: session.userId                                                                                  // 290
        }],                                                                                                            // 290
        frontendRef: value                                                                                             // 291
      };                                                                                                               // 279
    },                                                                                                                 // 293
    query: function (db, userId, value) {                                                                              // 295
      var encryption = {                                                                                               // 296
        none: null                                                                                                     // 296
      };                                                                                                               // 296
                                                                                                                       //
      if (value) {                                                                                                     // 297
        encryption = Capnp.parse(IpRpc.IpNetwork.PowerboxTag, value).encryption || encryption;                         // 298
      }                                                                                                                // 299
                                                                                                                       //
      if (userId && Meteor.users.findOne(userId).isAdmin) {                                                            // 301
        return [{                                                                                                      // 302
          _id: "frontendref-ipnetwork",                                                                                // 304
          frontendRef: {                                                                                               // 305
            ipNetwork: {                                                                                               // 305
              encryption: encryption                                                                                   // 305
            }                                                                                                          // 305
          },                                                                                                           // 305
          cardTemplate: "ipNetworkPowerboxCard"                                                                        // 306
        }];                                                                                                            // 303
      } else {                                                                                                         // 309
        return [];                                                                                                     // 310
      }                                                                                                                // 311
    }                                                                                                                  // 312
  });                                                                                                                  // 262
});                                                                                                                    // 314
                                                                                                                       //
var IpRemoteHostImpl = function () {                                                                                   //
  function IpRemoteHostImpl(address, tls) {                                                                            // 317
    (0, _classCallCheck3.default)(this, IpRemoteHostImpl);                                                             // 317
    this.address = address;                                                                                            // 318
    this.tls = tls;                                                                                                    // 319
  }                                                                                                                    // 320
                                                                                                                       //
  IpRemoteHostImpl.prototype.getTcpPort = function () {                                                                //
    function getTcpPort(portNum) {                                                                                     //
      return {                                                                                                         // 323
        port: new TcpPortImpl(this.address, portNum, this.tls)                                                         // 323
      };                                                                                                               // 323
    }                                                                                                                  // 324
                                                                                                                       //
    return getTcpPort;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  IpRemoteHostImpl.prototype.getUdpPort = function () {                                                                //
    function getUdpPort(portNum) {                                                                                     //
      if (this.tls) {                                                                                                  // 327
        var error = new Error("Datagram Transport Layer Security is not yet supported");                               // 328
        error.kjType = "unimplemented";                                                                                // 329
        throw error;                                                                                                   // 330
      }                                                                                                                // 331
                                                                                                                       //
      return {                                                                                                         // 333
        port: new UdpPortImpl(this.address, portNum)                                                                   // 333
      };                                                                                                               // 333
    }                                                                                                                  // 334
                                                                                                                       //
    return getUdpPort;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  return IpRemoteHostImpl;                                                                                             //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 335
                                                                                                                       //
TcpPortImpl = function () {                                                                                            // 337
  function TcpPortImpl(address, portNum, tls) {                                                                        // 338
    (0, _classCallCheck3.default)(this, TcpPortImpl);                                                                  // 338
    this.address = address;                                                                                            // 339
    this.port = portNum;                                                                                               // 340
    this.tls = tls;                                                                                                    // 341
  }                                                                                                                    // 342
                                                                                                                       //
  TcpPortImpl.prototype.connect = function () {                                                                        // 337
    function connect(downstream) {                                                                                     // 337
      var _this = this;                                                                                                // 345
                                                                                                                       //
      var resolved = false;                                                                                            // 346
      var connectMethod = Net.connect;                                                                                 // 347
                                                                                                                       //
      if (this.tls) {                                                                                                  // 348
        connectMethod = Tls.connect;                                                                                   // 349
      }                                                                                                                // 350
                                                                                                                       //
      return new Promise(function (resolve, reject) {                                                                  // 352
        var client = connectMethod({                                                                                   // 353
          host: _this.address,                                                                                         // 353
          port: _this.port                                                                                             // 353
        }, function () {                                                                                               // 353
          resolved = true;                                                                                             // 354
          resolve({                                                                                                    // 355
            upstream: new ByteStreamConnection(client)                                                                 // 355
          });                                                                                                          // 355
        });                                                                                                            // 356
        client.on("data", function (data) {                                                                            // 358
          downstream.write(data);                                                                                      // 359
        });                                                                                                            // 360
        client.on("close", function (hadError) {                                                                       // 362
          downstream.done();                                                                                           // 363
        });                                                                                                            // 364
        client.on("error", function (err) {                                                                            // 366
          if (resolved) {                                                                                              // 367
            client.write = errorWrite;                                                                                 // 368
          } else {                                                                                                     // 369
            // upstream hasn't been resolved yet, so it's safe to reject                                               // 370
            reject(err);                                                                                               // 371
          }                                                                                                            // 372
        });                                                                                                            // 373
      });                                                                                                              // 374
    }                                                                                                                  // 375
                                                                                                                       //
    return connect;                                                                                                    // 337
  }();                                                                                                                 // 337
                                                                                                                       //
  return TcpPortImpl;                                                                                                  // 337
}();                                                                                                                   // 337
                                                                                                                       //
var errorWrite = function (data) {                                                                                     // 378
  throw new Error("error occurred in connection");                                                                     // 379
};                                                                                                                     // 380
                                                                                                                       //
UdpPortImpl = function () {                                                                                            // 382
  function UdpPortImpl(address, portNum) {                                                                             // 383
    (0, _classCallCheck3.default)(this, UdpPortImpl);                                                                  // 383
    this.address = address;                                                                                            // 384
    this.port = portNum;                                                                                               // 385
    var type = addressType(address);                                                                                   // 387
    this.socket = Dgram.createSocket(type); // TODO(someday): close socket after a certain time of inactivity?         // 388
    // This may be pointless since grains are killed frequently when not in use anyways                                // 391
    // Temporary hack. We only expect clients to pass in a single return port, so we'll store it                       // 393
    // and only send replies here.                                                                                     // 394
    // This will be changed to be correct when equality comparisons are added to capabilities.                         // 395
                                                                                                                       //
    this.returnPort = null;                                                                                            // 396
                                                                                                                       //
    var _this = this;                                                                                                  // 398
                                                                                                                       //
    this.socket.on("message", function (msg, rinfo) {                                                                  // 399
      if (_this.returnPort) {                                                                                          // 400
        _this.returnPort.send(msg, _this);                                                                             // 401
      }                                                                                                                // 402
    });                                                                                                                // 403
  }                                                                                                                    // 404
                                                                                                                       //
  UdpPortImpl.prototype.send = function () {                                                                           // 382
    function send(message, returnPort) {                                                                               // 382
      this.returnPort = returnPort;                                                                                    // 407
      this.socket.send(message, 0, message.length, this.port, this.address); // TODO(someday): use callback to catch errors and do something with them
    }                                                                                                                  // 411
                                                                                                                       //
    return send;                                                                                                       // 382
  }();                                                                                                                 // 382
                                                                                                                       //
  return UdpPortImpl;                                                                                                  // 382
}();                                                                                                                   // 382
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"mail.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/mail.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _this = this;                                                                                                      //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var SMTPServer = void 0;                                                                                               // 1
module.watch(require("smtp-server"), {                                                                                 // 1
  SMTPServer: function (v) {                                                                                           // 1
    SMTPServer = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var MailParser = void 0;                                                                                               // 1
module.watch(require("mailparser"), {                                                                                  // 1
  MailParser: function (v) {                                                                                           // 1
    MailParser = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var rawSend = void 0;                                                                                                  // 1
module.watch(require("/imports/server/email.js"), {                                                                    // 1
  rawSend: function (v) {                                                                                              // 1
    rawSend = v;                                                                                                       // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var shouldRestartGrain = void 0;                                                                                       // 1
module.watch(require("/imports/server/backend.js"), {                                                                  // 1
  shouldRestartGrain: function (v) {                                                                                   // 1
    shouldRestartGrain = v;                                                                                            // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var inMeteor = void 0;                                                                                                 // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 25
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 26
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 27
                                                                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");                                                            // 29
var EmailImpl = Capnp.importSystem("sandstorm/email-impl.capnp");                                                      // 30
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;                        // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 32
var EmailSendPort = EmailRpc.EmailSendPort;                                                                            // 33
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 35
                                                                                                                       //
var ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                        // 37
var HOSTNAME = ROOT_URL.hostname;                                                                                      // 38
var RECIPIENT_LIMIT = 20;                                                                                              // 40
var CLIENT_TIMEOUT = 15000; // 15s                                                                                     // 42
// Every day, reset all per-user sent counts to zero.                                                                  // 44
// TODO(cleanup): Consider a more granular approach. For example, each user could have a timer                         // 45
//   after which their count will reset. We'd only check the timer when that user is trying to                         // 46
//   send a new message. This avoids a global query.                                                                   // 47
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 48
  // only first replica                                                                                                // 48
  SandstormDb.periodicCleanup(86400000, function () {                                                                  // 49
    Meteor.users.update({                                                                                              // 50
      dailySentMailCount: {                                                                                            // 50
        $exists: true                                                                                                  // 50
      }                                                                                                                // 50
    }, {                                                                                                               // 50
      $unset: {                                                                                                        // 51
        dailySentMailCount: ""                                                                                         // 51
      }                                                                                                                // 51
    }, {                                                                                                               // 51
      multi: true                                                                                                      // 52
    });                                                                                                                // 52
  });                                                                                                                  // 53
}                                                                                                                      // 54
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 56
  var server = new SMTPServer({                                                                                        // 57
    banner: "Sandstorm Mail Server",                                                                                   // 58
    authOptional: true,                                                                                                // 59
    disabledCommands: ["STARTTLS", "AUTH"],                                                                            // 60
    onData: function (req, session, callback) {                                                                        // 61
      var mailparser = new MailParser();                                                                               // 62
      req.pipe(mailparser);                                                                                            // 63
      mailparser.on("end", function (mail) {                                                                           // 65
        // Wrap in outer promise for easier error handling.                                                            // 66
        Promise.resolve().then(function () {                                                                           // 67
          // Extract the 'from' address.                                                                               // 68
          var from = void 0;                                                                                           // 69
                                                                                                                       //
          if (mail.from && mail.from.length > 0) {                                                                     // 70
            // It's theoretically possible for the message to have multiple 'from' headers, but this                   // 71
            // never really happens in legitimate practice so we'll just take the first one.                           // 72
            from = mail.from[0];                                                                                       // 73
          } else {                                                                                                     // 74
            // The mail body is missing a 'From:' header. We'll use the bounce address instead,                        // 75
            // with the caveat that this is sometimes *not* the original sender                                        // 76
            // but rather some intermediate agent (e.g. a mailing list daemon). See:                                   // 77
            //   http://en.wikipedia.org/wiki/Bounce_address                                                           // 78
            // TODO(someday): Is this really right, or should we report a blank address instead?                       // 79
            from = {                                                                                                   // 80
              address: session.envelope.mailFrom,                                                                      // 80
              name: false                                                                                              // 80
            };                                                                                                         // 80
          }                                                                                                            // 81
                                                                                                                       //
          var attachments = [];                                                                                        // 83
                                                                                                                       //
          if (mail.attachments) {                                                                                      // 84
            attachments = mail.attachments.map(function (attachment) {                                                 // 85
              var disposition = attachment.contentDisposition || "attachment";                                         // 86
              disposition += ';\n\tfilename="' + (attachment.fileName || attachment.generatedFileName) + '"';          // 87
              return {                                                                                                 // 88
                contentType: attachment.contentType,                                                                   // 89
                contentDisposition: disposition,                                                                       // 90
                contentId: attachment.contentId,                                                                       // 91
                content: attachment.content                                                                            // 92
              };                                                                                                       // 88
            });                                                                                                        // 94
          }                                                                                                            // 95
                                                                                                                       //
          if (mail.replyTo && mail.replyTo.length > 1) {                                                               // 97
            console.error("More than one reply-to address address was received in an email.");                         // 98
          }                                                                                                            // 99
                                                                                                                       //
          var mailMessage = {                                                                                          // 101
            // Note that converting the date to nanoseconds actually goes outside the range of                         // 102
            // integers that Javascript can represent precisely. But this is OK because dates aren't                   // 103
            // precise anyway.                                                                                         // 104
            date: (mail.date || new Date()).getTime() * 1000000,                                                       // 105
            from: from,                                                                                                // 106
            to: mail.to,                                                                                               // 107
            cc: mail.cc || [],                                                                                         // 108
            bcc: mail.bcc || [],                                                                                       // 109
            replyTo: mail.replyTo && mail.replyTo[0] || {},                                                            // 110
            messageId: mail.headers["message-id"] || Meteor.uuid() + "@" + HOSTNAME,                                   // 111
            references: mail.references || [],                                                                         // 112
            inReplyTo: mail.inReplyTo || [],                                                                           // 113
            subject: mail.subject || "",                                                                               // 114
            text: mail.text || null,                                                                                   // 115
            html: mail.html || null,                                                                                   // 116
            attachments: attachments                                                                                   // 117
          }; // Get list of grain IDs.                                                                                 // 101
                                                                                                                       //
          var grainPublicIds = _.uniq(session.envelope.rcptTo.map(function (deliverTo) {                               // 121
            // smtp-server already validates that the address contains an @.                                           // 122
            // To simplify things, we ignore the hostname part of the address and assume that the                      // 123
            // message would not have been sent here if it weren't intended for our host. Usually                      // 124
            // there will be an nginx frontend verifying hostnames anyway. Grain public IDs are                        // 125
            // globally unique anyway, so an e-mail meant for another server presumably won't match                    // 126
            // any ID at this one anyway.                                                                              // 127
            var address = deliverTo.address;                                                                           // 121
            return address.slice(0, address.indexOf("@"));                                                             // 129
          })); // Deliver to each grain in parallel.                                                                   // 130
                                                                                                                       //
                                                                                                                       //
          return Promise.all(grainPublicIds.map(function (publicId) {                                                  // 133
            // Wrap in a function so that we can call it recursively to retry.                                         // 134
            var tryDeliver = function (retryCount) {                                                                   // 135
              var grainId = void 0;                                                                                    // 136
              return inMeteor(function () {                                                                            // 137
                var grain = Grains.findOne({                                                                           // 138
                  publicId: publicId                                                                                   // 138
                }, {                                                                                                   // 138
                  fields: {}                                                                                           // 138
                });                                                                                                    // 138
                                                                                                                       //
                if (grain) {                                                                                           // 139
                  grainId = grain._id;                                                                                 // 140
                  return globalBackend.continueGrain(grainId, retryCount > 0);                                         // 141
                } else {                                                                                               // 142
                  // TODO(someday): We really ought to rig things up so that the 'RCPT TO' SMTP command                // 143
                  // fails in this case, by adding an onRcptTo() callback.                                             // 144
                  throw new Error("No such grain: " + publicId);                                                       // 145
                }                                                                                                      // 146
              }).then(function (grainInfo) {                                                                           // 147
                var supervisor = grainInfo.supervisor;                                                                 // 148
                var uiView = supervisor.getMainView().view; // Create an arbitrary struct to use as the session params. E-mail sessions actually
                // require no params, but node-capnp won't let us pass null and we don't have an                       // 152
                // EmptyStruct type available, so we just use EmailAddress, but any struct type would                  // 153
                // work.                                                                                               // 154
                // TODO(cleanup): Fix node-capnp to accept null.                                                       // 155
                                                                                                                       //
                var emptyParams = Capnp.serialize(EmailRpc.EmailAddress, {}); // Create a new session of type HackEmailSession. This is a short-term hack until
                // persistent capabilities and the Powerbox are implemented. We need to pass along a                   // 159
                // HackSessionContext because old versions of sandstorm-http-bridge always use                         // 160
                // the context of the most recent session; if an old RoundCube grain has a                             // 161
                // WebSession open, receives an email, and then tries to send an email, the request                    // 162
                // will go out on the SessionContext associated with the HackEmailSession that                         // 163
                // delivered the email.                                                                                // 164
                                                                                                                       //
                var session = uiView.newSession({}, makeHackSessionContext(grainId), "0xc3b5ced7344b04a6", emptyParams).session.castAs(EmailSendPort);
                return session.send(mailMessage);                                                                      // 169
              }).catch(function (err) {                                                                                // 170
                if (shouldRestartGrain(err, retryCount)) {                                                             // 171
                  return tryDeliver(retryCount + 1);                                                                   // 172
                } else {                                                                                               // 173
                  throw err;                                                                                           // 174
                }                                                                                                      // 175
              });                                                                                                      // 176
            };                                                                                                         // 177
                                                                                                                       //
            return tryDeliver(0);                                                                                      // 179
          }));                                                                                                         // 180
        }).then(function () {                                                                                          // 181
          callback(null, "Message delivered");                                                                         // 182
        }, function (err) {                                                                                            // 183
          console.error("E-mail delivery failure:", err.stack);                                                        // 184
          var mailErr = new Error(err.message);                                                                        // 185
          err.responseCode = 451;                                                                                      // 186
          callback(err.message);                                                                                       // 187
        });                                                                                                            // 188
      });                                                                                                              // 189
    }                                                                                                                  // 190
  });                                                                                                                  // 57
                                                                                                                       //
  if (global.SANDSTORM_SMTP_LISTEN_HANDLE) {                                                                           // 193
    server.listen(global.SANDSTORM_SMTP_LISTEN_HANDLE);                                                                // 194
  } else {                                                                                                             // 195
    // We must be running `run-dev.sh`.                                                                                // 196
    var BIND_IP = process.env.BIND_IP || "127.0.0.1";                                                                  // 197
    var SMTP_LISTEN_PORT = Meteor.settings.public.smtpListenPort || 30025;                                             // 198
    server.listen(SMTP_LISTEN_PORT, BIND_IP);                                                                          // 199
  }                                                                                                                    // 200
});                                                                                                                    // 201
                                                                                                                       //
hackSendEmail = function (session, email) {                                                                            // 203
  return inMeteor(function () {                                                                                        // 204
    var recipientCount = 0;                                                                                            // 205
    recipientCount += email.to ? email.to.length : 0;                                                                  // 206
    recipientCount += email.cc ? email.cc.length : 0;                                                                  // 207
    recipientCount += email.bcc ? email.bcc.length : 0;                                                                // 208
                                                                                                                       //
    if (recipientCount > RECIPIENT_LIMIT) {                                                                            // 209
      throw new Error("Sorry, Sandstorm currently only allows you to send an e-mail to " + RECIPIENT_LIMIT + " recipients at a time, for spam control. Consider setting up a mailing list. " + "Please feel free to contact us if this is a problem for you.");
    }                                                                                                                  // 214
                                                                                                                       //
    var grainAddress = session._getAddress();                                                                          // 216
                                                                                                                       //
    var userAddress = session._getUserAddress(); // Overwrite the 'from' address with the grain's address.             // 217
                                                                                                                       //
                                                                                                                       //
    if (!email.from) {                                                                                                 // 220
      email.from = {                                                                                                   // 221
        address: grainAddress                                                                                          // 222
      };                                                                                                               // 221
    }                                                                                                                  // 224
                                                                                                                       //
    if (email.from.address !== grainAddress && email.from.address !== userAddress.address) {                           // 226
      throw new Error("FROM header in outgoing emails need to equal either " + grainAddress + " or " + userAddress.address + ". Yours was: " + email.from.address);
    } // Unpack fields                                                                                                 // 230
                                                                                                                       //
                                                                                                                       //
    var from = email.from,                                                                                             // 204
        to = email.to,                                                                                                 // 204
        cc = email.cc,                                                                                                 // 204
        bcc = email.bcc,                                                                                               // 204
        replyTo = email.replyTo,                                                                                       // 204
        subject = email.subject,                                                                                       // 204
        text = email.text,                                                                                             // 204
        html = email.html;                                                                                             // 204
    var options = {                                                                                                    // 235
      from: from,                                                                                                      // 236
      to: to,                                                                                                          // 237
      cc: cc,                                                                                                          // 238
      bcc: bcc,                                                                                                        // 239
      replyTo: replyTo,                                                                                                // 240
      subject: subject,                                                                                                // 241
      text: text,                                                                                                      // 242
      html: html,                                                                                                      // 243
      envelope: {                                                                                                      // 244
        from: grainAddress,                                                                                            // 245
        to: to,                                                                                                        // 246
        cc: cc,                                                                                                        // 247
        bcc: bcc                                                                                                       // 248
      }                                                                                                                // 244
    };                                                                                                                 // 235
    var headers = {};                                                                                                  // 252
                                                                                                                       //
    if (email.messageId) {                                                                                             // 253
      headers["message-id"] = email.messageId;                                                                         // 254
    }                                                                                                                  // 255
                                                                                                                       //
    if (email.references) {                                                                                            // 257
      headers["references"] = email.references; // jscs:ignore requireDotNotation                                      // 258
    }                                                                                                                  // 259
                                                                                                                       //
    if (email.inReplyTo) {                                                                                             // 261
      headers["in-reply-to"] = email.inReplyTo;                                                                        // 262
    }                                                                                                                  // 263
                                                                                                                       //
    if (email.date) {                                                                                                  // 265
      var date = new Date(email.date / 1000000);                                                                       // 266
                                                                                                                       //
      if (!isNaN(date.getTime())) {                                                                                    // 267
        // Check to make sure date is valid                                                                            // 267
        headers["date"] = date.toUTCString(); // jscs:ignore requireDotNotation                                        // 268
      }                                                                                                                // 269
    }                                                                                                                  // 270
                                                                                                                       //
    options.headers = headers;                                                                                         // 272
                                                                                                                       //
    if (email.attachments) {                                                                                           // 274
      var attachments = [];                                                                                            // 275
      email.attachments.forEach(function (attachment) {                                                                // 276
        attachments.push({                                                                                             // 277
          filename: false,                                                                                             // 278
          // Tell nodemailer not to generate a new filename.                                                           // 278
          cid: attachment.contentId,                                                                                   // 279
          contentType: attachment.contentType,                                                                         // 280
          contentDisposition: attachment.contentDisposition,                                                           // 281
          content: attachment.content                                                                                  // 282
        });                                                                                                            // 277
      });                                                                                                              // 284
      options.attachments = attachments;                                                                               // 285
    }                                                                                                                  // 286
                                                                                                                       //
    var grain = Grains.findOne(session.grainId);                                                                       // 288
    if (!grain) throw new Error("Grain does not exist.");                                                              // 289
    globalDb.incrementDailySentMailCount(grain.userId);                                                                // 290
    rawSend(options);                                                                                                  // 292
  }.bind(_this)).catch(function (err) {                                                                                // 293
    console.error("Error sending e-mail:", err.stack);                                                                 // 294
    throw err;                                                                                                         // 295
  });                                                                                                                  // 296
};                                                                                                                     // 297
                                                                                                                       //
var EmailVerifierImpl = function (_PersistentImpl) {                                                                   //
  (0, _inherits3.default)(EmailVerifierImpl, _PersistentImpl);                                                         //
                                                                                                                       //
  function EmailVerifierImpl(db, saveTemplate, params) {                                                               // 300
    (0, _classCallCheck3.default)(this, EmailVerifierImpl);                                                            // 300
                                                                                                                       //
    var _this2 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 300
                                                                                                                       //
    _this2._id = params.id;                                                                                            // 302
    _this2._services = params.services;                                                                                // 303
    return _this2;                                                                                                     // 300
  }                                                                                                                    // 304
                                                                                                                       //
  EmailVerifierImpl.prototype.getId = function () {                                                                    //
    function getId() {                                                                                                 //
      return {                                                                                                         // 307
        id: new Buffer(this._id, "base64")                                                                             // 307
      };                                                                                                               // 307
    }                                                                                                                  // 308
                                                                                                                       //
    return getId;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  EmailVerifierImpl.prototype.verifyEmail = function () {                                                              //
    function verifyEmail(tabId, verification) {                                                                        //
      var _this3 = this;                                                                                               // 310
                                                                                                                       //
      return unwrapFrontendCap(verification, "verifiedEmail", function (verification) {                                // 311
        if (verification.tabId !== tabId.toString("hex")) {                                                            // 312
          throw new Error("VerifiedEmail is from a different tab");                                                    // 313
        }                                                                                                              // 314
                                                                                                                       //
        if (_this3._services) {                                                                                        // 316
          // Since this verifier is restricted to specific services, only indicate a match if the                      // 317
          // VerifiedEmail was for the correct verifier ID. (If our _services is null, then we                         // 318
          // match all services, and therefore all VerifiedEmails.)                                                    // 319
          if (verification.verifierId !== _this3._id) {                                                                // 320
            throw new Error("VerifierEmail is for a different EmailVerifier.");                                        // 321
          }                                                                                                            // 322
        }                                                                                                              // 323
                                                                                                                       //
        return verification.address;                                                                                   // 325
      });                                                                                                              // 326
    }                                                                                                                  // 327
                                                                                                                       //
    return verifyEmail;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  return EmailVerifierImpl;                                                                                            //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
;                                                                                                                      // 328
                                                                                                                       //
var VerifiedEmailImpl = function (_PersistentImpl2) {                                                                  //
  (0, _inherits3.default)(VerifiedEmailImpl, _PersistentImpl2);                                                        //
                                                                                                                       //
  function VerifiedEmailImpl(db, saveTemplate) {                                                                       // 331
    (0, _classCallCheck3.default)(this, VerifiedEmailImpl);                                                            // 331
    return (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));              // 331
  }                                                                                                                    // 333
                                                                                                                       //
  return VerifiedEmailImpl;                                                                                            //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
function getVerifiedEmails(db, userId, verifierId) {                                                                   // 336
  // Get all of the email addresses verified as belonging to the given user using the given                            // 337
  // verifier.                                                                                                         // 338
  var services = null;                                                                                                 // 340
  var couldAddAnother = true;                                                                                          // 341
                                                                                                                       //
  if (verifierId) {                                                                                                    // 343
    var verifier = db.collections.apiTokens.findOne({                                                                  // 344
      "frontendRef.emailVerifier.id": verifierId                                                                       // 345
    });                                                                                                                // 345
    if (!verifier) return []; // invalid verifier                                                                      // 346
                                                                                                                       //
    var verifierInfo = verifier.frontendRef.emailVerifier;                                                             // 347
                                                                                                                       //
    if (verifierInfo.services) {                                                                                       // 349
      // Limit to the listed services.                                                                                 // 350
      services = {};                                                                                                   // 351
      verifierInfo.services.forEach(function (service) {                                                               // 352
        return services[service] = true;                                                                               // 352
      });                                                                                                              // 352
      couldAddAnother = !!services.email;                                                                              // 353
    }                                                                                                                  // 354
  }                                                                                                                    // 355
                                                                                                                       //
  var user = Meteor.users.findOne(userId);                                                                             // 357
  var emails = {}; // map address -> true, for uniquification                                                          // 358
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 360
    _id: {                                                                                                             // 360
      $in: SandstormDb.getUserCredentialIds(user)                                                                      // 360
    }                                                                                                                  // 360
  }).forEach(function (credential) {                                                                                   // 360
    if (!services || services[SandstormDb.getServiceName(credential)]) {                                               // 361
      SandstormDb.getVerifiedEmailsForCredential(credential).forEach(function (email) {                                // 362
        emails[email.email] = true;                                                                                    // 363
      });                                                                                                              // 363
    }                                                                                                                  // 364
  });                                                                                                                  // 365
  return {                                                                                                             // 367
    emails: Object.keys(emails),                                                                                       // 367
    couldAddAnother: couldAddAnother                                                                                   // 367
  };                                                                                                                   // 367
} // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 368
//   (ugh).                                                                                                            // 371
                                                                                                                       //
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 372
  globalFrontendRefRegistry.register({                                                                                 // 373
    frontendRefField: "emailVerifier",                                                                                 // 374
    typeId: EmailRpc.EmailVerifier.typeId,                                                                             // 375
    restore: function (db, saveTemplate, params) {                                                                     // 377
      return new Capnp.Capability(new EmailVerifierImpl(db, saveTemplate, params), EmailImpl.PersistentEmailVerifier);
    },                                                                                                                 // 380
    validate: function (db, session, value) {                                                                          // 382
      check(value, {                                                                                                   // 383
        services: Match.Optional([String])                                                                             // 383
      });                                                                                                              // 383
      var services = value.services;                                                                                   // 385
                                                                                                                       //
      if (services) {                                                                                                  // 386
        services.forEach(function (service) {                                                                          // 387
          if (!Accounts.loginServices[service]) {                                                                      // 388
            throw new Error("No such login service: " + service);                                                      // 389
          }                                                                                                            // 390
        });                                                                                                            // 391
      }                                                                                                                // 392
                                                                                                                       //
      value.id = Crypto.randomBytes(16).toString("base64");                                                            // 394
      return {                                                                                                         // 396
        descriptor: {                                                                                                  // 397
          tags: [{                                                                                                     // 397
            id: EmailRpc.EmailVerifier.typeId                                                                          // 397
          }]                                                                                                           // 397
        },                                                                                                             // 397
        requirements: [],                                                                                              // 398
        frontendRef: value                                                                                             // 399
      };                                                                                                               // 396
    },                                                                                                                 // 401
    query: function (db, userId, value) {                                                                              // 403
      var results = [];                                                                                                // 404
      results.push({                                                                                                   // 406
        _id: "emailverifier-all",                                                                                      // 407
        frontendRef: {                                                                                                 // 408
          emailVerifier: {}                                                                                            // 408
        },                                                                                                             // 408
        cardTemplate: "emailVerifierPowerboxCard"                                                                      // 409
      });                                                                                                              // 406
                                                                                                                       //
      for (var name in meteorBabelHelpers.sanitizeForInObject(Accounts.loginServices)) {                               // 412
        if (Accounts.loginServices[name].isEnabled()) {                                                                // 413
          results.push({                                                                                               // 414
            _id: "emailverifier-" + name,                                                                              // 415
            frontendRef: {                                                                                             // 416
              emailVerifier: {                                                                                         // 416
                services: [name]                                                                                       // 416
              }                                                                                                        // 416
            },                                                                                                         // 416
            cardTemplate: "emailVerifierPowerboxCard"                                                                  // 417
          });                                                                                                          // 414
        }                                                                                                              // 419
      }                                                                                                                // 420
                                                                                                                       //
      ;                                                                                                                // 420
      return results;                                                                                                  // 422
    }                                                                                                                  // 423
  });                                                                                                                  // 373
  globalFrontendRefRegistry.register({                                                                                 // 426
    frontendRefField: "verifiedEmail",                                                                                 // 427
    typeId: EmailRpc.VerifiedEmail.typeId,                                                                             // 428
    restore: function (db, saveTemplate) {                                                                             // 430
      return new Capnp.Capability(new VerifiedEmailImpl(db, saveTemplate), EmailImpl.PersistentVerifiedEmail);         // 431
    },                                                                                                                 // 433
    validate: function (db, session, value) {                                                                          // 435
      check(value, {                                                                                                   // 436
        verifierId: Match.Optional(String),                                                                            // 436
        address: String                                                                                                // 436
      });                                                                                                              // 436
                                                                                                                       //
      if (!session.userId) {                                                                                           // 438
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 439
      } // Verify that the address actually belongs to the user.                                                       // 440
                                                                                                                       //
                                                                                                                       //
      if (!_.contains(getVerifiedEmails(db, session.userId, value.verifierId).emails, value.address)) {                // 444
        throw new Meteor.Error(403, "User has no such verified address");                                              // 446
      } // Add the session's tabId.                                                                                    // 447
                                                                                                                       //
                                                                                                                       //
      value.tabId = session.tabId; // Build the descriptor, which contains the verifier ID.                            // 450
                                                                                                                       //
      var tagValue = value.verifierId && Capnp.serialize(EmailRpc.VerifiedEmail.PowerboxTag, {                         // 453
        verifierId: new Buffer(value.verifierId, "hex")                                                                // 455
      });                                                                                                              // 455
      return {                                                                                                         // 457
        descriptor: {                                                                                                  // 458
          tags: [{                                                                                                     // 458
            id: EmailRpc.VerifiedEmail.typeId,                                                                         // 458
            value: tagValue                                                                                            // 458
          }]                                                                                                           // 458
        },                                                                                                             // 458
        requirements: [],                                                                                              // 459
        frontendRef: value                                                                                             // 460
      };                                                                                                               // 457
    },                                                                                                                 // 462
    query: function (db, userId, value) {                                                                              // 464
      var verifierId = value && Capnp.parse(EmailRpc.VerifiedEmail.PowerboxTag, value).verifierId.toString("base64");  // 465
      var verified = getVerifiedEmails(db, userId, verifierId);                                                        // 467
      var result = verified.emails.map(function (address) {                                                            // 468
        return {                                                                                                       // 468
          _id: "email-" + address,                                                                                     // 469
          frontendRef: {                                                                                               // 470
            verifiedEmail: {                                                                                           // 470
              verifierId: verifierId,                                                                                  // 470
              address: address                                                                                         // 470
            }                                                                                                          // 470
          },                                                                                                           // 470
          cardTemplate: "verifiedEmailPowerboxCard",                                                                   // 471
          searchTerms: [address]                                                                                       // 472
        };                                                                                                             // 468
      });                                                                                                              // 468
                                                                                                                       //
      if (verified.couldAddAnother) {                                                                                  // 475
        result.push({                                                                                                  // 476
          _id: "email-add-a-new-one",                                                                                  // 477
          frontendRef: {                                                                                               // 478
            verifiedEmail: {                                                                                           // 478
              verifierId: verifierId                                                                                   // 478
            }                                                                                                          // 478
          },                                                                                                           // 478
          cardTemplate: "addNewVerifiedEmailPowerboxCard",                                                             // 479
          configureTemplate: "addNewVerifiedEmailPowerboxConfiguration"                                                // 480
        });                                                                                                            // 476
      }                                                                                                                // 482
                                                                                                                       //
      return result;                                                                                                   // 484
    }                                                                                                                  // 485
  });                                                                                                                  // 426
});                                                                                                                    // 487
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"00-startup.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/00-startup.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var module1 = module;                                                                                                  // 1
module1.watch(require("/imports/db-deprecated.js"));                                                                   // 1
var FrontendRefRegistry = void 0;                                                                                      // 1
module1.watch(require("/imports/server/frontend-ref.js"), {                                                            // 1
  FrontendRefRegistry: function (v) {                                                                                  // 1
    FrontendRefRegistry = v;                                                                                           // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module1.watch(require("/imports/server/persistent.js"), {                                                              // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var migrateToLatest = void 0;                                                                                          // 1
module1.watch(require("/imports/server/migrations.js"), {                                                              // 1
  migrateToLatest: function (v) {                                                                                      // 1
    migrateToLatest = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var ACCOUNT_DELETION_SUSPENSION_TIME = void 0;                                                                         // 1
module1.watch(require("/imports/constants.js"), {                                                                      // 1
  ACCOUNT_DELETION_SUSPENSION_TIME: function (v) {                                                                     // 1
    ACCOUNT_DELETION_SUSPENSION_TIME = v;                                                                              // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var onInMeteor = void 0;                                                                                               // 1
module1.watch(require("/imports/server/async-helpers.js"), {                                                           // 1
  onInMeteor: function (v) {                                                                                           // 1
    onInMeteor = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var monkeyPatchHttp = void 0;                                                                                          // 1
module1.watch(require("/imports/server/networking.js"), {                                                              // 1
  monkeyPatchHttp: function (v) {                                                                                      // 1
    monkeyPatchHttp = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var Fiber = void 0;                                                                                                    // 1
module1.watch(require("fibers"), {                                                                                     // 1
  "default": function (v) {                                                                                            // 1
    Fiber = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
                                                                                                                       //
var url = require("url");                                                                                              // 24
                                                                                                                       //
var Capnp = require("capnp");                                                                                          // 25
                                                                                                                       //
globalFrontendRefRegistry = new FrontendRefRegistry();                                                                 // 27
SandstormPowerbox.registerUiViewQueryHandler(globalFrontendRefRegistry);                                               // 29
                                                                                                                       //
if (global.BlackrockPayments && BlackrockPayments.registerPaymentsApi) {                                               // 31
  // TODO(cleanup): Meteor.startup() needed because unwrapFrontendCap is not defined yet when this                     // 32
  //   first runs. Move it into an import.                                                                             // 33
  Meteor.startup(function () {                                                                                         // 34
    BlackrockPayments.registerPaymentsApi(globalFrontendRefRegistry, PersistentImpl, unwrapFrontendCap);               // 35
  });                                                                                                                  // 37
}                                                                                                                      // 38
                                                                                                                       //
getWildcardOrigin = globalDb.getWildcardOrigin.bind(globalDb);                                                         // 40
Meteor.onConnection(function (connection) {                                                                            // 42
  // TODO(cleanup): This is the best way I've thought of so far to allow methods declared in                           // 43
  //   packages to actually use the DB, but it's pretty sad.                                                           // 44
  connection.sandstormDb = globalDb;                                                                                   // 45
  connection.frontendRefRegistry = globalFrontendRefRegistry;                                                          // 46
});                                                                                                                    // 47
SandstormDb.periodicCleanup(5 * 60 * 1000, SandstormPermissions.cleanupSelfDestructing(globalDb));                     // 48
SandstormDb.periodicCleanup(10 * 60 * 1000, SandstormPermissions.cleanupClientPowerboxTokens(globalDb));               // 49
SandstormDb.periodicCleanup(60 * 60 * 1000, function () {                                                              // 51
  globalDb.cleanupExpiredAssetUploads();                                                                               // 52
});                                                                                                                    // 53
SandstormDb.periodicCleanup(24 * 60 * 60 * 1000, function () {                                                         // 54
  SandstormAutoupdateApps.updateAppIndex(globalDb);                                                                    // 55
});                                                                                                                    // 56
var deleteAccount = global.BlackrockPayments && global.BlackrockPayments.deleteAccount;                                // 57
SandstormDb.periodicCleanup(24 * 60 * 60 * 1000, function () {                                                         // 58
  globalDb.deletePendingAccounts(ACCOUNT_DELETION_SUSPENSION_TIME, globalBackend, deleteAccount);                      // 59
});                                                                                                                    // 61
monkeyPatchHttp(globalDb, HTTP);                                                                                       // 63
Meteor.startup(function () {                                                                                           // 65
  migrateToLatest(globalDb, globalBackend);                                                                            // 65
}); // If there are multiple replicas, prefix every log message with our replica number.                               // 65
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 68
  var patchConsole = function (name) {                                                                                 // 68
    var old = console[name];                                                                                           // 72
                                                                                                                       //
    console[name] = function () {                                                                                      // 73
      // Meteor in dev mode writes "LISTENING" to tell the dev runner that it's ready to accept                        // 74
      // connections.                                                                                                  // 75
      if (arguments.length == 1 && arguments[0] == "LISTENING") {                                                      // 76
        old.apply(this, arguments);                                                                                    // 77
      } else {                                                                                                         // 78
        old.apply(this, [prefix].concat(Array.prototype.slice.call(arguments)));                                       // 79
      }                                                                                                                // 80
    };                                                                                                                 // 81
  };                                                                                                                   // 82
                                                                                                                       //
  var prefix = "replica" + Meteor.settings.replicaNumber.toString() + ":";                                             // 69
  patchConsole("log");                                                                                                 // 84
  patchConsole("info");                                                                                                // 85
  patchConsole("warn");                                                                                                // 86
  patchConsole("error");                                                                                               // 87
} // Make the fiber pool size infinite(ish)!                                                                           // 88
//                                                                                                                     // 91
// Each fiber created adds an entry to `v8::Isolate::ThreadDataTable`. Unfortunatley, items are                        // 92
// never deleted from this table. So if we create and delete a lot of fibers, then we leak memory.                     // 93
// Worse yet, the table is represented as a linked list, and v8 performs a linear scan of this                         // 94
// linked list every time we switch fibers. We've seen cases where Sandstorm was spending 65% of                       // 95
// its CPU time just scanning this list! The v8 people say this is "working as intended".                              // 96
//                                                                                                                     // 97
// The fibers package has implemented a work-around by maintaining a fiber pool. Fibers are not                        // 98
// deleted; they are returned to the pool. Unfortunately the pool has a default size of 120. So                        // 99
// if we pass 120 simultaneous fibers, we start leaking and slowing down. It's very easy to hit                        // 100
// this number with a few dozen users present.                                                                         // 101
//                                                                                                                     // 102
// Up until it hits the limit, the fibers module will grow the pool dynamically, starting with an                      // 103
// empty pool and adding each new fiber to it. Therefore, if we set the pool size to an impossibly                     // 104
// large number, we effectively get a pool size equal to the maximum number of simultaneous fibers                     // 105
// seen. This is exactly what we want! Now no fibers are ever deleted, so we never leak. A                             // 106
// Sandstorm server that sees a brief surge of traffic may end up holding on to unused RAM                             // 107
// long-term, but this is a relatively obscure problem.                                                                // 108
//                                                                                                                     // 109
// I initially tried to use the value `Infinity` here, but somehow when this made its way down into                    // 110
// the C++ code and was converted to an integer, that integer ended up being zero. So instead we                       // 111
// use 1e9 (1 billion), which ought to be enough for anyone.                                                           // 112
//                                                                                                                     // 113
// Third-party issues, for reference:                                                                                  // 114
//                                                                                                                     // 115
//    https://bugs.chromium.org/p/v8/issues/detail?id=5338                                                             // 116
//    https://bugs.chromium.org/p/v8/issues/detail?id=3777                                                             // 117
//    https://github.com/laverdet/node-fibers/issues/305                                                               // 118
                                                                                                                       //
                                                                                                                       //
Fiber.poolSize = 1e9; // Special debugging enabled on Blackrock only.                                                  // 120
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 123
  var sampleStackIfTooManyFibers = function () {                                                                       // 123
    if (Fiber.fibersCreated > TOO_MANY_FIBERS && sampleCounter++ % FIBER_SAMPLING_PERIOD == 0) {                       // 148
      var stack = new Error().stack;                                                                                   // 149
      console.warn("There are too many fibers!", stack);                                                               // 150
    }                                                                                                                  // 151
  }; // Sample calls to inMeteor().                                                                                    // 152
                                                                                                                       //
                                                                                                                       //
  // Also take samples from various Meteor infrastructure that schedules fibers.                                       // 157
  var addFiberSampling = function (prototype, functionName) {                                                          // 123
    var old = prototype[functionName];                                                                                 // 159
                                                                                                                       //
    prototype[functionName] = function () {                                                                            // 160
      sampleStackIfTooManyFibers();                                                                                    // 161
      return old.apply(this, arguments);                                                                               // 162
    };                                                                                                                 // 163
  };                                                                                                                   // 164
                                                                                                                       //
  console.warn("Fiber bomb defense enabled."); // TEMPORARY: Monitor the number of fibers created and kill the process any time it goes over
  //   2000. Unfortunately, due to the aforementioned linked list in ThreadDataTable, the process                      // 127
  //   will become unreasonably slow once the list gets this big. It's better to kill the process                      // 128
  //   so that it restarts fresh rather than to let the Sandstorm server become unresponsive.                          // 129
  // TODO(soon): Remove this when the bug is fixed.                                                                    // 130
                                                                                                                       //
  setInterval(function () {                                                                                            // 131
    if (Fiber.fibersCreated > 2000) {                                                                                  // 132
      console.error("Process has allocated more than 2000 concurrent fibers. Due to " + "https://bugs.chromium.org/p/v8/issues/detail?id=5338 it will become extremely slow " + "unless we restart it. ABORTING");
      process.abort();                                                                                                 // 137
    }                                                                                                                  // 138
  }, 5000); // Let's also log some stack traces when fiber count gets high so we can find out what's                   // 139
  // happening.                                                                                                        // 142
                                                                                                                       //
  var TOO_MANY_FIBERS = 1000; // Start sampling when we have more than this many fibers.                               // 143
                                                                                                                       //
  var FIBER_SAMPLING_PERIOD = 100; // Sample one stack per this many fibers created.                                   // 144
                                                                                                                       //
  var sampleCounter = 0;                                                                                               // 146
  onInMeteor(sampleStackIfTooManyFibers);                                                                              // 155
  addFiberSampling(Meteor._SynchronousQueue.prototype, "runTask");                                                     // 166
  addFiberSampling(Meteor._SynchronousQueue.prototype, "queueTask");                                                   // 167
  addFiberSampling(Meteor, "bindEnvironment");                                                                         // 168
}                                                                                                                      // 169
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 171
  // On Blackrock, dump debug info about local capability counts every minute.                                         // 172
  Meteor.setInterval(Capnp.dumpLocalCapTypeCounts, 60000);                                                             // 173
}                                                                                                                      // 174
                                                                                                                       //
OAuth._checkRedirectUrlOrigin = function (redirectUrl) {                                                               // 176
  // Mostly copied from meteor/packages/oauth/oauth_server.js                                                          // 177
  // We override this method in order to support login from stand-alone grain domains.                                 // 178
  var appHost = Meteor.absoluteUrl();                                                                                  // 179
  var appHostReplacedLocalhost = Meteor.absoluteUrl(undefined, {                                                       // 180
    replaceLocalhost: true                                                                                             // 181
  });                                                                                                                  // 180
  var redirectParsed = url.parse(redirectUrl);                                                                         // 184
  return !(redirectUrl.substr(0, appHost.length) === appHost || redirectUrl.substr(0, appHostReplacedLocalhost.length) === appHostReplacedLocalhost || globalDb.hostIsStandalone(redirectParsed.hostname));
};                                                                                                                     // 191
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"account-suspension.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/account-suspension.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var send = void 0;                                                                                                     // 1
module.watch(require("/imports/server/email.js"), {                                                                    // 1
  send: function (v) {                                                                                                 // 1
    send = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
function sendDeletionEmails(db, deletedUserId, byAdminUserId, feedback) {                                              // 19
  var deletedUser = db.getUser(deletedUserId);                                                                         // 20
                                                                                                                       //
  var userEmail = _.findWhere(SandstormDb.getUserEmails(deletedUser), {                                                // 22
    primary: true                                                                                                      // 22
  });                                                                                                                  // 22
                                                                                                                       //
  if (!byAdminUserId) {                                                                                                // 23
    // This was initiated by the user, send them an email                                                              // 23
    if (!userEmail) {                                                                                                  // 24
      console.error("Couldn't send deletion email. No address found for user with userId:", deletedUser._id);          // 25
    } else {                                                                                                           // 27
      var _emailOptions = {                                                                                            // 28
        from: db.getReturnAddress(),                                                                                   // 29
        subject: "Your account on " + db.getServerTitle() + " will be deleted in 7 days.",                             // 30
        text: "You have requested that your Sandstorm account on " + db.getServerTitle() + " be deleted. Your account has been suspended and will be fully deleted in seven days. If you change your mind, log into " + process.env.ROOT_URL + " to cancel the process.\n\nIf you did not request this deletion, please contact the server administrator immediately."
      };                                                                                                               // 28
                                                                                                                       //
      try {                                                                                                            // 35
        _emailOptions.to = userEmail.email;                                                                            // 36
        send(_emailOptions);                                                                                           // 37
      } catch (err) {                                                                                                  // 38
        console.error("Failed to send deletion email to user (id=" + deletedUser._id + ") with error: " + err);        // 39
      }                                                                                                                // 41
    }                                                                                                                  // 42
  }                                                                                                                    // 43
                                                                                                                       //
  var deleteUserString = userEmail && userEmail.email || "userId=" + deletedUser._id;                                  // 45
  var emailOptions = {                                                                                                 // 46
    from: db.getReturnAddress(),                                                                                       // 47
    subject: "Account for " + deleteUserString + " on " + db.getServerTitle() + " will be deleted in 7 days"           // 48
  };                                                                                                                   // 46
                                                                                                                       //
  if (byAdminUserId) {                                                                                                 // 51
    var initiatingAdmin = db.getUser(byAdminUserId);                                                                   // 52
    var adminName = initiatingAdmin.profile.name;                                                                      // 53
    emailOptions.text = adminName + " has requested that the Sandstorm account held by " + deleteUserString + " on " + db.getServerTitle() + " be deleted. The account has been suspended and will be fully deleted in seven days. To cancel the deletion, go to: " + process.env.ROOT_URL + "/admin/users/" + deletedUser._id;
  } else {                                                                                                             // 55
    emailOptions.text = deleteUserString + " has requested that their account be deleted on " + db.getServerTitle() + ". The account has been suspended and will be fully deleted in seven days. To cancel the deletion, go to: " + process.env.ROOT_URL + "/admin/users/" + deletedUser._id;
                                                                                                                       //
    if (feedback) {                                                                                                    // 57
      emailOptions.text += "\nUser gave the following feedback: " + feedback;                                          // 58
    }                                                                                                                  // 59
  }                                                                                                                    // 60
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 62
    isAdmin: true                                                                                                      // 62
  }).forEach(function (user) {                                                                                         // 62
    var email = _.findWhere(SandstormDb.getUserEmails(user), {                                                         // 63
      primary: true                                                                                                    // 63
    });                                                                                                                // 63
                                                                                                                       //
    if (!email) {                                                                                                      // 64
      console.error("No email found for admin with userId:", user._id);                                                // 65
      return;                                                                                                          // 66
    }                                                                                                                  // 67
                                                                                                                       //
    try {                                                                                                              // 69
      emailOptions.to = email.email;                                                                                   // 70
      send(emailOptions);                                                                                              // 71
    } catch (err) {                                                                                                    // 72
      console.error("Failed to send deletion email to admin (id=" + user._id + ") with error: " + err);                // 73
    }                                                                                                                  // 75
  });                                                                                                                  // 76
}                                                                                                                      // 77
                                                                                                                       //
Meteor.methods({                                                                                                       // 79
  suspendAccount: function (userId, willDelete) {                                                                      // 80
    check(userId, String);                                                                                             // 81
    check(willDelete, Boolean);                                                                                        // 82
                                                                                                                       //
    if (!isAdmin()) {                                                                                                  // 84
      throw new Meteor.Error(403, "Only admins can suspend other users.");                                             // 85
    }                                                                                                                  // 86
                                                                                                                       //
    if (userId === Meteor.userId()) {                                                                                  // 88
      throw new Meteor.Error(400, "Admins cannot suspend their own accounts from the admin page. Please go to your account setttings.");
    }                                                                                                                  // 90
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 92
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 94
      BlackrockPayments.suspendAccount(db, userId);                                                                    // 95
    }                                                                                                                  // 96
                                                                                                                       //
    db.suspendAccount(userId, Meteor.userId(), willDelete);                                                            // 98
                                                                                                                       //
    if (willDelete) {                                                                                                  // 100
      sendDeletionEmails(db, userId, Meteor.userId());                                                                 // 101
    }                                                                                                                  // 102
  },                                                                                                                   // 103
  deleteOwnAccount: function (feedback) {                                                                              // 105
    var db = this.connection.sandstormDb;                                                                              // 106
                                                                                                                       //
    if (!Meteor.userId()) {                                                                                            // 107
      throw new Meteor.Error(403, "Must be logged in to delete an account");                                           // 108
    }                                                                                                                  // 109
                                                                                                                       //
    if (db.isUserInOrganization(Meteor.user())) {                                                                      // 111
      throw new Meteor.Error(403, "Users in an organization cannot delete their own account. " + "Please ask your admin to do it for you.");
    }                                                                                                                  // 114
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 116
      BlackrockPayments.suspendAccount(db, Meteor.userId());                                                           // 117
    }                                                                                                                  // 118
                                                                                                                       //
    db.suspendAccount(Meteor.userId(), null, true);                                                                    // 120
    sendDeletionEmails(db, Meteor.userId(), null, feedback);                                                           // 122
  },                                                                                                                   // 123
  unsuspendAccount: function (userId) {                                                                                // 125
    check(userId, String);                                                                                             // 126
                                                                                                                       //
    if (!isAdmin()) {                                                                                                  // 128
      throw new Meteor.Error(403, "Only admins can unsuspend other users.");                                           // 129
    }                                                                                                                  // 130
                                                                                                                       //
    this.connection.sandstormDb.unsuspendAccount(userId, Meteor.userId());                                             // 132
  },                                                                                                                   // 133
  unsuspendOwnAccount: function () {                                                                                   // 135
    if (!Meteor.userId()) {                                                                                            // 136
      throw new Meteor.Error(403, "Must be logged in to unsuspend an account");                                        // 137
    }                                                                                                                  // 138
                                                                                                                       //
    this.connection.sandstormDb.unsuspendAccount(Meteor.userId());                                                     // 140
  }                                                                                                                    // 141
});                                                                                                                    // 79
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"admin-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _objectWithoutProperties2 = require("babel-runtime/helpers/objectWithoutProperties");                              //
                                                                                                                       //
var _objectWithoutProperties3 = _interopRequireDefault(_objectWithoutProperties2);                                     //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.watch(require("fs"), {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.watch(require("crypto"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Heapdump = void 0;                                                                                                 // 1
module.watch(require("heapdump"), {                                                                                    // 1
  "default": function (v) {                                                                                            // 1
    Heapdump = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var SANDSTORM_LOGDIR = void 0;                                                                                         // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_LOGDIR: function (v) {                                                                                     // 1
    SANDSTORM_LOGDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var clearAdminToken = void 0,                                                                                          // 1
    checkAuth = void 0,                                                                                                // 1
    tokenIsValid = void 0,                                                                                             // 1
    tokenIsSetupSession = void 0;                                                                                      // 1
module.watch(require("/imports/server/auth.js"), {                                                                     // 1
  clearAdminToken: function (v) {                                                                                      // 1
    clearAdminToken = v;                                                                                               // 1
  },                                                                                                                   // 1
  checkAuth: function (v) {                                                                                            // 1
    checkAuth = v;                                                                                                     // 1
  },                                                                                                                   // 1
  tokenIsValid: function (v) {                                                                                         // 1
    tokenIsValid = v;                                                                                                  // 1
  },                                                                                                                   // 1
  tokenIsSetupSession: function (v) {                                                                                  // 1
    tokenIsSetupSession = v;                                                                                           // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var sendEmail = void 0;                                                                                                // 1
module.watch(require("/imports/server/email.js"), {                                                                    // 1
  send: function (v) {                                                                                                 // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var fillUndefinedForChangedDoc = void 0;                                                                               // 1
module.watch(require("/imports/server/observe-helpers.js"), {                                                          // 1
  fillUndefinedForChangedDoc: function (v) {                                                                           // 1
    fillUndefinedForChangedDoc = v;                                                                                    // 1
  }                                                                                                                    // 1
}, 7);                                                                                                                 // 1
var publicAdminSettings = ["google", "github", "ldap", "saml", "emailToken", "splashUrl", "signupDialog", "adminAlert", "adminAlertTime", "adminAlertUrl", "termsUrl", "privacyUrl", "appMarketUrl", "appIndexUrl", "appUpdatesEnabled", "serverTitle", "returnAddress", "ldapNameField", "organizationMembership", "organizationSettings", "whitelabelCustomLoginProviderName", "whitelabelCustomLogoAssetId", "whitelabelHideSendFeedback", "whitelabelHideTroubleshooting", "whiteLabelHideAbout", "whitelabelUseServerTitleForHomeText", "quotaEnabled", "quotaLdapEnabled", "billingPromptUrl"];
var smtpConfigShape = {                                                                                                // 43
  hostname: String,                                                                                                    // 44
  port: Number,                                                                                                        // 45
  auth: {                                                                                                              // 46
    user: String,                                                                                                      // 47
    pass: String                                                                                                       // 48
  },                                                                                                                   // 46
  returnAddress: String                                                                                                // 50
};                                                                                                                     // 43
Meteor.methods({                                                                                                       // 53
  setAccountSetting: function (token, serviceName, value) {                                                            // 54
    checkAuth(token);                                                                                                  // 55
    check(serviceName, String);                                                                                        // 56
    check(value, Boolean); // Only check configurations for OAuth services.                                            // 57
                                                                                                                       //
    var oauthServices = ["google", "github"];                                                                          // 60
                                                                                                                       //
    if (value && oauthServices.indexOf(serviceName) != -1) {                                                           // 61
      var config = ServiceConfiguration.configurations.findOne({                                                       // 62
        service: serviceName                                                                                           // 62
      });                                                                                                              // 62
                                                                                                                       //
      if (!config) {                                                                                                   // 63
        throw new Meteor.Error(403, "You must configure the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                                                                // 66
                                                                                                                       //
      if (!config.clientId || !config.secret) {                                                                        // 68
        throw new Meteor.Error(403, "You must provide a non-empty clientId and secret for the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                                                                // 71
    }                                                                                                                  // 72
                                                                                                                       //
    Settings.upsert({                                                                                                  // 74
      _id: serviceName                                                                                                 // 74
    }, {                                                                                                               // 74
      $set: {                                                                                                          // 74
        value: value                                                                                                   // 74
      }                                                                                                                // 74
    });                                                                                                                // 74
                                                                                                                       //
    if (value) {                                                                                                       // 75
      Settings.update({                                                                                                // 76
        _id: serviceName                                                                                               // 76
      }, {                                                                                                             // 76
        $unset: {                                                                                                      // 76
          automaticallyReset: 1                                                                                        // 76
        }                                                                                                              // 76
      });                                                                                                              // 76
    }                                                                                                                  // 77
  },                                                                                                                   // 78
  setSmtpConfig: function (token, config) {                                                                            // 80
    checkAuth(token);                                                                                                  // 81
    check(config, smtpConfigShape);                                                                                    // 82
    Settings.upsert({                                                                                                  // 84
      _id: "smtpConfig"                                                                                                // 84
    }, {                                                                                                               // 84
      $set: {                                                                                                          // 84
        value: config                                                                                                  // 84
      }                                                                                                                // 84
    });                                                                                                                // 84
  },                                                                                                                   // 85
  disableEmail: function (token) {                                                                                     // 87
    checkAuth(token);                                                                                                  // 88
    var db = this.connection.sandstormDb;                                                                              // 90
    db.collections.settings.update({                                                                                   // 91
      _id: "smtpConfig"                                                                                                // 91
    }, {                                                                                                               // 91
      $set: {                                                                                                          // 91
        "value.hostname": ""                                                                                           // 91
      }                                                                                                                // 91
    });                                                                                                                // 91
  },                                                                                                                   // 92
  setSetting: function (token, name, value) {                                                                          // 94
    checkAuth(token);                                                                                                  // 95
    check(name, String);                                                                                               // 96
    check(value, Match.OneOf(null, String, Date, Boolean));                                                            // 97
    Settings.upsert({                                                                                                  // 99
      _id: name                                                                                                        // 99
    }, {                                                                                                               // 99
      $set: {                                                                                                          // 99
        value: value                                                                                                   // 99
      }                                                                                                                // 99
    });                                                                                                                // 99
  },                                                                                                                   // 100
  saveOrganizationSettings: function (token, params) {                                                                 // 102
    checkAuth(token);                                                                                                  // 103
    check(params, {                                                                                                    // 104
      membership: {                                                                                                    // 105
        emailToken: {                                                                                                  // 106
          enabled: Boolean,                                                                                            // 107
          domain: String                                                                                               // 108
        },                                                                                                             // 106
        google: {                                                                                                      // 110
          enabled: Boolean,                                                                                            // 111
          domain: String                                                                                               // 112
        },                                                                                                             // 110
        ldap: {                                                                                                        // 114
          enabled: Boolean                                                                                             // 115
        },                                                                                                             // 114
        saml: {                                                                                                        // 117
          enabled: Boolean                                                                                             // 118
        }                                                                                                              // 117
      },                                                                                                               // 105
      settings: {                                                                                                      // 121
        disallowGuests: Boolean,                                                                                       // 122
        shareContacts: Boolean                                                                                         // 123
      }                                                                                                                // 121
    });                                                                                                                // 104
    this.connection.sandstormDb.collections.settings.upsert({                                                          // 127
      _id: "organizationMembership"                                                                                    // 127
    }, {                                                                                                               // 127
      value: params.membership                                                                                         // 127
    });                                                                                                                // 127
    this.connection.sandstormDb.collections.settings.upsert({                                                          // 128
      _id: "organizationSettings"                                                                                      // 128
    }, {                                                                                                               // 128
      value: params.settings                                                                                           // 128
    });                                                                                                                // 128
  },                                                                                                                   // 129
  adminConfigureLoginService: function (token, options) {                                                              // 131
    checkAuth(token);                                                                                                  // 132
    check(options, Match.ObjectIncluding({                                                                             // 133
      service: String                                                                                                  // 133
    }));                                                                                                               // 133
    ServiceConfiguration.configurations.upsert({                                                                       // 135
      service: options.service                                                                                         // 135
    }, options);                                                                                                       // 135
  },                                                                                                                   // 136
  clearResumeTokensForService: function (token, serviceName) {                                                         // 138
    checkAuth(token);                                                                                                  // 139
    check(serviceName, String);                                                                                        // 140
    var query = {};                                                                                                    // 142
    query["services." + serviceName] = {                                                                               // 143
      $exists: true                                                                                                    // 143
    };                                                                                                                 // 143
    Meteor.users.find(query).forEach(function (credential) {                                                           // 144
      if (credential.services.resume && credential.services.resume.loginTokens && credential.services.resume.loginTokens.length > 0) {
        Meteor.users.update({                                                                                          // 147
          _id: credential._id                                                                                          // 147
        }, {                                                                                                           // 147
          $set: {                                                                                                      // 147
            "services.resume.loginTokens": []                                                                          // 147
          }                                                                                                            // 147
        });                                                                                                            // 147
      }                                                                                                                // 148
                                                                                                                       //
      Meteor.users.update({                                                                                            // 150
        "loginCredentials.id": credential._id                                                                          // 150
      }, {                                                                                                             // 150
        $set: {                                                                                                        // 151
          "services.resume.loginTokens": []                                                                            // 151
        }                                                                                                              // 151
      });                                                                                                              // 151
    });                                                                                                                // 152
  },                                                                                                                   // 153
  adminUpdateUser: function (token, userInfo) {                                                                        // 155
    checkAuth(token);                                                                                                  // 156
    check(userInfo, {                                                                                                  // 157
      userId: String,                                                                                                  // 158
      signupKey: Boolean,                                                                                              // 159
      isAdmin: Boolean                                                                                                 // 160
    });                                                                                                                // 157
    var userId = userInfo.userId;                                                                                      // 163
                                                                                                                       //
    if (userId === Meteor.userId() && !userInfo.isAdmin) {                                                             // 164
      throw new Meteor.Error(403, "User cannot remove admin permissions from itself.");                                // 165
    }                                                                                                                  // 166
                                                                                                                       //
    Meteor.users.update({                                                                                              // 168
      _id: userId                                                                                                      // 168
    }, {                                                                                                               // 168
      $set: _.omit(userInfo, ["_id", "userId"])                                                                        // 168
    });                                                                                                                // 168
  },                                                                                                                   // 169
  testSend: function (token, smtpConfig, to) {                                                                         // 171
    checkAuth(token);                                                                                                  // 172
    check(smtpConfig, smtpConfigShape);                                                                                // 173
    check(to, String);                                                                                                 // 174
    var returnAddress = smtpConfig.returnAddress,                                                                      // 171
        restConfig = (0, _objectWithoutProperties3.default)(smtpConfig, ["returnAddress"]);                            // 171
                                                                                                                       //
    try {                                                                                                              // 177
      sendEmail({                                                                                                      // 178
        to: to,                                                                                                        // 179
        from: {                                                                                                        // 180
          name: globalDb.getServerTitle(),                                                                             // 180
          address: returnAddress                                                                                       // 180
        },                                                                                                             // 180
        subject: "Testing your Sandstorm's SMTP setting",                                                              // 181
        text: "Success! Your outgoing SMTP is working.",                                                               // 182
        smtpConfig: restConfig                                                                                         // 183
      });                                                                                                              // 178
    } catch (e) {                                                                                                      // 185
      // Attempt to give more accurate error messages for a variety of known failure modes,                            // 186
      // and the actual exception data in the event a user hits a new failure mode.                                    // 187
      if (e.syscall === "getaddrinfo") {                                                                               // 188
        if (e.code === "EIO" || e.code === "ENOTFOUND") {                                                              // 189
          throw new Meteor.Error("getaddrinfo " + e.code, "Couldn't resolve \"" + smtpConfig.hostname + "\" - check for typos or broken DNS.");
        }                                                                                                              // 191
      } else if (e.syscall === "connect") {                                                                            // 192
        if (e.code === "ECONNREFUSED") {                                                                               // 193
          throw new Meteor.Error("connect ECONNREFUSED", "Server at " + smtpConfig.hostname + ":" + smtpConfig.port + " refused connection.  Check your settings, firewall rules, and that your mail server is up.");
        }                                                                                                              // 195
      } else if (e.name === "AuthError") {                                                                             // 196
        throw new Meteor.Error("auth error", "Authentication failed.  Check your credentials.  Message from " + smtpConfig.hostname + ": " + e.data);
      }                                                                                                                // 199
                                                                                                                       //
      throw new Meteor.Error("other-email-sending-error", "Error while trying to send test email: " + JSON.stringify(e));
    }                                                                                                                  // 202
  },                                                                                                                   // 203
  createSignupKey: function (token, note, quota) {                                                                     // 205
    checkAuth(token);                                                                                                  // 206
    check(note, String);                                                                                               // 207
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 208
    var key = Random.id();                                                                                             // 210
    var content = {                                                                                                    // 211
      _id: key,                                                                                                        // 211
      used: false,                                                                                                     // 211
      note: note                                                                                                       // 211
    };                                                                                                                 // 211
    if (typeof quota === "number") content.quota = quota;                                                              // 212
    SignupKeys.insert(content);                                                                                        // 213
    return key;                                                                                                        // 214
  },                                                                                                                   // 215
  sendInvites: function (token, origin, from, list, subject, message, quota) {                                         // 217
    checkAuth(token);                                                                                                  // 218
    check(from, {                                                                                                      // 219
      name: String,                                                                                                    // 219
      address: String                                                                                                  // 219
    });                                                                                                                // 219
    check([origin, list, subject, message], [String]);                                                                 // 220
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 221
                                                                                                                       //
    if (!from.address.trim()) {                                                                                        // 223
      throw new Meteor.Error(403, "Must enter 'from' address.");                                                       // 224
    }                                                                                                                  // 225
                                                                                                                       //
    if (!list.trim()) {                                                                                                // 227
      throw new Meteor.Error(403, "Must enter 'to' addresses.");                                                       // 228
    }                                                                                                                  // 229
                                                                                                                       //
    this.unblock();                                                                                                    // 231
    list = list.split("\n");                                                                                           // 233
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(list)) {                                                      // 234
      var email = list[i].trim();                                                                                      // 235
                                                                                                                       //
      if (email) {                                                                                                     // 237
        var key = Random.id();                                                                                         // 238
        var content = {                                                                                                // 240
          _id: key,                                                                                                    // 241
          used: false,                                                                                                 // 242
          note: "E-mail invite to " + email,                                                                           // 243
          email: email,                                                                                                // 244
          definitelySent: false                                                                                        // 245
        };                                                                                                             // 240
        if (typeof quota === "number") content.quota = quota;                                                          // 247
        SignupKeys.insert(content);                                                                                    // 248
        sendEmail({                                                                                                    // 249
          to: email,                                                                                                   // 250
          from: from,                                                                                                  // 251
          envelopeFrom: globalDb.getReturnAddress(),                                                                   // 252
          subject: subject,                                                                                            // 253
          text: message.replace(/\$KEY/g, origin + "/signup/" + key)                                                   // 254
        });                                                                                                            // 249
        SignupKeys.update(key, {                                                                                       // 256
          $set: {                                                                                                      // 256
            definitelySent: true                                                                                       // 256
          }                                                                                                            // 256
        });                                                                                                            // 256
      }                                                                                                                // 257
    }                                                                                                                  // 258
                                                                                                                       //
    return {                                                                                                           // 260
      sent: true                                                                                                       // 260
    };                                                                                                                 // 260
  },                                                                                                                   // 261
  adminToggleDisableCap: function (token, capId, value) {                                                              // 263
    checkAuth(token);                                                                                                  // 264
    check(capId, String);                                                                                              // 265
    check(value, Boolean);                                                                                             // 266
                                                                                                                       //
    if (value) {                                                                                                       // 268
      ApiTokens.update({                                                                                               // 269
        _id: capId                                                                                                     // 269
      }, {                                                                                                             // 269
        $set: {                                                                                                        // 269
          revoked: true                                                                                                // 269
        }                                                                                                              // 269
      });                                                                                                              // 269
    } else {                                                                                                           // 270
      ApiTokens.update({                                                                                               // 271
        _id: capId                                                                                                     // 271
      }, {                                                                                                             // 271
        $set: {                                                                                                        // 271
          revoked: false                                                                                               // 271
        }                                                                                                              // 271
      });                                                                                                              // 271
    }                                                                                                                  // 272
  },                                                                                                                   // 273
  updateQuotas: function (token, list, quota) {                                                                        // 275
    checkAuth(token);                                                                                                  // 276
    check(list, String);                                                                                               // 277
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 278
                                                                                                                       //
    if (!list.trim()) {                                                                                                // 280
      throw new Meteor.Error(400, "Must enter addresses.");                                                            // 281
    }                                                                                                                  // 282
                                                                                                                       //
    var items = list.split("\n");                                                                                      // 284
    var invalid = [];                                                                                                  // 285
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(items)) {                                                     // 286
      var modifier = typeof quota === "number" ? {                                                                     // 287
        $set: {                                                                                                        // 287
          quota: quota                                                                                                 // 287
        }                                                                                                              // 287
      } : {                                                                                                            // 287
        $unset: {                                                                                                      // 288
          quota: ""                                                                                                    // 288
        }                                                                                                              // 288
      };                                                                                                               // 288
      var n = SignupKeys.update({                                                                                      // 289
        email: items[i]                                                                                                // 289
      }, modifier, {                                                                                                   // 289
        multi: true                                                                                                    // 289
      });                                                                                                              // 289
      n += Meteor.users.update({                                                                                       // 290
        signupEmail: items[i]                                                                                          // 290
      }, modifier, {                                                                                                   // 290
        multi: true                                                                                                    // 290
      });                                                                                                              // 290
      if (n < 1) invalid.push(items[i]);                                                                               // 292
    }                                                                                                                  // 293
                                                                                                                       //
    if (invalid.length > 0) {                                                                                          // 295
      throw new Meteor.Error(404, "These addresses did not map to any user nor invite: " + invalid.join(", "));        // 296
    }                                                                                                                  // 298
  },                                                                                                                   // 299
  dismissAdminStatsNotifications: function (token) {                                                                   // 301
    checkAuth(token);                                                                                                  // 302
    globalDb.collections.notifications.remove({                                                                        // 303
      "admin.type": "reportStats"                                                                                      // 303
    });                                                                                                                // 303
  },                                                                                                                   // 304
  signUpAsAdmin: function (token) {                                                                                    // 306
    check(token, String);                                                                                              // 307
    checkAuth(token);                                                                                                  // 308
                                                                                                                       //
    if (!this.userId) {                                                                                                // 309
      throw new Meteor.Error(403, "Must be logged in to sign up as admin.");                                           // 310
    }                                                                                                                  // 311
                                                                                                                       //
    if (!Meteor.user().loginCredentials) {                                                                             // 313
      throw new Meteor.Error(403, "Must be logged into an account to sign up as admin.");                              // 314
    }                                                                                                                  // 315
                                                                                                                       //
    Meteor.users.update({                                                                                              // 317
      _id: this.userId                                                                                                 // 317
    }, {                                                                                                               // 317
      $set: {                                                                                                          // 317
        isAdmin: true,                                                                                                 // 317
        signupKey: "admin"                                                                                             // 317
      }                                                                                                                // 317
    });                                                                                                                // 317
    clearAdminToken(token);                                                                                            // 318
  },                                                                                                                   // 319
  redeemSetupToken: function (token) {                                                                                 // 321
    // Redeem an admin token into a setup session.                                                                     // 322
    check(token, String);                                                                                              // 323
                                                                                                                       //
    if (tokenIsValid(token)) {                                                                                         // 324
      var sessId = Random.secret();                                                                                    // 325
      var creationDate = new Date();                                                                                   // 326
      var hashedSessionId = Crypto.createHash("sha256").update(sessId).digest("base64");                               // 327
      this.connection.sandstormDb.collections.setupSession.upsert({                                                    // 328
        _id: "current-session"                                                                                         // 329
      }, {                                                                                                             // 328
        creationDate: creationDate,                                                                                    // 331
        hashedSessionId: hashedSessionId                                                                               // 332
      }); // Then, invalidate the token, so one one else can use it.                                                   // 330
                                                                                                                       //
      clearAdminToken(token);                                                                                          // 335
      return sessId;                                                                                                   // 336
    } else {                                                                                                           // 337
      throw new Meteor.Error(401, "Invalid setup token");                                                              // 338
    }                                                                                                                  // 339
  },                                                                                                                   // 340
  heapdump: function () {                                                                                              // 342
    // Requests a heap dump. Intended for use by Sandstorm developers. Requires admin.                                 // 343
    //                                                                                                                 // 344
    // Call this from the JS console like:                                                                             // 345
    //   Meteor.call("heapdump");                                                                                      // 346
    checkAuth(); // We use /var/log because it's a location in the container to which the front-end is allowed         // 348
    // to write.                                                                                                       // 351
                                                                                                                       //
    var name = "/var/log/" + Date.now() + ".heapsnapshot";                                                             // 352
    Heapdump.writeSnapshot(name);                                                                                      // 353
    console.log("Wrote heapdump: /opt/sandstorm" + name);                                                              // 354
    return name;                                                                                                       // 355
  },                                                                                                                   // 356
  setPreinstalledApps: function (appAndPackageIds) {                                                                   // 358
    checkAuth();                                                                                                       // 359
    check(appAndPackageIds, [{                                                                                         // 360
      appId: String,                                                                                                   // 360
      packageId: String                                                                                                // 360
    }]);                                                                                                               // 360
    this.connection.sandstormDb.setPreinstalledApps(appAndPackageIds);                                                 // 362
  }                                                                                                                    // 363
});                                                                                                                    // 53
                                                                                                                       //
var authorizedAsAdmin = function (token, userId) {                                                                     // 366
  return Match.test(token, Match.OneOf(undefined, null, String)) && (userId && isAdminById(userId) || tokenIsValid(token) || tokenIsSetupSession(token));
};                                                                                                                     // 369
                                                                                                                       //
Meteor.publish("admin", function (token) {                                                                             // 371
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 372
  return Settings.find();                                                                                              // 373
});                                                                                                                    // 374
Meteor.publish("adminServiceConfiguration", function (token) {                                                         // 376
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 377
  return ServiceConfiguration.configurations.find();                                                                   // 378
});                                                                                                                    // 379
Meteor.publish("publicAdminSettings", function () {                                                                    // 381
  return Settings.find({                                                                                               // 382
    _id: {                                                                                                             // 382
      $in: publicAdminSettings                                                                                         // 382
    }                                                                                                                  // 382
  });                                                                                                                  // 382
});                                                                                                                    // 383
Meteor.publish("adminToken", function (token) {                                                                        // 385
  check(token, String);                                                                                                // 386
  this.added("adminToken", "adminToken", {                                                                             // 387
    tokenIsValid: tokenIsValid(token) || tokenIsSetupSession(token)                                                    // 387
  });                                                                                                                  // 387
  this.ready();                                                                                                        // 388
});                                                                                                                    // 389
Meteor.publish("allUsers", function (token) {                                                                          // 391
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 392
  return Meteor.users.find();                                                                                          // 393
});                                                                                                                    // 394
Meteor.publish("adminUserDetails", function (userId) {                                                                 // 396
  var _this2 = this;                                                                                                   // 396
                                                                                                                       //
  if (!authorizedAsAdmin(undefined, this.userId)) return []; // Reactive publish of any credentials owned by the account with id userId,
  // as well as that user object itself.                                                                               // 400
                                                                                                                       //
  var credentialSubs = {};                                                                                             // 401
  var accountId = userId;                                                                                              // 402
                                                                                                                       //
  var unrefCredential = function (credentialId) {                                                                      // 404
    if (!credentialSubs[credentialId]) {                                                                               // 405
      // should never happen, but if somehow you attempt to unref an credential that we don't have a                   // 406
      // subscription to, then don't crash                                                                             // 407
      console.error("attempted to unref untracked credential id:", credentialId);                                      // 408
      return;                                                                                                          // 409
    }                                                                                                                  // 410
                                                                                                                       //
    var observeHandle = credentialSubs[credentialId];                                                                  // 412
    delete credentialSubs[credentialId];                                                                               // 413
    observeHandle.stop();                                                                                              // 414
                                                                                                                       //
    _this2.removed("users", credentialId);                                                                             // 415
  };                                                                                                                   // 416
                                                                                                                       //
  var refCredential = function (credentialId) {                                                                        // 418
    if (credentialSubs[credentialId]) {                                                                                // 419
      // should never happen, but if somehow an account wound up with a duplicate credential ID,                       // 420
      // avoid leaking a subscription                                                                                  // 421
      console.error("duplicate credential id:", credentialId);                                                         // 422
      return;                                                                                                          // 423
    }                                                                                                                  // 424
                                                                                                                       //
    var cursor = Meteor.users.find({                                                                                   // 426
      _id: credentialId                                                                                                // 426
    });                                                                                                                // 426
    var observeHandle = cursor.observe({                                                                               // 427
      added: function (doc) {                                                                                          // 428
        _this2.added("users", doc._id, doc);                                                                           // 429
      },                                                                                                               // 430
      changed: function (newDoc, oldDoc) {                                                                             // 432
        fillUndefinedForChangedDoc(newDoc, oldDoc);                                                                    // 433
                                                                                                                       //
        _this2.changed("users", newDoc._id, newDoc);                                                                   // 434
      },                                                                                                               // 435
      removed: function (oldDoc) {                                                                                     // 437
        _this2.removed("users", oldDoc._id);                                                                           // 438
      }                                                                                                                // 439
    });                                                                                                                // 427
    credentialSubs[credentialId] = observeHandle;                                                                      // 442
  };                                                                                                                   // 443
                                                                                                                       //
  var accountCursor = Meteor.users.find({                                                                              // 445
    _id: accountId                                                                                                     // 445
  });                                                                                                                  // 445
  var accountSubHandle = accountCursor.observe({                                                                       // 446
    added: function (newDoc) {                                                                                         // 447
      var newCredentials = SandstormDb.getUserCredentialIds(newDoc);                                                   // 448
      newCredentials.forEach(function (credentialId) {                                                                 // 449
        refCredential(credentialId);                                                                                   // 450
      });                                                                                                              // 451
                                                                                                                       //
      _this2.added("users", newDoc._id, newDoc);                                                                       // 453
    },                                                                                                                 // 454
    changed: function (newDoc, oldDoc) {                                                                               // 456
      var newCredentials = SandstormDb.getUserCredentialIds(newDoc);                                                   // 457
      var oldCredentials = SandstormDb.getUserCredentialIds(oldDoc); // Those in newDoc - oldDoc, ref.                 // 458
                                                                                                                       //
      var credentialsAdded = _.difference(newCredentials, oldCredentials);                                             // 461
                                                                                                                       //
      credentialsAdded.forEach(function (credentialId) {                                                               // 462
        refCredential(credentialId);                                                                                   // 463
      }); // Those in oldDoc - newDoc, unref.                                                                          // 464
                                                                                                                       //
      var credentialsRemoved = _.difference(oldCredentials, newCredentials);                                           // 467
                                                                                                                       //
      credentialsRemoved.forEach(function (credentialId) {                                                             // 468
        unrefCredential(credentialId);                                                                                 // 469
      });                                                                                                              // 470
      fillUndefinedForChangedDoc(newDoc, oldDoc);                                                                      // 472
                                                                                                                       //
      _this2.changed("users", newDoc._id, newDoc);                                                                     // 474
    },                                                                                                                 // 475
    removed: function (oldDoc) {                                                                                       // 477
      _this2.removed("users", oldDoc._id);                                                                             // 478
                                                                                                                       //
      var oldCredentials = SandstormDb.getUserCredentialIds(oldDoc);                                                   // 479
      oldCredentials.forEach(function (credentialId) {                                                                 // 480
        unrefCredential(credentialId);                                                                                 // 481
      });                                                                                                              // 482
    }                                                                                                                  // 483
  });                                                                                                                  // 446
  this.onStop(function () {                                                                                            // 486
    accountSubHandle.stop(); // Also stop all the credential subscriptions.                                            // 487
                                                                                                                       //
    var subs = _.values(credentialSubs);                                                                               // 489
                                                                                                                       //
    subs.forEach(function (sub) {                                                                                      // 490
      sub.stop();                                                                                                      // 491
    });                                                                                                                // 492
  }); // Meteor's cursor.observe() will synchronously call all of the added() callbacks from the initial               // 493
  // query, so by the time we get here we can report readiness.                                                        // 496
                                                                                                                       //
  this.ready();                                                                                                        // 497
});                                                                                                                    // 498
Meteor.publish("activityStats", function (token) {                                                                     // 500
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 501
  return ActivityStats.find();                                                                                         // 502
});                                                                                                                    // 503
Meteor.publish("statsTokens", function (token) {                                                                       // 505
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 506
  return StatsTokens.find();                                                                                           // 507
});                                                                                                                    // 508
Meteor.publish("allPackages", function (token) {                                                                       // 510
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 511
  return Packages.find({                                                                                               // 512
    manifest: {                                                                                                        // 512
      $exists: true                                                                                                    // 512
    }                                                                                                                  // 512
  }, {                                                                                                                 // 512
    fields: {                                                                                                          // 513
      appId: 1,                                                                                                        // 513
      "manifest.appVersion": 1,                                                                                        // 513
      "manifest.actions": 1,                                                                                           // 514
      "manifest.appTitle": 1,                                                                                          // 514
      progress: 1,                                                                                                     // 514
      status: 1                                                                                                        // 514
    }                                                                                                                  // 513
  });                                                                                                                  // 513
});                                                                                                                    // 515
Meteor.publish("realTimeStats", function (token) {                                                                     // 517
  if (!authorizedAsAdmin(token, this.userId)) return []; // Last five minutes.                                         // 518
                                                                                                                       //
  this.added("realTimeStats", "now", computeStats(new Date(Date.now() - 5 * 60 * 1000))); // Since last sample.        // 521
                                                                                                                       //
  var lastSample = ActivityStats.findOne({}, {                                                                         // 524
    sort: {                                                                                                            // 524
      timestamp: -1                                                                                                    // 524
    }                                                                                                                  // 524
  });                                                                                                                  // 524
  var lastSampleTime = lastSample ? lastSample.timestamp : new Date(0);                                                // 525
  this.added("realTimeStats", "today", computeStats(lastSampleTime)); // TODO(someday): Update every few minutes?      // 526
                                                                                                                       //
  this.ready();                                                                                                        // 530
});                                                                                                                    // 531
Meteor.publish("adminLog", function (token) {                                                                          // 533
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 534
  var logfile = SANDSTORM_LOGDIR + "/sandstorm.log";                                                                   // 536
  var fd = Fs.openSync(logfile, "r");                                                                                  // 538
  var startSize = Fs.fstatSync(fd).size; // Difference between the current file offset and the subscription offset. Can be non-zero when
  // logs have rotated.                                                                                                // 542
                                                                                                                       //
  var extraOffset = 0;                                                                                                 // 543
                                                                                                                       //
  if (startSize < 8192) {                                                                                              // 545
    // Log size is less than window size. Check for rotated log and grab its tail.                                     // 546
    var logfile1 = SANDSTORM_LOGDIR + "/sandstorm.log.1";                                                              // 547
                                                                                                                       //
    if (Fs.existsSync(logfile1)) {                                                                                     // 548
      var fd1 = Fs.openSync(logfile1, "r");                                                                            // 549
      var startSize1 = Fs.fstatSync(fd1).size;                                                                         // 550
      var amountFromLog1 = Math.min(startSize1, 8192 - startSize);                                                     // 551
      var offset1 = startSize1 - amountFromLog1;                                                                       // 552
      var buf = new Buffer(amountFromLog1);                                                                            // 553
      var n = Fs.readSync(fd1, buf, 0, buf.length, offset);                                                            // 554
                                                                                                                       //
      if (n > 0) {                                                                                                     // 555
        this.added("adminLog", 0, {                                                                                    // 556
          text: buf.toString("utf8", 0, n)                                                                             // 556
        });                                                                                                            // 556
        extraOffset += n;                                                                                              // 557
      }                                                                                                                // 558
    }                                                                                                                  // 559
  } // Start tailing at EOF - 8k.                                                                                      // 560
                                                                                                                       //
                                                                                                                       //
  var offset = Math.max(0, startSize - 8192);                                                                          // 563
                                                                                                                       //
  var _this = this;                                                                                                    // 565
                                                                                                                       //
  function doTail() {                                                                                                  // 566
    if (Fs.fstatSync(fd).size < offset) {                                                                              // 567
      extraOffset += offset;                                                                                           // 568
      offset = 0;                                                                                                      // 569
    }                                                                                                                  // 570
                                                                                                                       //
    for (;;) {                                                                                                         // 572
      var _buf = new Buffer(Math.max(1024, startSize - offset));                                                       // 573
                                                                                                                       //
      var _n = Fs.readSync(fd, _buf, 0, _buf.length, offset);                                                          // 574
                                                                                                                       //
      if (_n <= 0) break;                                                                                              // 575
                                                                                                                       //
      _this.added("adminLog", offset + extraOffset, {                                                                  // 576
        text: _buf.toString("utf8", 0, _n)                                                                             // 576
      });                                                                                                              // 576
                                                                                                                       //
      offset += _n;                                                                                                    // 577
    }                                                                                                                  // 578
  } // Watch the file for changes.                                                                                     // 579
                                                                                                                       //
                                                                                                                       //
  var watcher = Fs.watch(logfile, {                                                                                    // 582
    persistent: false                                                                                                  // 582
  }, Meteor.bindEnvironment(doTail)); // When the subscription stops, stop watching the file.                          // 582
                                                                                                                       //
  this.onStop(function () {                                                                                            // 585
    watcher.close();                                                                                                   // 586
    Fs.closeSync(fd);                                                                                                  // 587
  }); // Read initial 8k tail data immediately.                                                                        // 588
                                                                                                                       //
  doTail(); // Notify ready.                                                                                           // 591
                                                                                                                       //
  this.ready();                                                                                                        // 594
});                                                                                                                    // 595
Meteor.publish("adminApiTokens", function (token) {                                                                    // 597
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 598
  return ApiTokens.find({                                                                                              // 599
    $or: [{                                                                                                            // 600
      "frontendRef.ipNetwork": {                                                                                       // 601
        $exists: true                                                                                                  // 601
      }                                                                                                                // 601
    }, {                                                                                                               // 601
      "frontendRef.ipInterface": {                                                                                     // 602
        $exists: true                                                                                                  // 602
      }                                                                                                                // 602
    }]                                                                                                                 // 602
  }, {                                                                                                                 // 599
    fields: {                                                                                                          // 605
      frontendRef: 1,                                                                                                  // 606
      created: 1,                                                                                                      // 607
      requirements: 1,                                                                                                 // 608
      revoked: 1,                                                                                                      // 609
      owner: 1                                                                                                         // 610
    }                                                                                                                  // 605
  });                                                                                                                  // 604
});                                                                                                                    // 613
Meteor.publish("hasAdmin", function (token) {                                                                          // 615
  var _this3 = this;                                                                                                   // 615
                                                                                                                       //
  // Like hasUsers, but for admins, and with token auth required.                                                      // 616
  if (!authorizedAsAdmin(token, this.userId)) return []; // Query if there are any admin users.                        // 617
                                                                                                                       //
  var cursor = Meteor.users.find({                                                                                     // 620
    isAdmin: true                                                                                                      // 620
  });                                                                                                                  // 620
                                                                                                                       //
  if (cursor.count() > 0) {                                                                                            // 621
    this.added("hasAdmin", "hasAdmin", {                                                                               // 622
      hasAdmin: true                                                                                                   // 622
    });                                                                                                                // 622
  } else {                                                                                                             // 623
    var handle = cursor.observeChanges({                                                                               // 624
      added: function (id) {                                                                                           // 625
        _this3.added("hasAdmin", "hasAdmin", {                                                                         // 626
          hasAdmin: true                                                                                               // 626
        });                                                                                                            // 626
                                                                                                                       //
        handle.stop();                                                                                                 // 627
        handle = null;                                                                                                 // 628
      }                                                                                                                // 629
    });                                                                                                                // 624
    this.onStop(function () {                                                                                          // 631
      if (handle) handle.stop();                                                                                       // 632
    });                                                                                                                // 633
  }                                                                                                                    // 634
                                                                                                                       //
  this.ready();                                                                                                        // 636
});                                                                                                                    // 637
Meteor.publish("appIndexAdmin", function (token) {                                                                     // 639
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 640
  return globalDb.collections.appIndex.find();                                                                         // 641
});                                                                                                                    // 642
                                                                                                                       //
function observeOauthService(name) {                                                                                   // 644
  Settings.find({                                                                                                      // 645
    _id: name,                                                                                                         // 645
    value: true                                                                                                        // 645
  }).observe({                                                                                                         // 645
    added: function () {                                                                                               // 646
      // Tell the oauth library it should accept login attempts from this service.                                     // 647
      Accounts.oauth.registerService(name);                                                                            // 648
    },                                                                                                                 // 649
    removed: function () {                                                                                             // 651
      // Tell the oauth library it should deny login attempts from this service.                                       // 652
      Accounts.oauth.unregisterService(name);                                                                          // 653
    }                                                                                                                  // 654
  });                                                                                                                  // 645
}                                                                                                                      // 656
                                                                                                                       //
observeOauthService("github");                                                                                         // 658
observeOauthService("google");                                                                                         // 659
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"backup.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/backup.js                                                                                                    //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var ChildProcess = Npm.require("child_process");                                                                       // 19
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 20
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 21
                                                                                                                       //
var GrainInfo = Capnp.importSystem("sandstorm/grain.capnp").GrainInfo;                                                 // 23
var TOKEN_CLEANUP_MINUTES = 120; // Give enough time for large uploads on slow connections.                            // 25
                                                                                                                       //
var TOKEN_CLEANUP_TIMER = TOKEN_CLEANUP_MINUTES * 60 * 1000;                                                           // 26
                                                                                                                       //
function cleanupToken(tokenId) {                                                                                       // 28
  check(tokenId, String);                                                                                              // 29
  FileTokens.remove({                                                                                                  // 30
    _id: tokenId                                                                                                       // 30
  });                                                                                                                  // 30
  waitPromise(globalBackend.cap().deleteBackup(tokenId));                                                              // 31
}                                                                                                                      // 32
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 34
  // Cleanup tokens every TOKEN_CLEANUP_MINUTES                                                                        // 35
  SandstormDb.periodicCleanup(TOKEN_CLEANUP_TIMER, function () {                                                       // 36
    var queryDate = new Date(Date.now() - TOKEN_CLEANUP_TIMER);                                                        // 37
    FileTokens.find({                                                                                                  // 39
      timestamp: {                                                                                                     // 39
        $lt: queryDate                                                                                                 // 39
      }                                                                                                                // 39
    }).forEach(function (token) {                                                                                      // 39
      cleanupToken(token._id);                                                                                         // 40
    });                                                                                                                // 41
  });                                                                                                                  // 42
});                                                                                                                    // 43
Meteor.methods({                                                                                                       // 45
  backupGrain: function (grainId) {                                                                                    // 46
    check(grainId, String);                                                                                            // 47
    var grain = Grains.findOne(grainId);                                                                               // 48
                                                                                                                       //
    if (!grain || !this.userId || grain.userId !== this.userId) {                                                      // 49
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");                              // 50
    }                                                                                                                  // 51
                                                                                                                       //
    this.unblock();                                                                                                    // 53
    var token = {                                                                                                      // 55
      _id: Random.id(),                                                                                                // 56
      timestamp: new Date(),                                                                                           // 57
      name: grain.title                                                                                                // 58
    }; // TODO(soon): does the grain need to be offline?                                                               // 55
                                                                                                                       //
    var grainInfo = _.pick(grain, "appId", "appVersion", "title");                                                     // 63
                                                                                                                       //
    grainInfo.ownerIdentityId = grain.identityId;                                                                      // 64
    FileTokens.insert(token);                                                                                          // 66
    waitPromise(globalBackend.cap().backupGrain(token._id, this.userId, grainId, grainInfo));                          // 67
    return token._id;                                                                                                  // 69
  },                                                                                                                   // 70
  newRestoreToken: function () {                                                                                       // 72
    if (!isSignedUpOrDemo()) {                                                                                         // 73
      throw new Meteor.Error(403, "Unauthorized", "Only invited users can restore backups.");                          // 74
    }                                                                                                                  // 75
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 77
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 78
    }                                                                                                                  // 80
                                                                                                                       //
    var token = {                                                                                                      // 82
      _id: Random.id(),                                                                                                // 83
      timestamp: new Date()                                                                                            // 84
    };                                                                                                                 // 82
    FileTokens.insert(token);                                                                                          // 87
    return token._id;                                                                                                  // 89
  },                                                                                                                   // 90
  restoreGrain: function (tokenId, obsolete) {                                                                         // 92
    check(tokenId, String);                                                                                            // 93
    var token = FileTokens.findOne(tokenId);                                                                           // 94
                                                                                                                       //
    if (!token || !isSignedUpOrDemo()) {                                                                               // 95
      throw new Meteor.Error(403, "Unauthorized", "Token was not found, or user cannot create grains");                // 96
    }                                                                                                                  // 98
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 100
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 101
    }                                                                                                                  // 103
                                                                                                                       //
    this.unblock();                                                                                                    // 105
    var grainId = Random.id(22);                                                                                       // 107
                                                                                                                       //
    try {                                                                                                              // 109
      var grainInfo = waitPromise(globalBackend.cap().restoreGrain(tokenId, this.userId, grainId).catch(function (err) {
        console.error("Unzip failure:", err.message);                                                                  // 112
        throw new Meteor.Error(500, "Invalid backup file.");                                                           // 113
      })).info;                                                                                                        // 114
                                                                                                                       //
      if (!grainInfo.appId) {                                                                                          // 115
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 116
        throw new Meteor.Error(500, "Metadata object for uploaded grain has no AppId");                                // 117
      }                                                                                                                // 118
                                                                                                                       //
      var action = UserActions.findOne({                                                                               // 120
        appId: grainInfo.appId,                                                                                        // 120
        userId: this.userId                                                                                            // 120
      }); // Create variables we'll use for later Mongo query.                                                         // 120
                                                                                                                       //
      var packageId = void 0;                                                                                          // 123
      var appVersion = void 0; // DevPackages are system-wide, so we do not check the user ID.                         // 124
                                                                                                                       //
      var devPackage = DevPackages.findOne({                                                                           // 127
        appId: grainInfo.appId                                                                                         // 127
      });                                                                                                              // 127
                                                                                                                       //
      if (devPackage) {                                                                                                // 128
        // If the dev app package exists, it should override the user action.                                          // 129
        packageId = devPackage.packageId;                                                                              // 130
        appVersion = devPackage.manifest.appVersion;                                                                   // 131
      } else if (action) {                                                                                             // 132
        // The app is installed, so we can continue restoring this                                                     // 133
        // grain.                                                                                                      // 134
        packageId = action.packageId;                                                                                  // 135
        appVersion = action.appVersion;                                                                                // 136
      } else {                                                                                                         // 137
        // If the package isn't installed at all, bail out.                                                            // 138
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 139
        throw new Meteor.Error(500, "App id for uploaded grain not installed", "App Id: " + grainInfo.appId);          // 140
      }                                                                                                                // 143
                                                                                                                       //
      if (appVersion < grainInfo.appVersion) {                                                                         // 145
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 146
        throw new Meteor.Error(500, "App version for uploaded grain is newer than any " + "installed version. You need to upgrade your app first", "New version: " + grainInfo.appVersion + ", Old version: " + appVersion);
      }                                                                                                                // 152
                                                                                                                       //
      console.log(grainInfo);                                                                                          // 154
      Grains.insert({                                                                                                  // 156
        _id: grainId,                                                                                                  // 157
        packageId: packageId,                                                                                          // 158
        appId: grainInfo.appId,                                                                                        // 159
        appVersion: appVersion,                                                                                        // 160
        userId: this.userId,                                                                                           // 161
        // For older backups that don't have the owner's identity ID, use the owner's identicon                        // 162
        // ID which is based on old-style global identity IDs.                                                         // 163
        identityId: grainInfo.ownerIdentityId || (Meteor.user().profile || {}).identicon || SandstormDb.generateIdentityId(),
        title: grainInfo.title,                                                                                        // 167
        "private": true,                                                                                               // 168
        size: 0                                                                                                        // 169
      });                                                                                                              // 156
    } finally {                                                                                                        // 171
      cleanupToken(tokenId);                                                                                           // 172
    }                                                                                                                  // 173
                                                                                                                       //
    return grainId;                                                                                                    // 175
  }                                                                                                                    // 176
});                                                                                                                    // 45
Router.map(function () {                                                                                               // 179
  this.route("downloadBackup", {                                                                                       // 180
    where: "server",                                                                                                   // 181
    path: "/downloadBackup/:tokenId",                                                                                  // 182
    action: function () {                                                                                              // 183
      var token = FileTokens.findOne(this.params.tokenId);                                                             // 184
      var response = this.response;                                                                                    // 185
                                                                                                                       //
      if (!token) {                                                                                                    // 186
        response.writeHead(404, {                                                                                      // 187
          "Content-Type": "text/plain"                                                                                 // 187
        });                                                                                                            // 187
        return response.end("File does not exist");                                                                    // 188
      }                                                                                                                // 189
                                                                                                                       //
      var started = false;                                                                                             // 191
      var encodedFilename = encodeURIComponent(token.name || "backup") + ".zip";                                       // 192
      var sawEnd = false;                                                                                              // 193
      var stream = {                                                                                                   // 195
        expectSize: function (size) {                                                                                  // 196
          if (!started) {                                                                                              // 197
            started = true;                                                                                            // 198
            response.writeHead(200, {                                                                                  // 199
              "Content-Length": size,                                                                                  // 200
              "Content-Type": "application/zip",                                                                       // 201
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 202
            });                                                                                                        // 199
          }                                                                                                            // 204
        },                                                                                                             // 205
        write: function (data) {                                                                                       // 207
          if (!started) {                                                                                              // 208
            started = true;                                                                                            // 209
            response.writeHead(200, {                                                                                  // 210
              "Content-Type": "application/zip",                                                                       // 211
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 212
            });                                                                                                        // 210
          }                                                                                                            // 214
                                                                                                                       //
          response.write(data);                                                                                        // 216
        },                                                                                                             // 217
        done: function (data) {                                                                                        // 219
          if (!started) {                                                                                              // 220
            started = true;                                                                                            // 221
            response.writeHead(200, {                                                                                  // 222
              "Content-Length": 0,                                                                                     // 223
              "Content-Type": "application/zip",                                                                       // 224
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 225
            });                                                                                                        // 222
          }                                                                                                            // 227
                                                                                                                       //
          sawEnd = true;                                                                                               // 229
          response.end();                                                                                              // 230
        }                                                                                                              // 231
      };                                                                                                               // 195
      waitPromise(globalBackend.cap().downloadBackup(this.params.tokenId, stream));                                    // 234
                                                                                                                       //
      if (!sawEnd) {                                                                                                   // 236
        console.error("backend failed to call done() when downloading backup");                                        // 237
                                                                                                                       //
        if (!started) {                                                                                                // 238
          throw new Meteor.Error(500, "backend failed to produce data");                                               // 239
        }                                                                                                              // 240
                                                                                                                       //
        response.end();                                                                                                // 242
      }                                                                                                                // 243
                                                                                                                       //
      cleanupToken(this.params.tokenId);                                                                               // 245
    }                                                                                                                  // 246
  });                                                                                                                  // 180
  this.route("uploadBackup", {                                                                                         // 249
    where: "server",                                                                                                   // 250
    path: "/uploadBackup/:token",                                                                                      // 251
    action: function () {                                                                                              // 252
      if (this.request.method === "POST") {                                                                            // 253
        var token = FileTokens.findOne(this.params.token);                                                             // 254
                                                                                                                       //
        if (!this.params.token || !token) {                                                                            // 255
          this.response.writeHead(403, {                                                                               // 256
            "Content-Type": "text/plain",                                                                              // 257
            "Access-Control-Allow-Origin": "*"                                                                         // 258
          });                                                                                                          // 256
          this.response.write("Invalid upload token.");                                                                // 260
          this.response.end();                                                                                         // 261
          return;                                                                                                      // 262
        }                                                                                                              // 263
                                                                                                                       //
        var request = this.request;                                                                                    // 265
                                                                                                                       //
        try {                                                                                                          // 266
          var stream = globalBackend.cap().uploadBackup(this.params.token).stream;                                     // 267
          waitPromise(new Promise(function (resolve, reject) {                                                         // 269
            request.on("data", function (data) {                                                                       // 270
              stream.write(data);                                                                                      // 271
            });                                                                                                        // 272
            request.on("end", function () {                                                                            // 273
              resolve(stream.done());                                                                                  // 274
            });                                                                                                        // 275
            request.on("error", function (err) {                                                                       // 276
              stream.close();                                                                                          // 277
            });                                                                                                        // 278
          }));                                                                                                         // 279
          this.response.writeHead(204, {                                                                               // 281
            "Access-Control-Allow-Origin": "*"                                                                         // 282
          });                                                                                                          // 281
          this.response.end();                                                                                         // 284
        } catch (error) {                                                                                              // 285
          console.error(error.stack);                                                                                  // 286
          this.response.writeHead(500, {                                                                               // 287
            "Content-Type": "text/plain",                                                                              // 288
            "Access-Control-Allow-Origin": "*"                                                                         // 289
          });                                                                                                          // 287
          this.response.write(error.stack);                                                                            // 291
          this.response.end();                                                                                         // 292
        }                                                                                                              // 293
      } else if (this.request.method == "OPTIONS") {                                                                   // 294
        // Allow cross-origin posts to uploadBackup so that uploads can occur on the DDP host                          // 295
        // rather than the main host. In theory we could have Access-Control-Allow-Origin specify                      // 296
        // the main host rather than "*", but an uploadBackup request already requires a valid                         // 297
        // upload token, which is plenty of access control in itself.                                                  // 298
        var requestedHeaders = this.request.headers["access-control-request-headers"];                                 // 299
                                                                                                                       //
        if (requestedHeaders) {                                                                                        // 300
          this.response.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                   // 301
        }                                                                                                              // 302
                                                                                                                       //
        this.response.writeHead(204, {                                                                                 // 304
          "Access-Control-Allow-Origin": "*",                                                                          // 305
          "Access-Control-Allow-Methods": "POST, OPTIONS",                                                             // 306
          "Access-Control-Max-Age": "3600"                                                                             // 307
        });                                                                                                            // 304
        this.response.end();                                                                                           // 309
      } else {                                                                                                         // 310
        this.response.writeHead(405, {                                                                                 // 311
          "Content-Type": "text/plain"                                                                                 // 312
        });                                                                                                            // 311
        this.response.write("You can only POST here.");                                                                // 314
        this.response.end();                                                                                           // 315
      }                                                                                                                // 316
    }                                                                                                                  // 317
  });                                                                                                                  // 249
});                                                                                                                    // 319
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"contacts-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/contacts-server.js                                                                                           //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.publish("contactProfiles", function (showAll) {                                                                 // 17
  var db = this.connection.sandstormDb;                                                                                // 18
                                                                                                                       //
  var _this = this;                                                                                                    // 19
                                                                                                                       //
  var userId = this.userId; // We maintain a map from account IDs to live query handles that track profile changes.    // 20
                                                                                                                       //
  var contactAccounts = {};                                                                                            // 23
  var disallowGuests = db.getOrganizationDisallowGuests();                                                             // 24
                                                                                                                       //
  function addAccountOfContact(accountId) {                                                                            // 26
    if (!(accountId in contactAccounts)) {                                                                             // 27
      var user = Meteor.users.findOne({                                                                                // 28
        _id: accountId                                                                                                 // 28
      });                                                                                                              // 28
                                                                                                                       //
      if (disallowGuests && !showAll) {                                                                                // 30
        if (!db.isUserInOrganization(user)) {                                                                          // 31
          return;                                                                                                      // 32
        }                                                                                                              // 33
      }                                                                                                                // 34
                                                                                                                       //
      if (user) {                                                                                                      // 36
        var filteredUser = _.pick(user, "_id", "profile");                                                             // 37
                                                                                                                       //
        filteredUser.intrinsicNames = db.getAccountIntrinsicNames(user, false);                                        // 38
                                                                                                                       //
        _this.added("contactProfiles", user._id, filteredUser);                                                        // 39
      }                                                                                                                // 40
                                                                                                                       //
      contactAccounts[accountId] = Meteor.users.find({                                                                 // 42
        _id: accountId                                                                                                 // 43
      }, {                                                                                                             // 43
        fields: {                                                                                                      // 43
          profile: 1                                                                                                   // 43
        }                                                                                                              // 43
      }).observeChanges({                                                                                              // 43
        changed: function (id, fields) {                                                                               // 44
          _this.changed("contactProfiles", id, fields);                                                                // 45
        }                                                                                                              // 46
      });                                                                                                              // 43
    }                                                                                                                  // 48
  }                                                                                                                    // 49
                                                                                                                       //
  var cursor = db.collections.contacts.find({                                                                          // 51
    ownerId: userId                                                                                                    // 51
  });                                                                                                                  // 51
  var handle = cursor.observe({                                                                                        // 53
    added: function (contact) {                                                                                        // 54
      addAccountOfContact(contact.accountId);                                                                          // 55
    },                                                                                                                 // 56
    changed: function (contact) {                                                                                      // 58
      addAccountOfContact(contact.accountId);                                                                          // 59
    },                                                                                                                 // 60
    removed: function (contact) {                                                                                      // 62
      _this.removed("contactProfiles", contact.accountId);                                                             // 63
                                                                                                                       //
      var contactAccount = contactAccounts[contact.accountId];                                                         // 64
      if (contactAccount) contactAccounts[contact.accountId].stop();                                                   // 65
      delete contactAccounts[contact.accountId];                                                                       // 66
    }                                                                                                                  // 67
  });                                                                                                                  // 53
  var orgHandle = void 0;                                                                                              // 70
                                                                                                                       //
  if (db.getOrganizationShareContacts() && db.isUserInOrganization(db.collections.users.findOne({                      // 72
    _id: userId                                                                                                        // 73
  }))) {                                                                                                               // 73
    var orgCursor = db.collections.users.find({                                                                        // 74
      type: "account"                                                                                                  // 74
    }); // TODO(perf): make a mongo query that can find all accounts in an organization and add                        // 74
    // indices for it. Currently, we do some case insensitive matching which mongo can't                               // 76
    // handle well.                                                                                                    // 77
                                                                                                                       //
    orgHandle = orgCursor.observe({                                                                                    // 79
      added: function (user) {                                                                                         // 80
        if (db.isUserInOrganization(user) && user._id !== userId) {                                                    // 81
          addAccountOfContact(user._id);                                                                               // 82
        }                                                                                                              // 83
      },                                                                                                               // 84
      changed: function (user) {                                                                                       // 86
        if (db.isUserInOrganization(user) && user._id !== userId) {                                                    // 87
          addAccountOfContact(user._id);                                                                               // 88
        }                                                                                                              // 89
      },                                                                                                               // 90
      removed: function (user) {                                                                                       // 92
        if (db.isUserInOrganization(user) && user._id !== userId) {                                                    // 93
          _this.removed("contactProfiles", user._id);                                                                  // 94
                                                                                                                       //
          var contactAccount = contactAccounts[contact.accountId];                                                     // 95
          if (contactAccount) contactAccounts[contact.accountId].stop();                                               // 96
          delete contactAccounts[user._id];                                                                            // 97
        }                                                                                                              // 98
      }                                                                                                                // 99
    });                                                                                                                // 79
  }                                                                                                                    // 101
                                                                                                                       //
  this.ready();                                                                                                        // 103
  this.onStop(function () {                                                                                            // 105
    handle.stop();                                                                                                     // 106
                                                                                                                       //
    if (orgHandle) {                                                                                                   // 107
      orgHandle.stop();                                                                                                // 108
    }                                                                                                                  // 109
                                                                                                                       //
    Object.keys(contactAccounts).forEach(function (accountId) {                                                        // 111
      contactAccounts[accountId].stop();                                                                               // 112
      delete contactAccounts[accountId];                                                                               // 113
    });                                                                                                                // 114
  });                                                                                                                  // 115
});                                                                                                                    // 116
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"core.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/core.js                                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var StaticAssetImpl = void 0,                                                                                          // 1
    IdenticonStaticAssetImpl = void 0;                                                                                 // 1
module.watch(require("/imports/server/static-asset.js"), {                                                             // 1
  StaticAssetImpl: function (v) {                                                                                      // 1
    StaticAssetImpl = v;                                                                                               // 1
  },                                                                                                                   // 1
  IdenticonStaticAssetImpl: function (v) {                                                                             // 1
    IdenticonStaticAssetImpl = v;                                                                                      // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var PersistentImpl = void 0,                                                                                           // 1
    hashSturdyRef = void 0,                                                                                            // 1
    generateSturdyRef = void 0,                                                                                        // 1
    checkRequirements = void 0,                                                                                        // 1
    fetchApiToken = void 0,                                                                                            // 1
    insertApiToken = void 0;                                                                                           // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  },                                                                                                                   // 1
  hashSturdyRef: function (v) {                                                                                        // 1
    hashSturdyRef = v;                                                                                                 // 1
  },                                                                                                                   // 1
  generateSturdyRef: function (v) {                                                                                    // 1
    generateSturdyRef = v;                                                                                             // 1
  },                                                                                                                   // 1
  checkRequirements: function (v) {                                                                                    // 1
    checkRequirements = v;                                                                                             // 1
  },                                                                                                                   // 1
  fetchApiToken: function (v) {                                                                                        // 1
    fetchApiToken = v;                                                                                                 // 1
  },                                                                                                                   // 1
  insertApiToken: function (v) {                                                                                       // 1
    insertApiToken = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 19
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 20
                                                                                                                       //
var PersistentHandle = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentHandle;                              // 24
var SandstormCore = Capnp.importSystem("sandstorm/supervisor.capnp").SandstormCore;                                    // 25
var SandstormCoreFactory = Capnp.importSystem("sandstorm/backend.capnp").SandstormCoreFactory;                         // 26
var PersistentOngoingNotification = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentOngoingNotification;    // 27
var PersistentUiView = Capnp.importSystem("sandstorm/persistentuiview.capnp").PersistentUiView;                        // 28
var StaticAsset = Capnp.importSystem("sandstorm/util.capnp").StaticAsset;                                              // 29
var SystemPersistent = Capnp.importSystem("sandstorm/supervisor.capnp").SystemPersistent;                              // 30
                                                                                                                       //
var SandstormCoreImpl = function () {                                                                                  //
  function SandstormCoreImpl(db, grainId) {                                                                            // 33
    (0, _classCallCheck3.default)(this, SandstormCoreImpl);                                                            // 33
    this.db = db;                                                                                                      // 34
    this.grainId = grainId;                                                                                            // 35
  }                                                                                                                    // 36
                                                                                                                       //
  SandstormCoreImpl.prototype.restore = function () {                                                                  //
    function restore(sturdyRef) {                                                                                      //
      var _this = this;                                                                                                // 38
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 39
        sturdyRef = sturdyRef.toString("utf8");                                                                        // 40
        var token = fetchApiToken(_this.db, sturdyRef, {                                                               // 41
          "owner.grain.grainId": _this.grainId                                                                         // 42
        });                                                                                                            // 42
                                                                                                                       //
        if (!token) {                                                                                                  // 43
          throw new Error("no such token");                                                                            // 44
        }                                                                                                              // 45
                                                                                                                       //
        return restoreInternal(_this.db, sturdyRef, {                                                                  // 47
          grain: Match.ObjectIncluding({                                                                               // 48
            grainId: _this.grainId                                                                                     // 48
          })                                                                                                           // 48
        }, [], token);                                                                                                 // 48
      });                                                                                                              // 50
    }                                                                                                                  // 51
                                                                                                                       //
    return restore;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.drop = function () {                                                                     //
    function drop(sturdyRef) {                                                                                         //
      var _this2 = this;                                                                                               // 53
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 54
        sturdyRef = sturdyRef.toString("utf8");                                                                        // 55
        return dropInternal(_this2.db, sturdyRef, {                                                                    // 56
          grain: Match.ObjectIncluding({                                                                               // 56
            grainId: _this2.grainId                                                                                    // 56
          })                                                                                                           // 56
        });                                                                                                            // 56
      });                                                                                                              // 57
    }                                                                                                                  // 58
                                                                                                                       //
    return drop;                                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.makeToken = function () {                                                                //
    function makeToken(ref, owner, requirements) {                                                                     //
      var _this3 = this;                                                                                               // 60
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 61
        var sturdyRef = insertApiToken(_this3.db, {                                                                    // 62
          grainId: _this3.grainId,                                                                                     // 63
          objectId: ref,                                                                                               // 64
          owner: owner,                                                                                                // 65
          created: new Date(),                                                                                         // 66
          requirements: requirements                                                                                   // 67
        });                                                                                                            // 62
        return {                                                                                                       // 70
          token: new Buffer(sturdyRef)                                                                                 // 71
        };                                                                                                             // 70
      });                                                                                                              // 73
    }                                                                                                                  // 74
                                                                                                                       //
    return makeToken;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.makeChildToken = function () {                                                           //
    function makeChildToken(parent, owner, requirements) {                                                             //
      var _this4 = this;                                                                                               // 76
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 77
        // Compute the save ApiToken template.                                                                         // 78
        return makeSaveTemplateForChild(_this4.db, parent.toString(), requirements);                                   // 79
      }).then(function (saveTemplate) {                                                                                // 80
        // Create a dummy PersistentImpl and invoke its own save() method.                                             // 81
        return new PersistentImpl(_this4.db, saveTemplate).save({                                                      // 82
          sealFor: owner                                                                                               // 82
        });                                                                                                            // 82
      }).then(function (saveResult) {                                                                                  // 83
        // Transform to expected result structure.                                                                     // 84
        return {                                                                                                       // 85
          token: saveResult.sturdyRef                                                                                  // 85
        };                                                                                                             // 85
      });                                                                                                              // 86
    }                                                                                                                  // 87
                                                                                                                       //
    return makeChildToken;                                                                                             //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.getOwnerNotificationTarget = function () {                                               //
    function getOwnerNotificationTarget() {                                                                            //
      var _this5 = this;                                                                                               // 89
                                                                                                                       //
      var grainId = this.grainId;                                                                                      // 90
      return {                                                                                                         // 91
        owner: {                                                                                                       // 92
          addOngoing: function (displayInfo, notification) {                                                           // 93
            return inMeteor(function () {                                                                              // 94
              var grain = _this5.db.collections.grains.findOne({                                                       // 95
                _id: grainId                                                                                           // 95
              });                                                                                                      // 95
                                                                                                                       //
              if (!grain) {                                                                                            // 96
                throw new Error("Grain not found.");                                                                   // 97
              }                                                                                                        // 98
                                                                                                                       //
              var castedNotification = notification.castAs(PersistentOngoingNotification);                             // 100
              var wakelockToken = waitPromise(castedNotification.save()).sturdyRef.toString("utf8"); // We have to close both the casted cap and the original. Perhaps this should be fixed in
              // node-capnp?                                                                                           // 104
                                                                                                                       //
              castedNotification.close();                                                                              // 105
              notification.close();                                                                                    // 106
                                                                                                                       //
              var notificationId = _this5.db.collections.notifications.insert({                                        // 107
                ongoing: wakelockToken,                                                                                // 108
                grainId: grainId,                                                                                      // 109
                userId: grain.userId,                                                                                  // 110
                text: displayInfo.caption,                                                                             // 111
                timestamp: new Date(),                                                                                 // 112
                isUnread: true                                                                                         // 113
              });                                                                                                      // 107
                                                                                                                       //
              return {                                                                                                 // 116
                handle: globalFrontendRefRegistry.create(_this5.db, {                                                  // 117
                  notificationHandle: notificationId                                                                   // 118
                })                                                                                                     // 118
              };                                                                                                       // 116
            });                                                                                                        // 120
          }                                                                                                            // 121
        }                                                                                                              // 92
      };                                                                                                               // 91
    }                                                                                                                  // 124
                                                                                                                       //
    return getOwnerNotificationTarget;                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.backgroundActivity = function () {                                                       //
    function backgroundActivity(event) {                                                                               //
      var _this6 = this;                                                                                               // 126
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 127
        logActivity(_this6.grainId, null, event);                                                                      // 128
      });                                                                                                              // 129
    }                                                                                                                  // 130
                                                                                                                       //
    return backgroundActivity;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.reportGrainSize = function () {                                                          //
    function reportGrainSize(bytes) {                                                                                  //
      bytes = parseInt(bytes); // int64s are stringified but precision isn't critical here                             // 133
                                                                                                                       //
      var result = this.db.collections.grains.findAndModify({                                                          // 135
        query: {                                                                                                       // 136
          _id: this.grainId                                                                                            // 136
        },                                                                                                             // 136
        update: {                                                                                                      // 137
          $set: {                                                                                                      // 137
            size: bytes                                                                                                // 137
          }                                                                                                            // 137
        },                                                                                                             // 137
        fields: {                                                                                                      // 138
          _id: 1,                                                                                                      // 138
          userId: 1,                                                                                                   // 138
          size: 1                                                                                                      // 138
        }                                                                                                              // 138
      });                                                                                                              // 135
                                                                                                                       //
      if (!result.ok) {                                                                                                // 141
        throw new Error("Grain not found.");                                                                           // 142
      } // If the grain did not have a "size" field before the update, then it is a grain created                      // 143
      // before per-grain size tracking was implemented. In that case, we don't want to update the                     // 146
      // user record because it may already be counting the grain (specifically on Blackrock, where                    // 147
      // whole-user size counting has existed for some time).                                                          // 148
                                                                                                                       //
                                                                                                                       //
      if (this.db.isQuotaEnabled() && "size" in result.value) {                                                        // 149
        // Update the user record, too. Note that we periodically recompute the user's storage usage                   // 150
        // from scratch as well, so this doesn't have to be perfectly reliable.                                        // 151
        var diff = bytes - (result.value.size || 0);                                                                   // 152
        this.db.collections.users.update(result.value.userId, {                                                        // 153
          $inc: {                                                                                                      // 153
            storageUsage: diff                                                                                         // 153
          }                                                                                                            // 153
        });                                                                                                            // 153
      }                                                                                                                // 154
    }                                                                                                                  // 155
                                                                                                                       //
    return reportGrainSize;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.getIdentityId = function () {                                                            //
    function getIdentityId(identity) {                                                                                 //
      var _this7 = this;                                                                                               // 157
                                                                                                                       //
      return unwrapFrontendCap(identity, "identity", function (accountId) {                                            // 158
        var grain = _this7.db.getGrain(_this7.grainId);                                                                // 159
                                                                                                                       //
        if (!grain) {                                                                                                  // 160
          throw new Error("Grain not found.");                                                                         // 161
        }                                                                                                              // 162
                                                                                                                       //
        var identityId = _this7.db.getOrGenerateIdentityId(accountId, grain);                                          // 163
                                                                                                                       //
        return {                                                                                                       // 164
          id: new Buffer(identityId, "hex")                                                                            // 164
        };                                                                                                             // 164
      });                                                                                                              // 165
    }                                                                                                                  // 166
                                                                                                                       //
    return getIdentityId;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormCoreImpl;                                                                                            //
}();                                                                                                                   //
                                                                                                                       //
var makeSandstormCore = function (db, grainId) {                                                                       // 169
  return new Capnp.Capability(new SandstormCoreImpl(db, grainId), SandstormCore);                                      // 170
};                                                                                                                     // 171
                                                                                                                       //
var NotificationHandle = function (_PersistentImpl) {                                                                  //
  (0, _inherits3.default)(NotificationHandle, _PersistentImpl);                                                        //
                                                                                                                       //
  // TODO(cleanup): Move to a different file.                                                                          // 174
  function NotificationHandle(db, saveTemplate, notificationId) {                                                      // 176
    (0, _classCallCheck3.default)(this, NotificationHandle);                                                           // 176
                                                                                                                       //
    var _this8 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 176
                                                                                                                       //
    _this8.notificationId = notificationId;                                                                            // 178
    _this8.db = db;                                                                                                    // 179
    return _this8;                                                                                                     // 176
  }                                                                                                                    // 180
                                                                                                                       //
  NotificationHandle.prototype.close = function () {                                                                   //
    function close() {                                                                                                 //
      var _this9 = this;                                                                                               // 182
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 183
        if (!_this9.isSaved()) {                                                                                       // 184
          dismissNotification(_this9.db, _this9.notificationId);                                                       // 185
        }                                                                                                              // 186
      });                                                                                                              // 187
    }                                                                                                                  // 188
                                                                                                                       //
    return close;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  return NotificationHandle;                                                                                           //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
var PersistentUiViewImpl = function (_PersistentImpl2) {                                                               //
  (0, _inherits3.default)(PersistentUiViewImpl, _PersistentImpl2);                                                     //
                                                                                                                       //
  // TODO(cleanup): Move out of core.js.                                                                               // 192
  function PersistentUiViewImpl(db, saveTemplate, grainId) {                                                           // 194
    (0, _classCallCheck3.default)(this, PersistentUiViewImpl);                                                         // 194
                                                                                                                       //
    var _this10 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));       // 194
                                                                                                                       //
    check(grainId, String);                                                                                            // 196
    _this10._db = db;                                                                                                  // 197
    _this10._grainId = grainId;                                                                                        // 198
    return _this10;                                                                                                    // 194
  }                                                                                                                    // 199
                                                                                                                       //
  PersistentUiViewImpl.prototype.getViewInfo = function () {                                                           //
    function getViewInfo() {                                                                                           //
      var _this11 = this;                                                                                              // 201
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 202
        var grain = _this11._db.getGrain(_this11._grainId);                                                            // 203
                                                                                                                       //
        if (!grain || grain.trashed) {                                                                                 // 204
          throw new Error("grain no longer exists");                                                                   // 205
        }                                                                                                              // 206
                                                                                                                       //
        var pkg = _this11._db.getPackage(grain.packageId) || DevPackages.findOne({                                     // 208
          appId: grain.appId                                                                                           // 209
        }) || {};                                                                                                      // 209
        var manifest = pkg.manifest || {};                                                                             // 212
        var viewInfo = grain.cachedViewInfo || {};                                                                     // 214
                                                                                                                       //
        if (!viewInfo.appTitle) {                                                                                      // 216
          viewInfo.appTitle = manifest.appTitle || {};                                                                 // 217
        }                                                                                                              // 218
                                                                                                                       //
        if (!viewInfo.grainIcon) {                                                                                     // 220
          var grainIcon = ((manifest.metadata || {}).icons || {}).grain;                                               // 221
                                                                                                                       //
          if (grainIcon) {                                                                                             // 222
            viewInfo.grainIcon = new Capnp.Capability(new StaticAssetImpl(grainIcon.assetId), StaticAsset);            // 223
          } else {                                                                                                     // 225
            var hash = Identicon.hashAppIdForIdenticon(pkg.appId);                                                     // 226
            viewInfo.grainIcon = new Capnp.Capability(new IdenticonStaticAssetImpl(hash, 24), StaticAsset);            // 227
          }                                                                                                            // 229
        }                                                                                                              // 230
                                                                                                                       //
        return viewInfo;                                                                                               // 232
      });                                                                                                              // 233
    }                                                                                                                  // 234
                                                                                                                       //
    return getViewInfo;                                                                                                //
  }(); // All other UiView methods are currently unimplemented, which, while not strictly correct,                     //
  // results in the same overall behavior, since users can't call restore() on a PersistentUiView,                     // 237
  // and grains can't call methods on UiViews because they lack the "is human" pseudopermission.                       // 238
                                                                                                                       //
                                                                                                                       //
  return PersistentUiViewImpl;                                                                                         //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
var makePersistentUiView = function (db, saveTemplate, grainId) {                                                      // 241
  check(grainId, String); // Verify that the grain exists and hasn't been trashed.                                     // 242
                                                                                                                       //
  var grain = db.getGrain(grainId);                                                                                    // 245
                                                                                                                       //
  if (!grain || grain.trashed) {                                                                                       // 246
    throw new Meteor.Error(404, "grain not found");                                                                    // 247
  }                                                                                                                    // 248
                                                                                                                       //
  return new Capnp.Capability(new PersistentUiViewImpl(db, saveTemplate, grainId), PersistentUiView);                  // 250
};                                                                                                                     // 252
                                                                                                                       //
globalFrontendRefRegistry.register({                                                                                   // 254
  frontendRefField: "notificationHandle",                                                                              // 255
  restore: function (db, saveTemplate, notificationId) {                                                               // 257
    return new Capnp.Capability(new NotificationHandle(db, saveTemplate, notificationId), PersistentHandle);           // 258
  }                                                                                                                    // 260
});                                                                                                                    // 254
                                                                                                                       //
function dismissNotification(db, notificationId, callCancel) {                                                         // 263
  var notification = db.collections.notifications.findOne({                                                            // 264
    _id: notificationId                                                                                                // 264
  });                                                                                                                  // 264
                                                                                                                       //
  if (notification) {                                                                                                  // 265
    db.collections.notifications.remove({                                                                              // 266
      _id: notificationId                                                                                              // 266
    });                                                                                                                // 266
                                                                                                                       //
    if (notification.ongoing) {                                                                                        // 267
      var id = notification.ongoing;                                                                                   // 268
                                                                                                                       //
      if (!callCancel) {                                                                                               // 270
        dropInternal(db, id, {                                                                                         // 271
          frontend: null                                                                                               // 271
        });                                                                                                            // 271
      } else {                                                                                                         // 272
        var notificationCap = restoreInternal(db, id, {                                                                // 273
          frontend: null                                                                                               // 273
        }, []).cap;                                                                                                    // 273
        var castedNotification = notificationCap.castAs(PersistentOngoingNotification);                                // 274
        dropInternal(db, id, {                                                                                         // 275
          frontend: null                                                                                               // 275
        });                                                                                                            // 275
                                                                                                                       //
        try {                                                                                                          // 276
          waitPromise(castedNotification.cancel());                                                                    // 277
          castedNotification.close();                                                                                  // 278
          notificationCap.close();                                                                                     // 279
        } catch (err) {                                                                                                // 280
          if (err.kjType !== "disconnected") {                                                                         // 281
            // ignore disconnected errors, since cancel may shutdown the grain before the supervisor                   // 282
            // responds.                                                                                               // 283
            throw err;                                                                                                 // 284
          }                                                                                                            // 285
        }                                                                                                              // 286
      }                                                                                                                // 287
    } else if (notification.appUpdates) {                                                                              // 288
      _.forEach(notification.appUpdates, function (app, appId) {                                                       // 289
        db.deleteUnusedPackages(appId);                                                                                // 290
      });                                                                                                              // 291
    }                                                                                                                  // 292
  }                                                                                                                    // 293
}                                                                                                                      // 294
                                                                                                                       //
Meteor.methods({                                                                                                       // 296
  dismissNotification: function (notificationId) {                                                                     // 297
    // This will remove notifications from the database and from view of the user.                                     // 298
    // For ongoing notifications, it will begin the process of cancelling and dropping them from                       // 299
    // the app.                                                                                                        // 300
    check(notificationId, String);                                                                                     // 302
    var db = this.connection.sandstormDb;                                                                              // 304
    var notification = db.collections.notifications.findOne({                                                          // 306
      _id: notificationId                                                                                              // 306
    });                                                                                                                // 306
                                                                                                                       //
    if (!notification) {                                                                                               // 307
      throw new Meteor.Error(404, "Notification id not found.");                                                       // 308
    } else if (notification.userId !== Meteor.userId()) {                                                              // 309
      throw new Meteor.Error(403, "Notification does not belong to current user.");                                    // 310
    } else {                                                                                                           // 311
      dismissNotification(db, notificationId, true);                                                                   // 312
    }                                                                                                                  // 313
  },                                                                                                                   // 314
  readAllNotifications: function () {                                                                                  // 316
    // Marks all notifications as read for the current user.                                                           // 317
    if (!Meteor.userId()) {                                                                                            // 318
      throw new Meteor.Error(403, "User not logged in.");                                                              // 319
    }                                                                                                                  // 320
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 322
    db.collections.notifications.update({                                                                              // 323
      userId: Meteor.userId()                                                                                          // 324
    }, {                                                                                                               // 324
      $set: {                                                                                                          // 325
        isUnread: false                                                                                                // 325
      }                                                                                                                // 325
    }, {                                                                                                               // 325
      multi: true                                                                                                      // 326
    });                                                                                                                // 326
  },                                                                                                                   // 327
  acceptPowerboxOffer: function (sessionId, sturdyRef, sessionToken) {                                                 // 329
    var db = this.connection.sandstormDb;                                                                              // 330
    check(sessionId, String);                                                                                          // 331
    check(sturdyRef, String);                                                                                          // 332
    check(sessionToken, Match.OneOf(String, null, undefined));                                                         // 333
    var sessionQuery = {                                                                                               // 335
      _id: sessionId                                                                                                   // 335
    };                                                                                                                 // 335
                                                                                                                       //
    if (sessionToken) {                                                                                                // 336
      sessionQuery.hashedToken = hashSturdyRef(sessionToken);                                                          // 337
    } else {                                                                                                           // 338
      sessionQuery.userId = this.userId;                                                                               // 339
    }                                                                                                                  // 340
                                                                                                                       //
    var session = db.collections.sessions.findOne(sessionQuery);                                                       // 342
                                                                                                                       //
    if (!session) {                                                                                                    // 343
      throw new Meteor.Error(404, "No matching session found.");                                                       // 344
    }                                                                                                                  // 345
                                                                                                                       //
    var apiToken = fetchApiToken(db, sturdyRef, {                                                                      // 347
      "owner.clientPowerboxOffer.sessionId": sessionId                                                                 // 348
    });                                                                                                                // 348
                                                                                                                       //
    if (!apiToken) {                                                                                                   // 349
      throw new Meteor.Error(404, "No such token.");                                                                   // 350
    }                                                                                                                  // 351
                                                                                                                       //
    var tokenId = apiToken._id;                                                                                        // 353
    var newSturdyRef = void 0;                                                                                         // 355
                                                                                                                       //
    if (sessionToken && apiToken.parentToken) {                                                                        // 356
      // An anonymous user is being offered a child token. To avoid bloating the database,                             // 357
      // we deterministically derive the sturdyref from the session token and the hashed parent                        // 358
      // token.                                                                                                        // 359
      //                                                                                                               // 360
      // Note that an attacker with read access to the database might be able to derive                                // 361
      // `newSturdyRef` from `sessionToken` and gain access to the capability without ever having                      // 362
      // been explicitly offered it by the grain. Therefore, grains must assume that offering                          // 363
      // a capability to an anonymous user can possibly make that capability accessible to any                         // 364
      // other anonymous user who has connected through the same URL.                                                  // 365
      newSturdyRef = Crypto.createHash("sha256").update(sessionToken).update(apiToken.parentToken).digest("base64").slice(0, -1) // removing trailing "="
      .replace(/\+/g, "-").replace(/\//g, "_"); // make URL-safe                                                       // 367
                                                                                                                       //
      if (fetchApiToken(db, newSturdyRef)) {                                                                           // 374
        // We have already generated this token.                                                                       // 375
        db.removeApiTokens({                                                                                           // 376
          _id: tokenId                                                                                                 // 376
        });                                                                                                            // 376
        return newSturdyRef;                                                                                           // 377
      }                                                                                                                // 378
    } else {                                                                                                           // 379
      newSturdyRef = generateSturdyRef();                                                                              // 380
    }                                                                                                                  // 381
                                                                                                                       //
    apiToken.owner = {                                                                                                 // 383
      webkey: null                                                                                                     // 383
    };                                                                                                                 // 383
    insertApiToken(db, apiToken, newSturdyRef);                                                                        // 385
    db.removeApiTokens({                                                                                               // 386
      _id: tokenId                                                                                                     // 386
    });                                                                                                                // 386
    return newSturdyRef;                                                                                               // 387
  }                                                                                                                    // 388
});                                                                                                                    // 296
                                                                                                                       //
var makeSaveTemplateForChild = function (db, parentToken, requirements, parentTokenInfo) {                             // 392
  // Constructs (part of) an ApiToken record appropriate to be used when save()ing a capability                        // 393
  // that was originally created by restore()ing `parentToken`. This fills in everything that is                       // 394
  // appropriate to fill in based only on the parent. Some fields -- especially `owner`, `created`,                    // 395
  // and `_id` -- obviously cannot be filled in until `save()` is called.                                              // 396
  //                                                                                                                   // 397
  // `parentToken` is the raw SturdyRef of the parent.                                                                 // 398
  //                                                                                                                   // 399
  // `requirements` is a list of MembraneRequirements that should be added to any children.                            // 400
  //                                                                                                                   // 401
  // `parentTokenInfo` is the ApiToken record for `parentToken`. Provide this only if you have                         // 402
  // it handy; if omitted it will be looked up.                                                                        // 403
  parentTokenInfo = parentTokenInfo || fetchApiToken(db, parentToken);                                                 // 405
                                                                                                                       //
  if (!parentTokenInfo) {                                                                                              // 406
    throw new Error("no such token");                                                                                  // 407
  }                                                                                                                    // 408
                                                                                                                       //
  var parentOwner = parentTokenInfo.owner || {};                                                                       // 410
  var saveTemplate = void 0;                                                                                           // 412
                                                                                                                       //
  if (parentOwner.clientPowerboxRequest || parentOwner.clientPowerboxOffer) {                                          // 413
    // Saving this token should make a copy of the restored token, rather than make a child                            // 414
    // token.                                                                                                          // 415
    saveTemplate = _.clone(parentTokenInfo); // Don't copy over fields that should be determined at save() time.       // 417
                                                                                                                       //
    delete saveTemplate._id;                                                                                           // 420
    delete saveTemplate.owner;                                                                                         // 421
    delete saveTemplate.created;                                                                                       // 422
  } else {                                                                                                             // 423
    if (parentTokenInfo.accountId) {                                                                                   // 424
      // A UiView token. Need to denormalize some fields from the parent.                                              // 425
      saveTemplate = _.pick(parentTokenInfo, "grainId", "accountId"); // By default, a save()d copy should have the same permissions, so set an allAccess role
      // assignment.                                                                                                   // 429
                                                                                                                       //
      saveTemplate.roleAssignment = {                                                                                  // 430
        allAccess: null                                                                                                // 430
      };                                                                                                               // 430
    } else {                                                                                                           // 431
      // A non-UiView child token.                                                                                     // 432
      saveTemplate = {};                                                                                               // 433
    } // Saved token should be a child of the restored token.                                                          // 434
                                                                                                                       //
                                                                                                                       //
    saveTemplate.parentToken = parentTokenInfo._id; // Will be encrypted on save().                                    // 437
                                                                                                                       //
    saveTemplate.parentTokenKey = parentToken;                                                                         // 440
  }                                                                                                                    // 441
                                                                                                                       //
  if (requirements) {                                                                                                  // 443
    // Append additional requirements requested by caller.                                                             // 444
    saveTemplate.requirements = (saveTemplate.requirements || []).concat(requirements);                                // 445
  }                                                                                                                    // 446
                                                                                                                       //
  return saveTemplate;                                                                                                 // 448
};                                                                                                                     // 449
                                                                                                                       //
restoreInternal = function (db, originalToken, ownerPattern, requirements, originalTokenInfo, currentTokenId, currentTokenKey) {
  // Restores the token `originalToken`, which is a Buffer.                                                            // 453
  //                                                                                                                   // 454
  // `ownerPattern` is a match pattern (i.e. used with check()) that the token's owner must match.                     // 455
  // This is used to enforce than an entity can't use tokens owned by some other entity.                               // 456
  //                                                                                                                   // 457
  // `requirements` is a list of additional MembraneRequirements to add to the returned capability,                    // 458
  // beyond what's already stored in ApiTokens. This is often an empty list.                                           // 459
  //                                                                                                                   // 460
  // `originalTokenInfo` is optional. If specified, it should be the ApiToken record associated                        // 461
  // with `originalToken`. Provide this if you happen to have looked it up already.                                    // 462
  //                                                                                                                   // 463
  // `currentTokenId` is only provided when this function calls itself recursively. It is the                          // 464
  // _id (hashed token) of an ancestor of originalToken. The function recurses until it gets to                        // 465
  // the top-level token.                                                                                              // 466
  requirements = requirements || [];                                                                                   // 468
                                                                                                                       //
  if (!originalTokenInfo) {                                                                                            // 470
    originalTokenInfo = fetchApiToken(db, originalToken);                                                              // 471
                                                                                                                       //
    if (!originalTokenInfo) {                                                                                          // 472
      throw new Meteor.Error(403, "No token found to restore");                                                        // 473
    }                                                                                                                  // 474
  }                                                                                                                    // 475
                                                                                                                       //
  var token = !currentTokenId ? originalTokenInfo : typeof currentTokenKey === "string" ? fetchApiToken(db, currentTokenKey) : db.collections.apiTokens.findOne(currentTokenId);
                                                                                                                       //
  if (!token) {                                                                                                        // 480
    if (!originalTokenInfo) {                                                                                          // 481
      throw new Meteor.Error(403, "Couldn't restore token because parent token has been deleted");                     // 482
    }                                                                                                                  // 483
  }                                                                                                                    // 484
                                                                                                                       //
  if (token.revoked) {                                                                                                 // 486
    throw new Meteor.Error(403, "Token has been revoked");                                                             // 487
  } // The ownerPattern should specify the appropriate user or grain involved, if appropriate.                         // 488
                                                                                                                       //
                                                                                                                       //
  check(token.owner, ownerPattern); // Check requirements on the token.                                                // 491
                                                                                                                       //
  checkRequirements(db, token.requirements); // Check expiration.                                                      // 494
                                                                                                                       //
  if (token.expires && token.expires.getTime() <= Date.now()) {                                                        // 497
    throw new Meteor.Error(403, "Authorization token expired");                                                        // 498
  }                                                                                                                    // 499
                                                                                                                       //
  if (token.expiresIfUnused) {                                                                                         // 501
    if (token.expiresIfUnused.getTime() <= Date.now()) {                                                               // 502
      throw new Meteor.Error(403, "Authorization token expired");                                                      // 503
    } else {                                                                                                           // 504
      // It's getting used now, so clear the expiresIfUnused field.                                                    // 505
      db.collections.apiTokens.update(token._id, {                                                                     // 506
        $unset: {                                                                                                      // 506
          expiresIfUnused: ""                                                                                          // 506
        }                                                                                                              // 506
      });                                                                                                              // 506
    }                                                                                                                  // 507
  } // If this token has a parent, go ahead and recurse to it now.                                                     // 508
                                                                                                                       //
                                                                                                                       //
  if (token.parentToken) {                                                                                             // 511
    // A token which chains to some parent token.  Restore the parent token (possibly recursively),                    // 512
    // checking requirements on the way up.                                                                            // 513
    return restoreInternal(db, originalToken, Match.Any, requirements, originalTokenInfo, token.parentToken, token.parentTokenKey);
  } // Check the passed-in `requirements`.                                                                             // 516
                                                                                                                       //
                                                                                                                       //
  checkRequirements(db, requirements);                                                                                 // 519
                                                                                                                       //
  if (token.objectId) {                                                                                                // 521
    // A token which represents a specific capability exported by a grain.                                             // 522
    // Fix Mongo converting Buffers to Uint8Arrays.                                                                    // 524
    if (token.objectId.appRef) {                                                                                       // 525
      token.objectId.appRef = new Buffer(token.objectId.appRef);                                                       // 526
    } // Ensure the grain is running, then restore the capability.                                                     // 527
                                                                                                                       //
                                                                                                                       //
    return waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {                                   // 530
      // Note that in this case it is the supervisor's job to implement SystemPersistent, so we                        // 531
      // don't generate a saveTemplate here.                                                                           // 532
      return supervisor.restore(token.objectId, requirements, new Buffer(originalToken, "utf8"));                      // 533
    }));                                                                                                               // 534
  } else {                                                                                                             // 535
    // Construct a template ApiToken for use if the restored capability is save()d later.                              // 536
    var saveTemplate = makeSaveTemplateForChild(db, originalToken, requirements, originalTokenInfo);                   // 537
                                                                                                                       //
    if (token.frontendRef) {                                                                                           // 539
      // A token which represents a capability implemented by a pseudo-driver.                                         // 540
      var cap = globalFrontendRefRegistry.restore(db, saveTemplate, token.frontendRef);                                // 542
      return {                                                                                                         // 543
        cap: cap                                                                                                       // 543
      };                                                                                                               // 543
    } else if (token.grainId) {                                                                                        // 544
      // It's a UiView.                                                                                                // 545
      // If a grain is attempting to restore a UiView, it gets a UiView which filters out all                          // 547
      // the method calls.  In the future, we may allow grains to restore UiViews that pass along the                  // 548
      // "is human" pseudopermission (say, to allow an app to proxy all requests to some grain and                     // 549
      // do some transformation), which will return a different capability.                                            // 550
      return {                                                                                                         // 551
        cap: makePersistentUiView(db, saveTemplate, token.grainId)                                                     // 551
      };                                                                                                               // 551
    } else {                                                                                                           // 552
      throw new Meteor.Error(500, "Unknown token type. ID: " + token._id);                                             // 553
    }                                                                                                                  // 554
  }                                                                                                                    // 555
};                                                                                                                     // 556
                                                                                                                       //
function dropInternal(db, sturdyRef, ownerPattern) {                                                                   // 558
  // Drops `sturdyRef`, checking first that its owner matches `ownerPattern`.                                          // 559
  var token = fetchApiToken(db, sturdyRef);                                                                            // 561
                                                                                                                       //
  if (!token) {                                                                                                        // 562
    return;                                                                                                            // 563
  }                                                                                                                    // 564
                                                                                                                       //
  var hashedSturdyRef = token._id;                                                                                     // 566
                                                                                                                       //
  if (!Match.test(token.owner, ownerPattern)) {                                                                        // 568
    // The caller of `drop()` does not own the token. From the caller's perspective, this means                        // 569
    // that there is actually nothing to be dropped. For example, perhaps a grain got backed up                        // 570
    // and restored, and now is trying to drop one of its pre-backup tokens. The call to                               // 571
    // `SandstormApi.drop()` should succeed, behaving exactly as if the token was not found.                           // 572
    return;                                                                                                            // 573
  } // TODO(soon): Revoke OAuth tokens.                                                                                // 574
                                                                                                                       //
                                                                                                                       //
  if (token.frontendRef && token.frontendRef.notificationHandle) {                                                     // 578
    var notificationId = token.frontendRef.notificationHandle;                                                         // 579
    db.removeApiTokens({                                                                                               // 580
      _id: hashedSturdyRef                                                                                             // 580
    });                                                                                                                // 580
    var anyToken = db.collections.apiTokens.findOne({                                                                  // 581
      "frontendRef.notificationHandle": notificationId                                                                 // 582
    });                                                                                                                // 582
                                                                                                                       //
    if (!anyToken) {                                                                                                   // 583
      // No other tokens referencing this notification exist, so dismiss the notification                              // 584
      dismissNotification(db, notificationId);                                                                         // 585
    }                                                                                                                  // 586
  } else if (token.objectId) {                                                                                         // 587
    waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {                                          // 588
      return supervisor.drop(token.objectId);                                                                          // 589
    }));                                                                                                               // 590
    db.removeApiTokens({                                                                                               // 592
      _id: hashedSturdyRef                                                                                             // 592
    });                                                                                                                // 592
  } else {                                                                                                             // 593
    db.removeApiTokens({                                                                                               // 594
      _id: hashedSturdyRef                                                                                             // 594
    });                                                                                                                // 594
  }                                                                                                                    // 595
}                                                                                                                      // 596
                                                                                                                       //
var SandstormCoreFactoryImpl = function () {                                                                           //
  function SandstormCoreFactoryImpl(db) {                                                                              // 599
    (0, _classCallCheck3.default)(this, SandstormCoreFactoryImpl);                                                     // 599
    this.db = db;                                                                                                      // 600
  }                                                                                                                    // 601
                                                                                                                       //
  SandstormCoreFactoryImpl.prototype.getSandstormCore = function () {                                                  //
    function getSandstormCore(grainId) {                                                                               //
      return {                                                                                                         // 604
        core: makeSandstormCore(this.db, grainId)                                                                      // 604
      };                                                                                                               // 604
    }                                                                                                                  // 605
                                                                                                                       //
    return getSandstormCore;                                                                                           //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormCoreFactoryImpl;                                                                                     //
}();                                                                                                                   //
                                                                                                                       //
makeSandstormCoreFactory = function (db) {                                                                             // 608
  return new Capnp.Capability(new SandstormCoreFactoryImpl(db), SandstormCoreFactory);                                 // 609
};                                                                                                                     // 610
                                                                                                                       //
unwrapFrontendCap = function (cap, type, callback) {                                                                   // 612
  // Expect that `cap` is a Cap'n Proto capability implemented by the frontend as a frontendRef                        // 613
  // with the given type (the name of one of the fields of frontendRef). Unwraps the capability                        // 614
  // and then calls callback() with the `frontendRef[type]` descriptor object as                                       // 615
  // the paramater. The callback runs in a Meteor fiber, but this function can be called from                          // 616
  // anywhere. The function returns a Promise for the result of the callback.                                          // 617
  //                                                                                                                   // 618
  // (The reason for the callback, rather than just returning a Promise for the descriptor, is                         // 619
  // so that you don't have to do your own inMeteor() dance.)                                                          // 620
  // For now, we save() the capability and then dig through ApiTokens to find where it leads.                          // 622
  // TODO(cleanup): In theory we should be using something like CapabilityServerSet, but it is                         // 623
  //   not available in Javascript yet and even if it were, it wouldn't work in the case where                         // 624
  //   there are multiple front-end replicas, since the capability could be on a different                             // 625
  //   replica.                                                                                                        // 626
  return cap.castAs(SystemPersistent).save({                                                                           // 628
    frontend: null                                                                                                     // 628
  }).then(function (saveResult) {                                                                                      // 628
    return inMeteor(function () {                                                                                      // 629
      var tokenInfo = fetchApiToken(globalDb, saveResult.sturdyRef.toString("utf8")); // Delete the token now since it's not needed.
                                                                                                                       //
      ApiTokens.remove(tokenInfo._id);                                                                                 // 633
                                                                                                                       //
      for (;;) {                                                                                                       // 635
        if (!tokenInfo) throw new Error("missing token?");                                                             // 636
        if (!tokenInfo.parentToken) break;                                                                             // 637
                                                                                                                       //
        if (typeof tokenInfo.parentTokenKey === "string") {                                                            // 638
          tokenInfo = fetchApiToken(globalDb, tokenInfo.parentTokenKey);                                               // 639
        } else {                                                                                                       // 640
          // Hmm, parentTokenKey doesn't exist or is still encrypted. We can't decrypt the parent                      // 641
          // but we can still fetch it in encrypted format.                                                            // 642
          tokenInfo = ApiTokens.findOne(tokenInfo.parentToken);                                                        // 643
        }                                                                                                              // 644
      }                                                                                                                // 645
                                                                                                                       //
      if (!tokenInfo.frontendRef || !tokenInfo.frontendRef[type]) {                                                    // 647
        throw new Error("not a " + type + " capability");                                                              // 648
      }                                                                                                                // 649
                                                                                                                       //
      return callback(tokenInfo.frontendRef[type]);                                                                    // 651
    });                                                                                                                // 652
  });                                                                                                                  // 653
};                                                                                                                     // 654
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"demo-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/demo-server.js                                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var allowDemo = void 0;                                                                                                // 1
module.watch(require("/imports/demo.js"), {                                                                            // 1
  allowDemo: function (v) {                                                                                            // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var DEMO_EXPIRATION_MS = 60 * 60 * 1000;                                                                               // 20
var DEMO_GRACE_MS = 10 * 60 * 1000; // time between expiration and deletion                                            // 21
                                                                                                                       //
Accounts.validateLoginAttempt(function (attempt) {                                                                     // 23
  // Enforce expiration times for demo accounts.                                                                       // 24
  if (attempt.user && attempt.user.expires) {                                                                          // 26
    var expireIn = attempt.user.expires.getTime() - Date.now() + DEMO_GRACE_MS;                                        // 27
                                                                                                                       //
    if (expireIn < 0) {                                                                                                // 28
      throw new Meteor.Error(403, "This demo account has expired.");                                                   // 29
    }                                                                                                                  // 30
                                                                                                                       //
    if (!allowDemo) {                                                                                                  // 32
      // An expiration time implies that this is a demo account, but those have been disabled.                         // 33
      // Perhaps the config was just changed.                                                                          // 34
      throw new Meteor.Error(403, "Demo accounts have been disabled.");                                                // 35
    } // Force connection close when account expires, so that the client reconnects and                                // 36
    // re-authenticates, which fails.                                                                                  // 39
                                                                                                                       //
                                                                                                                       //
    var connection = attempt.connection;                                                                               // 40
    var handle = Meteor.setTimeout(function () {                                                                       // 41
      connection.close();                                                                                              // 41
    }, expireIn);                                                                                                      // 41
    connection.onClose(function () {                                                                                   // 43
      Meteor.clearTimeout(handle);                                                                                     // 43
    });                                                                                                                // 43
  }                                                                                                                    // 44
                                                                                                                       //
  return true;                                                                                                         // 46
});                                                                                                                    // 47
                                                                                                                       //
function cleanupExpiredUsers() {                                                                                       // 49
  // Delete expired demo accounts and all their grains.                                                                // 50
  var now = new Date(Date.now() - DEMO_GRACE_MS);                                                                      // 52
  Meteor.users.find({                                                                                                  // 53
    expires: {                                                                                                         // 53
      $lt: now                                                                                                         // 53
    },                                                                                                                 // 53
    loginCredentials: {                                                                                                // 53
      $exists: true                                                                                                    // 53
    }                                                                                                                  // 53
  }, {                                                                                                                 // 53
    fields: {                                                                                                          // 54
      _id: 1,                                                                                                          // 54
      loginCredentials: 1,                                                                                             // 54
      lastActive: 1,                                                                                                   // 54
      appDemoId: 1,                                                                                                    // 54
      experiments: 1                                                                                                   // 54
    }                                                                                                                  // 54
  }).forEach(function (user) {                                                                                         // 54
    console.log("delete demo user: " + user._id);                                                                      // 56
    globalDb.deleteAccount(user._id, globalBackend); // Record stats about demo accounts.                              // 57
                                                                                                                       //
    var deleteStatsType = "demoUser";                                                                                  // 60
    var isAppDemoUser = !!user.appDemoId;                                                                              // 61
                                                                                                                       //
    if (isAppDemoUser) {                                                                                               // 62
      deleteStatsType = "appDemoUser";                                                                                 // 63
    } // Intentionally record deleted users at time of deletion to avoid miscounting users that                        // 64
    // were demoing just before the day rolled over.                                                                   // 67
                                                                                                                       //
                                                                                                                       //
    var record = {                                                                                                     // 68
      type: deleteStatsType,                                                                                           // 68
      lastActive: new Date(),                                                                                          // 68
      appId: user.appDemoId                                                                                            // 68
    };                                                                                                                 // 68
                                                                                                                       //
    if (user.experiments) {                                                                                            // 69
      record.experiments = user.experiments;                                                                           // 70
    }                                                                                                                  // 71
                                                                                                                       //
    DeleteStats.insert(record);                                                                                        // 73
  }); // All demo credentials should have been deleted as part of deleting the demo users, but just in                 // 74
  // case, check for them too.                                                                                         // 77
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 78
    expires: {                                                                                                         // 78
      $lt: now                                                                                                         // 78
    },                                                                                                                 // 78
    loginCredentials: {                                                                                                // 78
      $exists: false                                                                                                   // 78
    }                                                                                                                  // 78
  }, {                                                                                                                 // 78
    fields: {                                                                                                          // 79
      _id: 1                                                                                                           // 79
    }                                                                                                                  // 79
  }).forEach(function (user) {                                                                                         // 79
    console.log("delete demo credential: " + user._id);                                                                // 81
    globalDb.deleteCredential(user._id);                                                                               // 82
  });                                                                                                                  // 83
}                                                                                                                      // 84
                                                                                                                       //
if (allowDemo) {                                                                                                       // 86
  Meteor.methods({                                                                                                     // 87
    createDemoUser: function (displayName, appDemoId) {                                                                // 88
      // This is a login method that creates a new temporary user                                                      // 89
      // every time it is used.                                                                                        // 90
      //                                                                                                               // 91
      // appDemoId is important for stats; see cleanupExpiredUsers().                                                  // 92
      check(displayName, String);                                                                                      // 93
      check(appDemoId, Match.OneOf(undefined, null, String)); // Create the new user.                                  // 94
                                                                                                                       //
      var newUser = {                                                                                                  // 97
        expires: new Date(Date.now() + DEMO_EXPIRATION_MS)                                                             // 97
      };                                                                                                               // 97
                                                                                                                       //
      if (appDemoId) {                                                                                                 // 98
        newUser.appDemoId = appDemoId;                                                                                 // 99
      }                                                                                                                // 100
                                                                                                                       //
      var userId = Accounts.insertUserDoc({                                                                            // 102
        profile: {                                                                                                     // 102
          name: displayName                                                                                            // 102
        }                                                                                                              // 102
      }, newUser); // Log them in on this connection.                                                                  // 102
                                                                                                                       //
      return Accounts._loginMethod(this, "createDemoUser", arguments, "demo", function () {                            // 105
        return {                                                                                                       // 106
          userId: userId                                                                                               // 106
        };                                                                                                             // 106
      });                                                                                                              // 106
    },                                                                                                                 // 107
    testExpireDemo: function () {                                                                                      // 109
      if (!isDemoUser()) throw new Meteor.Error(403, "not a demo user");                                               // 110
      var newExpires = new Date(Date.now() + 15000);                                                                   // 112
                                                                                                                       //
      if (Meteor.user().expires.getTime() < newExpires.getTime()) {                                                    // 113
        throw new Meteor.Error(403, "can't exend demo");                                                               // 114
      }                                                                                                                // 115
                                                                                                                       //
      Meteor.users.update(this.userId, {                                                                               // 117
        $set: {                                                                                                        // 117
          expires: newExpires                                                                                          // 117
        }                                                                                                              // 117
      });                                                                                                              // 117
    }                                                                                                                  // 118
  }); // If demo mode is enabled, we permit the client to subscribe to information about an app by                     // 87
  // appId. If this were available in non-demo mode, then anonymous users could effectively ask the                    // 122
  // server which apps are installed.                                                                                  // 123
                                                                                                                       //
  Meteor.publish("appDemoInfo", function (appId) {                                                                     // 124
    // This publishes info about an app, including the latest version of it. Once you log in, it                       // 125
    // also publishes your list of UserActions.                                                                        // 126
    check(appId, String); // Get data about this app from the app index. Note that the app index cache on this server  // 127
    // is typically 6-24 hours delayed from reality, so if an app is newly-available                                   // 130
    // in the app market, it won't be in the app index. Therefore we don't bail-out if the app                         // 131
    // is missing from the AppIndex collection.                                                                        // 132
                                                                                                                       //
    var appIndexData = globalDb.collections.appIndex.findOne({                                                         // 133
      appId: appId                                                                                                     // 133
    }); // Prepare a helper function we can use to transform a package cursor into an appropriate return               // 133
    // value for this function.                                                                                        // 136
                                                                                                                       //
    var packageCursorAndMaybeUserActions = function (userId, appId, packageCursor) {                                   // 137
      // This allows us to avoid creating duplicate UserActions.                                                       // 138
      if (userId) {                                                                                                    // 139
        return [packageCursor, UserActions.find({                                                                      // 140
          userId: userId,                                                                                              // 142
          appId: appId                                                                                                 // 142
        })];                                                                                                           // 142
      }                                                                                                                // 144
                                                                                                                       //
      return packageCursor;                                                                                            // 146
    }; // If the app is in the app index, and the current version is installed, always return that. If                 // 147
    // that package isn't installed, store the version number so we can filter on it later.                            // 150
                                                                                                                       //
                                                                                                                       //
    var packageQuery = {};                                                                                             // 151
                                                                                                                       //
    if (appIndexData) {                                                                                                // 152
      var appIndexPackageQuery = Packages.find({                                                                       // 153
        _id: appIndexData.packageId                                                                                    // 153
      });                                                                                                              // 153
                                                                                                                       //
      if (appIndexPackageQuery.count() > 0) {                                                                          // 154
        return packageCursorAndMaybeUserActions(this.userId, appId, appIndexPackageQuery);                             // 155
      } // If the app index version isn't present, insist on a lower version than the app index                        // 156
      // version. This avoids accidentally catching some development version of the app that has the                   // 159
      // same version number as the app market version but isn't the app market version.                               // 160
                                                                                                                       //
                                                                                                                       //
      packageQuery["manifest.appVersion"] = {                                                                          // 161
        $lt: appIndexData.versionNumber                                                                                // 161
      };                                                                                                               // 161
    } // If the specific package from the app index isn't installed, or the app isn't there at all, do                 // 162
    // our best.                                                                                                       // 165
                                                                                                                       //
                                                                                                                       //
    packageQuery.appId = appId;                                                                                        // 166
    var packageCursor = Packages.find(packageQuery, {                                                                  // 167
      sort: {                                                                                                          // 169
        "manifest.appVersion": -1                                                                                      // 169
      },                                                                                                               // 169
      limit: 1                                                                                                         // 170
    });                                                                                                                // 169
    return packageCursorAndMaybeUserActions(this.userId, appId, packageCursor);                                        // 172
  });                                                                                                                  // 173
  SandstormDb.periodicCleanup(DEMO_EXPIRATION_MS, cleanupExpiredUsers);                                                // 175
} else {                                                                                                               // 176
  // Just run once, in case the config just changed from allowing demos to prohibiting them.                           // 177
  Meteor.setTimeout(cleanupExpiredUsers, DEMO_EXPIRATION_MS + DEMO_GRACE_MS);                                          // 178
}                                                                                                                      // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"desktop-notifications-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/desktop-notifications-server.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.watch(require("meteor/meteor"), {                                                                               // 1
  Meteor: function (v) {                                                                                               // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
SandstormDb.periodicCleanup(120000, function () {                                                                      // 19
  // Remove old desktop notfications regularly.                                                                        // 20
  // periodicCleanup doesn't like runs more frequent than once every two minutes                                       // 21
  var now = Date.now();                                                                                                // 22
  var then = new Date(now - 30000); // Clear tokens older than 30 seconds.                                             // 23
                                                                                                                       //
  globalDb.collections.desktopNotifications.find({                                                                     // 24
    creationDate: {                                                                                                    // 25
      $lt: then                                                                                                        // 25
    }                                                                                                                  // 25
  }).forEach(function (doc) {                                                                                          // 24
    globalDb.collections.desktopNotifications.remove({                                                                 // 27
      _id: doc._id                                                                                                     // 28
    });                                                                                                                // 27
  });                                                                                                                  // 30
});                                                                                                                    // 31
Meteor.publish("desktopNotifications", function () {                                                                   // 33
  var _this = this;                                                                                                    // 33
                                                                                                                       //
  var subscribeTime = new Date();                                                                                      // 34
                                                                                                                       //
  if (!this.userId) {                                                                                                  // 36
    // No desktop notifications for anonymous users.                                                                   // 37
    return [];                                                                                                         // 38
  }                                                                                                                    // 39
                                                                                                                       //
  var db = this.connection.sandstormDb;                                                                                // 41
  var callbacks = {                                                                                                    // 42
    added: function (doc) {                                                                                            // 43
      _this.added("desktopNotifications", doc._id, doc);                                                               // 44
    },                                                                                                                 // 45
    changed: function (newDoc, oldDoc) {                                                                               // 47
      _this.changed("desktopNotifications", newDoc._id, newDoc);                                                       // 48
    },                                                                                                                 // 49
    removed: function (doc) {                                                                                          // 51
      _this.removed("desktopNotifications", doc._id);                                                                  // 52
    }                                                                                                                  // 53
  };                                                                                                                   // 42
  var sub = db.collections.desktopNotifications.find({                                                                 // 56
    userId: this.userId,                                                                                               // 57
    creationDate: {                                                                                                    // 58
      $gt: subscribeTime                                                                                               // 58
    }                                                                                                                  // 58
  }).observe(callbacks);                                                                                               // 56
  this.onStop(function () {                                                                                            // 61
    if (sub) {                                                                                                         // 62
      sub.stop();                                                                                                      // 63
    }                                                                                                                  // 64
  });                                                                                                                  // 65
  this.ready();                                                                                                        // 67
});                                                                                                                    // 68
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"dev-accounts-server.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/dev-accounts-server.js                                                                                       //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
Meteor.methods({                                                                                                       // 19
  createDevAccount: function (displayName, isAdmin, profile, unverifiedEmail) {                                        // 20
    // This is a login method that creates or logs in a dev account with the given displayName                         // 21
    check(displayName, String);                                                                                        // 23
    check(isAdmin, Match.OneOf(undefined, null, Boolean));                                                             // 24
    check(profile, Match.OneOf(undefined, null, Object));                                                              // 25
    check(unverifiedEmail, Match.OneOf(undefined, null, String));                                                      // 26
                                                                                                                       //
    if (!this.connection.sandstormDb.allowDevAccounts()) {                                                             // 28
      throw new Meteor.Error(404, "Dev accounts are not enabled on this server");                                      // 29
    }                                                                                                                  // 30
                                                                                                                       //
    isAdmin = isAdmin || false;                                                                                        // 32
    profile = profile || {};                                                                                           // 34
    profile.name = profile.name || displayName;                                                                        // 35
    var hasCompletedSignup = !!unverifiedEmail && !!profile.pronoun && !!profile.handle;                               // 36
    var user = Meteor.users.findOne({                                                                                  // 38
      "services.dev.name": displayName                                                                                 // 38
    });                                                                                                                // 38
    var userId = void 0;                                                                                               // 39
                                                                                                                       //
    if (user) {                                                                                                        // 41
      userId = user._id;                                                                                               // 42
    } else {                                                                                                           // 43
      userId = Accounts.insertUserDoc({                                                                                // 44
        profile: profile,                                                                                              // 45
        unverifiedEmail: unverifiedEmail                                                                               // 46
      }, {                                                                                                             // 44
        services: {                                                                                                    // 48
          dev: {                                                                                                       // 49
            name: displayName,                                                                                         // 50
            isAdmin: isAdmin,                                                                                          // 51
            hasCompletedSignup: hasCompletedSignup                                                                     // 52
          }                                                                                                            // 49
        }                                                                                                              // 48
      });                                                                                                              // 47
    } // Log them in on this connection.                                                                               // 56
                                                                                                                       //
                                                                                                                       //
    return Accounts._loginMethod(this, "createDevAccount", arguments, "dev", function () {                             // 58
      return {                                                                                                         // 59
        userId: userId                                                                                                 // 59
      };                                                                                                               // 59
    });                                                                                                                // 59
  }                                                                                                                    // 60
});                                                                                                                    // 19
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"grain-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/grain-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var sendEmail = void 0;                                                                                                // 1
module.watch(require("/imports/server/email.js"), {                                                                    // 1
  send: function (v) {                                                                                                 // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
var ROOT_URL = process.env.ROOT_URL;                                                                                   // 20
                                                                                                                       //
var emailLinkWithInlineStyle = function (url, text) {                                                                  // 22
  return "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#762F87;color:white'>" + text + "</a>";
};                                                                                                                     // 27
                                                                                                                       //
Meteor.publish("grainTopBar", function (grainId) {                                                                     // 29
  check(grainId, String);                                                                                              // 30
  var result = [Grains.find({                                                                                          // 32
    _id: grainId,                                                                                                      // 34
    $or: [{                                                                                                            // 35
      userId: this.userId                                                                                              // 36
    }, {                                                                                                               // 36
      "private": {                                                                                                     // 37
        $ne: true                                                                                                      // 37
      }                                                                                                                // 37
    }]                                                                                                                 // 37
  }, {                                                                                                                 // 33
    fields: {                                                                                                          // 40
      title: 1,                                                                                                        // 41
      userId: 1,                                                                                                       // 42
      "private": 1                                                                                                     // 43
    }                                                                                                                  // 40
  })];                                                                                                                 // 39
                                                                                                                       //
  if (this.userId) {                                                                                                   // 47
    result.push(ApiTokens.find({                                                                                       // 48
      grainId: grainId,                                                                                                // 49
      $or: [{                                                                                                          // 50
        "owner.user.accountId": this.userId                                                                            // 51
      }, {                                                                                                             // 51
        accountId: this.userId                                                                                         // 52
      }]                                                                                                               // 52
    }));                                                                                                               // 48
  }                                                                                                                    // 55
                                                                                                                       //
  return result;                                                                                                       // 57
}); // We allow users to learn package information about a grain they own.                                             // 58
// This is used for obtaining icon and app title information for grains                                                // 61
// you own, which is used in the sidebar. It is not a security/privacy                                                 // 62
// risk since it only exposes this information for grains the user owns.                                               // 63
                                                                                                                       //
Meteor.publish("packageByGrainId", function (grainId) {                                                                // 64
  check(grainId, String);                                                                                              // 65
  var publishThis = []; // We need to publish the packageId so that client-side code can                               // 66
  // find the right package.                                                                                           // 68
                                                                                                                       //
  var thisGrainCursor = Grains.find({                                                                                  // 69
    _id: grainId,                                                                                                      // 70
    userId: this.userId                                                                                                // 71
  }, {                                                                                                                 // 69
    fields: {                                                                                                          // 73
      packageId: 1                                                                                                     // 73
    }                                                                                                                  // 73
  });                                                                                                                  // 72
  publishThis.push(thisGrainCursor);                                                                                   // 75
                                                                                                                       //
  if (thisGrainCursor.count()) {                                                                                       // 77
    var thisGrain = thisGrainCursor.fetch()[0];                                                                        // 78
    var thisPackageCursor = Packages.find({                                                                            // 79
      _id: thisGrain.packageId                                                                                         // 79
    });                                                                                                                // 79
    publishThis.push(thisPackageCursor);                                                                               // 80
  }                                                                                                                    // 81
                                                                                                                       //
  return publishThis;                                                                                                  // 83
});                                                                                                                    // 84
Meteor.publish("tokenInfo", function (token, isStandalone) {                                                           // 86
  // Allows the client side to map a raw token to its entry in ApiTokens, and the additional                           // 87
  // metadata that it will need to display the app icon and title.  We do not care about making                        // 88
  // the metadata reactive.                                                                                            // 89
  check(token, String);                                                                                                // 90
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                        // 92
  var apiToken = ApiTokens.findOne({                                                                                   // 93
    _id: hashedToken                                                                                                   // 94
  }, {                                                                                                                 // 93
    fields: {                                                                                                          // 96
      grainId: 1,                                                                                                      // 97
      accountId: 1,                                                                                                    // 98
      owner: 1,                                                                                                        // 99
      revoked: 1                                                                                                       // 100
    }                                                                                                                  // 96
  });                                                                                                                  // 95
                                                                                                                       //
  if (!apiToken) {                                                                                                     // 103
    this.added("tokenInfo", token, {                                                                                   // 104
      invalidToken: true                                                                                               // 104
    });                                                                                                                // 104
  } else if (apiToken.revoked) {                                                                                       // 105
    this.added("tokenInfo", token, {                                                                                   // 106
      revoked: true                                                                                                    // 106
    });                                                                                                                // 106
  } else {                                                                                                             // 107
    var grainId = apiToken.grainId;                                                                                    // 108
    var grain = Grains.findOne({                                                                                       // 109
      _id: grainId                                                                                                     // 110
    }, {                                                                                                               // 109
      fields: {                                                                                                        // 112
        packageId: 1,                                                                                                  // 113
        appId: 1,                                                                                                      // 114
        userId: 1                                                                                                      // 115
      }                                                                                                                // 112
    });                                                                                                                // 111
                                                                                                                       //
    if (!grain) {                                                                                                      // 118
      this.added("tokenInfo", token, {                                                                                 // 119
        invalidToken: true                                                                                             // 119
      });                                                                                                              // 119
    } else {                                                                                                           // 120
      if (apiToken.owner && apiToken.owner.user) {                                                                     // 121
        var account = Meteor.users.findOne({                                                                           // 122
          _id: apiToken.owner.user.accountId                                                                           // 122
        });                                                                                                            // 122
        var metadata = apiToken.owner.user.denormalizedGrainMetadata;                                                  // 123
                                                                                                                       //
        if (account && metadata) {                                                                                     // 124
          account.credentials = Meteor.users.find({                                                                    // 125
            _id: {                                                                                                     // 126
              $in: account.loginCredentials.map(function (cred) {                                                      // 126
                return cred.id;                                                                                        // 126
              })                                                                                                       // 126
            }                                                                                                          // 126
          }).map(function (credential) {                                                                               // 126
            return {                                                                                                   // 128
              serviceName: SandstormDb.getServiceName(credential),                                                     // 129
              intrinsicName: SandstormDb.getIntrinsicName(credential, true),                                           // 130
              loginId: SandstormDb.getLoginId(credential)                                                              // 131
            };                                                                                                         // 128
          });                                                                                                          // 133
          account.intrinsicNames = globalDb.getAccountIntrinsicNames(account, false);                                  // 135
          this.added("tokenInfo", token, {                                                                             // 137
            accountOwner: _.pick(account, "_id", "profile", "credentials", "intrinsicNames"),                          // 138
            grainId: grainId,                                                                                          // 139
            grainMetadata: metadata                                                                                    // 140
          });                                                                                                          // 137
        } else {                                                                                                       // 142
          this.added("tokenInfo", token, {                                                                             // 143
            invalidToken: true                                                                                         // 143
          });                                                                                                          // 143
        }                                                                                                              // 144
      } else if (!apiToken.owner || "webkey" in apiToken.owner) {                                                      // 145
        if (this.userId && !isStandalone) {                                                                            // 146
          var childToken = ApiTokens.findOne({                                                                         // 147
            "owner.user.accountId": this.userId,                                                                       // 148
            parentToken: apiToken._id                                                                                  // 149
          });                                                                                                          // 147
                                                                                                                       //
          if (childToken || this.userId === grain.userId || this.userId === apiToken.accountId) {                      // 151
            this.added("tokenInfo", token, {                                                                           // 153
              alreadyRedeemed: true,                                                                                   // 153
              grainId: apiToken.grainId                                                                                // 153
            });                                                                                                        // 153
            this.ready();                                                                                              // 154
            return;                                                                                                    // 155
          }                                                                                                            // 156
        }                                                                                                              // 157
                                                                                                                       //
        var pkg = Packages.findOne({                                                                                   // 159
          _id: grain.packageId                                                                                         // 159
        }, {                                                                                                           // 159
          fields: {                                                                                                    // 159
            manifest: 1                                                                                                // 159
          }                                                                                                            // 159
        });                                                                                                            // 159
        var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || {                                               // 160
          defaultText: ""                                                                                              // 160
        };                                                                                                             // 160
        var appIcon = undefined;                                                                                       // 161
                                                                                                                       //
        if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                             // 162
          var icons = pkg.manifest.metadata.icons;                                                                     // 163
          appIcon = icons.grain || icons.appGrid;                                                                      // 164
        }                                                                                                              // 165
                                                                                                                       //
        var denormalizedGrainMetadata = {                                                                              // 167
          appTitle: appTitle,                                                                                          // 168
          icon: appIcon,                                                                                               // 169
          appId: appIcon ? undefined : grain.appId                                                                     // 170
        };                                                                                                             // 167
        this.added("tokenInfo", token, {                                                                               // 172
          webkey: true,                                                                                                // 173
          grainId: grainId,                                                                                            // 174
          grainMetadata: denormalizedGrainMetadata                                                                     // 175
        });                                                                                                            // 172
      } else {                                                                                                         // 177
        this.added("tokenInfo", token, {                                                                               // 178
          invalidToken: true                                                                                           // 178
        });                                                                                                            // 178
      }                                                                                                                // 179
    }                                                                                                                  // 180
  }                                                                                                                    // 181
                                                                                                                       //
  this.ready();                                                                                                        // 183
  return;                                                                                                              // 184
});                                                                                                                    // 185
Meteor.publish("requestingAccess", function (grainId) {                                                                // 187
  check(grainId, String);                                                                                              // 188
                                                                                                                       //
  if (!this.userId) {                                                                                                  // 190
    throw new Meteor.Error(403, "Must be logged in to request access.");                                               // 191
  }                                                                                                                    // 192
                                                                                                                       //
  var grain = globalDb.getGrain(grainId);                                                                              // 194
                                                                                                                       //
  if (!grain) {                                                                                                        // 195
    throw new Meteor.Error(404, "Grain not found.");                                                                   // 196
  }                                                                                                                    // 197
                                                                                                                       //
  if (grain.userId === this.userId) {                                                                                  // 199
    this.added("grantedAccessRequests", Random.id(), {                                                                 // 200
      grainId: grainId                                                                                                 // 200
    });                                                                                                                // 200
  }                                                                                                                    // 201
                                                                                                                       //
  var _this = this;                                                                                                    // 203
                                                                                                                       //
  var query = ApiTokens.find({                                                                                         // 204
    grainId: grainId,                                                                                                  // 205
    accountId: grain.userId,                                                                                           // 206
    parentToken: {                                                                                                     // 207
      $exists: false                                                                                                   // 207
    },                                                                                                                 // 207
    "owner.user.accountId": this.userId,                                                                               // 208
    revoked: {                                                                                                         // 209
      $ne: true                                                                                                        // 209
    }                                                                                                                  // 209
  });                                                                                                                  // 204
  var handle = query.observe({                                                                                         // 211
    added: function (apiToken) {                                                                                       // 212
      _this.added("grantedAccessRequests", Random.id(), {                                                              // 213
        grainId: grainId                                                                                               // 213
      });                                                                                                              // 213
    }                                                                                                                  // 214
  });                                                                                                                  // 211
  this.onStop(function () {                                                                                            // 217
    return handle.stop();                                                                                              // 217
  });                                                                                                                  // 217
});                                                                                                                    // 218
var GRAIN_DELETION_MS = 1000 * 60 * 60 * 24 * 30; // thirty days                                                       // 220
                                                                                                                       //
SandstormDb.periodicCleanup(86400000, function () {                                                                    // 221
  var trashExpiration = new Date(Date.now() - GRAIN_DELETION_MS);                                                      // 222
  globalDb.removeApiTokens({                                                                                           // 223
    trashed: {                                                                                                         // 223
      $lt: trashExpiration                                                                                             // 223
    }                                                                                                                  // 223
  });                                                                                                                  // 223
  globalDb.deleteGrains({                                                                                              // 224
    trashed: {                                                                                                         // 224
      $lt: trashExpiration                                                                                             // 224
    }                                                                                                                  // 224
  }, globalBackend, "grain");                                                                                          // 224
});                                                                                                                    // 225
Meteor.methods({                                                                                                       // 227
  updateGrainTitle: function (grainId, newTitle, obsolete) {                                                           // 228
    check(grainId, String);                                                                                            // 229
    check(newTitle, String);                                                                                           // 230
                                                                                                                       //
    if (this.userId) {                                                                                                 // 231
      var grain = Grains.findOne(grainId);                                                                             // 232
                                                                                                                       //
      if (grain) {                                                                                                     // 233
        if (grain.userId === this.userId) {                                                                            // 234
          Grains.update({                                                                                              // 235
            _id: grainId,                                                                                              // 235
            userId: this.userId                                                                                        // 235
          }, {                                                                                                         // 235
            $set: {                                                                                                    // 235
              title: newTitle                                                                                          // 235
            }                                                                                                          // 235
          }); // Denormalize new title out to all sharing tokens.                                                      // 235
                                                                                                                       //
          ApiTokens.update({                                                                                           // 238
            grainId: grainId,                                                                                          // 238
            "owner.user": {                                                                                            // 238
              $exists: true                                                                                            // 238
            }                                                                                                          // 238
          }, {                                                                                                         // 238
            $set: {                                                                                                    // 239
              "owner.user.upstreamTitle": newTitle                                                                     // 239
            }                                                                                                          // 239
          }, {                                                                                                         // 239
            multi: true                                                                                                // 240
          });                                                                                                          // 240
        } else {                                                                                                       // 241
          var token = ApiTokens.findOne({                                                                              // 242
            grainId: grainId,                                                                                          // 243
            objectId: {                                                                                                // 244
              $exists: false                                                                                           // 244
            },                                                                                                         // 244
            "owner.user.accountId": this.userId                                                                        // 245
          }, {                                                                                                         // 242
            sort: {                                                                                                    // 247
              created: 1                                                                                               // 247
            } // The oldest token is our source of truth for the name.                                                 // 247
                                                                                                                       //
          });                                                                                                          // 246
                                                                                                                       //
          if (token && token.owner.user.title !== newTitle) {                                                          // 249
            if (token.owner.user.upstreamTitle === newTitle) {                                                         // 250
              // User renamed grain to match upstream title. Act like they never renamed it at                         // 251
              // all.                                                                                                  // 252
              ApiTokens.update({                                                                                       // 253
                grainId: grainId,                                                                                      // 254
                "owner.user.accountId": this.userId                                                                    // 255
              }, {                                                                                                     // 253
                $set: {                                                                                                // 257
                  "owner.user.title": newTitle                                                                         // 257
                },                                                                                                     // 257
                $unset: {                                                                                              // 258
                  "owner.user.upstreamTitle": 1,                                                                       // 258
                  "owner.user.renamed": 1                                                                              // 258
                }                                                                                                      // 258
              }, {                                                                                                     // 256
                multi: true                                                                                            // 260
              });                                                                                                      // 259
            } else {                                                                                                   // 262
              var modification = {                                                                                     // 263
                "owner.user.title": newTitle,                                                                          // 264
                "owner.user.renamed": true                                                                             // 265
              };                                                                                                       // 263
                                                                                                                       //
              if (!token.owner.user.upstreamTitle) {                                                                   // 268
                // If `upstreamTitle` isn't present then it is equal to the old title.                                 // 269
                modification["owner.user.upstreamTitle"] = token.owner.user.title;                                     // 270
              }                                                                                                        // 271
                                                                                                                       //
              ApiTokens.update({                                                                                       // 273
                grainId: grainId,                                                                                      // 273
                "owner.user.accountId": this.userId                                                                    // 273
              }, {                                                                                                     // 273
                $set: modification                                                                                     // 274
              }, {                                                                                                     // 274
                multi: true                                                                                            // 275
              });                                                                                                      // 275
            }                                                                                                          // 276
          }                                                                                                            // 277
        }                                                                                                              // 278
      }                                                                                                                // 279
    }                                                                                                                  // 280
  },                                                                                                                   // 281
  privatizeGrain: function (grainId) {                                                                                 // 283
    check(grainId, String);                                                                                            // 284
                                                                                                                       //
    if (this.userId) {                                                                                                 // 285
      Grains.update({                                                                                                  // 286
        _id: grainId,                                                                                                  // 286
        userId: this.userId                                                                                            // 286
      }, {                                                                                                             // 286
        $set: {                                                                                                        // 286
          "private": true                                                                                              // 286
        }                                                                                                              // 286
      });                                                                                                              // 286
    }                                                                                                                  // 287
  },                                                                                                                   // 288
  inviteUsersToGrain: function (_origin, obsolete, grainId, title, roleAssignment, contacts, message) {                // 290
    if ((typeof message === "undefined" ? "undefined" : (0, _typeof3.default)(message)) === "object") {                // 292
      // Older versions of the client passed an object here, but we only care about the `text`                         // 293
      // parameter. (This block can eventually be removed.)                                                            // 294
      message = message.text;                                                                                          // 295
    }                                                                                                                  // 296
                                                                                                                       //
    if (!this.isSimulation) {                                                                                          // 298
      check(_origin, String);                                                                                          // 299
      check(grainId, String);                                                                                          // 300
      check(title, String);                                                                                            // 301
      check(roleAssignment, roleAssignmentPattern);                                                                    // 302
      check(contacts, [{                                                                                               // 303
        _id: String,                                                                                                   // 305
        isDefault: Match.Optional(Boolean),                                                                            // 306
        profile: Match.ObjectIncluding({                                                                               // 307
          service: Match.Optional(String),                                                                             // 308
          name: String,                                                                                                // 309
          intrinsicName: Match.Optional(String)                                                                        // 310
        }),                                                                                                            // 307
        intrinsicNames: [Object]                                                                                       // 312
      }]);                                                                                                             // 304
      check(message, String);                                                                                          // 315
                                                                                                                       //
      if (!this.userId) {                                                                                              // 316
        throw new Meteor.Error(403, "Must be logged in to share by email.");                                           // 317
      }                                                                                                                // 318
                                                                                                                       //
      if (contacts.length === 0) {                                                                                     // 320
        throw new Meteor.Error(400, "No contacts were provided.");                                                     // 321
      }                                                                                                                // 322
                                                                                                                       //
      if (globalDb.isDemoUser()) {                                                                                     // 324
        throw new Meteor.Error(403, "Demo users are not allowed to share by email.");                                  // 325
      }                                                                                                                // 326
                                                                                                                       //
      var escapedMessage = message.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;").replace(/"/g, "&quot;").replace(/\n/g, "<br>");
      var accountId = this.userId;                                                                                     // 334
      var outerResult = {                                                                                              // 335
        successes: [],                                                                                                 // 335
        failures: []                                                                                                   // 335
      };                                                                                                               // 335
      var fromEmail = globalDb.getReturnAddressWithDisplayName(accountId);                                             // 336
      var replyTo = globalDb.getPrimaryEmail(accountId);                                                               // 337
      contacts.forEach(function (contact) {                                                                            // 338
        if (contact.isDefault) {                                                                                       // 339
          var emailAddress = contact.profile.name;                                                                     // 340
          var result = SandstormPermissions.createNewApiToken(globalDb, {                                              // 341
            accountId: accountId                                                                                       // 342
          }, grainId, "email invitation for " + emailAddress, roleAssignment, {                                        // 342
            webkey: {                                                                                                  // 344
              forSharing: true                                                                                         // 344
            }                                                                                                          // 344
          });                                                                                                          // 344
          var url = ROOT_URL + "/shared/" + result.token;                                                              // 345
          var html = escapedMessage + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
                                                                                                                       //
          try {                                                                                                        // 351
            globalDb.incrementDailySentMailCount(accountId);                                                           // 352
            sendEmail({                                                                                                // 353
              to: emailAddress,                                                                                        // 354
              from: fromEmail,                                                                                         // 355
              replyTo: replyTo,                                                                                        // 356
              subject: title + " - Invitation to collaborate",                                                         // 357
              text: message + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
              html: html                                                                                               // 361
            });                                                                                                        // 353
          } catch (e) {                                                                                                // 363
            console.error(e.stack);                                                                                    // 364
            outerResult.failures.push({                                                                                // 365
              contact: contact,                                                                                        // 365
              error: e.toString()                                                                                      // 365
            });                                                                                                        // 365
          }                                                                                                            // 366
        } else {                                                                                                       // 367
          var _result = SandstormPermissions.createNewApiToken(globalDb, {                                             // 368
            accountId: accountId                                                                                       // 369
          }, grainId, "direct invitation to " + contact.profile.name, roleAssignment, {                                // 369
            user: {                                                                                                    // 371
              accountId: contact._id,                                                                                  // 371
              title: title                                                                                             // 371
            }                                                                                                          // 371
          });                                                                                                          // 371
                                                                                                                       //
          var _url = ROOT_URL + "/shared/" + _result.token;                                                            // 372
                                                                                                                       //
          try {                                                                                                        // 373
            var account = Meteor.users.findOne({                                                                       // 374
              _id: contact._id                                                                                         // 374
            });                                                                                                        // 374
                                                                                                                       //
            var email = _.findWhere(SandstormDb.getUserEmails(account), {                                              // 375
              primary: true                                                                                            // 376
            });                                                                                                        // 376
                                                                                                                       //
            if (email) {                                                                                               // 377
              var intrinsicName = contact.profile.intrinsicName;                                                       // 378
                                                                                                                       //
              var _html = escapedMessage + "<br><br>" + emailLinkWithInlineStyle(_url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>";
                                                                                                                       //
              globalDb.incrementDailySentMailCount(accountId);                                                         // 383
              sendEmail({                                                                                              // 384
                to: email.email,                                                                                       // 385
                from: fromEmail,                                                                                       // 386
                replyTo: replyTo,                                                                                      // 387
                subject: title + " - Invitation to collaborate",                                                       // 388
                text: message + "\n\nFollow this link to open the shared grain:\n\n" + _url,                           // 389
                html: _html                                                                                            // 390
              });                                                                                                      // 384
            } else {                                                                                                   // 392
              outerResult.failures.push({                                                                              // 393
                contact: contact,                                                                                      // 393
                warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + _url + " with them."
              });                                                                                                      // 393
            }                                                                                                          // 396
          } catch (e) {                                                                                                // 397
            console.error(e.stack);                                                                                    // 398
            outerResult.failures.push({                                                                                // 399
              contact: contact,                                                                                        // 399
              error: e.toString(),                                                                                     // 399
              warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + _url + " with them."
            });                                                                                                        // 399
          }                                                                                                            // 402
        }                                                                                                              // 403
      });                                                                                                              // 404
      return outerResult;                                                                                              // 406
    }                                                                                                                  // 407
  },                                                                                                                   // 408
  requestAccess: function (_origin, grainId, obsolete) {                                                               // 410
    check(_origin, String);                                                                                            // 411
    check(grainId, String);                                                                                            // 412
                                                                                                                       //
    if (!this.isSimulation) {                                                                                          // 413
      if (!this.userId) {                                                                                              // 414
        throw new Meteor.Error(403, "Must be logged in to request access.");                                           // 415
      }                                                                                                                // 416
                                                                                                                       //
      var grain = Grains.findOne(grainId);                                                                             // 418
                                                                                                                       //
      if (!grain) {                                                                                                    // 419
        throw new Meteor.Error(404, "No such grain");                                                                  // 420
      }                                                                                                                // 421
                                                                                                                       //
      var grainOwner = globalDb.getUser(grain.userId);                                                                 // 423
                                                                                                                       //
      var email = _.findWhere(SandstormDb.getUserEmails(grainOwner), {                                                 // 424
        primary: true                                                                                                  // 424
      });                                                                                                              // 424
                                                                                                                       //
      if (!email) {                                                                                                    // 425
        throw new Meteor.Error("no email", "Grain owner has no email address.");                                       // 426
      }                                                                                                                // 427
                                                                                                                       //
      var emailAddress = email.email;                                                                                  // 429
      globalDb.addContact(grainOwner._id, this.userId);                                                                // 431
      var fromEmail = globalDb.getReturnAddressWithDisplayName(this.userId);                                           // 433
      var replyTo = globalDb.getPrimaryEmail(this.userId); // TODO(soon): In the HTML version, we should display an identity card.
                                                                                                                       //
      var identityNotes = [];                                                                                          // 437
      globalDb.getAccountIntrinsicNames(Meteor.user(), true).forEach(function (intrinsic) {                            // 438
        // TODO(cleanup): Don't switch on service here; extend getAccountIntrinsicNames or                             // 439
        //   Account.loginServices to cover what we need.                                                              // 440
        if (intrinsic.service === "google") {                                                                          // 441
          identityNotes.push(intrinsic.name);                                                                          // 442
        } else if (intrinsic.service === "github") {                                                                   // 443
          identityNotes.push(intrinsic.name + " on GitHub");                                                           // 444
        } else if (intrinsic.service === "email") {                                                                    // 445
          identityNotes.push(intrinsic.name);                                                                          // 446
        } else if (intrinsic.service === "ldap") {                                                                     // 447
          identityNotes.push(intrinsic.name + " on LDAP");                                                             // 448
        } else if (intrinsic.service === "saml") {                                                                     // 449
          identityNotes.push(intrinsic.name + " on SAML");                                                             // 450
        }                                                                                                              // 451
      });                                                                                                              // 452
      var identityNote = identityNotes.length === 0 ? "" : " (" + identityNotes.join(", ") + ")";                      // 454
      var message = Meteor.user().profile.name + identityNote + " is requesting access to your grain: " + grain.title + ".";
      var url = ROOT_URL + "/share/" + grainId + "/" + this.userId;                                                    // 460
      var html = message + "<br><br>" + emailLinkWithInlineStyle(url, "Open Sharing Menu");                            // 462
      var user = Meteor.user();                                                                                        // 465
      var ACCESS_REQUEST_LIMIT = 10;                                                                                   // 466
      var resetCount = true;                                                                                           // 467
                                                                                                                       //
      if (user.accessRequests) {                                                                                       // 468
        if (user.accessRequests.resetOn < new Date()) {                                                                // 469
          Meteor.users.update({                                                                                        // 470
            _id: user._id                                                                                              // 470
          }, {                                                                                                         // 470
            $unset: {                                                                                                  // 470
              accessRequests: 1                                                                                        // 470
            }                                                                                                          // 470
          });                                                                                                          // 470
        } else if (user.accessRequests.count >= ACCESS_REQUEST_LIMIT) {                                                // 471
          throw new Meteor.Error(403, "For spam control reasons, you are not allowed to make " + "more than " + ACCESS_REQUEST_LIMIT + " access requests per day.");
        } else {                                                                                                       // 475
          resetCount = false;                                                                                          // 476
        }                                                                                                              // 477
      }                                                                                                                // 478
                                                                                                                       //
      var modifier = {                                                                                                 // 480
        $inc: {                                                                                                        // 480
          "accessRequests.count": 1                                                                                    // 480
        }                                                                                                              // 480
      };                                                                                                               // 480
                                                                                                                       //
      if (resetCount) {                                                                                                // 481
        var tomorrow = new Date(Date.now() + 86400000);                                                                // 482
        modifier.$set = {                                                                                              // 483
          "accessRequests.resetOn": tomorrow                                                                           // 483
        };                                                                                                             // 483
      }                                                                                                                // 484
                                                                                                                       //
      Meteor.users.update({                                                                                            // 486
        _id: user._id                                                                                                  // 486
      }, modifier);                                                                                                    // 486
      sendEmail({                                                                                                      // 488
        to: emailAddress,                                                                                              // 489
        from: fromEmail,                                                                                               // 490
        replyTo: replyTo,                                                                                              // 491
        subject: grain.title + " - Request for access",                                                                // 492
        text: message + "\n\nFollow this link to share access:\n\n" + url,                                             // 493
        html: html                                                                                                     // 494
      });                                                                                                              // 488
    }                                                                                                                  // 496
  }                                                                                                                    // 497
});                                                                                                                    // 227
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"hack-session.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/hack-session.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var hashSturdyRef = void 0,                                                                                            // 1
    checkRequirements = void 0,                                                                                        // 1
    fetchApiToken = void 0;                                                                                            // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  hashSturdyRef: function (v) {                                                                                        // 1
    hashSturdyRef = v;                                                                                                 // 1
  },                                                                                                                   // 1
  checkRequirements: function (v) {                                                                                    // 1
    checkRequirements = v;                                                                                             // 1
  },                                                                                                                   // 1
  fetchApiToken: function (v) {                                                                                        // 1
    fetchApiToken = v;                                                                                                 // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var ssrfSafeLookup = void 0;                                                                                           // 1
module.watch(require("/imports/server/networking.js"), {                                                               // 1
  ssrfSafeLookup: function (v) {                                                                                       // 1
    ssrfSafeLookup = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 18
                                                                                                                       //
var Https = Npm.require("https");                                                                                      // 19
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 20
                                                                                                                       //
var Dgram = Npm.require("dgram");                                                                                      // 21
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 22
                                                                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");                                                            // 27
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;                        // 28
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 29
var SystemPersistent = Capnp.importSystem("sandstorm/supervisor.capnp").SystemPersistent;                              // 30
var IpRpc = Capnp.importSystem("sandstorm/ip.capnp");                                                                  // 31
var EmailSendPort = EmailRpc.EmailSendPort;                                                                            // 32
var Grain = Capnp.importSystem("sandstorm/grain.capnp");                                                               // 33
var Powerbox = Capnp.importSystem("sandstorm/powerbox.capnp");                                                         // 34
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 36
                                                                                                                       //
ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                            // 38
HOSTNAME = ROOT_URL.hostname;                                                                                          // 39
                                                                                                                       //
SessionContextImpl = function () {                                                                                     // 41
  function SessionContextImpl(grainId, sessionId, accountId, tabId) {                                                  // 42
    (0, _classCallCheck3.default)(this, SessionContextImpl);                                                           // 42
    this.grainId = grainId;                                                                                            // 43
    this.sessionId = sessionId;                                                                                        // 44
    this.accountId = accountId;                                                                                        // 45
    this.tabId = tabId;                                                                                                // 46
  }                                                                                                                    // 47
                                                                                                                       //
  SessionContextImpl.prototype.claimRequest = function () {                                                            // 41
    function claimRequest(sturdyRef, requiredPermissions) {                                                            // 41
      var _this2 = this;                                                                                               // 49
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 50
        var token = fetchApiToken(globalDb, sturdyRef, {                                                               // 51
          "owner.clientPowerboxRequest.sessionId": _this2.sessionId                                                    // 52
        });                                                                                                            // 52
                                                                                                                       //
        if (!token) {                                                                                                  // 54
          throw new Error("no such token");                                                                            // 55
        }                                                                                                              // 56
                                                                                                                       //
        var session = Sessions.findOne({                                                                               // 58
          _id: _this2.sessionId                                                                                        // 58
        });                                                                                                            // 58
                                                                                                                       //
        if (!session) {                                                                                                // 60
          throw new Error("no such session");                                                                          // 61
        } // Honor `requiredPermissions`.                                                                              // 62
                                                                                                                       //
                                                                                                                       //
        var requirements = [];                                                                                         // 65
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 66
          // Session is authorized by token. Note that it's important to check this before accountId                   // 67
          // e.g. in the case of standalone domains.                                                                   // 68
          requirements.push({                                                                                          // 69
            permissionsHeld: {                                                                                         // 70
              permissions: requiredPermissions || [],                                                                  // 71
              tokenId: session.hashedToken,                                                                            // 72
              grainId: _this2.grainId                                                                                  // 73
            }                                                                                                          // 70
          });                                                                                                          // 69
        } else if (session.identityId) {                                                                               // 76
          // Session is authorized by identity. (We check session.identityId rather than                               // 77
          // session.userId because session.userId is set even for incognito sessions.)                                // 78
          requirements.push({                                                                                          // 79
            permissionsHeld: {                                                                                         // 80
              permissions: requiredPermissions || [],                                                                  // 81
              accountId: session.userId,                                                                               // 82
              grainId: _this2.grainId                                                                                  // 83
            }                                                                                                          // 80
          });                                                                                                          // 79
        } else {                                                                                                       // 86
          // This can only happen with old-style sharing which has been deprecated for years. If                       // 87
          // we wanted to support this, I suppose we could do so as long as requiredPermissions is                     // 88
          // all-false? But probably this will never come up.                                                          // 89
          throw new Error("Cannot accept powerbox request from anonymous session that " + "doesn't have a token.");    // 90
        }                                                                                                              // 92
                                                                                                                       //
        return restoreInternal(globalDb, sturdyRef, {                                                                  // 94
          clientPowerboxRequest: Match.ObjectIncluding({                                                               // 96
            sessionId: _this2.sessionId                                                                                // 96
          })                                                                                                           // 96
        }, requirements, token);                                                                                       // 96
      });                                                                                                              // 98
    }                                                                                                                  // 99
                                                                                                                       //
    return claimRequest;                                                                                               // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype._offerOrFulfill = function () {                                                         // 41
    function _offerOrFulfill(isFulfill, cap, requiredPermissions, descriptor, displayInfo) {                           // 41
      var _this3 = this;                                                                                               // 101
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 102
        var session = Sessions.findOne({                                                                               // 104
          _id: _this3.sessionId                                                                                        // 104
        });                                                                                                            // 104
                                                                                                                       //
        if (!session.identityId && !session.hashedToken) {                                                             // 106
          throw new Error("Session has neither an identityId nor a hashedToken.");                                     // 107
        }                                                                                                              // 108
                                                                                                                       //
        if (isFulfill && !session.powerboxRequest) {                                                                   // 110
          // Not a request session, so treat fulfillRequest() same as offer().                                         // 111
          isFulfill = false;                                                                                           // 112
        }                                                                                                              // 113
                                                                                                                       //
        var castedCap = cap.castAs(SystemPersistent);                                                                  // 115
        var apiTokenOwner = {                                                                                          // 116
          clientPowerboxOffer: {                                                                                       // 116
            sessionId: _this3.sessionId                                                                                // 116
          }                                                                                                            // 116
        };                                                                                                             // 116
        var isUiView = descriptor && descriptor.tags && descriptor.tags.length === 1 && descriptor.tags[0] && descriptor.tags[0].id && descriptor.tags[0].id === Grain.UiView.typeId;
                                                                                                                       //
        if (isUiView) {                                                                                                // 120
          var tagValue = {};                                                                                           // 121
                                                                                                                       //
          if (descriptor.tags[0].value) {                                                                              // 122
            tagValue = Capnp.parse(Grain.UiView.PowerboxTag, descriptor.tags[0].value);                                // 123
          }                                                                                                            // 124
                                                                                                                       //
          if (!tagValue.title) {                                                                                       // 126
            throw new Error("No value provided for UiView.PowerboxTag.title.");                                        // 127
          }                                                                                                            // 128
                                                                                                                       //
          if (session.identityId) {                                                                                    // 130
            apiTokenOwner = {                                                                                          // 131
              user: {                                                                                                  // 132
                accountId: _this3.accountId,                                                                           // 133
                title: tagValue.title                                                                                  // 134
              }                                                                                                        // 132
            };                                                                                                         // 131
          }                                                                                                            // 137
        }                                                                                                              // 138
                                                                                                                       //
        if (isFulfill) {                                                                                               // 140
          // The capability will pass directly to the requesting grain.                                                // 141
          apiTokenOwner = {                                                                                            // 142
            clientPowerboxRequest: {                                                                                   // 143
              sessionId: session.powerboxRequest.requestingSession                                                     // 144
            }                                                                                                          // 143
          };                                                                                                           // 142
        } // TODO(soon): This will eventually use SystemPersistent.addRequirements when membranes                      // 147
        // are fully implemented for supervisors.                                                                      // 150
                                                                                                                       //
                                                                                                                       //
        var permissionsHeld = {                                                                                        // 151
          grainId: _this3.grainId,                                                                                     // 152
          permissions: requiredPermissions                                                                             // 153
        }; // Note that if both hashedToken and identityId are present, this indicates that the session                // 151
        // is authorized by a token which the user has NOT redeemed. Thus the identity does not                        // 157
        // directly have access to the grain; the token does. (This is the case in particular for                      // 158
        // standalone domains.) So clearly we want our requirement to be based on the token, not the                   // 159
        // identity.                                                                                                   // 160
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 161
          permissionsHeld.tokenId = session.hashedToken;                                                               // 162
        } else if (session.identityId) {                                                                               // 163
          permissionsHeld.accountId = session.userId;                                                                  // 164
        } else {                                                                                                       // 165
          throw new Error("Cannot offer to anonymous session that does not have a token.");                            // 166
        }                                                                                                              // 167
                                                                                                                       //
        var requirement = {                                                                                            // 169
          permissionsHeld: permissionsHeld                                                                             // 169
        };                                                                                                             // 169
        checkRequirements(globalDb, [requirement]);                                                                    // 171
        var save = castedCap.save(apiTokenOwner);                                                                      // 173
        var sturdyRef = waitPromise(save).sturdyRef;                                                                   // 174
        ApiTokens.update({                                                                                             // 175
          _id: hashSturdyRef(sturdyRef)                                                                                // 175
        }, {                                                                                                           // 175
          $push: {                                                                                                     // 175
            requirements: requirement                                                                                  // 175
          }                                                                                                            // 175
        });                                                                                                            // 175
        var powerboxView = void 0;                                                                                     // 177
                                                                                                                       //
        if (isFulfill) {                                                                                               // 178
          powerboxView = {                                                                                             // 179
            fulfill: {                                                                                                 // 180
              token: sturdyRef.toString(),                                                                             // 181
              descriptor: Capnp.serializePacked(Powerbox.PowerboxDescriptor, descriptor).toString("base64")            // 182
            }                                                                                                          // 180
          };                                                                                                           // 179
        } else if (isUiView) {                                                                                         // 186
          if (session.identityId) {                                                                                    // 187
            // Deduplicate.                                                                                            // 188
            var newApiToken = fetchApiToken(globalDb, sturdyRef.toString());                                           // 189
            var tokenId = newApiToken._id;                                                                             // 190
                                                                                                                       //
            var dupeQuery = _.pick(newApiToken, "grainId", "roleAssignment", "requirements", "parentToken", "parentTokenKey", "accountId", "accountId");
                                                                                                                       //
            dupeQuery._id = {                                                                                          // 193
              $ne: newApiToken._id                                                                                     // 193
            };                                                                                                         // 193
            dupeQuery["owner.user.accountId"] = _this3.accountId;                                                      // 194
            dupeQuery.trashed = {                                                                                      // 195
              $exists: false                                                                                           // 195
            };                                                                                                         // 195
            dupeQuery.revoked = {                                                                                      // 196
              $exists: false                                                                                           // 196
            };                                                                                                         // 196
            var dupeToken = ApiTokens.findOne(dupeQuery);                                                              // 198
                                                                                                                       //
            if (dupeToken) {                                                                                           // 199
              globalDb.removeApiTokens({                                                                               // 200
                _id: tokenId                                                                                           // 200
              });                                                                                                      // 200
              tokenId = dupeToken._id;                                                                                 // 201
            }                                                                                                          // 202
                                                                                                                       //
            powerboxView = {                                                                                           // 204
              offer: {                                                                                                 // 204
                uiView: {                                                                                              // 204
                  tokenId: tokenId                                                                                     // 204
                }                                                                                                      // 204
              }                                                                                                        // 204
            };                                                                                                         // 204
          } else {                                                                                                     // 205
            powerboxView = {                                                                                           // 206
              offer: {                                                                                                 // 206
                uiView: {                                                                                              // 206
                  token: sturdyRef.toString()                                                                          // 206
                }                                                                                                      // 206
              }                                                                                                        // 206
            };                                                                                                         // 206
          }                                                                                                            // 207
        } else {                                                                                                       // 208
          powerboxView = {                                                                                             // 209
            offer: {                                                                                                   // 210
              token: sturdyRef.toString()                                                                              // 211
            }                                                                                                          // 210
          };                                                                                                           // 209
        }                                                                                                              // 214
                                                                                                                       //
        Sessions.update({                                                                                              // 216
          _id: _this3.sessionId                                                                                        // 216
        }, {                                                                                                           // 216
          $set: {                                                                                                      // 218
            powerboxView: powerboxView                                                                                 // 219
          }                                                                                                            // 218
        });                                                                                                            // 217
      });                                                                                                              // 223
    }                                                                                                                  // 224
                                                                                                                       //
    return _offerOrFulfill;                                                                                            // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.offer = function () {                                                                   // 41
    function offer(cap, requiredPermissions, descriptor, displayInfo) {                                                // 41
      return this._offerOrFulfill(false, cap, requiredPermissions, descriptor, displayInfo);                           // 227
    }                                                                                                                  // 228
                                                                                                                       //
    return offer;                                                                                                      // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.fulfillRequest = function () {                                                          // 41
    function fulfillRequest(cap, requiredPermissions, descriptor, displayInfo) {                                       // 41
      return this._offerOrFulfill(true, cap, requiredPermissions, descriptor, displayInfo);                            // 231
    }                                                                                                                  // 232
                                                                                                                       //
    return fulfillRequest;                                                                                             // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.activity = function () {                                                                // 41
    function activity(event) {                                                                                         // 41
      var _this4 = this;                                                                                               // 234
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 235
        logActivity(_this4.grainId, _this4.accountId || "anonymous", event);                                           // 236
      });                                                                                                              // 237
    }                                                                                                                  // 238
                                                                                                                       //
    return activity;                                                                                                   // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  return SessionContextImpl;                                                                                           // 41
}();                                                                                                                   // 41
                                                                                                                       //
Meteor.methods({                                                                                                       // 241
  finishPowerboxRequest: function (sessionId, webkeyUrl, saveLabel, obsolete, grainId) {                               // 242
    check(sessionId, String);                                                                                          // 243
    check(webkeyUrl, String);                                                                                          // 244
    check(saveLabel, Match.OneOf(undefined, null, String));                                                            // 245
    check(grainId, String);                                                                                            // 246
    var db = this.connection.sandstormDb;                                                                              // 248
    var userId = Meteor.userId();                                                                                      // 250
                                                                                                                       //
    if (!userId) {                                                                                                     // 251
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 252
    }                                                                                                                  // 253
                                                                                                                       //
    var parsedWebkey = Url.parse(webkeyUrl.trim());                                                                    // 255
    var hostId = matchWildcardHost(parsedWebkey.host);                                                                 // 256
                                                                                                                       //
    if (!hostId || !db.isApiHostId(hostId)) {                                                                          // 257
      throw new Meteor.Error(500, "Hostname does not match this server. External webkeys are not " + "supported (yet)");
    }                                                                                                                  // 260
                                                                                                                       //
    var token = parsedWebkey.hash;                                                                                     // 262
                                                                                                                       //
    if (!token) {                                                                                                      // 263
      throw new Meteor.Error(400, "Invalid webkey. You must pass in the full webkey, " + "including domain name and hash fragment");
    }                                                                                                                  // 266
                                                                                                                       //
    token = token.slice(1);                                                                                            // 268
                                                                                                                       //
    if (db.isTokenSpecificHostId(hostId) && hostId !== db.apiHostIdForToken(token)) {                                  // 269
      // Note: This case is not security-sensitive. The client could easily compute the correct                        // 270
      //   host ID for this token. But since they've passed one that doesn't match, we assume                          // 271
      //   something is wrong and stop here.                                                                           // 272
      throw new Meteor.Error(400, "Invalid webkey: token doesn't match hostname.");                                    // 273
    }                                                                                                                  // 274
                                                                                                                       //
    var cap = restoreInternal(db, token, Match.Optional({                                                              // 276
      webkey: Match.Optional(Match.Any)                                                                                // 277
    }), []).cap;                                                                                                       // 277
    var castedCap = cap.castAs(SystemPersistent);                                                                      // 278
    var owner = {                                                                                                      // 279
      clientPowerboxRequest: {                                                                                         // 280
        grainId: grainId,                                                                                              // 281
        sessionId: sessionId                                                                                           // 282
      }                                                                                                                // 280
    };                                                                                                                 // 279
                                                                                                                       //
    if (saveLabel) {                                                                                                   // 285
      grainOwner.saveLabel = {                                                                                         // 286
        defaultText: saveLabel                                                                                         // 287
      };                                                                                                               // 286
    }                                                                                                                  // 289
                                                                                                                       //
    var save = castedCap.save(owner);                                                                                  // 291
    var sturdyRef = waitPromise(save).sturdyRef;                                                                       // 292
    return sturdyRef.toString();                                                                                       // 293
  },                                                                                                                   // 294
  finishPowerboxOffer: function (sessionId) {                                                                          // 296
    check(sessionId, String);                                                                                          // 297
    Sessions.update({                                                                                                  // 299
      _id: sessionId                                                                                                   // 299
    }, {                                                                                                               // 299
      $unset: {                                                                                                        // 299
        powerboxView: null                                                                                             // 299
      }                                                                                                                // 299
    });                                                                                                                // 299
  },                                                                                                                   // 300
  getViewInfoForApiToken: function (apiTokenId) {                                                                      // 302
    check(apiTokenId, String);                                                                                         // 303
    var token = globalDb.collections.apiTokens.findOne(apiTokenId);                                                    // 304
                                                                                                                       //
    if (!token) {                                                                                                      // 305
      throw new Meteor.Error(400, "No such token.");                                                                   // 306
    }                                                                                                                  // 307
                                                                                                                       //
    var grain = globalDb.collections.grains.findOne(token.grainId);                                                    // 309
                                                                                                                       //
    if (!grain) {                                                                                                      // 310
      throw new Meteor.Error(500, "No grain found for token.");                                                        // 311
    }                                                                                                                  // 312
                                                                                                                       //
    return grain.cachedViewInfo;                                                                                       // 314
  }                                                                                                                    // 315
});                                                                                                                    // 241
                                                                                                                       //
HackSessionContextImpl = function (_SessionContextImpl) {                                                              // 318
  (0, _inherits3.default)(HackSessionContextImpl, _SessionContextImpl);                                                // 318
                                                                                                                       //
  function HackSessionContextImpl(grainId, sessionId, accountId, tabId) {                                              // 319
    (0, _classCallCheck3.default)(this, HackSessionContextImpl);                                                       // 319
    return (0, _possibleConstructorReturn3.default)(this, _SessionContextImpl.call(this, grainId, sessionId, accountId, tabId));
  }                                                                                                                    // 321
                                                                                                                       //
  HackSessionContextImpl.prototype._getPublicId = function () {                                                        // 318
    function _getPublicId() {                                                                                          // 318
      // Get the grain's public ID, assigning a new one if it doesn't yet have one.                                    // 324
      //                                                                                                               // 325
      // Must be called in a Meteor context.                                                                           // 326
      while (!this.publicId) {                                                                                         // 328
        // We haven't looked up the public ID yet.                                                                     // 329
        var grain = Grains.findOne(this.grainId, {                                                                     // 330
          fields: {                                                                                                    // 330
            publicId: 1                                                                                                // 330
          }                                                                                                            // 330
        });                                                                                                            // 330
        if (!grain) throw new Error("Grain does not exist.");                                                          // 331
                                                                                                                       //
        if (grain.publicId) {                                                                                          // 333
          this.publicId = grain.publicId;                                                                              // 334
        } else {                                                                                                       // 335
          // The grain doesn't have a public ID yet. Generate one.                                                     // 336
          var candidate = generateRandomHostname(20);                                                                  // 337
                                                                                                                       //
          if (Grains.findOne({                                                                                         // 339
            publicId: candidate                                                                                        // 339
          })) {                                                                                                        // 339
            // This should never ever happen.                                                                          // 340
            console.error("CRITICAL PROBLEM: Public ID collision. " + "CSPRNG is bad or has insufficient entropy.");   // 341
            continue;                                                                                                  // 343
          } // Carefully perform an update that becomes a no-op if anyone else has assigned a public ID                // 344
          // simultaneously.                                                                                           // 347
                                                                                                                       //
                                                                                                                       //
          if (Grains.update({                                                                                          // 348
            _id: this.grainId,                                                                                         // 348
            publicId: {                                                                                                // 348
              $exists: false                                                                                           // 348
            }                                                                                                          // 348
          }, {                                                                                                         // 348
            $set: {                                                                                                    // 349
              publicId: candidate                                                                                      // 349
            }                                                                                                          // 349
          }) > 0) {                                                                                                    // 349
            // We won the race.                                                                                        // 350
            this.publicId = candidate;                                                                                 // 351
          }                                                                                                            // 352
        }                                                                                                              // 353
      }                                                                                                                // 354
                                                                                                                       //
      return this.publicId;                                                                                            // 356
    }                                                                                                                  // 357
                                                                                                                       //
    return _getPublicId;                                                                                               // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype._getAddress = function () {                                                         // 318
    function _getAddress() {                                                                                           // 318
      // Get the grain's outgoing e-mail address.                                                                      // 360
      //                                                                                                               // 361
      // Must be called in a Meteor context.                                                                           // 362
      return this._getPublicId() + "@" + HOSTNAME;                                                                     // 364
    }                                                                                                                  // 365
                                                                                                                       //
    return _getAddress;                                                                                                // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype._getUserAddress = function () {                                                     // 318
    function _getUserAddress() {                                                                                       // 318
      // Get the user's e-mail address.                                                                                // 368
      //                                                                                                               // 369
      // Must be called in a Meteor context.                                                                           // 370
      var grain = Grains.findOne(this.grainId, {                                                                       // 372
        fields: {                                                                                                      // 372
          userId: 1                                                                                                    // 372
        }                                                                                                              // 372
      });                                                                                                              // 372
      var user = Meteor.users.findOne({                                                                                // 374
        _id: grain.userId                                                                                              // 374
      });                                                                                                              // 374
                                                                                                                       //
      var email = _.findWhere(SandstormDb.getUserEmails(user), {                                                       // 376
        primary: true                                                                                                  // 376
      });                                                                                                              // 376
                                                                                                                       //
      var result = {};                                                                                                 // 378
                                                                                                                       //
      if (email) {                                                                                                     // 379
        result.address = email.email;                                                                                  // 380
      }                                                                                                                // 381
                                                                                                                       //
      if (user.profile.name) {                                                                                         // 383
        result.name = user.profile.name;                                                                               // 384
      }                                                                                                                // 385
                                                                                                                       //
      return result;                                                                                                   // 387
    }                                                                                                                  // 388
                                                                                                                       //
    return _getUserAddress;                                                                                            // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.send = function () {                                                                // 318
    function send(email) {                                                                                             // 318
      return hackSendEmail(this, email);                                                                               // 391
    }                                                                                                                  // 392
                                                                                                                       //
    return send;                                                                                                       // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.getPublicId = function () {                                                         // 318
    function getPublicId() {                                                                                           // 318
      return inMeteor(function () {                                                                                    // 395
        var result = {};                                                                                               // 396
        result.publicId = this._getPublicId();                                                                         // 398
        result.hostname = HOSTNAME;                                                                                    // 399
        result.autoUrl = ROOT_URL.protocol + "//" + makeWildcardHost(result.publicId);                                 // 400
        var grain = Grains.findOne(this.grainId, {                                                                     // 402
          fields: {                                                                                                    // 402
            userId: 1                                                                                                  // 402
          }                                                                                                            // 402
        });                                                                                                            // 402
        result.isDemoUser = Meteor.users.findOne(grain.userId).expires ? true : false;                                 // 403
        return result;                                                                                                 // 405
      }.bind(this));                                                                                                   // 406
    }                                                                                                                  // 407
                                                                                                                       //
    return getPublicId;                                                                                                // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.httpGet = function () {                                                             // 318
    function httpGet(url) {                                                                                            // 318
      var _this = this;                                                                                                // 410
                                                                                                                       //
      var session = _this;                                                                                             // 411
      return inMeteor(function () {                                                                                    // 413
        return ssrfSafeLookup(globalDb, url);                                                                          // 414
      }).then(function (safe) {                                                                                        // 415
        return new Promise(function (resolve, reject) {                                                                // 416
          var requestMethod = Http.request;                                                                            // 417
                                                                                                                       //
          if (safe.url.indexOf("https://") === 0) {                                                                    // 418
            requestMethod = Https.request;                                                                             // 419
          } else if (safe.url.indexOf("http://") !== 0) {                                                              // 420
            err = new Error("Protocol not recognized.");                                                               // 421
            err.nature = "precondition";                                                                               // 422
            reject(err);                                                                                               // 423
          }                                                                                                            // 424
                                                                                                                       //
          var options = Url.parse(safe.url);                                                                           // 426
          options.headers = {                                                                                          // 427
            host: safe.host                                                                                            // 427
          };                                                                                                           // 427
          options.servername = safe.host.split(":")[0];                                                                // 428
          req = requestMethod(options, function (resp) {                                                               // 430
            var buffers = [];                                                                                          // 431
            var err = void 0;                                                                                          // 432
                                                                                                                       //
            switch (Math.floor(resp.statusCode / 100)) {                                                               // 434
              case 2:                                                                                                  // 435
                // 2xx response -- OK.                                                                                 // 435
                resp.on("data", function (buf) {                                                                       // 436
                  buffers.push(buf);                                                                                   // 437
                });                                                                                                    // 438
                resp.on("end", function () {                                                                           // 440
                  resolve({                                                                                            // 441
                    content: Buffer.concat(buffers),                                                                   // 442
                    mimeType: resp.headers["content-type"] || null                                                     // 443
                  });                                                                                                  // 441
                });                                                                                                    // 445
                break;                                                                                                 // 446
                                                                                                                       //
              case 3:                                                                                                  // 447
                // 3xx response -- redirect.                                                                           // 447
                resolve(session.httpGet(resp.headers.location));                                                       // 448
                break;                                                                                                 // 449
                                                                                                                       //
              case 4:                                                                                                  // 450
                // 4xx response -- client error.                                                                       // 450
                err = new Error("Status code " + resp.statusCode + " received in response.");                          // 451
                err.nature = "precondition";                                                                           // 452
                reject(err);                                                                                           // 453
                break;                                                                                                 // 454
                                                                                                                       //
              case 5:                                                                                                  // 455
                // 5xx response -- internal server error.                                                              // 455
                err = new Error("Status code " + resp.statusCode + " received in response.");                          // 456
                err.nature = "localBug";                                                                               // 457
                reject(err);                                                                                           // 458
                break;                                                                                                 // 459
                                                                                                                       //
              default:                                                                                                 // 460
                // ???                                                                                                 // 460
                err = new Error("Invalid status code " + resp.statusCode + " received in response.");                  // 461
                err.nature = "localBug";                                                                               // 462
                reject(err);                                                                                           // 463
                break;                                                                                                 // 464
            }                                                                                                          // 434
          });                                                                                                          // 466
          req.on("error", function (e) {                                                                               // 468
            e.nature = "networkFailure";                                                                               // 469
            reject(e);                                                                                                 // 470
          });                                                                                                          // 471
          req.setTimeout(15000, function () {                                                                          // 473
            req.abort();                                                                                               // 474
            err = new Error("Request timed out.");                                                                     // 475
            err.nature = "localBug";                                                                                   // 476
            err.durability = "overloaded";                                                                             // 477
            reject(err);                                                                                               // 478
          });                                                                                                          // 479
          req.end();                                                                                                   // 481
        });                                                                                                            // 482
      });                                                                                                              // 483
    }                                                                                                                  // 484
                                                                                                                       //
    return httpGet;                                                                                                    // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.getUserAddress = function () {                                                      // 318
    function getUserAddress() {                                                                                        // 318
      return inMeteor(function () {                                                                                    // 487
        return this._getUserAddress();                                                                                 // 488
      }.bind(this));                                                                                                   // 489
    }                                                                                                                  // 490
                                                                                                                       //
    return getUserAddress;                                                                                             // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteGenerateApiToken = function () {                                            // 318
    function obsoleteGenerateApiToken(petname, userInfo, expires) {                                                    // 318
      throw new Error("generateApiToken() has been removed. Use offer templates instead.");                            // 493
    }                                                                                                                  // 494
                                                                                                                       //
    return obsoleteGenerateApiToken;                                                                                   // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteListApiTokens = function () {                                               // 318
    function obsoleteListApiTokens() {                                                                                 // 318
      throw new Error("listApiTokens() has been removed. Use offer templates instead.");                               // 497
    }                                                                                                                  // 498
                                                                                                                       //
    return obsoleteListApiTokens;                                                                                      // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteRevokeApiToken = function () {                                              // 318
    function obsoleteRevokeApiToken(tokenId) {                                                                         // 318
      throw new Error("revokeApiToken() has been removed. Use offer templates instead.");                              // 501
    }                                                                                                                  // 502
                                                                                                                       //
    return obsoleteRevokeApiToken;                                                                                     // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  HackSessionContextImpl.prototype.getUiViewForEndpoint = function () {                                                // 318
    function getUiViewForEndpoint(url) {                                                                               // 318
      var parsedUrl = Url.parse(url);                                                                                  // 505
                                                                                                                       //
      if (parsedUrl.hash) {                                                                                            // 507
        // Assume that anything with a fragment is a webkey                                                            // 507
        if (parsedUrl.pathname && parsedUrl.pathname !== "/") {                                                        // 508
          throw new Error("Webkey urls cannot contain a path.");                                                       // 509
        }                                                                                                              // 510
                                                                                                                       //
        var token = parsedUrl.hash.slice(1); // Get rid of # which is always the first character                       // 512
                                                                                                                       //
        var hostId = matchWildcardHost(parsedUrl.host); // Connecting to a remote server with a bearer token.          // 513
        // TODO(someday): Negotiate server-to-server Cap'n Proto connection.                                           // 515
                                                                                                                       //
        return {                                                                                                       // 516
          view: new ExternalUiView(url, token)                                                                         // 516
        };                                                                                                             // 516
      } else {                                                                                                         // 517
        return {                                                                                                       // 518
          view: new ExternalUiView(url)                                                                                // 518
        };                                                                                                             // 518
      }                                                                                                                // 519
    }                                                                                                                  // 520
                                                                                                                       //
    return getUiViewForEndpoint;                                                                                       // 318
  }();                                                                                                                 // 318
                                                                                                                       //
  return HackSessionContextImpl;                                                                                       // 318
}(SessionContextImpl);                                                                                                 // 318
                                                                                                                       //
makeHackSessionContext = function (grainId, sessionId, accountId, tabId) {                                             // 523
  return new Capnp.Capability(new HackSessionContextImpl(grainId, sessionId, accountId, tabId), HackSessionContext);   // 524
};                                                                                                                     // 526
                                                                                                                       //
var HOSTNAME_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789";                                                           // 528
                                                                                                                       //
generateRandomHostname = function (length) {                                                                           // 530
  // Generate a random unique name suitable for use in a hostname.                                                     // 531
  var digits = [];                                                                                                     // 533
                                                                                                                       //
  for (var i = 0; i < length; i++) {                                                                                   // 534
    digits[i] = Random.choice(HOSTNAME_CHARS);                                                                         // 535
  }                                                                                                                    // 536
                                                                                                                       //
  return digits.join("");                                                                                              // 538
};                                                                                                                     // 539
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"identity.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/identity.js                                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var PersistentImpl = void 0;                                                                                           // 1
module.watch(require("/imports/server/persistent.js"), {                                                               // 1
  PersistentImpl: function (v) {                                                                                       // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var StaticAssetImpl = void 0,                                                                                          // 1
    IdenticonStaticAssetImpl = void 0;                                                                                 // 1
module.watch(require("/imports/server/static-asset.js"), {                                                             // 1
  StaticAssetImpl: function (v) {                                                                                      // 1
    StaticAssetImpl = v;                                                                                               // 1
  },                                                                                                                   // 1
  IdenticonStaticAssetImpl: function (v) {                                                                             // 1
    IdenticonStaticAssetImpl = v;                                                                                      // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Capnp = Npm.require("capnp");                                                                                      // 17
                                                                                                                       //
var IdentityRpc = Capnp.importSystem("sandstorm/identity-impl.capnp");                                                 // 18
var Identity = Capnp.importSystem("sandstorm/identity.capnp").Identity;                                                // 19
var StaticAsset = Capnp.importSystem("sandstorm/util.capnp").StaticAsset;                                              // 22
                                                                                                                       //
var IdentityImpl = function (_PersistentImpl) {                                                                        //
  (0, _inherits3.default)(IdentityImpl, _PersistentImpl);                                                              //
                                                                                                                       //
  function IdentityImpl(db, saveTemplate, accountId) {                                                                 // 25
    (0, _classCallCheck3.default)(this, IdentityImpl);                                                                 // 25
                                                                                                                       //
    var _this = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));          // 25
                                                                                                                       //
    _this.accountId = accountId;                                                                                       // 27
    _this.db = db;                                                                                                     // 28
    return _this;                                                                                                      // 25
  }                                                                                                                    // 29
                                                                                                                       //
  IdentityImpl.prototype.getProfile = function () {                                                                    //
    function getProfile() {                                                                                            //
      var user = Meteor.users.findOne({                                                                                // 32
        _id: this.accountId                                                                                            // 32
      });                                                                                                              // 32
      var profile = {                                                                                                  // 34
        displayName: {                                                                                                 // 35
          defaultText: user.profile.name                                                                               // 35
        },                                                                                                             // 35
        preferredHandle: user.profile.handle,                                                                          // 36
        pronouns: user.profile.pronoun                                                                                 // 37
      };                                                                                                               // 34
                                                                                                                       //
      if (user.profile.picture) {                                                                                      // 40
        profile.picture = new Capnp.Capability(new StaticAssetImpl(user.profile.picture), StaticAsset);                // 41
      } else {                                                                                                         // 43
        var hash = user.profile.identicon;                                                                             // 44
        profile.picture = new Capnp.Capability(new IdenticonStaticAssetImpl(hash, 24), StaticAsset);                   // 45
      }                                                                                                                // 47
                                                                                                                       //
      return {                                                                                                         // 49
        profile: profile                                                                                               // 49
      };                                                                                                               // 49
    }                                                                                                                  // 50
                                                                                                                       //
    return getProfile;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  return IdentityImpl;                                                                                                 //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Find a better home for this.                                                                       // 51
                                                                                                                       //
var MembraneRequirement = Match.OneOf({                                                                                // 54
  tokenValid: String                                                                                                   // 55
}, {                                                                                                                   // 55
  permissionsHeld: {                                                                                                   // 56
    accountId: String,                                                                                                 // 57
    grainId: String,                                                                                                   // 57
    permissions: Match.Optional([Boolean])                                                                             // 57
  }                                                                                                                    // 57
}, {                                                                                                                   // 56
  permissionsHeld: {                                                                                                   // 58
    tokenId: String,                                                                                                   // 59
    grainId: String,                                                                                                   // 59
    permissions: Match.Optional([Boolean])                                                                             // 59
  }                                                                                                                    // 59
}, {                                                                                                                   // 58
  userIsAdmin: String                                                                                                  // 60
});                                                                                                                    // 60
                                                                                                                       //
makeIdentity = function (accountId, requirements) {                                                                    // 62
  var saveTemplate = {                                                                                                 // 63
    frontendRef: {                                                                                                     // 63
      identity: accountId                                                                                              // 63
    }                                                                                                                  // 63
  };                                                                                                                   // 63
                                                                                                                       //
  if (requirements) {                                                                                                  // 64
    check(requirements, [MembraneRequirement]);                                                                        // 65
    saveTemplate.requirements = requirements;                                                                          // 66
  }                                                                                                                    // 67
                                                                                                                       //
  return new Capnp.Capability(new IdentityImpl(globalDb, saveTemplate, accountId), IdentityRpc.PersistentIdentity);    // 69
};                                                                                                                     // 71
                                                                                                                       //
globalFrontendRefRegistry.register({                                                                                   // 73
  frontendRefField: "identity",                                                                                        // 74
  typeId: Identity.typeId,                                                                                             // 75
  restore: function (db, saveTemplate, accountId) {                                                                    // 77
    return new Capnp.Capability(new IdentityImpl(db, saveTemplate, accountId), IdentityRpc.PersistentIdentity);        // 78
  },                                                                                                                   // 80
  validate: function (db, session, value) {                                                                            // 82
    check(value, {                                                                                                     // 83
      id: String,                                                                                                      // 83
      roleAssignment: SandstormDb.prototype.roleAssignmentPattern                                                      // 83
    });                                                                                                                // 83
                                                                                                                       //
    if (!session.userId) {                                                                                             // 85
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 86
    }                                                                                                                  // 87
                                                                                                                       //
    var grain = db.getGrain(session.grainId);                                                                          // 89
    SandstormPermissions.createNewApiToken(db, {                                                                       // 90
      accountId: session.userId                                                                                        // 92
    }, session.grainId, "petname", value.roleAssignment, {                                                             // 92
      user: {                                                                                                          // 96
        accountId: value.id,                                                                                           // 96
        title: grain.title                                                                                             // 96
      }                                                                                                                // 96
    }); // TODO(soon): Somehow notify this user that they now have access.                                             // 96
    // TODO(perf): This permissions computation happens once here and then once again when the                         // 100
    //   `permissionsHeld` requirement is checked. Is there a way to avoid the duplicated work?                        // 101
                                                                                                                       //
    var permissions = SandstormPermissions.grainPermissions(db, {                                                      // 102
      grain: {                                                                                                         // 103
        _id: session.grainId,                                                                                          // 103
        accountId: value.id                                                                                            // 103
      }                                                                                                                // 103
    }, session.viewInfo || {}).permissions;                                                                            // 103
    return {                                                                                                           // 106
      descriptor: {                                                                                                    // 107
        tags: [{                                                                                                       // 108
          id: Identity.typeId,                                                                                         // 110
          value: Capnp.serialize(Identity.PowerboxTag, {                                                               // 111
            permissions: permissions                                                                                   // 114
          })                                                                                                           // 113
        }]                                                                                                             // 109
      },                                                                                                               // 107
      requirements: [{                                                                                                 // 119
        permissionsHeld: {                                                                                             // 121
          accountId: value.id,                                                                                         // 122
          grainId: session.grainId,                                                                                    // 123
          permissions: []                                                                                              // 124
        }                                                                                                              // 121
      }],                                                                                                              // 120
      frontendRef: value.id                                                                                            // 128
    };                                                                                                                 // 106
  },                                                                                                                   // 130
  query: function (db, userId, value) {                                                                                // 132
    var resultSet = {};                                                                                                // 133
    var requestedPermissions = [];                                                                                     // 134
                                                                                                                       //
    if (value) {                                                                                                       // 135
      requestedPermissions = Capnp.parse(Identity.PowerboxTag, value).permissions || [];                               // 136
    }                                                                                                                  // 137
                                                                                                                       //
    var resultForUser = function (user) {                                                                              // 139
      SandstormDb.fillInPictureUrl(user);                                                                              // 140
      return {                                                                                                         // 141
        _id: "frontendref-identity-" + user._id,                                                                       // 142
        frontendRef: {                                                                                                 // 143
          identity: user._id                                                                                           // 143
        },                                                                                                             // 143
        cardTemplate: "identityPowerboxCard",                                                                          // 144
        configureTemplate: "identityPowerboxConfiguration",                                                            // 145
        profile: user.profile,                                                                                         // 146
        requestedPermissions: requestedPermissions,                                                                    // 147
        searchTerms: [user.profile.name, user.profile.handle]                                                          // 148
      };                                                                                                               // 141
    };                                                                                                                 // 154
                                                                                                                       //
    db.collections.contacts.find({                                                                                     // 156
      ownerId: userId                                                                                                  // 156
    }).forEach(function (contact) {                                                                                    // 156
      var user = Meteor.users.findOne({                                                                                // 157
        _id: contact.accountId                                                                                         // 157
      });                                                                                                              // 157
                                                                                                                       //
      if (user) {                                                                                                      // 158
        resultSet[user._id] = resultForUser(user);                                                                     // 159
      }                                                                                                                // 160
    });                                                                                                                // 161
                                                                                                                       //
    if (db.getOrganizationShareContacts() && db.isUserInOrganization(db.collections.users.findOne({                    // 163
      _id: userId                                                                                                      // 164
    }))) {                                                                                                             // 164
      // TODO(perf): Add some way to efficiently fetch all members in an organization.                                 // 166
      db.collections.users.find({                                                                                      // 167
        type: "credential"                                                                                             // 167
      }).forEach(function (credential) {                                                                               // 167
        if (db.isCredentialInOrganization(credential)) {                                                               // 168
          var user = Meteor.users.findOne({                                                                            // 169
            "loginCredentials.id": credential._id                                                                      // 169
          });                                                                                                          // 169
                                                                                                                       //
          if (user) {                                                                                                  // 170
            resultSet[user._id] = resultForUser(user);                                                                 // 171
          }                                                                                                            // 172
        }                                                                                                              // 173
      });                                                                                                              // 174
    }                                                                                                                  // 175
                                                                                                                       //
    return _.values(resultSet);                                                                                        // 177
  }                                                                                                                    // 178
});                                                                                                                    // 73
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"install-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/install-server.js                                                                                            //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var allowDemo = void 0;                                                                                                // 1
module.watch(require("/imports/demo.js"), {                                                                            // 1
  allowDemo: function (v) {                                                                                            // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var promiseToFuture = void 0;                                                                                          // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  promiseToFuture: function (v) {                                                                                      // 1
    promiseToFuture = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var TOKEN_CLEANUP_MINUTES = 120; // Give enough time for large uploads on slow connections.                            // 20
                                                                                                                       //
var TOKEN_CLEANUP_TIMER = TOKEN_CLEANUP_MINUTES * 60 * 1000;                                                           // 21
                                                                                                                       //
function cleanupToken(tokenId) {                                                                                       // 23
  check(tokenId, String);                                                                                              // 24
  globalDb.collections.spkTokens.remove({                                                                              // 25
    _id: tokenId                                                                                                       // 25
  });                                                                                                                  // 25
  waitPromise(globalBackend.cap().deleteBackup(tokenId));                                                              // 26
}                                                                                                                      // 27
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 29
  // Cleanup tokens every TOKEN_CLEANUP_MINUTES                                                                        // 30
  SandstormDb.periodicCleanup(TOKEN_CLEANUP_TIMER, function () {                                                       // 31
    var queryDate = new Date(Date.now() - TOKEN_CLEANUP_TIMER);                                                        // 32
    globalDb.collections.spkTokens.find({                                                                              // 34
      timestamp: {                                                                                                     // 34
        $lt: queryDate                                                                                                 // 34
      }                                                                                                                // 34
    }).forEach(function (token) {                                                                                      // 34
      cleanupToken(token._id);                                                                                         // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
});                                                                                                                    // 38
var localizedTextPattern = {                                                                                           // 40
  defaultText: String,                                                                                                 // 41
  localizations: Match.Optional([{                                                                                     // 42
    locale: String,                                                                                                    // 42
    text: String                                                                                                       // 42
  }])                                                                                                                  // 42
}; // Not all users are allowed to upload apps. We need to manually implement authorization                            // 40
// because Meteor.userId() is not available in server-side routes.                                                     // 46
                                                                                                                       //
Meteor.methods({                                                                                                       // 48
  cancelDownload: function (packageId) {                                                                               // 49
    check(packageId, String); // TODO(security):  Only let user cancel download if they initiated it.                  // 50
                                                                                                                       //
    cancelDownload(packageId);                                                                                         // 53
  },                                                                                                                   // 54
  newUploadToken: function () {                                                                                        // 56
    if (!isSignedUp()) {                                                                                               // 57
      throw new Meteor.Error(403, "Unauthorized", "Only invited users can upload apps.");                              // 58
    }                                                                                                                  // 59
                                                                                                                       //
    if (globalDb.isUninvitedFreeUser()) {                                                                              // 61
      throw new Meteor.Error(403, "Unauthorized", "Only paid users can upload apps.");                                 // 62
    }                                                                                                                  // 63
                                                                                                                       //
    var token = Random.id(22);                                                                                         // 65
    globalDb.collections.spkTokens.insert({                                                                            // 66
      _id: token,                                                                                                      // 66
      timestamp: new Date()                                                                                            // 66
    });                                                                                                                // 66
    return token;                                                                                                      // 68
  },                                                                                                                   // 69
  upgradeGrains: function (appId, version, packageId) {                                                                // 71
    this.connection.sandstormDb.upgradeGrains(appId, version, packageId, globalBackend);                               // 72
  },                                                                                                                   // 73
  ensureInstalled: function (packageId, url, isRetry) {                                                                // 75
    check(packageId, String);                                                                                          // 76
    check(url, Match.OneOf(String, undefined, null));                                                                  // 77
    check(isRetry, Boolean);                                                                                           // 78
                                                                                                                       //
    if (!packageId.match(/^[a-zA-Z0-9]*$/)) {                                                                          // 80
      throw new Meteor.Error(400, "The package ID contains illegal characters.");                                      // 81
    }                                                                                                                  // 82
                                                                                                                       //
    if (!this.userId) {                                                                                                // 84
      if (allowDemo && isSafeDemoAppUrl(url)) {// continue on                                                          // 85
      } else {                                                                                                         // 87
        // jscs:ignore disallowEmptyBlocks                                                                             // 87
        throw new Meteor.Error(403, "You must be logged in to install packages.");                                     // 88
      }                                                                                                                // 89
    } else if (!isSignedUpOrDemo()) {                                                                                  // 90
      throw new Meteor.Error(403, "This Sandstorm server requires you to get an invite before installing apps.");      // 91
    }                                                                                                                  // 93
                                                                                                                       //
    var pkg = Packages.findOne(packageId);                                                                             // 95
                                                                                                                       //
    if (!pkg || pkg.status !== "ready") {                                                                              // 97
      if (!this.userId || isDemoUser() || globalDb.isUninvitedFreeUser()) {                                            // 98
        if (!isSafeDemoAppUrl(url)) {                                                                                  // 99
          // TODO(someday): Billing prompt on client side.                                                             // 100
          throw new Meteor.Error(403, "Sorry, demo and free users cannot upload custom apps; " + "they may only install apps from apps.sandstorm.io.");
        }                                                                                                              // 103
      }                                                                                                                // 104
    }                                                                                                                  // 105
                                                                                                                       //
    if (!pkg || isRetry) {                                                                                             // 107
      globalDb.startInstall(packageId, url, isRetry);                                                                  // 108
    }                                                                                                                  // 109
  }                                                                                                                    // 110
});                                                                                                                    // 48
Meteor.publish("packageInfo", function (packageId) {                                                                   // 113
  check(packageId, String);                                                                                            // 114
  var db = this.connection.sandstormDb;                                                                                // 115
  var pkgCursor = db.collections.packages.find(packageId);                                                             // 116
  var pkg = pkgCursor.fetch()[0];                                                                                      // 117
                                                                                                                       //
  if (pkg && this.userId) {                                                                                            // 118
    return [pkgCursor, db.collections.userActions.find({                                                               // 119
      userId: this.userId,                                                                                             // 121
      appId: pkg.appId                                                                                                 // 121
    }), db.collections.grains.find({                                                                                   // 121
      userId: this.userId,                                                                                             // 122
      appId: pkg.appId                                                                                                 // 122
    })];                                                                                                               // 122
  } else {                                                                                                             // 124
    return pkgCursor;                                                                                                  // 125
  }                                                                                                                    // 126
});                                                                                                                    // 127
Router.map(function () {                                                                                               // 129
  this.route("upload", {                                                                                               // 130
    where: "server",                                                                                                   // 131
    path: "/upload/:token",                                                                                            // 132
    action: function () {                                                                                              // 134
      if (typeof this.params.token !== "string" || !globalDb.collections.spkTokens.findOne(this.params.token)) {       // 135
        this.response.writeHead(403, {                                                                                 // 137
          "Content-Type": "text/plain",                                                                                // 138
          "Access-Control-Allow-Origin": "*"                                                                           // 139
        });                                                                                                            // 137
        this.response.write("Invalid upload token.");                                                                  // 141
        this.response.end();                                                                                           // 142
      } else if (this.request.method === "POST") {                                                                     // 143
        try {                                                                                                          // 144
          var packageId = promiseToFuture(doClientUpload(this.request)).wait();                                        // 145
          this.response.writeHead(200, {                                                                               // 146
            "Content-Length": packageId.length,                                                                        // 147
            "Content-Type": "text/plain",                                                                              // 148
            "Access-Control-Allow-Origin": "*"                                                                         // 149
          });                                                                                                          // 146
          this.response.write(packageId);                                                                              // 151
          this.response.end();                                                                                         // 152
          globalDb.collections.spkTokens.remove(this.params.token);                                                    // 153
        } catch (error) {                                                                                              // 154
          console.error(error.stack);                                                                                  // 155
          this.response.writeHead(500, {                                                                               // 156
            "Content-Type": "text/plain",                                                                              // 157
            "Access-Control-Allow-Origin": "*"                                                                         // 158
          });                                                                                                          // 156
          this.response.write("Unpacking SPK failed; is it valid?");                                                   // 160
          this.response.end();                                                                                         // 161
        }                                                                                                              // 162
                                                                                                                       //
        ;                                                                                                              // 162
      } else if (this.request.method == "OPTIONS") {                                                                   // 163
        // Allow cross-origin posts to upload so that uploads can occur on the DDP host                                // 164
        // rather than the main host. In theory we could have Access-Control-Allow-Origin specify                      // 165
        // the main host rather than "*", but an upload request already requires a valid                               // 166
        // upload token, which is plenty of access control in itself.                                                  // 167
        var requestedHeaders = this.request.headers["access-control-request-headers"];                                 // 168
                                                                                                                       //
        if (requestedHeaders) {                                                                                        // 169
          this.response.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                   // 170
        }                                                                                                              // 171
                                                                                                                       //
        this.response.writeHead(204, {                                                                                 // 173
          "Access-Control-Allow-Origin": "*",                                                                          // 174
          "Access-Control-Allow-Methods": "POST, OPTIONS",                                                             // 175
          "Access-Control-Max-Age": "3600"                                                                             // 176
        });                                                                                                            // 173
        this.response.end();                                                                                           // 178
      } else {                                                                                                         // 179
        this.response.writeHead(405, {                                                                                 // 180
          "Content-Type": "text/plain",                                                                                // 181
          "Access-Control-Allow-Origin": "*"                                                                           // 182
        });                                                                                                            // 180
        this.response.write("You can only POST here.");                                                                // 184
        this.response.end();                                                                                           // 185
      }                                                                                                                // 186
    }                                                                                                                  // 187
  });                                                                                                                  // 130
});                                                                                                                    // 189
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"installer.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/installer.js                                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var module1 = module;                                                                                                  // 1
var Fs = void 0;                                                                                                       // 1
module1.watch(require("fs"), {                                                                                         // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Path = void 0;                                                                                                     // 1
module1.watch(require("path"), {                                                                                       // 1
  "default": function (v) {                                                                                            // 1
    Path = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module1.watch(require("crypto"), {                                                                                     // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var ChildProcess = void 0;                                                                                             // 1
module1.watch(require("child_process"), {                                                                              // 1
  "default": function (v) {                                                                                            // 1
    ChildProcess = v;                                                                                                  // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module1.watch(require("url"), {                                                                                        // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module1.watch(require("/imports/server/async-helpers.js"), {                                                           // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var ssrfSafeLookupOrProxy = void 0;                                                                                    // 1
module1.watch(require("/imports/server/networking.js"), {                                                              // 1
  ssrfSafeLookupOrProxy: function (v) {                                                                                // 1
    ssrfSafeLookupOrProxy = v;                                                                                         // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var Request = HTTPInternals.NpmModules.request.module;                                                                 // 26
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 28
                                                                                                                       //
var Manifest = Capnp.importSystem("sandstorm/package.capnp").Manifest;                                                 // 30
var installers = void 0; // set to {} on main replica                                                                  // 32
// To protect against race conditions, we require that each row in the Packages                                        // 33
// collection have at most one writer at a time, as tracked by this `installers`                                       // 34
// map. Each key is a package ID and each value is either an AppInstaller object                                       // 35
// or the string 'uninstalling', indicating that some fiber is working on                                              // 36
// uninstalling the package.                                                                                           // 37
                                                                                                                       //
var verifyIsMainReplica = function () {                                                                                // 39
  if (Meteor.settings.replicaNumber) {                                                                                 // 40
    throw new Error("This can only be called on the main front-end replica.");                                         // 41
  }                                                                                                                    // 42
};                                                                                                                     // 43
                                                                                                                       //
var deletePackageInternal = function (pkg) {                                                                           // 45
  verifyIsMainReplica();                                                                                               // 46
  var packageId = pkg._id;                                                                                             // 48
                                                                                                                       //
  if (packageId in installers) {                                                                                       // 50
    return;                                                                                                            // 51
  }                                                                                                                    // 52
                                                                                                                       //
  installers[packageId] = "uninstalling";                                                                              // 54
                                                                                                                       //
  try {                                                                                                                // 56
    var action = UserActions.findOne({                                                                                 // 57
      packageId: packageId                                                                                             // 57
    });                                                                                                                // 57
    var grain = Grains.findOne({                                                                                       // 58
      packageId: packageId                                                                                             // 58
    });                                                                                                                // 58
    var notificationQuery = {};                                                                                        // 59
    notificationQuery["appUpdates." + pkg.appId + ".packageId"] = packageId;                                           // 60
                                                                                                                       //
    if (!grain && !action && !Notifications.findOne(notificationQuery) && !globalDb.getAppIdForPreinstalledPackage(packageId)) {
      Packages.update({                                                                                                // 63
        _id: packageId                                                                                                 // 64
      }, {                                                                                                             // 63
        $set: {                                                                                                        // 66
          status: "delete"                                                                                             // 66
        },                                                                                                             // 66
        $unset: {                                                                                                      // 67
          shouldCleanup: ""                                                                                            // 67
        }                                                                                                              // 67
      });                                                                                                              // 65
      waitPromise(globalBackend.cap().deletePackage(packageId));                                                       // 69
      Packages.remove(packageId); // Clean up assets (icon, etc).                                                      // 70
                                                                                                                       //
      getAllManifestAssets(pkg.manifest).forEach(function (assetId) {                                                  // 73
        globalDb.unrefStaticAsset(assetId);                                                                            // 74
      });                                                                                                              // 75
    } else {                                                                                                           // 76
      Packages.update({                                                                                                // 77
        _id: packageId                                                                                                 // 77
      }, {                                                                                                             // 77
        $unset: {                                                                                                      // 77
          shouldCleanup: ""                                                                                            // 77
        }                                                                                                              // 77
      });                                                                                                              // 77
    }                                                                                                                  // 78
                                                                                                                       //
    delete installers[packageId];                                                                                      // 80
  } catch (error) {                                                                                                    // 81
    delete installers[packageId];                                                                                      // 82
    throw error;                                                                                                       // 83
  }                                                                                                                    // 84
};                                                                                                                     // 85
                                                                                                                       //
var startInstallInternal = function (pkg) {                                                                            // 87
  verifyIsMainReplica();                                                                                               // 88
                                                                                                                       //
  if (pkg._id in installers) {                                                                                         // 90
    return;                                                                                                            // 91
  }                                                                                                                    // 92
                                                                                                                       //
  var installer = new AppInstaller(pkg._id, pkg.url, pkg.appId, pkg.isAutoUpdated);                                    // 94
  installers[pkg._id] = installer;                                                                                     // 95
  installer.start();                                                                                                   // 96
};                                                                                                                     // 97
                                                                                                                       //
cancelDownload = function (packageId) {                                                                                // 99
  Packages.remove({                                                                                                    // 100
    _id: packageId,                                                                                                    // 100
    status: "download"                                                                                                 // 100
  });                                                                                                                  // 100
};                                                                                                                     // 101
                                                                                                                       //
var cancelDownloadInternal = function (pkg) {                                                                          // 103
  verifyIsMainReplica();                                                                                               // 104
  var installer = installers[pkg._id]; // Don't do anything unless a download is in progress.                          // 106
                                                                                                                       //
  if (installer && installer.downloadRequest) {                                                                        // 109
    // OK, effect cancellation by faking an error.                                                                     // 110
    installer.wrapCallback(function () {                                                                               // 111
      throw new Error("Canceled");                                                                                     // 112
    })();                                                                                                              // 113
  }                                                                                                                    // 114
};                                                                                                                     // 115
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 117
  installers = {};                                                                                                     // 118
  Meteor.startup(function () {                                                                                         // 120
    // Restart any deletions that were killed while in-progress.                                                       // 121
    Packages.find({                                                                                                    // 122
      status: "delete"                                                                                                 // 122
    }).forEach(deletePackageInternal); // Watch for new installation requests and fulfill them.                        // 122
                                                                                                                       //
    Packages.find({                                                                                                    // 125
      status: {                                                                                                        // 125
        $in: ["download", "verify", "unpack", "analyze"]                                                               // 125
      }                                                                                                                // 125
    }).observe({                                                                                                       // 125
      added: startInstallInternal,                                                                                     // 126
      removed: cancelDownloadInternal                                                                                  // 127
    }); // Watch for new cleanup requests and fulfill them.                                                            // 125
                                                                                                                       //
    Packages.find({                                                                                                    // 131
      status: "ready",                                                                                                 // 131
      shouldCleanup: true                                                                                              // 131
    }).observe({                                                                                                       // 131
      added: deletePackageInternal                                                                                     // 132
    });                                                                                                                // 131
  });                                                                                                                  // 134
}                                                                                                                      // 135
                                                                                                                       //
doClientUpload = function (stream) {                                                                                   // 137
  return new Promise(function (resolve, reject) {                                                                      // 138
    var id = Random.id();                                                                                              // 139
    var backendStream = globalBackend.cap().installPackage().stream;                                                   // 141
    var hasher = Crypto.createHash("sha256");                                                                          // 142
    stream.on("data", function (chunk) {                                                                               // 144
      try {                                                                                                            // 145
        hasher.update(chunk);                                                                                          // 146
        backendStream.write(chunk);                                                                                    // 147
      } catch (err) {                                                                                                  // 148
        reject(err);                                                                                                   // 149
      }                                                                                                                // 150
    });                                                                                                                // 151
    stream.on("end", function () {                                                                                     // 153
      try {                                                                                                            // 154
        backendStream.done();                                                                                          // 155
        var packageId = hasher.digest("hex").slice(0, 32);                                                             // 156
        resolve(backendStream.saveAs(packageId).then(function () {                                                     // 157
          return packageId;                                                                                            // 158
        }));                                                                                                           // 159
        backendStream.close();                                                                                         // 160
      } catch (err) {                                                                                                  // 161
        reject(err);                                                                                                   // 162
      }                                                                                                                // 163
    });                                                                                                                // 164
    stream.on("error", function (err) {                                                                                // 166
      // TODO(soon):  This event does't seem to fire if the user leaves the page mid-upload.                           // 167
      try {                                                                                                            // 168
        backendStream.close();                                                                                         // 169
        reject(err);                                                                                                   // 170
      } catch (err2) {                                                                                                 // 171
        reject(err2);                                                                                                  // 172
      }                                                                                                                // 173
    });                                                                                                                // 174
  });                                                                                                                  // 175
};                                                                                                                     // 176
                                                                                                                       //
AppInstaller = function () {                                                                                           // 178
  function AppInstaller(packageId, url, appId, isAutoUpdated) {                                                        // 179
    (0, _classCallCheck3.default)(this, AppInstaller);                                                                 // 179
    verifyIsMainReplica();                                                                                             // 180
    this.packageId = packageId;                                                                                        // 182
    this.url = url;                                                                                                    // 183
    this.failed = false;                                                                                               // 184
    this.appId = appId;                                                                                                // 185
    this.isAutoUpdated = isAutoUpdated; // Serializes database writes.                                                 // 186
                                                                                                                       //
    this.writeChain = Promise.resolve();                                                                               // 189
  }                                                                                                                    // 190
                                                                                                                       //
  AppInstaller.prototype.updateProgress = function () {                                                                // 178
    function updateProgress(status, progress, error, manifest) {                                                       // 178
      // TODO(security):  On error, we should actually delete the package from the database and only                   // 193
      //   display the error to whomever was watching at the time.  Otherwise it's easy to confuse                     // 194
      //   people by 'pre-failing' packages.  (Actually, perhaps if a user tries to download an                        // 195
      //   already-downloading package but specifies a different URL, we really should initiate an                     // 196
      //   entirely separate download...  but cancel it if the first download succeeds.)                               // 197
      this.status = status;                                                                                            // 199
      this.progress = progress || -1;                                                                                  // 200
      this.error = error;                                                                                              // 201
      this.manifest = manifest || null;                                                                                // 202
                                                                                                                       //
      var _this = this; // The callback passed to inMeteor() runs in a new fiber. We need to make sure database writes
      // occur in exactly the order in which we generate them, so we use a promise chain to serialize                  // 207
      // them.                                                                                                         // 208
                                                                                                                       //
                                                                                                                       //
      this.writeChain = this.writeChain.then(function () {                                                             // 209
        return inMeteor(function () {                                                                                  // 210
          if (manifest) extractManifestAssets(manifest);                                                               // 211
          Packages.update(_this.packageId, {                                                                           // 213
            $set: {                                                                                                    // 214
              status: _this.status,                                                                                    // 215
              progress: _this.progress,                                                                                // 216
              error: _this.error ? _this.error.message : null,                                                         // 217
              manifest: _this.manifest,                                                                                // 218
              appId: _this.appId,                                                                                      // 219
              authorPgpKeyFingerprint: _this.authorPgpKeyFingerprint                                                   // 220
            }                                                                                                          // 214
          });                                                                                                          // 213
                                                                                                                       //
          if (_this.authorPgpKeyFingerprint) {                                                                         // 224
            globalDb.updateKeybaseProfileAsync(_this.authorPgpKeyFingerprint);                                         // 225
          }                                                                                                            // 226
        }).catch(function (err) {                                                                                      // 227
          console.error(err.stack);                                                                                    // 228
        });                                                                                                            // 229
      });                                                                                                              // 230
    }                                                                                                                  // 231
                                                                                                                       //
    return updateProgress;                                                                                             // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.wrapCallback = function () {                                                                  // 178
    function wrapCallback(method) {                                                                                    // 178
      // Note that the function below must not be an arrow function, since arrow functions do not have                 // 234
      // access to the context's arguments array.                                                                      // 235
      var _this = this;                                                                                                // 236
                                                                                                                       //
      return function () {                                                                                             // 237
        if (_this.failed) return;                                                                                      // 238
                                                                                                                       //
        try {                                                                                                          // 239
          return method.apply(_this, _.toArray(arguments));                                                            // 240
        } catch (err) {                                                                                                // 241
          _this.failed = true;                                                                                         // 242
                                                                                                                       //
          _this.cleanup();                                                                                             // 243
                                                                                                                       //
          _this.updateProgress("failed", 0, err);                                                                      // 244
                                                                                                                       //
          _this.writeChain = _this.writeChain.then(function () {                                                       // 245
            delete installers[_this.packageId];                                                                        // 246
          });                                                                                                          // 247
          console.error("Failed to install app:", err.stack);                                                          // 248
        }                                                                                                              // 249
      };                                                                                                               // 250
    }                                                                                                                  // 251
                                                                                                                       //
    return wrapCallback;                                                                                               // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.cleanup = function () {                                                                       // 178
    function cleanup() {                                                                                               // 178
      if (this.uploadStream) {                                                                                         // 254
        try {                                                                                                          // 255
          this.uploadStream.close();                                                                                   // 256
        } catch (err) {}                                                                                               // 257
                                                                                                                       //
        delete this.uploadStream;                                                                                      // 259
      }                                                                                                                // 260
                                                                                                                       //
      if (this.downloadRequest) {                                                                                      // 262
        try {                                                                                                          // 263
          this.downloadRequest.abort();                                                                                // 264
        } catch (err) {}                                                                                               // 265
                                                                                                                       //
        delete this.downloadRequest;                                                                                   // 267
      }                                                                                                                // 268
    }                                                                                                                  // 269
                                                                                                                       //
    return cleanup;                                                                                                    // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.start = function () {                                                                         // 178
    function start() {                                                                                                 // 178
      var _this2 = this;                                                                                               // 271
                                                                                                                       //
      return this.wrapCallback(function () {                                                                           // 272
        _this2.cleanup();                                                                                              // 273
                                                                                                                       //
        globalBackend.cap().tryGetPackage(_this2.packageId).then(_this2.wrapCallback(function (info) {                 // 275
          if (info.appId) {                                                                                            // 276
            _this2.appId = info.appId;                                                                                 // 277
            _this2.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;                                             // 278
                                                                                                                       //
            _this2.done(info.manifest);                                                                                // 279
          } else {                                                                                                     // 280
            _this2.doDownload();                                                                                       // 281
          }                                                                                                            // 282
        }), _this2.wrapCallback(function (err) {                                                                       // 283
          throw err;                                                                                                   // 284
        }));                                                                                                           // 285
      })();                                                                                                            // 286
    }                                                                                                                  // 287
                                                                                                                       //
    return start;                                                                                                      // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.doDownload = function () {                                                                    // 178
    function doDownload() {                                                                                            // 178
      if (!this.url) {                                                                                                 // 290
        throw new Error("Unknown package ID, and no URL was provided.");                                               // 291
      }                                                                                                                // 292
                                                                                                                       //
      console.log("Downloading app:", this.url);                                                                       // 294
      this.updateProgress("download");                                                                                 // 295
      this.uploadStream = globalBackend.cap().installPackage().stream;                                                 // 297
      return this.doDownloadTo(this.uploadStream);                                                                     // 298
    }                                                                                                                  // 299
                                                                                                                       //
    return doDownload;                                                                                                 // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.doDownloadTo = function () {                                                                  // 178
    function doDownloadTo(out) {                                                                                       // 178
      inMeteor(this.wrapCallback(function () {                                                                         // 302
        var _this3 = this;                                                                                             // 302
                                                                                                                       //
        var safe = ssrfSafeLookupOrProxy(globalDb, this.url);                                                          // 303
        var bytesExpected = undefined;                                                                                 // 305
        var bytesReceived = 0;                                                                                         // 306
        var hasher = Crypto.createHash("sha256");                                                                      // 307
        var done = false;                                                                                              // 308
                                                                                                                       //
        var updateDownloadProgress = _.throttle(this.wrapCallback(function () {                                        // 309
          if (!done) {                                                                                                 // 310
            if (bytesExpected) {                                                                                       // 311
              _this3.updateProgress("download", bytesReceived / bytesExpected);                                        // 312
            } else {                                                                                                   // 313
              _this3.updateProgress("download", bytesReceived);                                                        // 314
            }                                                                                                          // 315
          }                                                                                                            // 316
        }), 500);                                                                                                      // 317
                                                                                                                       //
        var request = safe.proxy ? Request.get(this.url, {                                                             // 319
          proxy: safe.proxy                                                                                            // 320
        }) : Request.get(safe.url, {                                                                                   // 320
          headers: {                                                                                                   // 322
            host: safe.host                                                                                            // 322
          },                                                                                                           // 322
          servername: safe.host.split(":")[0]                                                                          // 323
        });                                                                                                            // 321
        request.on("response", this.wrapCallback(function (response) {                                                 // 326
          if ("content-length" in response.headers) {                                                                  // 327
            bytesExpected = parseInt(response.headers["content-length"]);                                              // 328
          }                                                                                                            // 329
        }));                                                                                                           // 330
        request.on("data", this.wrapCallback(function (chunk) {                                                        // 332
          hasher.update(chunk);                                                                                        // 333
          out.write(chunk);                                                                                            // 334
          bytesReceived += chunk.length;                                                                               // 335
          updateDownloadProgress();                                                                                    // 336
        }));                                                                                                           // 337
        request.on("end", this.wrapCallback(function () {                                                              // 339
          out.done();                                                                                                  // 340
                                                                                                                       //
          if (hasher.digest("hex").slice(0, 32) !== _this3.packageId) {                                                // 342
            throw new Error("Package hash did not match.");                                                            // 343
          }                                                                                                            // 344
                                                                                                                       //
          done = true;                                                                                                 // 346
          delete _this3.downloadRequest;                                                                               // 347
                                                                                                                       //
          _this3.updateProgress("unpack");                                                                             // 349
                                                                                                                       //
          out.saveAs(_this3.packageId).then(_this3.wrapCallback(function (info) {                                      // 350
            _this3.appId = info.appId;                                                                                 // 351
            _this3.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;                                             // 352
                                                                                                                       //
            _this3.done(info.manifest);                                                                                // 353
          }), _this3.wrapCallback(function (err) {                                                                     // 354
            throw err;                                                                                                 // 355
          }));                                                                                                         // 356
        }));                                                                                                           // 357
        request.on("error", this.wrapCallback(function (err) {                                                         // 359
          throw err;                                                                                                   // 359
        }));                                                                                                           // 359
        this.downloadRequest = request;                                                                                // 361
      }));                                                                                                             // 362
    }                                                                                                                  // 363
                                                                                                                       //
    return doDownloadTo;                                                                                               // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.done = function () {                                                                          // 178
    function done(manifest) {                                                                                          // 178
      console.log("App ready:", this.packageId);                                                                       // 366
      this.updateProgress("ready", 1, undefined, manifest);                                                            // 367
                                                                                                                       //
      var _this = this;                                                                                                // 368
                                                                                                                       //
      _this.writeChain = _this.writeChain.then(function () {                                                           // 369
        return inMeteor(function () {                                                                                  // 370
          if (_this.isAutoUpdated) {                                                                                   // 371
            globalDb.sendAppUpdateNotifications(_this.appId, _this.packageId, manifest.appTitle && manifest.appTitle.defaultText, manifest.appVersion, manifest.appMarketingVersion && manifest.appMarketingVersion.defaultText);
          }                                                                                                            // 375
                                                                                                                       //
          if (globalDb.getPackageIdForPreinstalledApp(_this.appId) && globalDb.collections.appIndex.findOne({          // 377
            _id: _this.appId,                                                                                          // 379
            packageId: _this.packageId                                                                                 // 380
          })) {                                                                                                        // 378
            // Only mark app as preinstall ready if its appId is in the preinstalledApps setting                       // 382
            // and if it's the latest package version in the appIndex. The updateAppIndex function                     // 383
            // will always trigger updates of preinstalled apps, even if a concurrent download of                      // 384
            // an older package is going on.                                                                           // 385
            globalDb.setPreinstallAppAsReady(_this.appId, _this.packageId);                                            // 386
          }                                                                                                            // 387
        });                                                                                                            // 388
      }).then(function () {                                                                                            // 389
        delete installers[_this.packageId];                                                                            // 390
      });                                                                                                              // 391
    }                                                                                                                  // 392
                                                                                                                       //
    return done;                                                                                                       // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  return AppInstaller;                                                                                                 // 178
}();                                                                                                                   // 178
                                                                                                                       //
extractManifestAssets = function (manifest) {                                                                          // 395
  var metadata = manifest.metadata;                                                                                    // 396
  if (!metadata) return;                                                                                               // 397
  var icons = metadata.icons;                                                                                          // 399
                                                                                                                       //
  if (icons) {                                                                                                         // 400
    var handleIcon = function (icon) {                                                                                 // 401
      if (icon.svg) {                                                                                                  // 402
        icon.assetId = globalDb.addStaticAsset({                                                                       // 403
          mimeType: "image/svg+xml"                                                                                    // 403
        }, icon.svg);                                                                                                  // 403
        icon.format = "svg";                                                                                           // 404
        delete icon.svg;                                                                                               // 405
        return true;                                                                                                   // 406
      } else if (icon.png) {                                                                                           // 407
        // Use the 1x version for 'normal' DPI, unless 1x isn't provided, in which case use 2x.                        // 408
        var normalDpi = icon.png.dpi1x || icon.png.dpi2x;                                                              // 409
        if (!normalDpi) return false;                                                                                  // 410
        icon.format = "png";                                                                                           // 411
        icon.assetId = globalDb.addStaticAsset({                                                                       // 412
          mimeType: "image/png"                                                                                        // 412
        }, normalDpi);                                                                                                 // 412
                                                                                                                       //
        if (icon.png.dpi1x && icon.png.dpi2x) {                                                                        // 414
          // Icon specifies both resolutions, so also record a 2x DPI option.                                          // 415
          icon.assetId2xDpi = globalDb.addStaticAsset({                                                                // 416
            mimeType: "image/png"                                                                                      // 416
          }, icon.png.dpi2x);                                                                                          // 416
        }                                                                                                              // 417
                                                                                                                       //
        delete icon.png;                                                                                               // 419
        return true;                                                                                                   // 420
      } else {                                                                                                         // 421
        // Unknown icon. Filter it.                                                                                    // 422
        return false;                                                                                                  // 423
      }                                                                                                                // 424
    };                                                                                                                 // 425
                                                                                                                       //
    if (icons.appGrid && !handleIcon(icons.appGrid)) delete icons.appGrid;                                             // 427
    if (icons.grain && !handleIcon(icons.grain)) delete icons.grain; // We don't need the market icons.                // 428
                                                                                                                       //
    if (icons.market) delete icons.market;                                                                             // 431
    if (icons.marketBig) delete icons.marketBig;                                                                       // 432
  }                                                                                                                    // 433
                                                                                                                       //
  var handleLocalizedText = function (text) {                                                                          // 435
    if (text.defaultText) {                                                                                            // 436
      text.defaultTextAssetId = globalDb.addStaticAsset({                                                              // 437
        mimeType: "text/plain"                                                                                         // 437
      }, text.defaultText);                                                                                            // 437
      delete text.defaultText;                                                                                         // 438
    }                                                                                                                  // 439
                                                                                                                       //
    if (text.localizations) {                                                                                          // 441
      text.localizations.forEach(function (localization) {                                                             // 442
        if (localization.text) {                                                                                       // 443
          localization.assetId = globalDb.addStaticAsset({                                                             // 444
            mimeType: "text/plain"                                                                                     // 445
          }, localization.text);                                                                                       // 445
          delete localization.text;                                                                                    // 446
        }                                                                                                              // 447
      });                                                                                                              // 448
    }                                                                                                                  // 449
  };                                                                                                                   // 450
                                                                                                                       //
  var license = metadata.license;                                                                                      // 452
                                                                                                                       //
  if (license) {                                                                                                       // 453
    if (license.proprietary) license.proprietary = handleLocalizedText(license.proprietary);                           // 454
    if (license.publicDomain) license.publicDomain = handleLocalizedText(license.proprietary);                         // 455
    if (license.notices) license.notices = handleLocalizedText(license.notices);                                       // 456
  }                                                                                                                    // 457
                                                                                                                       //
  var author = metadata.author;                                                                                        // 459
                                                                                                                       //
  if (author) {                                                                                                        // 460
    // We remove the PGP signature since it was already verified down to a key ID in the back-end.                     // 461
    if (author.pgpSignature) delete author.pgpSignature;                                                               // 462
  } // Don't need the keyring either.                                                                                  // 463
                                                                                                                       //
                                                                                                                       //
  if (metadata.pgpKeyring) delete metadata.pgpKeyring; // Perhaps used by the 'about' page?                            // 466
                                                                                                                       //
  if (metadata.description) metadata.description = handleLocalizedText(metadata.description); // Screenshots are for app marketing; we don't use them post-install.
                                                                                                                       //
  if (metadata.screenshots) delete metadata.screenshots; // We might allow the user to view the changelog.             // 472
                                                                                                                       //
  if (metadata.changeLog) metadata.changeLog = handleLocalizedText(metadata.changeLog);                                // 475
};                                                                                                                     // 476
                                                                                                                       //
getAllManifestAssets = function (manifest) {                                                                           // 478
  // Returns a list of all asset IDs in the given manifest.                                                            // 479
  var metadata = manifest.metadata;                                                                                    // 481
  if (!metadata) return [];                                                                                            // 482
  var result = [];                                                                                                     // 484
  var icons = metadata.icons;                                                                                          // 486
                                                                                                                       //
  if (icons) {                                                                                                         // 487
    var handleIcon = function (icon) {                                                                                 // 488
      if (icon.assetId) {                                                                                              // 489
        result.push(icon.assetId);                                                                                     // 490
      }                                                                                                                // 491
    };                                                                                                                 // 492
                                                                                                                       //
    if (icons.appGrid) handleIcon(icons.appGrid);                                                                      // 494
    if (icons.grain) handleIcon(icons.grain);                                                                          // 495
  }                                                                                                                    // 496
                                                                                                                       //
  var handleLocalizedText = function (text) {                                                                          // 498
    if (text.defaultTextAssetId) {                                                                                     // 499
      result.push(defaultTextAssetId);                                                                                 // 500
    }                                                                                                                  // 501
                                                                                                                       //
    if (text.localizations) {                                                                                          // 503
      text.localizations.forEach(function (localization) {                                                             // 504
        if (localization.assetId) {                                                                                    // 505
          result.push(localization.assetId);                                                                           // 506
        }                                                                                                              // 507
      });                                                                                                              // 508
    }                                                                                                                  // 509
  };                                                                                                                   // 510
                                                                                                                       //
  var license = metadata.license;                                                                                      // 512
                                                                                                                       //
  if (license) {                                                                                                       // 513
    if (license.proprietary) handleLocalizedText(license.proprietary);                                                 // 514
    if (license.publicDomain) handleLocalizedText(license.publicDomain);                                               // 515
    if (license.notices) handleLocalizedText(license.notices);                                                         // 516
  }                                                                                                                    // 517
                                                                                                                       //
  if (metadata.description) handleLocalizedText(metadata.description);                                                 // 519
  if (metadata.changeLog) handleLocalizedText(metadata.changeLog);                                                     // 520
  return result;                                                                                                       // 522
};                                                                                                                     // 523
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"notifications-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/notifications-server.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var createAppActivityDesktopNotification = void 0;                                                                     // 1
module.watch(require("/imports/server/desktop-notifications.js"), {                                                    // 1
  createAppActivityDesktopNotification: function (v) {                                                                 // 1
    createAppActivityDesktopNotification = v;                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var SupervisorCapnp = Capnp.importSystem("sandstorm/supervisor.capnp");                                                // 21
var SystemPersistent = SupervisorCapnp.SystemPersistent;                                                               // 22
                                                                                                                       //
logActivity = function (grainId, accountIdOrAnonymous, event) {                                                        // 24
  // accountIdOrAnonymous is the string "anonymous" for an anonymous user, or is null for a                            // 25
  // non-user-initiated ("background") activity.                                                                       // 26
  check(grainId, String);                                                                                              // 28
  check(accountIdOrAnonymous, Match.Maybe(String)); // `event` is always an ActivityEvent parsed from Cap'n Proto but that's too complicated to check
  // here.                                                                                                             // 31
                                                                                                                       //
  var accountId = accountIdOrAnonymous === "anonymous" ? null : accountIdOrAnonymous; // TODO(perf): A cached copy of the grain from when the session opened would be fine to use
  //   here, rather than looking it up every time.                                                                     // 36
                                                                                                                       //
  grain = Grains.findOne(grainId);                                                                                     // 37
                                                                                                                       //
  if (!grain) {                                                                                                        // 38
    // Shouldn't be possible since activity events come from the grain.                                                // 39
    throw new Error("no such grain");                                                                                  // 40
  } // Look up the event typedef.                                                                                      // 41
                                                                                                                       //
                                                                                                                       //
  var eventType = ((grain.cachedViewInfo || {}).eventTypes || [])[event.type];                                         // 44
                                                                                                                       //
  if (!eventType) {                                                                                                    // 45
    throw new Error("No such event type in app's ViewInfo: " + event.type);                                            // 46
  }                                                                                                                    // 47
                                                                                                                       //
  if (!eventType.suppressUnread) {                                                                                     // 49
    // Clear the "seenAllActivity" bit for all users except the acting user.                                           // 50
    // TODO(perf): Consider throttling? Or should that be the app's responsibility?                                    // 51
    if (accountId != grain.userId) {                                                                                   // 52
      Grains.update(grainId, {                                                                                         // 53
        $unset: {                                                                                                      // 53
          ownerSeenAllActivity: true                                                                                   // 53
        }                                                                                                              // 53
      });                                                                                                              // 53
    } // Also clear on ApiTokens.                                                                                      // 54
                                                                                                                       //
                                                                                                                       //
    ApiTokens.update({                                                                                                 // 57
      "grainId": grainId,                                                                                              // 58
      "owner.user.seenAllActivity": true,                                                                              // 59
      "owner.user.accountId": {                                                                                        // 60
        $ne: accountId                                                                                                 // 60
      }                                                                                                                // 60
    }, {                                                                                                               // 57
      $unset: {                                                                                                        // 61
        "owner.user.seenAllActivity": true                                                                             // 61
      }                                                                                                                // 61
    }, {                                                                                                               // 61
      multi: true                                                                                                      // 61
    });                                                                                                                // 61
  }                                                                                                                    // 62
                                                                                                                       //
  if (accountId) {                                                                                                     // 64
    // Apply auto-subscriptions.                                                                                       // 65
    if (eventType.autoSubscribeToGrain) {                                                                              // 66
      globalDb.subscribeToActivity(accountId, grainId);                                                                // 67
    }                                                                                                                  // 68
                                                                                                                       //
    if (event.thread && eventType.autoSubscribeToThread) {                                                             // 70
      globalDb.subscribeToActivity(accountId, grainId, event.thread.path || "");                                       // 71
    }                                                                                                                  // 72
  } // Figure out whom we need to notify.                                                                              // 73
                                                                                                                       //
                                                                                                                       //
  var notifyMap = {};                                                                                                  // 76
                                                                                                                       //
  var addRecipient = function (recipient) {                                                                            // 77
    // Mutes take priority over subscriptions.                                                                         // 78
    if (recipient.mute) {                                                                                              // 79
      notifyMap[recipient.accountId] = false;                                                                          // 80
    } else {                                                                                                           // 81
      if (!(recipient.accountId in notifyMap)) {                                                                       // 82
        notifyMap[recipient.accountId] = true;                                                                         // 83
      }                                                                                                                // 84
    }                                                                                                                  // 85
  };                                                                                                                   // 86
                                                                                                                       //
  if (accountId) {                                                                                                     // 88
    // Don't notify self.                                                                                              // 89
    addRecipient({                                                                                                     // 90
      accountId: accountId,                                                                                            // 90
      mute: true                                                                                                       // 90
    });                                                                                                                // 90
  } // Notify subscribers, if desired.                                                                                 // 91
                                                                                                                       //
                                                                                                                       //
  if (eventType.notifySubscribers) {                                                                                   // 94
    // The grain owner is implicitly subscribed.                                                                       // 95
    addRecipient({                                                                                                     // 96
      accountId: grain.userId                                                                                          // 96
    }); // Add everyone subscribed to the grain.                                                                       // 96
                                                                                                                       //
    globalDb.getActivitySubscriptions(grainId).forEach(addRecipient);                                                  // 99
                                                                                                                       //
    if (event.thread) {                                                                                                // 101
      // Add everyone subscribed to the thread.                                                                        // 102
      globalDb.getActivitySubscriptions(grainId, event.thread.path || "").forEach(addRecipient);                       // 103
    }                                                                                                                  // 104
  } // Add everyone who is mentioned.                                                                                  // 105
                                                                                                                       //
                                                                                                                       //
  if (event.users && event.users.length > 0) {                                                                         // 108
    var promises = [];                                                                                                 // 109
    event.users.forEach(function (user) {                                                                              // 110
      if (user.identity && (user.mentioned || user.subscribed)) {                                                      // 111
        promises.push(unwrapFrontendCap(user.identity, "identity", function (targetId) {                               // 112
          addRecipient({                                                                                               // 113
            accountId: targetId                                                                                        // 113
          });                                                                                                          // 113
        }));                                                                                                           // 114
      }                                                                                                                // 115
    });                                                                                                                // 116
    waitPromise(Promise.all(promises).then(function (junk) {                                                           // 117
      return undefined;                                                                                                // 117
    }));                                                                                                               // 117
  } // Make a list of everyone to notify.                                                                              // 118
                                                                                                                       //
                                                                                                                       //
  var notify = [];                                                                                                     // 121
                                                                                                                       //
  for (var _accountId in meteorBabelHelpers.sanitizeForInObject(notifyMap)) {                                          // 122
    if (notifyMap[_accountId]) {                                                                                       // 123
      notify.push(_accountId);                                                                                         // 124
    }                                                                                                                  // 125
  }                                                                                                                    // 126
                                                                                                                       //
  if (notify.length > 0) {                                                                                             // 128
    var notification = {                                                                                               // 129
      grainId: grainId,                                                                                                // 130
      path: event.path || ""                                                                                           // 131
    }; // Fields we'll update even if the notification already exists.                                                 // 129
                                                                                                                       //
    var update = {                                                                                                     // 135
      isUnread: true,                                                                                                  // 136
      timestamp: new Date()                                                                                            // 137
    };                                                                                                                 // 135
                                                                                                                       //
    if (event.thread) {                                                                                                // 140
      notification.threadPath = event.thread.path || "";                                                               // 141
    }                                                                                                                  // 142
                                                                                                                       //
    if (accountId) {                                                                                                   // 144
      notification.initiatingAccount = accountId;                                                                      // 145
    } else if (accountIdOrAnonymous) {                                                                                 // 146
      notification.initiatorAnonymous = true;                                                                          // 147
    }                                                                                                                  // 148
                                                                                                                       //
    notification.eventType = event.type;                                                                               // 150
    update.text = eventType.verbPhrase;                                                                                // 151
    var body = event.notification && event.notification.caption || {                                                   // 153
      defaultText: ""                                                                                                  // 153
    };                                                                                                                 // 153
    var appActivity = {                                                                                                // 155
      grainId: notification.grainId,                                                                                   // 156
      path: notification.path,                                                                                         // 157
      body: body,                                                                                                      // 158
      actionText: eventType.verbPhrase                                                                                 // 159
    };                                                                                                                 // 155
                                                                                                                       //
    if (accountId) {                                                                                                   // 162
      // Look up icon urls for the responsible user and the app                                                        // 163
      var account = Meteor.users.findOne({                                                                             // 164
        _id: accountId                                                                                                 // 164
      });                                                                                                              // 164
                                                                                                                       //
      if (!account) {                                                                                                  // 165
        throw new Error("no such user");                                                                               // 166
      }                                                                                                                // 167
                                                                                                                       //
      SandstormDb.fillInPictureUrl(account);                                                                           // 169
      appActivity.user = {                                                                                             // 171
        accountId: account._id,                                                                                        // 172
        name: account.profile.name,                                                                                    // 173
        avatarUrl: account.profile.pictureUrl || ""                                                                    // 174
      };                                                                                                               // 171
    } else if (accountIdOrAnonymous) {                                                                                 // 176
      appActivity.user = {                                                                                             // 177
        anonymous: true                                                                                                // 177
      };                                                                                                               // 177
    }                                                                                                                  // 178
                                                                                                                       //
    notify.forEach(function (targetId) {                                                                               // 180
      // Notify the account.                                                                                           // 181
      // We need to know the ID of the inserted/updated document so we can embed it in the                             // 183
      // desktop notification to bind them.                                                                            // 184
      var idIfInserted = Random.id(17);                                                                                // 185
      var result = Notifications.findAndModify({                                                                       // 186
        query: _.extend({                                                                                              // 187
          userId: targetId                                                                                             // 187
        }, notification),                                                                                              // 187
        update: {                                                                                                      // 188
          $set: update,                                                                                                // 189
          $inc: {                                                                                                      // 190
            count: 1                                                                                                   // 190
          },                                                                                                           // 190
          $setOnInsert: {                                                                                              // 191
            _id: idIfInserted                                                                                          // 192
          }                                                                                                            // 191
        },                                                                                                             // 188
        upsert: true                                                                                                   // 195
      });                                                                                                              // 186
                                                                                                                       //
      if (!result.ok) {                                                                                                // 198
        console.error("Couldn't create notification!", result.lastErrorObject);                                        // 199
        return;                                                                                                        // 200
      }                                                                                                                // 201
                                                                                                                       //
      var notificationId = result.value._id || idIfInserted;                                                           // 203
      var desktopNotification = {                                                                                      // 205
        userId: targetId,                                                                                              // 206
        notificationId: notificationId,                                                                                // 207
        appActivity: appActivity                                                                                       // 208
      };                                                                                                               // 205
      createAppActivityDesktopNotification(desktopNotification);                                                       // 211
    });                                                                                                                // 212
  }                                                                                                                    // 213
};                                                                                                                     // 214
                                                                                                                       //
Meteor.methods({                                                                                                       // 216
  testNotifications: function () {                                                                                     // 217
    // Deliver a test notification of each non-grain-initiated type to the user.                                       // 218
    if (!this.userId) return;                                                                                          // 219
    Notifications.insert({                                                                                             // 221
      admin: {                                                                                                         // 222
        action: "/admin/stats",                                                                                        // 223
        type: "reportStats"                                                                                            // 224
      },                                                                                                               // 222
      userId: this.userId,                                                                                             // 226
      timestamp: new Date(),                                                                                           // 227
      isUnread: true                                                                                                   // 228
    });                                                                                                                // 221
    Notifications.insert({                                                                                             // 231
      userId: this.userId,                                                                                             // 232
      referral: true,                                                                                                  // 233
      timestamp: new Date(),                                                                                           // 234
      isUnread: true                                                                                                   // 235
    });                                                                                                                // 231
    Notifications.insert({                                                                                             // 238
      userId: this.userId,                                                                                             // 239
      identityChanges: true,                                                                                           // 240
      timestamp: new Date(),                                                                                           // 241
      isUnread: true                                                                                                   // 242
    });                                                                                                                // 238
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 245
      Notifications.insert({                                                                                           // 246
        userId: this.userId,                                                                                           // 247
        mailingListBonus: true,                                                                                        // 248
        timestamp: new Date(),                                                                                         // 249
        isUnread: true                                                                                                 // 250
      });                                                                                                              // 246
    }                                                                                                                  // 252
  }                                                                                                                    // 253
});                                                                                                                    // 216
Meteor.publish("notifications", function () {                                                                          // 256
  return Notifications.find({                                                                                          // 257
    userId: this.userId                                                                                                // 257
  });                                                                                                                  // 257
});                                                                                                                    // 258
Meteor.publish("notificationGrains", function (notificationIds) {                                                      // 260
  // Since publishes can't be reactive, we leave it to the client to subscribe to both                                 // 261
  // "notifications" and "notificationGrains" reactively.                                                              // 262
  check(notificationIds, [String]);                                                                                    // 263
  var notifications = Notifications.find({                                                                             // 264
    _id: {                                                                                                             // 265
      $in: notificationIds                                                                                             // 265
    },                                                                                                                 // 265
    userId: this.userId                                                                                                // 266
  }, {                                                                                                                 // 264
    fields: {                                                                                                          // 268
      grainId: 1,                                                                                                      // 268
      initiatingAccount: 1                                                                                             // 268
    }                                                                                                                  // 268
  });                                                                                                                  // 267
  var grainIds = notifications.map(function (row) {                                                                    // 271
    return row.grainId;                                                                                                // 272
  }).filter(function (x) {                                                                                             // 273
    return x;                                                                                                          // 273
  });                                                                                                                  // 273
  var accounts = notifications.map(function (row) {                                                                    // 275
    return row.initiatingAccount;                                                                                      // 276
  }).filter(function (x) {                                                                                             // 277
    return x;                                                                                                          // 277
  });                                                                                                                  // 277
  return [Meteor.users.find({                                                                                          // 279
    _id: {                                                                                                             // 280
      $in: accounts                                                                                                    // 280
    }                                                                                                                  // 280
  }, {                                                                                                                 // 280
    fields: {                                                                                                          // 280
      type: 1,                                                                                                         // 280
      profile: 1                                                                                                       // 280
    }                                                                                                                  // 280
  })];                                                                                                                 // 280
});                                                                                                                    // 282
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"pre-meteor.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/pre-meteor.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var module1 = module;                                                                                                  // 1
var inMeteor = void 0;                                                                                                 // 1
module1.watch(require("/imports/server/async-helpers.js"), {                                                           // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 21
                                                                                                                       //
var Fs = Npm.require("fs");                                                                                            // 22
                                                                                                                       //
var Dns = Npm.require("dns");                                                                                          // 23
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 24
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 25
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 26
                                                                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;                                                // 27
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;                                                               // 29
var DDP_HOSTNAME = process.env.DDP_DEFAULT_CONNECTION_URL && Url.parse(process.env.DDP_DEFAULT_CONNECTION_URL).hostname;
var CACHE_TTL_SECONDS = 30; // 30 seconds.  Cache-Control expects units of seconds, not millis.                        // 32
                                                                                                                       //
var DNS_CACHE_TTL = CACHE_TTL_SECONDS * 1000; // DNS cache is in millis.                                               // 33
                                                                                                                       //
var webPublishingHandlers = {}; // Maps grain public IDs to Connect handlers.                                          // 35
// TODO(perf): Garbage-collect this map?                                                                               // 37
// TODO(perf): Implement a more general cache not just of web publishing handlers but of all sorts                     // 39
//   of hosts, in order to avoid database lookups to determine how to handle requests. Host types                      // 40
//   to distinguish between:                                                                                           // 41
//   - Grain hosts                                                                                                     // 42
//   - Web publishing hosts                                                                                            // 43
//   - Standalone grain hosts                                                                                          // 44
                                                                                                                       //
var dnsCache = {}; // Unfortunately, node's DNS library doesn't cache results, so we do our own caching.               // 46
// Unfortunately, node's DNS library also dosen't give us TTLs. So, we'll cache for                                    // 48
// DNS_CACHE_TTL (a relatively small value) and rely on the upstream DNS server to implement                           // 49
// better caching.                                                                                                     // 50
                                                                                                                       //
function isSandstormShell(hostname) {                                                                                  // 52
  // Is this hostname mapped to the Sandstorm shell?                                                                   // 53
  return hostname === HOSTNAME || DDP_HOSTNAME && hostname === DDP_HOSTNAME;                                           // 55
} // We need to use connect. Let's make sure we're using the same version as Meteor's WebApp module                    // 56
// uses. Fortunately, they let us extract it.                                                                          // 59
                                                                                                                       //
                                                                                                                       //
var connect = WebAppInternals.NpmModules.connect.module;                                                               // 60
var mime = connect.static.mime;                                                                                        // 62
                                                                                                                       //
function wwwHandlerForGrain(grainId) {                                                                                 // 64
  return function (request, response) {                                                                                // 65
    var path = request.url; // Strip query.                                                                            // 66
                                                                                                                       //
    path = path.split("?")[0]; // If a directory, open "index.html".                                                   // 69
                                                                                                                       //
    if (path.slice(-1) === "/") {                                                                                      // 72
      path = path + "index.html";                                                                                      // 73
    } // Strip leading "/".                                                                                            // 74
                                                                                                                       //
                                                                                                                       //
    if (path[0] === "/") path = path.slice(1); // URI-decode the rest. Note that this allows filenames to contain spaces and question marks.
                                                                                                                       //
    var encoded = path;                                                                                                // 80
    path = decodeURIComponent(encoded);                                                                                // 81
    var type = mime.lookup(path);                                                                                      // 83
    var charset = mime.charsets.lookup(type);                                                                          // 84
                                                                                                                       //
    if (charset) {                                                                                                     // 85
      type = type + "; charset=" + charset;                                                                            // 86
    } else if (type === "application/json") {                                                                          // 87
      // HACK: Apparently the MIME module does not assume UTF-8 for JSON. :(                                           // 88
      type = type + "; charset=utf-8";                                                                                 // 89
    } // TODO(perf): Automatically gzip text content? Use Express's "compress" middleware for this?                    // 90
    //   Note that nginx will also auto-compress things...                                                             // 93
                                                                                                                       //
                                                                                                                       //
    response.setHeader("Content-Type", type);                                                                          // 95
    response.setHeader("Cache-Control", "public, max-age=" + CACHE_TTL_SECONDS);                                       // 96
                                                                                                                       //
    if (path === "apps/index.json" || path.match(/apps\/[a-z0-9]{52}[.]json/) || path === "experimental/index.json" || path.match(/experimental\/[a-z0-9]{52}[.]json/)) {
      // TODO(cleanup): Extra special terrible hack: The app index needs to serve these JSON files                     // 102
      //   cross-origin. We could almost just make all web sites allow cross-origin since generally                    // 103
      //   web publishing is meant to publish public content. There is one case where this is                          // 104
      //   problematic, though: sites behind a firewall. Those sites could potentially be read                         // 105
      //   by outside sites if CORS is enabled on them. Some day we should make it so apps can                         // 106
      //   explicitly opt-in to allowing cross-origin queries but that day is not today.                               // 107
      response.setHeader("Access-Control-Allow-Origin", "*");                                                          // 108
    }                                                                                                                  // 109
                                                                                                                       //
    var stream = new ResponseStream(response, 200, "OK"); // We hold our own capability to the stream becaues we don't want ResponseStream's close()
    // logic to happen if the response has no data.                                                                    // 114
    //                                                                                                                 // 115
    // TODO(cleanup): This is pretty ugly.                                                                             // 116
                                                                                                                       //
    var streamCap = new Capnp.Capability(stream, ByteStream);                                                          // 117
    globalBackend.useGrain(grainId, function (supervisor) {                                                            // 119
      var tryPath = function (path) {                                                                                  // 120
        return supervisor.getWwwFileHack(path, streamCap).then(function (result) {                                     // 120
          // jscs:disable disallowQuotedKeysInObjects                                                                  // 121
          var status = result.status;                                                                                  // 122
                                                                                                                       //
          if (status === "file") {// Nothing to do.                                                                    // 123
          } else if (status === "directory") {                                                                         // 125
            stream.sendingDirectResponse();                                                                            // 126
            response.writeHead(303, {                                                                                  // 127
              "Content-Type": "text/plain",                                                                            // 128
              "Location": "/" + path + "/",                                                                            // 129
              "Cache-Control": "public, max-age=" + CACHE_TTL_SECONDS                                                  // 130
            });                                                                                                        // 127
            response.end("redirect: /" + path + "/");                                                                  // 132
          } else if (status === "notFound") {                                                                          // 133
            if (path !== encoded) {                                                                                    // 134
              // Compatibility hack: Try getting the URL-encoded version of the path. We used to not                   // 135
              // URL-decode paths at all, hence the need for this fallback.                                            // 136
              // TODO(apibump): Move this hack to the compat layer.                                                    // 137
              return tryPath(encoded);                                                                                 // 138
            }                                                                                                          // 139
                                                                                                                       //
            stream.sendingDirectResponse();                                                                            // 141
            response.writeHead(404, {                                                                                  // 142
              "Content-Type": "text/plain"                                                                             // 143
            });                                                                                                        // 142
            response.end("404 not found: /" + path);                                                                   // 145
          } else {                                                                                                     // 146
            console.error("didn't understand result of getWwwFileHack:", status);                                      // 147
            stream.sendingDirectResponse();                                                                            // 148
            response.writeHead(500, {                                                                                  // 149
              "Content-Type": "text/plain"                                                                             // 150
            });                                                                                                        // 149
            response.end("Internal server error");                                                                     // 152
          }                                                                                                            // 153
                                                                                                                       //
          streamCap.close();                                                                                           // 155
        });                                                                                                            // 156
      };                                                                                                               // 120
                                                                                                                       //
      return tryPath(path);                                                                                            // 157
    }).catch(function (err) {                                                                                          // 158
      console.error(err.stack);                                                                                        // 159
      streamCap.close();                                                                                               // 160
    });                                                                                                                // 161
  };                                                                                                                   // 162
}                                                                                                                      // 163
                                                                                                                       //
function writeErrorResponse(res, err) {                                                                                // 165
  var status = 500;                                                                                                    // 166
                                                                                                                       //
  if (err instanceof Meteor.Error && typeof err.error === "number" && err.error >= 400 && err.error < 600) {           // 167
    status = err.error;                                                                                                // 169
  } else if (err.httpErrorCode) {                                                                                      // 170
    status = err.httpErrorCode;                                                                                        // 171
  } // Log errors that are our fault, but not errors that are the client's fault.                                      // 172
                                                                                                                       //
                                                                                                                       //
  if (status >= 500) console.error(err.stack);                                                                         // 175
  res.writeHead(status, {                                                                                              // 177
    "Content-Type": err.htmlMessage ? "text/html" : "text/plain"                                                       // 177
  });                                                                                                                  // 177
  res.end(err.htmlMessage || err.message);                                                                             // 178
}                                                                                                                      // 179
                                                                                                                       //
var PNG_MAGIC = new Buffer([0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A]);                                          // 181
var JPEG_MAGIC = new Buffer([0xFF, 0xD8, 0xFF]);                                                                       // 182
                                                                                                                       //
function checkMagic(buf, magic) {                                                                                      // 184
  if (buf.length < magic.length) return false;                                                                         // 185
                                                                                                                       //
  for (var i = 0; i < magic.length; i++) {                                                                             // 187
    if (buf[i] != magic[i]) return false;                                                                              // 188
  }                                                                                                                    // 189
                                                                                                                       //
  return true;                                                                                                         // 191
}                                                                                                                      // 192
                                                                                                                       //
function serveSelfTest(req, res) {                                                                                     // 194
  try {                                                                                                                // 195
    if (req.method === "GET" && req.url === "/") {                                                                     // 196
      var content = new Buffer("Self-test OK."); // Convert the ROOT_URL to something that is a valid origin.          // 198
                                                                                                                       //
      var rootUrlAsOrigin = process.env.ROOT_URL;                                                                      // 201
                                                                                                                       //
      if (rootUrlAsOrigin.slice(-1) === "/") {                                                                         // 202
        rootUrlAsOrigin = rootUrlAsOrigin.slice(0, -1);                                                                // 203
      }                                                                                                                // 204
                                                                                                                       //
      res.writeHead(200, {                                                                                             // 206
        "Content-Type": "text/plain",                                                                                  // 207
        "Content-Length": content.length,                                                                              // 208
        "Access-Control-Allow-Origin": rootUrlAsOrigin                                                                 // 209
      });                                                                                                              // 206
      res.end(content);                                                                                                // 211
    } else {                                                                                                           // 212
      res.writeHead(400, {                                                                                             // 213
        "Content-Type": "text/plain"                                                                                   // 213
      });                                                                                                              // 213
      res.end("Bad request to self-test subdomain.");                                                                  // 214
      return;                                                                                                          // 215
    }                                                                                                                  // 216
  } catch (err) {                                                                                                      // 217
    console.log("WARNING: An error occurred while serving self-test; proceeding anyway:", err);                        // 218
  }                                                                                                                    // 219
}                                                                                                                      // 220
                                                                                                                       //
function serveStaticAsset(req, res, hostId) {                                                                          // 222
  inMeteor(function () {                                                                                               // 223
    if (req.method === "GET") {                                                                                        // 224
      var assetCspHeader = "default-src 'none'; style-src 'unsafe-inline'; sandbox";                                   // 225
                                                                                                                       //
      if (req.headers["if-none-match"] === "permanent") {                                                              // 226
        // Cache never invalidates since we use a new URL for every resource.                                          // 227
        res.writeHead(304, {                                                                                           // 228
          "Cache-Control": "public, max-age=31536000",                                                                 // 229
          "ETag": "permanent",                                                                                         // 230
          // To be safe, send these again, although it shouldn't be necessary.                                         // 232
          "Content-Security-Policy": assetCspHeader,                                                                   // 233
          "Access-Control-Allow-Origin": "*",                                                                          // 234
          "X-Content-Type-Options": "nosniff"                                                                          // 235
        });                                                                                                            // 228
        res.end();                                                                                                     // 237
        return;                                                                                                        // 238
      }                                                                                                                // 239
                                                                                                                       //
      var url = Url.parse(req.url, true);                                                                              // 241
      var pathname = url.pathname.slice(1);                                                                            // 242
      var asset = void 0;                                                                                              // 243
                                                                                                                       //
      if (pathname.startsWith("identicon/")) {                                                                         // 244
        var size = parseInt((url.query || {}).s);                                                                      // 245
        asset = new Identicon(pathname.slice("identicon/".length), size).asAsset();                                    // 246
      } else if (pathname.indexOf("/") == -1) {                                                                        // 247
        asset = globalDb.getStaticAsset(pathname);                                                                     // 248
      }                                                                                                                // 249
                                                                                                                       //
      if (asset) {                                                                                                     // 251
        var headers = {                                                                                                // 252
          "Content-Type": asset.mimeType,                                                                              // 253
          "Content-Length": asset.content.length,                                                                      // 254
          // Assets can be cached forever because each one has a unique ID.                                            // 256
          "Cache-Control": "public, max-age=31536000",                                                                 // 257
          // Since different resources get different URLs, we can use a static etag.                                   // 259
          "ETag": "permanent",                                                                                         // 260
          // Set strict Content-Security-Policy to prevent static assets from executing any script                     // 262
          // or doing basically anything when browsed to directly. The static assets host is not                       // 263
          // intended to serve HTML. Mostly, it serves images and javascript -- note that setting                      // 264
          // the CSP header on Javascript files does not prevent other hosts from voluntarily                          // 265
          // specifying these scripts in <script> tags.                                                                // 266
          "Content-Security-Policy": assetCspHeader,                                                                   // 267
          // Allow any host to fetch these assets. This is safe since requests to this host are                        // 269
          // totally side-effect-free and the asset ID acts as a capability to prevent loading                         // 270
          // assets you're not supposed to know about.                                                                 // 271
          "Access-Control-Allow-Origin": "*",                                                                          // 272
          // Extra protection against content type trickery.                                                           // 274
          "X-Content-Type-Options": "nosniff"                                                                          // 275
        };                                                                                                             // 252
                                                                                                                       //
        if (asset.encoding) {                                                                                          // 278
          headers["Content-Encoding"] = asset.encoding;                                                                // 279
        }                                                                                                              // 280
                                                                                                                       //
        res.writeHead(200, headers);                                                                                   // 282
        res.end(asset.content);                                                                                        // 283
      } else {                                                                                                         // 284
        res.writeHead(404, {                                                                                           // 285
          "Content-Type": "text/plain"                                                                                 // 285
        });                                                                                                            // 285
        res.end("not found");                                                                                          // 286
      }                                                                                                                // 287
    } else if (req.method === "POST") {                                                                                // 288
      res.setHeader("Access-Control-Allow-Origin", "*");                                                               // 289
                                                                                                                       //
      var _url = Url.parse(req.url);                                                                                   // 291
                                                                                                                       //
      var purpose = globalDb.fulfillAssetUpload(_url.pathname.slice(1)); // Sanity check the purpose of this upload token.
                                                                                                                       //
      if (!purpose) {                                                                                                  // 295
        res.writeHead(404, {                                                                                           // 296
          "Content-Type": "text/plain"                                                                                 // 296
        });                                                                                                            // 296
        res.end("Upload token not found or expired.");                                                                 // 297
        return;                                                                                                        // 298
      } else if (purpose.profilePicture) {                                                                             // 299
        var userId = purpose.profilePicture.userId;                                                                    // 300
        check(userId, String);                                                                                         // 301
      } else if (purpose.loginLogo) {// no additional fields for loginLogo                                             // 302
      } else {                                                                                                         // 304
        res.writeHead(500, {                                                                                           // 305
          "Content-Type": "text/plain"                                                                                 // 305
        });                                                                                                            // 305
        res.end("Found invalid purpose for upload");                                                                   // 306
        return;                                                                                                        // 307
      }                                                                                                                // 308
                                                                                                                       //
      var buffers = [];                                                                                                // 310
      var totalSize = 0;                                                                                               // 311
      var done = new Future();                                                                                         // 312
      req.on("data", function (buf) {                                                                                  // 313
        totalSize += buf.length;                                                                                       // 314
                                                                                                                       //
        if (totalSize <= 64 * 1024) {                                                                                  // 315
          buffers.push(buf);                                                                                           // 316
        }                                                                                                              // 317
      });                                                                                                              // 318
      req.on("end", done.return.bind(done));                                                                           // 319
      req.on("error", done.throw.bind(done));                                                                          // 320
      done.wait();                                                                                                     // 321
                                                                                                                       //
      if (totalSize > 64 * 1024) {                                                                                     // 323
        // TODO(soon): Resize the image ourselves.                                                                     // 324
        res.writeHead(400, {                                                                                           // 325
          "Content-Type": "text/plain"                                                                                 // 325
        });                                                                                                            // 325
        res.end("Picture too large; please use an image under 64 KiB.");                                               // 326
        return;                                                                                                        // 327
      }                                                                                                                // 328
                                                                                                                       //
      var content = Buffer.concat(buffers);                                                                            // 330
      var type = void 0;                                                                                               // 331
                                                                                                                       //
      if (checkMagic(content, PNG_MAGIC)) {                                                                            // 332
        type = "image/png";                                                                                            // 333
      } else if (checkMagic(content, JPEG_MAGIC)) {                                                                    // 334
        type = "image/jpeg";                                                                                           // 335
      } else {                                                                                                         // 336
        res.writeHead(400, {                                                                                           // 337
          "Content-Type": "text/plain"                                                                                 // 337
        });                                                                                                            // 337
        res.end("Image must be PNG or JPEG.");                                                                         // 338
        return;                                                                                                        // 339
      }                                                                                                                // 340
                                                                                                                       //
      var assetId = globalDb.addStaticAsset({                                                                          // 342
        mimeType: type                                                                                                 // 342
      }, content);                                                                                                     // 342
                                                                                                                       //
      if (purpose.profilePicture) {                                                                                    // 344
        var accountId = purpose.profilePicture.userId;                                                                 // 345
        var result = Meteor.users.findAndModify({                                                                      // 346
          query: {                                                                                                     // 347
            _id: accountId                                                                                             // 347
          },                                                                                                           // 347
          update: {                                                                                                    // 348
            $set: {                                                                                                    // 348
              "profile.picture": assetId                                                                               // 348
            }                                                                                                          // 348
          },                                                                                                           // 348
          fields: {                                                                                                    // 349
            "profile.picture": 1                                                                                       // 349
          }                                                                                                            // 349
        });                                                                                                            // 346
                                                                                                                       //
        if (result.ok) {                                                                                               // 352
          var old = result.value;                                                                                      // 353
                                                                                                                       //
          if (old && old.profile && old.profile.picture) {                                                             // 354
            globalDb.unrefStaticAsset(old.profile.picture);                                                            // 355
          }                                                                                                            // 356
        } else {                                                                                                       // 357
          res.writeHead(500, {                                                                                         // 358
            "Content-Type": "text/plain"                                                                               // 358
          });                                                                                                          // 358
          res.end("Couldn't update profile picture.");                                                                 // 359
          return;                                                                                                      // 360
        }                                                                                                              // 361
      } else if (purpose.loginLogo) {                                                                                  // 362
        var _result = globalDb.collections.settings.findAndModify({                                                    // 363
          query: {                                                                                                     // 364
            _id: "whitelabelCustomLogoAssetId"                                                                         // 364
          },                                                                                                           // 364
          update: {                                                                                                    // 365
            $setOnInsert: {                                                                                            // 366
              _id: "whitelabelCustomLogoAssetId"                                                                       // 366
            },                                                                                                         // 366
            $set: {                                                                                                    // 367
              value: assetId                                                                                           // 367
            }                                                                                                          // 367
          },                                                                                                           // 365
          upsert: true,                                                                                                // 369
          fields: {                                                                                                    // 370
            value: 1                                                                                                   // 370
          }                                                                                                            // 370
        });                                                                                                            // 363
                                                                                                                       //
        if (_result.ok) {                                                                                              // 373
          var _old = _result.value;                                                                                    // 374
                                                                                                                       //
          if (_old && _old.value) {                                                                                    // 375
            globalDb.unrefStaticAsset(_old.value);                                                                     // 376
          }                                                                                                            // 377
        } else {                                                                                                       // 378
          res.writeHead(500, {                                                                                         // 379
            "Content-Type": "text/plain"                                                                               // 379
          });                                                                                                          // 379
          res.end("Couldn't update login logo.");                                                                      // 380
          return;                                                                                                      // 381
        }                                                                                                              // 382
      }                                                                                                                // 383
                                                                                                                       //
      res.writeHead(204, {});                                                                                          // 385
      res.end();                                                                                                       // 386
    } else if (req.method === "OPTIONS") {                                                                             // 387
      var requestedHeaders = req.headers["access-control-request-headers"];                                            // 388
                                                                                                                       //
      if (requestedHeaders) {                                                                                          // 389
        res.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                               // 390
      }                                                                                                                // 391
                                                                                                                       //
      res.writeHead(204, {                                                                                             // 393
        "Access-Control-Allow-Origin": "*",                                                                            // 394
        "Access-Control-Allow-Methods": "GET, HEAD, POST, OPTIONS",                                                    // 395
        "Access-Control-Max-Age": "3600"                                                                               // 396
      });                                                                                                              // 393
      res.end();                                                                                                       // 398
    } else {                                                                                                           // 399
      res.writeHead(405, "Method Not Allowed", {                                                                       // 400
        "Allow": "GET, POST, OPTIONS",                                                                                 // 401
        "Content-Type": "text/plain"                                                                                   // 402
      });                                                                                                              // 400
      res.end("405 Method Not Allowed: " + req.method);                                                                // 404
    }                                                                                                                  // 405
  }).catch(function (err) {                                                                                            // 406
    writeErrorResponse(res, err);                                                                                      // 407
  });                                                                                                                  // 408
}                                                                                                                      // 409
                                                                                                                       //
var canonicalizeShellOrWildcardUrl = function (hostname, url) {                                                        // 411
  // Start with ROOT_URL, apply host & path from inbound URL, then                                                     // 412
  // redirect.                                                                                                         // 413
  var targetUrl = Url.parse(process.env.ROOT_URL); // Retain the protocol & port from ROOT_URL but use the inbound     // 414
  // hostname.                                                                                                         // 417
                                                                                                                       //
  targetUrl.host = hostname + ":" + targetUrl.port; // The following allows to avoid decoding + re-encoding query      // 418
  // string parameters, if provided.                                                                                   // 421
                                                                                                                       //
  targetUrl = Url.resolve(Url.format(targetUrl), url);                                                                 // 422
  return targetUrl;                                                                                                    // 423
};                                                                                                                     // 424
                                                                                                                       //
var handleNonMeteorRequest = function (req, res, next, redirectIfInWildcard) {                                         // 426
  // if redirectIfInWildcard is true, and if the host is part of the wildcard host, then we will                       // 427
  // prefer to redirect to the canonical url for that request.  If false, we will respond to the                       // 428
  // request directly.                                                                                                 // 429
  var hostname = req.headers.host.split(":")[0];                                                                       // 431
  var publicIdPromise = void 0; // See if the request was for a host in the wildcard.                                  // 433
                                                                                                                       //
  var id = matchWildcardHost(req.headers.host);                                                                        // 435
                                                                                                                       //
  if (id) {                                                                                                            // 436
    // Match!                                                                                                          // 437
    if (redirectIfInWildcard) {                                                                                        // 438
      res.writeHead(302, {                                                                                             // 439
        "Location": canonicalizeShellOrWildcardUrl(hostname, req.url)                                                  // 439
      });                                                                                                              // 439
      res.end();                                                                                                       // 440
      return;                                                                                                          // 441
    }                                                                                                                  // 442
                                                                                                                       //
    if (id === "static") {                                                                                             // 444
      // Static assets domain.                                                                                         // 445
      serveStaticAsset(req, res);                                                                                      // 446
      return;                                                                                                          // 447
    }                                                                                                                  // 448
                                                                                                                       //
    if (id.match(/^selftest-/)) {                                                                                      // 450
      // Self test domain pattern. Starts w/ hyphen to avoid ambiguity with grain session/static                       // 451
      // publishing wildcard hosts.                                                                                    // 452
      serveSelfTest(req, res);                                                                                         // 453
      return;                                                                                                          // 454
    } // Try to route the request to a session.                                                                        // 455
                                                                                                                       //
                                                                                                                       //
    publicIdPromise = tryProxyRequest(id, req, res).then(function (handled) {                                          // 458
      if (handled) {                                                                                                   // 459
        return null;                                                                                                   // 460
      } else {                                                                                                         // 461
        return id;                                                                                                     // 462
      }                                                                                                                // 463
    });                                                                                                                // 464
  } else {                                                                                                             // 465
    // Not a wildcard host. Perhaps it is a custom host.                                                               // 466
    publicIdPromise = lookupPublicIdFromDns(hostname);                                                                 // 467
  }                                                                                                                    // 468
                                                                                                                       //
  publicIdPromise.then(function (publicId) {                                                                           // 470
    if (publicId) {                                                                                                    // 471
      return Promise.resolve(undefined).then(function () {                                                             // 472
        var handler = webPublishingHandlers[publicId];                                                                 // 473
                                                                                                                       //
        if (handler) {                                                                                                 // 474
          return handler;                                                                                              // 475
        } else {                                                                                                       // 476
          // We don't have a handler for this publicId, so look it up in the grain DB.                                 // 477
          return inMeteor(function () {                                                                                // 478
            var grain = Grains.findOne({                                                                               // 479
              publicId: publicId                                                                                       // 479
            }, {                                                                                                       // 479
              fields: {                                                                                                // 479
                _id: 1                                                                                                 // 479
              }                                                                                                        // 479
            });                                                                                                        // 479
                                                                                                                       //
            if (!grain) {                                                                                              // 480
              throw new Meteor.Error(404, "No such grain for public ID: " + publicId);                                 // 481
            }                                                                                                          // 482
                                                                                                                       //
            var grainId = grain._id;                                                                                   // 484
            return webPublishingHandlers[publicId] = wwwHandlerForGrain(grainId);                                      // 485
          });                                                                                                          // 486
        }                                                                                                              // 487
      }).then(function (handler) {                                                                                     // 488
        handler(req, res);                                                                                             // 489
      });                                                                                                              // 490
    } else {                                                                                                           // 491
      // Already handled by tryProxyRequest above, no further action needed.                                           // 492
      return Promise.resolve(undefined);                                                                               // 493
    }                                                                                                                  // 494
  }).catch(function (err) {                                                                                            // 495
    writeErrorResponse(res, err);                                                                                      // 496
  });                                                                                                                  // 497
}; // This function serves responses on Sandstorm's main HTTP/HTTPS                                                    // 498
// port.                                                                                                               // 501
                                                                                                                       //
                                                                                                                       //
var handleNonMeteorRequestDirectly = function (req, res, next) {                                                       // 502
  handleNonMeteorRequest(req, res, next, false);                                                                       // 503
}; // "Alternate ports" are ports other than the main HTTP or HTTPS port.                                              // 504
// They are handled as follows:                                                                                        // 507
//                                                                                                                     // 508
// * Requests to the shell & grains, we redirect to the main port.                                                     // 509
// * For requests to static publishing on hosts within the wildcard, we redirect to the main port.                     // 510
// * For static publishing glued in via DNS TXT record, we serve the request as-is.                                    // 511
//                                                                                                                     // 512
// Alternate ports are bound to FD #5 and higher.                                                                      // 513
                                                                                                                       //
                                                                                                                       //
var getNumberOfAlternatePorts = function () {                                                                          // 514
  var numPorts = process.env.PORT.split(",").length;                                                                   // 515
  var numAlternatePorts = numPorts - 1;                                                                                // 516
  return numAlternatePorts;                                                                                            // 517
};                                                                                                                     // 518
                                                                                                                       //
var handleNonMainPortRequest = function (req, res, next) {                                                             // 520
  try {                                                                                                                // 521
    if (!req.headers.host) {                                                                                           // 522
      res.writeHead(400, {                                                                                             // 523
        "Content-Type": "text/plain"                                                                                   // 523
      });                                                                                                              // 523
      res.end("Missing Host header");                                                                                  // 524
      return;                                                                                                          // 525
    } // If this request is intended for the shell, redirect to the canonical shell URL.                               // 526
                                                                                                                       //
                                                                                                                       //
    var hostname = req.headers.host.split(":")[0];                                                                     // 529
                                                                                                                       //
    if (isSandstormShell(hostname)) {                                                                                  // 530
      res.writeHead(302, {                                                                                             // 531
        "Location": canonicalizeShellOrWildcardUrl(hostname, req.url)                                                  // 531
      });                                                                                                              // 531
      res.end();                                                                                                       // 532
      return;                                                                                                          // 533
    } // TODO(someday): check for standalone hosts and do something reasonable                                         // 534
                                                                                                                       //
                                                                                                                       //
    handleNonMeteorRequest(req, res, next, true);                                                                      // 537
  } catch (e) {                                                                                                        // 538
    // This should never be reached, because all the request handlers should be catching                               // 539
    // exceptions, but you can never be too careful in a top-level request handler.                                    // 540
    console.err("Unhandled exception in request handler:", e.stack);                                                   // 541
    res.writeHead(500, {                                                                                               // 542
      "Content-Type": "text/plain"                                                                                     // 542
    });                                                                                                                // 542
    res.end("Unhandled exception: " + e.stack);                                                                        // 543
  }                                                                                                                    // 544
};                                                                                                                     // 545
                                                                                                                       //
var listenOnAlternatePorts = function () {                                                                             // 547
  for (var i = 0; i < getNumberOfAlternatePorts(); i++) {                                                              // 548
    // Call createServerForSandstorm() to skip our monkey patching.                                                    // 549
    var alternatePortServer = Http.createServerForSandstorm(handleNonMainPortRequest);                                 // 550
    alternatePortServer.listen({                                                                                       // 551
      fd: i + 5                                                                                                        // 551
    });                                                                                                                // 551
  }                                                                                                                    // 552
};                                                                                                                     // 553
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 555
  var meteorRequestListeners = WebApp.httpServer.listeners("request"); // Construct the middleware chain for requests to non-DDP, non-shell hosts.
                                                                                                                       //
  var nonMeteorRequestHandler = connect(); // BlackrockPayments is only defined in the Blackrock build of Sandstorm.   // 559
                                                                                                                       //
  if (global.BlackrockPayments) {                                                                                      // 561
    // Have to check with global, because it could be undefined.                                                       // 561
    nonMeteorRequestHandler.use(BlackrockPayments.makeConnectHandler(globalDb));                                       // 562
  }                                                                                                                    // 563
                                                                                                                       //
  nonMeteorRequestHandler.use(handleNonMeteorRequestDirectly);                                                         // 565
  WebApp.httpServer.removeAllListeners("request");                                                                     // 567
  WebApp.httpServer.on("request", function (req, res) {                                                                // 568
    Promise.resolve(undefined).then(function () {                                                                      // 569
      if (!req.headers.host) {                                                                                         // 570
        res.writeHead(400, {                                                                                           // 571
          "Content-Type": "text/plain"                                                                                 // 571
        });                                                                                                            // 571
        res.end("Missing Host header");                                                                                // 572
        return;                                                                                                        // 573
      }                                                                                                                // 574
                                                                                                                       //
      var hostname = req.headers.host.split(":")[0]; // strip port if it exists                                        // 576
                                                                                                                       //
      if (isSandstormShell(hostname)) {                                                                                // 577
        if (req.url.startsWith("/_oauth/")) {                                                                          // 578
          // Intercept oauth callbacks on the main host to check if the user was actually logging                      // 579
          // into a standalone host, in which case we need to redirect to that host instead.                           // 580
          var parsedUrl = Url.parse(req.url, true);                                                                    // 581
                                                                                                                       //
          if (parsedUrl.query && parsedUrl.query.state) {                                                              // 582
            var rawState = new Buffer(parsedUrl.query.state, "base64");                                                // 583
            var state = JSON.parse(rawState.toString());                                                               // 584
                                                                                                                       //
            if (state.redirectUrl) {                                                                                   // 585
              var parsedRedirect = Url.parse(state.redirectUrl);                                                       // 586
              var redirectHostname = parsedRedirect.hostname;                                                          // 587
                                                                                                                       //
              if (redirectHostname !== HOSTNAME) {                                                                     // 588
                return inMeteor(function () {                                                                          // 589
                  if (globalDb.hostIsStandalone(redirectHostname)) {                                                   // 590
                    res.writeHead(302, {                                                                               // 591
                      "Location": parsedRedirect.protocol + "//" + parsedRedirect.host + req.url                       // 591
                    });                                                                                                // 591
                    res.end();                                                                                         // 593
                    return;                                                                                            // 594
                  } else {                                                                                             // 595
                    throw new Meteor.Error(400, "redirectUrl in OAuth was for an unknown host: " + state.redirectUrl);
                  }                                                                                                    // 598
                });                                                                                                    // 599
              }                                                                                                        // 600
            }                                                                                                          // 601
          }                                                                                                            // 602
        } // If destined for the DDP host or the main host, pass on to Meteor                                          // 603
                                                                                                                       //
                                                                                                                       //
        for (var i = 0; i < meteorRequestListeners.length; i++) {                                                      // 605
          meteorRequestListeners[i](req, res);                                                                         // 606
        }                                                                                                              // 607
      } else {                                                                                                         // 608
        return inMeteor(function () {                                                                                  // 609
          if (globalDb.hostIsStandalone(hostname)) {                                                                   // 610
            // If it's a standalone host, also pass on to meteor                                                       // 611
            for (var _i = 0; _i < meteorRequestListeners.length; _i++) {                                               // 612
              meteorRequestListeners[_i](req, res);                                                                    // 613
            }                                                                                                          // 614
          } else {                                                                                                     // 615
            // Otherwise, dispatch to our own middleware proxy chain.                                                  // 616
            nonMeteorRequestHandler(req, res); // Adjust timeouts on proxied requests to allow apps to long-poll if needed.
                                                                                                                       //
            WebApp._timeoutAdjustmentRequestCallback(req, res);                                                        // 619
          }                                                                                                            // 620
        });                                                                                                            // 621
      }                                                                                                                // 622
    }).catch(function (e) {                                                                                            // 623
      // This should never be reached, because all the request handlers should be catching                             // 624
      // exceptions, but you can never be too careful in a top-level request handler.                                  // 625
      console.error("Unhandled exception in request handler:", e.stack);                                               // 626
      res.writeHead(500, {                                                                                             // 627
        "Content-Type": "text/plain"                                                                                   // 627
      });                                                                                                              // 627
      res.end("Unhandled exception: " + e.stack);                                                                      // 628
    });                                                                                                                // 629
  });                                                                                                                  // 630
  var meteorUpgradeListeners = WebApp.httpServer.listeners("upgrade");                                                 // 632
  WebApp.httpServer.removeAllListeners("upgrade");                                                                     // 633
  WebApp.httpServer.on("upgrade", function (req, socket, head) {                                                       // 635
    Promise.resolve(undefined).then(function () {                                                                      // 636
      if (!req.headers.host) {                                                                                         // 637
        throw new Meteor.Error(400, "Missing Host header");                                                            // 638
      } else if (isSandstormShell(req.headers.host.split(":")[0])) {                                                   // 639
        // Go on to Meteor.                                                                                            // 640
        for (var ii = 0; ii < meteorUpgradeListeners.length; ++ii) {                                                   // 641
          meteorUpgradeListeners[ii](req, socket, head);                                                               // 642
        }                                                                                                              // 643
                                                                                                                       //
        return true;                                                                                                   // 645
      } else {                                                                                                         // 646
        return inMeteor(function () {                                                                                  // 647
          // TODO(someday): This is slightly sad that some requests will now need a Meteor fiber                       // 648
          // If we disallow standalone domains from being in the wildcard host, then we can move                       // 649
          // this check to the bottom.                                                                                 // 650
          if (globalDb.hostIsStandalone(req.headers.host.split(":")[0])) {                                             // 651
            for (var _ii = 0; _ii < meteorUpgradeListeners.length; ++_ii) {                                            // 652
              meteorUpgradeListeners[_ii](req, socket, head);                                                          // 653
            }                                                                                                          // 654
                                                                                                                       //
            return true;                                                                                               // 656
          } else {                                                                                                     // 657
            var id = matchWildcardHost(req.headers.host);                                                              // 658
                                                                                                                       //
            if (id) {                                                                                                  // 659
              return tryProxyUpgrade(id, req, socket, head);                                                           // 660
            } else {                                                                                                   // 661
              return false;                                                                                            // 662
            }                                                                                                          // 663
          }                                                                                                            // 664
        });                                                                                                            // 665
      }                                                                                                                // 666
    }).then(function (handled) {                                                                                       // 667
      if (!handled) socket.destroy();                                                                                  // 668
    }).catch(function (err) {                                                                                          // 669
      console.error("WebSocket event handler failed:", err.stack);                                                     // 670
      socket.destroy();                                                                                                // 671
    });                                                                                                                // 672
  });                                                                                                                  // 673
  listenOnAlternatePorts();                                                                                            // 675
});                                                                                                                    // 676
var errorTxtMapping = {};                                                                                              // 678
errorTxtMapping[Dns.NOTFOUND] = "<p>" + "If you were trying to configure static publishing for a blog or website, powered " + "by a Sandstorm app hosted at this server, you either have not added DNS TXT records " + "correctly, or the DNS cache has not updated yet (may take a while, like 5 minutes to one " + "hour).</p>";
errorTxtMapping[Dns.NODATA] = errorTxtMapping[Dns.NOTFOUND];                                                           // 684
errorTxtMapping[Dns.TIMEOUT] = "<p>" + "The DNS query has timed out, which may be a sign of poorly configured DNS on the server.</p>";
errorTxtMapping[Dns.CONNREFUSED] = "<p>" + "The DNS server refused the connection, which means either your DNS server is down/unreachable, " + "or the server has misconfigured their DNS.</p>";
                                                                                                                       //
function lookupPublicIdFromDns(hostname) {                                                                             // 691
  // Given a hostname, determine its public ID.                                                                        // 692
  // We look for a TXT record indicating the public ID. Unfortunately, according to spec, a single                     // 693
  // hostname cannot have both a CNAME and a TXT record, because a TXT lookup on a CNAME'd host                        // 694
  // should actually be redirected to the CNAME, just like an A lookup would be. In practice lots                      // 695
  // of DNS software actually allows TXT records on CNAMEs, and it seems to work, but some software                    // 696
  // does not allow it and it's explicitly disallowed by the spec. Therefore, we instead look for                      // 697
  // the TXT record on a subdomain.                                                                                    // 698
  //                                                                                                                   // 699
  // I also considered having the CNAME itself point to <publicId>.<hostname>, where                                   // 700
  // *.<hostname> is in turn a CNAME for the Sandstorm server. This approach seemed elegant at                         // 701
  // first, but has a number of problems, the biggest being that it breaks the ability to place a                      // 702
  // CDN like CloudFlare in front of the site.                                                                         // 703
  var cache = dnsCache[hostname];                                                                                      // 705
                                                                                                                       //
  if (cache && Date.now() < cache.expiration) {                                                                        // 706
    return Promise.resolve(cache.value);                                                                               // 707
  }                                                                                                                    // 708
                                                                                                                       //
  return new Promise(function (resolve, reject) {                                                                      // 710
    Dns.resolveTxt("sandstorm-www." + hostname, function (err, records) {                                              // 711
      if (err) {                                                                                                       // 712
        var errorMsg = errorTxtMapping[err.code] || "";                                                                // 713
        var error = new Error('Error looking up DNS TXT records for host "' + hostname + '": ' + err.message);         // 714
        error.htmlMessage = '<style type="text/css">h2, h3, p { max-width: 600px; }</style>' + "<h2>Sandstorm static publishing needs further configuration (or wrong URL)</h2>" + errorMsg + "<p>To visit this Sandstorm server's main interface, go to: <a href='" + process.env.ROOT_URL + "'>" + process.env.ROOT_URL + "</a></p>" + "<h3>DNS details</h3>" + '<p>Error looking up DNS TXT records for host "' + hostname + '": ' + err.message + "</p>" + "<p>If you have the <tt>dig</tt> tool, you can run this command to learn more:</p>" + "<p><tt>dig TXT sandstorm-www." + hostname + "</tt></p>" + "<h3>Changing the server URL, or troubleshooting OAuth login</h3>" + "<p>If you are the server admin and want to use this address as the main interface, " + "edit /opt/sandstorm/sandstorm.conf, modify the BASE_URL setting, and restart " + "Sandstorm.</p>" + "<p>If you got here after trying to log in via OAuth (e.g. through GitHub or Google), " + "the problem is probably that the OAuth callback URL was set wrong. You need to " + "update it through the respective login provider's management console. The " + "easiest way to do that is to run <tt>sudo sandstorm admin-token</tt>, then " + "reconfigure the OAuth provider.</p>";
        error.httpErrorCode = _.contains(["ENOTFOUND", "ENODATA"], err.code) ? 404 : 500;                              // 735
        reject(error);                                                                                                 // 736
      } else if (records.length !== 1) {                                                                               // 737
        reject(new Error('Host "sandstorm-www.' + hostname + '" must have exactly one TXT record.'));                  // 738
      } else {                                                                                                         // 739
        // Under node 0.12+, `records` is an array-of-arrays. Each sub-array contains "chunks" of a                    // 740
        // record, which should be joined.                                                                             // 741
        var result = records[0].join("");                                                                              // 742
        dnsCache[hostname] = {                                                                                         // 743
          value: result,                                                                                               // 743
          expiration: Date.now() + DNS_CACHE_TTL                                                                       // 743
        };                                                                                                             // 743
        resolve(result);                                                                                               // 744
      }                                                                                                                // 745
    });                                                                                                                // 746
  });                                                                                                                  // 747
}                                                                                                                      // 748
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"proxy.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/proxy.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Bignum = void 0;                                                                                                   // 1
module.watch(require("bignum"), {                                                                                      // 1
  "default": function (v) {                                                                                            // 1
    Bignum = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_ALTHOME: function (v) {                                                                                    // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SandstormBackend = void 0,                                                                                         // 1
    shouldRestartGrain = void 0;                                                                                       // 1
module.watch(require("/imports/server/backend.js"), {                                                                  // 1
  SandstormBackend: function (v) {                                                                                     // 1
    SandstormBackend = v;                                                                                              // 1
  },                                                                                                                   // 1
  shouldRestartGrain: function (v) {                                                                                   // 1
    shouldRestartGrain = v;                                                                                            // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  waitPromise: function (v) {                                                                                          // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var REQUEST_HEADER_WHITELIST = void 0,                                                                                 // 1
    RESPONSE_HEADER_WHITELIST = void 0;                                                                                // 1
module.watch(require("/imports/server/header-whitelist.js"), {                                                         // 1
  REQUEST_HEADER_WHITELIST: function (v) {                                                                             // 1
    REQUEST_HEADER_WHITELIST = v;                                                                                      // 1
  },                                                                                                                   // 1
  RESPONSE_HEADER_WHITELIST: function (v) {                                                                            // 1
    RESPONSE_HEADER_WHITELIST = v;                                                                                     // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 23
                                                                                                                       //
var ChildProcess = Npm.require("child_process");                                                                       // 24
                                                                                                                       //
var Fs = Npm.require("fs");                                                                                            // 25
                                                                                                                       //
var Path = Npm.require("path");                                                                                        // 26
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 27
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 28
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 29
                                                                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;                                                // 31
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;                                         // 32
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;                                         // 33
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");                                                  // 34
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 35
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 36
var Powerbox = Capnp.importSystem("sandstorm/powerbox.capnp");                                                         // 37
var PARSED_ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                 // 39
var PROTOCOL = PARSED_ROOT_URL.protocol;                                                                               // 40
var PORT = PARSED_ROOT_URL.port;                                                                                       // 41
                                                                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, apiTokenAccountId) {                            // 43
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we                      // 44
  // update Bob's Account to say it's referredBy Alice's Account (which is apiTokenAccountId).                         // 45
  check(accountId, String);                                                                                            // 46
  check(apiTokenAccountId, String); // Bail out early if referrals aren't enabled                                      // 47
                                                                                                                       //
  if (!db.isReferralEnabled()) {                                                                                       // 50
    return;                                                                                                            // 51
  }                                                                                                                    // 52
                                                                                                                       //
  var aliceAccountId = apiTokenAccountId;                                                                              // 54
  var bobAccountId = accountId;                                                                                        // 55
                                                                                                                       //
  if (Meteor.users.find({                                                                                              // 57
    _id: bobAccountId,                                                                                                 // 58
    referredByComplete: {                                                                                              // 59
      $exists: true                                                                                                    // 59
    }                                                                                                                  // 59
  }).count() > 0) {                                                                                                    // 57
    return;                                                                                                            // 61
  } // Only actually update Bob's Account ID if there is no referredBy.                                                // 62
                                                                                                                       //
                                                                                                                       //
  Meteor.users.update({                                                                                                // 65
    _id: bobAccountId,                                                                                                 // 66
    referredBy: {                                                                                                      // 66
      $exists: false                                                                                                   // 66
    }                                                                                                                  // 66
  }, {                                                                                                                 // 66
    $set: {                                                                                                            // 67
      referredBy: aliceAccountId                                                                                       // 67
    }                                                                                                                  // 67
  });                                                                                                                  // 67
};                                                                                                                     // 68
                                                                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {                                                         // 70
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView                    // 71
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share                    // 72
  // access' interface. Let's call them Bob.                                                                           // 73
  //                                                                                                                   // 74
  // If Bob himself was referred by Alice, then Alice is now eligible for referral credit, as Bob                      // 75
  // has proven he knows how to share.                                                                                 // 76
  //                                                                                                                   // 77
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's referredBy -- let's                        // 78
  // call that Alice.                                                                                                  // 79
  //                                                                                                                   // 80
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's                           // 81
  // referredAccountIds to point at Bob's Account, and then remove the referredBy from Bob's                           // 82
  // Account since it has become redundant.                                                                            // 83
  //                                                                                                                   // 84
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a                       // 85
  // link with Bob, even if Bob already had an account.                                                                // 86
  // Bail out early if referrals aren't enabled                                                                        // 88
  if (!db.isReferralEnabled()) {                                                                                       // 89
    return;                                                                                                            // 90
  }                                                                                                                    // 91
                                                                                                                       //
  var bobAccount = Meteor.users.findOne({                                                                              // 93
    _id: bobAccountId                                                                                                  // 93
  }); // Bail out if Bob is already a complete referral.                                                               // 93
                                                                                                                       //
  if (bobAccount.referredByComplete) {                                                                                 // 96
    return;                                                                                                            // 97
  } // Bail out if Bob wasn't referred by anyone.                                                                      // 98
                                                                                                                       //
                                                                                                                       //
  if (!bobAccount.referredBy) {                                                                                        // 101
    return;                                                                                                            // 102
  }                                                                                                                    // 103
                                                                                                                       //
  var aliceAccountId = bobAccount.referredBy; // Store Bob's Account.referralCompletedBy.                              // 105
                                                                                                                       //
  var now = new Date();                                                                                                // 108
  Meteor.users.update({                                                                                                // 109
    _id: bobAccountId,                                                                                                 // 110
    referredBy: {                                                                                                      // 111
      $exists: true                                                                                                    // 111
    },                                                                                                                 // 111
    referredByComplete: {                                                                                              // 112
      $exists: false                                                                                                   // 112
    }                                                                                                                  // 112
  }, {                                                                                                                 // 109
    $rename: {                                                                                                         // 114
      referredBy: "referredByComplete"                                                                                 // 115
    },                                                                                                                 // 114
    $set: {                                                                                                            // 117
      referredCompleteDate: now                                                                                        // 118
    }                                                                                                                  // 117
  }); // Update Alice's Account.referredAccountIds.                                                                    // 113
                                                                                                                       //
  Meteor.users.update({                                                                                                // 123
    _id: aliceAccountId                                                                                                // 123
  }, {                                                                                                                 // 123
    $push: {                                                                                                           // 124
      referredAccountIds: bobAccountId                                                                                 // 124
    }                                                                                                                  // 124
  });                                                                                                                  // 123
} // User-agent strings that should be allowed to use http basic authentication.                                       // 126
// These are regex matches, so ensure they are escaped properly with double                                            // 129
// backslashes. For security reasons, we MUST NOT whitelist any user-agents                                            // 130
// that may render html and execute embedded scripts.                                                                  // 131
                                                                                                                       //
                                                                                                                       //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");                          // 141
var SESSION_PROXY_TIMEOUT = 60000;                                                                                     // 143
var sandstormCoreFactory = makeSandstormCoreFactory(globalDb);                                                         // 145
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;                                         // 146
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);                                  // 147
var sandstormBackend = sandstormBackendConnection.restore(null, Backend); // TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.               // 151
                                                                                                                       //
globalBackend = new SandstormBackend(globalDb, sandstormBackend);                                                      // 152
Meteor.onConnection(function (connection) {                                                                            // 153
  connection.sandstormBackend = globalBackend;                                                                         // 154
}); // We've observed a problem in production where occasionally the front-end stops talking to the                    // 155
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it                      // 158
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.                      // 159
// Restarting the node process (and only the node process) always fixes the problem.                                   // 160
//                                                                                                                     // 161
// Here, I've added some code that attempts to detect the problem by doing a health check                              // 162
// periodically and then remaking the connection if it seems broken. We'll see if this helps!                          // 163
                                                                                                                       //
var backendHealthy = true;                                                                                             // 164
var disconnectCount = 0;                                                                                               // 165
Meteor.setInterval(function () {                                                                                       // 166
  if (!backendHealthy) {                                                                                               // 167
    if (disconnectCount++ > 2) process.abort();                                                                        // 168
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");                                     // 169
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);                                        // 170
    sandstormBackendConnection.close();                                                                                // 171
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);                                  // 172
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);                                              // 173
    globalBackend._backendCap = sandstormBackend;                                                                      // 174
  }                                                                                                                    // 175
                                                                                                                       //
  var debugLog = !backendHealthy;                                                                                      // 177
  backendHealthy = false;                                                                                              // 178
  var promise = sandstormBackend.ping().then(function () {                                                             // 179
    backendHealthy = true;                                                                                             // 180
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);                                       // 181
  }, function (err) {                                                                                                  // 182
    console.error("error: Backend ping threw error!", err.stack); // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.                         // 185
    // So we don't expect this branch to execute in any case.                                                          // 186
  });                                                                                                                  // 187
                                                                                                                       //
  if (debugLog) {                                                                                                      // 189
    console.log("capnp.js: outer promise:", promise);                                                                  // 190
  }                                                                                                                    // 191
}, 30000); // =======================================================================================                  // 192
// API for creating / starting grains from Meteor methods.                                                             // 195
                                                                                                                       //
var proxiesByHostId = {};                                                                                              // 197
                                                                                                                       //
function parsePowerboxDescriptorList(list) {                                                                           // 199
  return list.map(function (packedDescriptor) {                                                                        // 200
    return Capnp.parse(Powerbox.PowerboxDescriptor, new Buffer(packedDescriptor, "base64"), {                          // 200
      packed: true                                                                                                     // 202
    });                                                                                                                // 202
  });                                                                                                                  // 200
}                                                                                                                      // 203
                                                                                                                       //
Meteor.methods({                                                                                                       // 205
  newGrain: function (packageId, command, title, obsolete) {                                                           // 206
    // Create and start a new grain.                                                                                   // 207
    check(packageId, String);                                                                                          // 209
    check(command, Object); // Manifest.Command from package.capnp.                                                    // 210
                                                                                                                       //
    check(title, String);                                                                                              // 211
                                                                                                                       //
    if (!this.userId) {                                                                                                // 213
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");                              // 214
    }                                                                                                                  // 215
                                                                                                                       //
    if (!isSignedUpOrDemo()) {                                                                                         // 217
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");              // 218
    }                                                                                                                  // 220
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 222
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 223
    }                                                                                                                  // 225
                                                                                                                       //
    var pkg = Packages.findOne(packageId);                                                                             // 227
    var isDev = false;                                                                                                 // 228
    var mountProc = false;                                                                                             // 229
                                                                                                                       //
    if (!pkg) {                                                                                                        // 230
      // Maybe they wanted a dev package.  Check there too.                                                            // 231
      pkg = DevPackages.findOne(packageId);                                                                            // 232
      isDev = true;                                                                                                    // 233
      mountProc = pkg.mountProc;                                                                                       // 234
    }                                                                                                                  // 235
                                                                                                                       //
    if (!pkg) {                                                                                                        // 237
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");                                       // 238
    }                                                                                                                  // 239
                                                                                                                       //
    var appId = pkg.appId;                                                                                             // 241
    var manifest = pkg.manifest;                                                                                       // 242
    var grainId = Random.id(22); // 128 bits of entropy                                                                // 243
                                                                                                                       //
    Grains.insert({                                                                                                    // 244
      _id: grainId,                                                                                                    // 245
      packageId: packageId,                                                                                            // 246
      appId: appId,                                                                                                    // 247
      appVersion: manifest.appVersion,                                                                                 // 248
      userId: this.userId,                                                                                             // 249
      identityId: SandstormDb.generateIdentityId(),                                                                    // 250
      title: title,                                                                                                    // 251
      "private": true,                                                                                                 // 252
      size: 0                                                                                                          // 253
    });                                                                                                                // 244
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev, mountProc);                // 256
    globalBackend.updateLastActive(grainId, this.userId);                                                              // 258
    return grainId;                                                                                                    // 259
  },                                                                                                                   // 260
  openSession: function (grainId, revealIdentity, cachedSalt, options) {                                               // 262
    // Open a new UI session on an existing grain.  Starts the grain if it is not already                              // 263
    // running.                                                                                                        // 264
    options = options || {};                                                                                           // 266
    check(grainId, String);                                                                                            // 268
    revealIdentity = !!revealIdentity;                                                                                 // 269
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 270
    check(options, {                                                                                                   // 271
      powerboxRequest: Match.Optional({                                                                                // 272
        descriptors: [String],                                                                                         // 273
        requestingSession: String                                                                                      // 274
      })                                                                                                               // 272
    });                                                                                                                // 271
    var accountId = revealIdentity ? this.userId : null;                                                               // 278
    var grain = Grains.findOne({                                                                                       // 280
      _id: grainId                                                                                                     // 280
    });                                                                                                                // 280
                                                                                                                       //
    if (!grain) {                                                                                                      // 281
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + grainId);                                          // 282
    } else if (grain.trashed) {                                                                                        // 283
      throw new Meteor.Error("grain-is-in-trash", "Grain is in trash", "Grain ID: " + grainId);                        // 284
    } else if (grain.suspended) {                                                                                      // 285
      throw new Meteor.Error("grain-owner-suspended", "Grain's owner is suspended", "Grain ID: " + grainId);           // 286
    }                                                                                                                  // 288
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 290
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 291
                                                                                                                       //
    if (!SandstormPermissions.mayOpenGrain(db, {                                                                       // 292
      grain: {                                                                                                         // 293
        _id: grainId,                                                                                                  // 293
        accountId: accountId                                                                                           // 293
      }                                                                                                                // 293
    })) {                                                                                                              // 293
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                       // 294
    }                                                                                                                  // 295
                                                                                                                       //
    var sessionFields = {};                                                                                            // 297
                                                                                                                       //
    if (options.powerboxRequest) {                                                                                     // 298
      sessionFields.powerboxRequest = {                                                                                // 299
        descriptors: parsePowerboxDescriptorList(options.powerboxRequest.descriptors),                                 // 300
        requestingSession: options.powerboxRequest.requestingSession                                                   // 301
      };                                                                                                               // 299
    }                                                                                                                  // 303
                                                                                                                       //
    var opened = globalBackend.openSessionInternal(grain, this.userId, revealIdentity, null, null, cachedSalt, sessionFields);
    var result = opened.methodResult;                                                                                  // 307
    var proxy = new Proxy(grain, result.sessionId, result.hostId, result.tabId, accountId, result.identityId, false, null, opened.supervisor);
                                                                                                                       //
    if (sessionFields.powerboxRequest) {                                                                               // 311
      proxy.powerboxRequest = sessionFields.powerboxRequest;                                                           // 312
    }                                                                                                                  // 313
                                                                                                                       //
    proxiesByHostId[result.hostId] = proxy;                                                                            // 315
    return result;                                                                                                     // 316
  },                                                                                                                   // 317
  openSessionFromApiToken: function (params, revealIdentity, cachedSalt, neverRedeem, parentOrigin, options) {         // 319
    // Given an API token, either opens a new WebSession to the underlying grain or returns a                          // 320
    // path to which the client should redirect in order to open such a session.                                       // 321
    //                                                                                                                 // 322
    // `parentOrigin` is the origin of the parent frame of the grain iframe. This may differ from                      // 323
    // ROOT_URL for standalone domains.                                                                                // 324
    neverRedeem = neverRedeem || false;                                                                                // 326
    parentOrigin = parentOrigin || process.env.ROOT_URL;                                                               // 327
    options = options || {};                                                                                           // 328
    check(params, {                                                                                                    // 330
      token: String,                                                                                                   // 331
      incognito: Match.Optional(Boolean) // obsolete, ignored                                                          // 332
                                                                                                                       //
    });                                                                                                                // 330
    revealIdentity = !!revealIdentity;                                                                                 // 334
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 335
    check(neverRedeem, Boolean);                                                                                       // 336
    check(parentOrigin, String);                                                                                       // 337
    check(options, {                                                                                                   // 338
      powerboxRequest: Match.Optional({                                                                                // 339
        descriptors: [String],                                                                                         // 340
        requestingSession: String                                                                                      // 341
      })                                                                                                               // 339
    });                                                                                                                // 338
    var accountId = revealIdentity ? this.userId : null;                                                               // 345
                                                                                                                       //
    if (!accountId && globalDb.getOrganizationDisallowGuests()) {                                                      // 347
      throw new Meteor.Error("guestDisallowed", "This Sandstorm server does not allow " + "guests or anonymous users");
    }                                                                                                                  // 350
                                                                                                                       //
    var token = params.token;                                                                                          // 352
    var incognito = !accountId;                                                                                        // 353
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                      // 354
    var apiToken = ApiTokens.findOne(hashedToken);                                                                     // 355
    validateWebkey(apiToken);                                                                                          // 356
    var grain = Grains.findOne({                                                                                       // 357
      _id: apiToken.grainId                                                                                            // 357
    });                                                                                                                // 357
                                                                                                                       //
    if (!grain) {                                                                                                      // 358
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);                                 // 359
    }                                                                                                                  // 360
                                                                                                                       //
    if (apiToken.accountId) {                                                                                          // 362
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);                                                 // 363
    }                                                                                                                  // 364
                                                                                                                       //
    var title = void 0;                                                                                                // 366
                                                                                                                       //
    if (grain.userId === apiToken.accountId) {                                                                         // 367
      title = grain.title;                                                                                             // 368
    } else {                                                                                                           // 369
      var sharerToken = apiToken.accountId && ApiTokens.findOne({                                                      // 370
        grainId: apiToken.grainId,                                                                                     // 372
        "owner.user.accountId": apiToken.accountId                                                                     // 373
      }, {                                                                                                             // 371
        sort: {                                                                                                        // 375
          lastUsed: -1                                                                                                 // 376
        }                                                                                                              // 375
      });                                                                                                              // 374
                                                                                                                       //
      if (sharerToken) {                                                                                               // 379
        title = sharerToken.owner.user.title;                                                                          // 380
      } else {                                                                                                         // 381
        title = "shared grain";                                                                                        // 382
      }                                                                                                                // 383
    }                                                                                                                  // 384
                                                                                                                       //
    if (this.userId && !incognito && !neverRedeem) {                                                                   // 386
      if (accountId != apiToken.accountId && accountId != grain.userId && !ApiTokens.findOne({                         // 387
        "owner.user.accountId": accountId,                                                                             // 388
        parentToken: hashedToken                                                                                       // 388
      })) {                                                                                                            // 388
        var owner = {                                                                                                  // 389
          user: {                                                                                                      // 389
            accountId: accountId,                                                                                      // 389
            title: title                                                                                               // 389
          }                                                                                                            // 389
        }; // Create a new API token for the account redeeming this token.                                             // 389
                                                                                                                       //
        var result = SandstormPermissions.createNewApiToken(globalDb, {                                                // 392
          rawParentToken: token                                                                                        // 393
        }, apiToken.grainId, apiToken.petname || "redeemed webkey", {                                                  // 393
          allAccess: null                                                                                              // 395
        }, owner);                                                                                                     // 395
        globalDb.addContact(apiToken.accountId, accountId); // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference                // 399
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For                        // 400
        // share-by-identity, this happens immediately.)                                                               // 401
                                                                                                                       //
        if (result.parentApiToken) {                                                                                   // 402
          var parentApiToken = result.parentApiToken;                                                                  // 403
                                                                                                                       //
          if (parentApiToken.forSharing && parentApiToken.accountId) {                                                 // 404
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, parentApiToken.accountId);                  // 405
          }                                                                                                            // 407
        }                                                                                                              // 408
      }                                                                                                                // 409
                                                                                                                       //
      return {                                                                                                         // 411
        redirectToGrain: apiToken.grainId                                                                              // 411
      };                                                                                                               // 411
    } else {                                                                                                           // 412
      if (!SandstormPermissions.mayOpenGrain(globalDb, {                                                               // 413
        token: apiToken                                                                                                // 413
      })) {                                                                                                            // 413
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                     // 414
      }                                                                                                                // 416
                                                                                                                       //
      var sessionFields = {};                                                                                          // 418
                                                                                                                       //
      if (options.powerboxRequest) {                                                                                   // 419
        sessionFields.powerboxRequest = {                                                                              // 420
          descriptors: parsePowerboxDescriptorList(options.powerboxRequest.descriptors),                               // 421
          requestingSession: options.powerboxRequest.requestingSession                                                 // 422
        };                                                                                                             // 420
      } // Even in incognito mode, we want to record the user ID on the session. The user ID is not                    // 424
      // leaked to the app in any way -- it is used to decide what options to show in a powerbox                       // 427
      // request. Even if the user is incognito, they should be able to offer their full range of                      // 428
      // capabilities.                                                                                                 // 429
      //                                                                                                               // 430
      // The identity ID passed here IS revealed to the app, but for incognito mode it is always                       // 431
      // null/undefined.                                                                                               // 432
                                                                                                                       //
                                                                                                                       //
      var opened = globalBackend.openSessionInternal(grain, this.userId, revealIdentity, title, apiToken, cachedSalt, sessionFields);
      var _result = opened.methodResult;                                                                               // 436
      var proxy = new Proxy(grain, _result.sessionId, _result.hostId, _result.tabId, accountId, _result.identityId, false, parentOrigin, opened.supervisor);
      proxy.apiToken = apiToken;                                                                                       // 439
                                                                                                                       //
      if (sessionFields.powerboxRequest) {                                                                             // 440
        proxy.powerboxRequest = sessionFields.powerboxRequest;                                                         // 441
      }                                                                                                                // 442
                                                                                                                       //
      proxiesByHostId[_result.hostId] = proxy;                                                                         // 444
      return _result;                                                                                                  // 445
    }                                                                                                                  // 446
  },                                                                                                                   // 447
  keepSessionAlive: function (sessionId) {                                                                             // 449
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared                    // 450
    //   by them.                                                                                                      // 451
    check(sessionId, String);                                                                                          // 452
    var result = Sessions.findAndModify({                                                                              // 454
      query: {                                                                                                         // 455
        _id: sessionId                                                                                                 // 455
      },                                                                                                               // 455
      update: {                                                                                                        // 456
        $set: {                                                                                                        // 456
          timestamp: new Date().getTime()                                                                              // 456
        }                                                                                                              // 456
      },                                                                                                               // 456
      fields: {                                                                                                        // 457
        grainId: 1,                                                                                                    // 457
        identityId: 1,                                                                                                 // 457
        hostId: 1                                                                                                      // 457
      }                                                                                                                // 457
    });                                                                                                                // 454
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 460
      return false;                                                                                                    // 461
    }                                                                                                                  // 462
                                                                                                                       //
    var session = result.value;                                                                                        // 464
                                                                                                                       //
    if (session) {                                                                                                     // 465
      // Session still present in database, so send keep-alive to backend.                                             // 466
      try {                                                                                                            // 467
        var grainId = session.grainId;                                                                                 // 468
        var hostId = session.hostId;                                                                                   // 469
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;                                // 470
                                                                                                                       //
        if (!supervisor) {                                                                                             // 471
          supervisor = globalBackend.continueGrain(grainId).supervisor;                                                // 472
        }                                                                                                              // 473
                                                                                                                       //
        waitPromise(supervisor.keepAlive());                                                                           // 475
        globalBackend.updateLastActive(grainId, this.userId);                                                          // 476
      } catch (err) {                                                                                                  // 477
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on                    // 478
        // the next request, so whatever.                                                                              // 479
        if (err.kjType !== "disconnected") {                                                                           // 480
          throw err;                                                                                                   // 481
        }                                                                                                              // 482
      }                                                                                                                // 483
                                                                                                                       //
      return true;                                                                                                     // 485
    } else {                                                                                                           // 486
      return false;                                                                                                    // 487
    }                                                                                                                  // 488
  },                                                                                                                   // 489
  shutdownGrain: function (grainId) {                                                                                  // 491
    check(grainId, String);                                                                                            // 492
    var grain = Grains.findOne(grainId);                                                                               // 493
                                                                                                                       //
    if (!grain || !this.userId || grain.userId !== this.userId) {                                                      // 494
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");                              // 495
    }                                                                                                                  // 496
                                                                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));                                             // 498
  }                                                                                                                    // 499
});                                                                                                                    // 205
                                                                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {                                                        // 502
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If                        // 503
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the                            // 504
  // `expiresIfUnused` field is reset to `refreshedExpiration`.                                                        // 505
  if (!apiToken) {                                                                                                     // 507
    throw new Meteor.Error(403, "Invalid authorization token");                                                        // 508
  }                                                                                                                    // 509
                                                                                                                       //
  if (apiToken.revoked) {                                                                                              // 511
    throw new Meteor.Error(403, "Authorization token has been revoked");                                               // 512
  }                                                                                                                    // 513
                                                                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {                                                               // 515
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");                                             // 516
  }                                                                                                                    // 517
                                                                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {                                                  // 519
    throw new Meteor.Error(403, "Authorization token expired");                                                        // 520
  }                                                                                                                    // 521
                                                                                                                       //
  if (apiToken.expiresIfUnused) {                                                                                      // 523
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {                                                            // 524
      throw new Meteor.Error(403, "Authorization token expired");                                                      // 525
    } else if (refreshedExpiration) {                                                                                  // 526
      ApiTokens.update(apiToken._id, {                                                                                 // 527
        $set: {                                                                                                        // 527
          expiresIfUnused: refreshedExpiration                                                                         // 527
        }                                                                                                              // 527
      });                                                                                                              // 527
    } else {                                                                                                           // 528
      // It's getting used now, so clear the expiresIfUnused field.                                                    // 529
      ApiTokens.update(apiToken._id, {                                                                                 // 530
        $set: {                                                                                                        // 530
          expiresIfUnused: null                                                                                        // 530
        }                                                                                                              // 530
      });                                                                                                              // 530
    }                                                                                                                  // 531
  }                                                                                                                    // 532
                                                                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                                                                     // 534
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");                                       // 535
  }                                                                                                                    // 536
};                                                                                                                     // 537
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 539
  var shutdownApp = function (appId) {                                                                                 // 540
    Grains.find({                                                                                                      // 541
      appId: appId                                                                                                     // 541
    }).forEach(function (grain) {                                                                                      // 541
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));                                               // 542
    });                                                                                                                // 543
  };                                                                                                                   // 544
                                                                                                                       //
  DevPackages.find().observe({                                                                                         // 546
    removed: function (devPackage) {                                                                                   // 547
      shutdownApp(devPackage.appId);                                                                                   // 547
    },                                                                                                                 // 547
    changed: function (newDevPackage, oldDevPackage) {                                                                 // 549
      shutdownApp(oldDevPackage.appId);                                                                                // 550
                                                                                                                       //
      if (oldDevPackage.appId !== newDevPackage.appId) {                                                               // 551
        shutdownApp(newDevPackage.appId);                                                                              // 552
      }                                                                                                                // 553
    },                                                                                                                 // 554
    added: function (devPackage) {                                                                                     // 556
      shutdownApp(devPackage.appId);                                                                                   // 556
    }                                                                                                                  // 556
  });                                                                                                                  // 546
  Sessions.find().observe({                                                                                            // 559
    removed: function (session) {                                                                                      // 560
      var proxy = proxiesByHostId[session.hostId];                                                                     // 561
      delete proxiesByHostId[session.hostId];                                                                          // 562
                                                                                                                       //
      if (proxy) {                                                                                                     // 563
        proxy.close();                                                                                                 // 564
      }                                                                                                                // 565
    }                                                                                                                  // 566
  });                                                                                                                  // 559
}); // Kill off sessions idle for >~3 minutes.                                                                         // 568
                                                                                                                       //
var TIMEOUT_MS = 180000;                                                                                               // 571
                                                                                                                       //
var gcSessions = function () {                                                                                         // 572
  var now = new Date().getTime();                                                                                      // 573
  Sessions.remove({                                                                                                    // 574
    timestamp: {                                                                                                       // 574
      $lt: now - TIMEOUT_MS                                                                                            // 574
    }                                                                                                                  // 574
  });                                                                                                                  // 574
};                                                                                                                     // 575
                                                                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                                                                   // 577
                                                                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {                                                           // 579
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it                       // 580
  // doesn't already exist. The first request on the session host will always create a new proxy.                      // 581
  // Later requests may create a proxy if they go to a different front-end replica or if the                           // 582
  // front-end was restarted.                                                                                          // 583
  check(hostId, String);                                                                                               // 584
  return Promise.resolve(undefined).then(function () {                                                                 // 586
    var proxy = proxiesByHostId[hostId];                                                                               // 587
                                                                                                                       //
    if (proxy) {                                                                                                       // 588
      return proxy;                                                                                                    // 589
    } else {                                                                                                           // 590
      // Set table entry to null for now so that we can detect if it is concurrently deleted.                          // 591
      proxiesByHostId[hostId] = null;                                                                                  // 592
      return inMeteor(function () {                                                                                    // 594
        var session = Sessions.findOne({                                                                               // 595
          hostId: hostId                                                                                               // 595
        });                                                                                                            // 595
                                                                                                                       //
        if (!session) {                                                                                                // 596
          if (isAlreadyOpened) {                                                                                       // 597
            var observer = void 0;                                                                                     // 598
            return new Promise(function (resolve, reject) {                                                            // 599
              var task = Meteor.setTimeout(function () {                                                               // 600
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                                                                               // 606
              observer = Sessions.find({                                                                               // 607
                hostId: hostId                                                                                         // 607
              }).observe({                                                                                             // 607
                added: function () {                                                                                   // 608
                  Meteor.clearTimeout(task);                                                                           // 609
                  resolve(getProxyForHostId(hostId, false));                                                           // 610
                }                                                                                                      // 611
              });                                                                                                      // 607
            }).then(function (v) {                                                                                     // 613
              observer.stop();                                                                                         // 614
              return v;                                                                                                // 615
            }, function (e) {                                                                                          // 616
              observer.stop();                                                                                         // 617
              throw e;                                                                                                 // 618
            });                                                                                                        // 619
          } else {                                                                                                     // 620
            // Does not appear to be a valid session host.                                                             // 621
            return undefined;                                                                                          // 622
          }                                                                                                            // 623
        }                                                                                                              // 624
                                                                                                                       //
        var apiToken = void 0;                                                                                         // 626
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 627
          apiToken = ApiTokens.findOne({                                                                               // 628
            _id: session.hashedToken                                                                                   // 628
          }); // We don't have to fully validate the API token here because if it changed the session                  // 628
          // would have been deleted.                                                                                  // 630
                                                                                                                       //
          if (!apiToken) {                                                                                             // 631
            throw new Meteor.Error(410, "ApiToken has been deleted");                                                  // 632
          }                                                                                                            // 633
        }                                                                                                              // 634
                                                                                                                       //
        var grain = Grains.findOne(session.grainId);                                                                   // 636
                                                                                                                       //
        if (!grain) {                                                                                                  // 637
          // Grain was deleted, I guess.                                                                               // 638
          throw new Meteor.Error(410, "Resource has been deleted");                                                    // 639
        } // Note that we don't need to call mayOpenGrain() because the existence of a session                         // 640
        // implies this check was already performed.                                                                   // 643
                                                                                                                       //
                                                                                                                       //
        var proxy = new Proxy(grain, session._id, hostId, session.tabId, session.userId, session.identityId, false);   // 645
        if (apiToken) proxy.apiToken = apiToken;                                                                       // 647
                                                                                                                       //
        if (session.powerboxRequest) {                                                                                 // 649
          proxy.powerboxRequest = session.powerboxRequest;                                                             // 650
        } // Only add the proxy to the table if it was not concurrently deleted (which could happen                    // 651
        // e.g. if the user's access was revoked).                                                                     // 654
                                                                                                                       //
                                                                                                                       //
        if (hostId in proxiesByHostId) {                                                                               // 655
          proxiesByHostId[hostId] = proxy;                                                                             // 656
        } else {                                                                                                       // 657
          throw new Meteor.Error(403, "Session was concurrently closed.");                                             // 658
        }                                                                                                              // 659
                                                                                                                       //
        return proxy;                                                                                                  // 661
      });                                                                                                              // 662
    }                                                                                                                  // 663
  });                                                                                                                  // 664
}; // =======================================================================================                          // 665
// API tokens                                                                                                          // 668
                                                                                                                       //
                                                                                                                       //
var ApiSessionProxies = function () {                                                                                  //
  // Class that caches proxies for requests that come in through HTTP API endpoints. Such                              // 671
  // requests are not associated with any entry in the `Sessions` collection, so we need                               // 672
  // some other means of detecting unused proxies and closing them. This class works by keeping                        // 673
  // the proxies in two buckets. When a proxy is used, it is put in the new bucket. Periodically,                      // 674
  // the old bucket is emptied and the contents of new bucket are moved to the old bucket.                             // 675
  //                                                                                                                   // 676
  // Each bucket is a two-level map, keyed by hashes of the tokens and their `ApiSession.Params`                       // 677
  // structs.                                                                                                          // 678
  function ApiSessionProxies(intervalMillis) {                                                                         // 680
    var _this2 = this;                                                                                                 // 680
                                                                                                                       //
    (0, _classCallCheck3.default)(this, ApiSessionProxies);                                                            // 680
    check(intervalMillis, Number);                                                                                     // 681
    this.newBucket = {};                                                                                               // 682
    this.oldBucket = {};                                                                                               // 683
    this.interval = Meteor.setInterval(function () {                                                                   // 684
      for (var oldHashedToken in meteorBabelHelpers.sanitizeForInObject(_this2.oldBucket)) {                           // 685
        for (var oldHashedParams in meteorBabelHelpers.sanitizeForInObject(_this2.oldBucket[oldHashedToken])) {        // 686
          var newProxy = _this2.newBucket[oldHashedToken] && _this2.newBucket[oldHashedToken][oldHashedParams];        // 687
          var oldProxy = _this2.oldBucket[oldHashedToken][oldHashedParams];                                            // 689
                                                                                                                       //
          if (oldProxy && !newProxy) {                                                                                 // 690
            if (Object.keys(oldProxy.websockets).length > 0 || Object.keys(oldProxy.streamingRequests).length > 0 || Object.keys(oldProxy.streamingResponses).length > 0) {
              // A client has an open stream. Keep the proxy around.                                                   // 694
              _this2.put(oldHashedToken, oldHashedParams, oldProxy);                                                   // 695
            } else {                                                                                                   // 696
              // We can close this proxy and forget about it.                                                          // 697
              oldProxy.close();                                                                                        // 698
            }                                                                                                          // 699
          }                                                                                                            // 700
        }                                                                                                              // 701
      }                                                                                                                // 702
                                                                                                                       //
      _this2.oldBucket = _this2.newBucket;                                                                             // 704
      _this2.newBucket = {};                                                                                           // 705
      var keepAlives = [];                                                                                             // 707
                                                                                                                       //
      for (var hashedToken in meteorBabelHelpers.sanitizeForInObject(_this2.oldBucket)) {                              // 708
        for (var hashedParams in meteorBabelHelpers.sanitizeForInObject(_this2.oldBucket[hashedToken])) {              // 709
          var proxy = _this2.oldBucket[hashedToken][hashedParams];                                                     // 710
          keepAlives.push(proxy.supervisor.keepAlive().catch(function (e) {                                            // 711
            console.error("supervisor.keepAlive() failed.", e);                                                        // 712
          }));                                                                                                         // 713
        }                                                                                                              // 714
      }                                                                                                                // 715
                                                                                                                       //
      waitPromise(Promise.all(keepAlives));                                                                            // 717
    }, intervalMillis);                                                                                                // 718
  }                                                                                                                    // 719
                                                                                                                       //
  ApiSessionProxies.prototype.get = function () {                                                                      //
    function get(hashedToken, hashedParams) {                                                                          //
      if (this.newBucket[hashedToken] && this.newBucket[hashedToken][hashedParams]) {                                  // 722
        return this.newBucket[hashedToken][hashedParams];                                                              // 723
      } else if (this.oldBucket[hashedToken] && this.oldBucket[hashedToken][hashedParams]) {                           // 724
        var proxy = this.oldBucket[hashedToken][hashedParams];                                                         // 725
        delete this.oldBucket[hashedToken][hashedParams];                                                              // 726
        this.put(hashedToken, hashedParams, proxy);                                                                    // 727
        return proxy;                                                                                                  // 728
      } else {                                                                                                         // 729
        return null;                                                                                                   // 730
      }                                                                                                                // 731
    }                                                                                                                  // 732
                                                                                                                       //
    return get;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  ApiSessionProxies.prototype.put = function () {                                                                      //
    function put(hashedToken, hashedParams, proxy) {                                                                   //
      if (!this.newBucket[hashedToken]) {                                                                              // 735
        this.newBucket[hashedToken] = {};                                                                              // 736
      }                                                                                                                // 737
                                                                                                                       //
      this.newBucket[hashedToken][hashedParams] = proxy;                                                               // 739
    }                                                                                                                  // 740
                                                                                                                       //
    return put;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  ApiSessionProxies.prototype.removeProxiesOfToken = function () {                                                     //
    function removeProxiesOfToken(hashedToken) {                                                                       //
      if (hashedToken in this.oldBucket) {                                                                             // 743
        for (var hashedParams in meteorBabelHelpers.sanitizeForInObject(this.oldBucket[hashedToken])) {                // 744
          this.oldBucket[hashedToken][hashedParams].close();                                                           // 745
        }                                                                                                              // 746
                                                                                                                       //
        delete this.oldBucket[hashedToken];                                                                            // 748
      }                                                                                                                // 749
                                                                                                                       //
      if (hashedToken in this.newBucket) {                                                                             // 751
        for (var _hashedParams in meteorBabelHelpers.sanitizeForInObject(this.newBucket[hashedToken])) {               // 752
          this.newBucket[hashedToken][_hashedParams].close();                                                          // 753
        }                                                                                                              // 754
                                                                                                                       //
        delete this.newBucket[hashedToken];                                                                            // 756
      }                                                                                                                // 757
    }                                                                                                                  // 758
                                                                                                                       //
    return removeProxiesOfToken;                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  return ApiSessionProxies;                                                                                            //
}(); // The interval of 60 seconds here deliberately matches the `keepSessionAlive()` interval                         //
// in the case of non-API proxies.                                                                                     // 762
                                                                                                                       //
                                                                                                                       //
var apiSessionProxies = new ApiSessionProxies(60 * 1000);                                                              // 763
Meteor.startup(function () {                                                                                           // 765
  Grains.find().observe({                                                                                              // 766
    changed: function (newGrain, oldGrain) {                                                                           // 767
      if (oldGrain.private != newGrain.private) {                                                                      // 768
        Sessions.remove({                                                                                              // 769
          grainId: oldGrain._id,                                                                                       // 769
          userId: {                                                                                                    // 769
            $ne: oldGrain.userId                                                                                       // 769
          }                                                                                                            // 769
        });                                                                                                            // 769
        ApiTokens.find({                                                                                               // 770
          grainId: oldGrain._id                                                                                        // 770
        }).forEach(function (apiToken) {                                                                               // 770
          apiSessionProxies.removeProxiesOfToken(apiToken._id);                                                        // 771
        });                                                                                                            // 772
      }                                                                                                                // 773
    }                                                                                                                  // 774
  });                                                                                                                  // 766
});                                                                                                                    // 776
                                                                                                                       //
function getApiSessionParams(request) {                                                                                // 778
  var params = {};                                                                                                     // 779
                                                                                                                       //
  if ("x-sandstorm-passthrough" in request.headers) {                                                                  // 780
    var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {                              // 781
      return s.trim();                                                                                                 // 783
    }); // The only currently supported passthrough value is 'address', but others could be useful in                  // 783
    // the future.                                                                                                     // 785
                                                                                                                       //
    if (optIns.indexOf("address") !== -1) {                                                                            // 787
      // Sadly, we can't use request.socket.remoteFamily because it's not available in the                             // 788
      // (rather-old) version of node that comes in the Meteor bundle we're using. Hence this                          // 789
      // hackery.                                                                                                      // 790
      var addressToPass = request.socket.remoteAddress;                                                                // 791
                                                                                                                       //
      if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {                                         // 792
        // Allow overriding the socket's remote address with X-Real-IP header if the request comes                     // 793
        // from either localhost or an RFC1918 address. These are not useful for geolocation anyway.                   // 794
        addressToPass = request.headers["x-real-ip"];                                                                  // 795
      }                                                                                                                // 796
                                                                                                                       //
      if (Net.isIPv4(addressToPass)) {                                                                                 // 798
        // Map IPv4 addresses in IPv6.                                                                                 // 799
        // This conveniently comes out to a 48-bit number, which is precisely representable in a                       // 800
        // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we                       // 801
        // might otherwise need to precisely represent 64-bit fields.                                                  // 802
        var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {                                       // 803
          return parseInt(x, 10);                                                                                      // 804
        }).reduce(function (a, b) {                                                                                    // 804
          return 256 * a + b;                                                                                          // 805
        });                                                                                                            // 805
        params.remoteAddress = {                                                                                       // 806
          lower64: v4Int,                                                                                              // 807
          upper64: 0                                                                                                   // 808
        };                                                                                                             // 806
      } else if (Net.isIPv6(addressToPass)) {                                                                          // 810
        // TODO(test): Unit test this                                                                                  // 811
        // Parse a valid v6 address.                                                                                   // 812
        // Split into groups, then insert an appropriate number of 0's if :: was used.                                 // 813
        var groups = addressToPass.split(":"); // Strip extra empty group in the case of a leading or trailing '::'.   // 814
                                                                                                                       //
        if (groups[0] === "") {                                                                                        // 817
          groups.shift();                                                                                              // 818
        }                                                                                                              // 819
                                                                                                                       //
        if (groups[groups.length - 1] === "") {                                                                        // 821
          groups.pop();                                                                                                // 822
        }                                                                                                              // 823
                                                                                                                       //
        var lastGroup = groups[groups.length - 1]; // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
        // should expand into two groups of 4-character hex strings, like the rest of the address.                     // 827
                                                                                                                       //
        if (Net.isIPv4(lastGroup)) {                                                                                   // 828
          groups.pop();                                                                                                // 829
          var quad = lastGroup.split(".").map(function (x) {                                                           // 830
            return parseInt(x, 10);                                                                                    // 830
          });                                                                                                          // 830
          groups.push((quad[0] * 256 + quad[1]).toString(16));                                                         // 831
          groups.push((quad[2] * 256 + quad[3]).toString(16));                                                         // 832
        }                                                                                                              // 833
                                                                                                                       //
        var groupsToAdd = 8 - groups.length;                                                                           // 835
        var emptyGroupIndex = groups.indexOf("");                                                                      // 836
        var cleanGroups = void 0;                                                                                      // 837
                                                                                                                       //
        if (emptyGroupIndex !== -1) {                                                                                  // 838
          var head = groups.slice(0, emptyGroupIndex); // groupsToAdd + 1 because we sliced out the empty element      // 839
                                                                                                                       //
          var mid = Array(groupsToAdd + 1);                                                                            // 841
                                                                                                                       //
          for (var i = 0; i < groupsToAdd + 1; i++) {                                                                  // 842
            mid[i] = "0";                                                                                              // 843
          }                                                                                                            // 844
                                                                                                                       //
          var tail = groups.slice(emptyGroupIndex + 1, groups.length);                                                 // 846
          cleanGroups = [].concat(head, mid, tail);                                                                    // 847
        } else {                                                                                                       // 848
          cleanGroups = groups;                                                                                        // 849
        }                                                                                                              // 850
                                                                                                                       //
        var ints = cleanGroups.map(function (x) {                                                                      // 852
          return parseInt(x, 16);                                                                                      // 852
        }); // We use strings because we'd lose data from loss of precision casting the 64-bit uints                   // 852
        // into 53-bit-mantissa doubles.                                                                               // 854
                                                                                                                       //
        params.remoteAddress = {                                                                                       // 855
          upper64: quadToIntString(ints.slice(0, 4)),                                                                  // 856
          lower64: quadToIntString(ints.slice(4, 8))                                                                   // 857
        };                                                                                                             // 855
      }                                                                                                                // 859
    }                                                                                                                  // 860
  }                                                                                                                    // 861
                                                                                                                       //
  return Capnp.serialize(ApiSession.Params, params);                                                                   // 863
} // Used by server/drivers/external-ui/view.js                                                                        // 864
                                                                                                                       //
                                                                                                                       //
getProxyForApiToken = function (token, request) {                                                                      // 867
  check(token, String);                                                                                                // 868
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                        // 869
  var tabId = Crypto.createHash("sha256").update("tab:").update(hashedToken).digest("hex").slice(0, 32);               // 870
  var serializedParams = getApiSessionParams(request);                                                                 // 872
  var hashedParams = Crypto.createHash("sha256").update(serializedParams).digest("base64");                            // 873
  return Promise.resolve(undefined).then(function () {                                                                 // 874
    var proxy = apiSessionProxies.get(hashedToken, hashedParams);                                                      // 875
                                                                                                                       //
    if (proxy) {                                                                                                       // 876
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {                                                    // 877
        throw new Meteor.Error(403, "Authorization token expired");                                                    // 878
      }                                                                                                                // 879
                                                                                                                       //
      return proxy;                                                                                                    // 881
    } else {                                                                                                           // 882
      return inMeteor(function () {                                                                                    // 883
        var tokenInfo = ApiTokens.findOne(hashedToken);                                                                // 884
        validateWebkey(tokenInfo);                                                                                     // 885
        var grain = Grains.findOne(tokenInfo.grainId);                                                                 // 887
                                                                                                                       //
        if (!grain) {                                                                                                  // 888
          // Grain was deleted, I guess.                                                                               // 889
          throw new Meteor.Error(410, "Resource has been deleted");                                                    // 890
        }                                                                                                              // 891
                                                                                                                       //
        var proxy = void 0;                                                                                            // 893
                                                                                                                       //
        if (tokenInfo.userInfo) {                                                                                      // 894
          throw new Error("API tokens created with arbitrary userInfo no longer supported");                           // 895
        } else {                                                                                                       // 896
          var identityId = null;                                                                                       // 897
          var accountId = null;                                                                                        // 898
                                                                                                                       //
          if (tokenInfo.accountId && !tokenInfo.forSharing) {                                                          // 899
            accountId = tokenInfo.accountId;                                                                           // 900
            identityId = globalDb.getOrGenerateIdentityId(tokenInfo.accountId, grain);                                 // 901
          }                                                                                                            // 902
                                                                                                                       //
          proxy = new Proxy(grain, null, null, tabId, accountId, identityId, true);                                    // 904
          proxy.apiToken = tokenInfo;                                                                                  // 905
          proxy.apiSessionParams = serializedParams;                                                                   // 906
        }                                                                                                              // 907
                                                                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, {                                                             // 909
          token: tokenInfo                                                                                             // 909
        })) {                                                                                                          // 909
          // Note that only public grains may be opened without a user ID.                                             // 910
          throw new Meteor.Error(403, "Unauthorized.");                                                                // 911
        }                                                                                                              // 912
                                                                                                                       //
        if (tokenInfo.expires) {                                                                                       // 914
          proxy.expires = tokenInfo.expires;                                                                           // 915
        }                                                                                                              // 916
                                                                                                                       //
        apiSessionProxies.put(hashedToken, hashedParams, proxy);                                                       // 918
        return proxy;                                                                                                  // 920
      });                                                                                                              // 921
    }                                                                                                                  // 922
  });                                                                                                                  // 923
};                                                                                                                     // 924
                                                                                                                       //
var apiUseBasicAuth = function (req, hostId) {                                                                         // 926
  // If the request was to a token-specific host *and* the request has no Origin header (meaning                       // 927
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we                       // 928
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,                         // 929
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack                         // 930
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we                       // 931
  // want to allow basic auth at all is because a lot of existing client apps support only basic                       // 932
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.                       // 933
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {                                                 // 934
    return true;                                                                                                       // 935
  } // Historically, all API tokens were served from the same host rather than have token-specific                     // 936
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was                        // 939
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same                    // 940
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when                         // 941
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud                       // 942
  // client app.                                                                                                       // 943
  //                                                                                                                   // 944
  // Since many clients in the wild have already been configured to use the shared API host, we                        // 945
  // must continue to support them, so this logic remains.                                                             // 946
                                                                                                                       //
                                                                                                                       //
  var agent = req.headers["user-agent"];                                                                               // 947
  return agent && agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                                                           // 948
};                                                                                                                     // 949
                                                                                                                       //
var apiTokenForRequest = function (req, hostId) {                                                                      // 951
  // Extract the API token from the request.                                                                           // 952
  var auth = req.headers.authorization;                                                                                // 954
  var token = void 0;                                                                                                  // 955
                                                                                                                       //
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {                                                          // 956
    token = auth.slice(7).trim();                                                                                      // 957
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {                    // 958
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];                                       // 960
  } else if (hostId !== "api" && req.url.startsWith("/.sandstorm-token/")) {                                           // 961
    // Disallow clients that use the old "api" hostId from using tokens in the path.                                   // 962
    // Allowing them would be a security problem in browsers that ignore CSP.                                          // 963
    var parts = req.url.slice(1).split("/"); // remove leading / and split                                             // 964
                                                                                                                       //
    if (parts.length < 2) {                                                                                            // 965
      token = undefined;                                                                                               // 966
    } else {                                                                                                           // 967
      token = parts[1];                                                                                                // 968
      req.url = "/" + parts.slice(2).join("/"); // remove .sandstorm-token/$TOKEN from path                            // 969
    }                                                                                                                  // 971
  } else {                                                                                                             // 972
    token = undefined;                                                                                                 // 973
  }                                                                                                                    // 974
                                                                                                                       //
  if (token && hostId !== "api") {                                                                                     // 976
    // Verify that the token matches the specific host.                                                                // 977
    if (hostId !== globalDb.apiHostIdForToken(token)) {                                                                // 978
      token = undefined;                                                                                               // 979
    }                                                                                                                  // 980
  }                                                                                                                    // 981
                                                                                                                       //
  return token;                                                                                                        // 983
}; // =======================================================================================                          // 984
// Routing to proxies.                                                                                                 // 987
//                                                                                                                     // 988
// pre-meteor.js calls this                                                                                            // 990
                                                                                                                       //
                                                                                                                       //
tryProxyUpgrade = function (hostId, req, socket, head) {                                                               // 991
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that                        // 992
  // resolves true if an appropriate grain is found, false if there was no match (but the caller                       // 993
  // should consider other host types, like static web publishing), or throws an error if the                          // 994
  // request is definitely invalid.                                                                                    // 995
  if (globalDb.isApiHostId(hostId)) {                                                                                  // 997
    var token = apiTokenForRequest(req, hostId);                                                                       // 998
                                                                                                                       //
    if (token) {                                                                                                       // 999
      return getProxyForApiToken(token, req).then(function (proxy) {                                                   // 1000
        // Meteor sets the timeout to five seconds. Change that back to two                                            // 1001
        // minutes, which is the default value.                                                                        // 1002
        socket.setTimeout(120000);                                                                                     // 1003
        proxy.upgradeHandler(req, socket, head);                                                                       // 1005
        return true;                                                                                                   // 1006
      });                                                                                                              // 1007
    } else {                                                                                                           // 1008
      return Promise.resolve(false);                                                                                   // 1009
    }                                                                                                                  // 1010
  } else {                                                                                                             // 1011
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;                   // 1012
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {                                          // 1013
      if (proxy) {                                                                                                     // 1014
        // Cross-origin requests are not allowed on UI session hosts.                                                  // 1015
        var origin = req.headers.origin;                                                                               // 1016
                                                                                                                       //
        if (origin !== PROTOCOL + "//" + req.headers.host) {                                                           // 1017
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);                      // 1018
        } // Meteor sets the timeout to five seconds. Change that back to two                                          // 1019
        // minutes, which is the default value.                                                                        // 1022
                                                                                                                       //
                                                                                                                       //
        socket.setTimeout(120000);                                                                                     // 1023
        proxy.upgradeHandler(req, socket, head);                                                                       // 1025
        return true;                                                                                                   // 1026
      } else {                                                                                                         // 1027
        return false;                                                                                                  // 1028
      }                                                                                                                // 1029
    });                                                                                                                // 1030
  }                                                                                                                    // 1031
}; // pre-meteor.js calls this                                                                                         // 1032
                                                                                                                       //
                                                                                                                       //
tryProxyRequest = function (hostId, req, res) {                                                                        // 1035
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that                            // 1036
  // resolves true if an appropriate grain is found, false if there was no match (but the caller                       // 1037
  // should consider other host types, like static web publishing), or throws an error if the                          // 1038
  // request is definitely invalid.                                                                                    // 1039
  var hostIdHash = globalDb.isApiHostId(hostId);                                                                       // 1041
                                                                                                                       //
  if (hostIdHash) {                                                                                                    // 1042
    // This is a request for the API host.                                                                             // 1043
    res.setHeader("Access-Control-Allow-Origin", "*");                                                                 // 1045
                                                                                                                       //
    if (req.method === "OPTIONS") {                                                                                    // 1047
      // Reply to CORS preflight request.                                                                              // 1048
      // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients                         // 1050
      // must send a valid Authorization header, and since cookies are not used for                                    // 1051
      // authorization, this is perfectly safe. In a sane world, we would only need to send back                       // 1052
      // 'Access-Control-Allow-Origin: *' and be done with it.                                                         // 1053
      //                                                                                                               // 1054
      // However, CORS demands that we explicitly whitelist individual methods and headers for                         // 1055
      // use cross-origin, as if this is somehow useful for implementing any practical security                        // 1056
      // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one                               // 1057
      // individually. We cannot just write '*' for these lists. WTF, CORS?                                            // 1058
      //                                                                                                               // 1059
      // Luckily, the request tells us exactly what method and headers are being requested, so we                      // 1060
      // only need to copy those over, rather than create an exhaustive list. But this is still                        // 1061
      // overly complicated.                                                                                           // 1062
      var accessControlHeaders = {                                                                                     // 1064
        "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",                                         // 1065
        "Access-Control-Max-Age": "3600"                                                                               // 1066
      }; // Copy all requested headers to the allowed headers list.                                                    // 1064
                                                                                                                       //
      var requestedHeaders = req.headers["access-control-request-headers"];                                            // 1070
                                                                                                                       //
      if (requestedHeaders) {                                                                                          // 1071
        accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;                                       // 1072
      } // Add the requested method to the allowed methods list, if it's not there already.                            // 1073
                                                                                                                       //
                                                                                                                       //
      var requestedMethod = req.headers["access-control-request-method"];                                              // 1076
                                                                                                                       //
      if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {        // 1077
        accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;                                // 1079
      }                                                                                                                // 1080
                                                                                                                       //
      for (header in meteorBabelHelpers.sanitizeForInObject(accessControlHeaders)) {                                   // 1082
        res.setHeader(header, accessControlHeaders[header]);                                                           // 1083
      }                                                                                                                // 1084
    }                                                                                                                  // 1085
                                                                                                                       //
    var errorHandler = function (err) {                                                                                // 1087
      if (err instanceof Meteor.Error) {                                                                               // 1088
        console.log("error: " + err);                                                                                  // 1089
        res.writeHead(err.error, err.reason, {                                                                         // 1090
          "Content-Type": "text/plain"                                                                                 // 1090
        });                                                                                                            // 1090
      } else {                                                                                                         // 1091
        res.writeHead(500, "Internal Server Error", {                                                                  // 1092
          "Content-Type": "text/plain"                                                                                 // 1092
        });                                                                                                            // 1092
      }                                                                                                                // 1093
                                                                                                                       //
      res.end(err.stack);                                                                                              // 1095
    };                                                                                                                 // 1096
                                                                                                                       //
    var token = apiTokenForRequest(req, hostId);                                                                       // 1098
                                                                                                                       //
    if (token && req.headers["x-sandstorm-token-keepalive"]) {                                                         // 1099
      inMeteor(function () {                                                                                           // 1100
        var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);                                  // 1101
        check(keepaliveDuration, Match.Integer);                                                                       // 1102
        var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                  // 1103
        validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));                      // 1104
      }).then(function () {                                                                                            // 1105
        res.writeHead(200, {});                                                                                        // 1106
        res.end();                                                                                                     // 1107
      }, errorHandler);                                                                                                // 1108
    } else if (token) {                                                                                                // 1109
      getProxyForApiToken(token, req).then(function (proxy) {                                                          // 1110
        proxy.requestHandler(req, res);                                                                                // 1111
      }, errorHandler);                                                                                                // 1112
    } else {                                                                                                           // 1113
      // No token. Look up static API host info.                                                                       // 1114
      inMeteor(function () {                                                                                           // 1115
        var apiHost = globalDb.collections.apiHosts.findOne(hostIdHash);                                               // 1116
                                                                                                                       //
        if (req.method === "OPTIONS") {                                                                                // 1118
          // OPTIONS request with no authorization token.                                                              // 1119
          var dav = ((apiHost || {}).options || {}).dav || [];                                                         // 1121
                                                                                                                       //
          if (dav.length > 0) {                                                                                        // 1122
            res.setHeader("DAV", dav.join(", "));                                                                      // 1123
            res.setHeader("Access-Control-Expose-Headers", "DAV");                                                     // 1124
          }                                                                                                            // 1125
                                                                                                                       //
          res.writeHead(200, {});                                                                                      // 1127
          res.end();                                                                                                   // 1128
        } else {                                                                                                       // 1129
          var resources = (apiHost || {}).resources || {};                                                             // 1130
          var path = SandstormDb.escapeMongoKey(req.url.split("?")[0]);                                                // 1131
                                                                                                                       //
          if (path in resources) {                                                                                     // 1132
            // Serve a static resource.                                                                                // 1133
            var resource = resources[path];                                                                            // 1134
            console.log(path, resources, resource);                                                                    // 1135
            if (resource.language) res.setHeader("Content-Language", resource.language);                               // 1136
            if (resource.encoding) res.setHeader("Content-Encoding", resource.encoding);                               // 1137
            res.writeHead(200, {                                                                                       // 1138
              "Content-Type": resource.type                                                                            // 1139
            });                                                                                                        // 1138
            res.end(resource.body);                                                                                    // 1141
          } else {                                                                                                     // 1142
            if (apiUseBasicAuth(req, hostId)) {                                                                        // 1143
              res.writeHead(401, {                                                                                     // 1144
                "Content-Type": "text/plain",                                                                          // 1145
                "WWW-Authenticate": "Basic realm='Sandstorm API'"                                                      // 1146
              });                                                                                                      // 1144
            } else {                                                                                                   // 1148
              // TODO(someday): Display some sort of nifty API browser.                                                // 1149
              res.writeHead(403, {                                                                                     // 1150
                "Content-Type": "text/plain"                                                                           // 1150
              });                                                                                                      // 1150
            }                                                                                                          // 1151
                                                                                                                       //
            res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
          }                                                                                                            // 1157
        }                                                                                                              // 1158
      }).then(function () {}, errorHandler);                                                                           // 1159
    }                                                                                                                  // 1160
                                                                                                                       //
    return Promise.resolve(true);                                                                                      // 1162
  } else {                                                                                                             // 1163
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;                   // 1164
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {                                          // 1167
      if (proxy) {                                                                                                     // 1168
        // Let's do some CSRF defense.                                                                                 // 1169
        var expectedOrigin = PROTOCOL + "//" + req.headers.host;                                                       // 1170
        var mainUrl = proxy.parentOrigin;                                                                              // 1171
        var origin = req.headers.origin;                                                                               // 1172
        var referer = req.headers.referer;                                                                             // 1173
                                                                                                                       //
        if (origin) {                                                                                                  // 1174
          // If an origin header was provided, then it must be accurate. Note that Chrome and                          // 1175
          // Safari always send an Origin header on non-GET requests (even same-origin), and                           // 1176
          // therefore on those browsers this rule will block all CSRF attacks. Unfortunately,                         // 1177
          // a patch to add this behavior to Firefox has stalled:                                                      // 1178
          //     https://bugzilla.mozilla.org/show_bug.cgi?id=446344                                                   // 1179
          // I haven't found any information on IE/Edge's behavior.                                                    // 1180
          if (origin !== expectedOrigin) {                                                                             // 1181
            // TODO(security): Alas, it turns out we have apps that have:                                              // 1182
            //   <meta name="referrer" content="no-referrer">                                                          // 1183
            // as Chrome sends "Origin: null" in these cases. :( These apps need to switch to:                         // 1184
            //   <meta name="referrer" content="same-origin">                                                          // 1185
            // It's important that we don't break apps, so we will accept null origins for now,                        // 1186
            // which of course completely defeats any CSRF protection. We should get the apps to                       // 1187
            // update or apply a whitelist soon.                                                                       // 1188
            if (origin === "null") {                                                                                   // 1189
              if (!proxy.wroteCsrfWarning) {                                                                           // 1190
                console.warn("Note: Observed null Origin header. App needs to be updated so that " + "we can apply automatic CSRF protection. grain ID:", proxy.grainId);
                proxy.wroteCsrfWarning = true;                                                                         // 1194
              }                                                                                                        // 1195
            } else {                                                                                                   // 1196
              throw new Meteor.Error(403, "Blocked illegal cross-origin request from: " + origin);                     // 1197
            }                                                                                                          // 1198
          }                                                                                                            // 1199
        } else if (referer) {                                                                                          // 1200
          // Mark that we've seed a Referer header on this host, which indicates that the user's                       // 1201
          // browser is not suppressing them.                                                                          // 1202
          proxy.seenReferer = true; // Deny requests coming from an external referer, since there is no legitimate use case
          // for these. Note of course that an attacker can trivially suppress the referer header                      // 1206
          // in practice, so this check alone does not give us any security benefit if we accept                       // 1207
          // requests that lack a referer.                                                                             // 1208
                                                                                                                       //
          if (referer != expectedOrigin && !referer.startsWith(expectedOrigin + "/") && referer != mainUrl && !referer.startsWith(mainUrl + "/")) {
            throw new Meteor.Error(403, "Blocked illegal cross-origin referral from: " + referer);                     // 1211
          }                                                                                                            // 1212
        } else {                                                                                                       // 1213
          // We saw neither an Origin nor a Referer header.                                                            // 1214
          if (req.method !== "GET" && req.method !== "HEAD") {                                                         // 1215
            // This is possibly a side-effecting request, meaning we should worry about CSRF.                          // 1216
            //                                                                                                         // 1217
            // Note that we'll never get to this branch for Chrome or Safari since they would have                     // 1218
            // sent an Origin header.                                                                                  // 1219
            if (!proxy.wroteCsrfWarning) {                                                                             // 1221
              console.warn("Note: Observed session for which auto-CSRF-protection couldn't be applied.", "seenReferer:", !!proxy.seenReferer, "method:", req.method, "path:", req.url, "UA:", req.headers["user-agent"]);
              proxy.wroteCsrfWarning = true;                                                                           // 1226
            }                                                                                                          // 1227
          }                                                                                                            // 1228
        }                                                                                                              // 1229
                                                                                                                       //
        proxy.requestHandler(req, res);                                                                                // 1231
        return true;                                                                                                   // 1232
      } else {                                                                                                         // 1233
        return false;                                                                                                  // 1234
      }                                                                                                                // 1235
    });                                                                                                                // 1236
  }                                                                                                                    // 1237
}; // =======================================================================================                          // 1238
// Proxy class                                                                                                         // 1241
//                                                                                                                     // 1242
// Connects to a grain and exports it on a wildcard host.                                                              // 1243
//                                                                                                                     // 1244
                                                                                                                       //
                                                                                                                       //
var Proxy = function () {                                                                                              //
  function Proxy(grain, sessionId, hostId, tabId, accountId, identityId, isApi, parentOrigin, supervisor) {            // 1247
    var _this3 = this;                                                                                                 // 1248
                                                                                                                       //
    (0, _classCallCheck3.default)(this, Proxy);                                                                        // 1248
    // `grain` is an entry in the `Grains` collection.                                                                 // 1249
    this.grainId = grain._id;                                                                                          // 1250
    this.ownerId = grain.userId;                                                                                       // 1251
    this.accountId = accountId;                                                                                        // 1252
    this.identityId = identityId;                                                                                      // 1253
    this.supervisor = supervisor; // note: optional parameter; we can reconnect                                        // 1254
                                                                                                                       //
    this.sessionId = sessionId;                                                                                        // 1255
    this.tabId = tabId;                                                                                                // 1256
    this.isApi = isApi;                                                                                                // 1257
    this.hasLoaded = false;                                                                                            // 1258
    this.websockets = {};                                                                                              // 1259
    this.websocketCounter = 0; // Used for generating unique socket IDs.                                               // 1260
                                                                                                                       //
    this.streamingRequests = {};                                                                                       // 1261
    this.streamingRequestCounter = 0; // Used for generating unique request IDs.                                       // 1262
                                                                                                                       //
    this.streamingResponses = {};                                                                                      // 1263
    this.streamingResponseCounter = 0; // Used for generating unique response IDs.                                     // 1264
                                                                                                                       //
    this.parentOrigin = parentOrigin || process.env.ROOT_URL;                                                          // 1265
                                                                                                                       //
    if (sessionId) {                                                                                                   // 1267
      if (!hostId) throw new Error("sessionId must come with hostId");                                                 // 1268
      if (isApi) throw new Error("API proxy shouldn't have sessionId");                                                // 1269
      this.hostId = hostId;                                                                                            // 1270
    } else {                                                                                                           // 1271
      if (!isApi) throw new Error("non-API proxy requires sessionId");                                                 // 1272
      if (hostId) throw new Error("API proxy sholudn't have hostId");                                                  // 1273
    }                                                                                                                  // 1274
                                                                                                                       //
    if (this.accountId) {                                                                                              // 1276
      var user = Meteor.users.findOne({                                                                                // 1277
        _id: this.accountId                                                                                            // 1277
      });                                                                                                              // 1277
                                                                                                                       //
      if (!user) {                                                                                                     // 1278
        throw new Error("user not found: " + this.accountId);                                                          // 1279
      }                                                                                                                // 1280
                                                                                                                       //
      SandstormDb.fillInPictureUrl(user); // The identity cap becomes invalid if the user no longer has access to the grain.
                                                                                                                       //
      var idCapRequirement = {                                                                                         // 1285
        permissionsHeld: {                                                                                             // 1286
          accountId: user._id,                                                                                         // 1286
          grainId: this.grainId                                                                                        // 1286
        }                                                                                                              // 1286
      };                                                                                                               // 1285
      this.userInfo = {                                                                                                // 1289
        displayName: {                                                                                                 // 1290
          defaultText: user.profile.name                                                                               // 1290
        },                                                                                                             // 1290
        preferredHandle: user.profile.handle,                                                                          // 1291
        identityId: new Buffer(this.identityId, "hex"),                                                                // 1292
        identity: makeIdentity(user._id, [idCapRequirement]),                                                          // 1293
        pictureUrl: user.profile.pictureUrl                                                                            // 1294
      };                                                                                                               // 1289
      if (user.profile.pronoun) this.userInfo.pronouns = user.profile.pronoun;                                         // 1297
    } else {                                                                                                           // 1298
      this.userInfo = {                                                                                                // 1299
        displayName: {                                                                                                 // 1300
          defaultText: "Anonymous User"                                                                                // 1300
        },                                                                                                             // 1300
        preferredHandle: "anonymous"                                                                                   // 1301
      };                                                                                                               // 1299
    }                                                                                                                  // 1303
                                                                                                                       //
    this.requestHandler = function (request, response) {                                                               // 1305
      if (_this3.sessionId) {                                                                                          // 1306
        // Implement /_sandstorm-init for setting the session cookie.                                                  // 1307
        var url = Url.parse(request.url, true);                                                                        // 1308
                                                                                                                       //
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this3.sessionId) {                         // 1309
          _this3.doSessionInit(request, response, url.query.path);                                                     // 1310
                                                                                                                       //
          return;                                                                                                      // 1311
        }                                                                                                              // 1312
      }                                                                                                                // 1313
                                                                                                                       //
      Promise.resolve(undefined).then(function () {                                                                    // 1315
        var contentLength = request.headers["content-length"];                                                         // 1316
                                                                                                                       //
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.                    // 1319
          return _this3.handleRequestStreaming(request, response, contentLength, 0);                                   // 1320
        } else {                                                                                                       // 1321
          return readAll(request).then(function (data) {                                                               // 1322
            return _this3.handleRequest(request, data, response, 0);                                                   // 1323
          });                                                                                                          // 1324
        }                                                                                                              // 1325
      }).catch(function (err) {                                                                                        // 1326
        _this3.setHasLoaded();                                                                                         // 1327
                                                                                                                       //
        var body = err.stack;                                                                                          // 1329
                                                                                                                       //
        if (err.cppFile) {                                                                                             // 1330
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");                                          // 1331
        }                                                                                                              // 1332
                                                                                                                       //
        if (err.kjType) {                                                                                              // 1334
          body += "\ntype: " + err.kjType;                                                                             // 1335
        }                                                                                                              // 1336
                                                                                                                       //
        if (response.headersSent) {                                                                                    // 1338
          // Unfortunately, it's too late to tell the client what happened.                                            // 1339
          console.error("HTTP request failed after response already sent:", body);                                     // 1340
          response.end();                                                                                              // 1341
        } else {                                                                                                       // 1342
          if (err instanceof Meteor.Error) {                                                                           // 1343
            response.writeHead(err.error, err.reason, {                                                                // 1344
              "Content-Type": "text/plain"                                                                             // 1344
            });                                                                                                        // 1344
          } else {                                                                                                     // 1345
            response.writeHead(500, "Internal Server Error", {                                                         // 1346
              "Content-Type": "text/plain"                                                                             // 1346
            });                                                                                                        // 1346
          }                                                                                                            // 1347
                                                                                                                       //
          response.end(body);                                                                                          // 1349
        }                                                                                                              // 1350
      });                                                                                                              // 1351
    };                                                                                                                 // 1352
                                                                                                                       //
    this.upgradeHandler = function (request, socket, head) {                                                           // 1354
      _this3.handleWebSocket(request, socket, head, 0).catch(function (err) {                                          // 1355
        console.error("WebSocket setup failed:", err.stack); // TODO(cleanup):  Manually send back a 500 response?     // 1356
                                                                                                                       //
        socket.destroy();                                                                                              // 1358
      });                                                                                                              // 1359
    };                                                                                                                 // 1360
  }                                                                                                                    // 1361
                                                                                                                       //
  Proxy.prototype.close = function () {                                                                                //
    function close() {                                                                                                 //
      for (var streamingResponseIdx in meteorBabelHelpers.sanitizeForInObject(this.streamingResponses)) {              // 1364
        this.streamingResponses[streamingResponseIdx].abort();                                                         // 1365
      }                                                                                                                // 1366
                                                                                                                       //
      this.streamingResponses = {};                                                                                    // 1368
                                                                                                                       //
      for (var streamingRequestIdx in meteorBabelHelpers.sanitizeForInObject(this.streamingRequests)) {                // 1370
        this.streamingRequests[streamingRequestIdx].destroy();                                                         // 1371
      }                                                                                                                // 1372
                                                                                                                       //
      this.streamingRequests = {};                                                                                     // 1374
                                                                                                                       //
      for (var socketIdx in meteorBabelHelpers.sanitizeForInObject(this.websockets)) {                                 // 1376
        this.websockets[socketIdx].destroy();                                                                          // 1377
      }                                                                                                                // 1378
                                                                                                                       //
      ;                                                                                                                // 1378
      this.websockets = {};                                                                                            // 1380
                                                                                                                       //
      if (this.session) {                                                                                              // 1382
        this.session.close();                                                                                          // 1383
        delete this.session;                                                                                           // 1384
      }                                                                                                                // 1385
                                                                                                                       //
      if (this.uiView) {                                                                                               // 1387
        this.uiView.close();                                                                                           // 1388
        delete this.uiView;                                                                                            // 1389
      }                                                                                                                // 1390
                                                                                                                       //
      if (this.supervisor) {                                                                                           // 1392
        this.supervisor.close();                                                                                       // 1393
        delete this.supervisor;                                                                                        // 1394
      }                                                                                                                // 1395
                                                                                                                       //
      if (this.permissionsObserver) {                                                                                  // 1397
        this.permissionsObserver.stop();                                                                               // 1398
        delete this.permissionsObserver;                                                                               // 1399
      }                                                                                                                // 1400
    }                                                                                                                  // 1401
                                                                                                                       //
    return close;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.getConnection = function () {                                                                        //
    function getConnection() {                                                                                         //
      if (!this.supervisor) {                                                                                          // 1404
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;                         // 1405
        this.uiView = null;                                                                                            // 1406
      }                                                                                                                // 1407
                                                                                                                       //
      if (!this.uiView) {                                                                                              // 1409
        this.uiView = this.supervisor.getMainView().view;                                                              // 1410
      }                                                                                                                // 1411
    }                                                                                                                  // 1412
                                                                                                                       //
    return getConnection;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewWebSession = function () {                                                                   //
    function _callNewWebSession(request, userInfo) {                                                                   //
      var params = Capnp.serialize(WebSession.Params, {                                                                // 1415
        basePath: PROTOCOL + "//" + request.headers.host,                                                              // 1416
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",               // 1417
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                                                                             // 1421
        }) : ["en-US", "en"]                                                                                           // 1421
      });                                                                                                              // 1415
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.accountId, this.tabId), WebSession.typeId, params, new Buffer(this.tabId, "hex")).session;
    }                                                                                                                  // 1427
                                                                                                                       //
    return _callNewWebSession;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewApiSession = function () {                                                                   //
    function _callNewApiSession(request, userInfo) {                                                                   //
      var _this4 = this;                                                                                               // 1429
                                                                                                                       //
      var serializedParams = this.apiSessionParams;                                                                    // 1430
                                                                                                                       //
      if (!serializedParams) {                                                                                         // 1431
        throw new Meteor.Error(500, "Should have already computed apiSessionParams.");                                 // 1432
      } // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon                 // 1433
      // calling newSession with an ApiSession._id.                                                                    // 1436
      // Eventually we'll remove this logic once we're sure apps have updated.                                         // 1437
                                                                                                                       //
                                                                                                                       //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.accountId, this.tabId), ApiSession.typeId, serializedParams, new Buffer(this.tabId, "hex")).then(function (session) {
        return session.session;                                                                                        // 1442
      }, function (err) {                                                                                              // 1443
        return _this4._callNewWebSession(request, userInfo);                                                           // 1444
      });                                                                                                              // 1445
    }                                                                                                                  // 1446
                                                                                                                       //
    return _callNewApiSession;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewRequestSession = function () {                                                               //
    function _callNewRequestSession(request, userInfo) {                                                               //
      var params = Capnp.serialize(WebSession.Params, {                                                                // 1449
        basePath: PROTOCOL + "//" + request.headers.host,                                                              // 1450
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",               // 1451
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                                                                             // 1455
        }) : ["en-US", "en"]                                                                                           // 1455
      });                                                                                                              // 1449
      return this.uiView.newRequestSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.accountId, this.tabId), WebSession.typeId, params, this.powerboxRequest.descriptors, new Buffer(this.tabId, "hex")).session;
    }                                                                                                                  // 1462
                                                                                                                       //
    return _callNewRequestSession;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewSession = function () {                                                                      //
    function _callNewSession(request, viewInfo) {                                                                      //
      var _this5 = this;                                                                                               // 1464
                                                                                                                       //
      var userInfo = _.clone(this.userInfo);                                                                           // 1465
                                                                                                                       //
      var promise = inMeteor(function () {                                                                             // 1466
        var vertex = void 0;                                                                                           // 1467
                                                                                                                       //
        if (_this5.apiToken) {                                                                                         // 1468
          vertex = {                                                                                                   // 1469
            token: _this5.apiToken                                                                                     // 1469
          };                                                                                                           // 1469
        } else {                                                                                                       // 1470
          // (this.accountId might be null; this is fine)                                                              // 1471
          vertex = {                                                                                                   // 1472
            grain: {                                                                                                   // 1472
              _id: _this5.grainId,                                                                                     // 1472
              accountId: _this5.accountId                                                                              // 1472
            }                                                                                                          // 1472
          };                                                                                                           // 1472
        }                                                                                                              // 1473
                                                                                                                       //
        var onInvalidated = function () {                                                                              // 1475
          Sessions.remove({                                                                                            // 1476
            _id: _this5.sessionId                                                                                      // 1476
          });                                                                                                          // 1476
        };                                                                                                             // 1477
                                                                                                                       //
        if (!_this5.sessionId) {                                                                                       // 1479
          onInvalidated = function () {                                                                                // 1480
            apiSessionProxies.removeProxiesOfToken(_this5.apiToken._id);                                               // 1481
          };                                                                                                           // 1482
        }                                                                                                              // 1483
                                                                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo, onInvalidated);            // 1485
        if (_this5.permissionsObserver) _this5.permissionsObserver.stop();                                             // 1488
        _this5.permissionsObserver = permissions.observeHandle;                                                        // 1489
                                                                                                                       //
        if (!permissions.permissions) {                                                                                // 1491
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                   // 1492
        }                                                                                                              // 1493
                                                                                                                       //
        Sessions.update({                                                                                              // 1495
          _id: _this5.sessionId                                                                                        // 1496
        }, {                                                                                                           // 1495
          $set: {                                                                                                      // 1498
            viewInfo: viewInfo,                                                                                        // 1499
            permissions: permissions.permissions                                                                       // 1500
          }                                                                                                            // 1498
        });                                                                                                            // 1497
        return permissions.permissions;                                                                                // 1504
      });                                                                                                              // 1505
      return promise.then(function (permissions) {                                                                     // 1507
        userInfo.permissions = permissions;                                                                            // 1508
        var numBytes = Math.ceil(permissions.length / 8);                                                              // 1510
        var buf = new Buffer(numBytes);                                                                                // 1511
                                                                                                                       //
        for (var ii = 0; ii < numBytes; ++ii) {                                                                        // 1512
          buf[ii] = 0;                                                                                                 // 1513
        }                                                                                                              // 1514
                                                                                                                       //
        for (var _ii = 0; _ii < permissions.length; ++_ii) {                                                           // 1516
          var byteNum = Math.floor(_ii / 8);                                                                           // 1517
          var bitNum = _ii % 8;                                                                                        // 1518
                                                                                                                       //
          if (permissions[_ii]) {                                                                                      // 1519
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                                                                 // 1520
          }                                                                                                            // 1521
        }                                                                                                              // 1522
                                                                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                                                                      // 1524
                                                                                                                       //
        if (_this5.isApi) {                                                                                            // 1526
          return _this5._callNewApiSession(request, userInfo);                                                         // 1527
        } else if (_this5.powerboxRequest) {                                                                           // 1528
          return _this5._callNewRequestSession(request, userInfo);                                                     // 1529
        } else {                                                                                                       // 1530
          return _this5._callNewWebSession(request, userInfo);                                                         // 1531
        }                                                                                                              // 1532
      });                                                                                                              // 1533
    }                                                                                                                  // 1534
                                                                                                                       //
    return _callNewSession;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.getSession = function () {                                                                           //
    function getSession(request) {                                                                                     //
      var _this6 = this;                                                                                               // 1536
                                                                                                                       //
      if (!this.session) {                                                                                             // 1537
        this.getConnection(); // make sure we're connected                                                             // 1538
                                                                                                                       //
        var promise = this.uiView.getViewInfo().then(function (viewInfo) {                                             // 1539
          return inMeteor(function () {                                                                                // 1540
            // For now, we don't allow grains to set `appTitle` or `grainIcon`.                                        // 1541
            var cachedViewInfo = _.omit(viewInfo, "appTitle", "grainIcon");                                            // 1542
                                                                                                                       //
            Grains.update(_this6.grainId, {                                                                            // 1543
              $set: {                                                                                                  // 1543
                cachedViewInfo: cachedViewInfo                                                                         // 1543
              }                                                                                                        // 1543
            });                                                                                                        // 1543
          }).then(function () {                                                                                        // 1544
            return _this6._callNewSession(request, viewInfo);                                                          // 1545
          });                                                                                                          // 1546
        }, function (error) {                                                                                          // 1547
          if (error.kjType === "failed" || error.kjType === "unimplemented") {                                         // 1548
            // Method not implemented.                                                                                 // 1549
            // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built                   // 1550
            //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have                   // 1551
            //   to accept 'failed' here at least until the next API bump.                                             // 1552
            return _this6._callNewSession(request, {});                                                                // 1553
          } else {                                                                                                     // 1554
            return Promise.reject(error);                                                                              // 1555
          }                                                                                                            // 1556
        });                                                                                                            // 1557
        this.session = new Capnp.Capability(promise, WebSession);                                                      // 1558
      }                                                                                                                // 1559
                                                                                                                       //
      return this.session;                                                                                             // 1561
    }                                                                                                                  // 1562
                                                                                                                       //
    return getSession;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.keepAlive = function () {                                                                            //
    function keepAlive() {                                                                                             //
      this.getConnection();                                                                                            // 1565
      return this.supervisor.keepAlive();                                                                              // 1566
    }                                                                                                                  // 1567
                                                                                                                       //
    return keepAlive;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.resetConnection = function () {                                                                      //
    function resetConnection() {                                                                                       //
      if (this.session) {                                                                                              // 1570
        this.session.close();                                                                                          // 1571
        delete this.session;                                                                                           // 1572
      }                                                                                                                // 1573
                                                                                                                       //
      if (this.uiView) {                                                                                               // 1575
        this.uiView.close();                                                                                           // 1576
        delete this.uiView;                                                                                            // 1577
      }                                                                                                                // 1578
                                                                                                                       //
      if (this.supervisor) {                                                                                           // 1580
        this.supervisor.close();                                                                                       // 1581
        delete this.supervisor;                                                                                        // 1582
      }                                                                                                                // 1583
                                                                                                                       //
      if (this.permissionsObserver) {                                                                                  // 1585
        this.permissionsObserver.stop();                                                                               // 1586
        delete this.permissionsObserver;                                                                               // 1587
      }                                                                                                                // 1588
    }                                                                                                                  // 1589
                                                                                                                       //
    return resetConnection;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.maybeRetryAfterError = function () {                                                                 //
    function maybeRetryAfterError(error, retryCount) {                                                                 //
      var _this7 = this;                                                                                               // 1591
                                                                                                                       //
      // If the error may be caused by the grain dying or a network failure, try to restart it,                        // 1592
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.                          // 1593
      // `retryCount` should be incremented for every successful retry as part of the same request;                    // 1594
      // we only want to retry once.                                                                                   // 1595
      if (shouldRestartGrain(error, retryCount)) {                                                                     // 1596
        this.resetConnection();                                                                                        // 1597
        return inMeteor(function () {                                                                                  // 1598
          _this7.supervisor = globalBackend.continueGrain(_this7.grainId).supervisor;                                  // 1599
        });                                                                                                            // 1600
      } else {                                                                                                         // 1601
        throw error;                                                                                                   // 1602
      }                                                                                                                // 1603
    }                                                                                                                  // 1604
                                                                                                                       //
    return maybeRetryAfterError;                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.doSessionInit = function () {                                                                        //
    function doSessionInit(request, response, requestPath) {                                                           //
      // jscs:disable disallowQuotedKeysInObjects                                                                      // 1607
      var path = requestPath || "/"; // Check that the path is relative (ie. starts with a /).                         // 1608
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative                       // 1611
                                                                                                                       //
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {                                         // 1612
        response.writeHead(400, "Invalid path supplied", {                                                             // 1613
          "Content-Type": "text/plain"                                                                                 // 1613
        });                                                                                                            // 1613
        response.end("Invalid path supplied.");                                                                        // 1614
        return;                                                                                                        // 1615
      } // Set the session ID.                                                                                         // 1616
                                                                                                                       //
                                                                                                                       //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));   // 1619
      response.setHeader("Cache-Control", "no-cache, private"); // Redirect to the app's root URL.                     // 1621
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially                // 1624
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67                     // 1625
                                                                                                                       //
      response.writeHead(303, "See Other", {                                                                           // 1626
        "Location": encodeURI(path)                                                                                    // 1626
      });                                                                                                              // 1626
      response.end();                                                                                                  // 1627
    }                                                                                                                  // 1628
                                                                                                                       //
    return doSessionInit;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.makeContext = function () {                                                                          //
    function makeContext(request, response, callback) {                                                                //
      // Parses the cookies from the request, checks that the session ID is present and valid, then                    // 1631
      // synchronously calls callback() passing it the request context object which contains the                       // 1632
      // other cookies. Throws an exception if the session ID is missing or invalid. The callback                      // 1633
      // must immediately use the context in an RPC. Once the callback returns, the context object                     // 1634
      // will be torn down.                                                                                            // 1635
      //                                                                                                               // 1636
      // `makeContext()` returns the result of `callback()`.                                                           // 1637
      var context = {};                                                                                                // 1639
                                                                                                                       //
      if (this.hostId) {                                                                                               // 1641
        var parseResult = parseCookies(request);                                                                       // 1642
                                                                                                                       //
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {                                      // 1643
          throw new Meteor.Error(403, "Unauthorized");                                                                 // 1644
        }                                                                                                              // 1645
                                                                                                                       //
        if (parseResult.cookies.length > 0) {                                                                          // 1647
          context.cookies = parseResult.cookies;                                                                       // 1648
        }                                                                                                              // 1649
      } else {// jscs:disable disallowEmptyBlocks                                                                      // 1650
        // This is an API request. Cookies are not supported.                                                          // 1652
      }                                                                                                                // 1653
                                                                                                                       //
      context.accept = parseAcceptHeader(request);                                                                     // 1655
      context.acceptEncoding = parseAcceptEncodingHeader(request);                                                     // 1657
      context.eTagPrecondition = parsePreconditionHeader(request);                                                     // 1659
      context.additionalHeaders = [];                                                                                  // 1661
                                                                                                                       //
      for (var headerName in meteorBabelHelpers.sanitizeForInObject(request.headers)) {                                // 1662
        if (REQUEST_HEADER_WHITELIST.matches(headerName)) {                                                            // 1663
          context.additionalHeaders.push({                                                                             // 1664
            name: headerName,                                                                                          // 1665
            value: request.headers[headerName]                                                                         // 1666
          });                                                                                                          // 1664
        }                                                                                                              // 1668
      }                                                                                                                // 1669
                                                                                                                       //
      if (response) {                                                                                                  // 1671
        var promise = new Promise(function (resolve, reject) {                                                         // 1672
          response.resolveResponseStream = resolve;                                                                    // 1673
          response.rejectResponseStream = reject;                                                                      // 1674
        });                                                                                                            // 1675
        context.responseStream = new Capnp.Capability(promise, ByteStream);                                            // 1677
                                                                                                                       //
        try {                                                                                                          // 1679
          return callback(context);                                                                                    // 1680
        } finally {                                                                                                    // 1681
          context.responseStream.close();                                                                              // 1682
        }                                                                                                              // 1683
      } else {                                                                                                         // 1684
        // WebSocket -- no response object.                                                                            // 1685
        return callback(context);                                                                                      // 1686
      }                                                                                                                // 1687
    }                                                                                                                  // 1688
                                                                                                                       //
    return makeContext;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.translateResponse = function () {                                                                    //
    function translateResponse(rpcResponse, response, request) {                                                       //
      var _this8 = this;                                                                                               // 1690
                                                                                                                       //
      if (this.hostId) {                                                                                               // 1691
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {                                             // 1692
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));                           // 1693
        }                                                                                                              // 1694
                                                                                                                       //
        var mainUrl = this.parentOrigin;                                                                               // 1696
        var grainHost = PROTOCOL + "//" + request.headers.host;                                                        // 1697
        response.setHeader("Content-Security-Policy", "frame-ancestors " + mainUrl + " " + grainHost);                 // 1698
        response.setHeader("X-Frame-Options", "ALLOW-FROM " + mainUrl); // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.                                        // 1702
        // (2) Prevents the app from navigating the parent frame.                                                      // 1703
        // (3) Prevents the app from opening popups.                                                                   // 1704
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup                       // 1705
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).                      // 1706
      } else {                                                                                                         // 1707
        // This is an API request. Cookies are not supported.                                                          // 1708
        // We need to make sure caches know that different bearer tokens get totally different results.                // 1710
        response.setHeader("Vary", "Authorization"); // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
                                                                                                                       //
        response.setHeader("Access-Control-Allow-Origin", "*"); // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
        // in a browser context. This policy should thoroughly neuter it.                                              // 1717
                                                                                                                       //
        response.setHeader("Content-Security-Policy", "default-src 'none'; sandbox");                                  // 1718
      }                                                                                                                // 1719
                                                                                                                       //
      if (rpcResponse.additionalHeaders && rpcResponse.additionalHeaders.length > 0) {                                 // 1721
        // Add any additional prefixed/whitelisted headers to the response.                                            // 1722
        // We cannot trust that the headers in the RPC response are actually                                           // 1723
        // allowed, so we check if they are prefixed or whitelisted.                                                   // 1724
        rpcResponse.additionalHeaders.forEach(function (header) {                                                      // 1725
          if (RESPONSE_HEADER_WHITELIST.matches(header.name)) {                                                        // 1726
            // If header is prefixed with appHeaderPrefix,                                                             // 1727
            // or if the header name is in whitelist, set it.                                                          // 1728
            response.setHeader(header.name, header.value);                                                             // 1729
          }                                                                                                            // 1730
        });                                                                                                            // 1731
      } // On first response, update the session to have hasLoaded=true                                                // 1732
                                                                                                                       //
                                                                                                                       //
      this.setHasLoaded(); // TODO(security): Set X-Content-Type-Options: nosniff?                                     // 1735
                                                                                                                       //
      if ("content" in rpcResponse) {                                                                                  // 1739
        var content = rpcResponse.content;                                                                             // 1740
        var code = successCodes[content.statusCode];                                                                   // 1741
                                                                                                                       //
        if (!code) {                                                                                                   // 1742
          throw new Error("Unknown status code: ", content.statusCode);                                                // 1743
        }                                                                                                              // 1744
                                                                                                                       //
        if (content.mimeType) {                                                                                        // 1746
          response.setHeader("Content-Type", content.mimeType);                                                        // 1747
        }                                                                                                              // 1748
                                                                                                                       //
        if (content.encoding) {                                                                                        // 1750
          response.setHeader("Content-Encoding", content.encoding);                                                    // 1751
        }                                                                                                              // 1752
                                                                                                                       //
        if (content.language) {                                                                                        // 1754
          response.setHeader("Content-Language", content.language);                                                    // 1755
        }                                                                                                              // 1756
                                                                                                                       //
        if (content.eTag) {                                                                                            // 1758
          response.setHeader("ETag", composeETag(content.eTag));                                                       // 1759
        }                                                                                                              // 1760
                                                                                                                       //
        if ("disposition" in content && "download" in content.disposition) {                                           // 1762
          response.setHeader("Content-Disposition", 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                                                                              // 1765
                                                                                                                       //
        if ("stream" in content.body) {                                                                                // 1767
          if (request.method === "HEAD") {                                                                             // 1768
            content.body.stream.close();                                                                               // 1769
            response.rejectResponseStream(new Error("HEAD request; content doesn't matter."));                         // 1770
          } else if (response.socket.destroyed) {                                                                      // 1771
            // The client has already closed the stream. No additional "close" event will be                           // 1772
            // emitted. Annoyingly, we cannot rely on `response.write()` throwing an exception                         // 1773
            // in this case, so if we don't handle it specially we could end up with a                                 // 1774
            // `ResponseStream` that gets stuck waiting for a "drain" event that never arrives.                        // 1775
            content.body.stream.close();                                                                               // 1776
            response.rejectResponseStream(new Error("client disconnected"));                                           // 1777
          } else {                                                                                                     // 1778
            var idx = this.streamingResponseCounter;                                                                   // 1779
            this.streamingResponseCounter += 1;                                                                        // 1780
            var streamHandle = content.body.stream;                                                                    // 1781
            var rawResponseStream = new ResponseStream(response, code.id, code.title, streamHandle, function () {      // 1782
              delete _this8.streamingResponses[idx];                                                                   // 1784
            });                                                                                                        // 1784
            this.streamingResponses[idx] = rawResponseStream;                                                          // 1785
            var responseStream = new Capnp.Capability(rawResponseStream, ByteStream);                                  // 1786
            response.resolveResponseStream(responseStream);                                                            // 1787
            return {                                                                                                   // 1788
              streamHandle: responseStream                                                                             // 1788
            };                                                                                                         // 1788
          }                                                                                                            // 1789
        } else {                                                                                                       // 1790
          response.rejectResponseStream(new Error("Response content body was not a stream."));                         // 1791
                                                                                                                       //
          if ("bytes" in content.body) {                                                                               // 1794
            response.setHeader("Content-Length", content.body.bytes.length);                                           // 1795
          } else {                                                                                                     // 1796
            throw new Error("Unknown content body type.");                                                             // 1797
          }                                                                                                            // 1798
        }                                                                                                              // 1799
                                                                                                                       //
        response.writeHead(code.id, code.title);                                                                       // 1801
                                                                                                                       //
        if ("bytes" in content.body && request.method !== "HEAD") {                                                    // 1803
          response.write(content.body.bytes);                                                                          // 1804
        }                                                                                                              // 1805
                                                                                                                       //
        response.end();                                                                                                // 1807
      } else if ("noContent" in rpcResponse) {                                                                         // 1808
        var noContent = rpcResponse.noContent;                                                                         // 1809
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];                                      // 1810
                                                                                                                       //
        if (noContent.eTag) {                                                                                          // 1811
          response.setHeader("ETag", composeETag(noContent.eTag));                                                     // 1812
        }                                                                                                              // 1813
                                                                                                                       //
        response.writeHead(noContentCode.id, noContentCode.title);                                                     // 1815
        response.end();                                                                                                // 1816
      } else if ("preconditionFailed" in rpcResponse) {                                                                // 1817
        var preconditionFailed = rpcResponse.preconditionFailed;                                                       // 1818
                                                                                                                       //
        if (request.method === "GET" && "if-none-match" in request.headers) {                                          // 1819
          if (preconditionFailed.matchingETag) {                                                                       // 1820
            response.setHeader("ETag", composeETag(preconditionFailed.matchingETag));                                  // 1821
          }                                                                                                            // 1822
                                                                                                                       //
          response.writeHead(304, "Not Modified");                                                                     // 1824
        } else {                                                                                                       // 1825
          response.writeHead(412, "Precondition Failed");                                                              // 1826
        }                                                                                                              // 1827
                                                                                                                       //
        response.end();                                                                                                // 1829
      } else if ("redirect" in rpcResponse) {                                                                          // 1830
        var redirect = rpcResponse.redirect;                                                                           // 1831
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];                             // 1832
        response.writeHead(redirectCode.id, redirectCode.title, {                                                      // 1833
          "Location": encodeURI(redirect.location)                                                                     // 1834
        });                                                                                                            // 1833
        response.end();                                                                                                // 1836
      } else if ("clientError" in rpcResponse) {                                                                       // 1837
        var clientError = rpcResponse.clientError;                                                                     // 1838
        var errorCode = errorCodes[clientError.statusCode];                                                            // 1839
                                                                                                                       //
        if (!errorCode) {                                                                                              // 1840
          throw new Error("Unknown status code: ", clientError.statusCode);                                            // 1841
        }                                                                                                              // 1842
                                                                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {                                                            // 1844
          "Content-Type": "text/html"                                                                                  // 1845
        });                                                                                                            // 1844
                                                                                                                       //
        if (request.method !== "HEAD") {                                                                               // 1848
          if (clientError.descriptionHtml) {                                                                           // 1849
            response.write(clientError.descriptionHtml);                                                               // 1850
          } else {                                                                                                     // 1851
            // TODO(someday):  Better default error page.                                                              // 1852
            response.write("<html><body><h1>" + errorCode.id + ": " + errorCode.title + "</h1></body></html>");        // 1853
          }                                                                                                            // 1855
        }                                                                                                              // 1856
                                                                                                                       //
        response.end();                                                                                                // 1858
      } else if ("serverError" in rpcResponse) {                                                                       // 1859
        response.writeHead(500, "Internal Server Error", {                                                             // 1860
          "Content-Type": "text/html"                                                                                  // 1861
        });                                                                                                            // 1860
                                                                                                                       //
        if (request.method !== "HEAD") {                                                                               // 1864
          if (rpcResponse.serverError.descriptionHtml) {                                                               // 1865
            response.write(rpcResponse.serverError.descriptionHtml);                                                   // 1866
          } else {                                                                                                     // 1867
            // TODO(someday):  Better default error page.                                                              // 1868
            response.write("<html><body><h1>500: Internal Server Error</h1></body></html>");                           // 1869
          }                                                                                                            // 1870
        }                                                                                                              // 1871
                                                                                                                       //
        response.end();                                                                                                // 1873
      } else {                                                                                                         // 1874
        throw new Error("Unknown HTTP response type:\n" + JSON.stringify(rpcResponse));                                // 1875
      }                                                                                                                // 1876
                                                                                                                       //
      return {};                                                                                                       // 1878
    }                                                                                                                  // 1879
                                                                                                                       //
    return translateResponse;                                                                                          //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleRequest = function () {                                                                        //
    function handleRequest(request, data, response, retryCount) {                                                      //
      var _this9 = this;                                                                                               // 1881
                                                                                                                       //
      return Promise.resolve(undefined).then(function () {                                                             // 1882
        return _this9.makeContext(request, response, function (context) {                                              // 1883
          // jscs:disable requireDotNotation                                                                           // 1884
          // Send the RPC.                                                                                             // 1885
          var path = request.url.slice(1); // remove leading '/'                                                       // 1886
                                                                                                                       //
          var session = _this9.getSession(request);                                                                    // 1887
                                                                                                                       //
          var requestContent = function () {                                                                           // 1889
            return {                                                                                                   // 1890
              content: data,                                                                                           // 1891
              encoding: request.headers["content-encoding"],                                                           // 1892
              mimeType: request.headers["content-type"]                                                                // 1893
            };                                                                                                         // 1890
          };                                                                                                           // 1895
                                                                                                                       //
          var xmlContent = function () {                                                                               // 1897
            var type = request.headers["content-type"] || "application/xml;charset=utf-8";                             // 1898
            var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);                                            // 1899
                                                                                                                       //
            if (!match) {                                                                                              // 1900
              response.writeHead(415, "Unsupported media type.", {                                                     // 1901
                "Content-Type": "text/plain"                                                                           // 1902
              });                                                                                                      // 1901
              response.end("expected XML request body");                                                               // 1904
              throw new Error("expected XML request body");                                                            // 1905
            }                                                                                                          // 1906
                                                                                                                       //
            var charset = match[2] || "ISO-8859-1";                                                                    // 1908
            var encoding = request.headers["content-encoding"];                                                        // 1910
                                                                                                                       //
            if (encoding && encoding !== "identity") {                                                                 // 1911
              if (encoding !== "gzip") throw new Error("unknown Content-Encoding: " + encoding);                       // 1912
              data = gunzipSync(data);                                                                                 // 1913
            }                                                                                                          // 1914
                                                                                                                       //
            return data.toString(charset.toLowerCase() === "utf-8" ? "utf8" : "binary");                               // 1916
          };                                                                                                           // 1917
                                                                                                                       //
          var propfindDepth = function () {                                                                            // 1919
            var depth = request.headers["depth"];                                                                      // 1920
            return depth === "0" ? "zero" : depth === "1" ? "one" : "infinity";                                        // 1921
          };                                                                                                           // 1924
                                                                                                                       //
          var shallow = function () {                                                                                  // 1926
            return request.headers["depth"] === "0";                                                                   // 1927
          };                                                                                                           // 1928
                                                                                                                       //
          var noOverwrite = function () {                                                                              // 1930
            return (request.headers["overwrite"] || "").toLowerCase() === "f";                                         // 1931
          };                                                                                                           // 1932
                                                                                                                       //
          var destination = function () {                                                                              // 1934
            var result = request.headers["destination"];                                                               // 1935
            if (!result) throw new Error("missing destination");                                                       // 1936
            return Url.parse(result).path.slice(1); // remove leading '/'                                              // 1937
          };                                                                                                           // 1938
                                                                                                                       //
          if (request.method === "GET" || request.method === "HEAD") {                                                 // 1940
            return session.get(path, context, request.method === "HEAD");                                              // 1941
          } else if (request.method === "POST") {                                                                      // 1942
            return session.post(path, requestContent(), context);                                                      // 1943
          } else if (request.method === "PUT") {                                                                       // 1944
            return session.put(path, requestContent(), context);                                                       // 1945
          } else if (request.method === "PATCH") {                                                                     // 1946
            return session.patch(path, requestContent(), context);                                                     // 1947
          } else if (request.method === "DELETE") {                                                                    // 1948
            return session.delete(path, context);                                                                      // 1949
          } else if (request.method === "PROPFIND") {                                                                  // 1950
            return session.propfind(path, xmlContent(), propfindDepth(), context);                                     // 1951
          } else if (request.method === "PROPPATCH") {                                                                 // 1952
            return session.proppatch(path, xmlContent(), context);                                                     // 1953
          } else if (request.method === "MKCOL") {                                                                     // 1954
            return session.mkcol(path, requestContent(), context);                                                     // 1955
          } else if (request.method === "COPY") {                                                                      // 1956
            return session.copy(path, destination(), noOverwrite(), shallow(), context);                               // 1957
          } else if (request.method === "MOVE") {                                                                      // 1958
            return session.move(path, destination(), noOverwrite(), context);                                          // 1959
          } else if (request.method === "LOCK") {                                                                      // 1960
            return session.lock(path, xmlContent(), shallow(), context);                                               // 1961
          } else if (request.method === "UNLOCK") {                                                                    // 1962
            return session.unlock(path, request.headers["lock-token"], context);                                       // 1963
          } else if (request.method === "ACL") {                                                                       // 1964
            return session.acl(path, xmlContent(), context);                                                           // 1965
          } else if (request.method === "REPORT") {                                                                    // 1966
            return session.report(path, requestContent(), context);                                                    // 1967
          } else if (request.method === "OPTIONS") {                                                                   // 1968
            return session.options(path, context).then(function (options) {                                            // 1969
              var dav = [];                                                                                            // 1970
              if (options.davClass1) dav.push("1");                                                                    // 1971
              if (options.davClass2) dav.push("2");                                                                    // 1972
              if (options.davClass3) dav.push("3");                                                                    // 1973
                                                                                                                       //
              if (options.davExtensions) {                                                                             // 1974
                options.davExtensions.forEach(function (token) {                                                       // 1975
                  if (token.match(/^([a-zA-Z0-9!#$%&'*+.^_`|~-]+|<[\x21-\x7E]*>)$/)) {                                 // 1976
                    dav.push(token);                                                                                   // 1977
                  }                                                                                                    // 1978
                });                                                                                                    // 1979
              }                                                                                                        // 1980
                                                                                                                       //
              if (dav.length > 0) {                                                                                    // 1982
                response.setHeader("DAV", dav.join(", "));                                                             // 1983
                response.setHeader("Access-Control-Expose-Headers", "DAV");                                            // 1984
              }                                                                                                        // 1985
                                                                                                                       //
              response.end(); // Return no response; we already handled everything.                                    // 1987
            }, function (err) {                                                                                        // 1989
              if (err.kjType !== "unimplemented") throw err;                                                           // 1990
              response.end(); // Return no response; we already handled everything.                                    // 1991
            });                                                                                                        // 1993
          } else {                                                                                                     // 1994
            throw new Error("Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.");
          }                                                                                                            // 1996
        });                                                                                                            // 1997
      }).then(function (rpcResponse) {                                                                                 // 1998
        if (rpcResponse !== undefined) {                                                                               // 1999
          // Will be undefined for OPTIONS request.                                                                    // 1999
          return _this9.translateResponse(rpcResponse, response, request);                                             // 2000
        }                                                                                                              // 2001
      }).catch(function (error) {                                                                                      // 2002
        return _this9.maybeRetryAfterError(error, retryCount).then(function () {                                       // 2003
          return _this9.handleRequest(request, data, response, retryCount + 1);                                        // 2004
        });                                                                                                            // 2005
      });                                                                                                              // 2006
    }                                                                                                                  // 2007
                                                                                                                       //
    return handleRequest;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleRequestStreaming = function () {                                                               //
    function handleRequestStreaming(request, response, contentLength, retryCount) {                                    //
      var _this10 = this;                                                                                              // 2009
                                                                                                                       //
      return this.makeContext(request, response, function (context) {                                                  // 2010
        var path = request.url.slice(1); // remove leading '/'                                                         // 2011
                                                                                                                       //
        var session = _this10.getSession(request);                                                                     // 2012
                                                                                                                       //
        var mimeType = request.headers["content-type"] || "application/octet-stream";                                  // 2014
        var encoding = request.headers["content-encoding"];                                                            // 2015
        var requestStreamPromise = void 0;                                                                             // 2017
                                                                                                                       //
        if (request.method === "POST") {                                                                               // 2018
          requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);                             // 2019
        } else if (request.method === "PUT") {                                                                         // 2020
          requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);                              // 2021
        } else {                                                                                                       // 2022
          throw new Error("Sandstorm only supports streaming POST and PUT requests.");                                 // 2023
        } // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to                // 2024
        //   handle the case where the app doesn't actually support streaming. We could pipeline while                 // 2027
        //   also buffering the data on the side in case we need it again later, but that's kind of                    // 2028
        //   complicated. We should fix the whole protocol to make streaming the standard.                             // 2029
                                                                                                                       //
                                                                                                                       //
        return requestStreamPromise.then(function (requestStreamResult) {                                              // 2030
          var requestStream = requestStreamResult.stream; // Initialized when getResponse() returns, if the response is streaming.
                                                                                                                       //
          var downloadStreamHandle = void 0; // Initialized if an upload-stream method throws.                         // 2034
                                                                                                                       //
          var uploadStreamError = void 0; // We call `getResponse()` immediately so that the app can start streaming data down even
          // while data is still being streamed up. This theoretically allows apps to perform                          // 2040
          // bidirectional streaming, though probably very few actually do that.                                       // 2041
          //                                                                                                           // 2042
          // Note that we need to be able to cancel `responsePromise` below, so it's important that                    // 2043
          // it is the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate                         // 2044
          // variable.                                                                                                 // 2045
                                                                                                                       //
          var responsePromise = requestStream.getResponse();                                                           // 2046
                                                                                                                       //
          var reportUploadStreamError = function (err) {                                                               // 2048
            // Called when an upload-stream method throws.                                                             // 2049
            if (!uploadStreamError) {                                                                                  // 2051
              uploadStreamError = err; // If we're still waiting on any response stuff, cancel it.                     // 2052
                                                                                                                       //
              responsePromise.cancel();                                                                                // 2055
              requestStream.close();                                                                                   // 2056
                                                                                                                       //
              if (downloadStreamHandle) {                                                                              // 2057
                downloadStreamHandle.close();                                                                          // 2058
              }                                                                                                        // 2059
            }                                                                                                          // 2060
          }; // If we have a Content-Length, pass it along to the app by calling `expectSize()`.                       // 2061
                                                                                                                       //
                                                                                                                       //
          if (contentLength !== undefined) {                                                                           // 2064
            requestStream.expectSize(contentLength).catch(function (err) {                                             // 2065
              // expectSize() is allowed to be unimplemented.                                                          // 2066
              if (err.kjType !== "unimplemented") {                                                                    // 2067
                reportUploadStreamError(err);                                                                          // 2068
              }                                                                                                        // 2069
            });                                                                                                        // 2070
          } // Pipe the input stream to the app.                                                                       // 2071
                                                                                                                       //
                                                                                                                       //
          var writesInFlight = 0;                                                                                      // 2074
          var readingPaused = false;                                                                                   // 2075
          request.on("data", function (buf) {                                                                          // 2076
            // TODO(someday): Coalesce small writes.                                                                   // 2077
            if (buf.length > 4194304) {                                                                                // 2079
              // Whoa, buffer is more than 4MB. We should be splitting it. But does this actually                      // 2080
              // happen? Let's try to detect this problem before we try to do anything about it.                       // 2081
              console.log("whoa, huge upload buffer:", buf.length);                                                    // 2082
              reportUploadStreamError(new Error("upload buffer too big (bug in Sandstorm)"));                          // 2083
              destructor();                                                                                            // 2084
              return;                                                                                                  // 2085
            }                                                                                                          // 2086
                                                                                                                       //
            if (!uploadStreamError) {                                                                                  // 2088
              requestStream.write(buf).then(function () {                                                              // 2089
                writesInFlight--;                                                                                      // 2090
                                                                                                                       //
                if (readingPaused) {                                                                                   // 2091
                  request.resume();                                                                                    // 2092
                  readingPaused = false;                                                                               // 2093
                }                                                                                                      // 2094
              }).catch(reportUploadStreamError);                                                                       // 2095
                                                                                                                       //
              if (writesInFlight++ > 16) {                                                                             // 2097
                request.pause();                                                                                       // 2098
                readingPaused = true;                                                                                  // 2099
              }                                                                                                        // 2100
            }                                                                                                          // 2101
          });                                                                                                          // 2102
          var idx = _this10.streamingRequestCounter;                                                                   // 2104
          _this10.streamingRequestCounter += 1;                                                                        // 2105
          _this10.streamingRequests[idx] = request;                                                                    // 2106
                                                                                                                       //
          destructor = function () {                                                                                   // 2107
            delete _this10.streamingRequests[idx];                                                                     // 2107
          };                                                                                                           // 2107
                                                                                                                       //
          request.on("end", function () {                                                                              // 2109
            if (!uploadStreamError) requestStream.done().catch(reportUploadStreamError); // We're all done making calls to requestStream.
                                                                                                                       //
            requestStream.close();                                                                                     // 2113
            destructor();                                                                                              // 2114
          });                                                                                                          // 2115
          request.on("close", function () {                                                                            // 2117
            reportUploadStreamError(new Error("HTTP connection unexpectedly closed during request."));                 // 2118
            destructor();                                                                                              // 2119
          });                                                                                                          // 2120
          request.on("error", function (err) {                                                                         // 2122
            reportUploadStreamError(err);                                                                              // 2123
            destructor();                                                                                              // 2124
          });                                                                                                          // 2125
          return responsePromise.then(function (rpcResponse) {                                                         // 2127
            // Stop here if the upload stream has already failed.                                                      // 2128
            if (uploadStreamError) throw uploadStreamError;                                                            // 2129
            downloadStreamHandle = _this10.translateResponse(rpcResponse, response, request).streamHandle;             // 2130
          });                                                                                                          // 2131
        }, function (err) {                                                                                            // 2132
          if (err.kjType === "failed" && err.message.indexOf("not implemented") !== -1) {                              // 2133
            // Hack to work around old apps using an old version of Cap'n Proto, before the                            // 2134
            // 'unimplemented' exception type was introduced. :(                                                       // 2135
            // TODO(cleanup): When we transition to API version 2, we can move this into the                           // 2136
            //   compatibility layer.                                                                                  // 2137
            err.kjType = "unimplemented";                                                                              // 2138
          }                                                                                                            // 2139
                                                                                                                       //
          if (shouldRestartGrain(err, 0)) {                                                                            // 2141
            // This is the kind of error that indicates we should retry. Note that we passed 0 for                     // 2142
            // the retry count above because we were just checking if this is a retriable error (vs.                   // 2143
            // possibly a method-not-implemented error); maybeRetryAfterError() will check again with                  // 2144
            // the proper retry count.                                                                                 // 2145
            return _this10.maybeRetryAfterError(err, retryCount).then(function () {                                    // 2146
              return _this10.handleRequestStreaming(request, response, contentLength, retryCount + 1);                 // 2147
            });                                                                                                        // 2148
          } else if (err.kjType === "unimplemented") {                                                                 // 2149
            // Streaming is not implemented. Fall back to non-streaming version.                                       // 2150
            return readAll(request).then(function (data) {                                                             // 2151
              return _this10.handleRequest(request, data, response, 0);                                                // 2152
            });                                                                                                        // 2153
          } else {                                                                                                     // 2154
            throw err;                                                                                                 // 2155
          }                                                                                                            // 2156
        });                                                                                                            // 2157
      });                                                                                                              // 2158
    }                                                                                                                  // 2159
                                                                                                                       //
    return handleRequestStreaming;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleWebSocket = function () {                                                                      //
    function handleWebSocket(request, socket, head, retryCount) {                                                      //
      var _this11 = this;                                                                                              // 2161
                                                                                                                       //
      return Promise.resolve(undefined).then(function () {                                                             // 2162
        return _this11.makeContext(request, undefined, function (context) {                                            // 2163
          var path = request.url.slice(1); // remove leading '/'                                                       // 2164
                                                                                                                       //
          var session = _this11.getSession(request);                                                                   // 2165
                                                                                                                       //
          if (!("sec-websocket-key" in request.headers)) {                                                             // 2167
            throw new Error("Missing Sec-WebSocket-Accept header.");                                                   // 2168
          }                                                                                                            // 2169
                                                                                                                       //
          var magic = request.headers["sec-websocket-key"] + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";                   // 2171
          var acceptKey = Crypto.createHash("sha1").update(magic).digest("base64");                                    // 2172
          var protocols = [];                                                                                          // 2174
                                                                                                                       //
          if ("sec-websocket-protocol" in request.headers) {                                                           // 2175
            protocols = request.headers["sec-websocket-protocol"].split(",").map(function (s) {                        // 2176
              return s.trim();                                                                                         // 2177
            });                                                                                                        // 2177
          }                                                                                                            // 2178
                                                                                                                       //
          var receiver = new WebSocketReceiver(socket);                                                                // 2180
          var promise = session.openWebSocket(path, context, protocols, receiver);                                     // 2182
                                                                                                                       //
          if (head.length > 0) {                                                                                       // 2184
            promise.serverStream.sendBytes(head);                                                                      // 2185
          }                                                                                                            // 2186
                                                                                                                       //
          var socketIdx = _this11.websocketCounter;                                                                    // 2188
          _this11.websockets[socketIdx] = socket;                                                                      // 2189
          _this11.websocketCounter += 1;                                                                               // 2190
          pumpWebSocket(socket, promise.serverStream, function () {                                                    // 2191
            delete _this11.websockets[socketIdx];                                                                      // 2191
          });                                                                                                          // 2191
          return promise.then(function (response) {                                                                    // 2193
            var headers = ["HTTP/1.1 101 Switching Protocols", "Upgrade: websocket", "Connection: Upgrade", "Sec-WebSocket-Accept: " + acceptKey];
                                                                                                                       //
            if (response.protocol && response.protocol.length > 0) {                                                   // 2200
              headers.push("Sec-WebSocket-Protocol: " + response.protocol.join(", "));                                 // 2201
            }                                                                                                          // 2202
                                                                                                                       //
            headers.push("");                                                                                          // 2204
            headers.push("");                                                                                          // 2205
            socket.write(headers.join("\r\n"));                                                                        // 2207
            receiver.go(); // Note:  At this point errors are out of our hands.                                        // 2208
          });                                                                                                          // 2211
        });                                                                                                            // 2212
      }).catch(function (error) {                                                                                      // 2213
        return _this11.maybeRetryAfterError(error, retryCount).then(function () {                                      // 2214
          return _this11.handleWebSocket(request, socket, head, retryCount + 1);                                       // 2215
        });                                                                                                            // 2216
      });                                                                                                              // 2217
    }                                                                                                                  // 2218
                                                                                                                       //
    return handleWebSocket;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.setHasLoaded = function () {                                                                         //
    function setHasLoaded() {                                                                                          //
      if (!this.hasLoaded) {                                                                                           // 2221
        this.hasLoaded = true;                                                                                         // 2222
        var sessionId = this.sessionId;                                                                                // 2223
        inMeteor(function () {                                                                                         // 2224
          Sessions.update({                                                                                            // 2225
            _id: sessionId                                                                                             // 2225
          }, {                                                                                                         // 2225
            $set: {                                                                                                    // 2225
              hasLoaded: true                                                                                          // 2225
            }                                                                                                          // 2225
          });                                                                                                          // 2225
        });                                                                                                            // 2226
      }                                                                                                                // 2227
    }                                                                                                                  // 2228
                                                                                                                       //
    return setHasLoaded;                                                                                               //
  }();                                                                                                                 //
                                                                                                                       //
  return Proxy;                                                                                                        //
}();                                                                                                                   //
                                                                                                                       //
var isRfc1918OrLocal = function (address) {                                                                            // 2231
  if (Net.isIPv4(address)) {                                                                                           // 2232
    var quad = address.split(".").map(function (x) {                                                                   // 2233
      return parseInt(x, 10);                                                                                          // 2233
    });                                                                                                                // 2233
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                                                                    // 2237
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks                                       // 2238
    return address === "::1" || address.lastIndexOf("fd", 0) === 0;                                                    // 2239
  } else {                                                                                                             // 2240
    // Ignore things that are neither IPv4 nor IPv6                                                                    // 2241
    return false;                                                                                                      // 2242
  }                                                                                                                    // 2243
};                                                                                                                     // 2244
                                                                                                                       //
var quadToIntString = function (quad) {                                                                                // 2246
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);          // 2247
  return num.toString();                                                                                               // 2251
}; // -----------------------------------------------------------------------------                                    // 2252
// Session cookie management                                                                                           // 2255
                                                                                                                       //
                                                                                                                       //
var parseCookies = function (request) {                                                                                // 2257
  // jscs:disable requireDotNotation                                                                                   // 2258
  var header = request.headers["cookie"];                                                                              // 2259
  var result = {                                                                                                       // 2261
    cookies: []                                                                                                        // 2261
  };                                                                                                                   // 2261
                                                                                                                       //
  if (header) {                                                                                                        // 2262
    var reqCookies = header.split(";");                                                                                // 2263
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(reqCookies)) {                                                // 2264
      var reqCookie = reqCookies[i];                                                                                   // 2265
      var equalsPos = reqCookie.indexOf("=");                                                                          // 2266
      var cookie = void 0;                                                                                             // 2267
                                                                                                                       //
      if (equalsPos === -1) {                                                                                          // 2268
        cookie = {                                                                                                     // 2269
          key: reqCookie.trim(),                                                                                       // 2269
          value: ""                                                                                                    // 2269
        };                                                                                                             // 2269
      } else {                                                                                                         // 2270
        cookie = {                                                                                                     // 2271
          key: reqCookie.slice(0, equalsPos).trim(),                                                                   // 2271
          value: reqCookie.slice(equalsPos + 1)                                                                        // 2271
        };                                                                                                             // 2271
      }                                                                                                                // 2272
                                                                                                                       //
      if (cookie.key === "sandstorm-sid") {                                                                            // 2274
        if (result.sessionId) {                                                                                        // 2275
          throw new Error("Multiple sandstorm session IDs?");                                                          // 2276
        }                                                                                                              // 2277
                                                                                                                       //
        result.sessionId = cookie.value;                                                                               // 2279
      } else {                                                                                                         // 2280
        result.cookies.push(cookie);                                                                                   // 2281
      }                                                                                                                // 2282
    }                                                                                                                  // 2283
  }                                                                                                                    // 2284
                                                                                                                       //
  return result;                                                                                                       // 2286
};                                                                                                                     // 2287
                                                                                                                       //
var parsePreconditionHeader = function (request) {                                                                     // 2289
  if (request.headers["if-match"]) {                                                                                   // 2290
    if (request.headers["if-match"].trim() === "*") {                                                                  // 2291
      return {                                                                                                         // 2292
        exists: null                                                                                                   // 2292
      };                                                                                                               // 2292
    }                                                                                                                  // 2293
                                                                                                                       //
    var matches = parseETagList(request.headers["if-match"]);                                                          // 2295
                                                                                                                       //
    if (matches.length > 0) {                                                                                          // 2296
      return {                                                                                                         // 2297
        matchesOneOf: matches                                                                                          // 2297
      };                                                                                                               // 2297
    }                                                                                                                  // 2298
  }                                                                                                                    // 2299
                                                                                                                       //
  if (request.headers["if-none-match"]) {                                                                              // 2301
    if (request.headers["if-none-match"].trim() === "*") {                                                             // 2302
      return {                                                                                                         // 2303
        doesntExist: null                                                                                              // 2303
      };                                                                                                               // 2303
    }                                                                                                                  // 2304
                                                                                                                       //
    var noneMatches = parseETagList(request.headers["if-none-match"]);                                                 // 2306
                                                                                                                       //
    if (noneMatches.length > 0) {                                                                                      // 2307
      return {                                                                                                         // 2308
        matchesNoneOf: noneMatches                                                                                     // 2308
      };                                                                                                               // 2308
    }                                                                                                                  // 2309
  }                                                                                                                    // 2310
                                                                                                                       //
  return {                                                                                                             // 2312
    none: null                                                                                                         // 2312
  };                                                                                                                   // 2312
};                                                                                                                     // 2313
                                                                                                                       //
var parseETagList = function (input) {                                                                                 // 2315
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to                          // 2316
  // indicate that it is weak. We are parsing a list of comma-delimited etags.                                         // 2317
  input = input.trim();                                                                                                // 2319
  var results = [];                                                                                                    // 2320
                                                                                                                       //
  while (input.length > 0) {                                                                                           // 2322
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);                                                    // 2323
    if (!match) throw new Meteor.Error(400, "invalid etag");                                                           // 2324
    input = input.slice(match[0].length).trim();                                                                       // 2326
    results.push({                                                                                                     // 2327
      weak: !!match[1],                                                                                                // 2327
      value: match[2].replace(/\\(.)/g, "$1")                                                                          // 2327
    });                                                                                                                // 2327
  }                                                                                                                    // 2328
                                                                                                                       //
  return results;                                                                                                      // 2330
};                                                                                                                     // 2331
                                                                                                                       //
var composeETag = function (tag) {                                                                                     // 2333
  var result = '"' + (tag.value || "").replace(/([\\"])/g, "\\$1") + '"';                                              // 2334
  if (tag.weak) result = "W/" + result;                                                                                // 2335
  return result;                                                                                                       // 2336
};                                                                                                                     // 2337
                                                                                                                       //
var parseAcceptHeaderString = function (header, field) {                                                               // 2339
  var result = [];                                                                                                     // 2340
                                                                                                                       //
  if (header) {                                                                                                        // 2341
    var acceptList = header.split(",");                                                                                // 2342
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(acceptList)) {                                                // 2343
      var acceptStr = acceptList[i];                                                                                   // 2344
      var tokensList = acceptStr.split(";");                                                                           // 2345
      var temp = {};                                                                                                   // 2347
      temp[field] = tokensList[0].trim();                                                                              // 2348
      var tokensListRest = tokensList.slice(1);                                                                        // 2350
                                                                                                                       //
      for (var j in meteorBabelHelpers.sanitizeForInObject(tokensListRest)) {                                          // 2351
        var token = tokensListRest[j];                                                                                 // 2352
        var equalsPos = token.indexOf("=");                                                                            // 2353
                                                                                                                       //
        if (equalsPos) {                                                                                               // 2354
          var key = token.slice(0, equalsPos).trim();                                                                  // 2355
          var value = token.slice(equalsPos + 1).trim();                                                               // 2356
                                                                                                                       //
          if (key === "q") {                                                                                           // 2358
            temp.qValue = +value;                                                                                      // 2359
          }                                                                                                            // 2360
        }                                                                                                              // 2361
      }                                                                                                                // 2362
                                                                                                                       //
      result.push(temp);                                                                                               // 2364
    }                                                                                                                  // 2365
  }                                                                                                                    // 2366
                                                                                                                       //
  return result;                                                                                                       // 2368
};                                                                                                                     // 2369
                                                                                                                       //
var parseAcceptHeader = function (request) {                                                                           // 2371
  // jscs:disable requireDotNotation                                                                                   // 2372
  var header = request.headers["accept"];                                                                              // 2373
  return parseAcceptHeaderString(header, "mimeType");                                                                  // 2374
};                                                                                                                     // 2375
                                                                                                                       //
var parseAcceptEncodingHeader = function (request) {                                                                   // 2377
  // jscs:disable requireDotNotation                                                                                   // 2378
  var header = request.headers["accept-encoding"];                                                                     // 2379
  return parseAcceptHeaderString(header, "contentCoding");                                                             // 2380
}; // -----------------------------------------------------------------------------                                    // 2381
// Regular HTTP request handling                                                                                       // 2384
                                                                                                                       //
                                                                                                                       //
var readAll = function (stream) {                                                                                      // 2386
  return new Promise(function (resolve, reject) {                                                                      // 2387
    var buffers = [];                                                                                                  // 2388
    stream.on("data", function (buf) {                                                                                 // 2389
      buffers.push(buf);                                                                                               // 2390
    });                                                                                                                // 2391
    stream.on("end", function () {                                                                                     // 2393
      resolve(Buffer.concat(buffers));                                                                                 // 2394
    });                                                                                                                // 2395
    stream.on("error", reject);                                                                                        // 2397
  });                                                                                                                  // 2398
};                                                                                                                     // 2399
                                                                                                                       //
var makeSetCookieHeader = function (cookie) {                                                                          // 2401
  var result = [cookie.name, "=", cookie.value];                                                                       // 2402
                                                                                                                       //
  if ("absolute" in cookie.expires) {                                                                                  // 2404
    result.push("; Expires=");                                                                                         // 2405
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());                                               // 2406
  } else if ("relative" in cookie.expires) {                                                                           // 2407
    result.push("; Max-Age=" + cookie.expires.relative);                                                               // 2408
  }                                                                                                                    // 2409
                                                                                                                       //
  if (cookie.path) {                                                                                                   // 2411
    result.push("; Path=" + cookie.path);                                                                              // 2412
  }                                                                                                                    // 2413
                                                                                                                       //
  if (cookie.httpOnly) {                                                                                               // 2415
    result.push("; HttpOnly");                                                                                         // 2416
  }                                                                                                                    // 2417
                                                                                                                       //
  return result.join("");                                                                                              // 2419
}; // TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.                                         // 2420
                                                                                                                       //
                                                                                                                       //
var successCodes = {                                                                                                   // 2423
  ok: {                                                                                                                // 2424
    id: 200,                                                                                                           // 2424
    title: "OK"                                                                                                        // 2424
  },                                                                                                                   // 2424
  created: {                                                                                                           // 2425
    id: 201,                                                                                                           // 2425
    title: "Created"                                                                                                   // 2425
  },                                                                                                                   // 2425
  accepted: {                                                                                                          // 2426
    id: 202,                                                                                                           // 2426
    title: "Accepted"                                                                                                  // 2426
  },                                                                                                                   // 2426
  multiStatus: {                                                                                                       // 2427
    id: 207,                                                                                                           // 2427
    title: "Multi-Status"                                                                                              // 2427
  }                                                                                                                    // 2427
};                                                                                                                     // 2423
var noContentSuccessCodes = [// Indexed by shouldResetForm * 1                                                         // 2429
{                                                                                                                      // 2431
  id: 204,                                                                                                             // 2431
  title: "No Content"                                                                                                  // 2431
}, {                                                                                                                   // 2431
  id: 205,                                                                                                             // 2432
  title: "Reset Content"                                                                                               // 2432
}];                                                                                                                    // 2432
var redirectCodes = [// Indexed by switchToGet * 2 + isPermanent                                                       // 2434
{                                                                                                                      // 2436
  id: 307,                                                                                                             // 2436
  title: "Temporary Redirect"                                                                                          // 2436
}, {                                                                                                                   // 2436
  id: 308,                                                                                                             // 2437
  title: "Permanent Redirect"                                                                                          // 2437
}, {                                                                                                                   // 2437
  id: 303,                                                                                                             // 2438
  title: "See Other"                                                                                                   // 2438
}, {                                                                                                                   // 2438
  id: 301,                                                                                                             // 2439
  title: "Moved Permanently"                                                                                           // 2439
}];                                                                                                                    // 2439
var errorCodes = {                                                                                                     // 2441
  badRequest: {                                                                                                        // 2442
    id: 400,                                                                                                           // 2442
    title: "Bad Request"                                                                                               // 2442
  },                                                                                                                   // 2442
  forbidden: {                                                                                                         // 2443
    id: 403,                                                                                                           // 2443
    title: "Forbidden"                                                                                                 // 2443
  },                                                                                                                   // 2443
  notFound: {                                                                                                          // 2444
    id: 404,                                                                                                           // 2444
    title: "Not Found"                                                                                                 // 2444
  },                                                                                                                   // 2444
  methodNotAllowed: {                                                                                                  // 2445
    id: 405,                                                                                                           // 2445
    title: "Method Not Allowed"                                                                                        // 2445
  },                                                                                                                   // 2445
  notAcceptable: {                                                                                                     // 2446
    id: 406,                                                                                                           // 2446
    title: "Not Acceptable"                                                                                            // 2446
  },                                                                                                                   // 2446
  conflict: {                                                                                                          // 2447
    id: 409,                                                                                                           // 2447
    title: "Conflict"                                                                                                  // 2447
  },                                                                                                                   // 2447
  gone: {                                                                                                              // 2448
    id: 410,                                                                                                           // 2448
    title: "Gone"                                                                                                      // 2448
  },                                                                                                                   // 2448
  requestEntityTooLarge: {                                                                                             // 2449
    id: 413,                                                                                                           // 2449
    title: "Request Entity Too Large"                                                                                  // 2449
  },                                                                                                                   // 2449
  requestUriTooLong: {                                                                                                 // 2450
    id: 414,                                                                                                           // 2450
    title: "Request-URI Too Long"                                                                                      // 2450
  },                                                                                                                   // 2450
  unsupportedMediaType: {                                                                                              // 2451
    id: 415,                                                                                                           // 2451
    title: "Unsupported Media Type"                                                                                    // 2451
  },                                                                                                                   // 2451
  imATeapot: {                                                                                                         // 2452
    id: 418,                                                                                                           // 2452
    title: "I'm a teapot"                                                                                              // 2452
  },                                                                                                                   // 2452
  unprocessableEntity: {                                                                                               // 2453
    id: 422,                                                                                                           // 2453
    title: "Unprocessable Entity"                                                                                      // 2453
  }                                                                                                                    // 2453
};                                                                                                                     // 2441
                                                                                                                       //
ResponseStream = function () {                                                                                         // 2456
  // Note: This class is used in pre-meteor.js as well as in this file.                                                // 2457
  function ResponseStream(response, httpCode, httpStatus, upstreamHandle, idempotentCleanup) {                         // 2459
    var _this12 = this;                                                                                                // 2459
                                                                                                                       //
    (0, _classCallCheck3.default)(this, ResponseStream);                                                               // 2459
    // This is stupidly complicated, because:                                                                          // 2460
    // - New versions of sandstorm-http-bridge (and other well-behaved apps) wait on write()                           // 2461
    //   completion for flow control, so we want write() to complete only when there is space                          // 2462
    //   available for more writes.                                                                                    // 2463
    // - Old versions of sandstorm-http-bridge would make a zillion simultaneous calls to write()                      // 2464
    //   with all of the data they wanted to send without waiting for previous calls to complete.                      // 2465
    //   While in-flight, these writes can consume a lot of RAM at multiple points in the system.                      // 2466
    // - Old versions of sandstorm-http-bridge would print any thrown error to stderr and then                         // 2467
    //   keep going. So if canceling a download early on causes all future write()s to fail, this                      // 2468
    //   can lead to very excessive log spam.                                                                          // 2469
    // - If a write() fails to complete (it returns a promise that never resolves), then later                         // 2470
    //   upon garbage collection the RPC will actually throw an exception about a PromiseFulfiller                     // 2471
    //   never having been fulfilled. Hence, if we simply stop responding to write()s after some                       // 2472
    //   point, we again get log spam.                                                                                 // 2473
    // - If the connection closes during a write, Node never calls the completion callback at all.                     // 2474
    //   We have to listen for the "close" event.                                                                      // 2475
    this.response = response;                                                                                          // 2477
    this.httpCode = httpCode;                                                                                          // 2478
    this.httpStatus = httpStatus;                                                                                      // 2479
    this.upstreamHandle = upstreamHandle;                                                                              // 2480
    this.idempotentCleanup = idempotentCleanup;                                                                        // 2481
    this.started = false;                                                                                              // 2483
    this.ended = false; // Invariant: if `ended` is true, then `started` is also true.                                 // 2484
                                                                                                                       //
    this.waiting = [];                                                                                                 // 2485
    response.on("drain", function () {                                                                                 // 2487
      _this12.waiting.forEach(function (f) {                                                                           // 2488
        return f();                                                                                                    // 2488
      });                                                                                                              // 2488
                                                                                                                       //
      _this12.waiting = [];                                                                                            // 2489
    });                                                                                                                // 2490
    response.on("close", function () {                                                                                 // 2492
      _this12.abort();                                                                                                 // 2492
    });                                                                                                                // 2492
  }                                                                                                                    // 2493
                                                                                                                       //
  ResponseStream.prototype.abort = function () {                                                                       // 2456
    function abort() {                                                                                                 // 2456
      this.aborted = true; // Resolve all outstanding writes. This is OK even if they didn't actually go through because
      // an incomplete download is the client's problem, not the server's. The important thing is                      // 2499
      // to cancel the stream if it is still being generated; to that end, the next new write() call                   // 2500
      // after this point will throw.                                                                                  // 2501
                                                                                                                       //
      this.waiting.forEach(function (f) {                                                                              // 2502
        return f();                                                                                                    // 2502
      });                                                                                                              // 2502
      this.waiting = [];                                                                                               // 2503
      this.close();                                                                                                    // 2505
    }                                                                                                                  // 2506
                                                                                                                       //
    return abort;                                                                                                      // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  ResponseStream.prototype.expectSize = function () {                                                                  // 2456
    function expectSize(size) {                                                                                        // 2456
      if (!this.started) {                                                                                             // 2509
        this.started = true;                                                                                           // 2510
        this.response.writeHead(this.httpCode, this.httpStatus, {                                                      // 2511
          "Content-Length": size                                                                                       // 2511
        });                                                                                                            // 2511
      }                                                                                                                // 2512
    }                                                                                                                  // 2513
                                                                                                                       //
    return expectSize;                                                                                                 // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  ResponseStream.prototype.write = function () {                                                                       // 2456
    function write(data) {                                                                                             // 2456
      var _this13 = this;                                                                                              // 2515
                                                                                                                       //
      if (!this.started) {                                                                                             // 2516
        this.started = true;                                                                                           // 2517
        this.response.writeHead(this.httpCode, this.httpStatus);                                                       // 2518
      }                                                                                                                // 2519
                                                                                                                       //
      if (this.aborted) {                                                                                              // 2521
        // Connection has been aborted.                                                                                // 2522
        if (this.alreadyThrew) {                                                                                       // 2523
          // We already threw an error from a previous write(). Do not throw more errors, because                      // 2524
          // older versions of sandstorm-http-bridge do not stop on the first error and will print                     // 2525
          // every error to the log, wasting space. Newer versions stop on the first failure.                          // 2526
          return;                                                                                                      // 2527
        } else {                                                                                                       // 2528
          this.alreadyThrew = true;                                                                                    // 2529
          throw new Error("client disconnected");                                                                      // 2530
        }                                                                                                              // 2531
      } else {                                                                                                         // 2532
        if (this.response.write(data)) {                                                                               // 2533
          // All written.                                                                                              // 2534
          return;                                                                                                      // 2535
        } else {                                                                                                       // 2536
          // Write buffer is full. Don't complete until it has more space.                                             // 2537
          return new Promise(function (resolve, reject) {                                                              // 2538
            _this13.waiting.push(resolve);                                                                             // 2539
          });                                                                                                          // 2540
        }                                                                                                              // 2541
      }                                                                                                                // 2542
    }                                                                                                                  // 2543
                                                                                                                       //
    return write;                                                                                                      // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  ResponseStream.prototype.done = function () {                                                                        // 2456
    function done() {                                                                                                  // 2456
      if (!this.started) {                                                                                             // 2546
        this.started = true;                                                                                           // 2547
        this.response.writeHead(this.httpCode, this.httpStatus, {                                                      // 2548
          "Content-Length": 0                                                                                          // 2548
        });                                                                                                            // 2548
      }                                                                                                                // 2549
                                                                                                                       //
      if (!this.ended) {                                                                                               // 2551
        this.ended = true;                                                                                             // 2552
        this.response.end();                                                                                           // 2553
      }                                                                                                                // 2554
                                                                                                                       //
      this.close();                                                                                                    // 2556
    }                                                                                                                  // 2557
                                                                                                                       //
    return done;                                                                                                       // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  ResponseStream.prototype.close = function () {                                                                       // 2456
    function close() {                                                                                                 // 2456
      if (this.upstreamHandle) {                                                                                       // 2560
        this.upstreamHandle.close();                                                                                   // 2561
      }                                                                                                                // 2562
                                                                                                                       //
      if (this.idempotentCleanup) {                                                                                    // 2564
        this.idempotentCleanup();                                                                                      // 2565
      } // If `this.done()` has not been called, we should report an error to the client. Unfortunately,               // 2566
      // the only case in which we can actually report the error is when we haven't done any writes                    // 2569
      // yet. So we handle that case and ignore the error otherwise. Note that `!this.started`                         // 2570
      // implies `!this.ended`.                                                                                        // 2571
                                                                                                                       //
                                                                                                                       //
      if (!this.started) {                                                                                             // 2572
        this.started = true;                                                                                           // 2573
        this.response.writeHead(500, "done() never called on response stream", {                                       // 2574
          "Content-Type": "text/plain"                                                                                 // 2575
        });                                                                                                            // 2575
      }                                                                                                                // 2576
                                                                                                                       //
      if (!this.ended) {                                                                                               // 2578
        this.ended = true;                                                                                             // 2579
        this.response.end();                                                                                           // 2580
      }                                                                                                                // 2581
    }                                                                                                                  // 2582
                                                                                                                       //
    return close;                                                                                                      // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  ResponseStream.prototype.sendingDirectResponse = function () {                                                       // 2456
    function sendingDirectResponse() {                                                                                 // 2456
      // For internal front-end use: Indicates that the caller is going to fill in the response                        // 2585
      // instead. If a response has already been started, this throws an exception.                                    // 2586
      //                                                                                                               // 2587
      // This is used in particular in pre-meteor.js, when dealing with static web publishing.                         // 2588
      if (this.started) {                                                                                              // 2590
        throw new Error("can't sendDirectResponse() because response already started");                                // 2591
      } else if (this.ended) {                                                                                         // 2592
        throw new Error("can't sendDirectResponse() because response already sent");                                   // 2593
      } else {                                                                                                         // 2594
        this.started = true;                                                                                           // 2595
        this.ended = true;                                                                                             // 2596
      }                                                                                                                // 2597
    }                                                                                                                  // 2598
                                                                                                                       //
    return sendingDirectResponse;                                                                                      // 2456
  }();                                                                                                                 // 2456
                                                                                                                       //
  return ResponseStream;                                                                                               // 2456
}(); // TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.                       // 2456
                                                                                                                       //
                                                                                                                       //
var Zlib = Npm.require("zlib");                                                                                        // 2602
                                                                                                                       //
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib); // -----------------------------------------------------------------------------
// WebSocket handling                                                                                                  // 2606
                                                                                                                       //
WebSocketReceiver = function () {                                                                                      // 2608
  function WebSocketReceiver(socket) {                                                                                 // 2609
    (0, _classCallCheck3.default)(this, WebSocketReceiver);                                                            // 2609
    this.socket = socket;                                                                                              // 2610
    this.queue = [];                                                                                                   // 2611
  }                                                                                                                    // 2612
                                                                                                                       //
  WebSocketReceiver.prototype.go = function () {                                                                       // 2608
    function go() {                                                                                                    // 2608
      for (var i in meteorBabelHelpers.sanitizeForInObject(this.queue)) {                                              // 2615
        this.socket.write(this.queue[i]);                                                                              // 2616
      }                                                                                                                // 2617
                                                                                                                       //
      this.queue = null;                                                                                               // 2619
    }                                                                                                                  // 2620
                                                                                                                       //
    return go;                                                                                                         // 2608
  }();                                                                                                                 // 2608
                                                                                                                       //
  WebSocketReceiver.prototype.sendBytes = function () {                                                                // 2608
    function sendBytes(message) {                                                                                      // 2608
      // TODO(someday):  Flow control of some sort?                                                                    // 2623
      if (this.queue === null) {                                                                                       // 2624
        this.socket.write(message);                                                                                    // 2625
      } else {                                                                                                         // 2626
        this.queue.push(message);                                                                                      // 2627
      }                                                                                                                // 2628
    }                                                                                                                  // 2629
                                                                                                                       //
    return sendBytes;                                                                                                  // 2608
  }();                                                                                                                 // 2608
                                                                                                                       //
  WebSocketReceiver.prototype.close = function () {                                                                    // 2608
    function close() {                                                                                                 // 2608
      this.socket.end();                                                                                               // 2632
    }                                                                                                                  // 2633
                                                                                                                       //
    return close;                                                                                                      // 2608
  }();                                                                                                                 // 2608
                                                                                                                       //
  return WebSocketReceiver;                                                                                            // 2608
}();                                                                                                                   // 2608
                                                                                                                       //
function pumpWebSocket(socket, rpcStream, destructor) {                                                                // 2636
  var writesInFlight = 0;                                                                                              // 2637
  var readingPaused = false;                                                                                           // 2638
  socket.on("data", function (chunk) {                                                                                 // 2639
    rpcStream.sendBytes(chunk).then(function () {                                                                      // 2640
      writesInFlight--;                                                                                                // 2641
                                                                                                                       //
      if (readingPaused) {                                                                                             // 2642
        socket.resume();                                                                                               // 2643
        readingPaused = false;                                                                                         // 2644
      }                                                                                                                // 2645
    }).catch(function (err) {                                                                                          // 2646
      if (err.kjType !== "disconnected") {                                                                             // 2647
        console.error("WebSocket sendBytes failed: " + err.stack);                                                     // 2648
      }                                                                                                                // 2649
                                                                                                                       //
      socket.destroy();                                                                                                // 2651
    });                                                                                                                // 2652
                                                                                                                       //
    if (writesInFlight++ > 16) {                                                                                       // 2654
      socket.pause();                                                                                                  // 2655
      readingPaused = true;                                                                                            // 2656
    }                                                                                                                  // 2657
  });                                                                                                                  // 2658
  socket.on("end", function (chunk) {                                                                                  // 2660
    rpcStream.close();                                                                                                 // 2661
  });                                                                                                                  // 2662
  socket.on("close", function () {                                                                                     // 2664
    destructor();                                                                                                      // 2665
  });                                                                                                                  // 2666
}                                                                                                                      // 2667
                                                                                                                       //
; // =======================================================================================                           // 2667
// Debug log access                                                                                                    // 2670
                                                                                                                       //
Meteor.publish("grainLog", function (grainId) {                                                                        // 2672
  check(grainId, String);                                                                                              // 2673
  var id = 0;                                                                                                          // 2674
  var grain = Grains.findOne(grainId);                                                                                 // 2675
                                                                                                                       //
  if (!grain || !this.userId || grain.userId !== this.userId) {                                                        // 2676
    this.added("grainLog", id++, {                                                                                     // 2677
      text: "Only the grain owner can view the debug log."                                                             // 2677
    });                                                                                                                // 2677
    this.ready();                                                                                                      // 2678
    return;                                                                                                            // 2679
  }                                                                                                                    // 2680
                                                                                                                       //
  var connected = false;                                                                                               // 2682
                                                                                                                       //
  var _this = this;                                                                                                    // 2683
                                                                                                                       //
  var receiver = {                                                                                                     // 2685
    write: function (data) {                                                                                           // 2686
      connected = true;                                                                                                // 2687
                                                                                                                       //
      _this.added("grainLog", id++, {                                                                                  // 2688
        text: data.toString("utf8")                                                                                    // 2688
      });                                                                                                              // 2688
    },                                                                                                                 // 2689
    close: function () {                                                                                               // 2691
      if (connected) {                                                                                                 // 2692
        _this.added("grainLog", id++, {                                                                                // 2693
          text: "*** lost connection to grain (probably because it shut down) ***"                                     // 2694
        });                                                                                                            // 2693
      }                                                                                                                // 2696
    }                                                                                                                  // 2697
  };                                                                                                                   // 2685
                                                                                                                       //
  try {                                                                                                                // 2700
    var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {                                   // 2701
      return supervisor.watchLog(8192, receiver);                                                                      // 2702
    })).handle;                                                                                                        // 2703
    connected = true;                                                                                                  // 2704
    this.onStop(function () {                                                                                          // 2705
      handle.close();                                                                                                  // 2706
    });                                                                                                                // 2707
  } catch (err) {                                                                                                      // 2708
    if (!connected) {                                                                                                  // 2709
      this.added("grainLog", id++, {                                                                                   // 2710
        text: "*** couldn't connect to grain (" + err + ") ***"                                                        // 2711
      });                                                                                                              // 2710
    }                                                                                                                  // 2713
  } // Notify ready.                                                                                                   // 2714
                                                                                                                       //
                                                                                                                       //
  this.ready();                                                                                                        // 2717
});                                                                                                                    // 2718
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"sandcats.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/sandcats.js                                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var pki = void 0,                                                                                                      // 1
    asn1 = void 0;                                                                                                     // 1
module.watch(require("node-forge"), {                                                                                  // 1
  pki: function (v) {                                                                                                  // 1
    pki = v;                                                                                                           // 1
  },                                                                                                                   // 1
  asn1: function (v) {                                                                                                 // 1
    asn1 = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  SANDSTORM_ALTHOME: function (v) {                                                                                    // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Sandcats = {};                                                                                                         // 17
                                                                                                                       //
var querystring = Npm.require("querystring");                                                                          // 20
                                                                                                                       //
var https = Npm.require("https");                                                                                      // 21
                                                                                                                       //
var fs = Npm.require("fs");                                                                                            // 22
                                                                                                                       //
var dgram = Npm.require("dgram");                                                                                      // 23
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 24
                                                                                                                       //
var SANDCATS_HOSTNAME = Meteor.settings && Meteor.settings.public && Meteor.settings.public.sandcatsHostname;          // 28
var SANDCATS_VARDIR = (SANDSTORM_ALTHOME || "") + "/var/sandcats"; // Figure out what IP address to send Sandcats requests from. For machines with multiple IPs, it
// is important to use the IP to which we're binding. However, some people set BIND_IP to 127.0.0.1                    // 33
// and put sniproxy in front of Sandstorm. In those cases, it won't work to send from BIND_IP;                         // 34
// we'll have to let the system choose.                                                                                // 35
                                                                                                                       //
var BIND_IP = process.env.BIND_IP && process.env.BIND_IP.startsWith("127.") ? null : process.env.BIND_IP;              // 36
var ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                        // 39
var HOSTNAME = ROOT_URL.hostname;                                                                                      // 40
var SANDCATS_NAME = void 0; // Look at `startup` below to see where this is set                                        // 41
                                                                                                                       //
var updateSandcatsIp = function () {                                                                                   // 43
  var responseCallback = function (res) {                                                                              // 44
    if (res.statusCode === 200) {                                                                                      // 45
      console.log("Successfully updated sandcats IP");                                                                 // 46
    } else {                                                                                                           // 47
      console.error("Failed to update sandcats IP:", res.headers);                                                     // 48
    }                                                                                                                  // 49
  };                                                                                                                   // 50
                                                                                                                       //
  var errorCallback = function (err) {                                                                                 // 52
    console.error("Couldn't send update sandcats hostname", err);                                                      // 53
  };                                                                                                                   // 54
                                                                                                                       //
  performSandcatsRequest("/update", SANDCATS_HOSTNAME, {                                                               // 56
    rawHostname: SANDCATS_NAME                                                                                         // 56
  }, errorCallback, responseCallback);                                                                                 // 56
};                                                                                                                     // 58
                                                                                                                       //
var pingUdp = function () {                                                                                            // 60
  var socket = dgram.createSocket("udp4");                                                                             // 61
  var secret = Random.secret(16);                                                                                      // 62
  message = new Buffer(SANDCATS_NAME + " " + secret);                                                                  // 64
  socket.on("message", function (buf) {                                                                                // 65
    if (buf.toString() === secret) {                                                                                   // 66
      updateSandcatsIp();                                                                                              // 67
    } else {                                                                                                           // 68
      console.error("Received unexpected response in UDP sandcats ping:", buf.toString());                             // 69
    }                                                                                                                  // 70
  });                                                                                                                  // 71
  socket.on("error", function (err) {                                                                                  // 73
    throw err;                                                                                                         // 74
  });                                                                                                                  // 75
                                                                                                                       //
  var callback = function () {                                                                                         // 77
    socket.send(message, 0, message.length, 8080, SANDCATS_HOSTNAME, function (err) {                                  // 78
      if (err) {                                                                                                       // 79
        console.error("Couldn't send UDP sandcats ping", err);                                                         // 80
      }                                                                                                                // 81
    });                                                                                                                // 82
    setTimeout(function () {                                                                                           // 84
      socket.close();                                                                                                  // 85
    }, 10 * 1000);                                                                                                     // 86
  };                                                                                                                   // 87
                                                                                                                       //
  if (BIND_IP) {                                                                                                       // 89
    socket.bind({                                                                                                      // 90
      address: BIND_IP                                                                                                 // 90
    }, callback);                                                                                                      // 90
  } else {                                                                                                             // 91
    callback();                                                                                                        // 92
  }                                                                                                                    // 93
};                                                                                                                     // 94
                                                                                                                       //
var performSandcatsRequest = function (path, hostname, postData, errorCallback, responseCallback) {                    // 96
  var options = {                                                                                                      // 97
    hostname: hostname,                                                                                                // 98
    path: path,                                                                                                        // 99
    method: "POST",                                                                                                    // 100
    agent: false,                                                                                                      // 101
    key: fs.readFileSync(SANDCATS_VARDIR + "/id_rsa"),                                                                 // 102
    cert: fs.readFileSync(SANDCATS_VARDIR + "/id_rsa.pub"),                                                            // 103
    headers: {                                                                                                         // 104
      "X-Sand": "cats",                                                                                                // 105
      "Content-Type": "application/x-www-form-urlencoded"                                                              // 106
    }                                                                                                                  // 104
  };                                                                                                                   // 97
                                                                                                                       //
  if (BIND_IP) {                                                                                                       // 110
    options.localAddress = BIND_IP;                                                                                    // 111
  }                                                                                                                    // 112
                                                                                                                       //
  if (postData.certificateSigningRequest) {                                                                            // 114
    console.log("Submitting certificate request for host", postData.rawHostname, "where the request has length", postData.certificateSigningRequest.length);
  }                                                                                                                    // 118
                                                                                                                       //
  var newPostData = querystring.stringify(postData);                                                                   // 120
                                                                                                                       //
  if (SANDCATS_HOSTNAME === "sandcats-dev.sandstorm.io" && !process.env.NODE_TLS_REJECT_UNAUTHORIZED) {                // 121
    console.log("You are using the Sandcats dev server but have full HTTPS " + "certificate checking enabled. Probably you will get an HTTPS " + "error. Proceeding anyway.");
  }                                                                                                                    // 126
                                                                                                                       //
  var req = https.request(options, responseCallback);                                                                  // 128
  req.write(newPostData);                                                                                              // 129
  req.end();                                                                                                           // 130
                                                                                                                       //
  if (errorCallback) {                                                                                                 // 132
    req.on("error", errorCallback);                                                                                    // 133
  }                                                                                                                    // 134
                                                                                                                       //
  return req;                                                                                                          // 136
};                                                                                                                     // 137
                                                                                                                       //
var generateKeyAndCsr = function (commonName) {                                                                        // 139
  check(commonName, String); // Generate key pair. Using Meteor.wrapAsync because forge supports                       // 140
  // a synchronous as well as an asynchronous API, and the synchronous                                                 // 143
  // one blocks for a while.                                                                                           // 144
                                                                                                                       //
  var wrappedGenerateKeyPair = Meteor.wrapAsync(pki.rsa.generateKeyPair); // I could pick an `e`[xponent] value for the resulting RSA key, but
  // I will refrain.                                                                                                   // 148
                                                                                                                       //
  var keys = wrappedGenerateKeyPair({                                                                                  // 149
    bits: 2048                                                                                                         // 149
  }); // Create a certificate request (CSR).                                                                           // 149
                                                                                                                       //
  var csr = pki.createCertificationRequest();                                                                          // 152
  csr.publicKey = keys.publicKey;                                                                                      // 153
  csr.setSubject([{                                                                                                    // 154
    name: "commonName",                                                                                                // 156
    value: commonName,                                                                                                 // 157
    valueTagClass: asn1.Type.UTF8 // We specify UTF8 to encode a UTF8String (rather than the default of PRINTABLESTRING) in the
    // commonName so that GlobalSign does not report a warning, and also because that happens to                       // 160
    // be what openssl(1) does when asked to create a CSR.                                                             // 161
                                                                                                                       //
  }]);                                                                                                                 // 155
  csr.sign(keys.privateKey);                                                                                           // 164
  console.log("generateKeyAndCsr created new key & certificate request for", commonName);                              // 165
  return {                                                                                                             // 166
    privateKeyAsPem: pki.privateKeyToPem(keys.privateKey),                                                             // 167
    csrAsPem: pki.certificationRequestToPem(csr)                                                                       // 168
  };                                                                                                                   // 166
};                                                                                                                     // 170
                                                                                                                       //
Sandcats.storeNewKeyAndCsr = function (hostname, basePath) {                                                           // 172
  // We use the current JS time (like UNIX timestamp but in                                                            // 173
  // milliseconds) as the key number. Note that the keyNumber is                                                       // 174
  // intended to be an opaque identifier; the only important thing is                                                  // 175
  // that it increases numerically over time. We use the current time                                                  // 176
  // just as a simplistic way to pick a filename that probably no one                                                  // 177
  // else has created yet.                                                                                             // 178
  var keyNumber = new Date().getTime();                                                                                // 179
  var keyFilename = basePath + "/" + keyNumber;                                                                        // 180
  var csrFilename = keyFilename + ".csr";                                                                              // 181
  var responseFilename = keyFilename + ".response-json";                                                               // 182
  var withWildcard = "*." + hostname;                                                                                  // 183
  var keyAndCsr = generateKeyAndCsr(withWildcard);                                                                     // 184
  fs.writeFileSync(keyFilename, keyAndCsr.privateKeyAsPem, {                                                           // 185
    mode: 0400                                                                                                         // 186
  });                                                                                                                  // 186
  fs.writeFileSync(csrFilename, keyAndCsr.csrAsPem);                                                                   // 187
  console.log("storeNewKeyAndCsr successfully saved key and certificate request to", keyFilename, "and", csrFilename, "respectively of length", keyAndCsr.privateKeyAsPem.length, "and", keyAndCsr.csrAsPem.length);
  return {                                                                                                             // 192
    csrFilename: csrFilename,                                                                                          // 193
    keyFilename: keyFilename,                                                                                          // 194
    responseFilename: responseFilename                                                                                 // 195
  };                                                                                                                   // 192
};                                                                                                                     // 197
                                                                                                                       //
Sandcats.renewHttpsCertificateIfNeeded = function () {                                                                 // 199
  var renewHttpsCertificate = function () {                                                                            // 200
    var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                           // 201
    var basePath = "/var/sandcats/https/" + hostname;                                                                  // 202
    var filenames = Sandcats.storeNewKeyAndCsr(hostname, basePath);                                                    // 203
                                                                                                                       //
    var errorCallback = function (err) {                                                                               // 205
      console.error("Error while renewing HTTPS certificate (will continue to retry)", err);                           // 206
    };                                                                                                                 // 207
                                                                                                                       //
    var responseCallback = function (res) {                                                                            // 209
      if (res.statusCode == 200) {                                                                                     // 210
        // Save the response, chunk by chunk, then store it on disk                                                    // 211
        // for later use.                                                                                              // 212
        var responseBody = "";                                                                                         // 213
        res.on("data", function (chunk) {                                                                              // 214
          responseBody += chunk;                                                                                       // 215
        });                                                                                                            // 216
        res.on("end", function () {                                                                                    // 218
          // For sanity, make sure it parses as JSON, since we're                                                      // 219
          // going to need it to do that.                                                                              // 220
          try {                                                                                                        // 221
            JSON.parse(responseBody);                                                                                  // 222
          } catch (e) {                                                                                                // 223
            console.error("JSON parse error receiving new HTTPS certificate. Discarding:", responseBody, "due to exception:", e); // Overwrite the responseBody with the empty JSON object
            // and continue with the process of saving it to disk.                                                     // 229
                                                                                                                       //
            responseBody = "{}";                                                                                       // 230
          }                                                                                                            // 231
                                                                                                                       //
          try {                                                                                                        // 233
            fs.writeFileSync(filenames.responseFilename, responseBody, "utf-8");                                       // 234
          } catch (err) {                                                                                              // 235
            return console.error("Failure while saveing new HTTPS certificate to", filenames.responseFilename, "will continue to retry. Exception was:", err);
          } // Tell Sandcats that now is a good time to update its info                                                // 240
          // about which keys are available.                                                                           // 243
                                                                                                                       //
                                                                                                                       //
          if (!global.sandcats) {                                                                                      // 244
            console.error("When getting new certificate, could not find callback to request re-keying! Certs will probably become invalid soon.");
          } else {                                                                                                     // 246
            // Call the sandcats rekeying function.                                                                    // 247
            global.sandcats.rekey(); // That's that.                                                                   // 248
                                                                                                                       //
            console.log("Successfully renewed HTTPS certificate into", filenames.responseFilename);                    // 250
          }                                                                                                            // 252
        });                                                                                                            // 253
      } else {                                                                                                         // 254
        console.log("Received HTTP error while renewing certificate (will keep retrying)", res.statusCode);            // 255
        res.on("data", function (chunk) {                                                                              // 257
          console.log("Error response contained information", chunk.toString("utf-8"));                                // 258
        });                                                                                                            // 259
      }                                                                                                                // 260
    };                                                                                                                 // 261
                                                                                                                       //
    var wrappedReadFile = Meteor.wrapAsync(fs.readFile);                                                               // 263
    performSandcatsRequest("/getcertificate", SANDCATS_HOSTNAME, {                                                     // 265
      rawHostname: SANDCATS_NAME,                                                                                      // 266
      certificateSigningRequest: wrappedReadFile(filenames.csrFilename, "utf-8")                                       // 267
    }, errorCallback, responseCallback);                                                                               // 265
  };                                                                                                                   // 269
                                                                                                                       //
  if (global.sandcats.shouldGetAnotherCertificate()) {                                                                 // 271
    console.log("renewHttpsCertificateIfNeeded: Happily choosing to renew certificate.");                              // 272
    return renewHttpsCertificate();                                                                                    // 273
  }                                                                                                                    // 274
};                                                                                                                     // 275
                                                                                                                       //
Sandcats.initializeSandcats = function () {                                                                            // 277
  var i = HOSTNAME.lastIndexOf(SANDCATS_HOSTNAME);                                                                     // 278
                                                                                                                       //
  if (i < 0) {                                                                                                         // 279
    console.error("SANDCATS_BASE_DOMAIN is configured but your HOSTNAME doesn't appear to contain it:", SANDCATS_HOSTNAME, HOSTNAME);
  } else {                                                                                                             // 282
    var oneMinute = 60 * 1000;                                                                                         // 283
    var oneHour = 60 * oneMinute;                                                                                      // 284
    var randomIntervalZeroToOneHour = Math.random() * oneHour; // All Sandcats installs need dyndns updating.          // 285
                                                                                                                       //
    SANDCATS_NAME = HOSTNAME.slice(0, i - 1);                                                                          // 287
    Meteor.setInterval(pingUdp, oneMinute); // If process.env.HTTPS_PORT is set, we need to auto-refresh our HTTPS certificate.
                                                                                                                       //
    if (process.env.HTTPS_PORT) {                                                                                      // 290
      // Always do a HTTPS certificate update check on Sandstorm start.                                                // 291
      Sandcats.renewHttpsCertificateIfNeeded(); // After that's done, schedule it for every approx 1-2 hours in        // 292
      // the future.                                                                                                   // 295
                                                                                                                       //
      Meteor.setInterval(Sandcats.renewHttpsCertificateIfNeeded, oneHour + randomIntervalZeroToOneHour);               // 296
    }                                                                                                                  // 298
  }                                                                                                                    // 299
};                                                                                                                     // 300
                                                                                                                       //
if (SANDCATS_HOSTNAME) {                                                                                               // 302
  Meteor.startup(Sandcats.initializeSandcats);                                                                         // 303
}                                                                                                                      // 304
                                                                                                                       //
;                                                                                                                      // 304
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"shell-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/shell-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var inMeteor = void 0;                                                                                                 // 1
module.watch(require("/imports/server/async-helpers.js"), {                                                            // 1
  inMeteor: function (v) {                                                                                             // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var staticAssetHost = void 0;                                                                                          // 1
module.watch(require("/imports/server/constants.js"), {                                                                // 1
  staticAssetHost: function (v) {                                                                                      // 1
    staticAssetHost = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
BrowserPolicy.framing.disallow(); // Disallow framing of the UI.                                                       // 38
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 39
  var frameSetter = function () {                                                                                      // 40
    BrowserPolicy.content.disallowFrame(); // This clears all the old rules                                            // 41
                                                                                                                       //
    BrowserPolicy.content.allowFrameOrigin(getWildcardOrigin());                                                       // 42
    var billingPromptUrl = globalDb.getBillingPromptUrl();                                                             // 43
                                                                                                                       //
    if (billingPromptUrl) {                                                                                            // 44
      BrowserPolicy.content.allowFrameOrigin(billingPromptUrl);                                                        // 45
    }                                                                                                                  // 46
  };                                                                                                                   // 47
                                                                                                                       //
  frameSetter(); // Call once on startup                                                                               // 49
                                                                                                                       //
  globalDb.collections.settings.find({                                                                                 // 50
    _id: "billingPromptUrl"                                                                                            // 50
  }).observe({                                                                                                         // 50
    added: frameSetter,                                                                                                // 51
    changed: frameSetter,                                                                                              // 52
    removed: frameSetter                                                                                               // 53
  });                                                                                                                  // 50
}); // Allow anything to be loaded from the static asset host.                                                         // 55
                                                                                                                       //
BrowserPolicy.content.allowImageOrigin(staticAssetHost);                                                               // 59
BrowserPolicy.content.allowScriptOrigin(staticAssetHost);                                                              // 60
BrowserPolicy.content.allowFontOrigin(staticAssetHost);                                                                // 61
BrowserPolicy.content.allowConnectOrigin(staticAssetHost);                                                             // 62
BrowserPolicy.content.allowConnectOrigin("wss:");                                                                      // 63
BrowserPolicy.content.allowConnectOrigin("ws:");                                                                       // 64
Meteor.publish("grainsMenu", function () {                                                                             // 66
  if (this.userId) {                                                                                                   // 67
    if (globalDb.isQuotaEnabled()) {                                                                                   // 68
      // Hack: Fire off an asynchronous update to the user's storage usage whenever they open the                      // 69
      //   front page.                                                                                                 // 70
      // TODO(someday): Implement the ability to reactively subscribe to storage usage from the                        // 71
      //   back-end?                                                                                                   // 72
      var userId = this.userId;                                                                                        // 73
      globalBackend.cap().getUserStorageUsage(userId).then(function (results) {                                        // 74
        inMeteor(function () {                                                                                         // 75
          Meteor.users.update(userId, {                                                                                // 76
            $set: {                                                                                                    // 76
              storageUsage: parseInt(results.size)                                                                     // 76
            }                                                                                                          // 76
          });                                                                                                          // 76
        });                                                                                                            // 77
      }).catch(function (err) {                                                                                        // 78
        if (err.kjType === "unimplemented") {                                                                          // 79
          // Compute based on sum of grain sizes instead.                                                              // 80
          var total = 0;                                                                                               // 81
          Grains.find({                                                                                                // 82
            userId: userId                                                                                             // 82
          }, {                                                                                                         // 82
            fields: {                                                                                                  // 82
              size: 1                                                                                                  // 82
            }                                                                                                          // 82
          }).forEach(function (grain) {                                                                                // 82
            return total += grain.size || 0;                                                                           // 83
          });                                                                                                          // 83
          Meteor.users.update(userId, {                                                                                // 84
            $set: {                                                                                                    // 84
              storageUsage: total                                                                                      // 84
            }                                                                                                          // 84
          });                                                                                                          // 84
        } else {                                                                                                       // 85
          console.error(err.stack);                                                                                    // 86
        }                                                                                                              // 87
      });                                                                                                              // 88
    }                                                                                                                  // 89
                                                                                                                       //
    return [UserActions.find({                                                                                         // 91
      userId: this.userId                                                                                              // 92
    }), Grains.find({                                                                                                  // 92
      userId: this.userId                                                                                              // 93
    }), ApiTokens.find({                                                                                               // 93
      "owner.user.accountId": this.userId                                                                              // 94
    })];                                                                                                               // 94
  } else {                                                                                                             // 96
    return [];                                                                                                         // 97
  }                                                                                                                    // 98
});                                                                                                                    // 99
Meteor.publish("sessions", function (sessionId) {                                                                      // 101
  // sessionId itself should be secret enough, but they are also not meant to be shared, so as                         // 102
  // a backup we only publish the session to its owner. Note that `userId` can be null if the                          // 103
  // user is not logged in or is using incognito mode.                                                                 // 104
  check(sessionId, String); // We exclude powerboxRequest because the client already has the descriptor list in packed
  // format, and the parsed format can be kind of large.                                                               // 108
                                                                                                                       //
  return Sessions.find({                                                                                               // 109
    _id: sessionId,                                                                                                    // 109
    $or: [{                                                                                                            // 109
      userId: this.userId                                                                                              // 109
    }, {                                                                                                               // 109
      userId: null                                                                                                     // 109
    }]                                                                                                                 // 109
  }, {                                                                                                                 // 109
    fields: {                                                                                                          // 110
      powerboxRequest: 0                                                                                               // 110
    }                                                                                                                  // 110
  });                                                                                                                  // 110
});                                                                                                                    // 111
Meteor.publish("devPackages", function () {                                                                            // 113
  return DevPackages.find();                                                                                           // 114
});                                                                                                                    // 115
Meteor.publish("hasUsers", function () {                                                                               // 117
  var _this = this;                                                                                                    // 117
                                                                                                                       //
  // Publish pseudo-collection which tells the client if there are any users at all.                                   // 118
  //                                                                                                                   // 119
  // TODO(cleanup):  This seems overcomplicated.  Does Meteor have a better way?                                       // 120
  var cursor = Meteor.users.find();                                                                                    // 121
                                                                                                                       //
  if (cursor.count() > 0) {                                                                                            // 122
    this.added("hasUsers", "hasUsers", {                                                                               // 123
      hasUsers: true                                                                                                   // 123
    });                                                                                                                // 123
  } else {                                                                                                             // 124
    var handle = cursor.observeChanges({                                                                               // 125
      added: function (id) {                                                                                           // 126
        _this.added("hasUsers", "hasUsers", {                                                                          // 127
          hasUsers: true                                                                                               // 127
        });                                                                                                            // 127
                                                                                                                       //
        handle.stop();                                                                                                 // 128
        handle = null;                                                                                                 // 129
      }                                                                                                                // 130
    });                                                                                                                // 125
    this.onStop(function () {                                                                                          // 132
      if (handle) handle.stop();                                                                                       // 133
    });                                                                                                                // 134
  }                                                                                                                    // 135
                                                                                                                       //
  this.ready();                                                                                                        // 137
});                                                                                                                    // 138
Meteor.publish("referralInfoPseudo", function () {                                                                     // 140
  var _this2 = this;                                                                                                   // 140
                                                                                                                       //
  // This publishes a pseudo-collection called referralInfo whose documents have the following                         // 141
  // form:                                                                                                             // 142
  //                                                                                                                   // 143
  // - id: (String) same as the User._id of an account this user has referred                                          // 144
  // - name: (String) the profile.name from that account                                                               // 145
  // - completed: (Boolean) if this referral is complete                                                               // 146
  //  If the user is not logged in, then we have no referralInfo.                                                      // 148
  if (!this.userId) {                                                                                                  // 149
    return [];                                                                                                         // 150
  } // Implementation note:                                                                                            // 151
  //                                                                                                                   // 154
  // This pseudo-collection is populated very differently for (1) the completed: false case versus                     // 155
  // the (2) completed: true case.                                                                                     // 156
  // Case 1. Publish information about not-yet-complete referrals.                                                     // 158
                                                                                                                       //
                                                                                                                       //
  var notCompletedReferralAccountsCursor = Meteor.users.find({                                                         // 159
    referredBy: this.userId,                                                                                           // 160
    referredByComplete: {                                                                                              // 161
      $exists: false                                                                                                   // 161
    },                                                                                                                 // 161
    type: "account",                                                                                                   // 162
    "profile.name": {                                                                                                  // 163
      $exists: true                                                                                                    // 163
    }                                                                                                                  // 163
  }, {                                                                                                                 // 159
    fields: {                                                                                                          // 165
      _id: 1,                                                                                                          // 166
      referredBy: 1,                                                                                                   // 167
      "profile.name": 1                                                                                                // 168
    }                                                                                                                  // 165
  });                                                                                                                  // 164
  var notCompletedReferralAccountsHandle = notCompletedReferralAccountsCursor.observeChanges({                         // 171
    // The added function gets called with the id of Bob when Alice refers Bob.                                        // 172
    added: function (id, fields) {                                                                                     // 173
      _this2.added("referralInfo", id, {                                                                               // 174
        name: fields.profile.name,                                                                                     // 174
        completed: false                                                                                               // 174
      });                                                                                                              // 174
    },                                                                                                                 // 175
    // The removed function gets called when Bob is no longer an uncompleted referral.  Note that                      // 176
    // this will get more complicated once we support sending completed referrals to the client.                       // 177
    removed: function (id) {                                                                                           // 178
      _this2.removed("referralInfo", id);                                                                              // 179
    },                                                                                                                 // 180
    // The modified function gets called when Bob's profile.name changed.                                              // 181
    modified: function (id, fields) {                                                                                  // 182
      _this2.modified("referralInfo", id, {                                                                            // 183
        name: fields.profile.name,                                                                                     // 183
        completed: false                                                                                               // 183
      });                                                                                                              // 183
    }                                                                                                                  // 184
  }); // Case 2. Handle completed referrals.                                                                           // 171
  //                                                                                                                   // 188
  // - Do a query for the current list of completed accounts.                                                          // 189
  //                                                                                                                   // 190
  // - Every time we see a new such account, we create a query that watches that one account in                        // 191
  //   case its profile.name changes.                                                                                  // 192
  //                                                                                                                   // 193
  // - Also watch the first query, since the list of completed accounts might change.                                  // 194
                                                                                                                       //
  var handleForProfileNameByAccountId = {};                                                                            // 195
                                                                                                                       //
  var stopWatchingAllAccounts = function () {                                                                          // 196
    Object.keys(handleForProfileNameByAccountId).forEach(function (accountId) {                                        // 197
      stopWatchingAccount(accountId);                                                                                  // 198
    });                                                                                                                // 199
  };                                                                                                                   // 200
                                                                                                                       //
  var stopWatchingAccount = function (accountId) {                                                                     // 202
    var handleForProfileName = handleForProfileNameByAccountId[accountId];                                             // 203
                                                                                                                       //
    if (handleForProfileName) {                                                                                        // 204
      _this2.removed("referralInfo", accountId);                                                                       // 205
                                                                                                                       //
      handleForProfileName.stop(); // delete is safe because we iterate across `Object.keys()` which returns a copy.   // 206
                                                                                                                       //
      delete handleForProfileNameByAccountId[accountId];                                                               // 208
    }                                                                                                                  // 209
  };                                                                                                                   // 210
                                                                                                                       //
  var watchAccountAndPublishReferralSuccess = function (accountId) {                                                   // 212
    var handleForProfileName = handleForProfileNameByAccountId[accountId];                                             // 213
                                                                                                                       //
    if (handleForProfileName) {                                                                                        // 214
      return;                                                                                                          // 215
    }                                                                                                                  // 216
                                                                                                                       //
    handleForProfileName = Meteor.users.find({                                                                         // 218
      _id: accountId                                                                                                   // 219
    }, {                                                                                                               // 218
      fields: {                                                                                                        // 221
        "profile.name": 1                                                                                              // 222
      }                                                                                                                // 221
    }).observeChanges({                                                                                                // 220
      added: function (id, fields) {                                                                                   // 225
        _this2.added("referralInfo", id, {                                                                             // 226
          name: fields.profile.name,                                                                                   // 226
          completed: true                                                                                              // 226
        });                                                                                                            // 226
      },                                                                                                               // 227
      changed: function (id, fields) {                                                                                 // 229
        _this2.changed("referralInfo", id, {                                                                           // 230
          name: fields.profile.name,                                                                                   // 230
          completed: true                                                                                              // 230
        });                                                                                                            // 230
      },                                                                                                               // 231
      removed: function (id) {                                                                                         // 233
        stopWatchingAccount(id);                                                                                       // 234
      }                                                                                                                // 235
    });                                                                                                                // 224
    handleForProfileNameByAccountId[accountId] = handleForProfileName;                                                 // 238
  };                                                                                                                   // 239
                                                                                                                       //
  var completedAccountIdsHandle = Meteor.users.find({                                                                  // 241
    _id: this.userId,                                                                                                  // 242
    referredAccountIds: {                                                                                              // 243
      $exists: true                                                                                                    // 243
    }                                                                                                                  // 243
  }, {                                                                                                                 // 241
    fields: {                                                                                                          // 245
      referredAccountIds: true                                                                                         // 246
    }                                                                                                                  // 245
  }).observeChanges({                                                                                                  // 244
    // `added` gets called when a user gets their first completed referral.                                            // 249
    added: function (id, fields) {                                                                                     // 250
      for (var i = 0; i < fields.referredAccountIds.length; i++) {                                                     // 251
        // Unconditionally mark these as successful referrals and start watching.                                      // 252
        watchAccountAndPublishReferralSuccess(fields.referredAccountIds[i]);                                           // 253
      }                                                                                                                // 255
    },                                                                                                                 // 256
    // `changed` gets called when a user adds/removes referredAccountIds, usually when a                               // 257
    // referral becomes complete.                                                                                      // 258
    changed: function (id, fields) {                                                                                   // 259
      // Two major tasks.                                                                                              // 260
      //                                                                                                               // 261
      // 1. Look for accountIds to unsubscribe from & send removed notices to the client.                              // 262
      //                                                                                                               // 263
      // 2. Look for accountIds to subscribe to.                                                                       // 264
      // Task 1. Unsubscribe where needed.                                                                             // 266
      var referredAccountIdsAsObject = {};                                                                             // 267
      fields.referredAccountIds.forEach(function (i) {                                                                 // 268
        referredAccountIdsAsObject[i] = true;                                                                          // 268
      });                                                                                                              // 268
      Object.keys(handleForProfileNameByAccountId).forEach(function (accountId) {                                      // 270
        // If the handle doesn't show up in the new list of referredAccountIds, then remove                            // 271
        // info from the client & stop it on the server & make it null.                                                // 272
        var handleForProfileName = handleForProfileNameByAccountId[accountId];                                         // 273
                                                                                                                       //
        if (referredAccountIdsAsObject.hasOwnProperty(accountId)) {                                                    // 274
          stopWatchingAccount(accountId);                                                                              // 275
        }                                                                                                              // 276
      }); // Task 2. Subscribe where needed.                                                                           // 277
                                                                                                                       //
      for (var i = 0; i < fields.referredAccountIds.length; i++) {                                                     // 280
        // The watch... function will avoid double-creating subscriptions, so this is safe.                            // 281
        watchAccountAndPublishReferralSuccess(fields.referredAccountIds[i]);                                           // 282
      }                                                                                                                // 283
    },                                                                                                                 // 284
    // `removed` gets called when a User suddenly has no referredAccountIds.                                           // 285
    removed: function () {                                                                                             // 286
      // Remove all data from client; stop all handles.                                                                // 287
      stopWatchingAllAccounts();                                                                                       // 288
    }                                                                                                                  // 289
  }); // With cases 1 and 2 handled, register a cleanup function, then declare victory.                                // 248
                                                                                                                       //
  this.onStop(function () {                                                                                            // 293
    stopWatchingAllAccounts();                                                                                         // 294
    notCompletedReferralAccountsHandle.stop();                                                                         // 295
    completedAccountIdsHandle.stop();                                                                                  // 296
  });                                                                                                                  // 297
  this.ready();                                                                                                        // 299
});                                                                                                                    // 300
Meteor.publish("backers", function () {                                                                                // 302
  var _this3 = this;                                                                                                   // 302
                                                                                                                       //
  var backers = Assets.getText("backers.txt");                                                                         // 303
  var anonCount = 0;                                                                                                   // 304
  var counter = 0;                                                                                                     // 305
  backers.split("\n").forEach(function (name) {                                                                        // 307
    name = name.trim();                                                                                                // 308
                                                                                                                       //
    if (name === "") {                                                                                                 // 309
      ++anonCount;                                                                                                     // 310
    } else {                                                                                                           // 311
      _this3.added("backers", counter++, {                                                                             // 312
        name: name                                                                                                     // 312
      });                                                                                                              // 312
    }                                                                                                                  // 313
  }); // Text file ends in \n but that shouldn't count.                                                                // 314
                                                                                                                       //
  --anonCount;                                                                                                         // 317
  this.added("backers", "anonymous", {                                                                                 // 319
    count: anonCount - 1                                                                                               // 319
  });                                                                                                                  // 319
  this.ready();                                                                                                        // 321
});                                                                                                                    // 322
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"signup-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/signup-server.js                                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file covers creation and consumption of invite keys (i.e. to invite people to become                           // 17
// users of the Sandstorm server).                                                                                     // 18
Meteor.publish("signupKey", function (key) {                                                                           // 20
  check(key, String);                                                                                                  // 21
  return SignupKeys.find(key);                                                                                         // 22
});                                                                                                                    // 23
Meteor.methods({                                                                                                       // 25
  useSignupKey: function (key) {                                                                                       // 26
    check(key, String);                                                                                                // 27
                                                                                                                       //
    if (!this.userId) {                                                                                                // 29
      throw new Meteor.Error(403, "Must be signed in.");                                                               // 30
    }                                                                                                                  // 31
                                                                                                                       //
    var user = Meteor.user();                                                                                          // 33
                                                                                                                       //
    if (user.signupKey) {                                                                                              // 34
      // Don't waste it.                                                                                               // 35
      return;                                                                                                          // 36
    }                                                                                                                  // 37
                                                                                                                       //
    if (isDemoUser()) {                                                                                                // 39
      throw new Meteor.Error(403, "Demo users cannot accept invite keys. Please sign in as a real user.");             // 40
    }                                                                                                                  // 42
                                                                                                                       //
    if (!user.loginCredentials) {                                                                                      // 44
      // Don't consume signup key as identity user.                                                                    // 45
      return;                                                                                                          // 46
    }                                                                                                                  // 47
                                                                                                                       //
    var keyInfo = SignupKeys.findOne(key);                                                                             // 49
                                                                                                                       //
    if (!keyInfo || keyInfo.used) {                                                                                    // 50
      throw new Meteor.Error(403, "Invalid key or already used.");                                                     // 51
    }                                                                                                                  // 52
                                                                                                                       //
    if (isSignedUp() && user.payments && user.payments.id) {                                                           // 54
      // This user is already signed up with a payment account. Possibly, they signed up before                        // 55
      // using their invite, and then went back and clicked on the invite. As a result they                            // 56
      // probably now have two payment accounts. Mark this invite as used but also add a special                       // 57
      // flag so we can find it later and cancel the dupe payment account. Record who tried to                         // 58
      // use it so that we can transfer credits over if needed.                                                        // 59
      SignupKeys.update(key, {                                                                                         // 60
        $set: {                                                                                                        // 60
          used: true,                                                                                                  // 60
          rejectedBy: this.userId                                                                                      // 60
        }                                                                                                              // 60
      });                                                                                                              // 60
      return;                                                                                                          // 61
    }                                                                                                                  // 62
                                                                                                                       //
    var userFields = {                                                                                                 // 64
      signupKey: key,                                                                                                  // 65
      signupNote: keyInfo.note                                                                                         // 66
    };                                                                                                                 // 64
                                                                                                                       //
    if (keyInfo.email) {                                                                                               // 68
      userFields.signupEmail = keyInfo.email;                                                                          // 69
    }                                                                                                                  // 70
                                                                                                                       //
    if ("quota" in keyInfo) {                                                                                          // 72
      userFields.quota = keyInfo.quota;                                                                                // 73
    }                                                                                                                  // 74
                                                                                                                       //
    if (keyInfo.plan) {                                                                                                // 76
      userFields.plan = keyInfo.plan;                                                                                  // 77
    }                                                                                                                  // 78
                                                                                                                       //
    if (keyInfo.payments) {                                                                                            // 80
      userFields["payments.id"] = keyInfo.payments.id;                                                                 // 81
    }                                                                                                                  // 82
                                                                                                                       //
    Meteor.users.update(this.userId, {                                                                                 // 84
      $set: userFields                                                                                                 // 84
    });                                                                                                                // 84
    SignupKeys.update(key, {                                                                                           // 85
      $set: {                                                                                                          // 85
        used: true                                                                                                     // 85
      }                                                                                                                // 85
    });                                                                                                                // 85
  }                                                                                                                    // 86
});                                                                                                                    // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"standalone-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/standalone-server.js                                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.publish("standaloneDomain", function (domain) {                                                                 // 17
  check(domain, String);                                                                                               // 18
  return globalDb.collections.standaloneDomains.find({                                                                 // 20
    _id: domain                                                                                                        // 20
  });                                                                                                                  // 20
});                                                                                                                    // 21
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"startup.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/startup.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file is for various startup code that doesn't fit neatly anywhere else                                         // 17
var ROOT_URL = process.env.ROOT_URL;                                                                                   // 19
Meteor.startup(function () {                                                                                           // 21
  var baseUrlRow = Misc.findOne({                                                                                      // 22
    _id: "BASE_URL"                                                                                                    // 22
  });                                                                                                                  // 22
                                                                                                                       //
  if (!baseUrlRow) {                                                                                                   // 24
    // Fill data with current value in the case this is a first run                                                    // 25
    baseUrlRow = {                                                                                                     // 26
      _id: "BASE_URL",                                                                                                 // 26
      value: ROOT_URL                                                                                                  // 26
    };                                                                                                                 // 26
    Misc.insert(baseUrlRow);                                                                                           // 27
  } else if (baseUrlRow.value !== ROOT_URL) {                                                                          // 28
    // If the only thing that happened is some trailing slashes got removed, then don't reset OAuth                    // 29
    // providers.                                                                                                      // 30
    var resetOAuth = true;                                                                                             // 31
                                                                                                                       //
    if (baseUrlRow.value.replace(/[/]*$/, "") === ROOT_URL) {                                                          // 32
      console.log("Not resetting OAuth because old ROOT_URL and new ROOT_URL differ only by trailing slashes.");       // 33
      resetOAuth = false;                                                                                              // 34
    }                                                                                                                  // 35
                                                                                                                       //
    if (resetOAuth) {                                                                                                  // 37
      console.log("resetting oauth");                                                                                  // 38
      Settings.find({                                                                                                  // 39
        _id: {                                                                                                         // 39
          $in: ["google", "github"]                                                                                    // 39
        }                                                                                                              // 39
      }).forEach(function (setting) {                                                                                  // 39
        if (!!setting.value) {                                                                                         // 40
          Settings.update({                                                                                            // 41
            _id: setting._id                                                                                           // 41
          }, {                                                                                                         // 41
            $set: {                                                                                                    // 42
              value: false,                                                                                            // 42
              automaticallyReset: {                                                                                    // 43
                baseUrlChangedFrom: baseUrlRow.value                                                                   // 43
              }                                                                                                        // 43
            }                                                                                                          // 42
          });                                                                                                          // 42
        }                                                                                                              // 44
      });                                                                                                              // 45
    }                                                                                                                  // 46
                                                                                                                       //
    baseUrlRow = {                                                                                                     // 48
      _id: "BASE_URL",                                                                                                 // 48
      value: ROOT_URL                                                                                                  // 48
    };                                                                                                                 // 48
    Misc.update({                                                                                                      // 49
      _id: "BASE_URL"                                                                                                  // 49
    }, {                                                                                                               // 49
      $set: {                                                                                                          // 49
        value: ROOT_URL                                                                                                // 49
      }                                                                                                                // 49
    });                                                                                                                // 49
  }                                                                                                                    // 50
});                                                                                                                    // 51
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"stats-server.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/stats-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var allowDemo = void 0;                                                                                                // 1
module.watch(require("/imports/demo.js"), {                                                                            // 1
  allowDemo: function (v) {                                                                                            // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var DAY_MS = 24 * 60 * 60 * 1000;                                                                                      // 19
                                                                                                                       //
if (Mongo.Collection.prototype.aggregate) {                                                                            // 21
  throw new Error("Looks like Meteor wrapped the Collection.aggregate() call. Make sure it " + "works then delete our own wrapper.");
}                                                                                                                      // 24
                                                                                                                       //
Mongo.Collection.prototype.aggregate = function () {                                                                   // 26
  // Meteor doesn't wrapp Mongo's aggregate() method.                                                                  // 27
  var raw = this.rawCollection();                                                                                      // 28
  return Meteor.wrapAsync(raw.aggregate, raw).apply(raw, arguments);                                                   // 29
};                                                                                                                     // 30
                                                                                                                       //
computeStats = function (since) {                                                                                      // 32
  // We'll need this for a variety of queries.                                                                         // 33
  var timeConstraint = {                                                                                               // 34
    $gt: since                                                                                                         // 34
  }; // This calculates the number of user accounts that have been used                                                // 34
  // during the requested time period.                                                                                 // 37
                                                                                                                       //
  var currentlyActiveUsersCount = Meteor.users.find({                                                                  // 38
    expires: {                                                                                                         // 39
      $exists: false                                                                                                   // 39
    },                                                                                                                 // 39
    loginCredentials: {                                                                                                // 40
      $exists: true                                                                                                    // 40
    },                                                                                                                 // 40
    lastActive: timeConstraint                                                                                         // 41
  }).count(); // This calculates the number of grains that have been used during                                       // 38
  // the requested time period.                                                                                        // 45
                                                                                                                       //
  var activeGrainsCount = Grains.find({                                                                                // 46
    lastUsed: timeConstraint                                                                                           // 46
  }).count(); // If Meteor.settings.allowDemoAccounts is true, DeleteStats                                             // 46
  // contains records of type `user` and `appDemoUser`, indicating                                                     // 49
  // the number of those types of accounts that were created and                                                       // 50
  // then auto-expired through the demo mode's auto-account-expiry.                                                    // 51
                                                                                                                       //
  var deletedDemoUsersCount = DeleteStats.find({                                                                       // 52
    type: "demoUser",                                                                                                  // 53
    lastActive: timeConstraint                                                                                         // 53
  }).count();                                                                                                          // 53
  var deletedAppDemoUsersCount = DeleteStats.find({                                                                    // 54
    type: "appDemoUser",                                                                                               // 55
    lastActive: timeConstraint                                                                                         // 55
  }).count(); // Similarly, if the demo is enabled, we auto-delete grains; we store that                               // 55
  // fact in DeleteStats with type: "grain".                                                                           // 58
                                                                                                                       //
  var deletedGrainsCount = DeleteStats.find({                                                                          // 59
    type: "grain",                                                                                                     // 60
    lastActive: timeConstraint                                                                                         // 60
  }).count();                                                                                                          // 60
  var apps = Grains.aggregate([{                                                                                       // 62
    $match: {                                                                                                          // 63
      lastUsed: timeConstraint                                                                                         // 63
    }                                                                                                                  // 63
  }, {                                                                                                                 // 63
    $group: {                                                                                                          // 65
      _id: "$appId",                                                                                                   // 66
      grains: {                                                                                                        // 67
        $sum: 1                                                                                                        // 67
      },                                                                                                               // 67
      userIds: {                                                                                                       // 68
        $addToSet: "$userId"                                                                                           // 68
      }                                                                                                                // 68
    }                                                                                                                  // 65
  }, {                                                                                                                 // 64
    $project: {                                                                                                        // 72
      grains: 1,                                                                                                       // 73
      owners: {                                                                                                        // 74
        $size: "$userIds"                                                                                              // 74
      }                                                                                                                // 74
    }                                                                                                                  // 72
  }]);                                                                                                                 // 71
  apps = _.indexBy(apps, "_id");                                                                                       // 78
                                                                                                                       //
  for (var appId in meteorBabelHelpers.sanitizeForInObject(apps)) {                                                    // 80
    // We need to count ApiTokens, which don't have appId denormalized into them. We therefore                         // 81
    // have to fetch a list of grainIds first.                                                                         // 82
    // TODO(perf): If stats are getting slow, denormalize appId into ApiTokens. Note that it is                        // 83
    //   already denormalized for the specific case of apps that don't have icons, but the data                        // 84
    //   for that use case is NOT safe to use here because it's intended that we might allow an                        // 85
    //   app to mimic another app's icon by spoofing that app ID. In other words, the existing                         // 86
    //   denormalization of appId should be considered "app ID for identicon purposes only". We'll                     // 87
    //   need to add a new denormalization for stats purposes -- and make sure that it is not                          // 88
    //   revealed to the user.                                                                                         // 89
    var app = apps[appId];                                                                                             // 90
    delete app._id;                                                                                                    // 91
    var grains = Grains.find({                                                                                         // 92
      lastUsed: timeConstraint,                                                                                        // 93
      appId: appId                                                                                                     // 94
    }, {                                                                                                               // 92
      fields: {                                                                                                        // 96
        _id: 1                                                                                                         // 96
      }                                                                                                                // 96
    }).fetch();                                                                                                        // 95
                                                                                                                       //
    var grainIds = _.pluck(grains, "_id");                                                                             // 98
                                                                                                                       //
    var counts = ApiTokens.aggregate([{                                                                                // 100
      $match: {                                                                                                        // 102
        "owner.user": {                                                                                                // 103
          $exists: true                                                                                                // 103
        },                                                                                                             // 103
        lastUsed: timeConstraint,                                                                                      // 104
        grainId: {                                                                                                     // 105
          $in: grainIds                                                                                                // 105
        }                                                                                                              // 105
      }                                                                                                                // 102
    }, {                                                                                                               // 101
      $group: {                                                                                                        // 108
        _id: "$owner.user.accountId"                                                                                   // 108
      }                                                                                                                // 108
    }, {                                                                                                               // 108
      $group: {                                                                                                        // 110
        _id: "count",                                                                                                  // 111
        count: {                                                                                                       // 112
          $sum: 1                                                                                                      // 112
        }                                                                                                              // 112
      }                                                                                                                // 110
    }]);                                                                                                               // 109
                                                                                                                       //
    if (counts.length > 0) {                                                                                           // 117
      if (counts.length !== 1) {                                                                                       // 118
        console.error("error: sharedUsers aggregation returned multiple rows");                                        // 119
      }                                                                                                                // 120
                                                                                                                       //
      app.sharedUsers = counts[0].count;                                                                               // 122
    }                                                                                                                  // 123
  } // Count per-app appdemo users and deleted grains.                                                                 // 124
                                                                                                                       //
                                                                                                                       //
  DeleteStats.aggregate([{                                                                                             // 127
    $match: {                                                                                                          // 129
      lastActive: timeConstraint,                                                                                      // 130
      appId: {                                                                                                         // 131
        $exists: true                                                                                                  // 131
      }                                                                                                                // 131
    }                                                                                                                  // 129
  }, {                                                                                                                 // 128
    $group: {                                                                                                          // 135
      _id: {                                                                                                           // 136
        appId: "$appId",                                                                                               // 137
        type: "$type"                                                                                                  // 138
      },                                                                                                               // 136
      count: {                                                                                                         // 140
        $sum: 1                                                                                                        // 140
      }                                                                                                                // 140
    }                                                                                                                  // 135
  }]).forEach(function (deletion) {                                                                                    // 134
    var app = apps[deletion._id.appId];                                                                                // 144
                                                                                                                       //
    if (!app) {                                                                                                        // 145
      app = apps[deletion.appId] = {};                                                                                 // 146
    }                                                                                                                  // 147
                                                                                                                       //
    if (deletion._id.type === "appDemoUser") {                                                                         // 149
      app.appDemoUsers = deletion.count;                                                                               // 150
    } else if (deletion._id.type === "grain") {                                                                        // 151
      app.deleted = deletion.count;                                                                                    // 152
    } else if (deletion._id.type === "demoGrain") {                                                                    // 153
      app.demoed = deletion.count;                                                                                     // 154
    }                                                                                                                  // 155
  });                                                                                                                  // 156
  return {                                                                                                             // 158
    activeUsers: currentlyActiveUsersCount,                                                                            // 159
    demoUsers: deletedDemoUsersCount,                                                                                  // 160
    appDemoUsers: deletedAppDemoUsersCount,                                                                            // 161
    activeGrains: activeGrainsCount + deletedGrainsCount,                                                              // 162
    apps: apps                                                                                                         // 163
  };                                                                                                                   // 158
};                                                                                                                     // 165
                                                                                                                       //
function recordStats() {                                                                                               // 167
  var postStats = function (record) {                                                                                  // 168
    HTTP.post("https://alpha-api.sandstorm.io/data", {                                                                 // 169
      data: record,                                                                                                    // 170
      headers: {                                                                                                       // 171
        Authorization: "Bearer aT-mGyNwsgwZBbZvd5FWr0Ma79O9IehI4NiEO94y_oR",                                           // 172
        "Content-Type": "application/json"                                                                             // 173
      }                                                                                                                // 171
    });                                                                                                                // 169
  };                                                                                                                   // 176
                                                                                                                       //
  var now = new Date();                                                                                                // 178
                                                                                                                       //
  var planStats = _.countBy(Meteor.users.find({                                                                        // 180
    expires: {                                                                                                         // 181
      $exists: false                                                                                                   // 181
    },                                                                                                                 // 181
    "payments.id": {                                                                                                   // 181
      $exists: true                                                                                                    // 181
    }                                                                                                                  // 181
  }, {                                                                                                                 // 181
    fields: {                                                                                                          // 182
      plan: 1                                                                                                          // 182
    }                                                                                                                  // 182
  }).fetch(), "plan");                                                                                                 // 182
                                                                                                                       //
  var record = {                                                                                                       // 186
    timestamp: now,                                                                                                    // 187
    daily: computeStats(new Date(now.getTime() - DAY_MS)),                                                             // 188
    weekly: computeStats(new Date(now.getTime() - 7 * DAY_MS)),                                                        // 189
    monthly: computeStats(new Date(now.getTime() - 30 * DAY_MS)),                                                      // 190
    forever: computeStats(new Date(0)),                                                                                // 191
    plans: planStats                                                                                                   // 192
  };                                                                                                                   // 186
  record.computeTime = Date.now() - now;                                                                               // 194
                                                                                                                       //
  if (global.BlackrockPayments && global.BlackrockPayments.getTotalCharges) {                                          // 195
    // This only exists under Blackrock                                                                                // 196
    record.totalCharges = global.BlackrockPayments.getTotalCharges();                                                  // 197
  }                                                                                                                    // 198
                                                                                                                       //
  ActivityStats.insert(record);                                                                                        // 200
  var age = ActivityStats.find().count(); // The stats page which the user agreed we can send actually displays the whole history
  // of the server, but we're only sending stats from the last day. Let's also throw in the                            // 203
  // length of said history. This is still strictly less information than what the user said                           // 204
  // we're allowed to send.                                                                                            // 205
                                                                                                                       //
  record.serverAge = age;                                                                                              // 206
                                                                                                                       //
  if (age > 3) {                                                                                                       // 208
    var reportSetting = Settings.findOne({                                                                             // 209
      _id: "reportStats"                                                                                               // 209
    });                                                                                                                // 209
                                                                                                                       //
    if (!reportSetting) {                                                                                              // 211
      // Setting not set yet, send out notifications and set it to false                                               // 212
      globalDb.sendAdminNotification("reportStats", "/admin/stats");                                                   // 213
      Settings.insert({                                                                                                // 214
        _id: "reportStats",                                                                                            // 214
        value: "unset"                                                                                                 // 214
      });                                                                                                              // 214
    } else if (reportSetting.value === true) {                                                                         // 215
      postStats(record);                                                                                               // 216
    }                                                                                                                  // 217
  }                                                                                                                    // 218
}                                                                                                                      // 219
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 221
  // Wait until 10:00 UTC (2:00 PST / 5:00 EST), then start recording stats every 24 hours.                            // 222
  // (Only on the first replica to avoid conflicts.)                                                                   // 223
  Meteor.setTimeout(function () {                                                                                      // 224
    Meteor.setInterval(function () {                                                                                   // 225
      recordStats();                                                                                                   // 226
    }, DAY_MS);                                                                                                        // 227
    recordStats();                                                                                                     // 229
  }, DAY_MS - (Date.now() - 10 * 60 * 60 * 1000) % DAY_MS);                                                            // 230
  Meteor.startup(function () {                                                                                         // 232
    if (StatsTokens.find().count() === 0) {                                                                            // 233
      StatsTokens.remove({});                                                                                          // 234
      StatsTokens.insert({                                                                                             // 235
        _id: Random.id(22)                                                                                             // 235
      });                                                                                                              // 235
    }                                                                                                                  // 236
  });                                                                                                                  // 237
}                                                                                                                      // 238
                                                                                                                       //
Meteor.methods({                                                                                                       // 240
  regenerateStatsToken: function () {                                                                                  // 241
    if (!isAdmin()) {                                                                                                  // 242
      throw new Meteor.Error(403, "Unauthorized", "User must be admin");                                               // 243
    }                                                                                                                  // 244
                                                                                                                       //
    StatsTokens.remove({});                                                                                            // 246
    var token = StatsTokens.insert({                                                                                   // 247
      _id: Random.id(22)                                                                                               // 247
    });                                                                                                                // 247
    return token._id;                                                                                                  // 248
  }                                                                                                                    // 249
});                                                                                                                    // 240
Router.map(function () {                                                                                               // 252
  this.route("fetchStats", {                                                                                           // 253
    where: "server",                                                                                                   // 254
    path: "/fetchStats/:tokenId",                                                                                      // 255
    action: function () {                                                                                              // 256
      var token = StatsTokens.findOne({                                                                                // 257
        _id: this.params.tokenId                                                                                       // 257
      });                                                                                                              // 257
                                                                                                                       //
      if (!token) {                                                                                                    // 259
        this.response.writeHead(404, {                                                                                 // 260
          "Content-Type": "text/plain"                                                                                 // 261
        });                                                                                                            // 260
        this.response.write("Token not found");                                                                        // 263
        return this.response.end();                                                                                    // 264
      }                                                                                                                // 265
                                                                                                                       //
      try {                                                                                                            // 267
        var stats = ActivityStats.find().fetch();                                                                      // 268
        var statsString = JSON.stringify(stats);                                                                       // 269
        this.response.writeHead(200, {                                                                                 // 271
          "Content-Type": "application/json"                                                                           // 272
        });                                                                                                            // 271
        this.response.write(statsString);                                                                              // 274
      } catch (error) {                                                                                                // 275
        console.error(error.stack);                                                                                    // 276
        this.response.writeHead(500, {                                                                                 // 277
          "Content-Type": "text/plain"                                                                                 // 278
        });                                                                                                            // 277
        this.response.write(error.stack);                                                                              // 280
      }                                                                                                                // 281
                                                                                                                       //
      return this.response.end();                                                                                      // 283
    }                                                                                                                  // 284
  });                                                                                                                  // 253
});                                                                                                                    // 286
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"shared":{"admin.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/admin.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// TODO(someday): split out fields by locus                                                                            // 17
// isEnabled is available on both client and server                                                                    // 18
// getLoginId is only used on server                                                                                   // 19
// initiateLogin and loginTemplate are only used on client                                                             // 20
function serviceEnabled(name) {                                                                                        // 22
  var setting = Settings.findOne({                                                                                     // 23
    _id: name                                                                                                          // 23
  });                                                                                                                  // 23
  return setting && !!setting.value;                                                                                   // 24
}                                                                                                                      // 25
                                                                                                                       //
Accounts.loginServices.github = {                                                                                      // 27
  isEnabled: function () {                                                                                             // 28
    return serviceEnabled("github");                                                                                   // 29
  },                                                                                                                   // 30
  getLoginId: function (identity) {                                                                                    // 32
    return identity.services.github.username;                                                                          // 33
  },                                                                                                                   // 34
  initiateLogin: function (loginId) {                                                                                  // 36
    Meteor.loginWithGithub();                                                                                          // 37
    return {                                                                                                           // 38
      oneClick: true                                                                                                   // 38
    };                                                                                                                 // 38
  },                                                                                                                   // 39
  loginTemplate: {                                                                                                     // 41
    name: "oauthLoginButton",                                                                                          // 42
    priority: 1,                                                                                                       // 43
    data: {                                                                                                            // 44
      method: "loginWithGithub",                                                                                       // 45
      name: "github",                                                                                                  // 46
      displayName: "GitHub",                                                                                           // 47
      linkingNewCredential: false                                                                                      // 48
    }                                                                                                                  // 44
  }                                                                                                                    // 41
};                                                                                                                     // 27
Accounts.loginServices.google = {                                                                                      // 53
  isEnabled: function () {                                                                                             // 54
    return serviceEnabled("google");                                                                                   // 55
  },                                                                                                                   // 56
  getLoginId: function (identity) {                                                                                    // 58
    return identity.services.google.email;                                                                             // 59
  },                                                                                                                   // 60
  initiateLogin: function (loginId) {                                                                                  // 62
    Meteor.loginWithGoogle({                                                                                           // 63
      loginHint: loginId                                                                                               // 63
    });                                                                                                                // 63
    return {                                                                                                           // 64
      oneClick: true                                                                                                   // 64
    };                                                                                                                 // 64
  },                                                                                                                   // 65
  loginTemplate: {                                                                                                     // 67
    name: "oauthLoginButton",                                                                                          // 68
    priority: 2,                                                                                                       // 69
    data: {                                                                                                            // 70
      method: "loginWithGoogle",                                                                                       // 71
      name: "google",                                                                                                  // 72
      displayName: "Google",                                                                                           // 73
      linkingNewCredential: false                                                                                      // 74
    }                                                                                                                  // 70
  }                                                                                                                    // 67
};                                                                                                                     // 53
Accounts.loginServices.email = {                                                                                       // 79
  isEnabled: function () {                                                                                             // 80
    return serviceEnabled("emailToken");                                                                               // 81
  },                                                                                                                   // 82
  getLoginId: function (identity) {                                                                                    // 84
    return identity.services.email.email;                                                                              // 85
  },                                                                                                                   // 86
  initiateLogin: function (loginId) {                                                                                  // 88
    return {                                                                                                           // 89
      form: true                                                                                                       // 89
    };                                                                                                                 // 89
  },                                                                                                                   // 90
  loginTemplate: {                                                                                                     // 92
    name: "emailLoginForm",                                                                                            // 93
    priority: 10                                                                                                       // 94
  }                                                                                                                    // 92
};                                                                                                                     // 79
Accounts.loginServices.ldap = {                                                                                        // 98
  isEnabled: function () {                                                                                             // 99
    return serviceEnabled("ldap");                                                                                     // 100
  },                                                                                                                   // 101
  getLoginId: function (identity) {                                                                                    // 103
    return identity.services.ldap.username;                                                                            // 104
  },                                                                                                                   // 105
  initiateLogin: function (loginId) {                                                                                  // 107
    return {                                                                                                           // 108
      form: true                                                                                                       // 108
    };                                                                                                                 // 108
  },                                                                                                                   // 109
  loginTemplate: {                                                                                                     // 111
    name: "ldapLoginForm",                                                                                             // 112
    priority: 21 // Put it at the bottom of the list.                                                                  // 113
                                                                                                                       //
  }                                                                                                                    // 111
};                                                                                                                     // 98
Accounts.loginServices.saml = {                                                                                        // 117
  isEnabled: function () {                                                                                             // 118
    return serviceEnabled("saml");                                                                                     // 119
  },                                                                                                                   // 120
  getLoginId: function (identity) {                                                                                    // 122
    return identity.services.saml.id;                                                                                  // 123
  },                                                                                                                   // 124
  initiateLogin: function (loginId) {                                                                                  // 126
    Meteor.loginWithSaml();                                                                                            // 127
    return {                                                                                                           // 128
      oneClick: true                                                                                                   // 128
    };                                                                                                                 // 128
  },                                                                                                                   // 129
  loginTemplate: {                                                                                                     // 131
    name: "samlLoginForm",                                                                                             // 132
    priority: 20 // Put it at the bottom of the list.                                                                  // 133
                                                                                                                       //
  }                                                                                                                    // 131
};                                                                                                                     // 117
Accounts.loginServices.demo = {                                                                                        // 137
  isEnabled: function () {                                                                                             // 138
    return false; // You can never authenticate as a demo user.                                                        // 139
  },                                                                                                                   // 140
  getLoginId: function (identity) {                                                                                    // 142
    return identity._id;                                                                                               // 143
  },                                                                                                                   // 144
  initiateLogin: function (loginId) {                                                                                  // 146
    return {                                                                                                           // 147
      demoCannotLogin: true                                                                                            // 147
    };                                                                                                                 // 147
  },                                                                                                                   // 148
  loginTemplate: {}                                                                                                    // 150
};                                                                                                                     // 137
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"dev-accounts.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/dev-accounts.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Accounts.loginServices.dev = {                                                                                         // 17
  isEnabled: function () {                                                                                             // 18
    return globalDb.allowDevAccounts();                                                                                // 19
  },                                                                                                                   // 20
  getLoginId: function (identity) {                                                                                    // 22
    return identity.services.dev.name;                                                                                 // 23
  },                                                                                                                   // 24
  initiateLogin: function (loginId) {                                                                                  // 26
    loginDevAccount(loginId);                                                                                          // 27
    return {                                                                                                           // 28
      oneClick: true                                                                                                   // 28
    };                                                                                                                 // 28
  },                                                                                                                   // 29
  loginTemplate: {                                                                                                     // 31
    name: "devLoginForm",                                                                                              // 32
    priority: -10 // Put it at the top.                                                                                // 33
                                                                                                                       //
  }                                                                                                                    // 31
};                                                                                                                     // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"grain-shared.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/grain-shared.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.methods({                                                                                                       // 17
  // Methods defined in this file have meaningful latency compensation (client-side prediction)                        // 18
  // potential.                                                                                                        // 19
  //                                                                                                                   // 20
  // Methods for which latency compensation makes no sense are defined in grain-server.js.                             // 21
  markActivityReadByOwner: function (grainId) {                                                                        // 23
    check(grainId, String);                                                                                            // 24
    check(this.userId, String);                                                                                        // 25
    Grains.update({                                                                                                    // 27
      _id: grainId,                                                                                                    // 27
      userId: this.userId                                                                                              // 27
    }, {                                                                                                               // 27
      $set: {                                                                                                          // 28
        "ownerSeenAllActivity": true                                                                                   // 28
      }                                                                                                                // 28
    });                                                                                                                // 28
  },                                                                                                                   // 29
  markActivityRead: function (grainId, obsolete) {                                                                     // 31
    check(grainId, String);                                                                                            // 32
                                                                                                                       //
    if (!this.userId) {                                                                                                // 34
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 35
    }                                                                                                                  // 36
                                                                                                                       //
    ApiTokens.update({                                                                                                 // 38
      "grainId": grainId,                                                                                              // 38
      "owner.user.accountId": Meteor.userId()                                                                          // 38
    }, {                                                                                                               // 38
      $set: {                                                                                                          // 39
        "owner.user.seenAllActivity": true                                                                             // 39
      }                                                                                                                // 39
    }, {                                                                                                               // 39
      multi: true                                                                                                      // 39
    });                                                                                                                // 39
  },                                                                                                                   // 40
  moveGrainsToTrash: function (grainIds) {                                                                             // 42
    var _this = this;                                                                                                  // 42
                                                                                                                       //
    check(grainIds, [String]);                                                                                         // 43
                                                                                                                       //
    if (this.userId) {                                                                                                 // 45
      Grains.update({                                                                                                  // 46
        userId: {                                                                                                      // 46
          $eq: this.userId                                                                                             // 46
        },                                                                                                             // 46
        _id: {                                                                                                         // 47
          $in: grainIds                                                                                                // 47
        },                                                                                                             // 47
        trashed: {                                                                                                     // 48
          $exists: false                                                                                               // 48
        }                                                                                                              // 48
      }, {                                                                                                             // 46
        $set: {                                                                                                        // 49
          trashed: new Date()                                                                                          // 49
        }                                                                                                              // 49
      }, {                                                                                                             // 49
        multi: true                                                                                                    // 50
      });                                                                                                              // 50
      ApiTokens.update({                                                                                               // 52
        grainId: {                                                                                                     // 52
          $in: grainIds                                                                                                // 52
        },                                                                                                             // 52
        "owner.user.accountId": Meteor.userId(),                                                                       // 53
        trashed: {                                                                                                     // 54
          $exists: false                                                                                               // 54
        }                                                                                                              // 54
      }, {                                                                                                             // 52
        $set: {                                                                                                        // 55
          "trashed": new Date()                                                                                        // 55
        }                                                                                                              // 55
      }, {                                                                                                             // 55
        multi: true                                                                                                    // 56
      });                                                                                                              // 56
                                                                                                                       //
      if (!this.isSimulation) {                                                                                        // 58
        var grainsOwned = Grains.find({                                                                                // 59
          userId: {                                                                                                    // 60
            $eq: this.userId                                                                                           // 60
          },                                                                                                           // 60
          _id: {                                                                                                       // 61
            $in: grainIds                                                                                              // 61
          }                                                                                                            // 61
        }, {                                                                                                           // 59
          fields: {                                                                                                    // 62
            _id: 1                                                                                                     // 62
          }                                                                                                            // 62
        });                                                                                                            // 62
        grainsOwned.forEach(function (grain) {                                                                         // 64
          Sessions.remove({                                                                                            // 65
            grainId: grain._id                                                                                         // 65
          });                                                                                                          // 65
                                                                                                                       //
          try {                                                                                                        // 66
            _this.connection.sandstormBackend.shutdownGrain(grain._id, _this.userId).await();                          // 67
          } catch (err) {                                                                                              // 68
            console.error("Failed to shutdown trashed grain", grain._id, err);                                         // 69
          }                                                                                                            // 70
        });                                                                                                            // 71
      }                                                                                                                // 72
    }                                                                                                                  // 73
  },                                                                                                                   // 74
  moveGrainsOutOfTrash: function (grainIds) {                                                                          // 76
    check(grainIds, [String]);                                                                                         // 77
                                                                                                                       //
    if (this.userId) {                                                                                                 // 79
      Grains.update({                                                                                                  // 80
        userId: {                                                                                                      // 80
          $eq: this.userId                                                                                             // 80
        },                                                                                                             // 80
        _id: {                                                                                                         // 81
          $in: grainIds                                                                                                // 81
        },                                                                                                             // 81
        trashed: {                                                                                                     // 82
          $exists: true                                                                                                // 82
        }                                                                                                              // 82
      }, {                                                                                                             // 80
        $unset: {                                                                                                      // 83
          trashed: 1                                                                                                   // 83
        }                                                                                                              // 83
      }, {                                                                                                             // 83
        multi: true                                                                                                    // 84
      });                                                                                                              // 84
      ApiTokens.update({                                                                                               // 86
        grainId: {                                                                                                     // 86
          $in: grainIds                                                                                                // 86
        },                                                                                                             // 86
        "owner.user.accountId": Meteor.userId(),                                                                       // 87
        "trashed": {                                                                                                   // 88
          $exists: true                                                                                                // 88
        }                                                                                                              // 88
      }, {                                                                                                             // 86
        $unset: {                                                                                                      // 89
          "trashed": 1                                                                                                 // 89
        }                                                                                                              // 89
      }, {                                                                                                             // 89
        multi: true                                                                                                    // 90
      });                                                                                                              // 90
    }                                                                                                                  // 91
  },                                                                                                                   // 92
  deleteGrain: function (grainId) {                                                                                    // 94
    check(grainId, String);                                                                                            // 95
                                                                                                                       //
    if (this.userId) {                                                                                                 // 97
      var grainsQuery = {                                                                                              // 98
        _id: grainId,                                                                                                  // 99
        userId: this.userId,                                                                                           // 100
        trashed: {                                                                                                     // 101
          $exists: true                                                                                                // 101
        }                                                                                                              // 101
      };                                                                                                               // 98
      var numDeleted = 0;                                                                                              // 104
                                                                                                                       //
      if (this.isSimulation) {                                                                                         // 105
        numDeleted = Grains.remove(grainsQuery);                                                                       // 106
      } else {                                                                                                         // 107
        numDeleted = globalDb.deleteGrains(grainsQuery, globalBackend, isDemoUser() ? "demoGrain" : "grain");          // 108
      } // Usually we don't automatically remove user-owned tokens that have become invalid,                           // 110
      // because if we did their owner might become confused as to why they have mysteriously                          // 113
      // disappeared. In this particular case, however, for tokens held by the grain owner,                            // 114
      // there should be no confusion. Indeed, it would be more confusing *not* to remove these                        // 115
      // tokens, because then the grain could still show up in the trash bin as a "shared with me"                     // 116
      // grain after the owner clicks "delete permanently".                                                            // 117
                                                                                                                       //
                                                                                                                       //
      var apiTokensQuery = {                                                                                           // 118
        grainId: grainId,                                                                                              // 119
        "owner.user.accountId": Meteor.userId(),                                                                       // 120
        "trashed": {                                                                                                   // 121
          $exists: true                                                                                                // 121
        }                                                                                                              // 121
      };                                                                                                               // 118
                                                                                                                       //
      if (numDeleted > 0) {                                                                                            // 124
        if (this.isSimulation) {                                                                                       // 125
          ApiTokens.remove(apiTokensQuery);                                                                            // 126
        } else {                                                                                                       // 127
          globalDb.removeApiTokens(apiTokensQuery);                                                                    // 128
        }                                                                                                              // 129
      }                                                                                                                // 130
    }                                                                                                                  // 131
  },                                                                                                                   // 132
  forgetGrain: function (grainId, obsolete) {                                                                          // 134
    check(grainId, String);                                                                                            // 135
                                                                                                                       //
    if (!this.userId) {                                                                                                // 137
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");                                             // 138
    }                                                                                                                  // 139
                                                                                                                       //
    var query = {                                                                                                      // 141
      grainId: grainId,                                                                                                // 142
      "owner.user.accountId": this.userId,                                                                             // 143
      "trashed": {                                                                                                     // 144
        $exists: true                                                                                                  // 144
      }                                                                                                                // 144
    };                                                                                                                 // 141
                                                                                                                       //
    if (this.isSimulation) {                                                                                           // 147
      ApiTokens.remove(query);                                                                                         // 148
    } else {                                                                                                           // 149
      globalDb.removeApiTokens(query);                                                                                 // 150
    }                                                                                                                  // 151
  }                                                                                                                    // 152
});                                                                                                                    // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"install.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/install.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
isSafeDemoAppUrl = function () {                                                                                       // 17
    function isSafeDemoAppUrl(url) {                                                                                   // 17
        // For demo accounts, we allow using a bare hash with no URL (which will never upload a new app)               // 18
        // and we allow specifying a sandstorm.io URL.                                                                 // 19
        return !url || url.lastIndexOf("http://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://alpha-j7uny7u376jnimcsx34c.sandstorm.io/", 0) === 0 || url.lastIndexOf("https://app-index.sandstorm.io/", 0) === 0;
    }                                                                                                                  // 25
                                                                                                                       //
    return isSafeDemoAppUrl;                                                                                           // 17
}();                                                                                                                   // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"testing.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/testing.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var isTesting = Meteor.settings && Meteor.settings.public && Meteor.settings.public.isTesting; // jscs:disable requireCamelCaseOrUpperCaseIdentifiers
                                                                                                                       //
if (isTesting) {                                                                                                       // 21
  if (Meteor.isServer) {                                                                                               // 22
    var clearUser = function (id) {                                                                                    // 22
      UserActions.remove({                                                                                             // 24
        userId: id                                                                                                     // 24
      });                                                                                                              // 24
      globalDb.removeApiTokens({                                                                                       // 25
        userId: id                                                                                                     // 25
      });                                                                                                              // 25
      Grains.find({                                                                                                    // 26
        userId: id                                                                                                     // 26
      }).forEach(function (grain) {                                                                                    // 26
        globalBackend.deleteGrain(grain._id);                                                                          // 27
      });                                                                                                              // 28
      Grains.remove({                                                                                                  // 30
        userId: id                                                                                                     // 30
      });                                                                                                              // 30
      Meteor.users.remove({                                                                                            // 31
        _id: id                                                                                                        // 31
      });                                                                                                              // 31
    };                                                                                                                 // 32
                                                                                                                       //
    Meteor.methods({                                                                                                   // 34
      createMockGithubUser: function () {                                                                              // 35
        Meteor.users.update({                                                                                          // 36
          _id: "Py8fwsaryQNGBuiXb"                                                                                     // 36
        }, {                                                                                                           // 36
          $set: {                                                                                                      // 37
            createdAt: new Date("2014-08-11T21:44:04.147Z"),                                                           // 37
            isAdmin: true,                                                                                             // 37
            lastActive: new Date("2014-08-19T09:58:39.676Z"),                                                          // 37
            profile: {                                                                                                 // 37
              name: "Github User"                                                                                      // 37
            },                                                                                                         // 37
            services: {                                                                                                // 37
              github: {                                                                                                // 37
                accessToken: "sometoken",                                                                              // 37
                id: 1595880,                                                                                           // 37
                username: "testuser"                                                                                   // 37
              },                                                                                                       // 37
              resume: {                                                                                                // 37
                loginTokens: [{                                                                                        // 37
                  when: new Date("2099-08-13T05:16:02.356Z"),                                                          // 37
                  hashedToken: "GriUSDp+uN/K4HptwSl1wsdWfHEpS8c9KjjdqwKNo0k="                                          // 37
                }]                                                                                                     // 37
              }                                                                                                        // 37
            },                                                                                                         // 37
            signupKey: "admin"                                                                                         // 37
          }                                                                                                            // 37
        }, {                                                                                                           // 37
          upsert: true                                                                                                 // 38
        });                                                                                                            // 38
      },                                                                                                               // 39
      clearMockGithubUser: function () {                                                                               // 41
        clearUser("Py8fwsaryQNGBuiXb");                                                                                // 42
      },                                                                                                               // 43
      createMockGoogleUser: function () {                                                                              // 45
        Meteor.users.update({                                                                                          // 46
          _id: "6WJcRo2gg2Ysuxsok"                                                                                     // 46
        }, {                                                                                                           // 46
          $set: {                                                                                                      // 47
            createdAt: new Date("2014-08-21T07:52:55.581Z"),                                                           // 47
            profile: {                                                                                                 // 47
              name: "Google User"                                                                                      // 47
            },                                                                                                         // 47
            services: {                                                                                                // 47
              google: {                                                                                                // 47
                accessToken: "sometoken",                                                                              // 47
                expiresAt: 4562182723000,                                                                              // 47
                id: "116893057283177439912",                                                                           // 47
                verified_email: true,                                                                                  // 47
                name: "Google User",                                                                                   // 47
                given_name: "Google",                                                                                  // 47
                family_name: "User",                                                                                   // 47
                picture: "https://lh3.googleusercontent.com/-XdUIqdMkCWA/AAAAAAAAAAI/AAAAAAAAAAA/4252rscbv5M/photo.jpg",
                locale: "en",                                                                                          // 47
                gender: "male"                                                                                         // 47
              },                                                                                                       // 47
              resume: {                                                                                                // 47
                loginTokens: [{                                                                                        // 47
                  when: new Date("2099-08-21T07:52:55.592Z"),                                                          // 47
                  hashedToken: "cbJGxLGKW3f0j7Ehit77hdK58W7xuPjzZhGHgKhyddo="                                          // 47
                }]                                                                                                     // 47
              }                                                                                                        // 47
            },                                                                                                         // 47
            signupKey: "admin"                                                                                         // 47
          }                                                                                                            // 47
        }, {                                                                                                           // 47
          upsert: true                                                                                                 // 48
        });                                                                                                            // 48
      },                                                                                                               // 49
      clearMockGoogleUser: function () {                                                                               // 51
        clearUser("6WJcRo2gg2Ysuxsok");                                                                                // 52
      },                                                                                                               // 53
      fetchAppIndexTest: function () {                                                                                 // 55
        globalDb.collections.appIndex.remove({});                                                                      // 56
        SandstormAutoupdateApps.updateAppIndex(this.connection.sandstormDb);                                           // 57
      }                                                                                                                // 58
    });                                                                                                                // 34
  }                                                                                                                    // 60
                                                                                                                       //
  mockLoginGithub = function () {                                                                                      // 62
    Meteor.call("createMockGithubUser", function (err) {                                                               // 63
      if (err) {                                                                                                       // 64
        console.log(err);                                                                                              // 65
      }                                                                                                                // 66
                                                                                                                       //
      window.localStorage.setItem("Meteor.loginToken", "F0S6luPxIJV--y_GAkBKOMuWMYnnqgG3UMv9M-DIs2f");                 // 68
      window.localStorage.setItem("Meteor.loginTokenExpires", "Mon Nov 10 2099 21:16:02 GMT-0800 (PST)");              // 69
      window.localStorage.setItem("Meteor.userId", "Py8fwsaryQNGBuiXb");                                               // 70
      window.location.reload();                                                                                        // 71
    });                                                                                                                // 72
  };                                                                                                                   // 73
                                                                                                                       //
  clearMockGithubUser = function () {                                                                                  // 75
    Meteor.call("clearMockGithubUser", function (err) {                                                                // 76
      if (err) {                                                                                                       // 77
        console.log(err);                                                                                              // 78
      }                                                                                                                // 79
    });                                                                                                                // 80
  };                                                                                                                   // 81
                                                                                                                       //
  mockLoginGoogle = function () {                                                                                      // 83
    Meteor.call("createMockGoogleUser", function (err) {                                                               // 84
      if (err) {                                                                                                       // 85
        console.log(err);                                                                                              // 86
      }                                                                                                                // 87
                                                                                                                       //
      window.localStorage.setItem("Meteor.loginToken", "P3ffUfVJtptyVX2IPUfNDZY0F3b-GIZ-WQf7w3GdL21");                 // 89
      window.localStorage.setItem("Meteor.loginTokenExpires", "Tue Nov 18 2099 23:52:55 GMT-0800 (PST)");              // 90
      window.localStorage.setItem("Meteor.userId", "6WJcRo2gg2Ysuxsok");                                               // 91
      window.location.reload();                                                                                        // 92
    });                                                                                                                // 93
  };                                                                                                                   // 94
                                                                                                                       //
  clearMockGoogleUser = function () {                                                                                  // 96
    Meteor.call("clearMockGoogleUser", function (err) {                                                                // 97
      if (err) {                                                                                                       // 98
        console.log(err);                                                                                              // 99
      }                                                                                                                // 100
    });                                                                                                                // 101
  };                                                                                                                   // 102
}                                                                                                                      // 103
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}},{
  "extensions": [
    ".js",
    ".json"
  ]
});
require("./server/accounts/credentials/credentials-server.js");
require("./server/accounts/email-token/token-server.js");
require("./server/accounts/ldap/ldap-server.js");
require("./server/accounts/saml/saml-server.js");
require("./server/accounts/accounts-server.js");
require("./server/accounts/accounts-ui-methods.js");
require("./server/accounts/accounts-ui-server.js");
require("./server/admin/admin-alert-server.js");
require("./server/admin/admin-user-invite.js");
require("./server/admin/network-capabilities-server.js");
require("./server/admin/personalization-server.js");
require("./server/admin/system-status-server.js");
require("./server/drivers/external-ui-view.js");
require("./server/drivers/ip.js");
require("./server/drivers/mail.js");
require("./server/00-startup.js");
require("./server/account-suspension.js");
require("./server/admin-server.js");
require("./server/backup.js");
require("./server/contacts-server.js");
require("./server/core.js");
require("./server/demo-server.js");
require("./server/desktop-notifications-server.js");
require("./server/dev-accounts-server.js");
require("./server/grain-server.js");
require("./server/hack-session.js");
require("./server/identity.js");
require("./server/install-server.js");
require("./server/installer.js");
require("./server/notifications-server.js");
require("./server/pre-meteor.js");
require("./server/proxy.js");
require("./server/sandcats.js");
require("./server/shell-server.js");
require("./server/signup-server.js");
require("./server/standalone-server.js");
require("./server/startup.js");
require("./server/stats-server.js");
require("./shared/admin.js");
require("./shared/dev-accounts.js");
require("./shared/grain-shared.js");
require("./shared/install.js");
require("./shared/testing.js");
//# sourceMappingURL=app.js.map
